#!/usr/bin/env python
#
# Copyright (c) 2020-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import ctypes
from ctypes import *

######################################################
#    defines
######################################################
WJH_SOCKET_BUFFER_SIZE = 1024
WJH_MONITOR_SERVER_PORT = 20000
WJH_MAX_PKT_SIZE = 10000
WJH_PCAP_NAME = "/tmp/wjh_raw.pcap"
WJH_TXT_NAME = "/tmp/wjh_agg.txt"
WJH_SERVER_LISTEN_QUEUE_SIZE = 1

WJH_USER_CHANNEL_TAILDROP_E = 0
WJH_USER_CHANNEL_CYCLIC_E = 1
WJH_USER_CHANNEL_AGGREGATE_E = 2
WJH_USER_CHANNEL_CYCLIC_AND_AGGREGATE_E = 3

WJH_DROP_REASON_GROUP_BUFFER_E = 0
WJH_DROP_REASON_GROUP_ACL_E = 1
WJH_DROP_REASON_GROUP_L1_E = 2
WJH_DROP_REASON_GROUP_L2_E = 3
WJH_DROP_REASON_GROUP_ROUTER_E = 4
WJH_DROP_REASON_GROUP_TUNNEL_E = 5
WJH_DROP_REASON_GROUP_ALL_E = 6

WJH_USER_CHANNEL_MODE_PUSH_E = 0
WJH_USER_CHANNEL_MODE_PULL_E = 1

WJH_SEVERITY_ALL_E = 0
WJH_SEVERITY_NOTICE_E = 1
WJH_SEVERITY_WARNING_E = 2
WJH_SEVERITY_ERROR_E = 3

WJH_ACL_BINDING_POINT_INGRESS = 0
WJH_ACL_BINDING_POINT_EGRESS = 1
WJH_ACL_BINDING_POINT_RIF_INGRESS = 2
WJH_ACL_BINDING_POINT_RIF_EGRESS = 3

RECYCLE_PORT = 0x10100

channel_type_to_str = {
    WJH_USER_CHANNEL_TAILDROP_E: 'Taildrop',
    WJH_USER_CHANNEL_CYCLIC_E: 'Cyclic',
    WJH_USER_CHANNEL_AGGREGATE_E: 'Aggregate',
    WJH_USER_CHANNEL_CYCLIC_AND_AGGREGATE_E: 'Cyclic-and-Aggregate'
}

drop_reason_group_to_str = {
    WJH_DROP_REASON_GROUP_BUFFER_E: "buffer",
    WJH_DROP_REASON_GROUP_ACL_E: "acl",
    WJH_DROP_REASON_GROUP_L1_E: "l1",
    WJH_DROP_REASON_GROUP_L2_E: "l2",
    WJH_DROP_REASON_GROUP_ROUTER_E: "router",
    WJH_DROP_REASON_GROUP_TUNNEL_E: "tunnel",
    WJH_DROP_REASON_GROUP_ALL_E: "all"
}

drop_reason_group_str_to_enum = {
    "buffer": WJH_DROP_REASON_GROUP_BUFFER_E,
    "acl": WJH_DROP_REASON_GROUP_ACL_E,
    "l1": WJH_DROP_REASON_GROUP_L1_E,
    "l2": WJH_DROP_REASON_GROUP_L2_E,
    "router": WJH_DROP_REASON_GROUP_ROUTER_E,
    "tunnel": WJH_DROP_REASON_GROUP_TUNNEL_E,
    "all": WJH_DROP_REASON_GROUP_ALL_E
}

channel_mode_str_to_enum = {
    "agent": WJH_USER_CHANNEL_MODE_PUSH_E,
    "manual": WJH_USER_CHANNEL_MODE_PULL_E
}

severity_to_str = {
    WJH_SEVERITY_NOTICE_E: 'notice',
    WJH_SEVERITY_WARNING_E: 'warning',
    WJH_SEVERITY_ERROR_E: 'error',
    WJH_SEVERITY_ALL_E: 'all'
}

severity_str_to_enum = {
    "notice": WJH_SEVERITY_NOTICE_E,
    "warning": WJH_SEVERITY_WARNING_E,
    "error": WJH_SEVERITY_ERROR_E,
    "all": WJH_SEVERITY_ALL_E,
}

binding_point_to_str = {
    WJH_ACL_BINDING_POINT_INGRESS: 'Ingress port ACL',
    WJH_ACL_BINDING_POINT_EGRESS: 'Egress port ACL',
    WJH_ACL_BINDING_POINT_RIF_INGRESS: 'Ingress router ACL',
    WJH_ACL_BINDING_POINT_RIF_EGRESS: 'Egress router ACL'
}

drop_reason_group_to_channel = {}
WJH_MIRROR_CONG_INVALID = 0xFFFF
WJH_MIRROR_TCLASS_INVALID = 0x1F
WJH_MIRROR_LATENCY_INVALID = 0xFFFFFF

MELLANOX_WJH_HEADER = 'Mellanox WJH'

WJH_META_DATA_SOURCE_PORT = 1
WJH_META_DATA_DEST_PORT = 2
WJH_META_DATA_ORIG_PACKET_SIZE = 3
WJH_META_DATA_DROP_GROUP = 4
WJH_META_DATA_DROP_REASON = 5
WJH_META_DATA_MIN = 1
WJH_META_DATA_MAX = 5

PCAP_VERSION_MAJOR = 2
PCAP_VERSION_MINOR = 4
PCAP_MAGIC = 0xa1b2c3d4
PCAP_ERRBUF_SIZE = 256
PCAP_MAX_SNAPLEN = 65535
LINKTYPE_USER15 = 162

WjhDeactiveCb = CFUNCTYPE(None)

raw_channel = 0
aggregation_channel = 0
cyclic_and_aggregation_channel = 0

enable_print = False
enable_save = False

pcap_fd = -1
txt_fd = -1

pcap_user_fd = -1
txt_user_fd = -1

wjh_raw_file = WJH_PCAP_NAME
wjh_agg_file = WJH_TXT_NAME


class WjhInitParam(Structure):
    _fields_ = [("force", c_uint8), ("deactive_cb", WjhDeactiveCb), ("bandwidth_percent", c_uint8),
                ("log_cb", c_void_p), ("conf_xml_path", c_char_p), ("debug_fs_path", c_char_p)]


class WjhDropCounter(Structure):
    _fields_ = [("dropped", c_uint64), ("received", c_uint64)]


class WjhTimespec(Structure):
    _fields_ = [("tv_sec", c_long), ("tv_nsec", c_long)]


class WjhDropReason(Structure):
    _fields_ = [("id", c_uint32),
                ("reason", c_char_p),
                ("severity", c_int),
                ("description", c_char_p)]


class WjhDropBasicAttr(Structure):
    _fields_ = [("recirculation_port", c_uint32),
                ("truncate_size", c_uint16)]


class WjhDropAdvancedAttr(Structure):
    _fields_ = [("trap_probability", c_uint32),
                ("trap_id", c_uint32),
                ("ptp_disabled", c_uint32)]


class WjhBufferDropAttr(Structure):
    _fields_ = [("basic_attr", WjhDropBasicAttr),
                ("advanced_attr", WjhDropAdvancedAttr)]


class WjhDropReasonGroupAttrUnion(Union):
    _fields_ = [("buffer_drop", WjhBufferDropAttr)]


class WjhDropReasonGroupAttr(Structure):
    _fields_ = [("attr", WjhDropReasonGroupAttrUnion),
                ("max_aggregation_entries_cnt", c_uint32)]


class WjhUserChannelAttr(Structure):
    _fields_ = [("polling_interval", c_uint32),
                ("mode", c_int)]


class WjhDropAggregateAttr(Structure):
    _fields_ = []


class WjhL2DropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhL2DropRawCb = CFUNCTYPE(c_int, POINTER(WjhL2DropRawInfo), POINTER(c_uint32))


class WjhRouterDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhRouterDropRawCb = CFUNCTYPE(c_int, POINTER(WjhRouterDropRawInfo), POINTER(c_uint32))


class WjhTunnelDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32)]


WjhTunnelDropRawCb = CFUNCTYPE(c_int, POINTER(WjhTunnelDropRawInfo), POINTER(c_uint32))


class WjhL1DropRawInfo(Structure):
    _fields_ = [("ingress_port", c_uint32),
                ("timestamp", WjhTimespec)]


WjhL1DropRawCb = CFUNCTYPE(c_int, POINTER(WjhL1DropRawInfo), POINTER(c_uint32))


class WjhL1DropAggregateKey(Structure):
    _fields_ = [("ingress_port", c_uint32)]


class WjhAggregateTimestamp(Structure):
    _fields_ = [("first_timestamp", WjhTimespec), ("last_timestamp", WjhTimespec)]


class WjhL1DropAggregateData(Structure):
    _fields_ = [("is_port_up", c_uint8),
                ("port_down_reason", c_char_p),
                ("description", c_char_p),
                ("state_change_count", c_uint64),
                ("symbol_error_count", c_uint64),
                ("crc_error_count", c_uint64),
                ("timestamp", WjhAggregateTimestamp)]


WjhL1DropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhL1DropAggregateKey), POINTER(WjhL1DropAggregateData), POINTER(c_uint32), POINTER(WjhDropAggregateAttr))


class WjhAclDropRawInfo(Structure):
    _fields_ = [("packet", c_void_p),
                ("packet_size", c_uint32),
                ("ingress_port", c_uint32),
                ("binding_point", c_int),
                ("rule_id", c_uint64),
                ("acl_name", c_char_p),
                ("rule", c_char_p),
                ("timestamp", WjhTimespec),
                ("is_lag_member", c_uint8),
                ("ingress_lag", c_uint32),
                ("drop_reason", WjhDropReason)]


WjhAclDropRawCb = CFUNCTYPE(c_int, POINTER(WjhAclDropRawInfo), POINTER(c_uint32))


class WjhBufferDropRawInfo(Structure):
    _fields_ = [("packet", ctypes.c_void_p),
                ("packet_size", ctypes.c_uint32),
                ("ingress_port", ctypes.c_uint32),
                ("egress_port", ctypes.c_uint32),
                ("tc", ctypes.c_uint8),
                ("drop_reason", WjhDropReason),
                ("timestamp", WjhTimespec),
                ("is_lag_member", ctypes.c_uint8),
                ("ingress_lag", ctypes.c_uint32),
                ("egress_port_valid", ctypes.c_uint8),
                ("original_occupancy", ctypes.c_uint16),
                ("original_latency", ctypes.c_uint32)]


WjhBufferDropRawCb = CFUNCTYPE(c_int, POINTER(WjhBufferDropRawInfo), POINTER(c_uint32))


class WjhAggregateFiveTuples(Structure):
    _fields_ = [("sip", c_char_p),
                ("dip", c_char_p),
                ("proto", c_uint8),
                ("sport", c_uint32),
                ("dport", c_uint32)]


class WjhBufferDropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("non_ip", c_uint8),
                ("reason", WjhDropReason)]


class WjhBufferDropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint64)]


class WjhAclDropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("non_ip", c_uint8),
                ("rule_id", c_uint64),
                ("acl_name", c_char_p),
                ("rule", c_char_p)]


class WjhAclDropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint64),
                ("flows_num", c_uint64)]


class WjhL2DropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("non_ip", c_uint8),
                ("dmac", c_uint8 * 6),
                ("smac", c_uint8 * 6),
                ("reason", WjhDropReason)]


class WjhL2DropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint64)]


class WjhRouterDropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("non_ip", c_uint8),
                ("reason", WjhDropReason)]


class WjhRouterDropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint64)]


class WjhTunnelDropAggregateKey(Structure):
    _fields_ = [("five_tuples", WjhAggregateFiveTuples),
                ("non_ip", c_uint8),
                ("dmac", c_uint8 * 6),
                ("smac", c_uint8 * 6),
                ("reason", WjhDropReason)]


class WjhTunnelDropAggregateData(Structure):
    _fields_ = [("timestamp", WjhAggregateTimestamp),
                ("count", c_uint64)]


WjhBufferDropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhBufferDropAggregateKey), POINTER(WjhBufferDropAggregateData), POINTER(c_uint32), POINTER(WjhDropAggregateAttr))
WjhAclDropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhAclDropAggregateKey), POINTER(WjhAclDropAggregateData), POINTER(c_uint32), POINTER(WjhDropAggregateAttr))
WjhL2DropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhL2DropAggregateKey), POINTER(WjhL2DropAggregateData), POINTER(c_uint32), POINTER(WjhDropAggregateAttr))
WjhRouterDropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhRouterDropAggregateKey), POINTER(WjhRouterDropAggregateData), POINTER(c_uint32), POINTER(WjhDropAggregateAttr))
WjhTunnelDropAggregateCb = CFUNCTYPE(c_int, POINTER(WjhTunnelDropAggregateKey), POINTER(WjhTunnelDropAggregateData), POINTER(c_uint32), POINTER(WjhDropAggregateAttr))


class RawCallback(Union):
    _fields_ = [("buffer", WjhBufferDropRawCb),
                ("acl", WjhAclDropRawCb),
                ("router", WjhRouterDropRawCb),
                ("L2", WjhL2DropRawCb),
                ("L1", WjhL1DropRawCb),
                ("tunnel", WjhTunnelDropRawCb)]


class AggregateCallback(Union):
    _fields_ = [("buffer", WjhBufferDropAggregateCb),
                ("acl", WjhAclDropAggregateCb),
                ("router", WjhRouterDropAggregateCb),
                ("L2", WjhL2DropAggregateCb),
                ("L1", WjhL1DropAggregateCb),
                ("tunnel", WjhTunnelDropAggregateCb)]


class WjhDropCallbacks(Structure):
    _fields_ = [("drop_reason_group", c_int),
                ("raw_cb", RawCallback),
                ("aggregate_cb", AggregateCallback)]



