#
# Copyright (c) 2016-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import bisect
from functools import partial
import os
import socket
import subprocess
from subprocess import TimeoutExpired
import time
import uuid
import inspect
import errno
import re
import random
import copy
import textwrap
import atexit
from multiprocessing import Process, cpu_count
import psutil
import eventlet
import eventlet.pools
import greenlet
import ujson as json

from netq_lib.common.utils import check_cl_version_onwards, \
    is_opta_server_cloud, is_active_systemd_svc, is_swp_kind, \
    netq_agent_process_status_write, netq_agent_get_subprocess_status_str
from netq_lib.common import configuration, utils
from netq_lib.common.nlfullstate import NetlinkFullStateObject, FullState
from netq_lib.common.enums import (
    Daemons,
    ConfKw,
    Heartbeat,
    NeighborState,
    LinkType,
    Platform,
    RunningService,
    CentosRunningService,
    RhelRunningService,
    SonicRunningService,
    RouteTable,
    ServiceStatus,
    StpPortState,
    PlatformType,
    AGENTSTATUS,
    ForwardingState,
    wjhSubProcessState,
    DefaultPaths,
    NetqAgentStatus,
    NetqAgentState,
    NetqLogGroup,
    ModelConstants as MC,
)

from netq_lib.common.service import Service
from netq_lib.common.agent_stats_monitor import AgentStatsMon
from netq_lib.orm import HOSTNAME
import netq_lib.orm.redisdb.models  # for purging all classes in file @ init
from netq_lib.orm.redisdb import (
    set_connection_settings,
    ping_server,
    server_info,
    is_backend_sentinel_aware
)
from netq_lib.orm.redisdb.models import ( #pylint: disable=unused-import
    Address,
    Description,
    Link,
    Liveness,
    MacFdb,
    Neighbor,
    Node,
    Route,
    Services,
    VxlanRemoteDest,
    Netconf,
)
from netq_lib.orm.redisdb import unsubscribe_listener
from netq_lib.common.netq_pq import create_message_queues, create_gnmi_message_queues, \
    create_agent_info_queue, setClAgentInst, NetQPriorityQueue
from netq_lib.common.utils import is_cl3, is_cl4, is_sonic, is_sonic_ib, is_ubuntu, is_aarch64, is_gnmi_enabled

from netq_agent.cmd.base_wjh_agent import DummyWjhClient
from netq_agent.handlers.sdk_utils import SdkUtils
from netq_agent.parsers.cumulus import dom, ptp
from netq_agent.parsers.sonic import proc as sonic_proc
from netq_agent.parsers.common import (
    netlink,
    package,
    ntp,
    systemctl
)
from netq_agent.parsers.cumulus.platform import process_platform_info as cl_process_platform_info
from netq_agent.parsers.sonic.platform import process_platform_info as sonic_process_platform_info
from netq_agent.parsers.sonic.featureStatus import process_feature_status
from netq_agent.parsers.cumulus.systemctl import process_systemctl
from netq_agent.parsers.sonic.systemctl import process_systemctl as sonic_process_systemctl
from netq_agent.parsers.bluefield.platform import process_platform_info as bluefield_process_platform_info
from netq_agent.parsers.common.netlink import NetlinkCopy
from netq_agent.cmd.netq_agent_commands import NetqCommands
# from netq_agent.cmd.netq_agent_monitor import NetqAgentMonitor
from netq_agent.cmd.netq_spice_agent import SpiceCollectionAgent
from netq_agent.cmd.docker_agent import DockerMonitor
from netq_agent.cmd.kube_agent import KubeMonitor
from netq_agent.cmd.netq_grpc_agent import NetQAgentGrpcClient
from netq_agent.cmd.agent_cli import AgentCli
from netq_agent.cmd.netq_agent_opta_discovery import NetQAgentOPTADiscovery
from netq_agent.cmd.netq_asic_monitor_service import NetQAsicMonitor
from netq_lib.common.configuration import load_configuration

if is_cl3() or is_cl4() or is_sonic() or is_ubuntu() or is_aarch64():
    from netq_agent.cmd.netq_agent_libnl import NetLinkProc, CommandServer, AgentCommands, NetlinkMsgType
if is_sonic():
    from netq_agent.parsers.sonic.redis_db.client import SonicDBSubscribeManager

_PROC_NAME = 'netq-agent'
_is_deliberate_wjh_disable = False

class _Command(object):
    """Command class.
    """
    _MIN_PERIOD = 15  # minimum period for command runs: 15s
    _CMD_TIMEOUT = 5  # End the cmd if not finished within this time
    _CMD_RANDOM_START = 20
    _CMD_RANDOM_STOP = 60
    _CMD_RANDOM_STEP = 10

    # pylint: disable=too-few-public-methods
    def __init__(self, period, key, command, service, callback=None, #pylint: disable=dangerous-default-value
                 always=False, tov=_CMD_TIMEOUT, fullstate=0, ignore=[]):
        try:
            period = int(period)
        except ValueError:
            period = self._MIN_PERIOD
        self.period = self._MIN_PERIOD if period < self._MIN_PERIOD else period
        self.key = key
        self.command = command
        self.service = service
        # function to parse & push out structured data
        self.callback = callback
        self.always = always
        try:
            tov = int(tov)
        except ValueError:
            tov = self._CMD_TIMEOUT
        self.timeout = tov
        self.errcnt = 0
        self.fullstate = fullstate
        self.fullstateCnt = fullstate
        self.ignore = ignore

    def __lt__(self, other):
        return (self.period - other.period) < (other.period - self.period)

    def __gt__(self, other):
        return (self.period - other.period) > (other.period - self.period)

    def __repr__(self):
        return '%s(period=%d, key=%s, command=%s)' % (
            self.__class__.__name__, self.period, self.key, self.command
        )

    def get_rand_period(self):
        # make sure commands are run within the first 60s of agent init
        if self.key == "ports":
            rand_period = self._CMD_RANDOM_START
        elif self.key == "dom":
            rand_period = self._CMD_RANDOM_STOP
        else:
            rand_period = random.randrange(self._CMD_RANDOM_START, self._CMD_RANDOM_STOP, self._CMD_RANDOM_STEP)
        return rand_period


class _Thread(object):
    """Thread class.
    """

    def __init__(self):
        self.__function = None
        self.__pool = None
        self.__start_time = None
        self.__stop_checker = None
        self.green_thread = None
        self.pipeline = None

    def kill(self):
        eventlet.kill(self.green_thread)

    def link(self, stop_checker):
        self.__stop_checker = stop_checker
        self.green_thread.link(stop_checker)

    @classmethod
    def spawn(cls, epoch, pool, function, pipeline=None):
        thread = cls()
        thread.__function = function  # pylint: disable=protected-access
        thread.__pool = pool  # pylint: disable=protected-access
        thread.__start_time = epoch  # pylint: disable=protected-access
        thread.pipeline = pipeline
        if pipeline is not None:
            thread.green_thread = pool.spawn(function, pipeline)
        else:
            thread.green_thread = pool.spawn(function)
        return thread

    def stale(self, now):
        return now - self.__start_time >= Heartbeat.HEARTBEAT_DEATH

    def respawn(self, epoch):
        self.kill()
        thread = self.spawn(epoch, self.__pool, self.__function,
                            pipeline=self.pipeline)
        return thread


class _ClDbPush(Service):
    """
    This class defines the main workhorse class of this file.
    It parses the args, fires off the command event monitoring
    threads, and processes the outputs posted by these threads
    and adds them to a redis database.
    """
    __BRIDGE_FBD_REGEX = (
        r'(?:(?P<deleted>^Deleted)\s+)?'
        r'(?P<mac_address>\S+)'
        r'\s+\S+\s+(?P<nexthop>\S+)'
        r'(?=.*vlan\s+(?P<vlan_id>\d+))?'
        r'(?=.*\s+(?P<origin>permanent$))?'
        r'(?=.*dst\s+(?P<dst>\S+))?'
        r'(?=.*master\s+(?P<port>\S+))?'
    )

    __SERVICES_RATE = 15
    __WJH_RESTART_HOLD_PERIOD = 30
    __WJH_MONITOR_RATE = 120
    __NLPOLL_RATE = 30
    __NLPOLL_RETRY_MAX = 2
    __NLPOLL_RETRY_MIN = 1
    __APPINFO_RATE = 300
    __HOUR_IN_SEC = 3600
    if not is_sonic():
        __PLATFORM_RATE = 60  # updated for agent cpu-limit
    else:
        __PLATFORM_RATE = 300

    """
    # update default command timeout to 30 for cpu_limit
    if 'arm' in utils.get_cpu_architecture():
        __CMD_TIMEOUT = 30
    else:
        __CMD_TIMEOUT = 15
    """
    __CMD_TIMEOUT = 30

    __AGENT_PAUSE = False
    __AGENT_PAUSE_FIRST_RUN = False

    __GNMI_AGENT_PAUSE = False
    __FIRST_RUN = False

    __NETLINK_INIT_DONE = False
    __NETLINK_RESTART = False

    __NETLINK_COUNT = 2
    __INT_CMDS = {}

    __MAX_REDIS_CONNECTIONS = 10

    __PIPELINE_RATE = 1

    __MAX_SERVER_INFO_FAIL_CNT = 3
    __SERVERINFO_RATE = 15

    __REDISTRIBUTE_NBR_TABLE_ID = 10

    __NON_CLS_MODELS = ['_RedisModel', 'RedisQuery', 'Connection', 'MemModel',
                        'Node', 'Liveness', 'ProcNetDevInterface', 'datetime',
                        'NetQPriorityQueue', 'QueueElem', 'NetQIPC']
    __NL_CLS_MODELS = ['Link', 'Address', 'Neighbor', 'Route', 'NetConf']

    __FRR_DOCKER_DEFAULT = 'cumulus-frr'

    def __init__(self, conf, grpc_stream_process, libnl_process, ng_agent_process):
        shutdown_cb = [unsubscribe_listener]
        if is_sonic():
            shutdown_cb.append(self._sonic_unsubscribe_listener)
        super(_ClDbPush, self).__init__(conf, _PROC_NAME,
                                        shutdown_callbacks=shutdown_cb)

        if not configuration.is_cl3():
            self._logger.debug('Removing BTRFS polling')
            _ClDbPush.__INT_CMDS.pop('btrfs', 'btrfs')

        self.__commands = []
        self.__command_buffer = {}
        self.__heartbeat_gt = None
        self.__hostname = HOSTNAME
        self.__prev_hostname = 'unknown'
        self.__ifname_by_index = None
        self.__lastboot = time.time()
        self.__last_reinit = time.time()
        self.__nlinit_gt = None
        self.__nlmonitor_gt = None
        self.__vxlan_vlan_map = {}
        self.__link_kind_map = {}
        self.__link_table_map = {}  # Link to route table map
        self.__link_vrf_map = {}  # Link to VRF map, may not be direct master
        self.__rt_protos_map = {}
        self.__link_master_map = {}  # Map of master for i/fs with a master
        self.__addr_ifindex_map = {}  # Map of addr to ifindex when ifname is none
        self.__to_update_master_links = {}
        self.__forwardingState = ForwardingState.FORWARDING_STATE_UNKNOWN
        self.__wjhSubProcessState = wjhSubProcessState.UNKNOWN
        self.__wjhLoggedOnce = False
        self.__logger_coalescer = utils.LogCoalescer("netq-agent", conf_dict=configuration.get_config(_PROC_NAME))
        self.__nlmonitor = netlink.Netlink(
            _PROC_NAME,
            self._conf,
            self._logger,
            {netlink.Netlink.NEWADDR: self.__rx_rtm_newaddr,
             netlink.Netlink.DELADDR: self.__rx_rtm_deladdr,
             netlink.Netlink.NEWLINK: self.__rx_rtm_newlink,
             netlink.Netlink.DELLINK: self.__rx_rtm_dellink,
             netlink.Netlink.NEWNEIGH: self.__rx_rtm_newneigh,
             netlink.Netlink.DELNEIGH: self.__rx_rtm_delneigh,
             netlink.Netlink.NEWROUTE: self.__rx_rtm_newroute,
             netlink.Netlink.DELROUTE: self.__rx_rtm_delroute,
             netlink.Netlink.NEWNETCONF: self.__rx_rtm_newnetconf},
            pool=self._pool
        )
        self.__prev_cmd = {}
        self.__prev_cntr = {}
        self.__prev_cntr = {}
        self.__redispool = eventlet.pools.TokenPool(
            max_size=self.__MAX_REDIS_CONNECTIONS
        )

        # Failed Commands
        self.__failed_commands = {}

        # Platform Commands
        self.__platform_commands = {}

        self.__redisthreads = set()
        self.__reinitpool = eventlet.pools.TokenPool(max_size=1)
        self.__heartbeatpool = eventlet.pools.TokenPool(max_size=1)
        self.__serverinfopool = eventlet.pools.TokenPool(max_size=1)
        self.__server_info_fail_cnt = 0
        self.__timers = {}
        self.__platform = self.__get_platform()
        self.__app_release_info = None
        self.__agent_version = self.__get_agent_version()
        self.__server_info = {}
        self.__ports = set()
        self.__optical_ports = set()
        self.__config_commands = NetqCommands(self._conf, self._logger)
        self.__docker = DockerMonitor(self._conf, self._logger, platform=self.__platform,
                                      pool=self._pool)
        self.__kubernetes = KubeMonitor(self._conf, self._logger, platform=self.__platform,
                                        pool=self._pool)

        self.__node_uptime = psutil.boot_time()

        # check if interface stats is enabled
        self.__intf_stats_enabled = self.__is_intf_stats_enabled()

        # check if opta config enabled
        self.__opta_enabled = self.__is_opta_enabled()

        # check if opta config enabled
        self.__gnmi_enabled = self.__is_gnmi_enabled()

        # check if cluster configured
        self.__cluster_enabled = self.__is_cluster_enabled()

        # check if randon platform config enabled
        self.__random_platform_info = self.__get_random_platform()

        # check if redis config enabled
        self.__redis_enabled = self.__is_redis_enabled()

        # init frr docker name if enabled
        self.__frr_docker_enabled = self.__init_frr_docker()

        # full_state period/timer
        self.__fullstate_period = self.__get_fullstate_period()

        self.__platform_fullstate_period = self.__fullstate_period + self.__HOUR_IN_SEC

        # full state period timer for platform data
        self.__platform_fullstate_timer = self.__platform_fullstate_period

        # full state period timer for netlink
        self.__netlink_fullstate_period = self.__fullstate_period + (2 * self.__HOUR_IN_SEC)


        # trace grpc Q full state
        self.__grpc_q_full = False
        self.__nl_poll_retry = 0
        self.__nl_poll_retry_max = self.__NLPOLL_RETRY_MAX

        self.__grpc_state = False
        self.__gnmi_grpc_state = False

        if self.__fullstate_period:
            self.full_state_object = NetlinkFullStateObject
        else:
            self.full_state_object = None

        # Log exporter process
        self.__log_exporter_process = self._start_log_exporter()

        self.__asic_monitor_service = None

        # Stream Process
        self.__grpc_stream_process = grpc_stream_process

        if grpc_stream_process.is_alive():
            msg_string = netq_agent_get_subprocess_status_str("gRPC stream process",
                                                              grpc_stream_process.pid,
                                                              "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)
            self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.GRPC_STATUS)
        else:
            msg_string = netq_agent_get_subprocess_status_str("gRPC stream process", -1,
                                                              "Failed")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                            msg_string)
            self.__logger_coalescer.info(msg_string, group=NetqLogGroup.GRPC_STATUS)

        # WJH Process
        msg_string = netq_agent_get_subprocess_status_str("WJH", -1, "NA/Not Enabled")
        netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_WJH_STATUS,
                                        NetqAgentState.NETQ_AGENT_STATE_NA,
                                        msg_string)

        self.__wjh_process = self._start_wjh()
        self.__wjh_restart_hold = 0
        self.__sx_sdk_pid = 0
        self.__sx_sdk_check = 10

        # SPICE process
        msg_string = netq_agent_get_subprocess_status_str("Spice", -1, "NA/Not Enabled")
        netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_SPICE_STATUS,
                                        NetqAgentState.NETQ_AGENT_STATE_NA,
                                        msg_string)

        self.__spice_process = self._start_spice()
        self.__spice_restart_hold = 0
        self.__SpiceLoggedOnce = False

        # The following parameters are for the activate agent resume state
        self.__backoff_timer = 0
        self.__next_reinit_time = 0
        self.__backoff_timer_step = 5
        self.__latest_pause_time = time.time()
        self.__agent_stable_time = self.__backoff_timer_step * 6

        self.__netlink_init_time = 0
        # In Seconds
        self.__max_netlink_init_time = 7 * 60


        setClAgentInst(self)

        # Update period
        self.__update_cmdlist_period(self.__INT_CMDS)

        # Agent Monitor
        # self.__agent_monitor = NetqAgentMonitor(conf)
        # self.__agent_monitor.init_client(conf)

        #Netlink Command
        self.__netlink_cmd_client = None
        self._is_netlink_enabled = self._conf.get('netq-agent').get(ConfKw.IS_NETLINK_ENABLED, ConfKw.DEFAULT_IS_NETLINK_ENABLED)
        self._is_libnl_enabled = self._conf.get('netq-agent').get(ConfKw.IS_LIBNL_ENABLED, ConfKw.DEFAULT_IS_LIBNL_ENABLED)

        self._logger.info("Libnl enabled {}".format(self._is_libnl_enabled))
        if self._is_libnl_enabled and not is_libnl_required():
            self._logger.info("Disabling Libnl")
            self._is_libnl_enabled = False

        if utils.is_hostname_default(self.__hostname):
            self._is_libnl_enabled = False

        if self._is_libnl_enabled:
            self.__netlink_cmd_client = CommandServer(conf, self._logger)
            self.__netlink_libnl_process = libnl_process
            self.__netlink_libnl_retry_timer = 0
            if libnl_process.is_alive():
                msg_string = netq_agent_get_subprocess_status_str("netlink process",
                                                                  libnl_process.pid,
                                                                  "Alive")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NL_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string)
                self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.NL_STATUS)
            else:
                msg_string = netq_agent_get_subprocess_status_str("netlink process", -1,
                                                                  "Failed")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NL_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                                msg_string)
                self.__logger_coalescer.info(msg_string, group=NetqLogGroup.NL_STATUS)
        else:
            self.__netlink_libnl_process = None
            msg_string = netq_agent_get_subprocess_status_str("netlink process", -1,
                                                              "libnl not Enabled")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NL_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)
            self.__logger_coalescer.info(msg_string, group=NetqLogGroup.NL_STATUS)

        # AGENT CLI
        self.__agent_cli = AgentCli(self._pool, self._logger, \
                self.__nlmonitor, self._is_libnl_enabled, self.__netlink_cmd_client)

        # NetQ-NG-Agent Service
        self.__ng_agent_process = ng_agent_process
        if self.__ng_agent_process and self.__is_ng_agent_process_active():
            msg_string = netq_agent_get_subprocess_status_str("ng agent process",
                                                              ng_agent_process.pid,
                                                              "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NG_AGENT_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)
            self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.NG_AGENT_STATUS)
        elif (is_cl4() or is_sonic()) and 'arm' not in utils.get_cpu_architecture():
            msg_string = netq_agent_get_subprocess_status_str("ng agent process", -1,
                                                              "Failed")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NG_AGENT_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                            msg_string)
            self.__logger_coalescer.info(msg_string, group=NetqLogGroup.NG_AGENT_STATUS)
        else:
            msg_string = netq_agent_get_subprocess_status_str("ng agent process", -1,
                                                              "NA")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NG_AGENT_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)

        self.__ng_agent_back_off = 60

        self.__subscribe_manager = None
        if is_sonic():
            # NetQ DB Subscribe Manager
            self.__subscribe_manager = SonicDBSubscribeManager(self._pool, self._logger)

        if Platform.VX in self.__platform:
            # Temporary workaround for netq-agent filling up systemd journal logs
            self.journald_system_custom_config = self._conf.get('netq-agent').get(ConfKw.JOURNALD_SYSTEM_CUSTOM_CONFIG,
                                                                                  ConfKw.DEFAULT_JOURNALD_SYSTEM_CUSTOM_CONFIG)
            if self.journald_system_custom_config:
                self.journald_system_max_use = self._conf.get('netq-agent').get(ConfKw.JOURNALD_SYSTEM_MAX_USE,
                                                                                ConfKw.DEFAULT_JOURNALD_SYSTEM_MAX_USE)
                self.journald_runtime_max_use = self._conf.get('netq-agent').get(ConfKw.JOURNALD_RUNTIME_MAX_USE,
                                                                                 ConfKw.DEFAULT_JOURNALD_RUNTIME_MAX_USE)
                self.journald_storage = self._conf.get('netq-agent').get(ConfKw.JOURNALD_STORAGE, ConfKw.DEFAULT_JOURNALD_STORAGE)
                self.limit_systemd_journalctl_size()
                self._logger.info("Updated systemd_journald settings SystemMaxUse={} RuntimeMaxUse={} Storage={}"
                                  .format(self.journald_system_max_use,
                                          self.journald_runtime_max_use,
                                          self.journald_storage))

    def limit_systemd_journalctl_size(self):
        config_file = '/etc/systemd/journald.conf.d/99-cumulus-journald.conf'
        try:
            # Read the existing configuration file
            with open(config_file, 'r') as file:
                config_lines = file.readlines()

            new_config_lines = []
            storage_config_added = False
            for line in config_lines:
                if line.startswith('RuntimeMaxUse='):
                    # Modify the RuntimeMaxUse option
                    line = 'RuntimeMaxUse='+self.journald_runtime_max_use+'\n'
                if line.startswith('SystemMaxUse='):
                    line = 'SystemMaxUse='+self.journald_system_max_use+'\n'
                if line.startswith('Storage='):
                    line = 'Storage='+self.journald_storage+'\n'
                    storage_config_added = True
                new_config_lines.append(line)

            if not storage_config_added:
                line = 'Storage=persistent\n'
                new_config_lines.append(line)

            # Write the modified configuration back to the file
            with open(config_file, 'w') as file:
                file.writelines(new_config_lines)

            subprocess.run(['systemctl', 'restart', 'systemd-journald'])
        except FileNotFoundError:
            self._logger.debug(f"The config file '{config_file}' does not exist. Using default limit for journalctl size")
        except Exception as e:
            self._logger.error(f"An unexpected error occurred: {str(e)}. Failed to limit journalctl size")


    def get_nlmonitor(self):
        return self.__nlmonitor

    def __run_safe_cmd(self, command, timeout=__CMD_TIMEOUT):
        """Run time-limited command and return status & cmd output
        return code is 0 for successful commands and non-zero otherwise
        """
        cmd_out = retcode = None
        try:
            if not command:
                return -1, ''
            if isinstance(command, str):
                cmd = subprocess.Popen(command.split(),
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       shell=False)
            else:
                cmd = subprocess.Popen(command,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       shell=False)
        except (subprocess.CalledProcessError, OSError):
            return -1, ''

        identifier = uuid.uuid4().hex
        green_thread = self._pool.spawn(cmd.communicate)
        green_thread.link(self.__stop_command_checker, identifier, command)
        endtime = time.time() + int(timeout)

        while True:
            if identifier in self.__command_buffer:
                cmd_out = self.__command_buffer.pop(identifier)
                if cmd_out is None:
                    retcode = -1
                else:
                    retcode = cmd.returncode
                break
            if time.time() > endtime:
                self._logger.info('Failed to execute command %s after %d '
                                  'seconds', command, timeout)
                cmd.terminate()
                eventlet.sleep(0.3)
                cmd.wait()
                eventlet.kill(green_thread, greenlet.GreenletExit)
                retcode = -1
                break
            eventlet.sleep(1)
        if retcode != 0:
            if isinstance(command, list):
                command = ",".join(command)
            if command not in self.__failed_commands:
                self._logger.debug('Command %s returned %d', command, retcode)
                self.__failed_commands[command] = retcode
        if cmd_out:
            return retcode, cmd_out.decode()
        else:
            return retcode, None

    def __docker_routing_svc_running(self, daemon):

        if not self.__frr_docker_name:
            return False

        dockerbin = self.__get_docker_bin()

        cmdstr = "".join((dockerbin, ' exec ', self.__frr_docker_name,
                          '/usr/lib/frr/frr', 'status'))

        try:
            retcode, cmd_out = utils.subprocess_wrapper(cmdstr.split())
        except (subprocess.CalledProcessError, OSError):
            return False

        if retcode:
            return False

        # FRR docker status will print something like this
        #
        #  * zebra is running
        #  * bgpd is running

        lines = cmd_out.splitlines()
        for line in lines:
            if daemon not in line:
                continue
            if 'running' in line:
                return True
        return False

    def __generic_routing_svc_running(self, daemon):

        if os.path.isdir('/etc/frr'):
            svc_name = RunningService.FRR
        else:
            svc_name = RunningService.QUAGGA

        cmdstr = '/bin/systemctl is-active %s.service' % svc_name

        try:
            utils.subprocess_wrapper(cmdstr.split())
        except (subprocess.CalledProcessError, OSError):
            return False

        # Quagga / FRR daemons need one more check
        cmdstr = '/bin/grep %s=yes /etc/%s/daemons' % (daemon, svc_name)
        try:
            utils.subprocess_wrapper(cmdstr.split())
            return True
        except (subprocess.CalledProcessError, OSError):
            return False

    def __is_routing_service_running(self, daemon):

        if self.__frr_docker_enabled:
            # we are asked to work with FRR docker
            return self.__docker_routing_svc_running(daemon)
        else:
            return self.__generic_routing_svc_running(daemon)

    def __is_service_running(self, daemon, vrf=None):
        """Check if service is running. Works across systemctl & service
        models.
        """
        routing_daemons = [Daemons.BGPD, Daemons.OSPFD, Daemons.OSPF6D,
                           Daemons.ZEBRA]

        if daemon in routing_daemons:
            return self.__is_routing_service_running(daemon)

        if os.path.isfile('/bin/systemctl') and not is_sonic_ib():
            if vrf:
                cmdstr = '/bin/systemctl is-active %s@%s' % (daemon, vrf)
            else:
                cmdstr = '/bin/systemctl is-active %s' % daemon
        else:
            cmdstr = '/usr/sbin/service %s status' % daemon

        try:
            utils.subprocess_wrapper(cmdstr.split())
        except (subprocess.CalledProcessError, OSError):
            return False

        return True

    def __docker_routing_svc_failed(self, daemon):

        if not self.__frr_docker_name:
            return True

        dockerbin = self.__get_docker_bin()
        cmdstr = "".join((dockerbin, ' exec ', self.__frr_docker_name,
                          '/usr/lib/frr/frr', 'status'))
        try:
            retcode, cmd_out = utils.subprocess_wrapper(cmdstr.split())
        except (subprocess.CalledProcessError, OSError):
            return True

        if retcode:
            return True

        # FRR docker status will print something like this
        #
        #  * zebra is running
        #  * bgpd is running

        lines = cmd_out.splitlines()
        for line in lines:
            if daemon not in line:
                continue
            if 'running' in line:
                return False
        return True

    def __generic_routing_svc_failed(self, daemon):

        if os.path.isdir('/etc/frr'):
            svc_name = RunningService.FRR
        else:
            svc_name = RunningService.QUAGGA

        cmdstr = '/bin/systemctl is-failed %s.service' % svc_name
        out = None
        try:
            utils.subprocess_wrapper(cmdstr.split())
            return True
        except (subprocess.CalledProcessError, OSError) as ex:
            # attrib 'output' only present for CalledProcessError
            out = getattr(ex, 'output', None)
        if out and 'active' in out.decode():
            cmdstr = '/usr/bin/vtysh -d %s -c quit' % daemon
            try:
                utils.subprocess_wrapper(cmdstr.split())
            except (subprocess.CalledProcessError, OSError):
                return True
        return False

    def __is_routing_service_failed(self, daemon):

        if self.__frr_docker_enabled:
            # we are asked to work with FRR docker
            return self.__docker_routing_svc_failed(daemon)
        else:
            return self.__generic_routing_svc_failed(daemon)

    def __has_service_failed(self, daemon):
        """Check if service has failed. Works only with systemctl for now
        """
        routing_daemons = [Daemons.BGPD, Daemons.OSPFD, Daemons.OSPF6D,
                           Daemons.ZEBRA]

        if daemon in routing_daemons:
            return self.__is_routing_service_failed(daemon)

        if os.path.isfile('/bin/systemctl'):
            cmdstr = '/bin/systemctl is-failed %s' % daemon
        else:
            cmdstr = '/usr/sbin/service %s status' % daemon

        try:
            utils.subprocess_wrapper(cmdstr.split())
            return True
        except (subprocess.CalledProcessError, OSError):
            return False

    def __is_opta(self):
        return os.path.isfile('/etc/app-release')

    def __update_cmdlist_period(self, cmdlist):
        arch = utils.get_cpu_architecture()
        if arch and "arm" in arch:
            for service, commands in list(cmdlist.items()):
                for command in commands:
                    if isinstance(command, list):
                        newcmds = command
                    else:
                        newcmds = [command]
                    for newcmd in newcmds:
                        newcmd['period'] = str(int(newcmd['period']) + 120)

    def __load_commands(self):
        """Determine what info to push based on box configuration.
        """
        cmdlist = {}
        self.__timers = {}

        if not self.__intf_stats_enabled:
            # remove interface polling command
            for entry in cmdlist['misc']:
                if entry.get("key") == "proc-net-dev":
                    cmdlist['misc'].remove(entry)

        self.__config_commands.init_commands(cmdlist)

        # ignore running these on opta only
        if self.__is_opta():
            if 'lldpd' in cmdlist:
                del cmdlist['lldpd']
            if 'smond' in cmdlist:
                del cmdlist['smond']

        # ignore running these on non-cumulus or opta
        if 'cumulus' not in self.__platform.lower() or self.__is_opta():
            if 'btrfs' in cmdlist:
                del cmdlist['btrfs']
            if 'switchd' in cmdlist:
                del cmdlist['switchd']

        # ignore running these on non-cumulus/non-sonic or opta
        if self.__is_opta() or ('cumulus' not in self.__platform.lower() and
                                'sonic' not in self.__platform.lower()):
            if 'ssdUtil' in cmdlist:
                del cmdlist['ssdUtil']
            if 'dpkg' in cmdlist:
                del cmdlist['dpkg']

        if not self.__is_opta() and ('cumulus' not in self.__platform.lower() and
                                     'sonic' not in self.__platform.lower()):
            if 'resourceUtil' in cmdlist:
                del cmdlist['resourceUtil']

        # ignore running these on VX
        if Platform.VX in self.__platform:
            if 'switchd' in cmdlist:
                del cmdlist['switchd']

        # enable ECMP only for Mellanox platform and on CL >= 5.1.0
        if not (check_cl_version_onwards(major=5, minor=1) and Platform.MELLANOX in self.__platform):
            if 'ecmp' in cmdlist:
                del cmdlist['ecmp']

        for service, commands in list(cmdlist.items()):
            for command in commands:
                if isinstance(command, list):
                    newcmds = command
                else:
                    newcmds = [command]
                for command in newcmds: #pylint: disable=redefined-outer-name
                    if service == 'inventory' and all(k in command
                                                      for k in ('key', 'command')):
                        key = str(command['key'].strip())
                        self.__platform_commands[key] = command['command']
                    elif all(k in command
                             for k in ('period', 'key', 'command')):
                        # last param is None because there is no
                        # command to parse out the info and push it out
                        self._logger.debug("Processing Service {}".format(str(service)))
                        key = str(command['key'].strip())
                        cmd = _Command(self.__get_command_period(key),
                                       key,
                                       command['command'],
                                       str(service),
                                       command.get('callback', None),
                                       command.get('run', None) ==
                                       'always',
                                       command.get('timeout', self.__CMD_TIMEOUT),
                                       self.__fullstate_period,
                                       command.get('ignore', []))
                        idx = bisect.bisect_left(self.__commands, cmd)
                        if (idx >= len(self.__commands) or
                                self.__commands[idx] != cmd):
                            bisect.insort(self.__commands, cmd)
                        else:
                            self._logger.info('Ignoring invalid line %r', command)

        epoch = int(time.time())
        for cmd in self.__commands:
            rand_period = cmd.get_rand_period()
            self.__timers.setdefault(epoch + rand_period, set())
            self.__timers[epoch + rand_period].add(cmd)

    def __get_docker_bin(self):

        if 'rhel' in self.__platform.lower():
            cmd = '/bin/docker'
        elif 'red hat' in self.__platform.lower():
            cmd = '/bin/docker'
        elif 'centos' in self.__platform.lower():
            cmd = '/bin/docker'
        else:
            cmd = '/usr/bin/docker'

        return cmd

    def __get_frr_docker_pids(self):

        routing_daemons = [Daemons.BGPD, Daemons.OSPFD, Daemons.OSPF6D,
                           Daemons.ZEBRA]

        if not self.__frr_docker_name and bool(self.__frr_docker_child_pid):
            # update services
            for daemon in routing_daemons:
                if self.__frr_docker_child_pid.get(daemon):
                    self.__frr_docker_child_pid[daemon] = 'n/a'
            try:
                systemctl.process_frr_docker_service(self.__frr_docker_child_pid,
                                                     self.__frr_docker_start_time)
            except Exception as ex:  # pylint: disable=broad-except
                self._logger.info('Failed to process frr docker services. Error: %s', ex)
            return

        dockerbin = self.__get_docker_bin()
        cmd = ' --format "{{.State.Pid}}"'
        dockercmd = dockerbin + ' inspect ' + self.__frr_docker_name + cmd

        retcode, cmd_out = self.__run_safe_cmd(dockercmd)
        if retcode:
            self._logger.info('Failed to get FRR docker pid %x', retcode)
            return

        self.__frr_docker_pid = cmd_out.strip('"\n')

        self._logger.debug('Found FRR docker pid %s', self.__frr_docker_pid)

        # now get the frr daemon pids
        # grab all child pids of this container
        cmd = (' --ppid %s --format pid,cmd' % self.__frr_docker_pid)
        dockercmd = dockerbin + ' top ' + self.__frr_docker_name + cmd

        retcode, cmd_out = self.__run_safe_cmd(dockercmd)

        if retcode:
            self._logger.info('Failed to get FRR docker child pids %x', retcode)
            return

        # ex. of docker top output from previous cmd
        # PID      CMD
        # 8426     /usr/lib/frr/zebra -M snmp -s 90000000 --daemon -A 127.0.0.1
        # 8442     /usr/lib/frr/bgpd -M snmp --daemon -A 127.0.0.1
        # 8448     /usr/lib/frr/watchfrr -adz -r /usr/sbin...

        routing_daemons = [Daemons.BGPD, Daemons.OSPFD, Daemons.OSPF6D,
                           Daemons.ZEBRA]

        self.__frr_docker_child_pid = {}
        lines = cmd_out.splitlines()
        for line in lines:
            if 'PID' in line:
                continue
            childpid = line.split()[0]
            cmdstr = line.split()[1]
            if 'watchfrr' in cmdstr:
                continue
            for daemon in routing_daemons:
                if daemon in cmdstr:
                    self.__frr_docker_child_pid[daemon] = childpid
                    self._logger.debug('Found FRR daemon %s pid %s',
                                       daemon, childpid)

        # update services
        try:
            systemctl.process_frr_docker_service(self.__frr_docker_child_pid,
                                                 self.__frr_docker_start_time)
        except Exception as ex:  # pylint: disable=broad-except
            self._logger.debug('Failed to process frr docker services. Error: %s', ex)
            self.__frr_docker_child_pid = {}
        return

    def __get_frr_docker_start_time(self):

        if not self.__frr_docker_name:
            self.__frr_docker_start_time = None
            return

        # get the frr docker start time
        cmd = ' --format="{{.State.StartedAt}}" '
        dockercmd = self.__get_docker_bin() + ' inspect ' + \
                    self.__frr_docker_name + cmd

        retcode, cmd_out = self.__run_safe_cmd(dockercmd)

        if retcode:
            self._logger.info('Failed to get FRR docker start time %x', retcode)
            self.__frr_docker_start_time = None
            return

        self._logger.debug('Updated FRR docker start time %s', cmd_out)

        # convert into unix timestamps
        xargcmd = 'xargs -n1 date +%s -d ' + cmd_out.strip('"\n')
        retcode, cmd_out = self.__run_safe_cmd(xargcmd)
        if retcode:
            self._logger.info('Failed to convert FRR docker start time %x', retcode)
            self.__frr_docker_start_time = None
            return

        self.__frr_docker_start_time = cmd_out.strip('"\n')
        self._logger.debug('Updated FRR docker start time utc %s',
                           self.__frr_docker_start_time)

    def __update_frr_docker_info(self):

        self.__get_frr_docker_name()

        if not self.__frr_docker_name:
            return

        # keep track of container pid
        # also get the bgp, zebra, ospf child pids
        if not bool(self.__frr_docker_child_pid):
            self.__get_frr_docker_pids()

    def __get_frr_docker_name(self):
        old_frr = self.__frr_docker_name
        self.__frr_docker_name = None

        cmd = ' ps --format "{{.Names}}"'
        dockercmd = self.__get_docker_bin() + cmd

        retcode, cmd_out = self.__run_safe_cmd(dockercmd)
        if retcode:
            self._logger.debug('Failed to run docker ps retcode - %x', retcode)
            return

        first_frr = None
        lines = cmd_out.splitlines()
        for line in lines:
            if self.__frr_docker_conf_name not in line:
                continue
            # ignore certain kube instances
            if 'POD' in line:
                continue
            # mark the first frr we see
            curr_frr = line.strip('"')
            if not first_frr:
                first_frr = curr_frr
            if not old_frr or curr_frr == old_frr:
                #  we found it !
                self.__frr_docker_name = curr_frr
                break

        if old_frr and not self.__frr_docker_name and first_frr:
            # we had a previous frr docker, which
            # we did not find in this run.
            # use whatever we got
            self.__frr_docker_name = first_frr

        change = True
        if old_frr and not self.__frr_docker_name:
            self._logger.info('No frr docker matching %s',
                              self.__frr_docker_conf_name)
        elif old_frr != self.__frr_docker_name:
            self._logger.info('Setting new frr docker to be %s',
                              self.__frr_docker_name)
        else:
            # no change to frr name
            self._logger.debug('Keeping frr docker to be %s',
                               self.__frr_docker_name)
            change = False

        # keep track of container start time
        if change:
            self.__get_frr_docker_start_time()

    def __get_fullstate_period(self):
        val = ConfKw.DEFAULT_FULLSTATE_PERIOD
        if self._conf.get('netq-agent'):
            val = self._conf.get('netq-agent').get(ConfKw.FULLSTATE_PERIOD, ConfKw.DEFAULT_FULLSTATE_PERIOD)
        if val:
            self._logger.info('Agent FULLSTATE period enabled %s', val)
        else:
            self._logger.info('Agent FULLSTATE disabled')
        return val

    def __is_redis_enabled(self):
        if (self._conf.get('netq-agent') and
                self._conf.get('netq-agent').get(ConfKw.IS_REDIS_ENABLED, ConfKw.DEFAULT_IS_REDIS_ENABLED)):
            self._logger.info('REDIS configuration Enabled')
            return True
        else:
            self._logger.info('REDIS configuration not found')
            return False

    def __get_random_platform(self):
        platform = 0
        if self._conf.get('netq-agent'):
            platform = self._conf.get('netq-agent').get(ConfKw.RANDOM_PLATFORM, 0)
        if platform:
            self._logger.info('Random Platform Enabled %d', platform)
        else:
            self._logger.debug('Random Platform not found')
        return platform

    def __is_opta_enabled(self):
        if (self._conf.get('netq-agent') and
                self._conf.get('netq-agent').get(ConfKw.IS_OPTA_ENABLED, ConfKw.DEFAULT_IS_OPTA_ENABLED)):
            self._logger.info('OPTA configuration Enabled')
            return True
        else:
            self._logger.info('OPTA configuration not found')
            return False

    def __is_gnmi_enabled(self):
        gnmi_enabled = is_gnmi_enabled(self._conf)
        if gnmi_enabled:
            self._logger.info('gNMI configuration Enabled')
            return True
        else:
            self._logger.info('gNMI configuration not found')
            return False

    def __is_cluster_enabled(self):
        if (self._conf.get('netq-agent') and
                self._conf['netq-agent'].get(ConfKw.CLUSTER_SERVERS)):
            self._logger.info('Cluster Enabled %s',
                              self._conf['netq-agent'].get(ConfKw.CLUSTER_SERVERS))
            return True
        else:
            self._logger.info('Cluster Disabled')
            return False

    def __is_intf_stats_enabled(self):
        stats = True
        if (self._conf.get('netq-agent') and
                self._conf['netq-agent'].get('stats')):
            stats = (self._conf['netq-agent']['stats'])
            if not isinstance(stats, bool):
                if str(stats).lower() in ('true', 'yes', 'on'):
                    stats = True
                else:
                    stats = False
        if not stats:
            self._logger.info('Interface stats collection disabled')
            return False

        self._logger.info('Interface stats collection enabled')
        return True

    def __is_frr_running(self):
        routing_dep_func = ['BGP', 'OSPF', 'EVPN']
        cmd = ['/usr/bin/vtysh', '-c', 'show version']
        if self.__frr_docker_enabled:
            if self.__frr_docker_name:
                cmd = self.__modify_frr_docker_cmd(cmd)
            else:
                return False
        try:
            output = utils.subprocess_wrapper(cmd).splitlines()
            rc = output[0].startswith('FRRouting ')
            if rc:
                msg_string = 'Routing daemon check: Active...starting {} data export (if enabled)'.format(routing_dep_func)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ROUTING_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string)
            else:
                msg_string = 'Routing daemon check: Inactive...{} data export impacted (if enabled)'.format(routing_dep_func)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ROUTING_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
            return rc

        except (subprocess.CalledProcessError, OSError):
            self._logger.debug("Failed to execute command to check FRR status")
            msg_string = 'Routing daemon check: Inactive...{} data export impacted (if enabled)'.format(routing_dep_func)
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ROUTING_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                            msg_string)
            return False
        except IndexError:
            self._logger.warn("Unable to find running status of FRR")
            msg_string = 'Routing daemon check: Inactive...{} data export impacted (if enabled)'.format(routing_dep_func)
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ROUTING_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                            msg_string)
            return False

    def __init_frr_docker(self):
        self.__frr_docker_name = None
        self.__frr_docker_conf_name = None
        self.__frr_docker_child_pid = {}
        frr_docker = False

        # we only support FRR docker monitoring
        # on hosts
        platform_str = self.__platform.lower()
        supported = False
        if 'ubuntu' in platform_str:
            supported = True
        elif 'centos' in platform_str:
            supported = True
        elif 'rhel' in platform_str:
            supported = True
        elif 'red hat' in platform_str:
            supported = True

        if (self._conf.get('netq-agent') and
                self._conf['netq-agent'].get('frr_docker')):
            frr_docker = (self._conf['netq-agent']['frr_docker'])
            if not isinstance(frr_docker, bool):
                if str(frr_docker).lower() in ('true', 'yes', 'on'):
                    frr_docker = True
                else:
                    frr_docker = False
        if not frr_docker:
            self._logger.info('FRR docker monitoring disabled')
            return False

        if not supported:
            self._logger.info('FRR docker monitoring not supported on %s',
                              platform_str)
            return False

        frr_name = self.__FRR_DOCKER_DEFAULT
        if self._conf['netq-agent'].get('frr_docker_name'):
            frr_name = str(self._conf['netq-agent']['frr_docker_name'])

        # save the config name
        self.__frr_docker_conf_name = frr_name

        # get the full docker name
        # filter out the k8 POD instances
        self.__get_frr_docker_name()

        self._logger.info('FRR Docker monitoring enabled for %s',
                          self.__frr_docker_conf_name)
        return True

    def __get_platform(self):
        """Get the platform name of this device. Skip if its not Cumulus.
        """
        retcode, cmd_out = self.__run_safe_cmd('/usr/bin/platform-detect')
        # We need to revisit this later after 3.3. The string returned for CL
        # and non CL are different.
        if retcode == 0:
            if 'nvidia' in cmd_out:
                box = '%s, %s, %s' % (Platform.CUMULUS, Platform.MELLANOX, cmd_out)
            else:
                box = '%s, %s' % (Platform.CUMULUS, cmd_out)
        elif os.path.exists('/etc/sonic/sonic_version.yml'):
            with open('/etc/sonic/sonic_version.yml', 'r') as fd:
                name, version = (Platform.SONIC, 'Unknown')
                asic_type = ''
                for line in fd.readlines():
                    if ':' not in line:
                        continue
                    (key, value) = line.strip().split(':', 1)
                    if key.lower() == 'build_version':
                        version = value.strip().strip("'")
                    if key.lower() == 'asic_type':
                        asic_type = value.strip()
                box = '%s, %s, %s' % (Platform.MELLANOX if asic_type != 'vs' else Platform.SONIC_VS, name, version)
        elif os.path.exists('/var/emu_param/fw_info'):
            with open('/var/emu_param/fw_info', 'r') as fd:
                name, version = ('Unknown', 'Unknown')
                for line in fd.readlines():
                    if 'Product Name' in line:
                        name = line.split(' ')[2]
                else: #pylint: disable=useless-else-on-loop
                    #  "Unable to get fw info since the network ports are not configured."
                    if name == 'Unknown':
                        name = 'BlueField'
                box = '%s, %s, %s' % (Platform.BLUEFIELD, name, version)
        elif os.path.exists('/etc/os-release'):
            with open('/etc/os-release', 'r') as fd:
                name, version = ('Unknown', 'Unknown')
                for line in fd.readlines():
                    if '=' not in line:
                        continue
                    (key, value) = line.strip().split('=')
                    if key.lower() == 'name':
                        name = value.strip('"')
                    if key.lower() == 'version':
                        version = value.strip('"')
                box = '%s, %s, %s' % (Platform.LINUX, name, version)
        return box

    def __get_agent_version(self):
        """Get netq-agent version
        """
        if os.path.exists('/usr/bin/rpm'):
            cmd = '/usr/bin/rpm -qi netq-agent'
        elif os.path.exists('/bin/rpm'):
            cmd = '/bin/rpm -qi netq-agent'
        elif os.path.exists('/usr/bin/dpkg'):
            cmd = '/usr/bin/dpkg -s netq-agent'
        else:
            return 'unknown'

        retcode, cmd_out = self.__run_safe_cmd(cmd)

        if retcode != 0:
            return 'unknown'

        return package.process_agent_version(cmd_out,
                                             self.__platform.lower())

    def __pipe_execute(self, pipeline):
        with self.__redispool.item():
            self._logger.debug('Executing pipeline %s of length %d',
                               id(pipeline), len(pipeline.command_stack))
            pipeline.execute()

    def __modify_frr_docker_cmd(self, cmd):
        dockerbin = self.__get_docker_bin()
        frr_cmd = [dockerbin, 'exec', self.__frr_docker_name]
        frr_cmd.extend(cmd)
        return frr_cmd

    def __update_command_timer(self, info):
        epoch = int(time.time())
        self.__timers.setdefault(epoch + info.period, set())
        self.__timers[epoch + info.period].add(info)

    def __process_command(self, info):
        """This updates a key based on new data from executing the command
        :param info: Command object
        :type info: _Command
        """

        retcode = ServiceStatus.OK
        cmd_out = None
        skip_cmd = False

        if 'ecmp-info' in info.key and not self.__is_forwarding_active():
            self.__update_command_timer(info)
            return
        if 'roce' in info.key and not self.__is_forwarding_active():
            self.__update_command_timer(info)
            return

        # For routing commands, check if we need to talk to FRR docker
        routing_daemons = [Daemons.BGPD, Daemons.OSPFD, Daemons.OSPF6D,
                           Daemons.ZEBRA]

        command = info.command
        if info.service in routing_daemons:
            if 'evpn' in info.key and not self.__is_frr_running():
                self.__update_command_timer(info)
                skip_cmd = True
            if self.__frr_docker_enabled:
                if self.__frr_docker_name:
                    command = self.__modify_frr_docker_cmd(command)
                else:
                    # no frr docker found
                    self.__update_command_timer(info)
                    return

        # Execute the command:
        if command and not skip_cmd:
            retcode, cmd_out = self.__run_safe_cmd(command, timeout=info.timeout)

        if retcode:
            if is_sonic() and cmd_out and re.search('Container (.+?) is not running', cmd_out):
                pass    # Container not running. Proceed ahead by invoking parser to perform object deletions
            elif retcode == 4 and 'smartctl' in info.command:
                pass    # Proceed ahead by invoking parser of smartctl
            else:
                info.errcnt += 1
                if info.errcnt < 3:
                    # ignore upto 3 consecutive errors
                    self.__update_command_timer(info)
                    return

            # we are going to let callbacks handle the error

            # FRR docker needs special handling
            # mark the docker name as none so that
            # timer handler will try and find a container
            if info.service in routing_daemons:
                if self.__frr_docker_enabled:
                    self.__frr_docker_name = None
        else:
            info.errcnt = 0

        # Update monitoring status
        status = retcode
        svcs = Services.memmodel()
        svc = svcs.query.get(hostname=HOSTNAME, name=info.service)
        newsvc = Services()

        if svc and info.service != Daemons.MISC and retcode != ServiceStatus.OK:
            svc_fail = self.__has_service_failed(info.service)
            if retcode > ServiceStatus.ERROR or svc_fail:
                status = ServiceStatus.FAIL
            else:
                status = ServiceStatus.ERROR

        if svc and info.service != Daemons.MISC and not info.callback:
            newsvc.copy(svc)
            newsvc.is_monitored = 1
            if newsvc.is_active:
                newsvc.status = status
            else:
                newsvc.status = ServiceStatus.NA
            newsvc.save()

        if not svc and info.always:
            newsvc.name = info.service
            newsvc.is_enabled = True
            newsvc.is_active = True
            newsvc.is_monitored = True
            if retcode == ServiceStatus.OK:
                status = ServiceStatus.OK
            elif retcode > ServiceStatus.ERROR:
                status = ServiceStatus.FAIL
            else:
                status = ServiceStatus.ERROR
            newsvc.status = status
            newsvc.pid = 'n/a'
            newsvc.start_time = 0.0
            newsvc.save()

        if not cmd_out:
            # I'm assuming that the parsers can all catch non-JSON input and
            # parse out an error
            cmd_out = ' '

        is_full_state = self.__is_full_state(info)

        # Don't push the data out unless it has changed
        if cmd_out:
            if info.callback is not None:
                try:
                    retcode = info.callback(
                        cmd_out, status, platform=self.__platform,
                        ports=copy.copy(self.__ports), fullstate=is_full_state, pool=self._pool,
                        random=self.__random_platform_info, optical_ports=self.__optical_ports,
                        period=self.__config_commands.get_command_period(info.key), logger=self._logger,
                        subscribe_manager=self.__subscribe_manager
                    )
                except subprocess.CalledProcessError:
                    self._logger.info("failed to execute command: {} ".format(info.command))
                except Exception as ex:  # pylint: disable=broad-except
                    self._logger.exception('Failed to execute command %r %r' %
                                           (info.command, ex))
                if svc and info.service != Daemons.MISC:
                    newsvc.copy(svc)
                    newsvc.is_monitored = 1
                    newsvc.save()
            else:
                self.__config_commands.process_cmd_out(cmd_out, info)

        self.__update_command_timer(info)

    def __is_full_state(self, info):

        if not self.__fullstate_period:
            return False
        if info.fullstateCnt >= info.fullstate:
            info.fullstateCnt = 0
            return True
        else:
            info.fullstateCnt = info.fullstateCnt + info.period
            return False

    def __process_counters(self):
        """Update Top talkers and such
        Push out counters per port so they can be sorted by redis.
        """

    def __check_ntp_status(self, ntp_sync=''):
        """Update ntp status
        """

        ntp_out = None
        retcode = -1
        ntp_cmd = None
        use_timedatectl = False
        use_chronyc = False

        if Platform.CUMULUS in self.__platform:
            ntp_svc = RunningService.NTP
            chrony_svc = RunningService.CHRONY
            if self.__is_service_running(ntp_svc):
                ntp_cmd = '/usr/bin/ntpq -p'
            elif self.__is_service_running(ntp_svc, vrf='default'):
                ntp_cmd = '/usr/bin/vrf task exec default ntpq -p'
            elif self.__is_service_running(ntp_svc, vrf='mgmt'):
                ntp_cmd = '/usr/bin/vrf task exec mgmt ntpq -p'
            elif self.__is_service_running(chrony_svc):
                if os.path.exists('/usr/bin/chronyc'):
                    ntp_cmd = '/usr/bin/chronyc sources'
                use_chronyc = True
            elif self.__is_service_running(chrony_svc, vrf='mgmt'):
                if os.path.exists('/usr/bin/chronyc'):
                    ntp_cmd = '/usr/bin/vrf task exec mgmt chronyc sources'
                use_chronyc = True

        else:
            # on non-CL platforms, we can support
            # ntpd, chronyd, timedatectl (we check in that order)
            if 'centos' in self.__platform.lower():
                ntp_svc = CentosRunningService.NTP
                chrony_svc = CentosRunningService.CHRONY
            elif 'rhel' in self.__platform.lower():
                ntp_svc = RhelRunningService.NTP
                chrony_svc = RhelRunningService.CHRONY
            elif 'red hat' in self.__platform.lower():
                ntp_svc = RhelRunningService.NTP
                chrony_svc = RhelRunningService.CHRONY
            else:
                ntp_svc = RunningService.NTP
                chrony_svc = RunningService.CHRONY
            if self.__is_service_running(ntp_svc):
                if os.path.exists('/usr/bin/ntpq'):
                    ntp_cmd = '/usr/bin/ntpq -p'
                elif os.path.exists('/usr/sbin/ntpq'):
                    ntp_cmd = '/usr/sbin/ntpq -p'
                elif os.path.exists('/sbin/ntpq'):
                    ntp_cmd = '/sbin/ntpq -p'
            elif (self.__is_service_running(chrony_svc) or
                  self.__is_service_running(chrony_svc, vrf='mgmt')):
                if os.path.exists('/usr/bin/chronyc'):
                    ntp_cmd = '/usr/bin/chronyc sources'
                elif os.path.exists('/bin/chronyc'):
                    ntp_cmd = '/bin/chronyc sources'
                use_chronyc = True
            elif os.path.exists('/bin/timedatectl'):
                ntp_cmd = '/bin/timedatectl status'
                use_timedatectl = True
            elif os.path.exists('/usr/bin/timedatectl'):
                ntp_cmd = '/usr/bin/timedatectl status'
                use_timedatectl = True

        if ntp_cmd:
            retcode, ntp_out = self.__run_safe_cmd(ntp_cmd, timeout=45)

        if use_timedatectl:
            return ntp.process_timedatectl(ntp_out)
        elif use_chronyc:
            return ntp.process_chronyc(ntp_out)
        else:
            return ntp.process_ntpq(ntp_out, self._logger)

    def __check_services(self, fullstate):
        """Update active/enabled/monitoring services status"""

        if self.__platform == Platform.UNKNOWN:
            return
        base_cumulus_services = []
        hardnode_services = []
        initd_services = []
        initd_svc_cmd = None
        initd_svc_cmd_out = None
        base_cumulus_extended_list = []
        overriden_extended_list = []

        if Platform.CUMULUS in self.__platform:
            base_cumulus_services = [
                RunningService.CLAGD,
                RunningService.MSTPD,
                RunningService.LLDPD,
                RunningService.FRR,
                RunningService.NETQD,
                RunningService.NETQAGENT,
                RunningService.NVUED,
                RunningService.PWMD,
                RunningService.SMOND,
                RunningService.QUAGGA
            ]
            overriden_extended_list = self.get_extended_service_list()
            base_cumulus_extended_list = overriden_extended_list if overriden_extended_list else [
                RunningService.NEIGHMGRD,
                RunningService.PTMD,
                RunningService.PTP4L,
                RunningService.DOCKER,
                RunningService.SNMPD,
                RunningService.RSYSLOG,
                RunningService.SSH,
                RunningService.DHCP_RELAY
            ]
            if Platform.VX not in self.__platform:
                hardnode_services = [
                    RunningService.WATCHDOG,
                    RunningService.SWITCHD,
                    RunningService.MLX_SDK
                ]
            base_linux_services = [
                RunningService.NTP
            ]
        elif PlatformType.SONIC in self.__platform:
            base_linux_services = [
                SonicRunningService.MGMT_FWK,
                SonicRunningService.TELEMETRY,
                SonicRunningService.SNMP,
                SonicRunningService.RADV,
                SonicRunningService.DHCP_RELAY,
                SonicRunningService.PMON,
                SonicRunningService.LLDP,
                SonicRunningService.SYNCD,
                SonicRunningService.TEAMD,
                SonicRunningService.SWSS,
                SonicRunningService.BGP,
                SonicRunningService.DATABASE,
                RunningService.NTP,
                RunningService.SSH,
                RunningService.RSYSLOG,
                RunningService.DOCKER,
                RunningService.NETQD,
                RunningService.NETQAGENT
            ]
        else:  # Platform.LINUX
            base_cumulus_services = [
                RunningService.NETQD,
                RunningService.NETQAGENT,
                RunningService.NETQNOTIFIER
            ]
            if 'ubuntu' in self.__platform.lower():
                # Non systemd services return UnitFileState=bad
                # Check for UnitFilePreset state
                initd_services = [
                    RunningService.NTP,
                ]

                prop = 'Names,ActiveState,UnitFilePreset,MainPID,ActiveEnterTimestamp'
                initd_svc_cmd = (
                    '%s show -p %s %s' % (
                        '/bin/systemctl', prop, ' '.join(initd_services)))

                base_linux_services = [
                    RunningService.CHRONY,
                    RunningService.SSH,
                    RunningService.RSYSLOG,
                    RunningService.DOCKER,
                    RunningService.LLDPD,
                    RunningService.QUAGGA,  # For routing on host
                    RunningService.FRR  # For routing on host
                ]
            elif 'centos' in self.__platform.lower():
                # Service names are different on different platform
                # e.g ntp service on centos is refered as - ntpd.service
                base_linux_services = [
                    CentosRunningService.NTP,
                    CentosRunningService.CHRONY,
                    RunningService.SSH,
                    RunningService.RSYSLOG,
                    RunningService.DOCKER,
                    RunningService.QUAGGA,  # For routing on host
                    RunningService.FRR  # For routing on host
                ]
            elif 'red hat' in self.__platform.lower():
                base_linux_services = [
                    RhelRunningService.NTP,
                    RhelRunningService.CHRONY,
                    RunningService.SSH,
                    RunningService.RSYSLOG,
                    RunningService.DOCKER,
                    RunningService.QUAGGA,  # For routing on host
                    RunningService.FRR  # For routing on host
                ]
            elif 'rhel' in self.__platform.lower():
                base_linux_services = [
                    RhelRunningService.NTP,
                    RhelRunningService.CHRONY,
                    RunningService.SSH,
                    RunningService.RSYSLOG,
                    RunningService.DOCKER,
                    RunningService.QUAGGA,  # For routing on host
                    RunningService.FRR  # For routing on host
                ]
            else:
                base_linux_services = [
                    RunningService.NTP,
                    RunningService.SSH,
                    RunningService.RSYSLOG,
                    RunningService.DOCKER,
                    RunningService.QUAGGA,  # For routing on host
                    RunningService.FRR  # For routing on host
                ]

        all_svcs = ' '.join(base_cumulus_services +
                            base_linux_services +
                            hardnode_services +
                            base_cumulus_extended_list)

        if self.__frr_docker_enabled:
            # make sure we remove FRR from systemctl request
            try:
                all_svcs = all_svcs.replace(RunningService.FRR, '')
                all_svcs = all_svcs.replace(RunningService.QUAGGA, '')
            except Exception:
                pass

        if not os.path.exists('/bin/systemctl'):
            self._logger.error('Non systemctl systems not supported')
            return
        prop = 'Names,ActiveState,UnitFileState,MainPID,ActiveEnterTimestamp,InactiveEnterTimestamp'
        cmd = '%s show -p %s %s *@*' % ('/bin/systemctl', prop, all_svcs)
        retcode, cmd_out = self.__run_safe_cmd(cmd)

        if initd_svc_cmd:
            initd_svc_retcode, initd_svc_cmd_out = self.__run_safe_cmd(initd_svc_cmd)
            if initd_svc_retcode != 0:
                initd_svc_cmd_out = None

        if retcode == 0 and cmd_out:
            if initd_svc_cmd_out:
                cmd_out = cmd_out + '\n' + initd_svc_cmd_out
            try:
                if PlatformType.SONIC in self.__platform:
                    sonic_process_systemctl(cmd_out, fullstate, self._logger, self.__platform)
                else:
                    process_systemctl(cmd_out, fullstate, self._logger, self.__platform)
            except Exception as ex:  # pylint: disable=broad-except
                self._logger.info('Failed to process systemctl output. %s', ex)

    ''' Check if user has overriden any extended services for monitoring'''
    def get_extended_service_list(self):
        overriden_extended_list = []
        rc, current_config = load_configuration(DefaultPaths.CONF_FILE)
        if rc:
            self._logger.error('Could not load the {} file, Hence falling back to default extended list.'.format(DefaultPaths.CONF_FILE))
        if not current_config:
            current_config = {ConfKw.NETQ_AGENT: dict()}
        current_config.setdefault(ConfKw.NETQ_AGENT, {})
        services = current_config.get(ConfKw.NETQ_AGENT, {}).get(ConfKw.NETQ_SERVICES)
        if services:
            overriden_extended_list = services.split(",")
        return overriden_extended_list

    def __update_link_masters(self, master_ifindex, master_ifname, master_kind):
        '''Update the master ifindex for each of the saved links'''

        if master_ifindex not in self.__to_update_master_links:
            return

        for ifname in self.__to_update_master_links[master_ifindex]:
            m_link = Link.memmodel()
            old_entry = m_link.query.get(hostname=HOSTNAME, ifname=ifname)
            if old_entry:
                if (old_entry.active and (old_entry.master ==
                                          str(master_ifindex))):
                    link = Link()
                    link.copy(old_entry)
                    link.master = master_ifname
                    if master_kind == LinkType.VRF:
                        link.vrf = master_ifname
                    link.timestamp = old_entry.timestamp
                    del_old_entry = Link()
                    del_old_entry.copy(old_entry)
                    del_old_entry.delete()
                    link.save()

        self.__to_update_master_links.pop(master_ifindex, None)

    def __update_server_info(self):
        """ Update redis server pid
            check for connected_slaves
            if mismatch detected, set AGENT_PAUSE=True
        """
        try:
            new_server_info = server_info()
        except Exception as ex:
            self.__server_info_fail_cnt = self.__server_info_fail_cnt + 1
            self._logger.info('Could not get Redis server info (fail cnt %d). Error: %s',
                              self.__server_info_fail_cnt, ex)
            if self.__server_info_fail_cnt > self.__MAX_SERVER_INFO_FAIL_CNT:
                self._logger.info('Redis server info failed after %d attempts',
                                  self.__server_info_fail_cnt)
                # self.__pause_agent()
                return False
            return True

        new_role = new_server_info.get("role", None)
        new_connected_slaves = new_server_info.get("connected_slaves", 0)

        if new_role is None:
            self._logger.info('Redis role could not be determined')
            return False

        if new_role != 'master':
            self._logger.info('Redis server detected new master %s:%s',
                              new_server_info.get("master_host", None),
                              new_server_info.get("master_port", None))
            # self.__pause_agent()
            return False

        if is_backend_sentinel_aware(self._conf[ConfKw.BACKEND]):
            if new_connected_slaves == 0:
                if not self.__AGENT_PAUSE_FIRST_RUN:
                    self._logger.info('Redis master Detected but no slaves')
                # self.__pause_agent()
                return False

        if not self.__server_info:
            self.__server_info = new_server_info
            self._logger.info('Redis server initialized run_id [%s]',
                              self.__server_info.get("run_id", None))
            self.__server_info_fail_cnt = 0
            return True

        if (self.__server_info.get("run_id", None) !=
                new_server_info.get("run_id", None)):
            self._logger.info('Redis server run_id old [%s] != new [%s]',
                              self.__server_info.get("run_id", None),
                              new_server_info.get("run_id", None))
            # if have HA setup, ignore run_id changes
            if not is_backend_sentinel_aware(self._conf[ConfKw.BACKEND]):
                # if agent already in paused state - ignore
                if self.__AGENT_PAUSE is False:
                    # clear info - so we can start afresh
                    self.__server_info = {}
                    # self.__pause_agent()
                    return False

        # assign new server info
        self.__server_info = new_server_info
        if self.__server_info_fail_cnt:
            self._logger.info('Redis server info updated run_id [%s]',
                              self.__server_info.get("run_id", None))
            self.__server_info_fail_cnt = 0
        return True

    def __cleanup_nlmonitor_thread(self):

        if self.__nlmonitor_gt is not None:
            self._logger.info('Terminating netlink monitor thread')
            eventlet.kill(self.__nlmonitor_gt, eventlet.StopServe)
            self.__nlmonitor_gt = None

    def __flush_obj_cache(self, nl_only=False):
        try:
            for cln, cls in inspect.getmembers(netq_lib.orm.redisdb.models):
                if cln in self.__NON_CLS_MODELS:
                    continue
                if nl_only and cln not in self.__NL_CLS_MODELS:
                    continue
                if inspect.isclass(cls) and hasattr(cls, 'memmodel'):
                    mcls = cls.memmodel()
                    if mcls:
                        mcls.flush()
        except Exception as ex:
            self._logger.info('Flush failed. Error: %s', ex)

    def __cleanup_threads(self):

        self.__cleanup_nlmonitor_thread()

        # Flush the pipeline
        self._logger.debug('Flushing the pipeline')
        for entry in list(self.__redisthreads):
            entry.kill()
        self.__redisthreads = set()

        self.__flush_obj_cache()

    def __delete_netlink_from_redis(self):
        """ Delete netlink keys
        """
        self._logger.info('Deleting netlink keys from REDIS')
        try:
            for cln, cls in inspect.getmembers(netq_lib.orm.redisdb.models):
                if cln not in self.__NON_CLS_MODELS:
                    continue
                if not inspect.isclass(cls):
                    continue
                try:
                    if hasattr(cls, 'key_fmt'):
                        key = cls.key_fmt()
                        if not key:
                            continue
                    else:
                        continue
                except Exception:
                    continue
        except Exception:
            # ignore all exceptions
            pass

        return

    @FullState("Link")
    def __netlink_get_all_links(self):
        """ Get netlink link info
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_LINK)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._pool.spawn(self.__nlmonitor.get_all_links).wait()
            self.__nlmonitor.set_mcast_group('link')

    @FullState('Address')
    def __netlink_get_all_addresses(self):
        """ Get netlink addr
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_ADDR)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._pool.spawn(self.__nlmonitor.get_all_addresses).wait()
            self.__nlmonitor.set_mcast_group('addr')

    @FullState('Neighbor')
    def __netlink_get_all_neighbors(self):
        """ Get netlink neigh
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_NEIGHBOR)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._pool.spawn(self.__nlmonitor.get_all_neighbors).wait()
            self.__nlmonitor.set_mcast_group('neigh')

    @FullState('Routes')
    def __netlink_get_all_routes(self):
        """ Get netlink routes
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_ROUTE)
            self.__netlink_cmd_client.send_command(cmd)
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_VXLANREMOTEDEST)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._pool.spawn(self.__nlmonitor.get_all_routes).wait()
            self.__nlmonitor.set_mcast_group('routes')

    @FullState('MacFdb')
    def __netlink_get_all_macs(self):
        """ Get netlink mac
        neigh mcast group should already be set
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_MACFDB)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._pool.spawn(self.__nlmonitor.get_all_macs).wait()

    @FullState('Netconf')
    def __netlink_get_all_netconf(self):
        """ Get netlink netconf
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_NETCONF)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._pool.spawn(self.__nlmonitor.get_all_netconf).wait()
            self.__nlmonitor.set_mcast_group('netconf')

    @FullState('Nexthop')
    def __netlink_get_all_nexthop(self):
        """ Get netlink nexthop
        """
        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.GET_FULL_STATE_NEXTHOP)
            self.__netlink_cmd_client.send_command(cmd)
        else:
            self._logger.warn('Libnl is not enabled')

    def __reinitialize_nl(self):
        """ Reinitializing netlink subsystem
        """

        self._logger.info('Netlink - (Re)initializing (%d)', self.__nl_poll_retry)

        self.__cleanup_nlmonitor_thread()

        # Flush the netlink cache
        self.__nlmonitor.reset()

        if self.__redis_enabled:
            self.__delete_netlink_from_redis()

        if self.__opta_enabled:
            self.__flush_obj_cache()

        # Rebind netlink socket
        self._logger.info('Binding netlink socket')
        try:
            self.__nlmonitor.bind()
        except RuntimeError as ex:
            self._logger.warn('Fail Binding netlink socket %s', ex)
            self.__NETLINK_RESTART = True
            self.__nlmonitor.socket.close()
            os.kill(self.__grpc_stream_process.pid, 9)
            return

        self.__nlmonitor_gt = (
            self._pool.spawn(self._serve,
                             self.__nlmonitor.socket,
                             self.__nlmonitor.handle_netlink_msg,
                             bufsize=netlink.Netlink.NLSOCK_BYTES)
        )
        self.__nlmonitor_gt.link(self.__stop_monitor_checker)

        # Start the netlink handler threads
        intf_gt, nlevent_gt = self.__nlmonitor.run()
        intf_gt.link(self._stop_checker)
        nlevent_gt.link(self._stop_checker)

        # Repopulating netlink
        self.__ifname_by_index = {}
        self.__vxlan_vlan_map = {}
        try:
            self.__netlink_get_all_links()
            self.__netlink_get_all_addresses()
            self.__netlink_get_all_routes()
            self.__netlink_get_all_neighbors()
            if not is_sonic():
                self.__netlink_get_all_macs()
            self.__netlink_get_all_netconf()
            if configuration.is_cl4():
                self.__netlink_get_all_nexthop()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            self._logger.info("Greenlet exit called")
            self.__nlmonitor.reset()
            if self.__nlmonitor_gt:
                self.__nlmonitor_gt.kill()
            pass
        except (ConnectionRefusedError, IOError):
            self._logger.info("Netlink Process initializing...")
            self.__nlmonitor.reset()
            if self.__nlmonitor_gt:
                self.__nlmonitor_gt.kill()
            pass
        except Exception as ex:  # pylint: disable=broad-except
            self._logger.info('netlink get_all_links failed %s', ex)
            if ex.errno == errno.ENOBUFS: # pylint: disable=no-member
                self.__NETLINK_RESTART = True
            else:
                self.__pause_agent()
            return

        self.__NETLINK_INIT_DONE = True

        if not self._is_libnl_enabled:
            self.__send_netlink_fullstate()

        self._logger.info('netlink init done')

    def __get_mem_objects(self, objectname, generator=True):
        try:
            mcls = eval(objectname).memmodel() # pylint: disable=eval-used
        except Exception:
            self._logger.error("Unable to find obj {}".format(objectname))
            return None
        if generator:
            return mcls.query.all()
        else:
            all_objs = [obj for obj in mcls.query.all()]
        return all_objs

    def __send_netlink_fullstate(self):
        message_types = ['Link', 'Address', 'Netconf', 'Route', 'Neighbor', 'MacFdb', 'VxlanRemoteDest']
        for message_type in message_types:
            for obj in self.__get_mem_objects(message_type):
                self.full_state_object.push_message_queue(obj)
            self.full_state_object.flush_fullstate(message_type)

    def __send_netlink_libnl_fullstate(self):
        self._logger.info("Netlink Libnl: Initiating fullstate")
        try:
            self.__netlink_get_all_links()
            self.__netlink_get_all_addresses()
            self.__netlink_get_all_routes()
            time.sleep(1)  # add delay after routes
            self.__netlink_get_all_neighbors()
            if not is_sonic():
                self.__netlink_get_all_macs()
            self.__netlink_get_all_netconf()
            time.sleep(1)  # add delay after macs
            if configuration.is_cl4():
                self.__netlink_get_all_nexthop()
            self.__netlink_libnl_retry_timer = 0
        except (ConnectionRefusedError, IOError):
            self._logger.info("Netlink Process initializing...")
            self.__netlink_libnl_retry_timer = 30
            pass
        except Exception as ex:  # pylint: disable=broad-except
            self._logger.info('netlink get_all_links failed %s', ex)
            self.__netlink_libnl_retry_timer = 30
            self.__pause_agent()
            return


    def _proc_link_msg(self, json_input):
        if isinstance(json_input, list):
            json_out = json_input[0]
        else:
            json_out = json_input
        kind = json_out['kind']
        ifname = json_out['ifname']
        deleted = json_out['deleted']
        if not deleted:
            if kind in (LinkType.SWP, LinkType.ETH):
                self.__ports.add(ifname)
        else:
            if kind in (LinkType.SWP, LinkType.ETH):
                self.__ports.discard(ifname)


    def _proc_netlink_stats(self, json_output):

        agent_stat_string = json.loads(json_output) # pylint: disable=no-member
        agent_stats_object = AgentStatsMon.get_instance()

        #msg_count_stats = agent_stats_object.message_count
        msg_count_stats = {}

        msg_count_stats['total_msg_count'] = msg_count_stats.get('total_msg_count', 0) + agent_stat_string.get("total_msg_count", 0)
        msg_count_stats['total_msg_bytes'] = msg_count_stats.get('total_msg_bytes', 0) + agent_stat_string.get("total_msg_bytes", 0)
        msg_count_stats['full_state_count'] = msg_count_stats.get('full_state_count', 0) + agent_stat_string.get("full_state_count", 0)
        msg_count_stats['partial_state_count'] = msg_count_stats.get('partial_state_count', 0) + agent_stat_string.get("partial_state_count", 0)
        msg_count_stats['netlink_msg_count'] = msg_count_stats.get('netlink_msg_count', 0) + agent_stat_string.get("netlink_msg_count", 0)

        msg_type_count = agent_stat_string.get("message_type_count")
        agent_msg_type = agent_stats_object.msg_type_count
        if agent_msg_type is None:
            agent_msg_type = {}
        if NetlinkMsgType.NETLINK_MSG_TYPE_LINK.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_LINK.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_LINK.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_LINK.value]
        if NetlinkMsgType.NETLINK_MSG_TYPE_ROUTE.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_ROUTE.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_ROUTE.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_ROUTE.value]
        if NetlinkMsgType.NETLINK_MSG_TYPE_ADDRESS.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_ADDRESS.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_ADDRESS.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_ADDRESS.value]
        if NetlinkMsgType.NETLINK_MSG_TYPE_MACFDB.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_MACFDB.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_MACFDB.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_MACFDB.value]
        if NetlinkMsgType.NETLINK_MSG_TYPE_NETCONF.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_NETCONF.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_NETCONF.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_NETCONF.value]
        if NetlinkMsgType.NETLINK_MSG_TYPE_NEIGHBOR.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_NEIGHBOR.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_NEIGHBOR.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_NEIGHBOR.value]
        if NetlinkMsgType.NETLINK_MSG_TYPE_VXLANREMOTEDEST.value in msg_type_count:
            agent_msg_type[NetlinkMsgType.NETLINK_MSG_TYPE_VXLANREMOTEDEST.value] = agent_msg_type.get(NetlinkMsgType.NETLINK_MSG_TYPE_VXLANREMOTEDEST.value, 0) + msg_type_count[NetlinkMsgType.NETLINK_MSG_TYPE_VXLANREMOTEDEST.value]

        agent_stats_object.msg_type_count = agent_msg_type
        #agent_stats_object.msg_type_count.update(agent_stat_string.get("message_type_count", None))

        agent_stats_object.message_count = msg_count_stats


    def __receive_netlink_info(self):
        if NetQPriorityQueue.agent_PQ.len_all() > 0:
            for priority, message_type, state, msg in NetQPriorityQueue.agent_PQ.get_buffer(1):
                if message_type == 'Link':
                    self._proc_link_msg(msg.data)
                if message_type == "Stats":
                    self._proc_netlink_stats(msg.data)


    def __pause_agent(self):
        self._logger.info('Agent is Paused')
        agent_stats_obj = AgentStatsMon.get_instance()
        agent_stats_obj.agent_pause_count += 1
        self.__AGENT_PAUSE = True

    def __init_netlink_thread(self):
        if self.__nlinit_gt:
            self._logger.info("killing netlink thread")
            self.__nlinit_gt.kill()
            self.__nlinit_gt = None

        self.__NETLINK_INIT_DONE = False
        self.__nlinit_gt = _Thread.spawn(int(time.time()),
                                         self._pool,
                                         self.__reinitialize_nl)
        self.__nlinit_gt.link(self.__stop_thread_checker)
        self.__redisthreads.add(self.__nlinit_gt)
        self.__NETLINK_RESTART = False
        self.__netlink_init_time = time.time()

    def __redis_connect_init(self):
        try:
            server, port = set_connection_settings(self._conf[ConfKw.BACKEND],
                                                   agent=True, logger=self._logger)
            if not self.__AGENT_PAUSE_FIRST_RUN:
                self._logger.info('Redis server connect OK {}:{}'.format(server, port))
        except Exception:
            if not self.__AGENT_PAUSE_FIRST_RUN:
                self._logger.info('Redis server connect failed')
            self.__AGENT_PAUSE_FIRST_RUN = True
            return False

        try:
            ping_server()
            if not self.__AGENT_PAUSE_FIRST_RUN:
                self._logger.info('Redis Ping to server succeeded.')
        except Exception as ex:
            if not self.__AGENT_PAUSE_FIRST_RUN:
                self._logger.info('Redis Ping failed to server. Error: %s', ex)
            # use this for limiting logs
            self.__AGENT_PAUSE_FIRST_RUN = True
            # in non-HA init the server-info
            # in HA keep the server-info to perform connected_slave checks
            if not is_backend_sentinel_aware(self._conf[ConfKw.BACKEND]):
                self.__server_info = {}
            return False

        # check server info
        server_updated = self.__update_server_info()

        if not server_updated:
            if not self.__AGENT_PAUSE_FIRST_RUN:
                self._logger.info('Could not update server info')
            # use this for limiting logs
            self.__AGENT_PAUSE_FIRST_RUN = True
            return False

        return True

    def __get_grpc_connection_state(self):
        if not NetQPriorityQueue.bp_PQ.is_empty():
            for priority, message_type, state, msg in NetQPriorityQueue.bp_PQ.get_buffer(0):
                if self.__opta_enabled and message_type == MC.GRPC_CLIENT_OPTA:
                    self.__grpc_state = state
                    self._logger.info("Agent states: OPTA Connectivity "
                                      "state %s", state)
                if self.__gnmi_enabled and message_type == MC.GRPC_CLIENT_GNMI:
                    self.__gnmi_grpc_state = state
                    self._logger.info("Agent states: gNMI Connectivity "
                                      "state %s", state)
        return self.__grpc_state, self.__gnmi_grpc_state

    def __reinitialize_db(self):
        """ Reinitializing the database. This is called
        a. on startup
        b. when pipeline execute fails
        c. the bridge monitor or netlink monitor greenthread crash
        """

        with self.__reinitpool.item():
            # Check if hostname is still 'Cumulus'
            if utils.is_hostname_default(HOSTNAME):
                if self.__prev_hostname == 'unknown':
                    self._logger.error("Agent paused as hostname is default: {}."\
                            "Please change hostname and restart agent".format(HOSTNAME))
                    self.__prev_hostname = HOSTNAME
                return

            self.__hostname = HOSTNAME
            self.__prev_hostname = 'unknown'

            if not self.__AGENT_PAUSE_FIRST_RUN:
                self._logger.info('Agent - (re)initializing on %s', HOSTNAME)

            self.__cleanup_threads()

            # Terminate docker monitor
            self.__docker.halt()

            # Terminate Kubernetes monitor
            self.__kubernetes.halt()

            # see if we can connect with redis
            if self.__redis_enabled:
                self.__redis_connect_init()

            grpc_state, gnmi_state = self.__get_grpc_connection_state()
            if self.__opta_enabled:
                if not grpc_state:
                    self.__pause_agent()
                    msg_string = 'gRPC connection state invalid... Agent is paused'
                    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_CONN_STATUS,
                                                    NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                    msg_string)
                    self._logger.info(msg_string)
                    return
                else:
                    self.__AGENT_PAUSE = False
                    msg_string = 'gRPC connection state checked... OK'
                    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_CONN_STATUS,
                                                    NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                    msg_string)
                    self._logger.debug(msg_string)
            else:
                msg_string = 'OPTA not enabled: gRPC Connection state check skipped'
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_CONN_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_NA,
                                                msg_string)

            if self.__gnmi_enabled:
                if not gnmi_state:
                    self.__GNMI_AGENT_PAUSE = True
                    msg_string = 'gNMI connection state invalid... Agent is paused'
                    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GNMI_CONN_STATUS,
                                                    NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                    msg_string)

                    self._logger.info(msg_string)
                else:
                    self.__GNMI_AGENT_PAUSE = False
                    msg_string = 'gNMI connection state checked... OK'
                    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GNMI_CONN_STATUS,
                                                    NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                    msg_string)
                    self._logger.debug(msg_string)
            else:
                msg_string = 'gNMI not enabled: Connection state check skipped'
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GNMI_CONN_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_NA,
                                                msg_string)


            if (self.__opta_enabled and
                    (NetQPriorityQueue.PQ.is_full() or
                     not NetQPriorityQueue.bp_PQ.is_empty())):
                if NetQPriorityQueue.PQ.is_full():
                    rsn = "GRPC priority Q FULL"
                    self.__grpc_q_full = True
                elif not NetQPriorityQueue.bp_PQ.is_empty():
                    rsn = "GRPC backpressure"
                    NetQPriorityQueue.bp_PQ.empty_queues()
                if not self.__AGENT_PAUSE_FIRST_RUN:
                    self._logger.info("Agent pause due to %s", rsn)
                # use this for limiting logs
                self.__AGENT_PAUSE_FIRST_RUN = True
                return
            elif self.__grpc_q_full:
                self._logger.info("Agent grpc priority Q RESUME len %d",
                                  NetQPriorityQueue.PQ.len(1))
                self.__grpc_q_full = False

            # redis ping , server info checks passed
            self.__AGENT_PAUSE_FIRST_RUN = False

            self._logger.info('Reinitializing the database')

            # Flush docker keys from database
            # self.__docker.flushkeys()

            # kick off netlink init thread
            self.__nl_poll_retry = 0
            if not self._is_libnl_enabled and self._is_netlink_enabled:
                self.__init_netlink_thread()  # this thread is for py-netlink

            # update platform info like mem, cpu etc
            self.__update_platform_info(False)

            # Run Docker monitor
            self.__docker.start()

            # Run Kubernetes monitor
            self.__kubernetes.start()

            # looks good!
            self.__AGENT_PAUSE = False

    def __rx_rtm_newnetconf(self, info):
        self._logger.debug('RX RTM_NEWNETCONF for ip%s ifindex %s: %d',
                           'v6' if info.is_ipv6 else 'v4', info.ifindex,
                           info.forwarding)
        netconf = Netconf()
        NetlinkCopy.copy_netconf(netconf, info, self.__ifname_by_index)
        if netconf.ifname == 'None':
            return info.ifindex
        netconf.save()

        return info.ifindex

    def __rx_rtm_newaddr(self, info):
        """Adds a new address to the DB.
        :param info: address object
        :type info: netlink.Address
        """
        self._logger.debug('RX RTM_NEWADDR for ip_address %s, ifindex %s',
                           '/'.join((info.prefix, str(info.mask))),
                           info.ifindex)
        addr = Address()
        NetlinkCopy.copy_address(addr, info, self.__ifname_by_index)
        table = self.__link_table_map.get(info.ifindex, RouteTable.DEFAULT)
        addr.vrf = self.__link_vrf_map.get(table, 'default')
        self._logger.debug('RX RTM_NEWADDR for %s, ifindex %s, VRF %s, to DB',
                           '/'.join((info.prefix, str(info.mask))), info.ifindex,
                           addr.vrf)
        if addr.ifname == 'None':
            # save this address - not yet ready to go into Db
            # ifindex can have multiple ips (v4, v6)
            if not self.__addr_ifindex_map.get(info.ifindex):
                self.__addr_ifindex_map[info.ifindex] = {}
            self.__addr_ifindex_map[info.ifindex][info.prefix] = addr
            return info.ifindex
        addr.save()

        return info.ifindex

    def __rx_rtm_deladdr(self, info):
        """Deletes an address from the DB.
        :param info: address object
        :type info: netlink.Address
        """
        self._logger.debug('RX RTM_DELADDR for ip_address %s, ifindex %s',
                           '/'.join((info.prefix, str(info.mask))),
                           info.ifindex)
        addr = Address()
        NetlinkCopy.copy_address(addr, info, self.__ifname_by_index)
        table = self.__link_table_map.get(info.ifindex, RouteTable.DEFAULT)
        addr.vrf = self.__link_vrf_map.get(table, 'default')
        if self.__addr_ifindex_map.get(info.ifindex):
            self.__addr_ifindex_map[info.ifindex].pop(info.prefix)
        self._logger.debug('Deleting address %r from DB', addr)
        addr.delete()
        return info.ifindex

    @staticmethod
    def set_link_kind(link, info):
        if link.kind == '':
            if info.ifname == 'lo':
                link.kind = LinkType.LOOPBACK
            elif (info.ifname.startswith('eth') or
                  info.ifname.startswith('usb')):
                link.kind = LinkType.ETH
            elif (info.ifname == 'swid0_eth' or
                  info.ifname.startswith('mirror')):
                # Melllanox chip internal port
                link.kind = LinkType.ETH
            else:
                link.kind = LinkType.SWP

        if (link.kind == LinkType.BCM or
                link.kind == LinkType.MLX or
                (link.kind == LinkType.TUN and is_swp_kind(info.ifname))):
            link.kind = LinkType.SWP
        return link.kind

    def __rx_rtm_newlink(self, info):
        """Adds a new link to the DB.
        :param info: link object
        :type info: netlink.Link
        """
        self._logger.debug('RX RTM_NEWLINK for ifname %s, admin_state %s, '
                           'oper_state %s', info.ifname, info.admin_state,
                           info.oper_state)

        del_master_idx_entry = False
        link = Link()
        old_entry = None

        self.__ifname_by_index[info.ifindex] = info.ifname
        if info.kind == LinkType.VRF:
            self.__link_vrf_map[info.rt_table_id] = info.ifname

        self.__link_table_map[info.ifindex] = info.rt_table_id

        m_link = Link.memmodel()
        old_entry = m_link.query.get(hostname=HOSTNAME,
                                     ifname=info.ifname)

        if info.family == socket.AF_BRIDGE:
            if old_entry:
                link.copy(old_entry)
            else:
                # Being bold here. Discard any updates we get before we
                # determine link type. Assuming we can get it all back.
                self._logger.info('Ignored AF_BRIDGE link upd8 for %s, no '
                                  'cache', info.ifname)
                return info.ifindex
        else:
            # for a bridge port, we can't assume an update always comes in an
            # AF_UNSPEC message followed by an AF_BRIDGE message, need to
            # carry over the link object attributes if it exists
            if old_entry:
                link.copy(old_entry)

            NetlinkCopy.copy_link(link, info)

        # Some of the fields have to be always copied over
        NetlinkCopy.copy_link_always(link, info)

        if info.stp_state != StpPortState.UNINIT:
            link.stp_state = info.stp_state
        elif old_entry and old_entry.stp_state != StpPortState.UNINIT:
            link.stp_state = old_entry.stp_state
        else:
            link.stp_state = StpPortState.FORWARDING

        if info.master_ifindex:
            link.master = self.__ifname_by_index.get(info.master_ifindex, '')
            if not link.master:
                link.master = str(info.master_ifindex)
                if info.master_ifindex not in self.__to_update_master_links:
                    self.__to_update_master_links[info.master_ifindex] = []
                self.__to_update_master_links[info.master_ifindex].append(
                    info.ifname)
        else:
            # master_ifindex is not present
            # could be bridge port or bridge notification AF_BRIDGE
            old_master = link.master
            link.master = info.ifname

            # need to reset bridge port attributes if link is no longer
            # a bridge port
            if info.kind != LinkType.BRIDGE and \
                    (old_master and old_master != link.master):
                link.access_vlan = 0
                link.vlans = ''

        # Forget the past with the old master, and go live with the new
        if old_entry and link.master != old_entry.master:
            del_master_idx_entry = True
            if old_entry.master.isdigit():
                old_master = int(old_entry.master)
                if old_master in self.__to_update_master_links:
                    try:
                        self.__to_update_master_links[old_master].remove(
                            info.ifname)
                    except ValueError:
                        pass

        # Update all slaves who hadn't resolved their old master
        # ifindex to name
        self.__update_link_masters(info.ifindex, info.ifname, info.kind)
        self.set_link_kind(link, info)

        link.managed = True

        if info.parent_if:
            link.parent_if = self.__ifname_by_index.get(info.parent_if,
                                                        str(info.parent_if))
        else:
            link.parent_if = info.ifname

        if info.kind == LinkType.VLAN:
            link.access_vlan = info.vni

        if info.kind == LinkType.MACVLAN:
            m_link = Link.memmodel()
            for m_lentry in list(m_link.query.filter(hostname=HOSTNAME,
                                                     ifname=link.parent_if)):
                link.access_vlan = m_lentry.access_vlan
                break

        # When the VLAN->VNI mapping happens, the kind is missing and so
        # the block below will not be executed if we use info.kind. We
        # have to rely on the cached link.kind
        if link.kind == LinkType.VXLAN:
            # We use this cached value to handle MacFdb updates
            self.__vxlan_vlan_map[info.ifname] = {'vlan': link.access_vlan,
                                                  'vni': link.vni}
            if info.ifname not in self.__link_kind_map:
                self.__link_kind_map[info.ifname] = {}
            self.__link_kind_map[info.ifname]['vni'] = link.vni

        # We store the name of the master interface with an interface key.
        # However, the kernel merely supplies the ifindex leaving it an
        # exercise for the user to derive the name associated with that index.
        # Sometimes, we get the slave interface's info before we receive the
        # master's, and we don't have a way to derive the ifname for the master
        # In such a scenario, we create the entry with the ifindex with the
        # hope that we'll resolve it soon after when we receive another nlink
        # update. When that event update happens, we want to delete the old
        # entry. Thats what the following code does.
        if del_master_idx_entry and old_entry:
            del_old_entry = Link()
            del_old_entry.copy(old_entry)
            self._logger.debug('Deleting ifidx master for %s' % info.ifname)
            if del_old_entry.objid:
                # Delete associated description too
                descr = Description()
                descr.objtype = Link.__name__
                descr.descrid = del_old_entry.objid
                descr.delete()
            del_old_entry.delete()
        elif (old_entry and old_entry.objid and not info.ifalias and
              info.family != socket.AF_BRIDGE):
            descr = Description()
            descr.objtype = old_entry.__class__.__name__
            descr.descrid = old_entry.objid
            descr.delete()

        self.__link_master_map[info.ifname] = link.master
        if link.kind != LinkType.VRF:
            link.vrf = self.__link_vrf_map.get(info.rt_table_id, 'default')

        # If there are Addresses that did not get ifName mapped
        # map it now
        if self.__addr_ifindex_map.get(info.ifindex):
            for prefix in self.__addr_ifindex_map[info.ifindex]:
                addr = self.__addr_ifindex_map[info.ifindex][prefix]
                addr.ifname = info.ifname
                addr.save()

        # If the VRF associated with this device changes, rejig the addr
        # associated to carry the right VRF.
        m_address = Address.memmodel()
        if old_entry and old_entry.vrf != link.vrf:
            for m_aentry in list(m_address.query.filter(hostname=HOSTNAME,
                                                        ifname=info.ifname,
                                                        vrf=old_entry.vrf)):
                old_aentry = Address()
                old_aentry.copy(m_aentry)
                old_aentry.delete()
                new_aentry = Address()
                new_aentry.copy(m_aentry)
                new_aentry.vrf = link.vrf
                new_aentry.save()

        self._logger.debug('Adding new link %r to DB', link)
        if info.ifalias:
            descr = Description()
            descr.timestamp = link.timestamp
            descr.objtype = link.__class__.__name__
            descr.descrid = link.get_objid()
            descr.description = info.ifalias
            link.objid = descr.descrid
            descr.save()

        link.save()
        if link.kind == LinkType.SWP or link.kind == LinkType.ETH:
            self.__ports.add(link.ifname)
        return info.ifindex

    def __rx_rtm_dellink(self, info):
        """Deletes a link from the DB.
        :param info: link object
        :type info: netlink.Link
        """
        self._logger.debug('RX RTM_DELLINK for ifname %s, admin_state %s, '
                           'oper_state %s', info.ifname, info.admin_state,
                           info.oper_state)
        link = Link()
        NetlinkCopy.copy_link(link, info)
        link.master = self.__link_master_map.get(info.ifname, str(info.ifname))
        self.__to_update_master_links.pop(info.ifindex, None)
        self.set_link_kind(link, info)
        self._logger.debug('Deleting link %r from DB', link)
        # If there are Routes that use this interface as nexthop
        # delete them as well
        # no guarantees that netlink will be generated for those
        # routes
        mcls = Route.memmodel()
        entries = mcls.query.filter()
        for entry in entries:
            to_save = False
            for nh in entry.nexthops:
                if nh[1] == link.ifname:
                    entry.nexthops.remove(nh)
                    to_save = True
            if to_save:
                rt = Route()
                rt.copy(entry)
                if not rt.nexthops:
                    rt.delete()
                else:
                    rt.save()

        # If there are Addresses that did not get ifName mapped
        # delete them as well
        if self.__addr_ifindex_map.get(info.ifindex):
            self.__addr_ifindex_map.pop(info.ifindex)
        descr = Description()
        descr.objtype = link.__class__.__name__
        descr.descrid = link.get_objid()
        descr.delete()
        link.delete()
        if link.kind == LinkType.SWP or link.kind == LinkType.ETH:
            self.__ports.discard(link.ifname)
        return info.ifindex

    def __rx_rtm_newneigh(self, info):
        """Adds a new neighbor to the DB.
        Neighbors in the kernel go through a lot of state transitions that
        we are currently not interested in tracking.  Only update redis if a
        neighbor is new...ie ignore all of the RTM_NEWNEIGH messages
        generated from the state transitions
        :param info: neighbor object
        :type info: netlink.Neighbor
        """
        self._logger.debug('RX RTM_NEWNEIGH for ip_address %s, ifindex %s',
                           info.ip_address, info.ifindex)
        if info.family == socket.AF_BRIDGE:
            neighbor = MacFdb()
            self.__nlmonitor.stats.increment(netlink.NetlinkStat.RTM_NEWFDB,
                                             1)
            NetlinkCopy.copy_macfdb(neighbor, info, self.__ifname_by_index)
            if neighbor.dst:
                vinfo = self.__vxlan_vlan_map.get(neighbor.nexthop)
                if vinfo:
                    neighbor.vlan = int(vinfo['vlan'])
                    # get the replication list for the vni
                    if neighbor.mac_address == '00:00:00:00:00:00':
                        rdst = VxlanRemoteDest()
                        rdst.vni = int(vinfo['vni'])
                        rdst.rdst = neighbor.dst
                        rdst.save()
                else:
                    neighbor.vlan = -1

            else:
                if self.__link_kind_map.get(neighbor.nexthop, None):
                    # This is the second update to the same remote MAC
                    # but this time from the bridge, without the remote vtep
                    # address. Ignore it.
                    self._logger.debug('Ign new neighbor %r to DB', neighbor)
                    return info.ifindex
        else:
            neighbor = Neighbor()  # pylint3: disable=redefined-variable-type
            NetlinkCopy.copy_neighbor(neighbor, info, self.__ifname_by_index)
            table = self.__link_table_map.get(info.ifindex, RouteTable.DEFAULT)
            neighbor.vrf = self.__link_vrf_map.get(table, 'default')
            if neighbor.vrf == neighbor.ifname:
                # ignore nbr updates that happen on the vrf device
                return info.ifindex
        if info.state == NeighborState.UPDATE:
            self._logger.debug('Adding new neighbor %r to DB', neighbor)
            neighbor.save()
        elif info.state == NeighborState.DELETE:
            # In neighbor delete, the MAC address is not provided. So get
            # the mac address from the entry in memory
            mcls = Neighbor.memmodel()
            mentry = mcls.query.get(hostname=HOSTNAME,
                                    ifname=neighbor.ifname,
                                    ip_address=neighbor.ip_address,
                                    is_ipv6=neighbor.is_ipv6,
                                    vrf=neighbor.vrf)
            if mentry:
                neighbor.mac_address = mentry.mac_address
                self._logger.debug('Deleting neighbor %r from DB', neighbor)
                neighbor.delete()
            else:
                # We cannot process the delete of this entry because we don't
                # have the mac address
                self._logger.debug('Unable to delete neighbor %s', neighbor)
        return info.ifindex

    def __rx_rtm_delneigh(self, info):
        """Deletes a neighbor from the DB.
        :param info: neighbor object
        :type info: netlink.Neighbor
        """
        self._logger.debug('RX RTM_DELNEIGH for ip_address %s, ifindex %s',
                           info.ip_address, info.ifindex)
        if info.family == socket.AF_BRIDGE:
            neighbor = MacFdb()
            self.__nlmonitor.stats.increment(netlink.NetlinkStat.RTM_DELFDB,
                                             1)
            NetlinkCopy.copy_macfdb(neighbor, info, self.__ifname_by_index)
            if neighbor.dst and neighbor.mac_address == '00:00:00:00:00:00':
                vinfo = self.__vxlan_vlan_map.get(neighbor.nexthop)
                rdst = None
                if vinfo:
                    vnid = int(vinfo['vni'])
                    rdsts = VxlanRemoteDest.memmodel()
                    rdst = rdsts.query.get(hostname=HOSTNAME, vni=vnid,
                                           rdst=neighbor.dst)
                if rdst:
                    old_rdst = VxlanRemoteDest()
                    old_rdst.copy(rdst)
                    old_rdst.delete()
        else:
            neighbor = Neighbor()  # pylint3: disable=redefined-variable-type
            NetlinkCopy.copy_neighbor(neighbor, info, self.__ifname_by_index)
            table = self.__link_table_map.get(info.ifindex, RouteTable.DEFAULT)
            neighbor.vrf = self.__link_vrf_map.get(table, 'default')
        self._logger.debug('Deleting neighbor %r from DB', neighbor)
        neighbor.delete()
        return info.ifindex

    def __rx_rtm_newroute(self, info):
        """Adds a new route to the DB.
        :param info: route object
        :type info: netlink.Route
        """
        # Skip redistribute neighbor routes in table 10
        if info.table == self.__REDISTRIBUTE_NBR_TABLE_ID:
            self._logger.debug('RX RTM_NEWROUTE ignore ip_address %s in redist nbr %s',
                               info.ip_address, info.table)
            return info.ifindex

        self._logger.debug('RX RTM_NEWROUTE for ip_address %s, nexthops %s',
                           info.ip_address, info.nexthops)

        if info.table is None or info.ifindex is None:
            #
            # This is a temporary fix to handle MPLS label path
            # NLManager does not ssupport parsing MPLS labeled multipath
            # resulting in invalid nexthop/table info
            #
            self._logger.info('RTM_NEWROUTE invalid table-id:%r or \
                ifindex:%r for %s, nexthops %s',
                              info.table, info.ifindex, info.ip_address, info.nexthops)
            return info.ifindex

        route = Route()
        NetlinkCopy.copy_route(route, info, self.__ifname_by_index,
                               self.__rt_protos_map)
        if route.rt_table_id == RouteTable.LOCAL and route.route_type != 3:
            # In case of VRFs, all local routes are shown as being in the VRF
            # table whereas for the default VRF, the local route (essentially
            # the interface address itself) is shown as being in table local.
            # This prevents us from efficiently using redis key structure to
            # filter entries.
            route.rt_table_id = RouteTable.DEFAULT
        route.vrf = self.__link_vrf_map.get(route.rt_table_id, 'default')

        # ignore certain route pfx
        skip_ipv6_addr = {'::1/128', 'ff00::/8'}
        if route.is_ipv6 and route.prefix in skip_ipv6_addr:
            return info.ifindex

        if route.is_ipv6:
            mcls = Route.memmodel()
            entries = mcls.query.filter(is_ipv6=route.is_ipv6, vrf=route.vrf,
                                        prefix=route.prefix, origin=route.origin)
            # check if lower metric detected
            for entry in entries:
                if route.priority > entry.priority:
                    # ignore higher metric
                    self._logger.debug('Ignoring route priority new %d > old %d',
                                       route.priority, entry.priority)
                    return info.ifindex

        self._logger.debug('Adding new route %r in VRF %s to DB', route,
                           route.vrf)
        route.save()
        return info.ifindex

    def __rx_rtm_delroute(self, info):
        """Delete a route from the DB.
        :param info: route object
        :type info: netlink.Route
        """
        self._logger.debug('RX RTM_DELROUTE for ip_address %s, nexthops %s',
                           info.ip_address, info.nexthops)
        route = Route()
        NetlinkCopy.copy_route(route, info, self.__ifname_by_index,
                               self.__rt_protos_map)
        if (route.rt_table_id == RouteTable.LOCAL and route.route_type != 3):
            # In case of VRFs, all local routes are shown as being in the VRF
            # table whereas for the default VRF, the local route (essentially
            # the interface address itself) is shown as being in table local.
            # This prevents us from efficiently using redis key structure to
            # filter entries.
            route.rt_table_id = RouteTable.DEFAULT
        route.vrf = self.__link_vrf_map.get(route.rt_table_id, 'default')
        self._logger.debug('Deleting route %r in VRF %s from DB', route,
                           route.vrf)
        route.delete()
        return info.ifindex

    def __stop_command_checker(self, green_thread, identifier, command):
        try:
            self.__command_buffer[identifier], _ = green_thread.wait()
        except greenlet.GreenletExit:
            pass
        except Exception as ex:  # pylint: disable=broad-except
            self._logger.info('Failed to execute command %s Error: %s', command, ex)

    def __stop_monitor_checker(self, green_thread):
        try:
            green_thread.wait()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            pass
        except Exception as ex:  # pylint: disable=broad-except
            if self.__nlmonitor_gt == green_thread:
                self.__nlmonitor_gt = None
                self._logger.info('Netlink monitor thread terminated unexpectedly. Error: %s', ex)
                if ex.errno == errno.ENOBUFS: # pylint: disable=no-member
                    self.__NETLINK_RESTART = True
                else:
                    self.__pause_agent()
            else:
                self._logger.exception('Monitor thread terminated unexpectedly. Error: %s', ex)
                self.__pause_agent()

    def __stop_thread_checker(self, green_thread):
        """ Propagates exceptions to the run green thread.
        """
        entry = next((thread for thread in self.__redisthreads
                      if thread.green_thread == green_thread), None)
        if entry is not None:
            self.__redisthreads.remove(entry)
            if entry == self.__heartbeat_gt:
                self.__heartbeat_gt = None
            elif entry == self.__nlinit_gt:
                self.__nlinit_gt = None
        try:
            green_thread.wait()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            pass
        except Exception as ex:
            if entry is not None:
                if entry.pipeline is not None:
                    self._logger.info('Failed to execute PIPELINE %s. Error: %s',
                                      id(entry.pipeline), ex)
                elif entry == self.__heartbeat_gt:
                    self._logger.info('Failed to send heartbeat. Error: %s', ex)
                elif entry == self.__nlinit_gt:
                    self._logger.info('Failed to execute netlink thread. Error: %s', ex)
                    self.__NETLINK_RESTART = True
            self.__pause_agent()
            return
        else:
            if entry is not None:
                if entry.pipeline is not None:
                    self._logger.debug('Successfully executed PIPELINE %s',
                                       id(entry.pipeline))
                elif entry == self.__heartbeat_gt:
                    self._logger.debug('Successfully sent heartbeat')
                elif entry == self.__nlinit_gt:
                    self.__nlinit_gt = None

    def __process_server_info(self):
        with self.__serverinfopool.item():
            self._logger.debug('Querying server info from server')
            self.__update_server_info()

    def __process_heartbeat(self):
        appliance_version = 'N/A'
        manifest_hash = 'N/A'

        try:
            if self.__app_release_info:
                for line in self.__app_release_info.splitlines():
                    if line.startswith('APPLIANCE_VERSION'):
                        _, appliance_version = line.strip().split('=')
                    if line.startswith('APPLIANCE_MANIFEST_HASH'):
                        _, manifest_hash = line.strip().split('=')
        except ValueError as ex:
            self._logger.error("Failed to process app release info: {}".format(ex))
            raise ex

        node = Node()
        node.last_reinit = self.__last_reinit
        node.lastboot = self.__lastboot
        node.sys_uptime = self.__node_uptime
        node.ntp_state = self.__check_ntp_status()
        if Platform.CUMULUS in self.__platform:
            node.ptp_state = ptp.ptp_render.is_ptp_enabled()
        else:
            node.ptp_state = False
        node.version = self.__agent_version
        node.appliance_version = appliance_version
        node.manifest_hash = manifest_hash
        #self.__agent_monitor.send_ping()
        with self.__heartbeatpool.item():
            if self.__AGENT_PAUSE is True:
                return
            self.__logger_coalescer.debug('Sending heartbeat to server, Netlink stats: %s',
                               json.dumps(repr(self.__nlmonitor.stats))) # pylint: disable=no-member
            node.save()

    def __init_rt_protos(self):
        '''Initialize routing protocol types'''

        self.__rt_protos_map = {}

        try:
            for file in ('/etc/iproute2/rt_protos',
                         '/etc/iproute2/rt_protos.d/frr.conf',
                         '/etc/iproute2/rt_protos.d/quagga.conf'):
                with open(file, 'r') as fh:
                    lines = fh.readlines()

                for line in lines:
                    line = line.strip()

                    if not line or line.startswith('#'):
                        continue

                    try:
                        table, proto = line.split()
                        self.__rt_protos_map[table] = proto
                    except ValueError:
                        pass
        except IOError:
            pass

    def __run_platform_cmd(self, command, timeout=__CMD_TIMEOUT):
        try:
            if not command:
                return None
            if isinstance(command, str):
                cmd = subprocess.Popen(command.split(),
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       shell=False)
            else:
                cmd = subprocess.Popen(command,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.STDOUT,
                                       shell=False)
        except (subprocess.CalledProcessError, OSError):
            return None

        identifier = uuid.uuid4().hex
        green_thread = self._pool.spawn(cmd.communicate)
        green_thread.link(self.__stop_command_checker, identifier, command)
        endtime = time.time() + int(timeout)
        return (cmd, identifier, green_thread, endtime, timeout)

    def __update_platform_info(self, is_full_state=False):
        # platform-detect
        box = self.__platform.split(',', 1)[1].strip()
        if 'nvidia' in box:
            box = box.split(',', 1)[1].strip()

        self.__logger_coalescer.debug("BOX: {} PLATFORM COMMANDS: {}".format(box, self.__platform_commands))

        output_dict = {}
        cmd_timeout = self.__CMD_TIMEOUT if self.__LAST_AGENT_STATE_RUNNING else self.__CMD_TIMEOUT + 15
        nvue_timeout = self.__CMD_TIMEOUT + 15

        gthread_dict = {}
        # OS Release
        result = self.__run_platform_cmd(self.__platform_commands.get('os-release'), timeout=cmd_timeout)
        if result:
            gthread_dict['os-release'] = result
        # Board Info
        result = self.__run_platform_cmd(self.__platform_commands.get('eprom'), timeout=cmd_timeout)
        if result:
            gthread_dict['eprom'] = result
        # lscpu
        result = self.__run_platform_cmd(self.__platform_commands.get('lscpu'), timeout=cmd_timeout)
        if result:
            gthread_dict['lscpu'] = result
        # meminfo
        result = self.__run_platform_cmd(self.__platform_commands.get('meminfo'), timeout=cmd_timeout)
        if result:
            gthread_dict['meminfo'] = result
        # lsblk
        result = self.__run_platform_cmd(self.__platform_commands.get('lsblk'), timeout=cmd_timeout)
        if result:
            gthread_dict['lsblk'] = result
        # dmidecode memory
        result = self.__run_platform_cmd(self.__platform_commands.get('dmicode'), timeout=cmd_timeout)
        if result:
            gthread_dict['dmicode'] = result

        result = self.__run_platform_cmd(self.__platform_commands.get('service-nvue-enabled'), timeout=cmd_timeout)
        if result:
            gthread_dict['service-nvue-enabled'] = result

        result = self.__run_platform_cmd(self.__platform_commands.get('service-nvue-active'), timeout=cmd_timeout)
        if result:
            gthread_dict['service-nvue-active'] = result

        result = self.__run_platform_cmd(self.__platform_commands.get('is_nvue_config_saved'), timeout=nvue_timeout)
        if result:
            gthread_dict['is_nvue_config_saved'] = result

        result = self.__run_platform_cmd(self.__platform_commands.get('service-nclu-enabled'), timeout=cmd_timeout)
        if result:
            gthread_dict['service-nclu-enabled'] = result

        result = self.__run_platform_cmd(self.__platform_commands.get('opta-on-switch-enabled'), timeout=cmd_timeout)
        if result:
            gthread_dict['opta-on-switch-enabled'] = result

        result = self.__run_platform_cmd(self.__platform_commands.get('lcm-executor-on-switch-enabled'), timeout=cmd_timeout)
        if result:
            gthread_dict['lcm-executor-on-switch-enabled'] = result


        for cmd_key, (cmd, identifier, green_thread, endtime, timeout) in gthread_dict.items():
            command = self.__platform_commands.get(cmd_key)
            cmd_out = None
            while True:
                if identifier in self.__command_buffer:
                    cmd_out = self.__command_buffer.pop(identifier)
                    if cmd_out is None:
                        retcode = -1
                    else:
                        retcode = cmd.returncode
                    break
                if time.time() > endtime:
                    self._logger.info('Failed to execute command %s after %d '
                                      'seconds', command, timeout)
                    cmd.terminate()
                    eventlet.sleep(0.3)
                    cmd.wait()
                    eventlet.kill(green_thread, greenlet.GreenletExit)
                    retcode = -1
                    break
                eventlet.sleep(1)

            if retcode != 0:
                if isinstance(command, list):
                    command = ",".join(command)
                if command not in self.__failed_commands:
                    self._logger.debug('Command %s returned %d', command, retcode)
                    self.__failed_commands[command] = retcode
            if cmd_out:
                output_dict[cmd_key] = cmd_out.decode()
            else:
                output_dict[cmd_key] = None

        platform_json = None
        if Platform.CUMULUS in self.__platform:
            paths = ['/usr/share/cumulus-platform/common/platform-info',
                     '/usr/share/cumulus/platform-info']
            for path in paths:
                if platform_json:
                    break
                for _, _, files in os.walk(path):
                    for filen in files:
                        try:
                            fp = open(os.path.join(path, filen))
                            plat = json.load(fp) # pylint: disable=no-member
                            fp.close()
                        except Exception:  # pylint: disable=broad-except
                            pass
                        else:
                            if box == 'dellemc,s5048f_c2538' and not plat.get(box):
                                box = 'dellemc,s5048f'
                            if "simx" in box and plat.get('mlnx,simx'):
                                platform_json = plat['mlnx,simx'][0]
                            if plat.get(box):
                                platform_json = plat[box][0]

        if Platform.SONIC in box:
            if not is_sonic_ib():
                sonic_process_platform_info(self._logger, self.__random_platform_info,
                                        self.__platform,
                                        output_dict.get('os-release'), output_dict.get('eprom'),
                                        output_dict.get('lscpu'), output_dict.get('meminfo'), output_dict.get('os-release'),
                                        output_dict.get('dmicode'), output_dict.get('lsblk'),
                                        self.__agent_version, box, is_full_state,
                                        self.__app_release_info, None, None, None, None) # NVUE fields are None for Sonic
        elif Platform.BLUEFIELD in box.lower():
            bluefield_process_platform_info(self._logger, self.__random_platform_info,
                                            self.__platform,
                                            output_dict.get('os-release'), output_dict.get('eprom'),
                                            output_dict.get('lscpu'), output_dict.get('meminfo'), platform_json,
                                            output_dict.get('dmicode'), output_dict.get('lsblk'),
                                            self.__agent_version, box, is_full_state,
                                            self.__app_release_info,
                                            output_dict.get('service-nvue-enabled'), output_dict.get('service-nvue-active'),
                                            output_dict.get('is_nvue_config_saved'), output_dict.get('service-nclu-enabled'))
        else:
            cl_process_platform_info(self._logger, self.__random_platform_info,
                                     self.__platform,
                                     output_dict.get('os-release'), output_dict.get('eprom'),
                                     output_dict.get('lscpu'), output_dict.get('meminfo'), platform_json,
                                     output_dict.get('dmicode'), output_dict.get('lsblk'),
                                     self.__agent_version, box, is_full_state, self.__app_release_info,
                                     output_dict.get('service-nvue-enabled'), output_dict.get('service-nvue-active'),
                                     output_dict.get('is_nvue_config_saved'), output_dict.get('service-nclu-enabled'),
                                     output_dict.get("opta-on-switch-enabled"), output_dict.get('lcm-executor-on-switch-enabled'))

    def _handle_asic_monitor_service_timer(self, timers, epoch):
        if self.__asic_monitor_service is None:
            return

        asic_monitor_collection_periodicity = self.__asic_monitor_service.get_monitor_collection_periodicity()
        if (timers['asic_monitor'] == 0) or (epoch - timers['asic_monitor'] >= asic_monitor_collection_periodicity):
            self.__logger_coalescer.debug("Spawning histogram processor thread")
            gt_thread = self._pool.spawn(self.__asic_monitor_service.histogram.histogram_snapshot_processor)
            gt_thread.link(self._stop_checker)

            gt_thread = self._pool.spawn(self.__asic_monitor_service.buffer_stats.buffer_stats_snapshot_processor)
            gt_thread.link(self._stop_checker)
            timers['asic_monitor'] = epoch

    def _handle_platform_timer(self, timers, epoch):
        if (timers['platform'] == 0) or \
                (epoch - timers['platform'] >= self.__PLATFORM_RATE):
            # update platform info like mem, cpu etc
            if (self.__platform_fullstate_timer >= self.__platform_fullstate_period and
                    self.__fullstate_period != 0):
                self.__update_platform_info(True)
                self.__platform_fullstate_timer = 0
            else:
                self.__platform_fullstate_timer += self.__PLATFORM_RATE
                self.__update_platform_info(False)
            timers['platform'] = epoch

    def _handle_appinfo_timer(self, timers, epoch):
        if (timers['appinfo'] == 0) or \
                (epoch - timers['appinfo'] >= self.__APPINFO_RATE):
            self.__app_release_info = None
            if self.__is_opta():
                retcode, self.__app_release_info = self.__run_safe_cmd(self.__platform_commands.get('is-opta'))
                if retcode != 0:
                    self.__app_release_info = None

            timers['appinfo'] = epoch


    def _handle_nlpoll_timer(self, timers, epoch):
        if (timers['nlpoll'] == 0) or \
                (epoch - timers['nlpoll'] >= self.__fullstate_period and self.__fullstate_period != 0) or \
                (self.__nl_poll_retry < self.__NLPOLL_RETRY_MAX):
            self.__init_netlink_thread()
            timers['nlpoll'] = epoch
            self.__nl_poll_retry += 1

    def _handle_netlink_libnl_timer(self, timers, epoch):
        if self._is_libnl_enabled and self.__is_libnl_process_active():
            if self.__netlink_libnl_retry_timer != 0 and \
                    (epoch - timers['nlpoll']) >= self.__netlink_libnl_retry_timer:
                self.__send_netlink_libnl_fullstate()
                timers['nlpoll'] = epoch
                return
            if timers['nlpoll'] == 0 or \
                    (self.__fullstate_period != 0 and
                     epoch - timers['nlpoll'] >= self.__netlink_fullstate_period):
                self.__send_netlink_libnl_fullstate()
                timers['nlpoll'] = epoch

    def _handle_ng_agent_timer(self, timers, epoch):
        if self.__ng_agent_process and not self.__is_ng_agent_process_active():
            if epoch - timers.get('ng_agent', 0) > \
                    self.__ng_agent_back_off:
                self._logger.info("Initializing netq-ng-agent process")
                self.__ng_agent_process = start_ng_agent_service(
                    self._logger, self.__gnmi_enabled,
                    self.__opta_enabled, self._conf)
                timers['ng_agent'] = epoch

    def _handle_hb_timer(self, timers, epoch):
        if (timers['liveliness'] == 0) or \
                (epoch - timers['liveliness'] >= Heartbeat.LIVELINESS_RATE):
            if self.__heartbeat_gt is None:
                self.__heartbeat_gt = _Thread.spawn(epoch,
                                                    self._pool,
                                                    self.__process_heartbeat)
                self.__heartbeat_gt.link(self.__stop_thread_checker)
                self.__redisthreads.add(self.__heartbeat_gt)
                timers['liveliness'] = epoch
            else:
                for thread in list(self.__redisthreads):
                    if thread.stale(epoch):
                        if thread == self.__heartbeat_gt:
                            self._logger.warn('Killing heartbeat thread : Inactive')
                            self.__heartbeat_gt = None
                            thread.kill()


    def _handle_server_info_timer(self, timers, epoch):
        if not self.__redis_enabled:
            return
        if (timers['serverinfo'] == 0) or \
                (epoch - timers['serverinfo'] >= self.__SERVERINFO_RATE):
            thread = _Thread.spawn(epoch, self._pool,
                                   self.__process_server_info)
            thread.link(self.__stop_thread_checker)
            self.__redisthreads.add(thread)
            timers['serverinfo'] = epoch

    def _handle_command_timer(self, timers, epoch):
        new_timers, self.__timers = self.__timers.copy(), {}
        for key in new_timers:
            if epoch > key:
                for cmd in new_timers[key]:
                    green_thread = self._pool.spawn(self.__process_command,
                                                    cmd)
                    green_thread.link(self._stop_checker)
            else:
                self.__timers[key] = new_timers[key]

    def _handle_inactivity_timer(self, timers, epoch):
        if (timers['inactivity'] == 0) or \
                (epoch - timers['inactivity'] >= Heartbeat.HEARTBEAT_DEATH):
            for thread in list(self.__redisthreads):
                if thread.stale(epoch):
                    if thread == self.__heartbeat_gt:
                        self._logger.warn('Killing heartbeat thread')
                        self.__heartbeat_gt = None
                    else:
                        self._logger.warn('Inactivity timeout exceeded '
                                          'for pipeline %s', id(thread.pipeline))
                    thread.kill()
                    #self.__pause_agent()
            timers['inactivity'] = epoch

    def _handle_services_timer(self, timers, epoch):
        if (timers['services'] == 0) or \
                (epoch - timers['services'] >= self.__SERVICES_RATE):
            # self.__process_counters()
            is_fullstate = timers['services'] == 0
            self.__check_services(is_fullstate)
            timers['services'] = epoch

    def _handle_wjh_service_monitor_timer(self, timers, epoch):
        if (timers['wjh_service'] == 0) or \
                (epoch - timers['wjh_service'] >= self.__WJH_MONITOR_RATE):
            if 'mlnx' in self.__platform:
                if self.__wjh_process and self.__is_wjh_process_active():
                    self.__monitor_wjh_service()
            timers['wjh_service'] = epoch

    def _handle_frr_docker_timer(self, timers, epoch):
        if (not self.__frr_docker_enabled or
                (self.__frr_docker_name and
                 bool(self.__frr_docker_child_pid))):
            timers['frr_docker'] = epoch
            return
        if (timers['frr_docker'] == 0) or \
                (epoch - timers['frr_docker'] >= self.__SERVICES_RATE):
            green_thread = self._pool.spawn(self.__update_frr_docker_info)
            green_thread.link(self._stop_checker)
            timers['frr_docker'] = epoch

    def _start_grpc_stream(self):
        pq, bp_pq = create_message_queues(conf=self._conf, logger=self._logger)
        gnmi_pq = create_gnmi_message_queues(conf=self._conf, logger=self._logger) \
            if self.__gnmi_enabled else None
        grpc_server_inst = NetQAgentGrpcClient(self._conf, self._logger, pq, bp_pq, gnmi_pq)
        self.__grpc_stream_process = \
            Process(target=grpc_server_inst.start_stream, args=(self._conf,))
        self.__grpc_stream_process.start()

    def __clean_up_procs(self):
        self._logger.info("Terminating Netlink process")
        if (self._is_libnl_enabled and self.__netlink_libnl_process.is_alive()):
            self.__netlink_libnl_process.terminate()
        self._logger.info("Terminating WJH process")
        if (self.__wjh_process and self.__is_wjh_process_active()):
            self.__wjh_process.terminate()
        self._logger.info("Terminating Spice process")
        if (self.__spice_process and self.__is_spice_process_active()):
            self.__spice_process.terminate()
        self._logger.info("Terminating gRPC Stream process")
        if self.__grpc_stream_process.is_alive():
            self.__grpc_stream_process.terminate()
        self._logger.info("Terminating NG Agent process")
        if self.__ng_agent_process and self.__is_ng_agent_process_active():
            self.__ng_agent_process.terminate()
            try:
                self.__ng_agent_process.wait(timeout=120)
            except TimeoutExpired:
                self._logger.warning("netq-ng-agent process termination timeout. Killing process")
                self.__ng_agent_process.kill()
        if (self.__log_exporter_process and self.__is_log_exporter_process_active()):
            self.__log_exporter_process.terminate()

    def __is_grpc_process_active(self):
        if self.__grpc_stream_process.is_alive():
            msg_string = netq_agent_get_subprocess_status_str("gRPC stream process",
                                                              self.__grpc_stream_process.pid,
                                                              "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)
            self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.GRPC_STATUS)
            return True
        else:
            msg_string = netq_agent_get_subprocess_status_str("gRPC stream process", -1,
                                                              "Not alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                            msg_string)
            self.__logger_coalescer.warn(msg_string, group=NetqLogGroup.GRPC_STATUS)
            self.__clean_up_procs()
            return False

    def __is_ng_agent_process_active(self):
        ret_code = self.__ng_agent_process.poll()
        if ret_code is None:
            msg_string = netq_agent_get_subprocess_status_str("ng agent process",
                                                              self.__ng_agent_process.pid,
                                                              "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NG_AGENT_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)
            self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.NG_AGENT_STATUS)
            return True
        else:
            if ret_code != 0:
                msg_sub_string = "Not alive. return code: {}, stdout: {}, stderr: {}"\
                    .format(ret_code,
                            self.__ng_agent_process.stdout.read(),
                            self.__ng_agent_process.stderr.read())
                msg_string = netq_agent_get_subprocess_status_str("ng agent process",
                                                                  self.__ng_agent_process.pid,
                                                                  msg_sub_string)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NG_AGENT_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                                msg_string)
                self._logger.warn(msg_string)
            else:
                msg_sub_string = "Ended. return code: {}, stdout: {}, stderr: {}"\
                    .format(ret_code,
                            self.__ng_agent_process.stdout.read(),
                            self.__ng_agent_process.stderr.read())
                msg_string = netq_agent_get_subprocess_status_str("ng agent process",
                                                                  self.__ng_agent_process.pid,
                                                                  msg_sub_string)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NG_AGENT_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string, True)
                self._logger.debug(msg_string)
            return False

    def __is_libnl_process_active(self):
        if not self._is_libnl_enabled or not self.__netlink_libnl_process:
            return True
        if self.__netlink_libnl_process.is_alive():
            msg_string = netq_agent_get_subprocess_status_str("netlink process",
                                                              self.__netlink_libnl_process.pid,
                                                              "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NL_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)
            self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.NL_STATUS)

            self.__receive_netlink_info()
            return True
        else:
            msg_string = netq_agent_get_subprocess_status_str("netlink process",
                                                              self.__netlink_libnl_process.pid,
                                                              "Not alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NL_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                            msg_string)
            self.__logger_coalescer.warn(msg_string, group=NetqLogGroup.NL_STATUS)
            self.__clean_up_procs()
            return False


    def _respawn_grpc_process(self):
        self._logger.info("Respawning ")
        self._start_grpc_stream()

    def _get_backoff_timer(self):
        self.__backoff_timer = self.__backoff_timer + self.__backoff_timer_step
        if self.__backoff_timer > self.__agent_stable_time:
            self.__backoff_timer = self.__agent_stable_time
        self.__logger_coalescer.debug("Backoff Timer {}".format(self.__backoff_timer))
        return self.__backoff_timer

    def _set_next_reinit_time(self):
        # Expiry in fibanocci series
        self.__next_reinit_time = time.time() + self._get_backoff_timer()
        self.__logger_coalescer.debug("Current Time", "{} next reinit time {}".format(time.time(), self.__next_reinit_time))

    def _monitor_connected_state(self):

        self.__logger_coalescer.debug("OPTA {} Q full {} empty {} ".format(self.__opta_enabled, NetQPriorityQueue.PQ.is_full(), NetQPriorityQueue.bp_PQ.is_empty()))

        if (self.__opta_enabled and
                (NetQPriorityQueue.PQ.is_full() or
                 not NetQPriorityQueue.bp_PQ.is_empty())):
            self._logger.info("Setting agent pause state to true")
            self.__AGENT_PAUSE = True

        # Agent is stable, no more fluctuations
        if self.__next_reinit_time and time.time() > self.__last_reinit + self.__agent_stable_time:
            self.__logger_coalescer.debug("Agent is stable for {}".format(self.__agent_stable_time))
            self.__next_reinit_time = 0

        # In case Agent is paused because of full or grpc connection is dead
        # Agent shall be paused and set next reinit time
        opta_state, gnmi_state = self.__get_grpc_connection_state()
        if self.__opta_enabled:
            if (NetQPriorityQueue.PQ.is_full() or not opta_state):
                self.__latest_pause_time = time.time()
                self.__AGENT_PAUSE = True
                msg_string = 'gRPC connection - PQ is full... Agent is paused'
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_PQ_FULL_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self.__logger_coalescer.info(msg_string, group=NetqLogGroup.GRPC_PQ_FULL_STATUS)
            else:
                msg_string = 'gRPC connection - PQ is not full...'
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_PQ_FULL_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string)
                self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.GRPC_PQ_FULL_STATUS)
        else:
            msg_string = 'OPTA not enabled... gRPC PQ check skipped ...'
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GRPC_PQ_FULL_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)

        if self.__gnmi_enabled:
            if (NetQPriorityQueue.GNMI_PQ.is_full() or not gnmi_state):
                self.__GNMI_AGENT_PAUSE = True
                msg_string = 'gNMI connection - PQ is full... Agent is paused'
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GNMI_PQ_FULL_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self._logger.info(msg_string)
            else:
                msg_string = 'gNMI connection - PQ is not full...'
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GNMI_PQ_FULL_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string)
                self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.GNMI_PQ_FULL_STATUS)
        else:
            msg_string = 'gNMI not enabled... PQ check skipped ...'
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_GNMI_PQ_FULL_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)

    def _is_reset_allowed(self):

        current_time = time.time()

        # Case 2: try resetting
        if self.__next_reinit_time < current_time:
            self._set_next_reinit_time()
            return True
        else:
            return False

    def _spawn_cli(self):
        green_thread = self._pool.spawn(self.__agent_cli.process_cli_server)
        green_thread.link(self._stop_checker)

    def _start_wjh(self):
        conf_dict = configuration.get_config(_PROC_NAME)
        enabled = conf_dict.get('netq-agent').get(ConfKw.IS_WJH_ENABLED, ConfKw.DEFAULT_IS_WJH_ENABLED)

        if not enabled:
            return None

        vx = Platform.VX in self.__platform or Platform.SONIC_VS in self.__platform
        mlx = 'mlnx' in self.__platform
        wjh_supported_platforms = (PlatformType.SONIC, PlatformType.CL4)
        if (mlx and utils.get_platform_type() in wjh_supported_platforms) or vx:
            self.__monitor_wjh_service()
            wjh_client_proc = Process(target=partial(self._init_wjh, mlx))
            wjh_client_proc.start()

            if wjh_client_proc.is_alive():
                msg_string = netq_agent_get_subprocess_status_str("WJH", wjh_client_proc.pid, "Alive")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_WJH_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string)
                self.__logger_coalescer.debug(msg_string, group=NetqLogGroup.WJH_STATUS)
            else:
                msg_string = netq_agent_get_subprocess_status_str("WJH", -1, "Failed")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_WJH_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self.__logger_coalescer.info(msg_string, group=NetqLogGroup.WJH_STATUS)

            return wjh_client_proc
        return None


    def _init_wjh(self, mlx):
        if mlx and is_cl4():
            try:
                if os.path.exists('/bin/systemctl'):
                    is_active = is_active_systemd_svc('sx_sdk.service')

                    if not is_active:
                        return

            except Exception as e:
                self._logger.debug("WJH process not running: check sx_sdk")
                return
        try:
            if mlx:
                conf_dict = configuration.get_config(_PROC_NAME)
                sdk_utils = SdkUtils(self._logger)
                sdk_utils.copy_sdk()
                from netq_agent.cmd.netq_wjh_agent import WjhClient # Import of WJH depends on platform Mellanox
                wjh_client_inst = WjhClient(conf_dict, self._logger, self.__platform, sdk_utils)
            else:
                wjh_client_inst = DummyWjhClient(self._logger)
            wjh_client_inst.start_wjh_client()
        except Exception as e:
            if 'sx_sdk.service' in str(e):
                self._logger.debug("WJH process not running: check sx_sdk")
            else:
                self._logger.error("WJH process failed: %s", str(e))


    def __is_wjh_process_active(self):
        if self.__wjh_process.is_alive():
            # self._logger.debug("WJH Process is alive {}".
            #                   format(self.__wjh_process.pid))
            if self.__wjhSubProcessState == wjhSubProcessState.UNKNOWN or \
                self.__wjhSubProcessState == wjhSubProcessState.INACTIVE:
                self.__wjhSubProcessState = wjhSubProcessState.ACTIVE

            if self.__wjhLoggedOnce is True:
                self._logger.info("WJH Process is alive {}".
                                  format(self.__wjh_process.pid))
                self.__wjhLoggedOnce = False
            return True
        else:
            doRestart = False
            if self.__wjhSubProcessState == wjhSubProcessState.ACTIVE:
                #Set this to true for 4.3.0
                doRestart = False

            self.__wjhSubProcessState = wjhSubProcessState.INACTIVE
            if self.__wjh_restart_hold == 0:
                if self.__wjhLoggedOnce is False:
                    self._logger.info("WJH Process is not alive {}".
                                      format(self.__wjh_process.pid))
                    self.__wjhLoggedOnce = True
                if doRestart is True:
                    self.__clean_up_procs()

            return False

    def __is_forwarding_active(self):
        fwding_dep_func = ['RoCE (without NVUE)', 'ECMP', 'AdaptiveRouting']
        if (not is_cl3() and not is_cl4()):
            # Implemented for CL alone
            msg_string = 'Forwarding daemon check: NA/Not Implemented'
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_FORWARDING_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)
            self._logger.debug(msg_string)
            return True

        if 'mlnx' not in self.__platform:
            # Check applicable for Mellanox platform
            msg_string = 'Forwarding daemon check: NA/Not Implemented'
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_FORWARDING_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)
            self._logger.debug(msg_string)
            return True
        try:
            cmd = '/bin/systemctl is-active switchd.service'
            isActive = utils.subprocess_wrapper(cmd.split())
            if isActive == "active":
                if self.__forwardingState != ForwardingState.FORWARDING_STATE_ACTIVE:
                    # transition from unknown or inactive to active
                    self.__forwardingState = ForwardingState.FORWARDING_STATE_ACTIVE
                    msg_string = 'Forwarding daemon check: Active...starting {} data export (if enabled)'.format(fwding_dep_func)
                    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_FORWARDING_STATUS,
                                                    NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                    msg_string)
                    self._logger.info(msg_string)
                return True

            if self.__forwardingState != ForwardingState.FORWARDING_STATE_INACTIVE:
                #transition from unknown or active to inactive
                self.__forwardingState = ForwardingState.FORWARDING_STATE_INACTIVE
                msg_string = 'Forwarding daemon check: Inactive... {} data export impacted (if enabled)'.format(fwding_dep_func)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_FORWARDING_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self._logger.warning(msg_string)
            return False

        except Exception:
            if self.__forwardingState != ForwardingState.FORWARDING_STATE_INACTIVE:
                #transition from unknown or active to inactive
                self.__forwardingState = ForwardingState.FORWARDING_STATE_INACTIVE
                msg_string = 'Forwarding daemon check: Inactive... {} data export impacted (if enabled)'.format(fwding_dep_func)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_FORWARDING_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self._logger.warning(msg_string)
            return False
        return True

    def __is_sx_sdk_running(self):
        if self.__sx_sdk_check != 0:
            self.__sx_sdk_check -= 1
            return
        else:
            self.__sx_sdk_check = 10
            if not self.__wjh_process:
                return

        self.__logger_coalescer.debug('checking sx_sdk...', group=NetqLogGroup.SX_SDK_STATUS)
        try:
            cmd = '/usr/bin/pidof sx_sdk'
            sdk_pid = utils.subprocess_wrapper(cmd.split())
            if int(sdk_pid) != self.__sx_sdk_pid:
                if self.__sx_sdk_pid != 0:
                    self.__wjhSubProcessState = wjhSubProcessState.SDK_RESTARTED
                    self.__wjh_process.terminate()
                    if self.__spice_process and self.__spice_process.is_alive():
                        self.__spice_process.terminate()
                    self._logger.info('')
                    self.__logger_coalescer.info('sx_sdk restarted. Terminated WJH client', group=NetqLogGroup.SX_SDK_STATUS)
                self.__sx_sdk_pid = int(sdk_pid)
        except Exception:
            if self.__sx_sdk_pid != 0:
                self.__wjhSubProcessState = wjhSubProcessState.SDK_NOT_RUNNING
                self.__wjh_process.terminate()
                if self.__spice_process and self.__spice_process.is_alive():
                    self.__spice_process.terminate()
                self.__logger_coalescer.info('sx_sdk not active. Terminated WJH client', group=NetqLogGroup.SX_SDK_STATUS)
                self.__sx_sdk_pid = 0

    def __is_netlink_freeze(self):
        difftime = time.time() - self.__netlink_init_time
        return difftime > self.__max_netlink_init_time

    def _run_command(self, cmd):
        try:
            proc = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except OSError:
            return None
        return proc

    def _sonic_unsubscribe_listener(self):
        # Subscribe manager will be present only for sonic.
        # To stop the listener threads on redis db
        if self.__subscribe_manager:
            self.__subscribe_manager.exit_gracefully()

    def __monitor_wjh_service(self):
        global _is_deliberate_wjh_disable # pylint: disable=global-statement
        if 'mlnx' not in self.__platform:
            return
        platform_type = utils.get_platform_type()
        if platform_type == PlatformType.CL4:
            if not os.path.exists('/bin/systemctl'):
                return
            # Check service state and disable
            try:
                cmd = '/bin/systemctl -q is-active what-just-happened.service'
                output = int(subprocess.call(cmd.split()))  # 0 - if active, non-zero if inactive
            except subprocess.CalledProcessError:
                self._logger.info("Failed to check status of what-just-happened.service")
                return

            if not output:      #  what-just-happened.service is running
                self._logger.info("Platform CL - what-just-happened.service is running, attempting to disable it")
                # Disable Service
                try:
                    stop_cmd = '/bin/systemctl stop what-just-happened.service'
                    disable_cmd = '/bin/systemctl disable what-just-happened.service'
                    utils.subprocess_wrapper(stop_cmd.split())
                    utils.subprocess_wrapper(disable_cmd.split())
                    if hasattr(self, '__wjh_process'):
                        self.__wjh_process.terminate()
                    _is_deliberate_wjh_disable = True
                except subprocess.CalledProcessError:
                    self._logger.info("Failed to stop what-just-happened.service")
                    return
        elif platform_type == PlatformType.SONIC:
            # Check service state and disable
            if process_feature_status(self._logger, 'what-just-happened'):
                self._logger.info("Platform SONiC - what-just-happened feature is enabled on SONiC, attempting to disable it")
                try:
                    disable_cmd = 'config feature state what-just-happened disabled'
                    utils.subprocess_wrapper(disable_cmd.split())
                    if hasattr(self, '__wjh_process'):
                        self.__wjh_process.terminate()
                except subprocess.CalledProcessError:
                    self._logger.info("Failed to disable what-just-happened feature")
                    return

    def _start_spice(self):
        agent_config = configuration.get_config(_PROC_NAME)
        opta_server = agent_config[ConfKw.NETQ_AGENT][ConfKw.SERVER]

        spice_enabled = agent_config.get(ConfKw.NETQ_AGENT).get(ConfKw.IS_SPICE_ENABLED,
                                                                ConfKw.DEFAULT_IS_SPICE_ENABLED)
        is_force_spice_enabled = agent_config.get(ConfKw.NETQ_AGENT).get(ConfKw.IS_FORCE_SPICE_ENABLED,
                                                                         ConfKw.DEFAULT_IS_FORCE_SPICE_ENABLED)
        if is_force_spice_enabled:
            return self.enable_spice()

        cli_config = configuration.get_config(ConfKw.NETQD)
        cli_server_ip = cli_config.get(ConfKw.NETQ_CLI).get(ConfKw.SERVER)
        is_cloud_cli_configured = "cumulusnetworks.com" in cli_server_ip or\
              "nvidia.com" in cli_server_ip

        if not spice_enabled:
            return None

        # check TS is cloud
        is_cloud = is_opta_server_cloud(opta_server)
        self._logger.debug("server {0} is cloud: {1}".format(opta_server, is_cloud))

        if not is_cloud and not is_cloud_cli_configured:
            self._logger.info("spice: Skipped enabling Amber collection")
            return None

        return self.enable_spice()

    def enable_spice(self):
        # check CL4
        if utils.get_platform_type() != PlatformType.CL4:
            return None
        # check CL Version
        if not check_cl_version_onwards(major=5, minor=1):
            return None
        # check Mellanox Platform
        mlx = 'mlnx' in self.__platform
        if not mlx:
            return None
        self._logger.info("spice: starting amber data collection")
        spice_agent = SpiceCollectionAgent(self._logger, self.__platform)
        spice_process = Process(target=spice_agent.run)
        spice_process.start()
        time.sleep(0.1)
        if spice_process.is_alive():
            msg_string = netq_agent_get_subprocess_status_str("Spice", spice_process.pid, "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_SPICE_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)
        else:
            msg_string = netq_agent_get_subprocess_status_str("Spice", -1, "Failed")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_SPICE_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                            msg_string)
            self._logger.info("spice: process didn't start")
        return spice_process

    def __is_spice_process_active(self):
        if self.__spice_process.is_alive():
            if self.__SpiceLoggedOnce is True:
                self._logger.debug("spice Process is alive {}".
                                   format(self.__spice_process.pid))
                self.__SpiceLoggedOnce = False
            return True
        else:
            if self.__spice_restart_hold == 0:
                if self.__SpiceLoggedOnce is False:
                    self._logger.info("spice Process is not alive {}".
                                      format(self.__spice_process.pid))
                    self.__SpiceLoggedOnce = True
            return False

    def write_log_exporter_conf_file(self, filename, input_path, exclude_path, output_file):
        # Use placeholders in the input and output strings

        input_str = textwrap.dedent(
            """
            [INPUT]
                Name        tail
                Path        {}
                Exclude_Path        {}
                Read_from_Head        True
            """.format(input_path, exclude_path)
            )

        output_str = textwrap.dedent(
            """
            [OUTPUT]
                Name         file
                Match        *
                File         {}
            """.format(output_file)
            )

        # Open the file in write mode
        with open(filename, "w") as conf_file:
            # Write the input and output strings to the file
            conf_file.write(input_str)
            conf_file.write(output_str)

        self._logger.debug("Configuration written to capture log from {}".format(input_path))
        conf_file.close()

    def _start_log_exporter(self):
        self._fluent_bit_path = DefaultPaths.FLUENT_BIT_PATH
        self._fluent_bit_conf = DefaultPaths.FLUENT_BIT_CONF_PATH

        conf_dict = configuration.get_config(_PROC_NAME)
        enabled = conf_dict.get('netq-agent').get(ConfKw.IS_LOG_EXPORTER_ENABLED, ConfKw.DEFAULT_IS_LOG_EXPORTER_ENABLED)

        msg_string = netq_agent_get_subprocess_status_str("Log Exporter", -1, "NA/Not Enabled")
        netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_LOG_EXPORTER_STATUS,
                                        NetqAgentState.NETQ_AGENT_STATE_NA,
                                        msg_string)

        if not enabled:
            return None

        if not os.path.exists(self._fluent_bit_path):
            self._logger.info("Unable to find {}".format(self._fluent_bit_path))
            return None

        # check TS is cloud
        opta_server = conf_dict[ConfKw.NETQ_AGENT][ConfKw.SERVER]
        is_cloud = is_opta_server_cloud(opta_server)
        self._logger.debug("server {0} is cloud: {1}".format(opta_server, is_cloud))
        if not is_cloud:
            self._logger.debug("server {0} is not cloud deployment.. skip exporting logs".format(opta_server))
            return None

        # Write the log export config file
        log_files_to_export = conf_dict.get('netq-agent').get(ConfKw.LOG_EXPORTER_FILES_TO_EXPORT,
                                                              ConfKw.DEFAULT_LOG_EXPORTER_FILES_TO_EXPORT)

        log_files_to_exclude = conf_dict.get('netq-agent').get(ConfKw.LOG_EXPORTER_FILES_TO_EXCLUDE,
                                                               ConfKw.DEFAULT_LOG_EXPORTER_FILES_TO_EXCLUDE)

        self.write_log_exporter_conf_file(self._fluent_bit_conf, log_files_to_export, log_files_to_exclude,
                                          DefaultPaths.LOG_EXPORTER_PRODUCER_FILE_PATH)

        #clear out the contents of any previous unexported logs
        #Scenario:
        #  say user was monitoring syslog, but want to modify only switchd.log,
        #  retaining old logs will send syslogs after restart.

        try:
            if os.path.exists(DefaultPaths.LOG_EXPORTER_PRODUCER_FILE_PATH):
                with open(DefaultPaths.LOG_EXPORTER_PRODUCER_FILE_PATH, "w") as logExporterProdFile:
                    logExporterProdFile.truncate(0)
        except: #pylint: disable=bare-except
            self._logger.debug("{} is not present, will create new one".
                               format(DefaultPaths.LOG_EXPORTER_PRODUCER_FILE_PATH))

        try:
            if os.path.exists(DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH):
                with open(DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH, "w") as logExporterConFile:
                    logExporterConFile.truncate(0)
        except: #pylint: disable=bare-except
            self._logger.debug("{} is not present, will create new one".
                               format(DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH))

        self._logger.debug("Starting Log collection")
        log_exporter_cmd = [self._fluent_bit_path, "-c", self._fluent_bit_conf]
        log_exporter_process = subprocess.Popen(log_exporter_cmd, stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE)

        if log_exporter_process.poll() is not None:
            msg_string = netq_agent_get_subprocess_status_str("Log Exporter", -1, "Failed")
            self._logger.debug(msg_string)
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_LOG_EXPORTER_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                            msg_string)
            return None
        else:
            msg_string = netq_agent_get_subprocess_status_str("Log Exporter", log_exporter_process.pid,
                                                              "Alive")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_LOG_EXPORTER_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                            msg_string)


        return log_exporter_process

    def __is_log_exporter_process_active(self):
        ret_code = self.__log_exporter_process.poll()
        if ret_code is not None:
            self._logger.warn("Log exporter process is not alive")
            return False

        self._logger.debug("Log Exporter Process is alive {}".
                           format(self.__log_exporter_process.pid))
        return True

    def _asic_monitor_service_start(self):
        asic_monitor_service = None
        conf_dict = configuration.get_config(_PROC_NAME)
        config_enabled = conf_dict.get('netq-agent').get(ConfKw.IS_ASIC_MONITOR_ENABLED, ConfKw.DEFAULT_IS_ASIC_MONITOR_ENABLED)
        simulation = Platform.VX in self.__platform
        platform_mellanox_or_simulation = 'mlnx' in self.__platform or simulation
        platform_supported = platform_mellanox_or_simulation and (PlatformType.CL4 in utils.get_platform_type())

        if platform_supported and config_enabled:
            monitor_min_file_count = conf_dict.get('netq-agent').get(ConfKw.ASIC_MONITOR_MIN_FILE_COUNT, ConfKw.DEFAULT_ASIC_MONITOR_MIN_FILE_COUNT)
            monitor_collection_periodicity = conf_dict.get('netq-agent').get(ConfKw.ASIC_MONITOR_COLLECTION_PERIODICITY, ConfKw.DEFAULT_ASIC_MONITOR_COLLECTION_PERIODICITY)
            asic_monitor_service = NetQAsicMonitor(self._logger, monitor_min_file_count=monitor_min_file_count, monitor_collection_periodicity=monitor_collection_periodicity, simulation=simulation)
            asic_monitor_service.monitor_init()
            if asic_monitor_service is not None:
                msg_string = 'Asic monitor service - started'
                self._logger.info(msg_string)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ASIC_MONITOR_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                                msg_string)
            else:
                msg_string = 'Asic monitor service - not started'
                self._logger.info(msg_string)
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ASIC_MONITOR_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
        else:
            msg_string = 'Asic monitor service - enabled: {}'.format(config_enabled and platform_supported)
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_ASIC_MONITOR_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_NA,
                                            msg_string)

        return asic_monitor_service

    def _asic_monitor_service_stop(self):
        if self.__asic_monitor_service is None:
            return
        self.__asic_monitor_service.monitor_deinit()
        self._logger.info("asic-monitor service - stopped")
        return

    def _run(self):
        """Once upon a time...
        """
        # start with agent in paused state
        self.__AGENT_PAUSE = True
        self.__GNMI_AGENT_PAUSE = True
        self.__AGENT_PAUSE_FIRST_RUN = False
        self.__FIRST_RUN = False
        self.__LAST_AGENT_STATE_RUNNING = False
        self.__NETLINK_INIT_DONE = False
        self.__NETLINK_RESTART = False

        self._old_log_level = 'info'

        self._logger.info('{} started'.format(_PROC_NAME))

        cpu_quota = utils.get_cpu_quota()
        if cpu_quota:
            self._logger.info("Configured CPU Quota : {:.1f}%".format(cpu_quota))
            if not utils.is_cpuquota_supported():
                if utils.is_cpulimit_installed():
                    actual_cpu_quota = int(cpu_quota/cpu_count())
                    if actual_cpu_quota == 0:
                        actual_cpu_quota = 1
                    cmd = "cpulimit -p " + str(os.getpid()) + " -l " + str(actual_cpu_quota)
                    self._run_command(cmd)
                else:
                    self._logger.warn("cpulimit not installed")
        else:
            self._logger.info("CPU Quota not configured")
        self._logger.info("CPU Architecture {}".format(utils.get_cpu_architecture()))

        self._spawn_cli()

        # Start asic monitor service
        self.__asic_monitor_service = self._asic_monitor_service_start()
        atexit.register(self._asic_monitor_service_stop)

        self.__config_commands.validate_command_files()

        epoch = time.time()
        timers = {'pipeline': 0, 'services': 0, 'liveliness': 0,
                  'inactivity': 0, 'serverinfo': 0, 'frr_docker': 0,
                  'platform': epoch, 'nlpoll': 0, 'ng_agent': 0,
                  'wjh_service': 0, 'appinfo': 0}

        while True:
            eventlet.sleep(1)
            epoch = time.time()

            if not self.__is_grpc_process_active() or not self.__is_libnl_process_active():
                if self.__wjh_process:
                    self.__wjh_process.join()
                return

            # perform hb check
            self._handle_hb_timer(timers, epoch)

            # perform common processes for OPTA and gNMI
            self._handle_ng_agent_timer(timers, epoch)
            self._handle_wjh_service_monitor_timer(timers, epoch)

            conf_dict = configuration.get_config(_PROC_NAME)
            new_log_level = conf_dict.get('netq-agent').get(ConfKw.LOG_LEVEL, ConfKw.DEFAULT_LOG_LEVEL)
            self.__logger_coalescer.update(conf_dict)

            if self._old_log_level != new_log_level:
                logger = utils.get_logger(_PROC_NAME)
                self._logger.info('updating log level from {} to {}'\
                    .format(self._old_log_level, new_log_level))
                utils.update_logger_level(logger, new_log_level)
                self._old_log_level = new_log_level

            if self.__config_commands.poll_command_file():
                self.__commands = []
                self.__load_commands()

            self.__is_sx_sdk_running()
            if self.__wjh_process and not self.__is_wjh_process_active():
                msg_string = netq_agent_get_subprocess_status_str("WJH", -1, "Not alive")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_WJH_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self.__logger_coalescer.info(msg_string, group=NetqLogGroup.WJH_STATUS)
                if self.__wjh_restart_hold == 0:
                    self.__wjh_restart_hold = epoch
                elif epoch - self.__wjh_restart_hold > self.__WJH_RESTART_HOLD_PERIOD:
                    self.__wjh_process = self._start_wjh()
                    self.__wjh_restart_hold = 0
            if self.__spice_process and not self.__is_spice_process_active():
                msg_string = netq_agent_get_subprocess_status_str("Spice", -1, "Not alive")
                netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_SPICE_STATUS,
                                                NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                                msg_string)
                self.__logger_coalescer.info(msg_string, group=NetqLogGroup.SPICE_STATUS)
                if self.__spice_restart_hold == 0:
                    self.__spice_restart_hold = epoch
                elif epoch - self.__spice_restart_hold > 10:
                    self.__spice_process = self._start_spice()
                    self.__spice_restart_hold = 0
            if not self.__AGENT_PAUSE or not self.__GNMI_AGENT_PAUSE:
                self._monitor_connected_state()
            if (self.__AGENT_PAUSE or self.__GNMI_AGENT_PAUSE) and not self._is_reset_allowed():
                continue

            if (self.__opta_enabled and self.__AGENT_PAUSE) or \
                    (self.__gnmi_enabled and self.__GNMI_AGENT_PAUSE):
                if self.__LAST_AGENT_STATE_RUNNING:
                    self.__last_reinit = time.time()

                # Initialize the routing protocol map
                self.__init_rt_protos()

                self.__commands = []
                self.__load_commands()

                # some event set the agent pause to true
                self.__reinitialize_db()

                if self.__AGENT_PAUSE:
                    # we hit a snag while re-init the db.
                    # retry after sleeping...zzz
                    eventlet.sleep(random.randrange(5, (Heartbeat.LIVELINESS_RATE / 2), 2))
                    continue
                elif not self.__FIRST_RUN:
                    # Assemble the data to be pushed
                    self.__FIRST_RUN = True

                self.__LAST_AGENT_STATE_RUNNING = True
                # Set epoch only for platform, as it is read as part of __reinitialize_db()
                timers = {'pipeline': 0, 'services': 0, 'liveliness': 0,
                          'inactivity': 0, 'serverinfo': 0, 'frr_docker': 0,
                          'platform': epoch, 'nlpoll': 0, 'ng_agent': 0,
                          'wjh_service': 0, 'appinfo': 0, 'asic_monitor': 0}

            elif self.__NETLINK_RESTART and not self._is_libnl_enabled and self._is_netlink_enabled:
                # kick off netlink thread
                self._logger.info("Initializing netlink thread")
                self.__init_netlink_thread()

            self._handle_appinfo_timer(timers, epoch)

            # always perform server info check
            self._handle_server_info_timer(timers, epoch)

            if not self.__NETLINK_INIT_DONE and not self._is_libnl_enabled and self._is_netlink_enabled:
                # netlink sync is not done
                # ignore remainder loop tasks
                continue

            self._handle_command_timer(timers, epoch)

            self._handle_inactivity_timer(timers, epoch)

            if self.__AGENT_PAUSE:
                continue

            self._handle_platform_timer(timers, epoch)

            self._handle_asic_monitor_service_timer(timers, epoch)

            if not self._is_libnl_enabled and self._is_netlink_enabled:
                self._handle_nlpoll_timer(timers, epoch)
            else:
                self._handle_netlink_libnl_timer(timers, epoch)

            self._handle_frr_docker_timer(timers, epoch)

            self._handle_services_timer(timers, epoch)

            # self._handle_nlflush_timer(timers, epoch)

    def __get_command_period(self, cmd_key):
        if cmd_key == "dom":
            return dom.dom.DOM_PROCESS_DURATION
        elif is_sonic() and cmd_key == "proc-net-dev":
            return sonic_proc.proc.PROC_PROCESS_DURATION
        else:
            return self.__config_commands.get_command_period(cmd_key)


def start_ng_agent_service(logger, gnmi_enabled, opta_enabled, conf_dict):
    try:
        logger.info("Starting netq-ng-agent service")
        args = ["/usr/sbin/netq-ng-agent"]
        if opta_enabled:
            opta_server = conf_dict[ConfKw.NETQ_AGENT][ConfKw.SERVER]
            opta_port = conf_dict[ConfKw.NETQ_AGENT][ConfKw.PORT]
            args.extend(["-opta_address", opta_server,
                         "-opta_port", str(opta_port)])
        popen = subprocess.Popen(args, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        return popen


    except Exception as e:
        logger.error("Failed to run netq-ng-agent process: %s", str(e))


def is_libnl_required():
    is_opta = os.path.isfile('/etc/app-release')
    if is_opta:
        return False
    return (is_cl3() or is_cl4() or is_sonic() or is_ubuntu() or is_aarch64())

def main2():
    conf_dict = configuration.get_config(_PROC_NAME)
    log_obj = utils.get_logger(_PROC_NAME, conf_dict)

    # Parent Process
    if conf_dict.get('netq-agent'):
        vrf = conf_dict.get('netq-agent').get('vrf')

    log_obj.info("Starting Agent - Vrf: {}".format(vrf))

    opta_enabled = (conf_dict.get('netq-agent', {}).get(
        ConfKw.IS_OPTA_ENABLED, ConfKw.DEFAULT_IS_OPTA_ENABLED))
    gnmi_enabled = is_gnmi_enabled(conf_dict)
    if not opta_enabled and not gnmi_enabled:
        log_obj.error("One of the following OPTA or gNMI "
                      "agent must be enabled to start netq-agent")
        netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_CONFIG_VALIDATION_STATUS,
                                        NetqAgentState.NETQ_AGENT_STATE_FAILURE,
                                        'One of the following OPTA or gNMI agent must be enabled to start netq-agent')
        return -1

    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_CONFIG_VALIDATION_STATUS,
                                    NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                    'OPTA enabled: {}, gNMI enabled: {}. Config checked... OK'.format(opta_enabled, gnmi_enabled))
    pq, bp_pq = create_message_queues(conf=conf_dict, logger=log_obj)
    gnmi_pq = create_gnmi_message_queues(conf=conf_dict, logger=log_obj) \
        if gnmi_enabled else None

    if opta_enabled and \
            conf_dict.get('netq-agent').get(ConfKw.OPTA_DISCOVERY_SERVERS):
        # Perform OPTA Discovery
        log_obj.info("Initiating OPTA Discovery")
        conf_dict[ConfKw.NETQ_AGENT][ConfKw.ENABLE_OPTA_DISCOVERY] = True
        discovery_client = NetQAgentOPTADiscovery(conf_dict, log_obj, pq, bp_pq)

        while True:
            result = discovery_client.send_discovery_request()
            if result == AGENTSTATUS.SUCCESS:
                break
            time.sleep(3)   # Adding delay upon failure

        conf_dict[ConfKw.NETQ_AGENT][ConfKw.SERVER] = discovery_client.opta_instance_ip
        conf_dict[ConfKw.NETQ_AGENT][ConfKw.PORT] = discovery_client.opta_instance_port

    ng_agent_process = None
    if (is_cl4() or is_sonic()) and 'arm' not in utils.get_cpu_architecture():
        ng_agent_process = start_ng_agent_service(
            log_obj, gnmi_enabled, opta_enabled, conf_dict)

    stream_process = None

    is_libnl_enabled = conf_dict.get('netq-agent').get(ConfKw.IS_LIBNL_ENABLED, ConfKw.DEFAULT_IS_LIBNL_ENABLED)
    if is_libnl_enabled and not is_libnl_required():
        is_libnl_enabled = False

    if utils.is_hostname_default(HOSTNAME):
        is_libnl_enabled = False

    netlink_process = None
    grpc_server_inst = NetQAgentGrpcClient(conf_dict, log_obj, pq, bp_pq, gnmi_pq)
    stream_process = \
        Process(target=grpc_server_inst.start_stream, args=(conf_dict,))
    stream_process.start()
    if is_libnl_enabled:
        agent_pq = create_agent_info_queue(conf=conf_dict, logger=log_obj)
        netlink_inst = NetLinkProc(conf_dict, log_obj, pq, bp_pq, agent_pq, gnmi_pq)
        netlink_process = \
            Process(target=netlink_inst.run, args=(conf_dict,))
        netlink_process.start()
        cldbpush_inst = _ClDbPush(conf_dict, stream_process, netlink_process, ng_agent_process)
    else:
        cldbpush_inst = _ClDbPush(conf_dict, stream_process, None, None)
    ret_val = cldbpush_inst.run()

    # For graceful shutdown
    if stream_process:
        stream_process.join()

    if is_libnl_enabled and netlink_process:
        netlink_process.join()

    if ng_agent_process:
        ng_agent_process.wait()
        log_obj.debug("netq-ng-agent process finished")

    global _is_deliberate_wjh_disable  #pylint: disable=global-statement

    if _is_deliberate_wjh_disable:
        try:
            _start_cl_wjh_service()
        except Exception as e:
            log_obj.info("Start CL WJH failed: {}".format(e))

        _is_deliberate_wjh_disable = False

    return ret_val

def _start_cl_wjh_service():
    enable_cmd = '/bin/systemctl enable what-just-happened.service'
    start_cmd = '/bin/systemctl start what-just-happened.service'
    utils.subprocess_wrapper(enable_cmd.split())
    utils.subprocess_wrapper(start_cmd.split())

def _start_netlink():

    netq_cmd = ['/usr/share/venvs/netq-agent/bin/python', '/usr/sbin/netq-agent-netlink', '>', '/dev/null 2>&1']

    time.sleep(3)
    out = subprocess.call(netq_cmd)

    return out

def spawn_process():

    conf_dict = configuration.get_config(_PROC_NAME)
    if configuration.is_cl4() or configuration.is_cl3() or is_sonic() or is_aarch64():
        if conf_dict.get('netq-agent'):
            config_vrf = conf_dict.get('netq-agent').get('vrf')
        if not config_vrf:
            config_vrf = 'default'
        cmd = ['ip', 'vrf', 'exec', config_vrf, '/usr/share/venvs/netq-agent/bin/python', '/usr/sbin/netq-sub-agent']
        #cmd = ['ip', 'vrf', 'exec', config_vrf, '/usr/local/bin/valgrind', '--leak-check=summary',    '--log-file=valgrind-out.txt', '/usr/share/venvs/netq-agent/bin/python', '/usr/sbin/netq-sub-agent']
    else:
        cmd = ['/usr/share/venvs/netq-agent/bin/python', '/usr/sbin/netq-sub-agent']

    # agent_monitor = NetqAgentMonitor(conf_dict)
    # agent_monitor.start()

    out = subprocess.call(cmd)
    return out


def main():
    spawn_process()
    return
