#
# Copyright (c) 2019-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import socket
import json
import time
import os
import inspect
import netq_lib.orm.redisdb.models
from netq_lib.common.nlfullstate import NetlinkFullStateObject, FullState
from netq_lib.orm.redisdb.models import ( #pylint: disable=unused-import
    Address,
    ACLResource,
    AgentStats,
    ARConfigGlobalV2,
    ARConfigIntf,
    AREcmpInfo,
    ASIC,
    AsicMonitorBufferStats,
    AsicMonitorHistogram,
    BgpSession,
    Board,
    BtrFsInfo,
    Channel,
    CheckCatalog,
    CheckFilter,
    CheckFilterObject,
    ClagSession,
    ClSupport,
    Command,
    ConfigMon,
    CPU,
    CustomerPreference,
    Description,
    Disk,
    DockerContainer,
    DockerHost,
    DockerNetwork,
    DockerPortMap,
    DockerService,
    DockerSwarmCluster,
    DockerSwarmNode,
    Dom,
    EcmpHash,
    EcmpInfo,
    EsxHost,
    EsxVmAttr,
    EsxVmBind,
    EthtoolInterface,
    Event,
    EventIds,
    EventsConfig,
    EventsConfigObject,
    EventsConfigScope,
    EventsHierarchy,
    EvpnVniInfo,
    Fan,
    Filter,
    ForwardingResource,
    Interface,
    Inventory,
    KubeCluster,
    KubeDaemonSet,
    KubeDeployment,
    KubeNode,
    KubePod,
    KubeReplicaSet,
    KubeReplicationController,
    KubeService,
    Label,
    License,
    Link,
    Liveness,
    Lldp,
    LogExporter,
    Manifest,
    MacFdb,
    Memory,
    MstpInfo,
    Neighbor,
    Netconf,
    Nexthop,
    Node,
    Notification,
    Notifier,
    Ntp,
    NvueMon,
    NvueRoceConfig,
    OS,
    OspfIf,
    OspfNbr,
    PackageInfo,
    Port,
    PortStatsInterface,
    Power,
    ProcNetDevInterface,
    PSU,
    Ptm,
    PtpClockInfo,
    PtpCorrection,
    PtpGlobalConfig,
    PtpPortStatistics,
    PtpPortStatus,
    ResourceUtil,
    Roce,
    RoceConfig,
    Route,
    Rule,
    RunningConfigMon,
    Search,
    Services,
    ServiceStatistics,
    ServiceStatisticsAggrgate,
    SpiceAmberGlobal,
    SpiceAmberPortLevel,
    SsdUtil,
    StreamingCheck,
    StreamingCheckAgent,
    StreamingCheckBgp,
    StreamingCheckEvpn,
    StreamingCheckInterface,
    StreamingCheckMlag,
    StreamingCheckMtu,
    StreamingCheckNtp,
    StreamingCheckRoCE,
    StreamingCheckSensors,
    StreamingCheckVlan,
    StreamingCheckVxlan,
    Tca,
    TcaConfig,
    TcaScope,
    Temp,
    Trace,
    Validation,
    Vlan,
    VxlanRemoteDest,
    WjhDropAggregate,
    WjhDropRawInfo,
    WjhL1DropAggregate,
    _Inventory,
    _Kube,
    _Sensor,
    HostDevicesRoce,
    IbInterface,
    StreamingCheckArConfig
)
from netq_lib.common.utils import is_cl3, is_cl4, is_sonic, is_ubuntu, is_aarch64

if is_cl3() or is_cl4() or is_sonic() or is_ubuntu() or is_aarch64():
    from netq_agent.cmd.netq_agent_libnl import NetLinkProc, CommandServer, AgentCommands #pylint: disable=unused-import



class AgentCli(object):
    __MAX_LENGTH = 4096

    __SERVER = '/run/netq_cli_server'

    __NON_CLS_MODELS = ['_RedisModel', 'RedisQuery', 'Connection', 'MemModel',
                        'Node', 'Liveness', 'ProcNetDevInterface', 'datetime',
                        'NetQPriorityQueue', 'QueueElem', 'NetQIPC']

    __NETLINK_MODELS = ['Link', 'Route', 'MacFdb', 'VxlanRemoteDest', 'Neighbor', 'Netconf', 'Address']

    def __init__(self, pool, logger, nlmonitor, is_libnl_enabled, cmd_client):
        self.__commands = [""]
        self.__supported_models = []
        self.__pool = pool
        self._logger = logger
        self.__nlmonitor = nlmonitor
        self.full_state_object = NetlinkFullStateObject
        self._is_libnl_enabled = is_libnl_enabled
        self._cmd_client = cmd_client
        if is_sonic:
            self.__NETLINK_MODELS.remove('MacFdb')

    def _write_to_file(self, buf, objs):

        # now = datetime.now()
        # file_name = buf + "_" + now.strftime("%m_%d_%Y_%H_%M_%S")
        abs_path = "/var/log/netq-model-" + buf
        try:
            out_file = open(abs_path, "w+")
        except Exception as e:
            self._logger.error("Unable to write to dump file {}".format(e))
            return

        out_file.write("[")
        for obj in objs:
            outstr = json.dumps(obj.to_json())
            out_file.write(outstr)
            out_file.write(",\n")
        out_file.write("]")

        out_file.close()
        return

    @FullState("Route")
    def __netlink_get_all_routes(self):
        """ Get netlink neigh
        """
        self.__pool.spawn(self.__nlmonitor.get_all_routes).wait()
        self.__nlmonitor.set_mcast_group('route')

    def get_nlmonitor(self):
        return self.__nlmonitor


    def __get_model_object(self, nl_only=False):
        try:
            for cln, cls in inspect.getmembers(netq_lib.orm.redisdb.models):
                if cln in self.__NON_CLS_MODELS:
                    continue
                if nl_only and cln not in self.__NL_CLS_MODELS: #pylint: disable=no-member
                    continue
                if inspect.isclass(cls) and hasattr(cls, 'memmodel'):
                    mcls = cls.memmodel()
                    yield cls
        except Exception as ex:
            self._logger.info('Could not find Class. Error: %s', ex)

    def _dump_all_objects(self):

        for model_obj in self.__get_model_object():
            if (self._is_libnl_enabled and model_obj.__name__ in self.__NETLINK_MODELS):
                continue
            model_objs = self.__get_mem_objects(model_obj.__name__)
            if model_objs:
                self._write_to_file(model_obj.__name__, model_objs)

        if self._is_libnl_enabled:
            cmd = "{}".format(AgentCommands.DUMP_NETLINK_MSG)
            self._cmd_client.send_command(cmd)

    def __process_cli_command(self, clientsocket):
        try:
            while True:
                buf = clientsocket.recv(self.__MAX_LENGTH).decode('utf-8')
                if buf == '':
                    return  # client terminated connection
                self._logger.info("Received Command to dump {}".format(buf))
                if buf == "all":
                    self._dump_all_objects()
                else:
                    all_objs = self.__get_mem_objects(buf)
                    if all_objs:
                        self._write_to_file(buf, all_objs)
        finally:
            clientsocket.close()

    def __get_mem_objects(self, objectname, generator=True):
        try:
            if objectname in ("Route", "Routes"):
                mcls = Route.memmodel()
            elif objectname == "Command":
                mcls = Command.memmodel()
            elif objectname == "Description":
                mcls = Description.memmodel()
            elif objectname == "Link":
                mcls = Link.memmodel()
            elif objectname == "MacFdb":
                mcls = MacFdb.memmodel()
            elif objectname == "Neighbor":
                mcls = Neighbor.memmodel()
            elif objectname == "Node":
                mcls = Node.memmodel()
            elif objectname == "Services":
                mcls = Services.memmodel()
            elif objectname == "VxlanRemoteDest":
                mcls = VxlanRemoteDest.memmodel()
            elif objectname == "Netconf":
                mcls = Netconf.memmodel()
            elif objectname == "Address":
                mcls = Address.memmodel()
            elif objectname == "HostDevicesRoce":
                mcls = HostDevicesRoce.memmodel()
            elif objectname == "IbInterface":
                mcls = IbInterface.memmodel()
            elif objectname == "StreamingCheckArConfig":
                mcls = StreamingCheckArConfig.memmodel()
            else:
                mcls = eval(objectname).memmodel() #pylint: disable=eval-used
        except Exception:
            self._logger.error("Unable to find obj {}".format(objectname))
            return None
        if generator:
            return mcls.query.all()
        else:
            all_objs = [obj for obj in mcls.query.all()]
        return all_objs

    def __compare_route(self, old_obj, new_obj):

        if ((old_obj.vrf == new_obj.vrf) and (old_obj.prefix == new_obj.prefix)):
            return True
        else:
            return False

    def __compare_routes(self, old_objs, new_objs):

        old_stale = []

        if not old_objs or not new_objs:
            self._logger.info("Ignoring initial runs")
            return

        valid_routes = new_objs
        inner_compare = False

        for old_obj in old_objs:
            inner_compare = False
            for idx, new_obj in enumerate(valid_routes):
                cmp_res = self.__compare_route(old_obj, new_obj)
                if cmp_res:
                    inner_compare = True
                    del valid_routes[idx]
                    break
            if inner_compare is False:
                self._logger.info("Old Object no match prefix {} vrf {}".format(old_obj.prefix, old_obj.vrf))
                old_stale.append(old_obj)

        if ((len(old_stale) > 0) or (len(valid_routes) > 0)): #pylint: disable=len-as-condition
            self._logger.info("Error found \n")
            self._write_to_file("old_stale", old_stale)
            self._write_to_file("new_stale", valid_routes)
            return
        else:
            self._logger.info("Old objects matches recent ones")

    def __process_periodic_runner(self):
        self.__RUN_PERIOD = 30
        object_name = "Route"

        while True:

            time.sleep(5)

            old_objs = self.__get_mem_objects(object_name, False)
            if old_objs is None:
                continue

            if object_name == "Route":
                self.__netlink_get_all_routes()

                new_objs = self.__get_mem_objects(object_name, False)

                self.__compare_routes(old_objs, new_objs)

    def process_cli_server(self):

        # Make sure the socket does not already exist
        try:
            os.unlink(self.__SERVER)
        except OSError:
            if os.path.exists(self.__SERVER):
                raise

        serversocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        serversocket.bind(self.__SERVER)
        serversocket.listen(10)

        # debug_thread = self.__pool.spawn(self.__process_periodic_runner)

        while True:
            (clientsocket, address) = serversocket.accept()
            self.__pool.spawn(self.__process_cli_command, clientsocket)
