#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import random
import signal
import time
from abc import ABCMeta, abstractmethod

import yaml

from netq_lib.orm.redisdb.models import WjhDropAggregate, WjhL1DropAggregate



class BaseWjhClient(object, metaclass=ABCMeta):
    def __init__(self, logger):
        self._logger = logger
        self.agg_cache = {}
        self.running = False
        signal.signal(signal.SIGTERM, self.sig_handler)
        signal.signal(signal.SIGINT, self.sig_handler)

    def flush_agg_cache(self):
        for key, agg in list(self.agg_cache.items()):
            self._logger.debug("Aggregate type: {}, count: {} first {} last {}".format(agg.drop_type, agg.agg_count, agg.first_timestamp, agg.last_timestamp))
            agg.save()

        self.agg_cache.clear()
        mcls = WjhDropAggregate.memmodel()
        mcls.flush()

    @abstractmethod
    def start_wjh_client(self):
        raise NotImplementedError

    def sig_handler(self, signum, frame):
        if self.running:
            self._logger.info("Interrupt/Terminate received "
                              "for wjh process, terminating process")
            self.running = False



class DummyWjhClient(BaseWjhClient):

    def start_wjh_client(self):
        class _raw_info(object):
            def __init__(self, reason='', reason_id=0, drop_type='', ingress_port='',
                         egress_port='', traffic_class=0, acl_rule_id=0,
                         acl_bind_point=0, acl_name='', acl_rule='', smac='',
                         dmac='', sip='', dip='', proto=0, sport=0, dport=0,
                         fts=0.0, lts=0.0, agg_count=0, severity='', tc=0,
                         latency=0, occupancy=0):
                self.drop_agg = WjhDropAggregate()
                self.drop_agg.reason = reason
                self.drop_agg.reason_id = reason_id
                self.drop_agg.drop_type = drop_type
                self.drop_agg.ingress_port = ingress_port
                self.drop_agg.egress_port = egress_port
                self.drop_agg.traffic_class = traffic_class
                self.drop_agg.acl_rule_id = acl_rule_id
                self.drop_agg.acl_bind_point = acl_bind_point
                self.drop_agg.acl_name = acl_name
                self.drop_agg.acl_rule = acl_rule
                self.drop_agg.smac = smac
                self.drop_agg.dmac = dmac
                self.drop_agg.sip = sip
                self.drop_agg.dip = dip
                self.drop_agg.proto = proto
                self.drop_agg.sport = sport
                self.drop_agg.dport = dport
                self.drop_agg.agg_count = agg_count
                self.drop_agg.first_timestamp = fts
                self.drop_agg.last_timestamp = lts
                self.drop_agg.severity = severity
                if drop_type == 'Buffer':
                    self.drop_agg.traffic_class = tc
                    self.drop_agg.original_latency = latency
                    self.drop_agg.original_occupancy = occupancy

        self._logger.info('Started DUMMY wjh agent')
        try:
            data_file = '/etc/netq/.wjh.data'
            with open(data_file, 'r') as f:
                drop_data = yaml.load(f, Loader=yaml.FullLoader) #pylint: disable=no-member
                # allow all processes to start
                time.sleep(5)
                if drop_data:
                    l1_data = drop_data.get('L1-drops')
                    other_data = drop_data.get('Other-drops')

                    t1 = time.time()
                    if other_data:
                        for ddict in other_data:
                            self._logger.debug('--- Others')
                            d = ddict['drop']
                            self._logger.debug('{}'.format(d))
                            fts = t1 - (random.random() * 20)
                            lts = t1 + (random.random() * 20)
                            raw_info = _raw_info(reason=d.get('reason', ''),
                                                 reason_id=d.get('reason_id', 0),
                                                 drop_type=d.get('drop_type', ''),
                                                 ingress_port=d.get('ingress_port', ''),
                                                 egress_port=d.get('egress_port', ''),
                                                 traffic_class=0 if not d.get('traffic_class') else int(d.get('traffic_class')),
                                                 acl_rule_id=0 if not d.get('acl_rule_id') else int(d.get('acl_rule_id')),
                                                 acl_bind_point=0 if not d.get('acl_bind_point') else int(d.get('acl_bind_point')),
                                                 acl_name=d.get('acl_name', ''),
                                                 acl_rule=d.get('acl_rule', ''),
                                                 smac=d.get('smac', ''),
                                                 dmac=d.get('dmac', ''),
                                                 proto=0 if not d.get('proto') else int(d.get('proto')),
                                                 sport=0 if not d.get('sport') else int(d.get('sport')),
                                                 dport=0 if not d.get('dport') else int(d.get('dport')),
                                                 sip=d.get('sip', ''),
                                                 dip=d.get('dip', ''),
                                                 fts=fts,
                                                 lts=lts,
                                                 agg_count=0 if not d.get('count') else int(d.get('count')),
                                                 severity=d.get('severity', ''),
                                                 tc=d.get('tc', 0),
                                                 latency=d.get('original_latency', 0),
                                                 occupancy=d.get('original_occupancy', 0))
                            drop_agg = raw_info.drop_agg
                            key = (drop_agg.reason, drop_agg.ingress_port, drop_agg.sip, drop_agg.dip,
                                   drop_agg.proto, drop_agg.sport, drop_agg.dport, drop_agg.smac,
                                   drop_agg.dmac)
                            self.agg_cache[key] = drop_agg
                            self._logger.debug('--- Others saved')
                    if l1_data:
                        for ddict in l1_data:
                            self._logger.debug('--- L1')
                            d = ddict['drop']
                            self._logger.debug('{}'.format(d))
                            fts = t1 - (random.random() * 20)
                            lts = t1 + (random.random() * 20)

                            L1drop = WjhL1DropAggregate()
                            L1drop.ingress_port = d.get('ingress_port', '')
                            L1drop.is_port_up = d.get('is_port_up') == 'True' or 'False'
                            L1drop.port_down_reason = d.get('port_down_reason', '')
                            L1drop.description = ''
                            if d.get('state_change_count'):
                                L1drop.state_change_count = int(d.get('state_change_count'))
                            if d.get('symbol_error_count'):
                                L1drop.symbol_error_count = int(d.get('symbol_error_count'))
                            if d.get('crc_error_count'):
                                L1drop.crc_error_count = int(d.get('crc_error_count'))
                            L1drop.first_timestamp = fts
                            L1drop.last_timestamp = lts
                            L1drop.save()
                            self._logger.debug('--- L1 saved')

        except Exception as e:
            self._logger.error('Error parsing wjh data file: {}'.format(e))
        self.running = True

        # Flush events every 15 seconds
        check_flush = 0
        flush_interval = 3

        sleep_interval_seconds = 5
        while self.running:
            time.sleep(sleep_interval_seconds)
        while self.running:
            time.sleep(sleep_interval_seconds)
            check_flush += 1
            if check_flush == flush_interval:
                check_flush = 0
                self.flush_agg_cache()
