#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



import eventlet
import eventlet.pools
import os
import logging
from netq_lib.orm import HOSTNAME

from eventlet import greenthread
from netq_lib.orm.redisdb.models import (
    KubeCluster,
    KubeDaemonSet,
    KubeDeployment,
    KubeNode,
    KubePod,
    KubeReplicationController,
    KubeReplicaSet,
    KubeService
)

SUCCESS = True
FAILED = False

class KubeMonitor(object):
    """Class KubeMonitor
    This class provides methods to monitor kubernetes cluster
    - The monitoring is done local to a kubernetes master
    - It listens to kubernetes events through API server
        - Tracks Nodes, Services, Pods
    """
    #Default Kubernetes Polling Period
    _KUBERNETES_DEFAULT_POLL_PERIOD = 30
    _KUBERNETES_MIN_POLL_PERIOD = 10

    def parse_config(self, config):
        '''
        Validate user hasn't specified any parameter that
        netq agent does not understand
        '''
        valid_keys = set(['poll_period', 'enable', 'config'])
        self.__enabled = False

        for key in list(config.keys()):
            if key not in valid_keys:
                self.__logger.error("Invalid configuration: %r" % key)
                return False

        self._poll_period = config.get('poll_period',
                                       KubeMonitor._KUBERNETES_DEFAULT_POLL_PERIOD)

        if self._poll_period < KubeMonitor._KUBERNETES_MIN_POLL_PERIOD:
            self.__logger.error("Configured poll_period:%r Support min poll_period:%r. Applying min poll" %
                                (self._poll_period, KubeMonitor._KUBERNETES_MIN_POLL_PERIOD))
            self._poll_period = KubeMonitor._KUBERNETES_MIN_POLL_PERIOD

        self.__enabled = config.get('enable', False)
        self.__config = config.get('config', '/etc/kubernetes/admin.conf')
        return


    @staticmethod
    def is_platform_supported(platform):
        supported = False
        platform_string = platform.lower()

        if 'ubuntu' in platform_string:
            supported = True
        elif 'centos' in platform_string:
            supported = True
        elif 'rhel' in platform_string:
            supported = True
        elif 'red hat' in platform_string:
            supported = True
        return supported


    def __init__(self, yaml_config, logger, platform=None, pool=None):
        self.__logger = logger
        self.__running = False
        self.__enabled = False
        self.__monitoring_events = False
        self.__api_client = None
        self.__api_server = ''
        self.__config = None
        self.__config_ts = 0
        self._host = HOSTNAME
        self._gt = None
        self._event_handler_gt = None
        self.__pool = None
        self.__log_once = True

        # 1. Check if Kubernetes monitoring is supported on the platform
        if not platform:
            self.__logger.info("Platform is not defined. Skip Kubernetes Monitoring...")
            return

        if not self.is_platform_supported(platform):
            self.__logger.info("Kubernetes Monitoring not Supported on %r..." % platform)
            return

        # 2. Check if docker monitoring is configured and enabled
        if yaml_config and 'kubernetes' in yaml_config:
            kube_config = yaml_config['kubernetes']
            self.parse_config(kube_config)

        if not self.__enabled:
            self.__logger.info("Kubernetes Monitoring not enabled...")
            return

        self.__pool = pool or eventlet.GreenPool()
        return

    @property
    def poll_period(self):
        return self._poll_period

    @property
    def enabled(self):
        return self.__enabled

    @property
    def running(self):
        return self.__running

    def flush_cluster_info(self, cluster_list):
        m_KubeCluster = KubeCluster.memmodel()

        for entry in m_KubeCluster.query.filter(hostname=HOSTNAME):
            if entry.cluster_name in cluster_list:
                continue
            old_entry = KubeCluster()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_node_info(self, node_list):
        m_KubeNode = KubeNode.memmodel()

        for entry in m_KubeNode.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.node_name) in node_list:
                continue
            old_entry = KubeNode()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_pod_info(self, pod_list):
        m_KubePod = KubePod.memmodel()

        for entry in m_KubePod.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.namespace, entry.name) in pod_list:
                continue
            old_entry = KubePod()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_service_info(self, service_list):
        m_KubeService = KubeService.memmodel()

        for entry in m_KubeService.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.namespace, entry.name) in service_list:
                continue
            old_entry = KubeService()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_replication_controller_info(self, rc_list):
        m_KubeReplicationController = KubeReplicationController.memmodel()

        for entry in m_KubeReplicationController.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.namespace, entry.name) in rc_list:
                continue
            old_entry = KubeReplicationController()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_replica_set_info(self, replica_set_list):
        m_KubeReplicaSet = KubeReplicaSet.memmodel()

        for entry in m_KubeReplicaSet.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.namespace, entry.name) in replica_set_list:
                continue
            old_entry = KubeReplicaSet()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_daemon_set_info(self, daemon_set_list):
        m_KubeDaemonSet = KubeDaemonSet.memmodel()

        for entry in m_KubeDaemonSet.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.namespace, entry.name) in daemon_set_list:
                continue
            old_entry = KubeDaemonSet()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def flush_deployment_info(self, deployment_list):
        m_KubeDeployment = KubeDeployment.memmodel()

        for entry in m_KubeDeployment.query.filter(hostname=HOSTNAME):
            if (entry.cluster_name, entry.namespace, entry.name) in deployment_list:
                continue
            old_entry = KubeDeployment()
            old_entry.copy(entry)
            old_entry.delete()
        return


    def update_cluster(self, node_list):
        try:
            v1api = globals()['client'].CoreV1Api(api_client=self.__api_client)
            component_statuses = v1api.list_component_status()
        except Exception as e:
            # 
            # This can happen when kubernetes API server
            # cannot be reached 
            #
            self.__logger.debug("Failed to get component status...{0}".format(e))
            # return kubeCluster_list

        etcd_status = ''
        controller_manager = ''
        scheduler = ''

        for cs in component_statuses.items:
            cluster_name = cs.metadata.cluster_name
            name = cs.metadata.name
            message = []
            status = []
            status_type = []
            for condition in cs.conditions:
                message.append(condition.message)
                status_type.append(condition.type)
                status.append(condition.status)
            if 'etcd' in name:
                etcd_status = str(' '.join(status_type))
            if 'controller-manager' in name:
                controller_manager = str(' '.join(status_type))
            if 'scheduler' in name:
                scheduler = str(' '.join(status_type))

        kubeCluster = KubeCluster()
        kubeCluster.api_server = self.__api_server
        kubeCluster.cluster_name = cluster_name or 'default'
        kubeCluster.nodes = str(' '.join(node_list))
        kubeCluster.etcd = etcd_status
        kubeCluster.controller_manager = controller_manager
        kubeCluster.scheduler = scheduler
        kubeCluster.save()
        return (kubeCluster.cluster_name)


    def update_node(self, node):
        spec = node.spec
        pod_cidr = spec.pod_cidr or ''
        metadata = node.metadata
        cluster_name = metadata.cluster_name

        name = metadata.name
        if not name:
            self.__logger.debug("Skip updating node - name undefined")
            return

        role = 'worker'
        label_list = []

        for label, value in list(metadata.labels.items()):
            if 'node-role.kubernetes.io' in label:
                label_tuple = label.split('/')
                if len(label_tuple) > 1:
                    role = label_tuple[1]

            label_str = "%s:%s" % (label.encode('ascii', 'ignore'),
                           value.encode('ascii', 'ignore'))
            label_list.append(label_str)

        node.status.node_info.machine_id.encode('ascii', 'ignore')
        kube_proxy_version = node.status.node_info.kube_proxy_version.encode('ascii', 'ignore')
        kubelet_version = node.status.node_info.kubelet_version.encode('ascii', 'ignore')
        container_runtime = node.status.node_info.container_runtime_version.encode('ascii', 'ignore')
        reason = 'Unknown'

        for condition in node.status.conditions:
            if condition.type == 'Ready':
                reason = condition.reason.encode('ascii', 'ignore')

        if reason == 'KubletReady':
            reason = 'Ready'

        kubeNode = KubeNode()
        kubeNode.api_server = self.__api_server
        kubeNode.cluster_name = cluster_name or 'default'
        kubeNode.node_name = name
        kubeNode.pod_cidr = pod_cidr.encode('ascii', 'ignore')
        kubeNode.role = role.encode('ascii', 'ignore')
        kubeNode.labels = ' '.join(label_list)
        kubeNode.status = reason
        kubeNode.kube_proxy_version = kube_proxy_version
        kubeNode.kubelet_version = kubelet_version
        kubeNode.container_runtime = container_runtime
        kubeNode.start_time = str(metadata.creation_timestamp.replace(tzinfo=None)) or ''
        kubeNode.save()
        if self.__log_once:
            self.__logger.info("Kubelet version {}: {}".format(name, kubelet_version))
            self.__log_once = False
        return (kubeNode.cluster_name, kubeNode.node_name)


    def update_pod(self, pod):
        metadata = pod.metadata
        name = metadata.name
        if not name:
            self.__logger.debug("Skipping pod, name not initialized")
            return

        cluster_name = metadata.cluster_name
        namespace = metadata.namespace

        owner_references = metadata.owner_references
        owner_kind = owner_name = ''

        if owner_references:
            for owner in owner_references:
                owner_kind = owner.kind
                owner_name = owner.name

        container_list = []
        container_statuses = pod.status.container_statuses or []
        for container in container_statuses:
            container_name = container.name
            container_id = container.container_id
            if container_id:
                container_id = container_id.split('/')
                c = '%s:%s' % (container_name, container_id[2])
                container_list.append(c)

        spec = pod.spec
        node_name = spec.node_name
        pod_ip = pod.status.pod_ip
        if not pod_ip:
            pod_ip = ''
        status = pod.status.phase
        reason = pod.status.reason
        if reason == 'NodeLost':
            status = reason
        pod_template_hash = ''

        label_list = []
        try:
            for label, value in list(metadata.labels.items()):
                if 'pod-template-hash' == label:
                    pod_template_hash = str(value)
                else:
                    label_str = "%s:%s" % (label, value)
                    label_list.append(label_str)
        except Exception:
            self.__logger.debug("pod labels are missing for pod {0}... ".format(name))

        kubePod = KubePod()
        kubePod.api_server = self.__api_server
        kubePod.cluster_name = cluster_name or 'default'
        kubePod.namespace = namespace or ''
        kubePod.name = name
        kubePod.template_hash = pod_template_hash
        kubePod.node_name = node_name or ''
        kubePod.containers = str(' '.join(container_list))
        kubePod.ip_address = str(pod_ip)
        kubePod.owner_kind = owner_kind or ''
        kubePod.owner_name = owner_name or ''
        kubePod.status = status or ''
        kubePod.start_time = str(pod.status.start_time.replace(tzinfo=None)) or ''
        kubePod.labels = str(' '.join(label_list))
        kubePod.save()
        return (kubePod.cluster_name, kubePod.namespace, kubePod.name)


    def update_service(self, service):
        spec = service.spec
        metadata = service.metadata
        cluster_name = metadata.cluster_name
        namespace = metadata.namespace
        if not metadata.name:
            self.__logger.debug("Skip Service update. Name is not defined")
            return

        label_list = []
        labels = spec.selector
        if labels:
            for label, value in list(labels.items()):
                label_list.append('%s:%s' % (label, value))

        port_list = []
        if spec.ports:
            for port in spec.ports:
                port_list.append('%s:%s' % (port.protocol, port.port))

        kubeService = KubeService()
        kubeService.api_server = self.__api_server
        kubeService.cluster_name = cluster_name or 'default'
        kubeService.namespace = namespace or ''
        kubeService.name = metadata.name
        kubeService.service_type = spec.type
        kubeService.labels = str(' '.join(label_list))
        kubeService.cluster_ip = spec.cluster_ip or ''
        kubeService.start_time = str(metadata.creation_timestamp.replace(tzinfo=None)) or ''
        if not spec.external_i_ps:
            external_ip = ''
        else:
            external_ip = spec.external_i_ps

        if type(external_ip) == list:
            kubeService.external_ip = str(','.join(external_ip))
        elif type(external_ip) == str:
            kubeService.external_ip = external_ip

        kubeService.ports = str(' '.join(port_list))
        kubeService.save()
        return (kubeService.cluster_name, kubeService.namespace, kubeService.name)


    def update_deployment(self, deployment):
        metadata = deployment.metadata
        name = metadata.name
        if not name:
            self.__logger.debug("Skip deployment, name not defined")
            return
        cluster_name = metadata.cluster_name
        namespace = metadata.namespace

        label_list = []
        if metadata.labels:
            for label, value in list(metadata.labels.items()):
                label_str = '%s:%s' % (label, value)
                label_list.append(label_str)

        spec = deployment.spec
        status = deployment.status

        kubeDeployment = KubeDeployment()
        kubeDeployment.api_server = self.__api_server
        kubeDeployment.cluster_name = cluster_name or 'default'
        kubeDeployment.namespace = namespace or ''
        kubeDeployment.name = name
        kubeDeployment.labels = str(' '.join(label_list))
        kubeDeployment.replicas = spec.replicas or 0
        kubeDeployment.ready_replicas = status.ready_replicas or 0
        kubeDeployment.start_time = str(metadata.creation_timestamp.replace(tzinfo=None)) or ''
        kubeDeployment.save()
        return (kubeDeployment.cluster_name, kubeDeployment.namespace, kubeDeployment.name)


    def update_replication_controller(self, rc):
        metadata = rc.metadata
        name = metadata.name
        if not name:
            self.__logger.debug("Skip replication controller, name not defined")
            return
        cluster_name = metadata.cluster_name
        namespace = metadata.namespace

        spec = rc.spec
        replicas = spec.replicas
        selector = spec.selector
        status = rc.status
        label_list = []
        if selector:
            for label, value in list(selector.items()):
                label_list.append('%s:%s' % (label, value))
        status = rc.status
        ready_replicas = status.ready_replicas

        kubeReplicationController = KubeReplicationController()
        kubeReplicationController.api_server = self.__api_server
        kubeReplicationController.cluster_name = cluster_name or 'default'
        kubeReplicationController.namespace = namespace or ''
        kubeReplicationController.name = name
        kubeReplicationController.labels = str(' '.join(label_list))
        kubeReplicationController.replicas = replicas or 0
        kubeReplicationController.ready_replicas = ready_replicas or 0
        kubeReplicationController.start_time = str(metadata.creation_timestamp.replace(tzinfo=None)) or ''
        kubeReplicationController.save()
        return (kubeReplicationController.cluster_name,
                kubeReplicationController.namespace,
                kubeReplicationController.name)


    def update_replica_set(self, rs):
        metadata = rs.metadata
        name = metadata.name
        if not name:
            self.__logger.debug("Skip replica set, name not defined")
            return
        cluster_name = metadata.cluster_name
        namespace = metadata.namespace

        spec = rs.spec
        replicas = spec.replicas
        selector = spec.selector
        status = rs.status

        label_list = []
        pod_template_hash = ''
        if selector and selector.match_labels:
            for label, value in list(selector.match_labels.items()):
                if 'pod-template-hash' == label:
                    pod_template_hash = str(value)
                else:
                    label_list.append('%s:%s' % (label, value))

        status = rs.status
        ready_replicas = status.ready_replicas

        kubeReplicaSet = KubeReplicaSet()
        kubeReplicaSet.api_server = self.__api_server
        kubeReplicaSet.cluster_name = cluster_name or 'default'
        kubeReplicaSet.namespace = namespace or ''
        kubeReplicaSet.name = name
        kubeReplicaSet.pod_template_hash = pod_template_hash
        kubeReplicaSet.labels = str(' '.join(label_list))
        kubeReplicaSet.replicas = replicas or 0
        kubeReplicaSet.ready_replicas = ready_replicas or 0
        kubeReplicaSet.start_time = str(metadata.creation_timestamp.replace(tzinfo=None)) or ''
        kubeReplicaSet.save()
        return (kubeReplicaSet.cluster_name,
                kubeReplicaSet.namespace,
                kubeReplicaSet.name)


    def update_daemon_set(self, ds):
        metadata = ds.metadata
        name = metadata.name
        if not name:
            self.__logger.debug("Skip daemon set, name not defined")
            return
        cluster_name = metadata.cluster_name
        namespace = metadata.namespace

        status = ds.status
        desired_count = status.desired_number_scheduled
        ready_count = status.number_ready

        label_list = []
        if metadata.labels:
            for label, value in list(metadata.labels.items()):
                label_list.append('%s:%s' % (label, value))

        kubeDaemonSet = KubeDaemonSet()
        kubeDaemonSet.api_server = self.__api_server
        kubeDaemonSet.cluster_name = cluster_name or 'default'
        kubeDaemonSet.namespace = namespace or ''
        kubeDaemonSet.name = name
        kubeDaemonSet.labels = str(' '.join(label_list))
        kubeDaemonSet.desired_count = desired_count or 0
        kubeDaemonSet.ready_count = ready_count or 0
        kubeDaemonSet.start_time = str(metadata.creation_timestamp.replace(tzinfo=None)) or ''
        kubeDaemonSet.save()
        return (kubeDaemonSet.cluster_name,
                kubeDaemonSet.namespace,
                kubeDaemonSet.name)


    def poll_kubernetes_nodes(self):
        node_list = []
        node_obj_list = []
        cluster_list = []

        try:
            v1api = globals()['client'].CoreV1Api(api_client=self.__api_client)
            nodes = v1api.list_node()
        except Exception as e:
            # 
            # Catch case where kube api server 
            # connection fails 
            #
            self.__logger.debug("Failed to get Node List: {0}...".format(e))
            return (FAILED, cluster_list, node_list)

        try:
            for node in nodes.items:
                node_obj = self.update_node(node)
                node_list.append(node_obj[1])
                node_obj_list.append(node_obj)

            if node_list:
                cluster_obj = self.update_cluster(node_list)
                cluster_list.append(cluster_obj)
            else:
                self.__logger.debug("Null Node List. Skip updating cluster...")

        except Exception as e:
            self.__logger.error("Failed to get Kubernetes nodes: %r" % e)

        return (SUCCESS, cluster_list, node_obj_list)


    def poll_kubernetes_pods(self):
        pod_list = []
        try:
            v1api = globals()['client'].CoreV1Api(api_client=self.__api_client)
            pods = v1api.list_pod_for_all_namespaces()
        except Exception as e:
            # 
            # Catch case where kube api server 
            # connection fails 
            #
            self.__logger.debug("Failed to get pods: {0}...".format(e))
            return (FAILED, pod_list)

        try:
            for pod in pods.items:
                pod_obj = self.update_pod(pod)
                pod_list.append(pod_obj)
        except Exception as e:
                self.__logger.debug("Failed to get pod information...")

        return (SUCCESS, pod_list)


    def poll_kubernetes_services(self):
        service_list = []

        try:
            v1api = globals()['client'].CoreV1Api(api_client=self.__api_client)
            services = v1api.list_service_for_all_namespaces()
        except Exception as e:
            # 
            # Catch case where kube api server 
            # connection fails 
            #
            self.__logger.debug("Failed to get Services: {0}...".format(e))
            return (FAILED, service_list)

        for service in services.items:
            service_obj = self.update_service(service)
            service_list.append(service_obj)

        return (SUCCESS, service_list)


    def poll_kubernetes_replication_controller(self):
        rc_list = []
        try:
            v1api = globals()['client'].CoreV1Api(api_client=self.__api_client)
            replication_controller = \
                    v1api.list_replication_controller_for_all_namespaces()
        except Exception as e:
            # 
            # Catch case where kube api server 
            # connection fails 
            #
            self.__logger.debug("Failed to get Replication Controller: {0}...".format(e))
            return (FAILED, rc_list)

        for rc in replication_controller.items:
            rc_obj = self.update_replication_controller(rc)
            rc_list.append(rc_obj)

        return (SUCCESS, rc_list)


    def poll_kubernetes_deployments(self):
        deployment_list = []

        try:
            v1api = globals()['client'].AppsV1Api(api_client=self.__api_client)
            deployments = v1api.list_deployment_for_all_namespaces()
        except Exception as e:
            # 
            # Catch case where kube api server 
            # connection fails 
            #
            self.__logger.debug("Failed to get deployment : {0}...".format(e))
            return (FAILED, deployment_list)

        for deployment in deployments.items:
            deployment_obj = self.update_deployment(deployment)
            deployment_list.append(deployment_obj)
        return(SUCCESS, deployment_list)


    def poll_kubernetes_replica_set(self):
        rs_list = []

        try:
            v1api = globals()['client'].AppsV1Api(api_client=self.__api_client)
            replica_set = v1api.list_replica_set_for_all_namespaces()
        except Exception as e:
            self.__logger.debug("Failed to get replica set: {0}...".format(e))
            return (FAILED, rs_list)

        for rs in replica_set.items:
            rs_obj = self.update_replica_set(rs)
            rs_list.append(rs_obj)

        return (SUCCESS, rs_list)


    def poll_kubernetes_daemon_sets(self):
        ds_list = []

        try:
            v1api = globals()['client'].AppsV1Api(api_client=self.__api_client)
            daemon_set = v1api.list_daemon_set_for_all_namespaces()
        except Exception as e:
            self.__logger.debug("Failed to get daemon set: {0}...".format(e))
            return (FAILED, ds_list)

        for ds in daemon_set.items:
            ds_obj = self.update_daemon_set(ds)
            ds_list.append(ds_obj)

        return (SUCCESS, ds_list)


    def _run(self):
        """
        Periodically monitor Kubernetes Master
        """
        self.__running = True
        self.__api_client = None
        KUBERNETES_INSTALLED = False

        while True:

            if self.__enabled:
                if not KUBERNETES_INSTALLED:
                    #
                    # Lazy import kubernetes python library 
                    try:
                        from kubernetes import client, config, watch  # noqa
                        kubernetes = __import__('kubernetes')
                        globals()['client'] = kubernetes.client
                        globals()['config'] = kubernetes.config
                        globals()['watch'] = kubernetes.watch
                        KUBERNETES_INSTALLED = True
                    except ImportError as e:
                        self.__logger.debug("Kubernetes module load failed  {0}".format(e))
                        KUBERNETES_INSTALLED = False
            else:
                KUBERNETES_INSTALLED = False
                self.__api_client = None
                self.__logger.debug("Kubernetes is not installed ...")


            if KUBERNETES_INSTALLED and os.path.exists(self.__config):
                # 
                # If Kubernetes is installed and kube configuration has 
                # changed load configuration again as certificates/credentials 
                # might have changed
                config_ts = os.stat(self.__config).st_mtime

                if self.__config_ts != config_ts:
                    self.__logger.debug("Kube Config has changed. Loading configuration {0}...".format(self.__config))
                    try:
                        self.__api_client = config.new_client_from_config(config_file=self.__config, persist_config=False)
                        self.__config_ts = config_ts
                        # Suppress warnings from urllib3 and request library 
                        logging.getLogger("urllib3").setLevel(logging.ERROR)
                        logging.getLogger("requests").setLevel(logging.ERROR)
                    except Exception as e:
                        self.__logger.error("Failed to initialize Kubernetes Client: {0}".format(e))
                        self.__api_client = None
                        self.__config_ts = 0

                    try:
                        self.__api_server = self.__api_client.configuration.host.split('//')[1].split(':')[0]
                        self.__logger.info("Connected to Kube API Server: {0}".format(self.__api_server))
                    except Exception as e:
                        self.__logger.error("Failed to Parse API Server info: {0}".format(e))
                        self.__api_server = ''
                        self.__api_client = None
                        self.__config_ts = 0
            else:
                self.__api_client = None
                self.__config_ts = 0
                self.__logger.debug("Kubernetes configuration not found...")
                #
                # Flush all kube objects from this master
                #
                self.flush_cluster_info([])
                self.flush_node_info([])
                self.flush_pod_info([])
                self.flush_service_info([])
                self.flush_replication_controller_info([])
                self.flush_replica_set_info([])
                self.flush_deployment_info([])
                self.flush_daemon_set_info([])


            if self.__api_client:
                #
                # Poll kubernetes resources only if client is initialized
                #
                cluster_nodes = self.poll_kubernetes_nodes()
                pods = self.poll_kubernetes_pods()
                services = self.poll_kubernetes_services()
                replica_controllers = self.poll_kubernetes_replication_controller()
                replica_sets = self.poll_kubernetes_replica_set()
                deployments = self.poll_kubernetes_deployments()
                daemon_sets = self.poll_kubernetes_daemon_sets()

                self.flush_cluster_info(cluster_nodes[1])
                self.flush_node_info(cluster_nodes[2])
                self.flush_pod_info(pods[1])
                self.flush_service_info(services[1])
                self.flush_replication_controller_info(replica_controllers[1])
                self.flush_replica_set_info(replica_sets[1])
                self.flush_deployment_info(deployments[1])
                self.flush_daemon_set_info(daemon_sets[1])

            try:
                greenthread.sleep(self._poll_period)
            except Exception as e:
                self.__logger.info("Kubernetes Poll Caught Exception {0}".format(e))
                break

        self.__logger.info('Poll loop terminated!!')
        return

    def start(self):
        """Kubernetes Monitor threads
        """
        # If Kubernetes monitoring isn't enabled skip
        if not self.__enabled:
            self.__logger.info('Kubernetes monitor is not enabled ')
            return

        if not self.__running:
            self.__logger.info('Starting Kubernetes Poll thread...')
            self.__config_ts = 0
            self.__api_client = None
            self._gt = self.__pool.spawn(self._run)
            self._gt.link(self.stop_poll_checker)
        else:
            self.__logger.info('Kubernetes Poll thread is already running...')
        return


    def stop_poll_checker(self, *args, **kwargs):
        """
        stop_poll_checker()
        should be called when Kubernetes poll thread is stopped/killed
        """
        if self.__running:
            self.__logger.info('Kubernetes Poll thread stopped')
            self.__running = False
        return


    def stop_monitor_checker(self, *args, **kwargs):
        """
        stop_monitor_checker()
        should be called when Kubernetes monitor event thread is stopped/killed
        """
        if self.__monitoring_events:
            self.__logger.info('Kubernetes Monitor event thread stopped')
            self.__monitoring_events = False
        return


    def halt(self):
        if self._gt:
            self.__logger.info('Terminating Kubernetes monitor: %r' % self._gt)
            eventlet.kill(self._gt, eventlet.StopServe)
        self._gt = None
        return
