#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import json
import importlib

from netq_lib.common.utils import save_fullstate
from netq_lib.orm.redisdb.models import Command
from netq_lib.common.netq_commands import NetqFileManager, NetqSupportedCommands, NetqCommandInfo
from netq_lib.common.enums import CommandParams


class NetqCommands(object):

    def __init__(self, conf, logger):
        self.__parsed_cmds = {}
        self._conf = conf
        self._logger = logger
        self.__prev_cmd = {}
        self.__fm = NetqFileManager(conf, logger)
        self.__sup_cmds = NetqSupportedCommands()
        self.__command_info = {}

    def __is_json(self, cmd_out):
        try:
            cmd_out = cmd_out.strip()
            if len(cmd_out) == 0: #pylint: disable=len-as-condition
                return False, cmd_out
            else:
                json_object = json.loads(cmd_out)
        except Exception as e:
            self._logger.error("Unsupported Command output {}".format(e))
            return False, cmd_out
        return True, json_object

    def __remove_ignored_objects(self, json_object, info):
        if isinstance(json_object, dict):
            for key in list(json_object.keys()):
                if key in info.ignore:
                    del json_object[key]
                else:
                    self.__remove_ignored_objects(json_object[key], info)
        return json_object

    def findDiff(self, d1, d2, path=""):
        for k in d1.keys():
            if k not in d2.keys():
                print(path, ":")
                print(k + " as key not in d2", "\n")
            else:
                if type(d1[k]) is dict: #pylint: disable=unidiomatic-typecheck
                    if path == "":
                        path = k
                    else:
                        path = path + "->" + k
                    self.findDiff(d1[k], d2[k], path)
                else:
                    if d1[k] != d2[k]:
                        print(path, ":")
                        print(" - ", k, " : ", d1[k])
                        print(" + ", k, " : ", d2[k])

    def __cache_n_save(self, json_out, info):
        self._logger.info("Cache and save for {}".format(info.key))
        newcmd = Command()
        newcmd.key = info.key
        newcmd.output = json_out
        newcmd.message_type = info.key
        newcmd.save()

    def process_cmd_out(self, cmd_out, info):
        """

        :param cmd_out:
        :param info:
        """
        is_json, json_object = self.__is_json(cmd_out)
        self._logger.debug("Key {}".format(info.key))
        if is_json:
            valid_result = self.__remove_ignored_objects(json_object, info)
            self.__cache_n_save(json.dumps(valid_result).encode("utf-8"), info)

    def __update_callback(self, commands):
        self._logger.debug("Commands {}".format(commands))
        for command in commands:
            if not command.get('isactive'):
                continue
            if command.get('parser') == "local":
                command_key = command.get('key')
                callback = self.__sup_cmds.service_callback(command_key)
                if command.get('callback'):
                    callback = command.get('callback')
                    self._logger.info("Using platform specific configured callback in /var/run/netq/netq_commands.yml for {} as {}".format(command_key, callback))
                if callback:
                    try:
                        module = importlib.import_module('netq_agent.parsers.' + callback.rsplit('.', 1)[0])
                        command['callback'] = getattr(module, callback.split('.')[-1])
                        if command['callback']:
                            self._logger.info("Using Local callback for {}".format(command_key))
                    except ModuleNotFoundError as err:
                        self._logger.error("Unable to load callback! - {}".format(err))
                else:
                    self._logger.debug("Unable to find the command {} in supported commands".format(command_key))

    @staticmethod
    def __get_deleted_entries(exist, new):
        exist_set = set(exist.keys())
        new_set = set(new.keys())
        return exist_set - new_set

    @staticmethod
    def __compare_cmd_info(old, new):
        modified_objs = []
        for key, value in new.items():
            if key in old and value != old[key]:
                modified_objs.append(key)
            elif key not in old:
                modified_objs.append(key)
        return modified_objs

    def __get_inactive_entries(self, cmd_info):
        inactive_set = set()
        for key, obj in cmd_info.items():
            if not obj.is_active:
                if key in self.__command_info and \
                        obj.is_active != self.__command_info[key].is_active:
                    inactive_set.add(key)
        return inactive_set

    @staticmethod
    def __dispatch_delete(model_array):
        for model in model_array:
            full_state_array = []
            # mcls = model.memmodel()
            # for obj in mcls.query.all():
            #    obj.active = False
            #    obj.deleted = True
            #    full_state_array.append(obj)
            save_fullstate(full_state_array, model().__class__.__name__)

    def __send_disable_fullstate(self, disable_keys):
        for key in disable_keys:
            model_array = self.__sup_cmds.get_models(key)
            if model_array:
                self.__dispatch_delete(model_array)

    def __update_cmd_info(self, cmd_info):
        self.__command_info = cmd_info

    def __handle_disabled_keys(self, cmd_info, reference):

        deleted_keys = self.__get_deleted_entries(reference, cmd_info)
        disabled_keys = self.__get_inactive_entries(cmd_info)
        disable_keys = deleted_keys.union(disabled_keys)
        self._logger.debug("Disabled keys {}".format(disable_keys))
        self.__send_disable_fullstate(disable_keys)
        return disable_keys

    def __process_file_modification(self):
        cmdlist = {}
        rv = self.__parse_commands(cmdlist, predicate=True)

        if rv is False:
            return False

        cmd_info = self.__get_command_info(cmdlist)
        disable_keys = self.__handle_disabled_keys(cmd_info, self.__command_info)
        modified_keys = self.__compare_cmd_info(self.__command_info, cmd_info)
        self._logger.debug("Modified keys {}".format(modified_keys))

        self.__update_cmd_info(cmd_info)

        return len(disable_keys) + len(modified_keys) > 0

    def poll_command_file(self): #pylint: disable=inconsistent-return-statements
        """

        :return: Boolean result whether the file has been modified or not
        """
        try:
            if self.__fm.is_file_modified():
                return self.__process_file_modification()
        except Exception as e:
            self._logger.info("Config File not found {}".format(e))
            self.__fm.get_configured_commands()
            return False
            # return self.__process_file_modification()

    def __get_command_info(self, cmdlist):

        cmd_info_dict = {}

        for service_name, commands in cmdlist.items():
            for command in commands:
                key = command.get("key")
                parser = command.get("parser")
                period = command.get("period")
                is_active = command.get("isactive")
                cmd_value = command.get("command")
                if cmd_value == "None":
                    command["command"] = None
                cmd = command.get("command")
                cmd_info_dict[key] = NetqCommandInfo(service_name, key, period, cmd, is_active, parser)

        return cmd_info_dict

    @staticmethod
    def get_valid_commands(commands):
        """

        :param commands: set of test commands
        :return: only active commands
        """
        out_cmd = []
        for command in commands:
            if command.get("isactive"):
                out_cmd.append(command)
        return out_cmd

    def __parse_commands(self, cmdlist, predicate=False):
        commands = self.__fm.get_configured_commands(predicate)
        if not commands:
            self._logger.error("Unable to get any command in the configuration")
            return False

        for entry in commands[CommandParams.NETQCOMANDS]:
            service = entry.get('service', 'misc')
            commands = entry.get('commands', [])
            if commands:
                self.__update_callback(commands)
                if service not in cmdlist:
                    cmdlist[service] = []
                cmdlist[service] = self.get_valid_commands(commands)
        if not self.__command_info:
            cmd_info = self.__get_command_info(cmdlist)
            self.__command_info = cmd_info

        return True

    def init_commands(self, cmdlist):
        """
        This the anchor function which is called when the agent boots
        :param cmdlist:
        :param prdicate:
        """
        support_cmds = {}
        rv = self.__parse_commands(cmdlist)

        if rv is False:
            return

        cmd_info = self.__get_command_info(cmdlist)

        for cmd in self.__sup_cmds.service_info:
            support_cmds[cmd.service_key] = cmd
        disable_keys = self.__handle_disabled_keys(cmd_info, support_cmds)

    def get_command_period(self, key):
        cmd_info = self.__command_info.get(key, None)
        if cmd_info:
            return int(cmd_info.period)
        else:
            return 0

    def validate_command_files(self):
        self.__fm.validate_command_files()
