#
# Copyright (c) 2018-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#




import socket

from netq_lib.common.enums import ConfKw, AGENTSTATUS
from netq_lib.common.utils import retry_on_timeout
from netq_agent.interface.netq_reg_pb2 import (
    ResponseStatus
)

# GRPC Specfic Imports
import grpc
from netq_agent.interface.netq_reg_pb2 import (
    NetqAgentDiscovery
)
from netq_agent.interface.netq_reg_pb2_grpc import (
    NetqAgentDiscoveryServiceStub
)
from netq_lib.orm import HOSTNAME


class ChannelState():
    IDLE = 1
    CONNECTING = 2
    CONNECTED = 3
    DISCONNECTED = 4
    SHUTDOWN = 5


RPC_CONNECTIVITY_MAPPING = {}
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.IDLE] = ChannelState.IDLE
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.CONNECTING] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.READY] = ChannelState.CONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.TRANSIENT_FAILURE] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.SHUTDOWN] = ChannelState.IDLE


class NetQAgentDiscovery(object):

    def __init__(self, conf, logger):
        self._conf = conf
        self._channel = None
        self._logger = logger
        self.channel_state = ChannelState.IDLE

    def callback_on_exception(self, exception, traceback):

        #allowed_exceptions = ['StatusCode.UNAVAILABLE']

        self._logger.error(traceback)
        self._logger.error(exception)

        # GRPC has very weired way of generating exception with _Rendezvous
        # if exception.code() not in allowed_exceptions:
        #    raise exception

    def _run():
        pass

    @property
    def channel_state(self):
        return self._channel_state

    @channel_state.setter
    def channel_state(self, new_state):
        if new_state in RPC_CONNECTIVITY_MAPPING:
            new_state = RPC_CONNECTIVITY_MAPPING[new_state]
        self._logger.debug("OPTA Discovery Setting new RPC Channel State {}".format(new_state))
        self._channel_state = new_state

    def _get_agent_connect_id(self, verbose=False):
        id = None
        try:
            opta_url = self._conf['netq-agent'][ConfKw.SERVER]
            opta_port = self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect((opta_url, opta_port))
            id = s.getsockname()[0]
        except Exception:
            self._logger.error("Failed to connect to OPTA")

        if verbose:
            if id:
                self._logger.info("OPTA Discovery Agent Connect IP {}".format(id))
            else:
                self._logger.info("OPTA Discovery No Connect IP determined")
        return id

    def get_agent_id(self):
        if not self.is_valid_hostname(HOSTNAME):
            return None
        return HOSTNAME

    def is_valid_hostname(self, hostname):
        if hostname.lower() in ["cumulus", "sonic"]:
            return False
        return True

    def get_client_url(self, verbose=False):
        # if user has provided an ip - use that
        agenturl = None
        try:
            agenturl = self._conf['netq-agent'].get(ConfKw.AGENT_URL)
        except Exception:
            pass

        if agenturl is None:
            # figure it out
            agenturl = self._get_agent_connect_id(verbose=verbose)
        return agenturl

    def get_client_port(self):
        # if user has provided an port - use that
        agentport = None
        try:
            agentport = int(self._conf['netq-agent'].get(ConfKw.AGENT_PORT))
        except Exception:
            pass

        if agentport is None:
            agentport = ConfKw.DEFAULT_AGENT_PORT

        return agentport

    def has_exhibit_url(self):
        try:
            if (self._conf['netq-agent'].get(ConfKw.EXHIBIT_URL) and
                self._conf['netq-agent'].get(ConfKw.EXHIBIT_PORT)):
                return True
        except Exception:
            return False

    def get_exhibit_url(self):
        #Guarded by has_exhibit_url function
        return (self._conf['netq-agent'].get(ConfKw.EXHIBIT_URL))

    def get_exhibit_port(self):
        #Guarded by has_exhibit_url function
        return (self._conf['netq-agent'].get(ConfKw.EXHIBIT_PORT))

    def build_register_message(self, verbose=False):
        aid = self.get_agent_id()
        if not aid:
            if verbose:
                self._logger.info("OPTA Discovery no valid ID %s", HOSTNAME)
            return None
        if (self.has_exhibit_url()):
            if verbose:
                self._logger.info("OPTA Discovery exhibit url %s port %s",
                                         self.get_exhibit_url(),
                                         self.get_exhibit_port())
            reg_message = NetqAgentDiscovery(aid=aid,
                                         host=self.get_exhibit_url(),
                                         port=self.get_exhibit_port())
        else:
            if verbose:
                self._logger.info("OPTA Discovery agent url %s port %s",
                                         self.get_client_url(verbose=verbose),
                                         self.get_client_port())
            client_url = self.get_client_url(verbose=verbose)
            if not client_url:
                return None
            reg_message = NetqAgentDiscovery(aid=aid,
                                         host=client_url,
                                         port=self.get_client_port())
        return reg_message

    def monitor_channel(self, state):
        self._logger.debug("Channel State {}".format(state))
        self.channel_state = state

    def handle_discovery_success(self):
        self.channel_state = ChannelState.CONNECTED
        return AGENTSTATUS.SUCCESS

    def _get_opta_addr(self):
        serv_addr = "{}:{}".format(self._conf['netq-agent'][ConfKw.SERVER],
                                   self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT))
        return serv_addr

    @retry_on_timeout
    def _reestablish_rpc_channel(self):
        #opta_addr = self._get_opta_addr()
        self._channel = None

    def handle_discovery_failure(self):
        self.channel_state = ChannelState.DISCONNECTED
        return AGENTSTATUS.FAILURE

    def handle_reconnect(self, reconnect_opta_addr):
        pass

    def _parse_discovery_response(self, response):

        self._logger.debug(
            "Received Discovery Response {}".format(response.status))
        if ResponseStatus.Value('Success') == response.status:
            ret = self.handle_discovery_success()
        elif ResponseStatus.Value('Failure') == response.status:
            ret = self.handle_discovery_failure()
        elif ResponseStatus.Value('Reconnect') == response.status:
            ret = self.handle_reconnect(response.reconnect_opta_addr)
        else:
            return AGENTSTATUS.FAILURE

        return ret

    # @retry_on_timeout(callback_on_exception=callback_on_exception)

    def close(self):
        if self._channel:
            self._channel.unsubscribe(self.monitor_channel)
            self._channel.close()
        self._channel = None

    def start_discovery(self, verbose=False):

        serv_addr = self._get_opta_addr()

        try:
            self._channel = grpc.insecure_channel(serv_addr)
        except Exception as e:
            self._logger.error(
                "Unable to establish connection with OPTA {}".format(e))
            return AGENTSTATUS.FAILURE

        if self._channel:
            self._channel.subscribe(self.monitor_channel, try_to_connect=True)
        else:
            self._logger.error("Unable to connect to OPTA")
            return AGENTSTATUS.FAILURE

        opta_stub = NetqAgentDiscoveryServiceStub(self._channel)

        message = self.build_register_message(verbose=verbose)

        if not message:
            if verbose:
                self._logger.error("Unable to construct OPTA discovery message")
            return AGENTSTATUS.FAILURE

        try:
            # Set timeout to 30 secs, after which call should timeout and
            # discovery should be retried
            rsp_message = opta_stub.DiscoverMessage(message, timeout=30)
        except grpc.RpcError as e:
            self._logger.error("OPTA Discovery GRPC Error Received {}".format(e))
            return AGENTSTATUS.FAILURE
        except Exception as e:
            self._logger.error("OPTA Discovery Unknown GRPC Exception {}".format(e))
            return AGENTSTATUS.FAILURE

        ret = self._parse_discovery_response(rsp_message)

        if ret == AGENTSTATUS.FAILURE:
            self._logger.error(
                "OPTA Discovery Unknown Message Received {}".format(rsp_message))

        return ret
