#
# Copyright (c) 2020-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

#pylint: disable=undefined-variable
#pylint: disable=protected-access
#pylint: disable=no-member
#Should not be done in general, but have to include this here due to wildchar import


import copy
import os
import socket
import sys
import time
import threading
from enum import Enum
from queue import Queue
from json import dumps as json_dumps, loads as json_loads
import grpc
import netq_netlink
from netq_lib.common.netq_pq import QueueElem
from netq_lib.orm import HOSTNAME
from netq_lib.common.utils import is_sonic, is_gnmi_enabled, netq_agent_misc_warn_write, netq_agent_misc_warn_resolve_write, LogCoalescer
from netq_lib.common.enums import NetlinkMsgType, ConfKw
from netq_lib.common.enums import AGENTSTATUS, TransMode, ModelConstants as MC
from netq_agent.cmd.netq_grpc_agent import * #pylint: disable=wildcard-import
from netq_agent.parsers.sonic.ifalias import update_ifalias_information

netlink_proc_obj = None
command_server_obj = None
global_working_sequence = {}
_PROC_NAME = 'netq-agent'
class AgentCommands(Enum):
    GET_FULL_STATE_ROUTE = 1
    GET_FULL_STATE_NETCONF = 2
    GET_FULL_STATE_NEIGHBOR = 3
    GET_FULL_STATE_ADDR = 4
    GET_FULL_STATE_MACFDB = 5
    GET_FULL_STATE_LINK = 6
    GET_FULL_STATE_VXLANREMOTEDEST = 7
    DUMP_NETLINK_MSG = 8
    GET_FULL_STATE_NEXTHOP = 9


msg_count_stats = dict()


class CommandServer(object):
    __MAX_LENGTH = 4096
    __SERVER = '/run/netq_agent_netlink'
    __DELIMITER = '$$$'

    def __init__(self, conf, logger):
        global command_server_obj #pylint: disable=global-statement
        self._conf = conf
        self._logger = logger
        self._ipc_q = Queue(512)
        command_server_obj = self

    def process_agent_command(self):
        try:
            if self._ipc_q.empty():
                return
            cmd = self._ipc_q.get()
            self.__process_agent_command(cmd)
        except Exception as e:
            return

    def __process_agent_command(self, command):
        command_args = command.split(",")
        command = eval(command_args[0]) #pylint: disable=eval-used
        if command == AgentCommands.GET_FULL_STATE_ROUTE:
            netq_netlink.get_full_state_route()
        elif command == AgentCommands.GET_FULL_STATE_NEIGHBOR:
            netq_netlink.get_full_state_neigh()
        elif command == AgentCommands.GET_FULL_STATE_LINK:
            netq_netlink.get_full_state_link()
        elif command == AgentCommands.GET_FULL_STATE_ADDR:
            netq_netlink.get_full_state_addr()
        elif command == AgentCommands.GET_FULL_STATE_NETCONF:
            netq_netlink.get_full_state_netconf()
        elif command == AgentCommands.GET_FULL_STATE_MACFDB:
            netq_netlink.get_full_state_macfdb()
        elif command == AgentCommands.GET_FULL_STATE_VXLANREMOTEDEST:
            netq_netlink.get_full_state_vxlanremotedest()
        elif command == AgentCommands.GET_FULL_STATE_NEXTHOP:
            netq_netlink.get_full_state_nexthop()
        elif command == AgentCommands.DUMP_NETLINK_MSG:
            netq_netlink.dump_netlink_msg()
        else:
            self._logger.info("Unsupported Command {}".format(command))

    def send_command(self, command_info):
        cmd_str = command_info + self.__DELIMITER
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(self.__SERVER)
        except Exception as e:
            self._logger.info("Failed to connect to Server {}".format(self.__SERVER))
            raise e
        self._logger.debug("Sending Command request for {}".format(cmd_str))
        sock.send(bytes(cmd_str, 'utf-8'))
        sock.close()
        return()

    def __process_socket_msg(self, clientsocket):

        try:
            buf = clientsocket.recv(self.__MAX_LENGTH).decode('utf-8')
            if buf == '':
                return
            self._logger.debug("Buffer recevied {}".format(buf))
            commands = buf.split(self.__DELIMITER)
            for cmd in commands:
                if cmd == '':
                    continue
                self._logger.info("Processing command {}".format(cmd))
                self._ipc_q.put(cmd)
        except Exception as e:
            self._logger.info("Failed to process command {}".format(e))
        finally:
            self._logger.debug("Client Socket Disconnected")
            clientsocket.close()

    def start_server(self):
        try:
            os.unlink(self.__SERVER)
        except OSError:
            if os.path.exists(self.__SERVER):
                raise

        serversocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
           serversocket.bind(self.__SERVER)
        except socket.error as msg:
            self._logger.warn("Libnl: Socket bind failed, error code: {} msg: {}".format(str(msg[0]), msg[1]))
            return

        serversocket.listen(10)
        self._logger.info("Netlink Server Init Complete")

        while True:
            self._logger.debug("Waiting for netlink connections")
            (clientsocket, address) = serversocket.accept()
            self.__process_socket_msg(clientsocket)

class NetLinkProc(object):

    def __init__(self, conf, logger, pq, backpressure_queue, agent_pq, gnmi_pq):
        global netlink_proc_obj #pylint: disable=global-statement
        self._conf = conf
        self._channel = None
        self._logger = logger
        self._pq = pq
        self._q = Queue(250000)
        self._agent_pq = agent_pq
        self._bp_pq = backpressure_queue
        self._rt_cnt = 0
        self._grpc_object = NetQAgentGrpcClient(conf, logger, pq, backpressure_queue)
        self._wait_time = 0.0
        netlink_proc_obj = self
        self.__connectivity_state = False
        self.lock = threading.Lock()
        self.msg_count_stats = {}
        self.msg_count_stats["message_type_count"] = {}
        self.__netlink_update_time = 0.0
        self._NETLINK_UPDATE_TIME_SEC = 30.0
        self._GRPC_UPDATE_TIME = 0.5
        self._NETLINK_UPDATE_TIME = self._NETLINK_UPDATE_TIME_SEC / self._GRPC_UPDATE_TIME
        self.__logger_coalescer = LogCoalescer("netq-agent", conf_dict=configuration.get_config(_PROC_NAME))

        self.opta_enabled = self._conf.get(ConfKw.NETQ_AGENT, {}).get(
            ConfKw.IS_OPTA_ENABLED, ConfKw.DEFAULT_IS_OPTA_ENABLED)
        self.gnmi_enabled = is_gnmi_enabled(self._conf)
        if self.gnmi_enabled:
            self._gnmi_q = gnmi_pq

    def __connect_reconnect_opta(self):
        logged_prev = False
        while True:
            if AGENTSTATUS.SUCCESS != self._grpc_object.reestablish_channel():
                msg_string = "Libnl: Failed to establish Connection with OPTA, retrying with backoff"
                self._logger.info(msg_string)
                netq_agent_misc_warn_write(msg_string)
                logged_prev = True

                if self._grpc_object.get_channel_retry_count() >= self._grpc_object.max_opta_retry:
                    self._grpc_object.reset_channel_retry_count()
                else:
                    self._grpc_object.update_channel_retry_count()
                timeVal = self._grpc_object.get_grpc_backoff_time()
                time.sleep(timeVal)
                continue
            else:
                if logged_prev:
                    msg_string = "Libnl: Established connection with OPTA (Warning resolved)"
                    netq_agent_misc_warn_resolve_write(msg_string)

            self._logger.info("Starting Stream")
            break

        self._opta_stub = NetqOPTAStreamServiceStub(self._grpc_object._channel)
        self.__connectivity_state = True


    def message_builder(self, msg_type, fullstate, json_out):

        seqno = 0
        try:
            if fullstate in (TransMode.FIRST_SEGMENT, TransMode.PARTIAL_TRANSFORM):
                seqno = self._grpc_object.get_sequence_number()
                global_working_sequence[msg_type] = seqno
            else:
                seqno = global_working_sequence[msg_type]
        except Exception as e:
            global_working_sequence[msg_type] = 0
        event_time = int(time.time() * 1000)


        if is_sonic() and NetlinkMsgType(msg_type) == NetlinkMsgType.NETLINK_MSG_TYPE_LINK:
            json_out = update_ifalias_information(json_out, self._logger)

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=seqno,
                                        ts=event_time,
                                        trans_mode=fullstate,
                                        type=0,
                                        message=json_out.encode('utf-8'))
        return netq_message


    def stream_libnl_msg(self, msg_type, json_out):
        try:
            rpc_msg = self.message_builder(msg_type, False, json_out)
            yield rpc_msg
            self._grpc_object._last_transmission_time = time.time()
        except Exception as e:
            self._logger.error("Libnl: GRPC StreamMessage failed %s", e)
            self._grpc_object._last_transmission_time = time.time()
            return

    def _is_json_string(self, json_out):
        try:
            json_dict = json_loads(json_out)
        except Exception as e:
            self._logger.error("Libnl: Failed to load json string {}".format(json_out))
            return 0
        else:
            return 1

    def should_send_gnmi(self, msg_type):
        if self.gnmi_enabled and msg_type in [MC.LINK]:
            return True
        return False

    def collect_data(self, msg_type, json_out, full_state):
        is_json = self._is_json_string(json_out)
        if self.should_send_gnmi(msg_type):
            if is_json:
                if is_sonic() and NetlinkMsgType(msg_type) == NetlinkMsgType.NETLINK_MSG_TYPE_LINK:
                    json_out = update_ifalias_information(json_out, self._logger)
                self._gnmi_q.putq(QueueElem(1, msg_type, full_state, json_out))
            else:
                self._logger.info("Libnl: Failed to decode json object {} for gNMI".format(json_out))
        if self.opta_enabled:
            if not self.__connectivity_state:
                self._logger.debug("Libnl: OPTA Connection failure")
                return
            if is_json:
                self._q.put((msg_type, full_state, json_out))
            else:
                self._logger.info("Failed to decode json object {}".format(json_out))
            if self._q.qsize() > 2000 or (full_state != TransMode.PARTIAL_TRANSFORM and self._q.qsize() > 40):
                self.process_grpc_msg()

            '''Update Message Stats''' #pylint: disable=pointless-string-statement
            try:
                self.msg_count_stats['total_msg_count'] = self.msg_count_stats.get('total_msg_count', 0) + 1
                self.msg_count_stats['total_msg_bytes'] = self.msg_count_stats.get('total_msg_bytes', 0) + sys.getsizeof(json_out)
                if not full_state == TransMode.PARTIAL_TRANSFORM:
                    self.msg_count_stats['full_state_count'] = \
                        self.msg_count_stats.get('full_state_count', 0) + 1
                else:
                    self.msg_count_stats['partial_state_count'] = \
                        self.msg_count_stats.get('partial_state_count', 0) + 1
                self.msg_count_stats['netlink_msg_count'] = self.msg_count_stats.get('netlink_msg_count', 0) + 1

                msg_type_dict = self.msg_count_stats["message_type_count"]
                msg_type_dict[msg_type] = msg_type_dict.get(msg_type, 0) + 1
                self.msg_count_stats["message_type_count"] = msg_type_dict
            except Exception as e:
                self._logger.error(" Caught Exception {}".format(e))


    def grpc_init_rate_control(self):
        try:
            self._wait_time = self._grpc_object._get_wait_time()
            self.__logger_coalescer.debug("Libnl: queue size {} ".format(self._q.qsize()),
                                          "elapsed time {} wait time {} min interval {}".format(self._grpc_object._elapsed_time,
                                                                                                self._wait_time,
                                                                                                self._grpc_object._min_interval))
            while self._wait_time > 0:
                q_len = self._q.qsize()
                if q_len > self._grpc_object._max_msg_threshold:
                    break
                else:
                    return False
                #self._grpc_object._wait_time = self._grpc_object._get_wait_time()

        except Exception as e:
            self._logger.info("Caught Wait time exception {}".format(e))
        return True

    def is_stream_available(self):
        if self.grpc_init_rate_control() and not self._q.empty():
            self._logger.debug("Libnl: Stream data is available, attempting rpc")
            return True

        return False

    def get_stream(self):
        msg_cnt = 0
        try:
            if not self.grpc_init_rate_control():
                self._grpc_object._last_transmission_time = time.time()
                return []
            while True:
                try:
                    if self._q.empty():
                        return []
                    (message_type, full_state, msg) = self._q.get()
                except Exception as e:
                    break
                rpc_msg = self.message_builder(message_type, full_state, msg)
                yield rpc_msg
                msg_cnt += 1
                if msg_cnt >= self._grpc_object._max_message_per_transaction:
                    self._logger.info("Libnl: Reached max messages per transaction")
                    break
            self._grpc_object._last_transmission_time = time.time()
            return []

        except Exception as e:
            self._logger.error("GRPC StreamMessage failed %s", e)
            self._grpc_object._last_transmission_time = time.time()
            return []


    def process_netlink_stats(self):
        self.__netlink_update_time += self._GRPC_UPDATE_TIME
        if self.__netlink_update_time >= self._NETLINK_UPDATE_TIME_SEC:
            self.push_netlink_stats()
            self.__netlink_update_time = 0.0

    def process_grpc_msg(self):
        if not self.opta_enabled:
            return AGENTSTATUS.SUCCESS
        metadata = [('aid', HOSTNAME)]
        if not self.__connectivity_state:
            self._logger.debug("Waiting for opta connection")
            return AGENTSTATUS.FAILURE
        while True:
            try:
                if not self.is_stream_available():
                    break
                with self.lock:
                    res, call = self._opta_stub.StreamMessage.with_call(self.get_stream(), metadata=metadata,
                                                                        timeout=self._grpc_object._grpc_call_timeout)
                    self._logger.debug("Libnl: [{}] RPC Completed {} status {}".format(threading.current_thread().name, res, call.code()))
                if call.code() != grpc.StatusCode.OK:
                    self._logger.error("Libnl: Agent Stream Failed {}".format(call.code()))
                    self.__connectivity_state = False
                    self._grpc_object._close()
                    return AGENTSTATUS.FAILURE
                break
            except grpc.RpcError as e:
                self._logger.error("Libnl: Agent Stream Failed Code {} Info: {}".format(e.code(), e.details()))
                self.__connectivity_state = False
                self._grpc_object._close()
                return AGENTSTATUS.FAILURE
            except Exception as e:
                self._logger.error("Libnl: Agent Stream failed error {}".format(e))
                self.__connectivity_state = False
                self._grpc_object._close()
                return AGENTSTATUS.FAILURE


    def run(self, config):
        if self.opta_enabled:
            self.__connect_reconnect_opta()

        cmd_srvr = CommandServer(config, self._logger)
        command_server = threading.Thread(target=cmd_srvr.start_server)
        command_server.start()

        netlink_thread = threading.Thread(target=netq_netlink.netlink_main, args=(self._logger.level, HOSTNAME))
        netlink_thread.start()

        while True:
            if (self.opta_enabled and (self._grpc_object.channel_state != ChannelState.CONNECTED or  \
                    not self.__connectivity_state)):
                self.__connectivity_state = False
                self._logger.info("OPTA not reachable")
                self.__connect_reconnect_opta()
            self.__logger_coalescer.debug("Libnl: Netlink active")
            if not command_server.is_alive():
                self._logger.info("Netlink Command server thread killed")
                break
            if not netlink_thread.is_alive():
                self._logger.info("Libnl: Netlink Parser thread killed")
                break
            time.sleep(3)


        if netlink_thread:
            netlink_thread.join()
        if command_server:
            command_server.join()
        return

    def activate(self):
        priority = 2
        msg_type = "NETQ_RATE_CONTROL"
        pq_data = json_loads("[]")
        if self._pq is None:
            #self._logger.error("Netq Not in ready state")*/
            return
        q_elem = QueueElem(priority, msg_type, False, pq_data)
        #self._logger.info("{}".format(libnl_msg))
        self._pq.putq(q_elem)


    def push_agent_info(self, msg_type, libnl_msg):
        priority = 1
        is_json = self._is_json_string(libnl_msg)
        if not is_json:
            return
        pq_data = json_loads(libnl_msg)
        if self._agent_pq is None:
            #self._logger.error("Netq Not in ready state")
            return
        q_elem = QueueElem(priority, msg_type, False, pq_data)
        self._agent_pq.putq(q_elem)

    def push_netlink_stats(self):
        priority = 1
        pq_data = json_dumps(self.msg_count_stats)
        #self._logger.info("Netlink Stats {}".format(pq_data))
        self.msg_count_stats = {}
        self.msg_count_stats["message_type_count"] = {}
        if self._agent_pq is None:
            #self._logger.error("Netq Not in ready state")
            return
        msg_type = "Stats"
        q_elem = QueueElem(priority, msg_type, False, pq_data)
        self._agent_pq.putq(q_elem)


    def log_netlink_msg(self, log_level, log_msg):
        if log_level == "info":
            self._logger.info(log_msg)
        elif log_level == "debug":
            self._logger.debug(log_msg)
        elif log_level == "warning":
            self._logger.warn(log_msg)
        else:
            self._logger.error(log_msg)


def proc_agent_info(msg_typ, libnl_msg):
    netlink_proc_obj.push_agent_info(msg_typ, libnl_msg)


def proc_libnl_msg(msg_type, libnl_msg, full_state):

    global netlink_proc_obj #pylint: disable=global-statement
    if netlink_proc_obj is None:
        print("Netlink Object not ready")
        return 0

    netlink_proc_obj.collect_data(msg_type, libnl_msg, copy.copy(full_state))
    return 0


def proc_grpc_msg():
    command_server_obj.process_agent_command()
    netlink_proc_obj.process_grpc_msg()
    netlink_proc_obj.process_netlink_stats()

def proc_logger_msg(log_level, log_msg):
    netlink_proc_obj.log_netlink_msg(log_level, log_msg)
