#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import threading
import socket
import time
import os
import signal
import psutil
from netq_lib.common import utils


class NetqAgentMonitor(object):
    """
        Code to Monitor netq-sub-agent process
    """
    __SERVER = '/run/netq_monitor_thread'

    __MAX_LENGTH = 4096

    def __init__(self, conf):
        self.__conf = conf
        self.__thread = None
        self.__active_time = time.time()
        self.__socket_timeout = 600
        self.__agent_name = "netq-sub-agent"
        self.__proc_name = 'netq-agent'
        self.__logger = utils.get_logger(self.__proc_name, conf)
        self.__logger.debug("Monitor thread Initialized")

    def init_client(self, conf):
        """

        :param conf: Configuration parameter
        """
        self.__socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.__socket.connect(self.__SERVER)

    def send_ping(self):
        try:
            self.__socket.send("ping".encode("utf-8"))
        except: #pylint: disable=bare-except
            self.__logger.info("heartbeat ping not sent... process termination in progress?")

    def __process_msg(self, clientsocket):
        try:
            clientsocket.settimeout(self.__socket_timeout)
            while True:
                buf = clientsocket.recv(self.__MAX_LENGTH).decode('utf-8')
                self.__logger.debug("Data {}".format(buf))
                self.__socket_timeout = time.time()
                if not buf:
                    self.__logger.info("Socket disconnection")
                    return -1 #pylint: disable=lost-exception
        except Exception as e:
            self.__logger.error("Agent Monitor: caught {} Exception ".format(e))
            self.__handle_socket_timeout()
            clientsocket.close()
            return -1
        finally:
            clientsocket.close()
            return -1 #pylint: disable=lost-exception

    def __handle_socket_timeout(self):

        self.__logger.debug("No ping received for {}".format(self.__socket_timeout))
        monitor_pid = os.getpid()
        parent = psutil.Process(monitor_pid)
        children = parent.children(recursive=True)
        if len(children) > 0: #pylint: disable=len-as-condition
            try:
                children = sorted(children, key=lambda proc: proc.pid, reverse=True)
            except Exception as e:
                self.__logger.info("sort except {}".format(e))
            self.__logger.debug("Cleaning processes {}".format(children))
            for process in children:
                self.__logger.info("Terminating agent pid {}".format(process.pid))
                process.send_signal(signal.SIGTERM)
        self.__logger.debug("Subprocess killed")
        return

    def _process_node(self):

        serversocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        serversocket.settimeout(self.__socket_timeout)

        try:
            os.unlink(self.__SERVER)
        except OSError:
            if os.path.exists(self.__SERVER):
                raise

        self.__logger.info("Binding to monitor port {}".format(self.__SERVER))
        serversocket.bind(self.__SERVER)
        serversocket.listen(10)

        while True:
            try:
                (clientsocket, address) = serversocket.accept()
            except socket.timeout:
                self.__logger.error("Socket timeout occurred")
                self.__handle_socket_timeout()
                return
            except Exception as e:
                raise e
            self.__process_msg(clientsocket)
            return

    def start(self):

        self.__thread = threading.Thread(target=self._process_node)
        self.__thread.start()
