#
# Copyright (c) 2020-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import time
from netq_agent.cmd.netq_grpc_agent import NetQAgentGrpcClient
from netq_lib.orm import HOSTNAME
from netq_lib.common.enums import ConfKw, AGENTSTATUS

# GRPC Specfic Imports
import grpc
from netq_agent.interface.netq_opta_discovery_pb2 import (
    AgentInfo,
    ResponseStatus
)
from netq_agent.interface.netq_opta_discovery_pb2_grpc import (
    OptaDiscoveryServicer,
    OptaDiscoveryStub
)


class NetQAgentOPTADiscovery(object):
    def __init__(self, conf, logger, pq, backpressure_queue):
        self._conf = conf
        self._logger = logger
        self._grpc_object = NetQAgentGrpcClient(conf, logger, pq, backpressure_queue)
        self._discovery_stub = None
        self.__channel_retry_count = 0
        self.opta_instance_ip = ''
        self.opta_instance_port = ConfKw.DEFAULT_OPTA_PORT

    def get_discovery_servers(self):
        discovery_servers = self._conf['netq-agent'].get(ConfKw.OPTA_DISCOVERY_SERVERS)
        if not discovery_servers:
            return []

        port = self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_DISCOVERY_SERVER_PORT)
        discovery_list = []
        self._logger.info('Configured ips: {}'.format(discovery_servers))
        for addr in discovery_servers.split(','):
            discovery_list.append('{}:{}'.format(addr, port))
        return discovery_list

    def message_builder(self):
        return AgentInfo(agentId=HOSTNAME, metaData={})

    def process_discovery_response(self, response):
        # update config object of Agent Monitor
        self.opta_instance_ip = response.instanceIP
        self.opta_instance_port = response.instancePort if response.instancePort else ConfKw.DEFAULT_OPTA_PORT

    def send_discovery_request(self):
        result_code = None
        discovery_servers = self.get_discovery_servers()
        discovery_iter = iter(discovery_servers)

        server_ip = next(discovery_iter)
        while True:
            try:
                # Set OPTA ip address information in NetQAgentGrpcClient
                self._grpc_object.set_opta_discovery_ip(server_ip)
                if AGENTSTATUS.SUCCESS != self._grpc_object.reestablish_channel():
                    self._logger.info("Failed to establish Connection with OPTA, retrying with backoff")
                    if self._grpc_object.get_channel_retry_count() >= self._grpc_object.max_opta_retry:
                        self._grpc_object.reset_channel_retry_count()
                        self._grpc_object._reset_grpc_backoff()
                        if self._grpc_object.is_channel_recreate_limit_reached():
                            server_ip = next(discovery_iter)
                            self._grpc_object._reset_grpc_backoff()
                            continue    # to next discovery server
                    else:
                        self._grpc_object.update_channel_retry_count()
                    timeVal = self._grpc_object.get_grpc_backoff_time()
                    time.sleep(timeVal)
                    continue    # to same discovery server

                self._logger.info("Starting Stream")
                self._grpc_object._reset_grpc_backoff()

                self._discovery_stub = OptaDiscoveryStub(self._grpc_object._channel)
                try:
                    res = self._discovery_stub.getOptaInstance(self.message_builder())
                    self._logger.info("OPTA Discovery Request RPC Completed {}, Status {}".format(res, res.status))
                    if res.status == ResponseStatus.Success:
                        self.process_discovery_response(res)
                        self._grpc_object._close()
                        result_code = AGENTSTATUS.SUCCESS
                        break   # Success
                    elif res.status == ResponseStatus.Failure:
                        self._logger.error("Discovery Request failed with error {}".format(res.errorMessage))
                        self._grpc_object._close()
                        result_code = AGENTSTATUS.FAILURE
                        server_ip = next(discovery_iter)
                        continue    # to next discovery server
                    elif res.status == ResponseStatus.Reconnect:
                        self._logger.info("Discovery Request failed, Reconnect requested by OPTA Discovery Server")
                        continue    # to same discovery server
                    else:
                        self._logger.error("Invalid Status received in Discovery Response {}".format(res.status))
                        self._grpc_object._close()
                        server_ip = next(discovery_iter)
                        continue    # to next discovery server
                except grpc.RpcError as e:
                    self._logger.error("RPC Error: OPTA Discovery Request failed Code {} Info: {}".format(e.code(), e.details()))
                    self._grpc_object._close()
                    result_code = AGENTSTATUS.FAILURE
                    server_ip = next(discovery_iter)
                    continue    # to next discovery server
                except Exception as e:
                    self._logger.error("OTPA Discovery Request failed with exception {}".format(e))
                    self._grpc_object._close()
                    result_code = AGENTSTATUS.FAILURE
                    server_ip = next(discovery_iter)
                    continue    # to next discovery server
            except StopIteration:
                self._logger.error("Unable to perform successful discovery procedure with any configured servers")
                result_code = AGENTSTATUS.FAILURE
                break    # Abort

        return result_code
