#
# Copyright (c) 2018-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#




import time
import json
import select
import random
import socket
from netq_lib.common.enums import ConfKw
from netq_agent.cmd.netq_agent_discovery import NetQAgentDiscovery, AGENTSTATUS, ChannelState
from netq_lib.common.netq_pq import QueueElem
from netq_lib.orm import HOSTNAME
from concurrent import futures

# GRPC Specfic Imports
import grpc
from netq_agent.interface.netq_pb2 import (
    NetqAgentMessage,
)
from netq_agent.interface.netq_pb2_grpc import (
    NetqAgentStreamServiceServicer,
    add_NetqAgentStreamServiceServicer_to_server
)

PARTIAL_TRANSFORM = 1
FULL_TRANSFORM = 0
FIRST_SEGMENT = 2
MID_SEGMENT = 3
LAST_SEGMENT = 4

class NetQAgentServicer(NetqAgentStreamServiceServicer):

    run_id = 0
    stop_rpc = False

    def __init__(self, logger, pq, bp_pq):
        self._logger = logger
        self.seq_no = 0
        self._pq = pq
        self._bp_pq = bp_pq
        self.stop_rpc = False
        self.running_rpc_cnt = 0

    def service_start_request(self):
        self._start = True

    def get_lldp_command_out(self):
        msg = ""
        return msg

    def get_sequence_number(self):
        self.seq_no = self.seq_no + 1
        return self.seq_no


    def full_state_message_builder(self, msg_type, segment, seq_no, json_out):

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=seq_no,
                                        ts=int(time.time() * 1000),
                                        trans_mode=segment,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def message_builder(self, msg_type, fullstate, json_out):

        full_state = FULL_TRANSFORM if fullstate else PARTIAL_TRANSFORM

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=self.get_sequence_number(),
                                        ts=int(time.time() * 1000),
                                        trans_mode=full_state,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def get_full_state_message(self, msg):

        num_msg_per_packet = 2
        if 'priority_qid' not in list(msg.keys()) or 'message_type' not in list(msg.keys()):
            self._logger.error('Invalid full state message')

        pq_idx = msg.get('priority_qid')
        queue_len = self._pq.len(pq_idx)
        num_segments = int(((queue_len - 1) / num_msg_per_packet) + 1)

        msg_data = []
        num_msgs = 0

        seq_no = self.get_sequence_number()
        segment_idx = 1

        for priority, message_type, fullstate, msg in self._pq.get_buffer(pq_idx):
            msg_data.append(msg.data)
            num_msgs += 1
            if (num_msgs == num_msg_per_packet):
                if (segment_idx == num_segments):
                    rpc_msg = self.full_state_message_builder(message_type, LAST_SEGMENT, seq_no, msg_data)
                elif(segment_idx == 1):
                    rpc_msg = self.full_state_message_builder(message_type, FIRST_SEGMENT, seq_no, msg_data)
                else:
                    rpc_msg = self.full_state_message_builder(message_type, MID_SEGMENT, seq_no, msg_data)
                msg_data = []
                num_msgs = 0
                segment_idx += 1
                yield rpc_msg

        if len(msg_data) > 0:
            rpc_msg = self.full_state_message_builder(message_type, LAST_SEGMENT, seq_no, msg_data)
            yield rpc_msg


    def StreamMessage(self, request, context):
        # pq = NetQPriorityQueue.PQ  # Wrap as utility function
        pq = self._pq
        fd_set = pq.get_reader_fds()
        my_run_id = random.randint(1, 1000)
        peername = 'unknown'
        try:
            peername = context.peer()
        except Exception:
            pass
        self._logger.info("Enable new GRPC %d peer %s pending Q %d", my_run_id, peername, pq.len(1))
        cnt = 0
        if self.running_rpc_cnt:
            self._logger.info("Detected existing GRPC %d", self.run_id)
            # terminate the other rpc
            # indicate to agent thread to go through agent pause resume
            bp_pq = self._bp_pq
            try:
                q_elem = QueueElem(0, 'bp', 'bp')
                bp_pq.putq(q_elem)
            except Exception as ex:
                self._logger.info('GRPC backpressure putq failed %s', ex)
                pass

            self.stop_rpc = True
            while cnt < 3:
                if self.stop_rpc:
                    time.sleep(10)
                else:
                    self.running_rpc_cnt = 0
                    break
                cnt = cnt + 1
        self.running_rpc_cnt = 1
        self.run_id = my_run_id
        log_once = False
        self.stop_rpc = False
        while True:
            if self.stop_rpc:
                self._logger.info("GRPC stream terminated %d", my_run_id)
                self.stop_rpc = False
                return
            try:
                (epoll_input, [], []) = select.select(fd_set, [], [], 10)
                if not len(epoll_input):
                    continue
                if not log_once:
                    self._logger.info("GRPC stream started %d", my_run_id)
                    log_once = True
                for queue_idx in range(0, self._pq.max_num_queue):
                    for priority, message_type, fullstate, msg in pq.get_buffer(queue_idx):
                        if (message_type == 'FULL_STATE_END'):
                            full_gen = self.get_full_state_message(msg.data)
                            for segment in full_gen:
                                yield segment
                        elif (isinstance(msg.data, list)):
                            rpc_msg = self.message_builder(message_type, fullstate, msg.data)
                            yield rpc_msg
                        else:
                            rpc_msg = self.message_builder(message_type, fullstate, [msg.data])
                            yield rpc_msg

            except Exception as e:
                self._logger.error("GRPC %d StreamMessage failed %s", my_run_id, e)
                self.run_id = 0
                self.running_rpc_cnt = 0
                return


class NetQAgentStream(object):

    def __init__(self, conf, logger, pq, bp_pq):
        self._conf = conf
        self._channel = None
        self._logger = logger
        self._pq = pq
        self.run_id = 0
        self._bp_pq = bp_pq
        self.discovery_agent = None

    def _get_server_addr(self):
        # always listen on *:port
        agent_url = "0.0.0.0"
        try:
            agent_port = self._conf['netq-agent'].get(ConfKw.AGENT_PORT, ConfKw.DEFAULT_AGENT_PORT)
        except Exception:
            agent_port = ConfKw.DEFAULT_AGENT_PORT
        server_addr = "{}:{}".format(agent_url, agent_port)
        return server_addr

    def _init_grpc_server(self):
        if getattr(self, 'server', None) is not None:
            self.server.stop(None)
        self.server = None
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        add_NetqAgentStreamServiceServicer_to_server(
            NetQAgentServicer(self._logger, self._pq, self._bp_pq), self.server)

        server_addr = self._get_server_addr()
        self._logger.info("Start Streaming server at {}".format(server_addr))

        self.server.add_insecure_port(server_addr)
        self.server.start()
        self._logger.info("Streaming Server Started")


    def _init_rpc_service(self):

        self._init_grpc_server()

        self.discovery_agent = NetQAgentDiscovery(
            self._conf, logger=self._logger)

        if self.discovery_agent is None:
            self._logger.error("Failed to start disovery process")
            self.server.stop(None)
            return AGENTSTATUS.FAILURE

        discovery_enabled = self._is_discovery_enabled()
        self.wait_time = 10
        self.discovery_retry_time = 60
        retry_count = 0
        prev_wait_time = time.time()
        period_cnt = 0
        while True:
            curr_time = time.time()
            if discovery_enabled:
                if self.get_discovery_status() == ChannelState.IDLE:
                    if not retry_count:
                        self._logger.info("OPTA Start Discovery IDLE")
                    elif retry_count < 3:
                        self._logger.info("OPTA Re-Start Discovery (retry %s) IDLE", retry_count)
                    self.discovery_agent.close()
                    ret = self.discovery_agent.start_discovery(verbose=True)
                    if ret == AGENTSTATUS.FAILURE:
                        # big hammer - start/stop the GRPC server
                        self._logger.info("OPTA Discovery failure - stop/start streaming server")
                        self._init_grpc_server()
                        self._logger.info("OPTA Restarted GRPC server")
                        self.set_discovery_status(ChannelState.IDLE)
                    retry_count += 1
                    # reset the periodic discovery logic
                    prev_wait_time = curr_time
                elif self.get_discovery_status() == ChannelState.CONNECTED:
                    if retry_count:
                        self._logger.info("OPTA Discovery OK READY")
                        retry_count = 0
                    if (curr_time - prev_wait_time) >= self.discovery_retry_time:
                        # calculate next retry time
                        prev_wait_time = curr_time
                        verbose = False
                        if period_cnt == 10:
                            self._logger.info("OPTA Periodic Discovery")
                            verbose = True
                            period_cnt = 0
                        period_cnt += 1
                        self.discovery_agent.close()
                        self.discovery_agent.start_discovery(verbose=verbose)
                        if ret == AGENTSTATUS.FAILURE:
                            # big hammer - start/stop the GRPC server
                            self._logger.info("OPTA Discovery failure - stop/start streaming server")
                            self._init_grpc_server()
                            self._logger.info("OPTA Restarted GRPC server")
                            self.set_discovery_status(ChannelState.IDLE)
                else:
                    # reset the periodic discovery logic
                    prev_wait_time = curr_time
                    if not retry_count:
                        self._logger.info("OPTA Discovery Channel not ready")
                        retry_count += 1
            time.sleep(self.wait_time)

    def start_stream_server(self):
        pass

    def start_stream(self):
        pass

    def stop_stream(self):
        pass

    def set_discovery_status(self, state):
        self.discovery_agent.channel_state = state

    def get_discovery_status(self):
        return self.discovery_agent.channel_state

    def _is_discovery_enabled(self):
        try:
            bool_obj = self._conf['netq-agent'].get(ConfKw.ENABLE_OPTA_DISCOVERY, ConfKw.DEFAULT_ENABLE_OPTA_DISCOVERY)
            if isinstance(bool_obj, bool):
                if bool_obj:
                    return True
                else:
                    return False
        except Exception:
            return True

        if bool_obj.lower() in ['yes', 'true', '1']:
            return True
        else:
            return False

    def stream_main(self):

        self._init_rpc_service()
        return
