#
# Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import signal
import json
import queue
import time
import datetime
import eventlet
import eventlet.pools
import threading
import inotify.adapters
from netq_lib.common.service import Service
from netq_lib.common.enums import Heartbeat
from netq_lib.common.utils import send_list_data
from netq_lib.orm.redisdb import unsubscribe_listener
from netq_lib.orm.redisdb.models import AsicMonitorHistogram, AsicMonitorBufferStats
from netq_agent.handlers.sdk_utils import SdkUtils, SX_CHIP_TYPE_SPECTRUM, SX_CHIP_TYPE_SPECTRUM2, SX_CHIP_TYPE_SPECTRUM4
from netq_lib.orm.redisdb.models import Lldp
from netq_lib.orm import HOSTNAME


class NetQAsicMonitorConfig(object):
    def __init__(self, logger, snapshot_min_file_count=120, snapshot_collection_periodicity=30):

        # configurations
        self.logger = logger
        self.config_monitor_file = '/etc/cumulus/datapath/monitor.conf'
        self.config_monitor_dir = '/etc/cumulus/datapath'

        # configuration keys
        self.config_key = "config"
        self.snapshot_key = "snapshot"
        self.snapshot_file_key = "snapshot.file"
        self.snapshot_file_count_key = "snapshot.file_count"
        self.action_list_key = "action_list"
        self.stat_type_key = "stat_type"
        self.port_group_list_key = "monitor.port_group_list"
        self.buffer_stats_key = 'buffer'
        self.histogram_key = 'histogram'
        self.histogram_min_bytes_key = "histogram.minimum_bytes_boundary"
        self.histogram_size_bytes_key = "histogram.histogram_size_bytes"
        self.histogram_sampling_time_ns_key = "histogram.sample_time_ns"

        # Snapshot collection periodicity
        self.snapshot_collection_periodicity = snapshot_collection_periodicity

        # Snapshot minimum supported file count
        self.snapshot_min_file_count = snapshot_min_file_count

        # Global configuration dictionary of each stat type
        self.port_group_dict = {}
        self.stat_type_dict = {self.histogram_key: {}, self.buffer_stats_key: {}}
        self.snapshot_to_port_group_dict = {}

        # Inotify watch list maintenance
        self.inotify_watch_dict = {}


    # GET methods
    def get_config_monitor_dir(self):
        return self.config_monitor_dir

    def get_config_monitor_file(self):
        return self.config_monitor_file

    def get_pg_from_snapshot_file(self, path, filename):
        filepath = os.path.join(path, filename).rsplit('_', 1)[0]
        if filepath in self.snapshot_to_port_group_dict:
            return self.snapshot_to_port_group_dict[filepath]
        return None


    # GET methods for histogram
    def get_histogram_snapshot_dir(self, pg):
        dirname = None
        if self.histogram_key in self.stat_type_dict and pg in self.stat_type_dict[self.histogram_key]:
            dirname = self.stat_type_dict[self.histogram_key][pg][self.snapshot_file_key]
            dirname = os.path.dirname(dirname)

        return dirname

    def get_histogram_snapshot_file(self, pg):
        filename = None
        if self.histogram_key in self.stat_type_dict and pg in self.stat_type_dict[self.histogram_key]:
            filename = self.stat_type_dict[self.histogram_key][pg][self.snapshot_file_key]
            filename = os.path.basename(filename)

        return filename

    def get_histogram_min_bytes_boundary(self, pg):
        value = 0
        if self.histogram_key in self.stat_type_dict and pg in self.stat_type_dict[self.histogram_key]:
            value = int(self.stat_type_dict[self.histogram_key][pg][self.histogram_min_bytes_key])
        return value

    def get_histogram_size_bytes(self, pg):
        value = 0
        if self.histogram_key in self.stat_type_dict and pg in self.stat_type_dict[self.histogram_key]:
            value = int(self.stat_type_dict[self.histogram_key][pg][self.histogram_size_bytes_key])
        return value

    def get_histogram_sample_time(self, pg):
        value = 0
        if self.histogram_key in self.stat_type_dict and pg in self.stat_type_dict[self.histogram_key]:
            value = int(self.stat_type_dict[self.histogram_key][pg][self.histogram_sampling_time_ns_key])
        return value

    def get_histogram_snapshot_file_count(self, pg):
        value = 0
        if self.histogram_key in self.stat_type_dict and pg in self.stat_type_dict[self.histogram_key]:
            value = int(self.stat_type_dict[self.histogram_key][pg][self.snapshot_file_count_key])
        return value

    def get_histogram_pg_list(self):
        if self.histogram_key in self.stat_type_dict:
            return self.stat_type_dict[self.histogram_key].keys()
        return list()


    # GET methods for buffer stats
    def get_buffer_stats_snapshot_dir(self, pg):
        dirname = None
        if self.buffer_stats_key in self.stat_type_dict and pg in self.stat_type_dict[self.buffer_stats_key]:
            dirname = self.stat_type_dict[self.buffer_stats_key][pg][self.snapshot_file_key]
            dirname = os.path.dirname(dirname)

        return dirname

    def get_buffer_stats_snapshot_file(self, pg):
        filename = None
        if self.buffer_stats_key in self.stat_type_dict and pg in self.stat_type_dict[self.buffer_stats_key]:
            filename = self.stat_type_dict[self.buffer_stats_key][pg][self.snapshot_file_key]
            filename = os.path.basename(filename)

        return filename

    def get_buffer_stats_snapshot_file_count(self, pg):
        value = 0
        if self.buffer_stats_key in self.stat_type_dict and pg in self.stat_type_dict[self.buffer_stats_key]:
            value = int(self.stat_type_dict[self.buffer_stats_key][pg][self.snapshot_file_count_key])
        return value

    def get_buffer_stats_pg_list(self):
        if self.buffer_stats_key in self.stat_type_dict:
            return self.stat_type_dict[self.buffer_stats_key].keys()
        return list()

    # Condition check methods
    def inotify_watch_has_users(self, watch_dir, user):
        if self.inotify_watch_dict != None:
            filter_function = lambda watchpath: os.path.dirname(watchpath[0]) == watch_dir and watchpath[1] != None
            return len(list(filter(filter_function, self.inotify_watch_dict.items()))) > 0
        return False

    def is_config_valid(self, conf_dict):
        if self.config_key in conf_dict:
            return True
        return False

    def is_stat_type_buffer_stats(self, conf_dict):
        if self.stat_type_key in conf_dict:
            stat_type = conf_dict.get(self.stat_type_key)
            if stat_type == self.buffer_stats_key:
                return True
        return False

    def is_stat_type_histogram(self, conf_dict):
        if self.stat_type_key in conf_dict:
            stat_type = conf_dict.get(self.stat_type_key)
            if stat_type == self.histogram_key:
                return True
        return False

    def is_stat_type_valid(self, conf_dict):
        if self.stat_type_key in conf_dict:
            stat_type = conf_dict.get(self.stat_type_key)
            if stat_type == self.histogram_key or stat_type == self.buffer_stats_key:
                return True
        return False

    def is_action_valid(self, conf_dict):
        if self.action_list_key in conf_dict:
            action = conf_dict.get(self.action_list_key)
            if action == self.snapshot_key:
                return True
        return False

    def is_snapshot_file_valid(self, conf_dict):
        if self.snapshot_file_key in conf_dict and len(conf_dict[self.snapshot_file_key]) != 0:
            snapshot_file = conf_dict[self.snapshot_file_key]
            snapshot_dir = os.path.dirname(snapshot_file)
            return os.path.exists(snapshot_dir)
        return False

    def is_snapshot_file_count_valid(self, conf_dict):
        if self.snapshot_file_count_key in conf_dict and conf_dict[self.snapshot_file_count_key] >= self.snapshot_min_file_count:
            return True
        return False

    # Condition check methods for histogram
    def is_pg_histogram_type(self, pg):
        if pg in self.port_group_dict and self.stat_type_key in self.port_group_dict[pg]:
            return self.port_group_dict[pg][self.stat_type_key] == self.histogram_key
        return False

    def is_histogram_snapshot_file_count_valid(self, conf_dict):
        if self.snapshot_file_count_key in conf_dict:
            val = conf_dict.get(self.snapshot_file_count_key)
            if int(val) >= self.snapshot_min_file_count:
                return True
        return False

    def is_histogram_min_bytes_boundary_valid(self, conf_dict):
        if self.histogram_min_bytes_key in conf_dict:
            val = conf_dict.get(self.histogram_min_bytes_key)
            if int(val) >= 0:
                return True
        return False

    def is_histogram_size_bytes_valid(self, conf_dict):
        if self.histogram_size_bytes_key in conf_dict:
            val = conf_dict.get(self.histogram_size_bytes_key)
            if int(val) > 0:
                return True
        return False

    def is_histogram_sample_time_valid(self, conf_dict):
        if self.histogram_sampling_time_ns_key in conf_dict:
            val = conf_dict.get(self.histogram_sampling_time_ns_key)
            if int(val) > 0:
                return True
        return False

    def is_histogram_config_enabled(self):
        if self.histogram_key in self.stat_type_dict:
            conf_dict = self.stat_type_dict[self.histogram_key]
            valid_list = []
            for pg in conf_dict:
                pg_conf_dict = conf_dict[pg]
                if self.is_config_valid(pg_conf_dict):
                    valid_list.append(pg)
            return len(valid_list) > 0
        return False

    # Condition check methods for buffer stats
    def is_pg_buffer_stats_type(self, pg):
        if pg in self.port_group_dict and self.stat_type_key in self.port_group_dict[pg]:
            return self.port_group_dict[pg][self.stat_type_key] == self.buffer_stats_key
        return False

    def is_buffer_stats_snapshot_file_count_valid(self, conf_dict):
        if self.snapshot_file_count_key in conf_dict:
            val = conf_dict.get(self.snapshot_file_count_key)
            if int(val) >= self.snapshot_min_file_count:
                return True
        return False

    def is_buffer_stats_config_enabled(self):
        if self.buffer_stats_key in self.stat_type_dict:
            conf_dict = self.stat_type_dict[self.buffer_stats_key]
            valid_list = []
            for pg in conf_dict:
                pg_conf_dict = conf_dict[pg]
                if self.is_config_valid(pg_conf_dict):
                    valid_list.append(pg)
            return len(valid_list) > 0
        return False


    # Condition check methods for configuration
    def has_config_changed(self, conf_dict):
        if self.port_group_dict != conf_dict:
            self.logger.info("Configuration changed")
            return True

        self.logger.info("Configuration has not changed")
        return False


    # Validation check methods
    def validate_common_config(self, conf_dict):
        action_valid = self.is_action_valid(conf_dict)
        snapshot_file_valid = self.is_snapshot_file_valid(conf_dict)
        stat_type_valid = self.is_stat_type_valid(conf_dict)
        config_valid = self.is_config_valid(conf_dict)

        valid = action_valid and snapshot_file_valid and stat_type_valid and config_valid
        return valid

    # Validation check methods for hisrogram
    def validate_histogram_config(self, conf_dict):
        valid = False
        common_config_valid = self.validate_common_config(conf_dict)
        min_bytes_boundary_valid = self.is_histogram_min_bytes_boundary_valid(conf_dict)
        size_bytes_valid = self.is_histogram_size_bytes_valid(conf_dict)
        sample_time_valid = self.is_histogram_sample_time_valid(conf_dict)
        file_count_valid = self.is_histogram_snapshot_file_count_valid(conf_dict)

        if common_config_valid and min_bytes_boundary_valid and size_bytes_valid and sample_time_valid and file_count_valid:
            min_bytes_val = int(conf_dict.get(self.histogram_min_bytes_key))
            size_bytes_val = int(conf_dict.get(self.histogram_size_bytes_key))
            if size_bytes_val < min_bytes_val:
                valid = False
            else:
                valid = True

        return valid

    # Validation check methods from buffer stats
    def validate_buffer_stats_config(self, conf_dict):
        common_config_valid = self.validate_common_config(conf_dict)
        file_count_valid = self.is_buffer_stats_snapshot_file_count_valid(conf_dict)

        if file_count_valid:
            return True

        return False

    # Global validation check methods
    def validate_config(self, conf_dict):
        valid = False
        invalid_config_list = []

        if len(conf_dict) == 0:
            return valid, conf_dict

        for pg in conf_dict:
            if self.config_key not in conf_dict[pg] or conf_dict[pg][self.config_key] == False:
                invalid_config_list.append(pg)
                continue

            conf_data_dict = conf_dict.get(pg)

            stat_type_valid = self.is_stat_type_valid(conf_data_dict)

            if stat_type_valid:
                if self.is_stat_type_histogram(conf_data_dict) and not self.validate_histogram_config(conf_data_dict):
                    self.logger.debug("Histogram config {} is invalid".format(conf_data_dict))
                    invalid_config_list.append(pg)

                if self.is_stat_type_buffer_stats(conf_data_dict) and not self.validate_buffer_stats_config(conf_data_dict):
                    self.logger.debug("Buffer stats config {} is invalid".format(conf_data_dict))
                    invalid_config_list.append(pg)

            else:
                self.logger.debug("Unsupported stat type {}".format(conf_data_dict))
                invalid_config_list.append(pg)

        self.logger.info("Valid count = {}, {}".format(len(conf_dict.keys()), len(invalid_config_list)))

        valid = len(invalid_config_list) < len(conf_dict.keys())

        # Remove all invalid configurations
        if valid:
            for pg in invalid_config_list:
                del(conf_dict[pg])

        return valid, conf_dict

    # Clear methods
    def clear_configs(self):
        del(self.port_group_dict)
        del(self.stat_type_dict)
        del(self.inotify_watch_dict)
        del(self.snapshot_to_port_group_dict)

        self.inotify_watch_dict = {}
        self.port_group_dict = {}
        self.stat_type_dict = {self.histogram_key: {}, self.buffer_stats_key: {}}
        self.snapshot_to_port_group_dict = {}


    # Update methods
    def update_port_group_list(self, line, conf_dict={}):
        pg_list = line.split('=')[1].strip(' []\n')

        if (len(pg_list) == 0):
            return conf_dict

        pg_list = list(map(str.strip, pg_list.split(',')))

        for pg in pg_list:
            if pg not in conf_dict:
                conf_dict[pg] = {self.config_key: True}
            else:
                conf_dict[pg] = {self.config_key: True}

        return conf_dict

    def update_port_group_stat_type(self, line, conf_dict={}):
        pg = line.split('=')[0].split('.')[1].strip()
        stat_type = line.split('=')[1].strip()

        if pg not in conf_dict:
            conf_dict[pg] = {self.stat_type_key: stat_type}
        else:
            conf_dict[pg][self.stat_type_key] = stat_type

        return conf_dict

    def update_port_group_action(self, line, conf_dict={}):
        pg = line.split('=')[0].split('.')[1].strip()
        action_list = line.split('=')[1].strip(' []\n')

        if self.snapshot_key in action_list:
            if pg not in conf_dict:
                conf_dict[pg] = {self.action_list_key: self.snapshot_key}
            else:
                conf_dict[pg][self.action_list_key] = self.snapshot_key

        return conf_dict

    def update_port_group_snapshot_file(self, line, conf_dict={}):
        pg = line.split('=')[0].split('.')[1].strip()
        snapshot_file = line.split('=')[1].strip()

        if len(snapshot_file) == 0:
            return conf_dict

        if pg not in conf_dict:
            conf_dict[pg] = {self.snapshot_file_key: snapshot_file}
        else:
            conf_dict[pg][self.snapshot_file_key] = snapshot_file

        return conf_dict

    def update_port_group_snapshot_file_count(self, line, conf_dict={}):
        pg = line.split('=')[0].split('.')[1].strip()
        snapshot_file_count = int(line.split('=')[1].strip())

        if pg not in conf_dict:
            conf_dict[pg] = {self.snapshot_file_count_key: snapshot_file_count}
        else:
            conf_dict[pg][self.snapshot_file_count_key] = snapshot_file_count

        return conf_dict

    # Update histogram methods
    def update_histogram_data_min_bytes(self, line, conf_dict):
        pg = line.split('=')[0].split('.')[1].strip()
        value = int(line.split('=')[1].strip())

        if pg not in conf_dict:
            conf_dict[pg] = {self.histogram_min_bytes_key: value}
        else:
            conf_dict[pg][self.histogram_min_bytes_key] = value

        return conf_dict

    def update_histogram_data_sample_time(self, line, conf_dict):
        pg = line.split('=')[0].split('.')[1].strip()
        value = int(line.split('=')[1].strip())

        if pg not in conf_dict:
            conf_dict[pg] = {self.histogram_sampling_time_ns_key: value}
        else:
            conf_dict[pg][self.histogram_sampling_time_ns_key] = value

        return conf_dict

    def update_histogram_data_size_bytes(self, line, conf_dict):
        pg = line.split('=')[0].split('.')[1].strip()
        value = int(line.split('=')[1].strip())

        if pg not in conf_dict:
            conf_dict[pg] = {self.histogram_size_bytes_key: value}
        else:
            conf_dict[pg][self.histogram_size_bytes_key] = value

        return conf_dict

    # Update config methods
    def update_config(self, conf_dict):
        self.logger.debug("Config: {}".format(conf_dict))
        if self.has_config_changed(conf_dict):
            self.port_group_dict.clear()
            self.port_group_dict.update(conf_dict)
            self.inotify_watch_dict = {}
            self.snapshot_to_port_group_dict = {}
            for pg in conf_dict:
                conf_data_dict = conf_dict.get(pg)
                self.stat_type_dict[conf_data_dict.get(self.stat_type_key)][pg] = conf_data_dict
                self.snapshot_to_port_group_dict[self.snapshot_file_key] = pg
                snapshot_file = conf_data_dict.get(self.snapshot_file_key)
                self.inotify_watch_dict[snapshot_file] = None
                if snapshot_file in self.snapshot_to_port_group_dict:
                    self.logger.warn("Duplicate snapshot file {} - already mapped to pg {}".format(snapshot_file, pg))
                self.snapshot_to_port_group_dict[snapshot_file] = pg
            return True

        return False

    # Inotify watch methods
    def add_to_inotify_watch_dir(self, watch_file, user):
        if watch_file == None:
            self.logger.warn("Not a valid watch path {}".format(watch_file))
            return

        if watch_file in self.inotify_watch_dict:
            self.inotify_watch_dict[watch_file] = user
            self.logger.info("Watchers list {}".format(self.inotify_watch_dict))
        else:
            self.logger.warn("Watch dir {} not present in watch list {}".format(watch_file, self.inotify_watch_dict))

    def remove_from_inotify_watch_dir(self, watch_file, user):
        if watch_file == None:
            self.logger.warn("Not a valid watch path {}".format(watch_file))
            return

        if watch_file in self.inotify_watch_dict:
            self.inotify_watch_dict[watch_file] = None
            self.logger.info("Watchers list {}".format(self.inotify_watch_dict))
        else:
            self.logger.warn("Watch user {} or path {} not present in watch list {}".format(user, watch_file, self.inotify_watch_dict))

    # Configuration methods
    def read_config(self, conf_dict={}, conf_file='/etc/cumulus/datapath/monitor.conf'):
        if not os.path.exists(conf_file):
            self.logger.info("Configuration file {} is not present. Skipping read".format(conf_file))
            return conf_dict

        with open(conf_file, "r") as conf:
            for line in conf:
                if not line.lstrip().startswith('#'):

                    if self.histogram_min_bytes_key in line:
                        self.update_histogram_data_min_bytes(line, conf_dict)

                    if self.histogram_size_bytes_key in line:
                        self.update_histogram_data_size_bytes(line, conf_dict)

                    if self.histogram_sampling_time_ns_key in line:
                        self.update_histogram_data_sample_time(line, conf_dict)

                    if self.port_group_list_key in line:
                        self.update_port_group_list(line, conf_dict)

                    if self.stat_type_key in line:
                        self.update_port_group_stat_type(line, conf_dict)

                    if self.action_list_key in line:
                        self.update_port_group_action(line, conf_dict)

                    if self.snapshot_file_count_key in line:
                        self.update_port_group_snapshot_file_count(line, conf_dict)

                    if self.snapshot_file_key in line and not self.snapshot_file_count_key in line:
                        self.update_port_group_snapshot_file(line, conf_dict)

        conf.close()

        self.logger.info("Configuration: {}".format(conf_dict))
        return conf_dict

    # Main config handler method
    def handle_config(self, conf_file='/etc/cumulus/datapath/monitor.conf'):
        new_config_dict = self.read_config({}, conf_file=conf_file)
        valid, new_config_dict = self.validate_config(new_config_dict)
        if valid:
            self.logger.info("Configuration is valid")
            return (True, self.update_config(new_config_dict))
        else:
            self.logger.info("Configuration is not valid")
            return (False, False)

class NetQAsicMonitorBufferStatsSnapshotReader(object):
    def __init__(self, logger, config):
        self.logger = logger
        self.config = config

    def read_buffer_stats(self, pg, filename=None):
        data_dict = {}
        if filename == None:
            filename = '/var/lib/cumulus/buffer_stats_0'

        if os.path.exists(filename):
            with open(filename, "r") as bs_file:
                try:
                    bs_json = json.load(bs_file)
                except:
                    bs_file.close()
                    self.logger.info("[BUFFER-STATS]: Buffer stats data is not populated yet for file {}".format(filename))
                    return data_dict

                data_dict['start_timestamp'] = bs_json['timestamp_info']['start_datetime']
                data_dict['end_timestamp'] = bs_json['timestamp_info']['end_datetime']
                data_dict['buffer_info'] = bs_json['buffer_info']

            bs_file.close()

        return data_dict

class NetQAsicMonitorBufferStats(object):
    def __init__(self, logger, config, inotify_monitor_handle, processor_periodicity=30):
        self.logger = logger
        self.config = config
        self.buffer_stats_enabled = False
        self.buffer_stats_snapshot_monitor_file = {}
        self.buffer_stats_snapshot_monitor_dir = {}
        self.buffer_stats_snapshot_monitor_count = {}
        self.buffer_stats_snapshot_monitor_processor_periodicity = processor_periodicity
        self.buffer_stats_snapshot_reader = None
        self.buffer_stats_snapshot_reference = {}
        self.buffer_stats_snapshot_last_read_reference = {}
        self.inotify_snapshot_handle = inotify_monitor_handle
        self.inotify_snapshot_mask = inotify.constants.IN_CLOSE_WRITE

    # GET methods
    def get_buffer_stats_snapshot_monitor_filepath(self, pg):
        return os.path.join(self.buffer_stats_snapshot_monitor_dir[pg], self.buffer_stats_snapshot_monitor_file[pg])

    def get_buffer_stats_snapshot_monitor_file(self, pg):
        return self.buffer_stats_snapshot_monitor_file[pg]

    def get_buffer_stats_snapshot_monitor_dir(self, pg):
        return self.buffer_stats_snapshot_monitor_dir[pg]

    def get_buffer_stats_snapshot_monitor_processor_periodicity(self):
        return self.buffer_stats_snapshot_monitor_processor_periodicity


    # Condition check methods
    def is_buffer_stats_snapshot_reference_captured(self, pg):
        assert len(self.buffer_stats_snapshot_reference[pg]) <= 2
        return len(self.buffer_stats_snapshot_reference[pg]) != 0

    def is_buffer_stats_enabled(self):
        return self.buffer_stats_enabled


    # Record reference method - records the first occurrence of the snapshot
    def buffer_stats_snapshot_record_reference(self, pg, path, filename):
        snapshot_file = filename
        snapshot_path = os.path.join(path, snapshot_file)
        if os.path.exists(snapshot_path):
            snapshot_ctime = os.path.getctime(snapshot_path)
            snapshot_ctime_str = datetime.datetime.fromtimestamp(snapshot_ctime)
            #self.logger.info("Recording reference snapshot file: {}, ctime: {}".format(snapshot_path, snapshot_ctime_str))
            self.logger.info("[BUFFER-STATS]: Recording reference snapshot file {}, ctime: {} for pg: {}".format(snapshot_path, snapshot_ctime_str, pg))
            self.buffer_stats_snapshot_reference[pg] = (snapshot_path, snapshot_ctime)


    # Buffer watcher methods
    def buffer_stats_snapshot_watch_init(self, pg):
        watch_dir = self.get_buffer_stats_snapshot_monitor_dir(pg)
        watch_file = self.get_buffer_stats_snapshot_monitor_filepath(pg)
        if not self.config.inotify_watch_has_users(watch_dir, self.config.buffer_stats_key):
            inotify_handle = self.inotify_snapshot_handle
            self.logger.info("[BUFFER-STATS]: Watching buffer stats snapshot dir: {} for pg: {}".format(watch_dir, pg))
            inotify_handle.add_watch(watch_dir, self.inotify_snapshot_mask)
        self.config.add_to_inotify_watch_dir(watch_file, self.config.buffer_stats_key)

    def buffer_stats_snapshot_watch_deinit(self, pg):
        watch_dir = self.get_buffer_stats_snapshot_monitor_dir(pg)
        watch_file = self.get_buffer_stats_snapshot_monitor_filepath(pg)
        self.config.remove_from_inotify_watch_dir(watch_file, self.config.buffer_stats_key)
        if not self.config.inotify_watch_has_users(watch_dir, self.config.buffer_stats_key):
            inotify_handle = self.inotify_snapshot_handle
            self.logger.info("[BUFFER-STATS]: Unwatching buffer stats snapshot dir: {} for pg: {}".format(watch_dir, pg))
            inotify_handle.remove_watch(watch_dir)

    # Buffer stats snapshot methods
    def buffer_stats_snapshot_init(self):
        self.buffer_stats_snapshot_reader = NetQAsicMonitorBufferStatsSnapshotReader(self.logger, self.config)
        for pg in self.config.get_buffer_stats_pg_list():
            self.buffer_stats_snapshot_monitor_file[pg] = self.config.get_buffer_stats_snapshot_file(pg)
            self.buffer_stats_snapshot_monitor_dir[pg] = self.config.get_buffer_stats_snapshot_dir(pg)
            self.buffer_stats_snapshot_monitor_count[pg] = self.config.get_buffer_stats_snapshot_file_count(pg)
            self.buffer_stats_snapshot_reference[pg] = ()
            self.buffer_stats_snapshot_last_read_reference[pg] = ()
            self.buffer_stats_snapshot_watch_init(pg)

    def buffer_stats_snapshot_deinit(self):
        self.buffer_stats_snapshot_reader = None
        for pg in self.config.get_buffer_stats_pg_list():
            self.buffer_stats_snapshot_watch_deinit(pg)

        self.buffer_stats_snapshot_monitor_dir = {}
        self.buffer_stats_snapshot_monitor_file = {}
        self.buffer_stats_snapshot_monitor_count = {}
        self.buffer_stats_snapshot_reference = {}
        self.buffer_stats_snapshot_last_read_reference = {}

    def buffer_stats_snapshot_reinit(self):
        self.buffer_stats_snapshot_deinit()
        self.buffer_stats_snapshot_init()

    def buffer_stats_snapshot_file_handler(self, path, filename):
        pg = self.config.get_pg_from_snapshot_file(path, filename)
        if len(self.buffer_stats_snapshot_reference[pg]) == 0:
            self.buffer_stats_snapshot_record_reference(pg, path, filename)
            self.buffer_stats_snapshot_watch_deinit(pg)

    def buffer_stats_snapshot_processor(self):
        for pg in self.config.get_buffer_stats_pg_list():
            if pg == None:
                break

            if len(self.buffer_stats_snapshot_reference[pg]) == 0:
                continue

            self.logger.debug("[BUFFER-STATS]: Processing buffer stats snapshots for pg: {}".format(pg))
            snapshot_count = self.buffer_stats_snapshot_monitor_count[pg]

            if len(self.buffer_stats_snapshot_last_read_reference[pg]) == 0:
                (snapshot_path, snapshot_ctime) = self.buffer_stats_snapshot_reference[pg]
                snapshot_file_path, snapshot_file_index = snapshot_path.rsplit('_', 1)
                snapshot_file_index = int(snapshot_file_index)
            else:
                (snapshot_path, snapshot_ctime) = self.buffer_stats_snapshot_last_read_reference[pg]
                snapshot_file_path, snapshot_file_index = snapshot_path.rsplit('_', 1)
                snapshot_file_index = (int(snapshot_file_index) + 1) % snapshot_count

            count = 0
            chunks = 10
            buffer_stats_list = []
            while (count <= snapshot_count):
                filepath = snapshot_file_path + '_' + str(snapshot_file_index)
                if (not os.path.exists(filepath)):
                    break

                ctime = os.path.getctime(filepath)
                if ctime < snapshot_ctime:
                    break

                buffer_stats = self.buffer_stats_snapshot_reader.read_buffer_stats(pg, filepath)
                if not buffer_stats:
                    # If the data is null that means, we have not populated the data yet. Its in mid of update.
                    # Better to read it in next iteration
                    break

                buffer_stats_list.append(buffer_stats)
                #self.logger.info("[BUFFER-STATS]: Read buffer stats data({}) - file: {}".format(len(data_dict), filepath))

                # Update snapshot reference
                self.buffer_stats_snapshot_reference[pg] = (filepath, ctime)
                self.buffer_stats_snapshot_last_read_reference[pg] = (filepath, ctime)
                snapshot_file_index = (snapshot_file_index + 1) % snapshot_count
                count += 1

                if (len(buffer_stats_list) != 0 and (count % chunks) == 0):
                    self.logger.debug("[BUFFER-STATS]: Sending data chunk {}{}".format(count / 10, len(buffer_stats_list)))
                    #json_obj = json.dumps(buffer_stats_list).encode('utf-8')
                    json_obj = buffer_stats_list
                    msg = AsicMonitorBufferStats()
                    msg.buffer_stats = json_obj
                    buffer_stats_list = []
                    msg.save()

            if len(buffer_stats_list) != 0:
                self.logger.debug("[BUFFER-STATS]: Sending last data chunk {}{}".format(count / 10, len(buffer_stats_list)))
                #json_obj = json.dumps(buffer_stats_list).encode('utf-8')
                json_obj = buffer_stats_list
                msg = AsicMonitorBufferStats()
                msg.buffer_stats = json_obj
                msg.save()

            if count > 0:
                self.logger.info("[BUFFER-STATS]: Sent {} data to grpc for pg {}".format(count, pg))

    def buffer_stats_snapshot_monitor_filter(self, header, event_type, path, filename):
        filepath = os.path.join(path, filename)
        pg = self.config.get_pg_from_snapshot_file(path, filename)
        if not self.is_buffer_stats_enabled() or not self.config.is_pg_buffer_stats_type(pg):
            return False

        inotify_mask = self.inotify_snapshot_mask
        if header.mask & inotify_mask:
            filepath = os.path.join(path, filename)

            if self.buffer_stats_snapshot_monitor_file[pg] in filepath:
                self.logger.info("[BUFFER-STATS]: Buffer stats snapshot was created/modified for pg: {}".format(pg))
                return True

        return False

    # Buffer stats methods
    def buffer_stats_init(self):
        self.buffer_stats_snapshot_init()

    def buffer_stats_deinit(self):
        self.buffer_stats_snapshot_deinit()

    def buffer_stats_reinit(self):
        self.buffer_stats_snapshot_reinit()

    # Buffer stats configuration handler method
    def buffer_stats_handle_config(self, force_disable=False):
        #if (force_disable):
        #    self.logger.info("Forcing buffer stats deinitialisation")
        #    self.buffer_stats_enabled = False
        #    self.buffer_stats_deinit()
        #    return

        buffer_stats_enabled = self.config.is_buffer_stats_config_enabled()
        if self.buffer_stats_enabled != buffer_stats_enabled:
            if (buffer_stats_enabled):
                self.logger.info("[BUFFER-STATS]: Buffer stats config is enabled")
                self.buffer_stats_init()
            else:
                self.logger.info("[BUFFER-STATS]: Buffer stats config is disabled")
                self.buffer_stats_deinit()
        elif self.buffer_stats_enabled:
            self.logger.info("[BUFFER-STATS]: Buffer stats config changed")
            self.buffer_stats_reinit()
        else:
            self.logger.info("[BUFFER-STATS]: Buffer stats in disabled state")

        self.buffer_stats_enabled = buffer_stats_enabled


class NetQAsicMonitorHistogramSnapshotReader(object):
    def __init__(self, logger, config, simulation=False):
        self.logger = logger
        self.config = config
        self.histogram_min_size_bytes = {}
        self.histogram_size_bytes = {}
        self.histogram_sample_time_ns = {}
        self.histogram_range = {}
        self.histogram_max_size_bytes = {}
        self.histogram_bins = {}
        self.histogram_range_dict = {}
        self.histogram_range_to_bin_dict = {}
        self.simulation = simulation
        if not simulation:
            self.sdk_utils = SdkUtils(self.logger)
            self.sdk_utils.copy_sdk()
        self.cell_byte_size = self.get_platform_cell_size()

        for pg in self.config.get_histogram_pg_list():
            min_size_bytes_adjusted = int(config.get_histogram_min_bytes_boundary(pg) / self.cell_byte_size) * self.cell_byte_size
            histogram_byte_size = config.get_histogram_size_bytes(pg)
            histogram_cells = float(histogram_byte_size / self.cell_byte_size)
            histogram_byte_size_adjusted = int(histogram_cells * self.cell_byte_size)
            bin_range = int(float(histogram_cells / 8)) * self.cell_byte_size

            self.histogram_min_size_bytes[pg] = min_size_bytes_adjusted
            self.histogram_size_bytes[pg] = histogram_byte_size_adjusted
            self.histogram_sample_time_ns[pg] = config.get_histogram_sample_time(pg)
            self.histogram_range[pg] = bin_range
            self.histogram_bins[pg] = 10
            self.histogram_max_size_bytes[pg] = min_size_bytes_adjusted + histogram_byte_size_adjusted
            self.histogram_range_dict[pg] = {'0:{}'.format(int(self.histogram_min_size_bytes[pg]) - 1): 0}
            self.histogram_range_to_bin_dict[pg] = {'0:{}'.format(int(self.histogram_min_size_bytes[pg] - 1)): 'bin0'}
            bin_value = self.histogram_min_size_bytes[pg]
            for bin in range(1, self.histogram_bins[pg] - 1):
                self.histogram_range_dict[pg]["{}:{}".format(bin_value, bin_value + self.histogram_range[pg] - 1)] = 0
                self.histogram_range_to_bin_dict[pg]["{}:{}".format(bin_value, bin_value + self.histogram_range[pg] - 1)] = 'bin{}'.format(bin)
                bin_value += self.histogram_range[pg]

            self.histogram_range_dict[pg]["{}:{}".format(bin_value, "*")] = 0
            self.histogram_range_to_bin_dict[pg]["{}:{}".format(bin_value, "*")] = 'bin9'

    # GET methods
    def get_platform_cell_size(self):
        if not self.simulation:
            cell_byte_size_map = {SX_CHIP_TYPE_SPECTRUM: 96, SX_CHIP_TYPE_SPECTRUM2: 144, SX_CHIP_TYPE_SPECTRUM4: 192}
            chip_type = self.sdk_utils.get_chip_type()
            cell_size = cell_byte_size_map[SX_CHIP_TYPE_SPECTRUM2]
            if chip_type in cell_byte_size_map.keys():
                cell_size = cell_byte_size_map[chip_type]
        else:
            chip_type = "SIM"
            cell_size = 144

        self.logger.info("Platform[{}] queue cell size in bytes {}".format(chip_type, cell_size))
        return cell_size

    # Helper methods to read timestamp
    def read_record_timestamp(self, str_time):
        result = None
        try:
            result = datetime.datetime.strptime(str_time, '%Y-%m-%d %H:%M:%S.%f')
        except:
            try:
                result = datetime.datetime.strptime(str_time, '%Y-%m-%d %H:%M:%S')
            except:
                self.logger.error("[HISTOGRAM]: Error parsing timestamp {}".format(str_time))

        return result

    def read_histogram_timestamps(self, hist_json):
        start_timestamp = self.read_record_timestamp(hist_json['timestamp_info']['start_datetime'])
        end_timestamp = self.read_record_timestamp(hist_json['timestamp_info']['end_datetime'])
        if (start_timestamp == None and end_timestamp == None):
            return None
        else:
            if start_timestamp == None:
                start_timestamp = end_timestamp
            elif end_timestamp == None:
                end_timestamp = start_timestamp

        return (start_timestamp, end_timestamp)

    # Main histogram file reader method
    def read_histogram(self, pg, lldp_info, filename=None):
        data_list = []
        if filename == None:
            filename = '/var/lib/cumulus/histogram_stats_0'

        if os.path.exists(filename):
            with open(filename, "r") as hfile:
                try:
                    hist_json = json.load(hfile)
                except:
                    hfile.close()
                    self.logger.info("[HISTOGRAM]: Histogram data is not populated yet for file {}".format(filename))
                    return data_list

                timestamps = self.read_histogram_timestamps(hist_json)
                if timestamps == None:
                    self.logger.error("[HISTOGRAM]: Error parsing record {} - skipping...".format(hist_json))
                    return data_list

                start_timestamp, end_timestamp = timestamps
                for ifname in hist_json['histogram_info']:
                    """
                    for bin_range in self.histogram_range_dict:
                        if not data_dict['histogram_info'][ifname].has_key(bin_range):
                            data_dict['histogram_info'][ifname][bin_range] = self.histogram_range_dict[pg][bin_range]
                    """
                    try:
                        bin_range_data = dict(self.histogram_range_dict[pg], **hist_json['histogram_info'][ifname])
                        bin_data = dict((self.histogram_range_to_bin_dict[pg][b], v) for(b, v) in bin_range_data.items())
                    except KeyError:
                        self.logger.warn("[HISTOGRAM]: Histogram range has changed. BinRange: {}, Histogram Record: {}. Skipping incompatible records...".format(self.histogram_range_to_bin_dict[pg], hist_json['histogram_info'][ifname]))
                        return data_list

                    peer_hostname = "N/A"
                    if not lldp_info or ifname not in lldp_info.keys():
                        self.logger.debug("[HISTOGRAM]: LLDP peer data not available for interface {}".format(ifname))
                    else:
                        peer_hostname = lldp_info[ifname].peer_hostname


                    msg = AsicMonitorHistogram()
                    msg.timestamp = float(start_timestamp.timestamp())
                    msg.start_timestamp = int(start_timestamp.timestamp())
                    msg.end_timestamp = int(end_timestamp.timestamp())
                    msg.ifname = ifname
                    msg.bin0 = bin_data.get('bin0')
                    msg.bin1 = bin_data.get('bin1')
                    msg.bin2 = bin_data.get('bin2')
                    msg.bin3 = bin_data.get('bin3')
                    msg.bin4 = bin_data.get('bin4')
                    msg.bin5 = bin_data.get('bin5')
                    msg.bin6 = bin_data.get('bin6')
                    msg.bin7 = bin_data.get('bin7')
                    msg.bin8 = bin_data.get('bin8')
                    msg.bin9 = bin_data.get('bin9')
                    msg.maximum_queue_bytes = self.histogram_size_bytes[pg]
                    msg.minimum_queue_bytes = self.histogram_min_size_bytes[pg]
                    msg.queue_bytes_range = self.histogram_range[pg]
                    msg.sampling_time = self.histogram_sample_time_ns[pg]
                    msg.peer_hostname = peer_hostname
                    msg.save()

                    # save
                    data_list.append(msg)

            hfile.close()

        return data_list


class NetQAsicMonitorHistogram(object):
    def __init__(self, logger, config, inotify_monitor_handle, processor_periodicity=30, simulation=False):
        self.logger = logger
        self.config = config
        self.histogram_enabled = False
        self.histogram_snapshot_monitor_file = {}
        self.histogram_snapshot_monitor_dir = {}
        self.histogram_snapshot_monitor_count = {}
        self.histogram_snapshot_monitor_processor_periodicity = processor_periodicity
        self.histogram_snapshot_reader = None
        self.histogram_snapshot_reference = {}
        self.histogram_snapshot_last_read_reference = {}
        self.inotify_snapshot_handle = inotify_monitor_handle
        self.inotify_snapshot_mask = inotify.constants.IN_CLOSE_WRITE
        self.simulation = simulation

    # GET methods
    def get_histogram_snapshot_monitor_filepath(self, pg):
        return os.path.join(self.histogram_snapshot_monitor_dir[pg], self.histogram_snapshot_monitor_file[pg])

    def get_histogram_snapshot_monitor_file(self, pg):
        return self.histogram_snapshot_monitor_file[pg]

    def get_histogram_snapshot_monitor_dir(self, pg):
        return self.histogram_snapshot_monitor_dir[pg]

    def get_histogram_snapshot_monitor_processor_periodicity(self):
        return self.histogram_snapshot_monitor_processor_periodicity

    # Condition check methods
    def is_histogram_snapshot_reference_captured(self, pg):
        assert len(self.histogram_snapshot_reference[pg]) <= 2
        return len(self.histogram_snapshot_reference[pg]) != 0

    def is_histogram_enabled(self):
        return self.histogram_enabled

    # Record reference method - records the first occurrence of the snapshot
    def histogram_snapshot_record_reference(self, pg, path, filename):
        snapshot_file = filename
        snapshot_path = os.path.join(path, snapshot_file)
        if os.path.exists(snapshot_path):
            snapshot_ctime = os.path.getctime(snapshot_path)
            snapshot_ctime_str = datetime.datetime.fromtimestamp(snapshot_ctime)
            self.logger.info("[HISTOGRAM]: Recording reference snapshot file: {}, ctime: {} for pg: {}".format(snapshot_path, snapshot_ctime_str, pg))
            self.histogram_snapshot_reference[pg] = (snapshot_path, snapshot_ctime)

    # Histogram watcher methods
    def histogram_snapshot_watch_init(self, pg):
        watch_dir = self.get_histogram_snapshot_monitor_dir(pg)
        watch_file = self.get_histogram_snapshot_monitor_filepath(pg)
        if not self.config.inotify_watch_has_users(watch_dir, self.config.histogram_key):
            inotify_handle = self.inotify_snapshot_handle
            self.logger.info("[HISTOGRAM]: Watching histogram snapshot dir: {} for pg: {}".format(watch_dir, pg))
            inotify_handle.add_watch(watch_dir, self.inotify_snapshot_mask)
        self.config.add_to_inotify_watch_dir(watch_file, self.config.histogram_key)

    def histogram_snapshot_watch_deinit(self, pg):
        watch_dir = self.get_histogram_snapshot_monitor_dir(pg)
        watch_file = self.get_histogram_snapshot_monitor_filepath(pg)
        self.config.remove_from_inotify_watch_dir(watch_file, self.config.histogram_key)
        if not self.config.inotify_watch_has_users(watch_dir, self.config.histogram_key):
            inotify_handle = self.inotify_snapshot_handle
            self.logger.info("[HISTOGRAM]: Unwatching histogram snapshot dir: {} for pg: {}".format(watch_dir, pg))
            inotify_handle.remove_watch(watch_dir)

    # Histogram snapshot methods
    def histogram_snapshot_init(self):
        self.histogram_snapshot_reader = NetQAsicMonitorHistogramSnapshotReader(self.logger, self.config, self.simulation)
        for pg in self.config.get_histogram_pg_list():
            self.histogram_snapshot_monitor_file[pg] = self.config.get_histogram_snapshot_file(pg)
            self.histogram_snapshot_monitor_dir[pg] = self.config.get_histogram_snapshot_dir(pg)
            self.histogram_snapshot_monitor_count[pg] = self.config.get_histogram_snapshot_file_count(pg)
            self.histogram_snapshot_reference[pg] = ()
            self.histogram_snapshot_last_read_reference[pg] = ()
            self.histogram_snapshot_watch_init(pg)

    def histogram_snapshot_deinit(self):
        self.histogram_snapshot_reader = None
        for pg in self.config.get_histogram_pg_list():
            self.histogram_snapshot_watch_deinit(pg)

        self.histogram_snapshot_monitor_dir = {}
        self.histogram_snapshot_monitor_file = {}
        self.histogram_snapshot_monitor_count = {}
        self.histogram_snapshot_reference = {}
        self.histogram_snapshot_last_read_reference = {}

    def histogram_snapshot_reinit(self):
        self.histogram_snapshot_deinit()
        self.histogram_snapshot_init()

    def histogram_snapshot_file_handler(self, path, filename):
        pg = self.config.get_pg_from_snapshot_file(path, filename)
        if len(self.histogram_snapshot_reference[pg]) == 0:
            self.histogram_snapshot_record_reference(pg, path, filename)
            self.histogram_snapshot_watch_deinit(pg)

    def histogram_snapshot_processor(self):
        lldp_info = {}
        if self.config.get_histogram_pg_list():
            lldp_info = {lldp.ifname: lldp for lldp in Lldp.memmodel().query.filter(hostname=HOSTNAME)}
            if not lldp_info:
                self.logger.debug("[HISTOGRAM]: LLDP peer data not populated")

        for pg in self.config.get_histogram_pg_list():
            if len(self.histogram_snapshot_reference[pg]) == 0:
                continue

            self.logger.debug("[HISTOGRAM]: Processing histogram snapshots for pg: {}".format(pg))
            snapshot_count = self.histogram_snapshot_monitor_count[pg]

            if len(self.histogram_snapshot_last_read_reference[pg]) == 0:
                (snapshot_path, snapshot_ctime) = self.histogram_snapshot_reference[pg]
                snapshot_file_path, snapshot_file_index = snapshot_path.rsplit('_', 1)
                snapshot_file_index = int(snapshot_file_index)
            else:
                (snapshot_path, snapshot_ctime) = self.histogram_snapshot_last_read_reference[pg]
                snapshot_file_path, snapshot_file_index = snapshot_path.rsplit('_', 1)
                snapshot_file_index = (int(snapshot_file_index) + 1) % snapshot_count

            count = 0
            chunks = 10
            histogram_list = []
            while (count <= snapshot_count):
                filepath = snapshot_file_path + '_' + str(snapshot_file_index)
                if (not os.path.exists(filepath)):
                    break

                ctime = os.path.getctime(filepath)
                if ctime < snapshot_ctime:
                    break

                histogram_iflist = self.histogram_snapshot_reader.read_histogram(pg, lldp_info, filepath)
                if not histogram_iflist:
                    # If the data is null that means, we have not populated the data yet. Its in mid of update.
                    # Better to read it in next iteration
                    break

                histogram_list = histogram_list + histogram_iflist

                #self.logger.info("[HISTOGRAM]: Read histogram data({}) - file: {}".format(len(data_dict), filepath))

                # Update snapshot reference
                self.histogram_snapshot_reference[pg] = (filepath, ctime)
                self.histogram_snapshot_last_read_reference[pg] = (filepath, ctime)
                snapshot_file_index = (snapshot_file_index + 1) % snapshot_count
                count += 1

                if (len(histogram_list) != 0 and (count % chunks) == 0):
                    self.logger.debug("[HISTOGRAM]: Sending data chunk {}{}".format(count / 10, len(histogram_list)))
                    send_list_data(histogram_list, AsicMonitorHistogram().__class__.__name__)
                    histogram_list = []

            if len(histogram_list) != 0:
                self.logger.debug("[HISTOGRAM]: Sending last data chunk {}{}".format(count / 10, len(histogram_list)))
                send_list_data(histogram_list, AsicMonitorHistogram().__class__.__name__)

            if count > 0:
                self.logger.info("[HISTOGRAM]: Sent {} data to grpc for pg {}".format(count, pg))


    def histogram_snapshot_monitor_filter(self, header, event_type, path, filename):
        filepath = os.path.join(path, filename)
        pg = self.config.get_pg_from_snapshot_file(path, filename)
        if not self.is_histogram_enabled() or not self.config.is_pg_histogram_type(pg):
            return False

        inotify_mask = self.inotify_snapshot_mask
        if header.mask & inotify_mask:
            if self.histogram_snapshot_monitor_file[pg] in filepath:
                self.logger.info("[HISTOGRAM]: Histogram snapshot was created/modified for pg: {}".format(pg))
                return True

        return False

    # Histogram methods
    def histogram_init(self):
        self.histogram_snapshot_init()

    def histogram_deinit(self):
        self.histogram_snapshot_deinit()

    def histogram_reinit(self):
        self.histogram_snapshot_reinit()


    # Histogram configuration handler method
    def histogram_handle_config(self, force_disable=False):
        #if (force_disable):
        #    self.logger.info("Forcing histogram deinitialisation")
        #    self.histogram_enabled = False
        #    self.histogram_deinit()
        #    return

        histogram_enabled = self.config.is_histogram_config_enabled()
        if self.histogram_enabled != histogram_enabled:
            if (histogram_enabled):
                self.logger.info("[HISTOGRAM]: Histogram config is enabled")
                self.histogram_init()
            else:
                self.logger.info("[HISTOGRAM]: Histogram config is disabled")
                self.histogram_deinit()
        elif self.histogram_enabled:
            self.logger.info("[HISTOGRAM]: Histogram config changed")
            self.histogram_reinit()
        else:
            self.logger.info("[HISTOGRAM]: Histogram in disabled state")

        self.histogram_enabled = histogram_enabled

class NetQAsicMonitor(object):
    def __init__(self, logger, thread_pool=None, thread_stop_callback=None, monitor_min_file_count=120, monitor_collection_periodicity=30, simulation=False):

        self.logger = logger
        self.thread_pool = thread_pool
        self.thread_stop_callback = thread_stop_callback
        self.config = None
        self.buffer_stats = None
        self.histogram = None

        self.monitor_config_dir = None
        self.monitor_config_file = None
        self.monitor_thread = None
        self.monitor_thread_event = threading.Event()

        self.monitor_min_file_count = monitor_min_file_count
        self.monitor_collection_periodicity = monitor_collection_periodicity

        self.inotify_handle = inotify.adapters.Inotify()
        self.inotify_mask = inotify.constants.IN_CLOSE_WRITE

        self.simulation = simulation

        self.initialised = False

    # GET methods
    def get_monitor_min_file_count(self):
        return self.monitor_min_file_count

    def get_monitor_collection_periodicity(self):
        return self.monitor_collection_periodicity

    # Monitor thread methods
    def monitor_thread_start(self):
        if self.monitor_thread == None:
            return

        self.monitor_thread_event.clear()
        if self.thread_pool == None:
            self.monitor_thread.start()

    def monitor_thread_stop(self):
        if self.monitor_thread == None:
            return

        self.monitor_thread_event.set()
        if self.thread_pool == None:
            self.monitor_thread.join()

    def monitor_thread_init(self, callback):
        if self.monitor_thread != None:
            return

        if self.thread_pool != None:
            self.monitor_thread = self.thread_pool.spawn(callback)
            self.monitor_thread.link(self.thread_stop_callback)
        else:
            self.monitor_thread = threading.Thread(target=callback)
            self.monitor_thread.daemon = True
            self.monitor_thread_start()
        self.logger.info("Monitor thread was started")

    def monitor_thread_deinit(self):
        self.monitor_thread_stop()
        self.monitor_thread = None
        self.logger.info("Monitor thread was stopped")

    # Monitor callback - for monitor all files
    def monitor_callback(self):
        inotify_handle = self.inotify_handle
        while (True):
            if self.monitor_thread_event.is_set():
                self.logger.info("Exiting monitor thread")
                break

            for event in inotify_handle.event_gen(yield_nones=True):
                if event == None:
                    break

                (header, event_type, path, filename) = event
                self.logger.debug("TYPE={}, PATH={}, FILENAME={}".format(type, path, filename))
                if self.monitor_config_filter(header, event_type, path, filename):
                    self.monitor_config_handler()

                if self.histogram.histogram_snapshot_monitor_filter(header, event_type, path, filename):
                    self.histogram.histogram_snapshot_file_handler(path, filename)

                if self.buffer_stats.buffer_stats_snapshot_monitor_filter(header, event_type, path, filename):
                    self.buffer_stats.buffer_stats_snapshot_file_handler(path, filename)

            time.sleep(10)

    # Monitor configuration methods
    def monitor_config_filter(self, header, event_type, path, filename):
        result = False
        inotify_mask = self.inotify_mask
        if header.mask & inotify_mask:
            created_or_modified_file = os.path.join(path, filename)
            if created_or_modified_file == self.config.get_config_monitor_file():
                self.logger.info("monitor.conf was created/modified")
                result = True

        return result

    def monitor_config_handler(self):
        valid, config_changed = self.config.handle_config()
        if (valid and config_changed):
            self.logger.info("Configuration was changed. Notifying users")
            self.histogram.histogram_handle_config()
            self.buffer_stats.buffer_stats_handle_config()
        elif not valid:
            self.logger.info("Configuration was invalid. Notifying users")
            self.config.clear_configs()
            self.histogram.histogram_handle_config(force_disable=True)
            self.buffer_stats.buffer_stats_handle_config(force_disable=True)

    def monitor_config_watch_init(self):
        watch_dir = self.config.get_config_monitor_dir()
        watch_file = self.config.get_config_monitor_file()
        if not self.config.inotify_watch_has_users(watch_dir, self.config.config_key):
            inotify_handle = self.inotify_handle
            self.logger.info("Watching asic monitor config dir: {}".format(watch_dir))
            inotify_handle.add_watch(watch_dir, self.inotify_mask)
        self.config.add_to_inotify_watch_dir(watch_file, self.config.config_key)

    def monitor_config_watch_deinit(self):
        watch_dir = self.config.get_config_monitor_dir()
        watch_file = self.config.get_config_monitor_file()
        self.config.remove_from_inotify_watch_dir(watch_file, self.config.config_key)
        if not self.config.inotify_watch_has_users(watch_dir, self.config.config_key):
            inotify_handle = self.inotify_handle
            self.logger.info("Unwatching asic monitor config dir: {}".format(watch_dir))
            inotify_handle.remove_watch(watch_dir)
            self.inotify_config_watch_on = False

    def monitor_config_watch_reinit(self):
        self.config_watch_deinit()
        self.config_watch_init()

    def monitor_config_init(self):
        if self.config == None:
            self.config = NetQAsicMonitorConfig(self.logger, self.monitor_min_file_count, self.monitor_collection_periodicity)

        self.monitor_config_file = self.config.get_config_monitor_file()
        self.monitor_config_dir = self.config.get_config_monitor_dir()
        valid, config_changed = self.config.handle_config()
        self.monitor_config_watch_init()
        self.monitor_thread_init(callback=self.monitor_callback)
        return valid

    def monitor_config_deinit(self):
        self.monitor_thread_deinit()
        self.monitor_config_watch_deinit()
        self.monitor_config_dir = None
        self.monitor_config_file = None
        self.config = None

    # Monitor methods - init/deinit
    def monitor_deinit(self):
        self.initialised = False
        self.histogram.histogram_deinit()
        self.buffer_stats.buffer_stats_deinit()
        self.monitor_config_deinit()

    def monitor_init(self):
        if (self.initialised == True):
            self.monitor_deinit()
            self.initialised = False

        valid = self.monitor_config_init()

        if self.histogram == None:
            self.histogram = NetQAsicMonitorHistogram(self.logger, self.config, self.inotify_handle, simulation=self.simulation)
            if valid:
                self.histogram.histogram_handle_config()

        if self.buffer_stats == None:
            self.buffer_stats = NetQAsicMonitorBufferStats(self.logger, self.config, self.inotify_handle)
            if valid:
                self.buffer_stats.buffer_stats_handle_config()

        self.initialised = True


class NetQAsicMonitorService(Service):
    def __init__(self, logger):
        self.logger = logger
        shutdown_cb = [unsubscribe_listener]
        super(NetQAsicMonitorService, self).__init__({}, "NetQAsicMonitorService", shutdown_callbacks=shutdown_cb)
        self.timers = {"asic-monitor-histogram-processor": 0, "asic-monitor-buffer-stats-processor": 0}
        self.histogram_timeout = 30
        self.buffer_stats_timeout = 30
        self.asic_monitor = NetQAsicMonitor(logger)
        self.inotify_config_thread = None
        self.inotify_snapshot_thread = None

    def _run(self):
        try:
            self.asic_monitor.monitor_init()

            while (True):
                epoch = time.time()
                if self.timers['asic-monitor-histogram-processor'] == 0 or (
                        epoch - self.timers['asic-monitor-histogram-processor']) > self.histogram_timeout:
                    self.logger.info("Spawning histogram processor thread")
                    gt_thread = self._pool.spawn(self.asic_monitor.histogram.histogram_snapshot_processor)
                    gt_thread.link(self._stop_checker)
                    self.timers['asic-monitor-histogram-processor'] = epoch

                if self.timers['asic-monitor-buffer-stats-processor'] == 0 or (
                        epoch - self.timers['asic-monitor-buffer-stats-processor']) > self.buffer_stats_timeout:
                    self.logger.info("Spawning buffer stats processor thread")
                    gt_thread = self._pool.spawn(self.asic_monitor.buffer_stats.buffer_stats_snapshot_processor)
                    gt_thread.link(self._stop_checker)
                    self.timers['asic-monitor-buffer-stats-processor'] = epoch

                time.sleep(10)

        except KeyboardInterrupt as e:
            self.logger.info("Exception in run ex={}".format(e))
            return


#if __name__ == "__main__":
#    try:
#        ams = NetQAsicMonitorService(print)
#        ams.run()
#    except KeyboardInterrupt as e:
#        self.logger.info("Exception encountered ex={}".format(e))
#        asic_monitor.monitor_deinit()
#        sys.exit(0)
