#
# Copyright (c) 2018-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import socket

from netq_agent.cmd.netq_agent_stream import NetQAgentStream
from netq_lib.common.netq_pq import QueueElem
from netq_lib.common import configuration
from netq_lib.common import utils
from netq_lib.common.enums import ConfKw
from threading import Thread
import time

# Logging
import logging
logger = logging.getLogger('netq-grpc')
consoleHandler = logging.StreamHandler()
logFormatter = logging.Formatter(
    "%(asctime)s [%(filename)s:%(lineno)d] [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
consoleHandler.setFormatter(logFormatter)
logger.addHandler(consoleHandler)
logger.setLevel(logging.DEBUG)

_PROC_NAME = 'netq-agent'
MAX_BACKLOG = 5
RECV_BUF_SIZE = 1024


class GRPCServer(object):

    def __init__(self, conf, priority_queue, backpressure_priority_queue):
        self._conf = conf
        self._logger = utils.get_logger(_PROC_NAME, self._conf)

        try:
            self._logger.info("creating message queues")
            self.__pq = priority_queue
            self.__bp_pq = backpressure_priority_queue

        except Exception as e:
            self._logger.error("Failed to create Message Queue {}".format(e))
            return

    @staticmethod
    def _get_netq_pq(self):
        return self._pq

    def _get_message_priority(self, data):
        return 1

    def process_ingress_data(self, conn):

        while True:
            try:
                ingress_data = conn.recv(RECV_BUF_SIZE)
                if not ingress_data:
                    conn.close()
                    break
            except KeyboardInterrupt:
                self._logger.error("Keyboard Interrupt")
                break

            except Exception:
                self._logger.error("Socket Connect error ")
                break

            pq = self._get_netq_pq()

            priority = self._get_message_priority(ingress_data)

            q_elem = QueueElem(priority, ingress_data)
            pq.putq(q_elem)

    def process_ingress(self, conf):

        HOST = conf['netq-agent'].get(ConfKw.AGENT_URL, ConfKw.DEFAULT_AGENT_URL)
        PORT = 8985
        server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.bind((HOST, PORT))
        server.listen(MAX_BACKLOG)
        self._logger.info("Starting to receive IPC data")
        while True:
            conn, addr = server.accept()
            self._logger.info("Ingress conn received from {}".format(addr))
            ingress_handler = Thread(
                target=self.process_ingress_data, args=(conn,))
            ingress_handler.start()
        conn.close()

    def process_stream_thread(self, conf):

        agent_stream = NetQAgentStream(conf, self._logger, self.__pq, self.__bp_pq)
        agent_stream.stream_main()

    def run(self, conf):

        try:
            ingress_thread = Thread(target=self.process_ingress, args=(conf,))

            stream_thread = Thread(
                target=self.process_stream_thread, args=(conf,))

            ingress_thread.start()
            stream_thread.start()
            ingress_thread.join()
            stream_thread.join()
            while True:
                time.sleep(1000)
        except KeyboardInterrupt:
            exit()

        except Exception as e:
            logger.error("NETQ Grpc: caught exception: {}".format(e))


def main():
    conf = configuration.get_config(_PROC_NAME)

    grpc_server_inst = GRPCServer(conf, logger)
    grpc_server_inst.run(conf)


if __name__ == '__main__':
    main()
