#
# Copyright (c) 2019-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import socket
import ssl
import time
import json
import sys
import threading
import random
from netq_lib.common.enums import AGENTSTATUS, ConfKw, TransMode, \
    ModelConstants as MC
from netq_lib.orm import HOSTNAME
from netq_lib.common.netq_pq import QueueElem
import psutil
import os
from collections import defaultdict
from multiprocessing.connection import wait
from netq_agent.cmd.grpc_client_base import GrpcClientBase, GrpcClientThread
from netq_agent.cmd.netq_gnmi_driver import GNMIDriver
from netq_lib.common import configuration
from netq_lib.common.utils import is_gnmi_enabled, netq_agent_misc_warn_write, \
    netq_agent_misc_warn_resolve_write, LogCoalescer

import grpc
if grpc._grpcio_metadata.__version__ < "1.22.0":
    from grpc._channel import _moot as unsubscribe_all
from netq_agent.interface.netq_opta_stream_pb2 import (
    NetqAgentMessage,
)

from netq_agent.interface.netq_opta_stream_pb2_grpc import (
    NetqOPTAStreamServiceStub
)

_PROC_NAME = 'netq-agent'

class ChannelState():
    IDLE = 1
    CONNECTING = 2
    CONNECTED = 3
    DISCONNECTED = 4
    SHUTDOWN = 5


RPC_CONNECTIVITY_MAPPING = {}
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.IDLE] = ChannelState.IDLE
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.CONNECTING] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.READY] = ChannelState.CONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.TRANSIENT_FAILURE] = ChannelState.DISCONNECTED
RPC_CONNECTIVITY_MAPPING[grpc.ChannelConnectivity.SHUTDOWN] = ChannelState.IDLE


class NetQAgentGrpcClient(GrpcClientBase):

    def __init__(self, conf, logger, pq, backpressure_queue, gnmi_pq=None):
        super(NetQAgentGrpcClient, self).__init__(
            logger, pq, backpressure_queue, MC.GRPC_CLIENT_OPTA,
            connection_callback=self.handle_opta_connection)
        self._conf = conf
        self._opta_enabled = self.__is_opta_enabled()
        self.channel_state = ChannelState.IDLE
        self._pq = pq
        self._bp_pq = backpressure_queue
        self.seq_no = 0
        self.max_opta_retry = 3  # max wait for ready retires per channel
        self.max_channel_recreate = 1 \
            if self._conf[ConfKw.NETQ_AGENT][ConfKw.ENABLE_OPTA_DISCOVERY] else 3   # max number of channels to be recreated
        self._grpc_backoff_time = 0

        self.__logger_coalescer = LogCoalescer("netq-agent", conf_dict=configuration.get_config(_PROC_NAME))

        self.__MAX_GRPC_BACKOFF = 30
        self.__chunk_count = {}
        self.__total_chunk_count = 0
        self.__opta_grps_reset_count = 0
        self.__util_stats = {}
        self.__pq_util_stats = defaultdict(int)

        # Rate Control */
        self._wait_time = 0.0
        self._last_transmission_time = 0.0
        self._elapsed_time = 0.0
        self._min_interval = 0.2
        self._max_message_per_transaction = 1000
        self._max_msg_threshold = 50
        self._wakeup_time = 0.2
        self._wait_for_ready_timeout = 10
        self.__primary_opta_addr = None
        self.__opta_server_list = []
        self.__channel_retry_count = 0  # wait for ready state per channel retry count
        self.__channel_recreate_count = 0   # number of channels recreated count

        # GRPC timeout
        self._grpc_call_timeout = 300

        self.__exit_stats_thread = False

        # gNMI agent
        self._gnmi_pq = gnmi_pq
        self._gnmi_enabled = is_gnmi_enabled(self._conf)
        if self._gnmi_enabled:
            address = self._conf.get(ConfKw.NETQ_AGENT, {}).get(
                ConfKw.NETQ_STREAM_ADDRESS, ConfKw.DEFAULT_NETQ_STREAM_ADDRESS)
            port = self._conf.get(ConfKw.NETQ_AGENT, {}).get(
                ConfKw.NETQ_STREAM_PORT, ConfKw.DEFAULT_NETQ_STREAM_PORT)
            self.gnmi_driver = GNMIDriver(address, port, self._gnmi_pq,
                                          self._bp_pq, self._logger)

        try:
            self._num_msg_per_packet = self._conf['netq-agent']['num_msgs_per_packet']
        except Exception:
            self._num_msg_per_packet = 30

        self._logger.info("Number of messages per packet {}".format(self._num_msg_per_packet))

    def __is_opta_enabled(self):
        return self._conf.get('netq-agent', {}).get(
            ConfKw.IS_OPTA_ENABLED, ConfKw.DEFAULT_IS_OPTA_ENABLED)

    def callback_on_exception(self, exception, traceback):

        self._logger.error(traceback)
        self._logger.error(exception)

    def _run(self):
        pass

    @property
    def channel_state(self):
        return self._channel_state

    @channel_state.setter
    def channel_state(self, new_state):
        if new_state in RPC_CONNECTIVITY_MAPPING:
            new_state = RPC_CONNECTIVITY_MAPPING[new_state]
        self._logger.debug("New RPC Channel State {}".format(new_state))
        self._channel_state = new_state

    def _get_agent_connect_id(self, verbose=False):
        id = None
        try:
            opta_url = self._conf['netq-agent'][ConfKw.SERVER]
            opta_port = self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect((opta_url, opta_port))
            id = s.getsockname()[0]
        except Exception:
            self._logger.error("Failed to connect to OPTA")

        if verbose:
            if id:
                self._logger.info("OPTA Discovery Agent Connect IP {}".format(id))
            else:
                self._logger.info("OPTA Discovery No Connect IP determined")
        return id

    def get_agent_id(self):
        if not self.is_valid_hostname(HOSTNAME):
            return None
        return HOSTNAME

    def is_valid_hostname(self, hostname):
        if hostname.lower() in ["cumulus", "sonic"]:
            return False
        return True

    def full_state_message_builder(self, msg_type, segment, seq_no, json_out):

        netq_message = NetqAgentMessage(aid=HOSTNAME,
                                        message_type=msg_type,
                                        seqno=seq_no,
                                        ts=int(time.time() * 1000),
                                        trans_mode=segment,
                                        type=0,
                                        message=json.dumps(json_out).encode('utf-8'))
        return netq_message

    def get_full_state_message(self, msg):

        num_msg_per_packet = self._num_msg_per_packet
        if 'priority_qid' not in list(msg.keys()) or 'message_type' not in list(msg.keys()):
            self._logger.error('Invalid full state message')

        pq_idx = msg.get('priority_qid')
        queue_len = self._pq.len(pq_idx)
        num_segments = int(((queue_len - 1) / num_msg_per_packet) + 1)
        counter = msg.get('counter')
        self._logger.debug("pq_idx = {} queue_len = {} num_segments = {}".format(pq_idx, queue_len, num_segments))

        msg_data = []
        num_msgs = 0

        seq_no = self.get_sequence_number()
        segment_idx = 1

        for priority, message_type, fullstate, msg in self._pq.get_buffer(pq_idx, counter):
            msg_data.append(msg.data)
            num_msgs += 1
            if (num_msgs == num_msg_per_packet):

                self.__chunk_count[message_type] = self.__chunk_count.get(message_type, 0) + 1
                self.__total_chunk_count = self.__total_chunk_count + 1

                if (segment_idx == num_segments):
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.LAST_SEGMENT, seq_no, msg_data)
                elif (segment_idx == 1):
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.FIRST_SEGMENT, seq_no, msg_data)
                else:
                    rpc_msg = self.full_state_message_builder(message_type, TransMode.MID_SEGMENT, seq_no, msg_data)
                msg_data = []
                num_msgs = 0
                segment_idx += 1
                yield rpc_msg

        if len(msg_data) > 0:
            rpc_msg = self.full_state_message_builder(message_type, TransMode.LAST_SEGMENT, seq_no, msg_data)
            yield rpc_msg

    def _get_wait_time(self):
        self._elapsed_time = time.time() - self._last_transmission_time
        return self._min_interval - self._elapsed_time

    def grpc_init_rate_control(self):
        self._wait_time = self._get_wait_time()
        self.__logger_coalescer.debug(
            "elapsed time {} wait time {} min interval {} queue size {}".format(self._elapsed_time, self._wait_time,
                                                                                self._min_interval, self._pq.len(1)))
        while self._wait_time > 0:
            q_len = self._pq.len_all()
            if q_len > self._max_msg_threshold:
                break
            else:
                self.__logger_coalescer.debug(
                    "Wakeup time {} Qlen {} max_threshold {}".format(self._wakeup_time, q_len, self._max_msg_threshold))
                time.sleep(self._wakeup_time)
            self._wait_time = self._get_wait_time()

    def _send_wakeup_message(self):
        message = {}
        message['message_type'] = 'NETQ_RATE_CONTROL'
        message['priority_qid'] = 1
        priority = 1
        try:
            queue_elem = QueueElem(priority, "NETQ_RATE_CONTROL", True, message)
            self._pq.putq(queue_elem)
        except Exception:
            return False

    def process_msg_data(self, msg_type, json_out):

        if msg_type == "NETQ_RATE_CONTROL":
            return True

        elif msg_type == "AgentStats":

            try:
                pq_util = {'average_pq_util': float(self.__pq_util_stats.get('sum_pq_util', 0)) / self.__pq_util_stats.get(
                    'count', 1),
                           'max_pq_util': self.__pq_util_stats.get('max_pq_util', 0),
                           'min_pq_util': self.__pq_util_stats.get("min_pq_util", 0)}

                json_out['pq_util'] = pq_util
                self.__pq_util_stats = defaultdict(int)

                json_out['message_count']['total_chunk_count'] = self.__total_chunk_count
                self.__total_chunk_count = 0
                json_out['chunk_count'] = self.__chunk_count
                self.__chunk_count = {}
                json_out['opta_grps_reset_count'] = self.__opta_grps_reset_count
                self.__opta_grps_reset_count = 0

                grpc_util = {'grpc_pid': self.__util_stats.get('grpc_pid', os.getpid()),
                             'max_cpu_util': self.__util_stats.get('max_cpu_util', 0.0),
                             'min_cpu_util': self.__util_stats.get('min_cpu_util', 0.0),
                             'avg_cpu_util': self.__util_stats.get('sum_cpu_util', 0.0) / self.__util_stats.get('count',
                                                                                                                1),
                             'max_mem_util': self.__util_stats.get('max_mem_util', 0.0),
                             'min_mem_util': self.__util_stats.get('min_mem_util', 0.0),
                             'avg_mem_util': self.__util_stats.get('sum_mem_util', 0.0) / self.__util_stats.get('count',
                                                                                                                1)}

                json_out['grpc_util'] = grpc_util
                self.__util_stats = {}
            except Exception as e:
                self._logger.debug("Agent Status Disabled..{}", e)
                return True

            return False

        else:
            return False

    def get_stream(self):

        pq = self._pq
        fd_set = pq.get_reader_fds()
        my_run_id = 1

        try:
            epoll_input = wait(fd_set, 10)
            if not len(epoll_input):
                return
            self.grpc_init_rate_control()
            msg_cnt = 0

            for queue_idx in range(0, self._pq.max_num_queue):
                for priority, message_type, fullstate, msg in pq.get_buffer(queue_idx):

                    ignore_msg_type = self.process_msg_data(message_type, msg.data)

                    if ignore_msg_type:
                        continue

                    if message_type == 'FULL_STATE_END':
                        full_gen = self.get_full_state_message(msg.data)
                        for segment in full_gen:
                            yield segment
                            msg_cnt += 1
                    else:
                        rpc_msg = self.message_builder(message_type, fullstate, msg.data)
                        yield rpc_msg
                    msg_cnt += 1
                    if msg_cnt >= self._max_message_per_transaction:
                        self._logger.info("GRPC: Reached max messages per transaction")
                        self._send_wakeup_message()
                        break
            self._last_transmission_time = time.time()

        except Exception as e:
            self._logger.error("GRPC %d StreamMessage failed %s", my_run_id, e)
            self._last_transmission_time = time.time()
            return

    def _get_opta_addr(self):
        if self.__primary_opta_addr in self.__opta_server_list:
            self.__opta_server_list.remove(self.__primary_opta_addr)

        if len(self.__opta_server_list) == 0:
            available_opta_servers = ''
            if self._conf['netq-agent'].get(ConfKw.CLUSTER_SERVERS):
                available_opta_servers = self._conf['netq-agent'][ConfKw.CLUSTER_SERVERS]
            elif self._conf['netq-agent'].get(ConfKw.SERVER):
                available_opta_servers = self._conf['netq-agent'][ConfKw.SERVER]
            opta_port = self._conf['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)

            for server_addr in available_opta_servers.split(','):
                opta_addr = "{}:{}".format(server_addr, opta_port)
                self.__opta_server_list.append(opta_addr)

        # pickup a random server address from healthy servers
        self.__primary_opta_addr = random.choice(self.__opta_server_list)
        return self.__primary_opta_addr

    def set_opta_discovery_ip(self, discovery_ip):
        self._logger.info('Setting discovery server: {}'.format(discovery_ip))
        self.__opta_server_list = [discovery_ip]
        # Resetting, special case fo status reconnect, where if not done it would remove ip from self.__opta_server_list
        self.__primary_opta_addr = None

    def monitor_channel(self, state):
        self._logger.info("Channel State {}".format(state))
        self.channel_state = state

    def wait_for_ready(self):
        try:
            grpc.channel_ready_future(self._channel).result(timeout=self._wait_for_ready_timeout)
            self._logger.debug("Channel connection successful")
            return AGENTSTATUS.SUCCESS
        except grpc.FutureTimeoutError:
            self._logger.info("Channel Failed to connect")
            return AGENTSTATUS.FAILURE

    def reestablish_channel(self):
        if self._channel:
            if self.__channel_retry_count < self.max_opta_retry:
                return self.wait_for_ready()
            else:
                try:
                    self._close()
                except Exception as e:
                    self._logger.info("Channel Close exception {}".format(e))
                if self.__channel_recreate_count >= self.max_channel_recreate:
                    self._logger.info("Max channel recreate attempts exhausted. Unable to connect with OPTA")
                    return AGENTSTATUS.FAILURE

        try:
            serv_addr = self._get_opta_addr()
            self.__opta_grps_reset_count += 1
            self._logger.info("Reestablishing OPTA Connection {}".format(serv_addr))
            if not self.is_ssl_enabled():
                self._channel = grpc.insecure_channel(serv_addr)
            else:
                credentials = self.get_credentials(serv_addr)
                self._channel = grpc.secure_channel(serv_addr, credentials)

        except Exception as e:
            self._logger.error(
                "Unable to establish connection with OPTA {}".format(e))
            return AGENTSTATUS.FAILURE

        try:
            if self._channel:
                self._channel.subscribe(self.monitor_channel, try_to_connect=True)
                self.wait_for_ready()
            else:
                self._logger.error("Unable to connect to OPTA")
                return AGENTSTATUS.FAILURE

        except Exception as e:
            self._logger.error("Unable to subscribe to OPTA")
            return AGENTSTATUS.FAILURE

        if self.channel_state != ChannelState.CONNECTED:
            return AGENTSTATUS.FAILURE

        return AGENTSTATUS.SUCCESS

    def _close(self):
        try:
            if self._channel:
                self._logger.info("Closing channel {}".format(self._channel))
                self._channel.unsubscribe(self.monitor_channel)
                if grpc._grpcio_metadata.__version__ < "1.22.0":
                    unsubscribe_all(self._channel._connectivity_state)
                self._channel.close()
            self._channel = None
        except Exception as e:
            self._logger.error("Channel Close Error {}".format(e))
            self._channel = None

    def run_contiguous_stream(self):
        try:
            sts = self.handle_connection()
            if sts != AGENTSTATUS.SUCCESS:
                self._logger.error("failed to established OPTA Stream connection, "
                                   "max retries reached. exiting thread")
                return AGENTSTATUS.FAILURE
        except Exception as e:
            self._logger.error("failed to established OPTA Stream connection, "
                               "max retries reached. exiting thread: {}".format(e))
            return AGENTSTATUS.FAILURE

        opta_stub = NetqOPTAStreamServiceStub(self._channel)
        metadata = [('aid', HOSTNAME)]

        while self.is_active():
            try:
                res, call = opta_stub.StreamMessage.with_call(self.get_stream(), metadata=metadata,
                                                              timeout=self._grpc_call_timeout)
                self.__logger_coalescer.debug("RPC Completed {} status {}".format(res, call.code()))
                if call.code() != grpc.StatusCode.OK:
                    self._logger.error("Agent Stream Failed {}".format(call.code()))
                    self._close()
                    self.send_pause_state(False)
                    return AGENTSTATUS.FAILURE
            except grpc.RpcError as e:
                self._logger.error("Agent Stream Failed Code {} Info: {}".format(e.code(), e.details()))
                self._close()
                self.send_pause_state(False)
                return AGENTSTATUS.FAILURE
            except Exception as e:
                self._logger.error("Agent Stream failed error {}".format(e))
                self._close()
                self.send_pause_state(False)
                return AGENTSTATUS.FAILURE

    def run_grpc_stats_collection(self):
        while not self.__exit_stats_thread:
            try:
                self.__pq_util_stats = self._pq.total_pq_usage(self.__pq_util_stats)

                agent_process = psutil.Process(os.getpid())
                self.__util_stats['grpc_pid'] = os.getpid()

                agent_cpu_usage = agent_process.cpu_percent(interval=0.1)
                total_mem = psutil.virtual_memory().total
                agent_mem_usage = float((agent_process.memory_info().vms * 100)) / total_mem

                self.__util_stats['count'] = self.__util_stats.get('count', 0) + 1

                if agent_cpu_usage >= self.__util_stats.get('max_cpu_util', 0.0):
                    self.__util_stats['max_cpu_util'] = agent_cpu_usage

                if agent_cpu_usage <= self.__util_stats.get('min_cpu_util', sys.maxsize):
                    self.__util_stats['min_cpu_util'] = agent_cpu_usage

                self.__util_stats['sum_cpu_util'] = self.__util_stats.get('sum_cpu_util', 0.0) + agent_cpu_usage

                if agent_mem_usage >= self.__util_stats.get('max_mem_util', 0.0):
                    self.__util_stats['max_mem_util'] = agent_mem_usage

                if agent_mem_usage <= self.__util_stats.get('min_mem_util', 0.0):
                    self.__util_stats['min_mem_util'] = agent_mem_usage

                self.__util_stats['sum_mem_util'] = self.__util_stats.get('sum_mem_util', 0.0) + agent_mem_usage

                time.sleep(15)

            except Exception as e:
                self._logger.error("Stats Collection Failed {}".format(e))
                return AGENTSTATUS.FAILURE

    def is_ssl_enabled(self):
        return self._conf['netq-agent'].get(ConfKw.IS_SSL_ENABLED, False)

    def get_credentials(self, serv_addr):
        if self._conf['netq-agent'].get(ConfKw.SSL_CERT):
            with open(self._conf['netq-agent'].get(ConfKw.SSL_CERT), 'rb') as f:
                trusted_certs = f.read()
                return grpc.ssl_channel_credentials(root_certificates=trusted_certs)
        if self._conf['netq-agent'].get(ConfKw.DOWNLOAD_SSL_CERT):
            return grpc.ssl_channel_credentials(ssl.get_server_certificate(serv_addr.split(':')).encode())
        else:
            return grpc.ssl_channel_credentials()

    def get_grpc_backoff_time(self):
        self._logger.info("Current Back off time {}".format(self._grpc_backoff_time))
        if self._grpc_backoff_time >= self.__MAX_GRPC_BACKOFF:
            return self.__MAX_GRPC_BACKOFF
        else:
            self._grpc_backoff_time = 1 + self._grpc_backoff_time * 2
            return self._grpc_backoff_time

    def reset_channel_retry_count(self):
        self.__channel_retry_count = 0

    def get_channel_retry_count(self):
        return self.__channel_retry_count

    def update_channel_retry_count(self):
        self.__channel_retry_count = self.__channel_retry_count + 1

    def is_channel_recreate_limit_reached(self):
        # This API has to be called to check the number of times channel recreation is allowed
        if self.__channel_recreate_count >= self.max_channel_recreate:
            # Max GRPC channel recreation limit exceeded
            return True
        else:
            self.__channel_recreate_count = self.__channel_recreate_count + 1
            return False

    def handle_opta_connection(self):
        logged_prev = False
        while True:
            if AGENTSTATUS.SUCCESS != self.reestablish_channel():
                msg_string = "Failed to establish Connection with OPTA, retrying with backoff"
                self._logger.info(msg_string)
                netq_agent_misc_warn_write(msg_string)
                logged_prev = True

                # Additional debug logs
                self._logger.info("GRPC channel recreate {}, channel retry {}".format(
                    self.__channel_recreate_count,
                    self.__channel_retry_count))

                if self.__channel_retry_count >= self.max_opta_retry:
                    self.send_pause_state(False)
                    self.__channel_retry_count = 0
                    # New Channel Created
                    if self.is_channel_recreate_limit_reached():
                        self.__exit_stats_thread = True
                        netq_agent_misc_warn_write('Channel recreate limit reached')
                        raise Exception("Channel recreate limit reached")
                else:
                    self.__channel_retry_count = self.__channel_retry_count + 1
                timeVal = self.get_grpc_backoff_time()
                time.sleep(timeVal)
                continue
            else:
                if logged_prev:
                    msg_string = "Established connection with OPTA (Warning resolved)"
                    netq_agent_misc_warn_resolve_write(msg_string)
                break
        self.send_pause_state(True)
        return AGENTSTATUS.SUCCESS

    def start_stream(self, verbose=False):

        stats_thread = threading.Thread(target=self.run_grpc_stats_collection, args=())
        stats_thread.start()
        opta_thread = None
        gnmi_thread = None
        self._logger.info("Agent Stream started")
        while True:
            try:
                if self._opta_enabled and not opta_thread:
                    self._logger.info("Starting OPTA Stream")
                    opta_thread = GrpcClientThread(
                        target=self.start_contiguous_stream)
                    # stream_response = opta_stub.StreamMessage(self.get_stream())
                    opta_thread.start()
                if self._gnmi_enabled and not gnmi_thread:
                    self._logger.info("Starting gNMI Stream")
                    gnmi_thread = GrpcClientThread(
                        target=self.gnmi_driver.start_contiguous_stream)
                    gnmi_thread.start()
                if opta_thread:
                    err = opta_thread.join(10)
                    if not opta_thread.is_alive():
                        self.stop()
                        self._logger.info("OPTA Stream ended")
                        opta_thread = None
                        if err:
                            raise Exception(err)
                        continue
                if gnmi_thread:
                    err = gnmi_thread.join(10)
                    if not gnmi_thread.is_alive():
                        self.gnmi_driver.stop()
                        self._logger.info("gNMI Stream ended")
                        gnmi_thread = None
                        if err:
                            raise Exception(err)
                        continue

                self._logger.debug("Thread processing complete")
            except Exception as e:
                self._logger.error("OPTA Stream Unknown GRPC Exception {}".format(e))
                self.__exit_stats_thread = True
                if opta_thread:
                    self.stop()
                    opta_thread.join()
                if gnmi_thread:
                    self.gnmi_driver.stop()
                    gnmi_thread.join()
                break

            self._logger.debug("Stream Completed")

        self._logger.info("Exiting GRPC Process!")
        stats_thread.join()

