#
# Copyright (c) 2019-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import binascii
import ctypes
import json
import os
import subprocess
import sys
import time
import signal
import ipaddress as ipaddr
from impacket.ImpactDecoder import *
from impacket.ImpactPacket import *
from ctypes import (
    c_uint32, c_char_p, byref)

from netq_agent.cmd.base_wjh_agent import BaseWjhClient
from netq_agent.handlers.port_handler import PortHandler
from netq_agent.handlers.sdk_utils import SX_CHIP_TYPE_SPECTRUM2
from netq_lib.common.enums import WjhConfKw, DefaultPaths
from netq_lib.common.utils import subprocess_wrapper, is_sonic, is_cl3, is_cl4, is_active_systemd_svc, send_list_data
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import (
    WjhDropRawInfo,
    WjhL1DropAggregate,
    WjhDropAggregate)

sys.path.append('/usr/lib/cumulus/wjh/')
from wjh_types import *

AGENTSTATUS_SUCCESS = 0
AGENTSTATUS_FAILURE = 1


WJH_TRAP_GROUP_ALLOCATE_MODE_STATIC_E = 0
WJH_TRAP_GROUP_ALLOCATE_MODE_DYNAMIC_E = 1


WJH_DROP_REASON_ID_PORT_LOOPBACK_FILTER = 208
WJH_DROP_REASON_ID_NON_ROUTABLE_PACKET_E = 301
WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX0_16_E = 305
WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX1_16_E = 306
WJH_DROP_REASON_ID_NON_IP_PACKET_E = 307

WJH_DROP_REASON_ID_MLAG_PORT_ISOLATION_E = 201
WJH_DROP_REASON_ID_INGRESS_SPANNING_TREE_FILTER_E = 205

default_filters = [WJH_DROP_REASON_ID_PORT_LOOPBACK_FILTER,
                   WJH_DROP_REASON_ID_NON_ROUTABLE_PACKET_E,
                   WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX0_16_E,
                   WJH_DROP_REASON_ID_IPV6_DESTINATION_IN_MULTICAST_SCOPE_FFX1_16_E,
                   WJH_DROP_REASON_ID_NON_IP_PACKET_E,
                   WJH_DROP_REASON_ID_MLAG_PORT_ISOLATION_E,
                   WJH_DROP_REASON_ID_INGRESS_SPANNING_TREE_FILTER_E]

WJH_MIRROR_CONG_INVALID = 0xFFFF
WJH_MIRROR_TCLASS_INVALID = 0x1F
WJH_MIRROR_LATENCY_INVALID = 0xFFFFFF
WJH_MIRROR_LATENCY_NSEC_UNIT = 64
WJH_BUFFER_OCCUPANCY_UNIT = 8

WJH_TELE_PATH = '/usr/sbin/netq-wjh-tele'
SDK_PID_LAST = os.path.join(DefaultPaths.RUN_DIR, 'sx_sdk.last')
THRESHOLDS_LAST = DefaultPaths.RUN_DIR
IPS = 'ips'

severity_to_str = {
    WJH_SEVERITY_NOTICE_E: 'Notice',
    WJH_SEVERITY_WARNING_E: 'Warning',
    WJH_SEVERITY_ERROR_E: 'Error'
}

acl_bp_to_str = {
    WJH_ACL_BINDING_POINT_INGRESS: 'Ingress ACL',
    WJH_ACL_BINDING_POINT_EGRESS: 'Egress ACL',
    WJH_ACL_BINDING_POINT_RIF_INGRESS: 'RIF Ingress ACL',
    WJH_ACL_BINDING_POINT_RIF_EGRESS: 'RIF Egress ACL'
}

# Remap record severity by reason ID
# ReasonId -> Expected severity
wjh_severity_remap = {
    505: WJH_SEVERITY_WARNING_E, # Port TC Congestion Threshold Crossed: Notice -> Warning
    506: WJH_SEVERITY_WARNING_E # Packet Latency Threshold Crossed: Notice -> Warning
}


class WjhClient(BaseWjhClient):

    def __init__(self, conf, logger, platform, sdk_utils):
        super(WjhClient, self).__init__(logger)
        self._conf = conf
        self.wjh_lib = None
        self.sdk_pid = 0
        self.mlx_platform = platform
        self.need_tele_init = False
        self.cleaned_up = False
        self.sdk_utils = sdk_utils
        self.port_handler = PortHandler.get_instance(logger, self.sdk_utils)
        self.chip_type = self.sdk_utils.get_chip_type()
        self.ip_to_network = self.get_prefix_network_config()
        self.wjh_polling_interval = self._conf.get('netq-agent', {}).get(WjhConfKw.WJH_POLLING_INTERVAL, WjhConfKw.DEFAULT_WJH_POLLING_INTERVAL)  # 1 minute interval
        self.wjh_4tuple_aggregation_enabled = self._conf.get('netq-agent', {}).get(WjhConfKw.WJH_4TUPLE_AGGREGATION_ENABLED, WjhConfKw.DEFAULT_WJH_4TUPLE_AGGREGATION_ENABLED)

    def create_user_channel(self, channel_type, polling_interval=None):
        user_channel = c_uint32()
        if self.wjh_lib.wjh_user_channel_create(channel_type,
                                                byref(user_channel)) != 0:
            self._logger.error("Failed to create user channel")
        else:
            self._logger.info("wjh - created channel id %d" % (user_channel.value))
        if polling_interval is not None:
            channel_attr = WjhUserChannelAttr()
            channel_attr.polling_interval = self.wjh_polling_interval
            if self.wjh_lib.wjh_user_channel_set(user_channel.value,
                                                 byref(channel_attr)) != 0:
                self._logger.error("failed to set user channel attr")
            else:
                self._logger.info("wjh - set user channel attr id %d interval %d 4-tuple aggregation enabled %s" % (user_channel.value, channel_attr.polling_interval, self.wjh_4tuple_aggregation_enabled))
        else:
            self._logger.info("wjh - set L1 user channel attr id %d" % user_channel.value)

        return user_channel.value


    def get_counter(self, user_channel_id):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_counter_get(user_channel_id,
                    byref(drop_counter), 0) != 0:

            self._logger.error("Failed to get counter")
        else:
            self._logger.info("channel[%d]: drop_counter [%d], recieve_counter [%d]" % (
                              user_channel_id, drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def get_global_counter(self):
        drop_counter = WjhDropCounter()

        if self.wjh_lib.wjh_global_counter_get(byref(drop_counter), 0) != 0:
            self._logger.error("Failed to get global counter")
        else:
            self._logger.error("global: drop_counter [%d], recieve_counter [%d]" % (
                               drop_counter.dropped, drop_counter.received))

        return drop_counter.dropped, drop_counter.received


    def configure_drop_reason_group(self, drop_reason_group, channel_id):
        if self.wjh_lib.wjh_drop_reason_group_bind(drop_reason_group, channel_id) != 0:
            self._logger.error("Failed to bind drop reason group {}".format(drop_reason_group))

        if self.wjh_lib.wjh_drop_reason_group_enable(drop_reason_group, WJH_SEVERITY_ALL_E) != 0:
            self._logger.error("Failed to enable drop reason group {}".format(drop_reason_group))


    def deconfigure_drop_reason_group(self, drop_reason_group):
        if self.wjh_lib.wjh_drop_reason_group_disable(drop_reason_group, WJH_SEVERITY_ALL_E) != 0:
            self._logger.info("Failed to disable drop reason group {}".format(drop_reason_group))

        if self.wjh_lib.wjh_drop_reason_group_unbind(drop_reason_group) != 0:
            self._logger.info("Failed to unbind drop reason group {}".format(drop_reason_group))

        if self.wjh_lib.wjh_drop_reason_group_deinit(drop_reason_group) != 0:
            self._logger.info("Failed to deinit drop reason group {}".format(drop_reason_group))


    def process_threshold_cfg(self, thresh, th_type):
        tele = WJH_TELE_PATH
        th_matrix = []
        old_th_matrix = []

        # get the previous settings
        old_th_file = '{}/{}.last'.format(THRESHOLDS_LAST, th_type)

        if not self.need_tele_init:
            try:
                with open(old_th_file, 'r') as json_file:
                    old_th_matrix = json.load(json_file)
            except Exception as e:
                pass

        for i in range(0, 8):
            th_matrix.append({'enabled': False, 'tc': i, 'ports': [], 'hi': 0, 'lo': 0})

        if thresh:
            #self._logger.info(thresh)
            for ths in thresh:
                tcs = ths.get('tc')
                if tcs is None:
                    continue

                ports = []
                if ths.get('ports') and not is_sonic():
                    for p in ths.get('ports').split(','):
                        lid = self.port_handler.get_port_lid(p)
                        if not lid:
                            continue
                        ports.append(lid)
                else:
                    ports = self.port_handler.get_all_port_lids()

                tcs = str(tcs) if isinstance(tcs, int) else tcs
                for a_tc in tcs.split(','):
                    r = a_tc.split('-')
                    s = int(r[0])
                    e = s
                    if len(r) > 1:
                        e = int(r[1])
                    if s > 7 or e > 7:
                        continue

                    for i in range(s, e + 1):
                        th_matrix[i]['enabled'] = True
                        th_matrix[i]['hi'] = ths.get('hi', 0)
                        th_matrix[i]['lo'] = ths.get('lo', 0)
                        th_matrix[i]['ports'] = ports

                        # the left over ports for the tc in the old setting will be cleaned up
                        if len(old_th_matrix) > i and old_th_matrix[i]['enabled']:
                            for p in ports:
                                if p in th_matrix[i]['ports'] and p in old_th_matrix[i]['ports']:
                                    old_th_matrix[i]['ports'].remove(p)

        for ths in th_matrix:
            for lid in self.port_handler.get_all_port_lids():
                try:
                    sub_cmd = ''
                    if lid not in ths['ports'] or not ths['enabled']:
                        continue

                    cmd = '{} {} --port {} --tc {} --type {}'.format(tele, '--get',
                          lid, ths['tc'], th_type)
                    output = subprocess_wrapper(cmd.split())
                    if output:
                        for l in output.splitlines():
                            if 'sx_api_tele_threshold_get' in l:
                                _, _, rc = l.split()
                                if rc == '0':
                                    sub_cmd = '--update'
                                else:
                                    sub_cmd = ''
                                break

                    cmd = '{} {} --port {} --tc {} --type {} --threshold_hi {} --threshold_lo {}'.format(
                            tele, sub_cmd, lid, ths['tc'], th_type, ths['hi'], ths['lo'])
                    self._logger.debug('wjh telemetry configuration: cmd {} output: {}'.format(cmd, output))
                    subprocess_wrapper(cmd.split())
                except subprocess.CalledProcessError as e:
                    self._logger.error(e)

        # clean up the old settings that are removed
        for ths in old_th_matrix:
            if ths['enabled'] and len(ths['ports']) > 0:
                for p in ths['ports']:
                    try:
                        lid = self.port_handler.get_port_lid(p)
                        if not lid:
                            continue
                        cmd = '{} {} --port {} --tc {} --type {} --threshold_hi 0 --threshold_lo 0'.format(
                              tele, '--clear', lid, ths['tc'], th_type)
                        output = subprocess_wrapper(cmd.split())
                        self._logger.debug('wjh telemetry configuration: cmd {} output: {}'.format(cmd, output))
                    except subprocess.CalledProcessError as e:
                        self._logger.error(e)

        # save the threshold settings
        try:
            with open(old_th_file, 'w') as json_file:
                json.dump(th_matrix, json_file)
        except Exception as e:
            pass


    def init_latency_and_congestion_thresholds(self):
        # NOTE: this should really be done in CL as switchd init
        #       for now, the config is put under agent as follows
        # netq-agent:
        #     log_level: info
        #     wjh: true
        #     wjh_latency_thresholds:
        #         - tc: 0-4,7
        #           ports: swp5,swp6
        #           hi: 1
        #           lo: 0
        #         - tc: 5
        #           hi: 2
        #           lo: 1
        #     wjh_congestion_thresholds:
        #         - tc: 4
        #           ports: swp5,swp6
        #           hi: 1
        #           lo: 0

        tele = WJH_TELE_PATH
        if self.need_tele_init:
            cmd = '{} --init'.format(tele)
            try:
                subprocess_wrapper(cmd.split())
            except subprocess.CalledProcessError as e:
                self._logger.error(e)
            self._logger.info('Initialize sx tele module')

        lat_thresh = self._conf.get('netq-agent').get('wjh_latency_thresholds', False)
        self.process_threshold_cfg(lat_thresh, 'latency')


        cng_thresh = self._conf.get('netq-agent').get('wjh_congestion_thresholds', False)
        self.process_threshold_cfg(cng_thresh, 'congestion')


    def latency_monitor_supported(self):
        rc = False

        if not is_sonic() and self.sdk_utils.is_sdk_version_equals_or_greater("4.3.3206"):
            self._logger.info('Platform supports latency and congestion threshold monitoring')
            rc = True
        return rc


    def buffer_cb_supported(self):
        plat_name = self.mlx_platform.split()[1]
        self._logger.info('PLATFORM: {} chip Type: {}'.format(plat_name, self.chip_type))
        return self.chip_type >= SX_CHIP_TYPE_SPECTRUM2


    def wjh_callback(self, is_agg_cb_supported=True):
        global l2_cb, router_cb, tunnel_cb, acl_cb, l1_cb, buffer_cb

        self._logger.info("WJH callback with is_agg_cb_supported: {}".format(is_agg_cb_supported))
        self.port_handler.refresh_port_map()
        attr = c_uint32()

        l2_cb = WjhDropCallbacks()
        l2_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L2_E

        router_cb = WjhDropCallbacks()
        router_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ROUTER_E

        tunnel_cb = WjhDropCallbacks()
        tunnel_cb.drop_reason_group = WJH_DROP_REASON_GROUP_TUNNEL_E

        acl_cb = WjhDropCallbacks()
        acl_cb.drop_reason_group = WJH_DROP_REASON_GROUP_ACL_E

        if is_agg_cb_supported:
            l2_cb.aggregate_cb.L2 = WjhL2DropAggregateCb(self.wjh_l2_drop_aggregate_cb)
            router_cb.aggregate_cb.router = WjhRouterDropAggregateCb(self.wjh_rtr_drop_aggregate_cb)
            tunnel_cb.aggregate_cb.tunnel = WjhTunnelDropAggregateCb(self.wjh_tun_drop_aggregate_cb)
            acl_cb.aggregate_cb.acl = WjhAclDropAggregateCb(self.wjh_acl_drop_aggregate_cb)
        else:
            attr = WjhDropReasonGroupAttr()
            l2_cb.raw_cb.L2 = WjhL2DropRawCb(self.wjh_l2_drop_raw_cb)
            router_cb.raw_cb.router = WjhRouterDropRawCb(self.wjh_router_drop_raw_cb)
            tunnel_cb.raw_cb.tunnel = WjhTunnelDropRawCb(self.wjh_tunnel_drop_raw_cb)
            acl_cb.raw_cb.acl = WjhAclDropRawCb(self.wjh_acl_drop_raw_cb)

        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L2_E, byref(attr), byref(l2_cb)) != 0:
            self._logger.error("Failed to init L2 drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ROUTER_E, byref(attr), byref(router_cb)) != 0:
            self._logger.error("Failed to init router drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_TUNNEL_E, byref(attr), byref(tunnel_cb)) != 0:
            self._logger.error("Failed to init tunnel drop reason group")

        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_ACL_E, byref(attr), byref(acl_cb)) != 0:
            self._logger.error("Failed to init acl drop reason group")

        if self.buffer_cb_supported():
            buffer_cb = WjhDropCallbacks()
            buffer_cb.drop_reason_group = WJH_DROP_REASON_GROUP_BUFFER_E
            if is_agg_cb_supported:
                buffer_cb.aggregate_cb.buffer = WjhBufferDropAggregateCb(self.wjh_buf_drop_aggregate_cb)
                attr = c_uint32()
            else:
                buffer_cb.raw_cb.buffer = WjhBufferDropRawCb(self.wjh_buffer_drop_raw_cb)
                attr = WjhDropReasonGroupAttr()
                attr.attr.buffer_drop.basic_attr.recirculation_port = 0
                attr.attr.buffer_drop.advanced_attr.trap_probability = 8
            if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_BUFFER_E, byref(attr), byref(buffer_cb)) != 0:
                self._logger.error("Failed to init buffer drop reason group")

            if self.latency_monitor_supported():
                self.init_latency_and_congestion_thresholds()

        # L1
        l1_cb = WjhDropCallbacks()
        l1_cb.drop_reason_group = WJH_DROP_REASON_GROUP_L1_E
        l1_cb.aggregate_cb.L1 = WjhL1DropAggregateCb(self.wjh_l1_drop_aggregate_cb)
        attr = c_uint32()
        if self.wjh_lib.wjh_drop_reason_group_init(WJH_DROP_REASON_GROUP_L1_E, byref(attr), byref(l1_cb)) != 0:
            self._logger.error("Failed to init L1 drop reason group")

        channel_id = self.aggregation_channel_id if is_agg_cb_supported else self.channel_id

        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E, channel_id)
        if self.buffer_cb_supported():
            self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E, channel_id)
        self.configure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E, self.l1_aggregation_channel_id)

        try:
            with open(SDK_PID_LAST, 'w') as f:
                f.write('{}'.format(self.sdk_pid))
        except Exception as e:
            pass
        self.running = True
        self._logger.info("Default wjh_lib_conf.xml path: /usr/etc/wjh_lib_conf.xml")
        # Update ports every 60 seconds
        check_portmap = 0
        check_port_map_interval = 12

        # Flush events every 15 seconds
        check_flush = 0
        flush_interval = 3

        sleep_interval_seconds = 5
        while self.running:
            time.sleep(sleep_interval_seconds)
            check_flush += 1
            if check_flush == flush_interval:
                check_flush = 0
                self.flush_agg_cache()
            check_portmap += 1
            if check_portmap == check_port_map_interval:
                check_portmap = 0
                self.port_handler.refresh_port_map()

    def wjh_deactive_cb(self):
        self._logger.warning("This WJH client is deactivated by another WJH client")
        self.running = False

    # TBD: the agent process should really have a link cache for efficient
    #      access with up to date information.  Reading from switchd is not
    #      workable as the information can chnage at any time
    '''
    def get_lag_map(self):
        cmd = 'cat {}/{}'.format(SWITCHD_SW_TABLE_PATH, self.lag_map_id)
        try:
            output = subprocess_wrapper(cmd.split())
        except Exception:
            self._logger.error('Cannot get lagmap')
            return AGENTSTATUS_FAILURE

        lag = ''
        lid = ''
        self.lagmap = {}
        self.lidmap = {}
        for l in output.splitlines():
            if 'bond-info-entry' in l:
                lag = l.split()[1]
            if 'lag-id' in l:
                lid = l.split()[1]
                self.lagmap[lid] = lag
                self.lidmap[lag] = lid

        return AGENTSTATUS_SUCCESS
    '''


    def start_wjh_client(self):
        try:
            ctypes.CDLL('libsxapi.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxadviser.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libsxaclhelper.so', mode=ctypes.RTLD_GLOBAL)
            ctypes.CDLL('libscew.so', mode=ctypes.RTLD_GLOBAL)
            self.wjh_lib = ctypes.CDLL('libwjh.so', mode=ctypes.RTLD_GLOBAL)
        except OSError:
            try:
                ctypes.CDLL('/usr/lib/x86_64-linux-gnu/libelf.so.1', mode=ctypes.RTLD_GLOBAL)
                self.wjh_lib = ctypes.CDLL('libwjh.so', mode=ctypes.RTLD_GLOBAL)
            except Exception as e:
                self._logger.error("libwjh load error: {}".format(e))
                return AGENTSTATUS_FAILURE

        #wjh_helper = ctypes.CDLL('libwjhhelper.so', mode=ctypes.RTLD_GLOBAL)

        try:
            cmd = '/usr/bin/pidof sx_sdk'
            sdk_pid = subprocess_wrapper(cmd.split())
            self.sdk_pid = int(sdk_pid)

            try:
                with open(SDK_PID_LAST, 'r') as f:
                    old_sdk_pid = int(f.read())

            except Exception:
                old_sdk_pid = 0

            self._logger.info('old sdk pid {}, new sdk pid {}'.format(old_sdk_pid, self.sdk_pid))

            if old_sdk_pid != self.sdk_pid:
                self.need_tele_init = True

        except Exception:
            self._logger.debug("sx_sdk not alive")
            return AGENTSTATUS_FAILURE

        global init_params
        init_param = WjhInitParam()
        init_param.force = 0
        init_param.deactive_cb = WjhDeactiveCb(self.wjh_deactive_cb)
        init_param.bandwidth_percent = 0
        init_param.log_cb = 0
        init_param.debug_fs_path = 0
        init_param.ingress_info_type = WJH_INGRESS_INFO_TYPE_IF_INDEX \
            if is_sonic() else WJH_INGRESS_INFO_TYPE_LOGPORT
        init_param.aggregation_key_mode = WJH_AGGREGATION_KEY_MODE_DETAILED_E
        try:
            init_param.driver_init_param.trap_group_allocate_mode = WJH_TRAP_GROUP_ALLOCATE_MODE_STATIC_E
        except Exception as e:
            pass

        if self.wjh_lib.wjh_init(byref(init_param)) != 0:
            if is_cl3() or is_cl4():
                try:
                    is_active = is_active_systemd_svc('switchd.service')
                    if not is_active:
                        self._logger.debug("switchd not alive")
                        return AGENTSTATUS_FAILURE
                except Exception as e:
                    self._logger.debug("switchd not alive")
                    return AGENTSTATUS_FAILURE

            self._logger.error("wjh_init failed")
            return AGENTSTATUS_FAILURE
        self.cleaned_up = False
        self.channel_id = self.create_user_channel(WJH_USER_CHANNEL_CYCLIC_E, self.wjh_polling_interval)
        self.aggregation_channel_id = self.create_user_channel(WJH_USER_CHANNEL_AGGREGATE_E, self.wjh_polling_interval)
        self.l1_aggregation_channel_id = self.create_user_channel(WJH_USER_CHANNEL_AGGREGATE_E)

        is_agg_cb_supported = False if is_sonic() else True
        self.wjh_callback(is_agg_cb_supported)

        self.clean_up()
        return AGENTSTATUS_SUCCESS


    #
    # Raw callback handlers
    #
    def wjh_raw_drop_fill(self, info):
        drop_reason = WjhDropRawInfo()
        drop_reason.reason_id = info.drop_reason.id
        drop_reason.reason = info.drop_reason.reason.decode('utf-8')
        drop_reason.description = info.drop_reason.description
        drop_reason.severity = severity_to_str.get(info.drop_reason.severity) or ''
        drop_reason.ingress_port = self.port_handler.get_port_name(info.ingress_port)
        drop_reason.timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)

        packet_size = info.packet_size
        packet = list((packet_size * ctypes.c_ubyte).from_address(info.packet))

        drop_reason.packet_size = packet_size
        if (packet_size > 0):
            drop_reason.packet = ''.join('%02x' % i for i in packet)

        return drop_reason

    def __parse_tuple(self, eth, tup):
        smac = eth.get_ether_shost()
        dmac = eth.get_ether_dhost()
        tup['smac'] = '%02x:%02x:%02x:%02x:%02x:%02x' % (smac[0], smac[1], smac[2],
                                                       smac[3], smac[4], smac[5])
        tup['dmac'] = '%02x:%02x:%02x:%02x:%02x:%02x' % (dmac[0], dmac[1], dmac[2],
                                                       dmac[3], dmac[4], dmac[5])
        try:
            iph = eth.child()
            if (eth.get_ether_type() == ImpactPacket.IP.ethertype):
                tup['proto'] = iph.get_ip_p()
            elif(eth.get_ether_type() == IP6.IP6.ethertype):
                tup['proto'] = iph.get_next_header()
            tup['src'] = iph.get_ip_src()
            tup['dst'] = iph.get_ip_dst()

            transport = iph.child()
            if (tup['proto'] == TCP.protocol):
                tup['sport'] = transport.get_th_sport()
                tup['dport'] = transport.get_th_dport()
            elif (tup['proto'] == UDP.protocol):
                tup['sport'] = transport.get_uh_sport()
                tup['dport'] = transport.get_uh_dport()
        except Exception as e:
            self._logger.debug('{}'.format(e))

        return tup

    def parse_7_tuples(self, packet, packet_size):
        tup = {'smac': '00:00:00:00:00:00',
               'dmac': '00:00:00:00:00:00',
               'src': '0.0.0.0',
               'dst': '0.0.0.0',
               'sport': 0,
               'dport': 0,
               'proto': 0}

        inner_tup = None

        if packet_size == 0:
            return tup

        pkt = list((packet_size * ctypes.c_ubyte).from_address(packet))
        try:
            eth = EthDecoder().decode(binascii.a2b_hex(''.join('%02x' % i for i in pkt)))
            tup = self.__parse_tuple(eth, tup)
        except Exception as e:
            self._logger.error('Exception in parse_7_tuples {}'.format(e))
            return tup, inner_tup

        tup = self.__parse_tuple(eth, tup)
        # Commenting the below block of code, to remove scapy dependency
        # VxLAN is mainly required for flow telemetry usecase
#        if tup and tup['proto'] == 17 and tup['dport'] == 4789:
#            try:
#                vxlan = eth[VXLAN]
#                ieth = vxlan[Ether]
#                inner_tup = {'smac': '00:00:00:00:00:00',
#                             'dmac': '00:00:00:00:00:00',
#                             'src': '0.0.0.0',
#                             'dst': '0.0.0.0',
#                             'sport': 0,
#                             'dport': 0,
#                             'proto': 0}
#
#                inner_tup = self.__parse_tuple(ieth, inner_tup)
#            except Exception as e:
#                self._logger.debug('{}'.format(e))

        self._logger.debug('smac: {}, dmac: {} -- src: {}, dst: {}, proto {}, sport {}, dport {}'.format(
                           tup['smac'], tup['dmac'], tup['src'], tup['dst'], tup['proto'], tup['sport'], tup['dport']))
        return tup, inner_tup


    def __fill_event_tuples(self, event, tuples, info):
        event.sip = tuples['src']
        event.dip = tuples['dst']
        event.proto = tuples['proto']
        event.sport = tuples.get('sport', 0)
        event.dport = tuples.get('dport', 0)
        event.smac = tuples['smac']
        event.dmac = tuples['dmac']
        key = (event.reason, event.ingress_port, event.sip, event.dip,
               event.proto, event.sport, event.dport, event.smac,
               event.dmac)
        self._logger.debug('ADD key: {}'.format(key))

        agg = self.agg_cache.get(key)
        if agg:
            agg.agg_count += 1
            agg.last_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
        else:
            event.agg_count = 1
            event.first_timestamp = float(info.timestamp.tv_sec) + float(info.timestamp.tv_nsec / 1000000000.0)
            event.last_timestamp = event.first_timestamp
            if len(self.agg_cache) < 1500:
                self.agg_cache[key] = event

        return event

    def wjh_raw_drop_fill_aggregate(self, drop_type, info, egress_port='',
                                    traffic_class=0, acl_rule_id=0,
                                    original_occupancy=0, original_latency=0):
        drop_reason = WjhDropAggregate()
        if drop_type == 'ACL':
            if info.binding_point == WJH_ACL_BINDING_POINT_INGRESS and info.acl_name.decode('utf-8') == 'monitor' and info.rule.decode('utf-8') == '-':
                r = 'Monitor'
            else:
                r = acl_bp_to_str.get(info.binding_point) or ''
            drop_reason.reason = '{}'.format(r)
            drop_reason.acl_bind_point = info.binding_point
            drop_reason.acl_name = info.acl_name.decode('utf-8') if info.acl_name else ''
            drop_reason.acl_rule = info.rule.decode('utf-8') if info.rule else ''
        else:
            drop_reason.reason = info.drop_reason.reason.decode('utf-8')
        drop_reason.reason_id = info.drop_reason.id
        drop_reason.drop_type = drop_type
        drop_reason.severity = severity_to_str.get(info.drop_reason.severity) or ''
        drop_reason.egress_port = self.port_handler.get_port_name(egress_port) if egress_port != '' else ''
        drop_reason.acl_rule_id = acl_rule_id
        if traffic_class != WJH_MIRROR_TCLASS_INVALID:
            drop_reason.traffic_class = traffic_class
        if original_occupancy != WJH_MIRROR_CONG_INVALID:
            drop_reason.original_occupancy = max(original_occupancy, drop_reason.original_occupancy)
        if original_latency != WJH_MIRROR_LATENCY_INVALID:
            drop_reason.original_latency = max(original_latency, drop_reason.original_latency) * WJH_MIRROR_LATENCY_NSEC_UNIT
        drop_reason.ingress_port = self.port_handler.get_port_name(info.ingress_port)


        p_tuples, p_inner_tuples = self.parse_7_tuples(info.packet, info.packet_size)
        self.__fill_event_tuples(drop_reason, p_tuples, info)
        if p_inner_tuples:
            inner_event = WjhDropAggregate()
            inner_event.drop_type = drop_reason.drop_type
            inner_event.reason = drop_reason.reason
            inner_event.acl_bind_point = drop_reason.acl_bind_point
            if drop_reason.reason == 'Monitor':
                inner_event.acl_name = 'tunneled'
            inner_event.acl_rule_id = drop_reason.acl_rule_id
            inner_event.severity = drop_reason.severity
            inner_event.egress_port = drop_reason.egress_port
            inner_event.traffic_class = drop_reason.traffic_class
            inner_event.original_occupancy = drop_reason.original_occupancy
            inner_event.original_latency = drop_reason.original_latency
            inner_event.ingress_port = drop_reason.ingress_port
            self.__fill_event_tuples(inner_event, p_inner_tuples, info)

        return drop_reason




    def wjh_l2_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if raw_info_list_p[i].drop_reason.id in default_filters:
                continue
            if self.is_wjh_drop_filtered_from_data(raw_info_list_p[i], WjhConfKw.L2):
                continue
            self._logger.debug("l2 drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str.get(raw_info_list_p[i].drop_reason.severity) or '',
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('L2', raw_info_list_p[i])

        return 0


    def wjh_router_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if raw_info_list_p[i].drop_reason.id in default_filters:
                continue
            if self.is_wjh_drop_filtered_from_data(raw_info_list_p[i], WjhConfKw.ROUTER):
                continue
            self._logger.debug("router drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str.get(raw_info_list_p[i].drop_reason.severity) or '',
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Router', raw_info_list_p[i])

        return 0


    def wjh_tunnel_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if self.is_wjh_drop_filtered_from_data(raw_info_list_p[i], WjhConfKw.TUNNEL):
                continue
            self._logger.debug("tunnel drop reason group: ingress port [0x%x], size[%d], reason_id[%u], reason[%s], description[%s], severity[%s], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].packet_size,
                      raw_info_list_p[i].drop_reason.id,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].drop_reason.description,
                      severity_to_str.get(raw_info_list_p[i].drop_reason.severity) or '',
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('Tunnel', raw_info_list_p[i])

        return 0


    def wjh_buffer_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if self.is_wjh_drop_filtered_from_data(raw_info_list_p[i], WjhConfKw.BUFFER):
                continue
            self._logger.debug("buffer drop reason group: ingress port [0x%x], reason[%s], egress port [%d: 0x%x], tc [%d], occupancy[%d], latency[%d], timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                      raw_info_list_p[i].drop_reason.reason,
                      raw_info_list_p[i].egress_port_valid,
                      raw_info_list_p[i].egress_port,
                      raw_info_list_p[i].tc,
                      raw_info_list_p[i].original_occupancy,
                      raw_info_list_p[i].original_latency,
                      raw_info_list_p[i].timestamp.tv_sec,
                      raw_info_list_p[i].timestamp.tv_nsec))

            if raw_info_list_p[i].egress_port_valid:
                egress_port = raw_info_list_p[i].egress_port
            else:
                egress_port = ''
            self.wjh_raw_drop_fill_aggregate('Buffer', raw_info_list_p[i],
                                   egress_port=egress_port,
                                   traffic_class=raw_info_list_p[i].tc,
                                   original_occupancy=raw_info_list_p[i].original_occupancy * WJH_BUFFER_OCCUPANCY_UNIT,
                                   original_latency=raw_info_list_p[i].original_latency * WJH_MIRROR_LATENCY_NSEC_UNIT)

        return 0


    def wjh_acl_drop_raw_cb(self, raw_info_list_p, raw_info_list_size_p):
        list_size = raw_info_list_size_p.contents.value
        for i in range(list_size):
            if self.is_wjh_drop_filtered_from_data(raw_info_list_p[i], WjhConfKw.ACL):
                continue
            self._logger.debug("acl drop reason group: ingress port [0x%x], size[%d], rule_id[0x%x], binding_point[0x%x], acl_name[%s], rule{%s}, timestamp[%d.%d]."
                  % (raw_info_list_p[i].ingress_port,
                     raw_info_list_p[i].packet_size,
                     raw_info_list_p[i].rule_id,
                     raw_info_list_p[i].binding_point,
                     raw_info_list_p[i].acl_name,
                     raw_info_list_p[i].rule,
                     raw_info_list_p[i].timestamp.tv_sec,
                     raw_info_list_p[i].timestamp.tv_nsec))

            self.wjh_raw_drop_fill_aggregate('ACL', raw_info_list_p[i], acl_rule_id=raw_info_list_p[i].rule_id)
        return 0


    #
    # Aggregate callback handlers
    #
    def wjh_agg_drop_fill(self, key, data):
        drop = WjhDropAggregate()
        drop.ingress_port = self.port_handler.get_port_name(key.ingress_port)
        if key.is_lag_member:
            drop.ingress_lag = key.ingress_lag
        else:
            drop.ingress_lag = ''

        drop.sip = key.five_tuples.sip.decode('utf-8') if key.five_tuples.sip else '0.0.0.0'
        drop.dip = key.five_tuples.dip.decode('utf-8') if key.five_tuples.dip else '0.0.0.0'
        drop.proto = key.five_tuples.proto if key.five_tuples.proto else 0
        drop.sport = key.five_tuples.sport if key.five_tuples.sport else 0
        drop.dport = key.five_tuples.dport if key.five_tuples.dport else 0

        drop.smac = '%02x:%02x:%02x:%02x:%02x:%02x' % (key.smac[0], key.smac[1], key.smac[2],
                                                       key.smac[3], key.smac[4], key.smac[5])
        drop.dmac = '%02x:%02x:%02x:%02x:%02x:%02x' % (key.dmac[0], key.dmac[1], key.dmac[2],
                                                       key.dmac[3], key.dmac[4], key.dmac[5])

        drop.vlan = 0
        if 'vlan' in dict(key._fields_) and key.vlan:
            drop.vlan = key.vlan

        #drop.ether_type = key.ether_type
        drop.agg_count = data.count
        drop.first_timestamp = float(data.timestamp.first_timestamp.tv_sec) + float(
                                     data.timestamp.first_timestamp.tv_nsec / 1000000000.0)
        drop.last_timestamp = float(data.timestamp.last_timestamp.tv_sec) + float(
                                           data.timestamp.last_timestamp.tv_nsec / 1000000000.0)
        return drop


    def wjh_acl_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p=None):
        # TBD: bind point is not available
        list_size = cnt_p.contents.value
        self._logger.debug('wjh ACL drop count {}'.format(list_size))
        drop_reason_list = []
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]
            if self.is_wjh_drop_filtered_from_key(key, WjhConfKw.ACL):
                continue

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'ACL'

            if key.acl_name.decode('utf-8') == 'monitor' and key.rule.decode('utf-8') == '-':
                r = 'Monitor'
            else:
                r = key.reason.reason.decode('utf-8')
                #r = acl_bp_to_str.get(data.binding_point) or ''

            drop_reason.reason = '{}'.format(r)
            drop_reason.acl_rule_id = key.rule_id
            drop_reason.acl_name = key.acl_name.decode('utf-8') if key.acl_name else ''
            drop_reason.acl_rule = key.rule.decode('utf-8') if key.rule else ''
            drop_reason.reason_id = key.reason.id if key.reason.id else 0
            drop_reason.severity = severity_to_str.get(key.reason.severity) or ''
            self._logger.debug("SDK computed ACL drop_reason: {}".format(drop_reason))
            # drop_reason.acl_bind_point = acl_bp_to_str.get(info.binding_point) or '' # not available with agg cb
            drop_reason_list.append(drop_reason)
        return self.wjh_aggregate_4tuple(drop_reason_list)

    def wjh_buf_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p=None):
        list_size = cnt_p.contents.value
        self._logger.debug('wjh Buffer drop count {}'.format(list_size))
        drop_reason_list = []
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]
            if self.is_wjh_drop_filtered_from_key(key, WjhConfKw.BUFFER):
                continue

            egress_port = self.port_handler.get_port_name(data.egress_data.egress_port) if data.egress_data_valid else ''
            traffic_class = data.egress_data.tc

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'Buffer'
            drop_reason.reason = key.reason.reason.decode('utf-8')
            drop_reason.reason_id = key.reason.id if key.reason.id else 0
            if drop_reason.reason_id in wjh_severity_remap:
                severity = wjh_severity_remap[drop_reason.reason_id]
            else:
                severity = key.reason.severity
            drop_reason.severity = severity_to_str.get(severity) or ''
            drop_reason.egress_port = egress_port

            # SDK aggregated Buffer Drops do not have original_occupancy and original_latency, hence
            # setting them as -1 to show that these params are N/A
            drop_reason.original_occupancy = -1
            drop_reason.original_latency = -1

            if traffic_class is not None and traffic_class != WJH_MIRROR_TCLASS_INVALID:
                drop_reason.traffic_class = traffic_class

            self._logger.debug("SDK computed Buffer drop_reason: {}".format(drop_reason))
            drop_reason_list.append(drop_reason)

        return self.wjh_aggregate_4tuple(drop_reason_list)

    def wjh_l1_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        self._logger.debug('wjh L1 drop count {}'.format(list_size))
        for i in range(list_size):
            ingress_port = self.port_handler.get_port_name(key_list_p[i].ingress_port)
            port_down_reason = data_list_p[i].port_down_reason
            if port_down_reason is not None:
                port_down_reason = port_down_reason.decode()
            description = data_list_p[i].description
            if description is not None:
                description = description.decode()
            '''
            self._logger.info("port %s: is_port_up: %u, port_down_reason: %s, description: %s, state_change_count: %u" %
                  (ingress_port, data_list_p[i].is_port_up, port_down_reason,
                   description, data_list_p[i].state_change_count))
            self._logger.info("           symbol_error_count: %u, crc_error_count: %u" %
                  (data_list_p[i].symbol_error_count, data_list_p[i].crc_error_count))
            self._logger.info("           timestamp[%ld.%ld, %ld.%ld]" % (data_list_p[i].timestamp.first_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.first_timestamp.tv_nsec, data_list_p[i].timestamp.last_timestamp.tv_sec,
                                                              data_list_p[i].timestamp.last_timestamp.tv_nsec))
            '''
            # L1 drop callback data is quite different from the others
            filter_drop_reasons, filter_drop_severity, filter_ips = self.get_wjh_filter_config(WjhConfKw.L1)
            if port_down_reason in filter_drop_reasons:
                self._logger.debug('L1 drop with reason {} filtered out'
                                   .format(port_down_reason))
                continue

            l1drop = WjhL1DropAggregate()
            l1drop.ingress_port = ingress_port
            l1drop.is_port_up = data_list_p[i].is_port_up == 1
            l1drop.port_down_reason = port_down_reason
            l1drop.description = description
            l1drop.state_change_count = data_list_p[i].state_change_count
            l1drop.symbol_error_count = data_list_p[i].symbol_error_count
            l1drop.crc_error_count = data_list_p[i].crc_error_count
            l1drop.first_timestamp = float(data_list_p[i].timestamp.first_timestamp.tv_sec) + float(
                                           data_list_p[i].timestamp.first_timestamp.tv_nsec / 1000000000.0)
            l1drop.last_timestamp = float(data_list_p[i].timestamp.last_timestamp.tv_sec) + float(
                                          data_list_p[i].timestamp.last_timestamp.tv_nsec / 1000000000.0)
            self._logger.debug("SDK computed wjh L1: {}".format(l1drop))
            l1drop.save()

        return 0


    def wjh_rtr_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        self._logger.debug('wjh Router drop count {}'.format(list_size))
        drop_reason_list = []
        for i in range(list_size):
            key = key_list_p[i]
            if self.is_wjh_drop_filtered_from_key(key, WjhConfKw.ROUTER):
                continue
            data = data_list_p[i]
            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'Router'
            drop_reason.reason_id = key.reason.id if key.reason.id else 0
            drop_reason.severity = severity_to_str.get(key.reason.severity) or ''
            # drop_reason.ingress_port = self.port_handler.get_port_name(data.ingress_port)
            drop_reason.reason = key.reason.reason.decode('utf-8')
            self._logger.debug("SDK computed Router drop_reason: {}".format(drop_reason))
            drop_reason_list.append(drop_reason)
        return self.wjh_aggregate_4tuple(drop_reason_list)


    def wjh_l2_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        self._logger.debug('wjh L2 drop count {}'.format(list_size))
        drop_reason_list = []
        for i in range(list_size):
            key = key_list_p[i]
            if self.is_wjh_drop_filtered_from_key(key, WjhConfKw.L2):
                continue
            info = data_list_p[i]
            drop_reason = self.wjh_agg_drop_fill(key, info)
            drop_reason.reason_id = key.reason.id if key.reason.id else 0
            drop_reason.reason = key.reason.reason.decode('utf-8')
            drop_reason.severity = severity_to_str.get(key.reason.severity) or ''
            drop_reason.drop_type = 'L2'
            self._logger.debug("SDK computed L2 drop_reason: {}".format(drop_reason))
            drop_reason_list.append(drop_reason)
        return self.wjh_aggregate_4tuple(drop_reason_list)


    def wjh_tun_drop_aggregate_cb(self, key_list_p, data_list_p, cnt_p, attr_p):
        list_size = cnt_p.contents.value
        self._logger.debug('wjh Tunnel drop count {}'.format(list_size))
        drop_reason_list = []
        for i in range(list_size):
            key = key_list_p[i]
            data = data_list_p[i]
            if self.is_wjh_drop_filtered_from_key(key, WjhConfKw.TUNNEL):
                continue

            drop_reason = self.wjh_agg_drop_fill(key, data)
            drop_reason.drop_type = 'Tunnel'
            drop_reason.reason_id = key.reason.id if key.reason.id else 0
            drop_reason.reason = key.reason.reason.decode('utf-8')
            drop_reason.severity = severity_to_str.get(key.reason.severity) or ''
            self._logger.debug("SDK computed TUNNEL drop_reason: {}".format(drop_reason))
            drop_reason_list.append(drop_reason)
        return self.wjh_aggregate_4tuple(drop_reason_list)

    # Handling the following use cases for 4/5 tuple aggregation
    # 1. When source port is static and destination port is static -  Emit record without any changes.
    # 2. When destination port is static and source port is dynamic - aggregating as 4 tuple with source port = 0.
    # 3. When source port is static and destination port is dynamic - aggregating as 4 tuple with destination port = 0.

    # Records aggregation example in 1-minute window
    # reason          src-ip     dst-ip   protocol   proto   ingress-port  src-port    dst-port  vlan   agg-count
    # Tail drop       1.1.1.1    2.2.2.2  TCP        17      swp22        2020,       3030,      1     1500
    # Tail drop       1.1.1.1    2.2.2.2  TCP        17      swp22        2020,       3030,      1     1500
    # Tail drop       3.3.3.3    4.4.4.4  TCP        17      swp22        2020,       3031,      1     3000
    # Tail drop       3.3.3.3    4.4.4.4  TCP        17      swp22        2020,       3032,      1     3000
    # Tail drop       3.3.3.3    4.4.4.4  TCP        17      swp22        2020,       3034,      1     3000
    # Tail drop       3.3.3.3    4.4.4.4  TCP        17      swp22        2020,       3035,      1     3000

    # Records aggregation to emit for the 1-minute window
    # reason          src-ip     dst-ip   protocol   proto   ingress-port  src-port    dst-port  vlan   agg-count
    # Tail drop       1.1.1.1    2.2.2.2  TCP        17      swp22        2020,       3030,      1     3000
    # Tail drop       3.3.3.3    4.4.4.4  TCP        17      swp22        2020,       0,         1     12000
    def wjh_aggregate_4tuple(self, drop_reason_list):
        if not self.wjh_4tuple_aggregation_enabled:
            send_list_data(drop_reason_list, WjhDropAggregate().__class__.__name__)
            return 0
        aggregated_drop_list = {}
        sport_count_map = {}
        dport_count_map = {}
        for drop_reason in drop_reason_list:
            sport_key = (drop_reason.ingress_port, drop_reason.reason_id,
                         drop_reason.sip, drop_reason.dip, drop_reason.proto,
                         drop_reason.vlan, drop_reason.sport)
            sport_count_map[sport_key] = sport_count_map.setdefault(sport_key, 0) + 1

            dport_key = (drop_reason.ingress_port, drop_reason.reason_id,
                         drop_reason.sip, drop_reason.dip, drop_reason.proto,
                         drop_reason.vlan, drop_reason.dport)
            dport_count_map[dport_key] = dport_count_map.setdefault(dport_key, 0) + 1
        self._logger.debug("wjh_aggregate_4tuple: dport_count_map: {} sport_count_map {}".format(dport_count_map, sport_count_map))
        for drop_reason in drop_reason_list:
            sport_key = (drop_reason.ingress_port, drop_reason.reason_id,
                            drop_reason.sip, drop_reason.dip, drop_reason.proto,
                            drop_reason.vlan, drop_reason.sport)
            dport_key = (drop_reason.ingress_port, drop_reason.reason_id,
                            drop_reason.sip, drop_reason.dip, drop_reason.proto,
                            drop_reason.vlan, drop_reason.dport)
            sport_count = sport_count_map.get(sport_key, 0)
            dport_count = dport_count_map.get(dport_key, 0)
            if dport_count > 1 and dport_count > sport_count:
                detailed_key = (drop_reason.ingress_port, drop_reason.reason_id,
                                drop_reason.sip, drop_reason.dip,
                                drop_reason.proto,
                                drop_reason.vlan, 0, drop_reason.dport)
                drop_reason.sport = 0
            elif sport_count > 1:
                detailed_key = (drop_reason.ingress_port, drop_reason.reason_id,
                                drop_reason.sip, drop_reason.dip,
                                drop_reason.proto,
                                drop_reason.vlan, drop_reason.sport, 0)
                drop_reason.dport = 0
            else:
                detailed_key = (drop_reason.ingress_port, drop_reason.reason_id,
                                drop_reason.sip, drop_reason.dip,
                                drop_reason.proto,
                                drop_reason.vlan, drop_reason.sport, drop_reason.dport)

            if detailed_key in aggregated_drop_list:
                aggregated_drop = aggregated_drop_list[detailed_key]
                aggregated_drop.agg_count += drop_reason.agg_count

                if aggregated_drop.first_timestamp > drop_reason.first_timestamp:
                    aggregated_drop.first_timestamp = drop_reason.first_timestamp
                if aggregated_drop.last_timestamp < drop_reason.last_timestamp:
                    aggregated_drop.last_timestamp = drop_reason.last_timestamp
            else:
                aggregated_drop = drop_reason
            aggregated_drop_list[detailed_key] = aggregated_drop

        send_list_data(aggregated_drop_list.values(), WjhDropAggregate().__class__.__name__)
        return 0

    def is_wjh_drop_filtered_from_data(self, raw_info, drop_type):
        drop_reason = raw_info.drop_reason.reason.decode('utf-8')
        severity = severity_to_str.get(raw_info.drop_reason.severity)
        drop_reason_id = raw_info.drop_reason.id
        p_tuples, p_inner_tuples = self.parse_7_tuples(raw_info.packet, raw_info.packet_size)
        sip = p_tuples['src'] if p_tuples['src'] else None
        dip = p_tuples['dst'] if p_tuples['dst'] else None
        return self._is_wjh_drop_filtered(drop_reason_id, drop_reason, severity, drop_type, sip, dip)

    def is_wjh_drop_filtered_from_key(self, key, drop_type):
        drop_reason = key.reason.reason.decode('utf-8')
        drop_reason_id = key.reason.id
        severity = severity_to_str.get(key.reason.severity) or ''
        sip = key.five_tuples.sip.decode('utf-8') if key.five_tuples.sip else None
        dip = key.five_tuples.dip.decode('utf-8') if key.five_tuples.dip else None
        return self._is_wjh_drop_filtered(drop_reason_id, drop_reason, severity, drop_type, sip, dip)

    def _is_wjh_drop_filtered(self, drop_reason_id, drop_reason, severity, drop_type, sip, dip):
        filter_drop_reasons, filter_drop_severity, filter_ips = self.get_wjh_filter_config(drop_type)
        if drop_reason in filter_drop_reasons:
            self._logger.debug('{} drop with reason {} filtered'.format(drop_type, drop_reason))
            return True
        if severity in filter_drop_severity:
            self._logger.debug('{} drop with severity {} filtered'.format(drop_type, severity))
            return True
        if drop_reason_id in default_filters:
            self._logger.debug('{} drop with reason {} filtered'.format(drop_type, drop_reason))
            return True
        if self.ip_in_filter_ips(sip, filter_ips):
            self._logger.debug('{} drop with ip {} filtered'.format(drop_type, sip))
            return True
        if self.ip_in_filter_ips(dip, filter_ips):
            self._logger.debug('{} drop with ip {} filtered'.format(drop_type, dip))
            return True
        return False

    def ip_in_filter_ips(self, ip, filter_ips):
        if ip is None:
            return False
        for filter_ip in filter_ips:
            # support prefix network, prefix network example: 192.168.0.0/24
            if '/' in filter_ip and self.check_ip_in_network(ip, filter_ip):
                return True
            elif ip == filter_ip:
                return True
        return False

    def check_ip_in_network(self, ip, network):
        ip_address = ipaddr.ip_address(ip)
        network_prefix = self.ip_to_network[network] if network in self.ip_to_network else None
        return ip_address in network_prefix

    def get_wjh_filter_config(self, filter_type):
        wjh_filter_list = self._conf.get('netq-agent').get('wjh_filter', [])
        filter_drop_reasons = []
        filter_drop_severity = []
        filter_ips = []

        for wjh_filter in wjh_filter_list:
            if IPS in wjh_filter.keys():
                filter_ips = wjh_filter.get(IPS, [])

        for wjh_filter in wjh_filter_list:
            if filter_type in wjh_filter.keys():
                filter_type_config = wjh_filter.get(filter_type, {})
                drop_reason_dict = WjhConfKw.DROP_REASONS_MAPPING.get(filter_type)
                configured_drop_reasons = filter_type_config.get('drop-reasons', [])
                filter_drop_reasons = [drop_reason_dict.get(reason) for reason in configured_drop_reasons]
                filter_drop_severity = filter_type_config.get('severity', [])
                break

        return filter_drop_reasons, filter_drop_severity, filter_ips

    def get_prefix_network_config(self):
        wjh_filter_list = self._conf.get('netq-agent').get('wjh_filter', [])
        ip_to_network = {}
        for wjh_filter in wjh_filter_list:
            if IPS in wjh_filter.keys():
                filter_ips = wjh_filter.get(IPS, [])
                for filter_ip in filter_ips:
                    if '/' in filter_ip:
                        ip_to_network[filter_ip] = ipaddr.ip_network(filter_ip, strict=False)
                break
        return ip_to_network

    def _run():
        pass

    def clean_up(self):
        if self.cleaned_up:
            return
        if is_cl3() or is_cl4():
            try:
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    self._logger.debug("sx_sdk not alive")
                    self._logger.info('SX_SDK Inactive: WJH DONE, EXITING')
                    self._logger.info('SX_SDK Inactive: WJH cleaned up')
                    self.cleaned_up = True
                    self._logger.info("wjh lib is deinitialized")
                    return
            except Exception as e:
                self._logger.debug("Unable to check systemctl status for sx_sdk.service")

        self._logger.info('WJH DONE, EXITING')
        self._logger.info('WJH clean up')
        if self.buffer_cb_supported():
            self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_BUFFER_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_ACL_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_L1_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_L2_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_ROUTER_E)
        self.deconfigure_drop_reason_group(WJH_DROP_REASON_GROUP_TUNNEL_E)
        if self.wjh_lib.wjh_user_channel_destroy(self.channel_id) != 0:
            self._logger.info("Failed to destroy user channel")

        if self.wjh_lib.wjh_user_channel_destroy(self.aggregation_channel_id) != 0:
            self._logger.info("Failed to destroy agg user channel")

        if self.wjh_lib.wjh_user_channel_destroy(self.l1_aggregation_channel_id) != 0:
            self._logger.info("Failed to destroy l1 agg user channel")

        if self.wjh_lib.wjh_deinit() != 0:
            self._logger.info("Failed to deinit wjh service lib")

        self._logger.info("wjh lib is deinitialized")
        self.cleaned_up = True



