#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



import eventlet
import eventlet.pools
from collections import namedtuple
from pyVim.connect import SmartConnect, Disconnect

EsxHostKey = namedtuple('EsxHostKey', ['vcenter', 'datacenter', 'host', 'pnic', 'dvs', 'pg', 'encap_type', 'encap'])
EsxHostValue = namedtuple('EsxHostValue', ['ip', 'mac', 'mtu', 'pnic_admin_state', 'pnic_oper_state',
                                           'peer_hostname', 'peer_ifname'])

class Vcenter(object):
    """Vcenter class maintains per vcenter configuration cache
    It interfaces with the vcenter and collects ESX host and VM specific information
    """
    #Default Vcenter Polling Period
    _DEFAULT_VCENTER_POLL_PERIOD = 15

    def __init__(self, config, logger):
        self._ip = config.get('vcenter_ip', '')
        self._port = config.get('vcenter_port', 443)
        self._user = config.get('vcenter_user', '')
        self._password = config.get('vcenter_password', '')
        self._poll_period = config.get('vcenter_poll_period',
                                       Vcenter._DEFAULT_VCENTER_POLL_PERIOD)
        self.datacenter = config.get('datacenter', [])
        self.exclude_datacenter = config.get('exclude_datacenter', [])
        self.logger = logger
        return

    @staticmethod
    def valid_config(config):
        valid_keys = set(['vcenter_ip', 'vcenter_port', 'vcenter_user', 'vcenter_password',
                          'vcenter_poll_period', 'datacenter', 'exclude_datacenter'])
        '''
        Validate user hasn't specified any parameter that
        netq agent does not understand
        '''
        for key in list(config.keys()):
            if key not in valid_keys:
                return False
        return True

    @property
    def ip(self):
        return self._ip

    @property
    def port(self):
        return self._port

    @property
    def user(self):
        return self._user

    @property
    def password(self):
        return self._password

    @property
    def poll_period(self):
        return self._poll_period

    def run(self):
        """Monitor Vcenter using IP, User/Password provided by the user
        """
        if self.ip == '':
            self._logger.error('Missing IP address configuration')
            return

        if self.user == '':
            self._logger.error('Missing user name onfiguration')
            return

        if self.password == '':
            self._logger.error('Missing password')
            return

        # Establish a persistent connection to Vcenter
        self._connection = SmartConnect(host=self.ip, user=self.user, pwd=self.password, port=self.port)

        if not self._connection:
            self._logger.error('Failed to connect to vcenter:%r:%r with username:%r' % self.ip, self.port, self.user)
            return

        #Log session id for debugging 
        session_id = self._connection.content.sessionManager.currentSession.key
        self._logger.info('Vcenter %r: Established session id=%r' % self.ip, session_id)

        while True:
            # periodically retrieve data from vcenter and update database
            self.getVcenterInfo()
            #greenthread.sleep(self.poll_period)
        return

    def stop_checker(self):
        """ stop_checker() should be called when vcenter monitor thread is stopped/killed
        This function Disconnect from Vcenter
        """
        if self._connection:
            Disconnect(self._connection)
            self._logger.info('Disconnected from Vcenter %r' % self.ip)
            self._connection = None
        return

    def getVcenterInfo(self):
        content = self.RetriveContent()
        datacenters = content.rootFolder.childEntity
        esxhost = EsxHostKey()
        esxhost.vcenter = self.ip
        for dc in datacenters:  # Iterate though DataCenters
            esxhost.datacenter = dc.name
            clusters = dc.hostFolder.childEntity
            for cluster in clusters:  # Iterate through the clusters in the DC
                # Add Cluster name
                esxhost.cluster = cluster.name
                hosts = cluster.host
                for host in hosts:  # Iterate through Hosts in the Cluster
                    esxhost.esxhost = host.summary.config.name
                    '''
                    # Add VMs to data dict by config name
                    vms = host.vm
                    for vm in vms:  # Iterate through each VM on the host
                        vmname = vm.summary.config.name
                    '''

        return


    @staticmethod
    def getNICs(summary, guest):
        nics = {}
        for nic in guest.net:
            if nic.network:  # Only return adapter backed interfaces
                if nic.ipConfig is not None and nic.ipConfig.ipAddress is not None:
                    nics[nic.macAddress] = {}  # Use mac as uniq ID for nic
                    nics[nic.macAddress]['netlabel'] = nic.network
                    ipconf = nic.ipConfig.ipAddress
                    for ip in ipconf:
                        if ":" not in ip.ipAddress:  # Only grab ipv4 addresses
                            nics[nic.macAddress]['ip'] = ip.ipAddress
                            nics[nic.macAddress]['prefix'] = ip.prefixLength
                            nics[nic.macAddress]['connected'] = nic.connected
        return nics

class VmwareMonitor(object):
    """Class VmwareHypervisor
    Interfaces with VCenter. It provides method to launch a thread per vcenter
    and update database with VMware VM/ESX information

    In future this class can be extended to interface with vRealize or vCloudSuite etc
    """
    def __init__(self, yaml_config, logger=None, pool=None):
        # Parse and initialize VMware Specific Configuration object
        self.__logger = logger
        self.__pool = pool or eventlet.GreenPool()
        self.__vcenters = []

        if yaml_config and 'vmware' in yaml_config:
            # Loop through VMware configuration list
            # The list should contain dictionary of VCenter configuration 
            for config in yaml_config['vmware']:
                if Vcenter.valid_config(config):
                    vcenter = Vcenter(config, logger)
                    self.__vcenters.append(vcenter)
        return

    @property
    def num_vcenters(self):
        return len(self.__vcenters)

    def halt(self):
        for i, gt in enumerate(self._threads):
            self._logger.info('Terminating VCenter monitor greenthread: %d' % i)
            eventlet.kill(gt, eventlet.StopServe)
        self._threads = []
        return

    def run(self, pool, epoch, pipeline=None):
        """Check if vcenter monitor threads aren't running
        Start a thread for each vcenter
        """
        if self._threads:
            self._logger.error('%d Vcenter monitoring Threads are already running' % len(self._threads))
            return

        for i, vcenter in enumerate(self.__vcenter):
            self._logger.info('Launching Vcenter %d' % i)
            gt = self.__pool.spawn(vcenter.run())
            gt.link(vcenter.stop_checker())
            self._threads.append(gt)
