#
# Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

from pathlib import PosixPath

import os
import ujson as json
from netq_lib.common import utils
from netq_lib.common.enums import PlatformType
from netq_lib.common.utils import subprocess_wrapper, get_cl_switchd_sw_tables_idx, SWITCHD_SW_TABLE_PATH, \
     is_active_systemd_svc

class PortHandler(object):
    __instance = None

    def __init__(self, logger, sdk_utils=None):
        if PortHandler.__instance != None:
            raise Exception("port handler class is already instanced")
        else:
            PortHandler.__instance = self
        self.platform = utils.get_platform_type()
        self._logger = logger
        self.port_map_id = None
        self.portmap = {}
        self.lidmap = {}
        self.roce_port_map = {}
        self.port_snapshot = set()
        self.sdk_utils = sdk_utils

    @staticmethod
    def get_instance(logger, sdk_utils):
        if PortHandler.__instance is None:
            PortHandler(logger, sdk_utils)
        return PortHandler.__instance

    def get_port_name(self, port_id):
        """
        Get port name
        SONiC - mapping ifindex => port name
        Cumulus - mapping logical port id => port name
        @param port_id: Port identifier. SDK log_port (lid) for CL or if_index for SONiC.
        @return: port name. (eg. 'swp1' in CL or 'Ethernet1' in SONiC)
        """
        if self.platform == PlatformType.SONIC:
            ifindex = str(port_id)
            return self.portmap.get(ifindex) or ifindex
        else:
            hex_port_id = hex(int(port_id))
            if hex_port_id not in self.lidmap:
                self._logger.info("not found {}".format(self.lidmap))
            return self.lidmap.get(hex_port_id) or port_id


    def get_port_lid(self, port_name):
        return self.portmap.get(port_name, port_name)


    def get_all_port_lids(self):
        return list(self.lidmap.keys())
    def get_all_int_local_ports(self):
        hex_lids = list(self.lidmap.keys())
        local_ports = [int(int(hlid, 16) - 65536) for hlid in hex_lids]
        return local_ports


    def _rebuild_port_maps(self):
        try:
            if self.platform == PlatformType.SONIC:
                return self._get_sonic_port_map()
            else:
                return self._get_cl_port_map_v2()
        except Exception as e:
            self._logger.error('Cannot get ports map: %s', str(e))


    def refresh_port_map(self):
        if self.platform == PlatformType.SONIC:
            ports = set(os.listdir('/sys/devices/virtual/net'))
        else:
            scandir_iterator = os.scandir('/sys/devices/virtual/net')
            port_list = [item.path.split('/')[-1] for item in scandir_iterator
                         if utils.is_valid_interface(item.path.split('/')[-1])]
            ports = set(port_list)
        self._logger.debug('port snapshot: %s', list(self.port_snapshot))
        if ports != self.port_snapshot:
            self.port_snapshot = ports
            self._logger.info('PortHandler: port map changed in a subprocess')
            self._rebuild_port_maps()
            return True
        return False


    def _get_sonic_port_map(self):
        output = self.sdk_utils.get_logical_ports()
        if output is None:
            return
        # As for now, SONiC port names cannot be translated to SDK log_port.
        # Hence setting latency/occupancy thresholds on specific ports are not supported.

        self.lidmap = {lid: lid for lid in output.split(',')}
        self.portmap = {}
        p = PosixPath('/sys/devices/virtual/net')
        for if_index_file in p.glob('*/ifindex'):
            if_index = if_index_file.read_text().strip('\n')
            if_name = if_index_file.parent.name
            self._logger.debug("WJH Adding port mapping {}: {}".format(if_index, if_name))
            self.portmap[if_index] = if_name


    def _get_cl_port_map(self):
        port = ''
        self.portmap = {}
        self.lidmap = {}

        if self.port_map_id is None:
            self.port_map_id = get_cl_switchd_sw_tables_idx('Backend Port Info (Linux) Table')
            self.lag_map_id = get_cl_switchd_sw_tables_idx('Backend Bond Lag Info Table')

        if self.port_map_id == 0:
            raise Exception('Cannot get port map table')

        cmd = 'cat {}/{}'.format(SWITCHD_SW_TABLE_PATH, self.port_map_id)
        output = subprocess_wrapper(cmd.split())
        for l in output.splitlines():
            if 'port-info' in l:
                port = l.split()[1]
            if 'lid' in l:
                lid = l.split()[1]
                self.portmap[lid] = port
                self.lidmap[port] = lid

    def _get_cl_port_map_v2(self):
        self.portmap = {}
        self.lidmap = {}

        if os.path.exists('/bin/systemctl'):
            is_active = is_active_systemd_svc('sx_sdk.service')
            if not is_active:
                return

        try:
            mlx_port_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json ports info'.split())
            try:
                mlx_port_info_json = json.loads(mlx_port_info) #pylint: disable=no-member
            except ValueError:
                raise
            for port_id, port_status in list(mlx_port_info_json.items()):
                hex_port_id = hex(int(port_id))
                self.portmap[port_status.get('ifname')] = hex_port_id
                self.lidmap[hex_port_id] = port_status.get('ifname')
                self.roce_port_map[port_id] = port_status.get('ifname')
        except Exception as e:
            raise e
