#
# Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
import re
import tarfile
import time
from io import BytesIO
from pathlib import Path

import docker

from netq_lib.common.enums import DefaultPaths
from netq_lib.common.utils import is_sonic, subprocess_wrapper, get_logger, is_active_systemd_svc, is_cl3, is_cl4, get_cl_platform_asic_type

SONIC_SYNCD_CONTAINER = "syncd"
VERSION_REGEX = re.compile(r'(\d+)[._](\d+)[._](\d+)')
SX_CHIP_TYPE_SPECTRUM = 6
SX_CHIP_TYPE_SPECTRUM2 = 9
SX_CHIP_TYPE_SPECTRUM3 = 11
SX_CHIP_TYPE_SPECTRUM4 = 13
SDK_VERSION_FILE = os.path.join(DefaultPaths.VAR_LIB_DIR, 'sdk_version.last')
SDK_UTILS_SCRIPT = '/usr/sbin/netq-sdk-utils'


class SdkUtils(object):

    def __init__(self, logger):
        self._logger = logger
        self._sdk_version = self.get_sdk_version()


    @property
    def sdk_version(self):
        return self._sdk_version


    def get_sdk_version(self):
        if is_sonic():
            return self._get_sonic_sdk_version()
        else:
            cmd = '/usr/bin/dpkg -l sx-sdk-eth'
            output = subprocess_wrapper(cmd.split())
            pkg = output.splitlines()[-1].split()[2]
            return VERSION_REGEX.search(pkg).group()


    def _get_sonic_sdk_version(self):
        syncd = self._get_sonic_syncd_container()
        res = syncd.exec_run("sx_sdk --version")
        if res.exit_code:
            raise Exception("Failed to retrieve SDK version: {res.output}")
        output = res.output.split()[2]
        if isinstance(output, bytes):
            output = output.decode()
        return output.rstrip(',')


    def _get_sonic_syncd_container(self):
        client = docker.from_env()
        syncd = client.containers.get(SONIC_SYNCD_CONTAINER)
        retries = 0
        while syncd.status != "running":
            if retries >= 60:
                raise Exception('{} is not running.'.format(SONIC_SYNCD_CONTAINER))
            self._logger.debug(
                'Waiting for syncd container to start. Current status: %s',
                syncd.status)
            time.sleep(3)
            retries += 1
        return syncd


    def is_sdk_version_equals_or_greater(self, version):
        return self._is_version_equals_or_greater(self.sdk_version, version)


    def _is_version_equals_or_greater(self, version1, version2):
        version1_tuple = tuple(map(int, VERSION_REGEX.search(version1).groups()))
        version2_tuple = tuple(map(int, VERSION_REGEX.search(version2).groups()))
        return version1_tuple >= version2_tuple


    @staticmethod
    def _link_sdk_sx_api_files():
        if not is_sonic():
            return
        sock_path = Path('/var/run/sx_api.sock')
        sdk_ready_path = Path('/tmp/sdk_ready')
        if not sock_path.exists() and not sock_path.is_symlink():
            sock_path.symlink_to(
                '/var/lib/docker/volumes/mlnx_sdk_socket/_data/sx_api.sock')
        if not sdk_ready_path.exists() and not sdk_ready_path.is_symlink():
            sdk_ready_path.symlink_to(
                '/var/lib/docker/volumes/mlnx_sdk_ready/_data/sdk_ready')


    def copy_sdk(self):
        self._logger.info("Copying SDK files")
        self._link_sdk_sx_api_files()
        sdk_version = None
        if os.path.isfile(SDK_VERSION_FILE):
            try:
                with open(SDK_VERSION_FILE, 'r') as f:
                    sdk_version = f.read()
            except Exception as e:
                self._logger.warning('Unable to read SDK file: %s', str(e))

        if sdk_version == self.sdk_version:
            return
        self.copy_sdk_libs()
        self.update_wjh_types()
        try:
            with open(SDK_VERSION_FILE, 'w') as f:
                f.write('{}'.format(self.sdk_version))
        except Exception as e:
            self._logger.warning(
                'Unable to update SDK version file: %s', str(e))


    def update_wjh_types(self):
        highest_supported_version_file = Path('/usr/bin/wjh_types.py')
        if not highest_supported_version_file.exists():
            wjh_types = Path('/usr/lib/cumulus/wjh')
            sorted_files = sorted(wjh_types.glob('wjh_types_*'),
                                  key=lambda file: file.name, reverse=True)
            highest_supported_version_file = sorted_files[-1]
            for wjh_types_file in sorted_files:
                file_ver = VERSION_REGEX.search(wjh_types_file.name).group()
                if self._is_version_equals_or_greater(self.sdk_version, file_ver):
                    highest_supported_version_file = wjh_types_file
                    break
        self._logger.info('Updating WJH types with %s', highest_supported_version_file.name)
        current_wjh_types = Path('/usr/lib/cumulus/wjh/wjh_types.py')
        if current_wjh_types.is_symlink() or current_wjh_types.exists():
            current_wjh_types.unlink()
        current_wjh_types.symlink_to(highest_supported_version_file)


    def copy_sdk_libs(self):
        if not is_sonic():
            return
        file_prefixes = ["libsx", "libsdk", "libsw", "libpolicer", "libbridge",
                         "librouter", "libcraccess", "libautoreg", "libwjh", "libscew"]
        pattern = r"/usr/lib/\\({}\\).*".format(r'\\|'.join(file_prefixes))
        syncd = self._get_sonic_syncd_container()
        # TODO: time can be improved #pylint: disable=fixme
        res = syncd.exec_run("find /usr/lib -maxdepth 1 -regex {}".format(pattern))
        if res.exit_code:
            raise Exception("Failed to retrieve SDK library file list: {}".format(res.output))
        for file_path in res.output.splitlines():
            if isinstance(file_path, bytes):
                file_path = file_path.decode()
            self._docker_cp(syncd, file_path)
        py_dist_packages = "/usr/lib/python2.7/dist-packages/"
        Path(py_dist_packages).mkdir(parents=True, exist_ok=True)
        self._docker_cp(syncd, "/usr/lib/python2.7/dist-packages/python_sdk_api")
        self._docker_cp(syncd, "/usr/etc/wjh_lib_conf.xml")


    def _docker_cp(self, container, file):
        self._logger.debug("Copying {} from {}".format(file, container.name))
        stream, stat = container.get_archive(file)
        tar_file = tarfile.open(fileobj=BytesIO(b''.join(chunk for chunk in stream)))
        tar_file.extractall(os.path.dirname(file.rstrip('/')))


    def get_logical_ports(self):
        try:
            if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    return None
        except Exception as e:
            self._logger.debug('Check sdk status %s', str(e))
            return None

        cmd = SDK_UTILS_SCRIPT + ' --get-logical-ports'
        output = subprocess_wrapper(cmd.split())
        if not output:
            raise Exception("failed to get logical ports from SDK")
        return output.splitlines()[-1]

    def get_chip_type_v3(self, platform):
        chip_type = None
        asic_type = get_cl_platform_asic_type(platform)
        if asic_type is None:
            self._logger.warning("Failed to retrieve chip type")
            asic_type_trim = None
        else:
            asic_type_trim = asic_type.lower().replace("-", "").strip()

        if asic_type_trim == 'spectrum':
            chip_type = SX_CHIP_TYPE_SPECTRUM
        elif asic_type_trim == 'spectrum2':
            chip_type = SX_CHIP_TYPE_SPECTRUM2
        elif asic_type_trim == 'spectrum3':
            chip_type = SX_CHIP_TYPE_SPECTRUM3
        elif asic_type_trim == 'spectrum4':
            chip_type = SX_CHIP_TYPE_SPECTRUM4

        return chip_type

    def get_chip_type_v2(self):
        chip_type = None
        cmd = SDK_UTILS_SCRIPT + ' --get-chip-type'

        try:
            if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    self._logger.warning("Inactive sx_sdk service, Chiptype not determined")
                    return chip_type
        except Exception as e:
            self._logger.debug('Check sdk status %s', str(e))
            return chip_type

        try:
            output = subprocess_wrapper(cmd.split())
            output = output.splitlines()[-1]
            chip_type = int(output)
        except: #pylint: disable=bare-except
            self._logger.warning("Failed to retrieve chip type, Chiptype not determined")
        return chip_type

    def get_chip_type(self):
        chip_type = SX_CHIP_TYPE_SPECTRUM
        cmd = SDK_UTILS_SCRIPT + ' --get-chip-type'

        try:
            if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    self._logger.warning("Inactive sx_sdk service, assuming Spectrum1")
                    return chip_type
        except Exception as e:
            self._logger.debug('Check sdk status %s', str(e))
            return chip_type

        try:
            output = subprocess_wrapper(cmd.split())
            output = output.splitlines()[-1]
            chip_type = int(output)
        except: #pylint: disable=bare-except
            self._logger.warning("Failed to retrieve chip type, assuming Spectrum1")
        return chip_type

    def get_ecmp_count(self):
        try:
            if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    return None
        except Exception as e:
            self._logger.debug('Check sdk status %s', str(e))
            return None

        cmd = SDK_UTILS_SCRIPT + ' --get-ecmp-count'
        output = subprocess_wrapper(cmd.split())
        if not output:
            self._logger.debug("failed to get ecmp count from SDK")
            return None
        return output.splitlines()[-1]

if __name__ == "__main__":
    sdk_logger = get_logger('sdk_utils')
    SdkUtils(sdk_logger).copy_sdk()
