#
# Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import ARConfigIntf, ARConfigGlobalV2
from netq_lib.common import configuration
from netq_lib.common.enums import ServiceStatus, DefaultPaths, ConfKw, Platform
from netq_lib.common.utils import save_fullstate
from netq_lib.common.utils import check_cl_version_onwards, get_logger
from netq_agent.parsers.common.render import BaseRender
from netq_agent.handlers.sdk_utils import SX_CHIP_TYPE_SPECTRUM2, SX_CHIP_TYPE_SPECTRUM3, SX_CHIP_TYPE_SPECTRUM4, SdkUtils

class AdaptiveRoutingConfigRender(BaseRender):
    def __init__(self):
        super(AdaptiveRoutingConfigRender, self).__init__()
        self.model = ARConfigGlobalV2
        self.__logger = get_logger(ConfKw.NETQ_AGENT, configuration)
        self.DEFAULT_LINK_UTIL_THRESHOLD = 70
        self.sdk_utils = None
        self.chip_type = None
        self.ar_logging_enabled = False

    def ar_logger(self, message):
        if not self.ar_logging_enabled:
            return

        self.__logger.debug(message)

    def is_ar_supported(self, platform):
        # ignore running these on VX
        if Platform.VX in platform:
            if os.path.exists(DefaultPaths.ARCONFIG_VX_FILE):
                # For testing alone, so scale AR config data for fabric without hardnodes
                return True
            return False

        # enable AR export only for Mellanox platform and on CL >= 5.1.0
        if not (check_cl_version_onwards(major=5, minor=1) and Platform.MELLANOX in platform):
            return False

        if Platform.CUMULUS not in platform:
            return False

        # enable AR export only for SPC2+
        if self.chip_type is None:
            self.sdk_utils = SdkUtils(self.__logger)
            self.chip_type = self.sdk_utils.get_chip_type_v3(platform)

        if self.chip_type and self.chip_type in (SX_CHIP_TYPE_SPECTRUM2, SX_CHIP_TYPE_SPECTRUM3, SX_CHIP_TYPE_SPECTRUM4):
            return True

        return False

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        ar_global_config = {}
        ar_intf_config = {}
        roce_enabled = False
        roce_mode = ""
        fullstate_ar_global_config_arr = []
        fullstate_ar_intf_config_arr = []

        platform = kwargs.get('platform', '')
        if not self.is_ar_supported(platform):
            self.ar_logger("AR: Unsupported platform {}".format(platform))
            return ServiceStatus.OK

        self.ar_logger("AR: fullstate: {}".format(fullstate))

        if os.path.exists(DefaultPaths.ROCE_STATUS_FILE_PATH):
            self.ar_logger("AR: File exists {}".format(DefaultPaths.ROCE_STATUS_FILE_PATH))
            try:
                with open(DefaultPaths.ROCE_STATUS_FILE_PATH, 'r') as fd:
                    self.ar_logger("AR: Open the file")
                    for line in fd.readlines():
                        self.ar_logger("AR: line: {}".format(line))
                        if 'traffic.roce_mode' in line:
                            mode = int(line.strip().split('=')[1].strip())
                            if mode != 0:
                                roce_enabled = True

                            if mode == 1:
                                roce_mode = "Lossy"
                            elif mode == 3:
                                roce_mode = "Lossless"

                            self.ar_logger("AR: roce_enable: {} roce_mode {}".format(roce_enabled, roce_mode))
                            break
            except Exception as e:
                self.__logger.info(e)
                pass

        old_arconfigintf_objs = []
        current_arconfigintf_objs = []
        m_arconfigintf_objs = ARConfigIntf.memmodel()
        if m_arconfigintf_objs:
            old_arconfigintf_objs = [old_obj for old_obj in m_arconfigintf_objs.query.filter(hostname=HOSTNAME)]

        if os.path.exists(DefaultPaths.ARCONFIG_FILE):
            self.ar_logger("AR: File exists {}".format(DefaultPaths.ARCONFIG_FILE))
            try:
                with open(DefaultPaths.ARCONFIG_FILE, 'r') as fd:
                    self.ar_logger("AR: opened the file")
                    for line in fd.readlines():
                        self.ar_logger("AR: {}".format(line))
                        if line.startswith('#') or line.strip() == "":
                            #commented line or empty line, Ignore
                            continue

                        parts = line.split(' = ')
                        self.ar_logger("AR: parts: {}".format(parts))

                        key = parts[0].strip()
                        value = parts[1].strip()

                        if value.isdigit():
                            value = int(value)
                        elif value.upper() == 'TRUE':
                            value = True
                        elif value.upper() == 'FALSE':
                            value = False

                        key0 = key.split('.')[0]
                        key1 = key.split('.')[1]
                        self.ar_logger("AR: key: {} value: {} key0: {} key1 {}".format(key, value, key0, key1))

                        if key0 == 'interface' and 'sw' in key1:
                            intf = key1  # interface name
                            param = key.split('.')[3]  # Extracting the param

                            if intf not in ar_intf_config:
                                ar_intf_config[intf] = {}

                            ar_intf_config[intf][param] = value
                        else:
                            key = key.replace('adaptive_routing.', '')
                            ar_global_config[key] = value
            except Exception as e:
                self.__logger.info(e)
                pass

            # Accessing individual variables per port
            for ifname, ifparam in ar_intf_config.items():
                ar_intf_config_obj = ARConfigIntf()
                ar_intf_enable = ifparam.get('enable', False)
                ar_intf_link_util_thresh = ifparam.get('link_util_thresh', self.DEFAULT_LINK_UTIL_THRESHOLD)
                ar_intf_link_util_thresh_disabled = ar_global_config.get('link_utilization_threshold_disabled', True)
                ar_intf_config_obj.ifname = ifname
                ar_intf_config_obj.is_enabled = ar_intf_enable
                ar_intf_config_obj.link_util_thresh_disabled = ar_intf_link_util_thresh_disabled
                ar_intf_config_obj.link_util_thresh = ar_intf_link_util_thresh
                current_arconfigintf_objs.append(ar_intf_config_obj)
                self.ar_logger("AR: {}".format(ar_intf_config_obj))
                if not fullstate:
                    ar_intf_config_obj.save()
                    self.ar_logger("AR: save interface config")
                else:
                    self.ar_logger("AR: fullstate append interface config")
                    fullstate_ar_intf_config_arr.append(ar_intf_config_obj)

            if fullstate:
                save_fullstate(fullstate_ar_intf_config_arr, ARConfigIntf().__class__.__name__)

            if current_arconfigintf_objs:
                for old_obj in old_arconfigintf_objs:
                    if old_obj.ifname not in [new_obj.ifname for new_obj in current_arconfigintf_objs]:
                        old_ent = ARConfigIntf()
                        old_ent.copy(old_obj)
                        old_ent.delete()
                        self.ar_logger("AR: delete old obj {}".format(old_obj))
                        old_arconfigintf_objs.remove(old_obj)
            else:
                self.ar_logger("AR: Delete all AR intf")
                self.delete_ar_intf_config()

            # Global config parsing
            ar_enable = ar_global_config.get('enable', False)
            link_util_thresh_disabled = ar_global_config.get('link_utilization_threshold_disabled', True)

            # Derive if all the config is written to ASIC:

            ar_global_config_obj = ARConfigGlobalV2()
            ar_global_config_obj.is_enabled = ar_enable
            ar_global_config_obj.is_roce_enabled = roce_enabled
            ar_global_config_obj.roce_mode = roce_mode
            self.ar_logger("AR: ar_global_config_obj: {}".format(ar_global_config_obj))

            if not fullstate:
                ar_global_config_obj.save()
                self.ar_logger("AR: ar_global_config_obj: save obj")
            else:
                fullstate_ar_global_config_arr.append(ar_global_config_obj)
                self.ar_logger("AR: ar_global_config_obj: fullstate append obj")

            if fullstate:
                save_fullstate(fullstate_ar_global_config_arr, ARConfigGlobalV2().__class__.__name__)
                self.ar_logger("AR: save_fullstate")

        else:
            self.delete_all_ar_config()

        return ServiceStatus.OK

    def delete_all_ar_config(self):
        self.delete_ar_global_config()
        self.delete_ar_intf_config()

    def delete_ar_global_config(self):
        m_ar_global_config = ARConfigGlobalV2.memmodel()
        if m_ar_global_config:
            for entry in m_ar_global_config.query.filter(hostname=HOSTNAME):
                self.ar_logger("AR: Delete global entry {}".format(entry))
                old_ar_global_config = ARConfigGlobalV2()
                old_ar_global_config.copy(entry)
                old_ar_global_config.delete()

    def delete_ar_intf_config(self):
        m_ar_intf_config = ARConfigIntf.memmodel()
        if m_ar_intf_config:
            for entry in m_ar_intf_config.query.filter(hostname=HOSTNAME):
                self.ar_logger("AR: Delete intf entry {}".format(entry))
                old_ar_intf_config = ARConfigIntf()
                old_ar_intf_config.copy(entry)
                old_ar_intf_config.delete()
