#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import ctypes
import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import BgpSession, EvpnVniInfo, Description
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.render import BaseRender

class BgpSessionRender(BaseRender):
    def __init__(self):
        super(BgpSessionRender, self).__init__()
        self.model = BgpSession

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        """Build summary of BGP node info, such as ASN & RID
        """
        retcode = ServiceStatus.OK

        fullstate = kwargs.get('fullstate')

        try:
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            # We treat invalid JSON as the infomation terminating
            json_out = []
            if fullstate is True:
                save_fullstate(json_out, BgpSession().__class__.__name__)
            retcode = ServiceStatus.FAIL

        # first process the entries already in the DB
        m_bgp_session = BgpSession.memmodel()
        for entry in list(m_bgp_session.query.filter(hostname=HOSTNAME)):
            if entry.vrf == 'default':
                vrf = 'Default'
                # FRR json output can have 'default' or 'Default'
                if 'default' in json_out:
                    vrf = 'default'
            else:
                vrf = entry.vrf
            old_bgp_entry = BgpSession()
            if (vrf not in json_out or
                    entry.peer_name not in json_out[vrf]):
                old_bgp_entry.copy(entry)
                if old_bgp_entry.objid:
                    old_descr = Description()
                    old_descr.objtype = old_bgp_entry.__class__.__name__
                    old_descr.descrid = old_bgp_entry.objid
                    old_descr.delete()
                old_bgp_entry.delete()
            else:
                # Check that ASN hasn't changed
                lasn = json_out[vrf][entry.peer_name]['localAs']
                if int(lasn) < 0:
                    lasn = ctypes.c_uint32(lasn).value
                if lasn != entry.asn:
                    old_bgp_entry.copy(entry)
                    if old_bgp_entry.objid:
                        old_descr = Description()
                        old_descr.objtype = old_bgp_entry.__class__.__name__
                        old_descr.descrid = old_bgp_entry.objid
                        old_descr.delete()
                    old_bgp_entry.delete()

                # Check that the description hasn't been deleted
                if (entry.objid and
                        not json_out[vrf][entry.peer_name].get('nbrDesc', '')):
                    descr = Description()
                    descr.objtype = entry.__class__.__name__
                    descr.descrid = entry.objid
                    descr.delete()

        bgp_nbr_arr = []
        bgp_desc_arr = []
        for vrf in json_out:
            vrf_id = json_out[vrf]['vrfId']
            for key in json_out[vrf]:
                if key in ('vrfName', 'vrfId'):
                    continue

                session = json_out[vrf][key]

                bgp_nbr = BgpSession()
                # Quagga/FRR names the default VRF 'Default' whereas Linux names it
                # nothing, and in netq we call it 'default'. Fix this.
                if vrf == 'Default':
                    bgp_nbr.vrf = 'default'
                else:
                    bgp_nbr.vrf = vrf
                bgp_nbr.vrfid = vrf_id
                bgp_nbr.peer_name = key

                # protect against new sections that are not 'sessions'
                if session.get('bgpState') is None:
                    continue
                bgp_nbr.state = session['bgpState']
                if bgp_nbr.state != 'Established':
                    bgp_nbr.state = 'NotEstd'
                bgp_nbr.asn = int(session['localAs'])
                if int(bgp_nbr.asn) < 0:
                    bgp_nbr.asn = ctypes.c_uint32(bgp_nbr.asn).value
                bgp_nbr.reason = 'N/A'
                if bgp_nbr.state == 'Established':
                    bgp_nbr.up_time = float(
                        session.get('bgpTimerUpEstablishedEpoch', 0.0))
                else:
                    bgp_nbr.up_time = 0.0
                bgp_nbr.hold_time = session.get('bgpTimerHoldTimeMsecs', 0)
                bgp_nbr.keep_alive_interval = session.get('bgpTimerKeepAliveIntervalMsecs', 0)

                if 'gracefulRestartInfo' in session:
                    graceful_restart_info = session['gracefulRestartInfo']
                    bgp_nbr.local_gr_mode = graceful_restart_info.get('localGrMode', '')
                    bgp_nbr.remote_gr_mode = graceful_restart_info.get('remoteGrMode', '')
                    bgp_nbr.graceful_restart_rbit = graceful_restart_info.get('rBit', False)

                    timers = graceful_restart_info.get('timers', {})
                    bgp_nbr.configured_restart_time = timers.get('configuredRestartTimer', 0) * 1000
                    bgp_nbr.received_restart_time = timers.get('receivedRestartTimer', 0) * 1000

                    bgp_nbr.ipv4_stale_path_time = ((graceful_restart_info.get('ipv4Unicast', {})).get('timers', {})).get('stalePathTimer', 0) * 1000
                    bgp_nbr.ipv6_stale_path_time = ((graceful_restart_info.get('ipv6Unicast', {})).get('timers', {})).get('stalePathTimer', 0) * 1000

                bgp_nbr.peer_router_id = session['remoteRouterId']
                bgp_nbr.peer_asn = session['remoteAs']
                if int(bgp_nbr.peer_asn) < 0:
                    bgp_nbr.peer_asn = ctypes.c_uint32(bgp_nbr.peer_asn).value
                bgp_nbr.peer_hostname = session.get('hostname', '')
                bgp_nbr.upd8_rx = session['messageStats']['updatesRecv']
                bgp_nbr.upd8_tx = session['messageStats']['updatesSent']
                afi_session = session['addressFamilyInfo']
                bgp_nbr.last_reset_time = 0.0
                if bgp_nbr.state == 'NotEstd':
                    if 'lastResetDueTo' in afi_session:
                        bgp_nbr.reason = (
                            afi_session['lastResetDueTo']
                        )
                        bgp_nbr.last_reset_time = \
                            float(afi_session['lastResetTimerMsecs'] / 1000)
                    elif 'lastResetDueTo' in session:
                        bgp_nbr.reason = (
                            session['lastResetDueTo']
                        )
                        if bgp_nbr.reason.startswith("BGP Notification"):
                            if 'lastNotificationReason' in session:
                                bgp_nbr.reason = (
                                    session['lastNotificationReason'])
                        bgp_nbr.last_reset_time = \
                            float(session['lastResetTimerMsecs'] / 1000)
                    else:
                        bgp_nbr.last_reset_time = 0.0
                json_key = None
                if 'IPv4 Unicast' in session['addressFamilyInfo']:
                    json_key = 'IPv4 Unicast'
                elif 'ipv4Unicast' in session['addressFamilyInfo']:
                    json_key = 'ipv4Unicast'
                if json_key:
                    bgp_nbr.ipv4_pfx_rcvd = (
                        int(afi_session[json_key]['acceptedPrefixCounter'])
                    )
                    bgp_nbr.ipv4_packet_queue_length = afi_session[json_key].get('packetQueueLength', 0)
                else:
                    bgp_nbr.ipv4_pfx_rcvd = 0
                    bgp_nbr.ipv4_packet_queue_length = 0
                json_key = None
                if 'IPv6 Unicast' in session['addressFamilyInfo']:
                    json_key = 'IPv6 Unicast'
                elif 'ipv6Unicast' in session['addressFamilyInfo']:
                    json_key = 'ipv6Unicast'
                if json_key:
                    bgp_nbr.ipv6_pfx_rcvd = (
                        int(afi_session[json_key]['acceptedPrefixCounter'])
                    )
                    bgp_nbr.ipv6_packet_queue_length = afi_session[json_key].get('packetQueueLength', 0)
                else:
                    bgp_nbr.ipv6_pfx_rcvd = 0
                    bgp_nbr.ipv6_packet_queue_length = 0
                json_key = None
                if 'L2VPN EVPN' in session['addressFamilyInfo']:
                    json_key = 'L2VPN EVPN'
                elif 'l2VpnEvpn' in session['addressFamilyInfo']:
                    json_key = 'l2VpnEvpn'
                if json_key:
                    bgp_nbr.evpn_pfx_rcvd = (
                        int(afi_session[json_key]['acceptedPrefixCounter']))
                    bgp_nbr.evpn_packet_queue_length = afi_session[json_key].get('packetQueueLength', 0)
                else:
                    bgp_nbr.evpn_pfx_rcvd = 0
                    bgp_nbr.evpn_packet_queue_length = 0
                if 'connectionsDropped' in afi_session:
                    bgp_nbr.conn_dropped = afi_session['connectionsDropped']
                    bgp_nbr.conn_estd = afi_session['connectionsEstablished']
                elif 'connectionsDropped' in session:
                    bgp_nbr.conn_dropped = session['connectionsDropped']
                    bgp_nbr.conn_estd = session['connectionsEstablished']

                # BGP Capabilities
                rx_families = []
                tx_families = []
                if 'neighborCapabilities' in session:

                    neighbour_capabilities = session['neighborCapabilities']
                    bgp_nbr.graceful_restart = neighbour_capabilities.get('gracefulRestart', '')
                    bgp_nbr.graceful_restart_remote_time = neighbour_capabilities.get('gracefulRestartRemoteTimerMsecs', 0)
                    bgp_nbr.address_families_by_peer = neighbour_capabilities.get('addressFamiliesByPeer', '')

                    if 'multiprotocolExtensions' in session['neighborCapabilities']:
                        caps_dict = session['neighborCapabilities']['multiprotocolExtensions']

                        json_key = None
                        if 'IPv4 Unicast' in caps_dict:
                            json_key = 'IPv4 Unicast'
                        elif 'ipv4Unicast' in caps_dict:
                            json_key = 'ipv4Unicast'

                        if json_key:
                            if caps_dict[json_key].get(
                                    'advertisedAndReceived', False):
                                tx_families.append('ipv4')
                                rx_families.append('ipv4')
                            elif caps_dict[json_key].get(
                                    'advertised', False):
                                tx_families.append('ipv4')
                            elif caps_dict[json_key].get(
                                    'received', False):
                                rx_families.append('ipv4')

                        json_key = None
                        if 'IPv6 Unicast' in caps_dict:
                            json_key = 'IPv6 Unicast'
                        elif 'ipv6Unicast' in caps_dict:
                            json_key = 'ipv6Unicast'

                        if json_key:
                            if caps_dict[json_key].get(
                                    'advertisedAndReceived', False):
                                tx_families.append('ipv6')
                                rx_families.append('ipv6')
                            elif caps_dict[json_key].get(
                                    'advertised', False):
                                tx_families.append('ipv6')
                            elif caps_dict[json_key].get(
                                    'received', False):
                                rx_families.append('ipv6')

                        json_key = None
                        if 'L2VPN EVPN' in caps_dict:
                            json_key = 'L2VPN EVPN'
                        elif 'l2VpnEvpn' in caps_dict:
                            json_key = 'l2VpnEvpn'

                        if json_key:
                            if caps_dict[json_key].get(
                                    'advertisedAndReceived', False):
                                tx_families.append('evpn')
                                rx_families.append('evpn')
                            elif caps_dict[json_key].get(
                                    'advertised', False):
                                tx_families.append('evpn')
                            elif caps_dict[json_key].get(
                                    'received', False):
                                rx_families.append('evpn')
                    bgp_nbr.tx_families = tx_families
                    bgp_nbr.rx_families = rx_families

                # Save description if present
                descr = session.get('nbrDesc', '')
                if descr:
                    bgp_descr = Description()
                    bgp_descr.objtype = bgp_nbr.__class__.__name__
                    bgp_descr.descrid = bgp_nbr.get_objid()
                    bgp_descr.description = descr
                    bgp_descr.timestamp = bgp_nbr.timestamp
                    bgp_nbr.objid = bgp_descr.descrid
                    bgp_descr.save()
                    bgp_desc_arr.append(bgp_descr)
                bgp_nbr.save()
                bgp_nbr_arr.append(bgp_nbr)

        if fullstate is True:
            save_fullstate(bgp_desc_arr, Description().__class__.__name__)
            save_fullstate(bgp_nbr_arr, BgpSession().__class__.__name__)

        return retcode


class BgpEvpnVniRender(BaseRender):
    def __init__(self):
        super(BgpEvpnVniRender, self).__init__()
        self.model = EvpnVniInfo

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        '''Process EVPN VNI command output'''
        retcode = ServiceStatus.OK
        fullstate = kwargs.get('fullstate')

        try:
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            # We treat invalid JSON as the infomation terminating
            json_out = []
            if fullstate is True:
                save_fullstate(json_out, EvpnVniInfo().__class__.__name__)
            retcode = ServiceStatus.FAIL

        # first process the entries already in the DB
        m_evpn_info = EvpnVniInfo.memmodel()
        no_entry = True
        for entry in list(m_evpn_info.query.filter(hostname=HOSTNAME)):
            no_entry = False
            if retcode == ServiceStatus.FAIL:
                new_entry = EvpnVniInfo()
                new_entry.copy(entry)
                new_entry.delete()
                continue

            vni = str(entry.vni)
            if vni == 0:
                continue
            if vni not in json_out:
                new_entry = EvpnVniInfo()
                new_entry.copy(entry)
                new_entry.delete()
                continue

            # Check that the RD hasn't changed (happens when a VXLAN dev is
            # deleted and added back
            if json_out[vni]['rd'] != entry.rd:
                new_entry = EvpnVniInfo()
                new_entry.copy(entry)
                new_entry.delete()

        if not json_out and no_entry:
            # There was no JSON output, but thats because there's no EVPN either
            retcode = ServiceStatus.OK

        if json_out:
            try:
                adv_all_vni = json_out['advertiseAllVnis'] == 'Enabled'
                adv_gw_ip = json_out['advertiseGatewayMacip']
                adv_svi_macip = json_out.get('advertiseSviMacIp', 'Disabled')
                flooding = json_out.get('flooding', 'Enabled')
            except ValueError:
                # assume defaults
                adv_all_vni = False
                adv_gw_ip = 'Disabed'
                adv_svi_macip = 'Disabled'
                flooding = 'Enabled'

        bgp_evpn_arr = []
        for key in json_out:
            # we only care about the vni info which is a dict
            # ignore everything else
            if not isinstance(json_out[key], dict):
                continue

            # protect against a dict that is not vni
            if json_out[key].get('vni') is None:
                continue
            new_entry = EvpnVniInfo()
            new_entry.adv_all_vni = adv_all_vni
            new_entry.flooding = flooding == 'Enabled'

            try:
                new_entry.vni = int(key)
                new_entry.origin_ip = json_out[key]['originatorIp']
                new_entry.rd = json_out[key]['rd']
                new_entry.export_rt = json_out[key]['exportRTs']
                new_entry.import_rt = json_out[key]['importRTs']
                new_entry.in_kernel = json_out[key]['inKernel']
                new_entry.is_l3 = json_out[key].get('type', 'L2') != 'L2'
                new_entry.adv_gw_ip = json_out[key].get('advertiseGatewayMacip', adv_gw_ip)
                new_entry.adv_svi_macip = json_out[key].get('advertiseSviMacIp', adv_svi_macip)
                new_entry.adv_pip = json_out[key].get('advertisePip', 'Disabled')
                new_entry.sys_ip = json_out[key].get('sysIP')
                new_entry.sys_mac = json_out[key].get('sysMAC')
                new_entry.rmac = json_out[key].get('rmac')
                new_entry.save()
                if fullstate is True:
                    bgp_evpn_arr.append(new_entry)
            except ValueError:
                # dont crash on invalid json!
                pass

        if fullstate is True:
            save_fullstate(bgp_evpn_arr, EvpnVniInfo().__class__.__name__)

        return retcode
