#
# Copyright (c) 2020-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import subprocess
import time
import re
import sys
import eventlet
import eventlet.pools
import greenlet

from netq_lib.orm import HOSTNAME

from netq_lib.orm.redisdb.models import Dom
from netq_lib.common.enums import ServiceStatus
from netq_agent.parsers.common.render import BaseRender
from netq_agent.parsers.common.render import ReturnGroup


class DomRender(BaseRender):
    DOM_PROCESS_DURATION = 120
    DOM_CMD_TIMEOUT = 20
    LASER_RX_POWER_PATTERN = re.compile(r'(?:Rcvr signal avg optical power|Receiver signal OMA)\((Channel [0-9]+)\)')
    LASER_OUTPUT_POWER_PATTERN = re.compile(r'Transmit avg optical power \((Channel [0-9]+)\)')
    LASER_BIAS_CURRENT_PATTERN = re.compile(r'Laser tx bias current \((Channel [0-9]+)\)')

    def __init__(self):
        super(DomRender, self).__init__()
        self.first_run = True
        self.if_name_list = list()
        self.process_if_list = list()
        self.batch_size = 0
        self.bucket_index = 0
        self.total_buckets = 1
        self.cmd_periodicity = 0
        self.interface_ethtool_op = {}
        self.gthread_dict = {}
        self.sub_process = {}
        self._logger = None

    def __compute_if_process_list(self, kwargs):

        if self.bucket_index == 0:
            self.cmd_periodicity = kwargs.get('period', self.DOM_PROCESS_DURATION)
            if_name_set = kwargs.get('optical_ports', None)
            if not if_name_set:
                # No optical ports to process
                return
            self.if_name_list = list(if_name_set)
            if self.first_run:
                self.total_buckets = 1
                self.batch_size = len(self.if_name_list)
                self.first_run = False
            else:
                self.handle_outdated_objs()
                if self.DOM_PROCESS_DURATION < self.cmd_periodicity:
                    self.total_buckets = (self.cmd_periodicity // self.DOM_PROCESS_DURATION)
                div, mod = divmod(len(self.if_name_list), self.total_buckets)
                self.batch_size = (div+1) if mod > 0 else div

        start_idx = self.bucket_index * self.batch_size
        end_idx = start_idx + self.batch_size
        # Update bucket index
        self.bucket_index += 1
        if self.bucket_index == self.total_buckets:
            self.bucket_index = 0
        # Boundary check for indexes
        if start_idx >= len(self.if_name_list):
            self.process_if_list = list()
            return
        elif end_idx > len(self.if_name_list):
            end_idx = len(self.if_name_list)
        self.process_if_list = self.if_name_list[start_idx:end_idx]

    def process_dom_interfaces(self, kwargs):
        self._logger = kwargs.get('logger')
        # command = "ethtool -m "
        command = self._get_command()
        thread_pool = kwargs.get('pool', '*')
        self.__compute_if_process_list(kwargs)

        for ifName in self.process_if_list:
            # Run ethtool command for each interface in a separate thread
            self.__run_command_safe(command + ifName, thread_pool, ifName)

        for ifName in self.process_if_list:
            cmd_time_out = time.time() + int(self.DOM_CMD_TIMEOUT)
            while True:
                cmd_out = ''
                if ifName in self.interface_ethtool_op:
                    cmd_out = self.interface_ethtool_op.pop(ifName)
                    ret_code = 0 if cmd_out.success else -1
                    break

                if time.time() > cmd_time_out:
                    ret_code = -1
                    try:
                        self.sub_process[ifName].terminate()
                        eventlet.sleep(0.3)
                        self.sub_process[ifName].wait()
                        eventlet.kill(gthread_dict[ifName], greenlet.GreenletExit)
                    except Exception as e:
                        break
                    break
                eventlet.sleep(0.3)

            if ret_code == 0:
                eth_m_elems, rc = self._process_command_output(cmd_out)
                if rc != ServiceStatus.OK:
                    continue

                dom_obj = self._create_dom_object(ifName, eth_m_elems)
                if dom_obj:
                    dom_obj.save()

    def handle_outdated_objs(self):
        m_dom = Dom.memmodel()
        old_dom_objs = []
        if m_dom:
            old_dom_objs = (
                list(m_dom.query.filter(hostname=HOSTNAME))
            )

        for old_dom_obj in old_dom_objs:
            if old_dom_obj.ifname not in self.if_name_list:
                old_dom_ent = Dom()
                old_dom_ent.copy(old_dom_obj)
                old_dom_ent.delete()

    def __stop_command_checker(self, green_thread, ifName, cmd):
        try:
            (out, err) = green_thread.wait()
            if cmd.returncode != 0:
                self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
            else:
                self.interface_ethtool_op[ifName] = ReturnGroup(success=True, op=out)
            return
        except greenlet.GreenletExit:
            self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
        except Exception:  # pylint: disable=broad-except
            self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
            eventlet.kill(None, *sys.exc_info())

    def __run_command_safe(self, command, thread_pool, ifName):
        """Run time-limited command and update the global proc dict self.interface_ethtool_op
        with the ethtool status and response
        """
        global gthread_dict
        global sub_process
        try:
            cmd = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        except (subprocess.CalledProcessError, OSError):
            self.interface_ethtool_op[ifName] = ReturnGroup(success=False, op="")
            return

        self.sub_process[ifName] = cmd
        green_thread = thread_pool.spawn(cmd.communicate)
        green_thread.link(self.__stop_command_checker, ifName, cmd)
        self.gthread_dict[ifName] = green_thread
