#
# Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import os
import shutil
import json
from netq_lib.orm.redisdb.models import LogExporter
from netq_lib.common import configuration
from netq_lib.common.utils import send_list_data
from netq_lib.common.enums import ServiceStatus, DefaultPaths, ConfKw
from netq_agent.parsers.common.render import BaseRender

class LogExporterRender(BaseRender):
    def __init__(self):
        super(LogExporterRender, self).__init__()
        self.model = LogExporter
        self._logger = None

        conf_dict = configuration.get_config('netq-agent')
        self.log_exporter_enabled = conf_dict.get('netq-agent').get(ConfKw.IS_LOG_EXPORTER_ENABLED, ConfKw.DEFAULT_IS_LOG_EXPORTER_ENABLED)

        self.log_exporter_max_lines = conf_dict.get('netq-agent').get(ConfKw.LOG_EXPORTER_MAX_LINES, ConfKw.DEFAULT_LOG_EXPORTER_MAX_LINES)

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        if not self.log_exporter_enabled:
            return None

        self._logger = kwargs.get('logger')
        max_lines_to_export = self.log_exporter_max_lines

        try:
            total_lines = 0
            if os.path.exists(DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH):
                with open(DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH, 'r') as consumer_file:
                    lines = consumer_file.readlines()
                    total_lines = len(lines)

            if total_lines == 0:
                # Consumer file is fully exported or it is yet to be created
                if os.path.exists(DefaultPaths.LOG_EXPORTER_PRODUCER_FILE_PATH):
                    # New logs to export present in producer file
                    shutil.move(DefaultPaths.LOG_EXPORTER_PRODUCER_FILE_PATH, \
                                DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH)
                else:
                    # nothing to export present in producer file
                    return ServiceStatus.OK
            else:
                # Consumer file still had data to be exported in chunks of self.log_exporter_max_lines
                self.__logger_coalescer.debug("Exporting logs in chunks of utmost {}".format(self.log_exporter_max_lines))

        except Exception as exc:
            self._logger.debug(str(exc))
            return ServiceStatus.OK

        try:
            with open(DefaultPaths.LOG_EXPORTER_CONSUMER_FILE_PATH, 'r+') as consumer_file:
                # Export utmost self.log_exporter_max_lines at a time to prevent burst of gRPCs towards opta
                lines = consumer_file.readlines()
                total_lines = len(lines)

                if max_lines_to_export > total_lines:
                    #Last chunk
                    max_lines_to_export = total_lines

                # Remove the first self.log_exporter_max_lines lines from the consumer file
                consumer_file.seek(0)
                consumer_file.truncate()
                consumer_file.write(''.join(lines[max_lines_to_export:]))
                log_exporter_obj_list = []
                log_exporter_send_list = False

                # Iterate over the max_lines_to_export lines
                for idx, line in enumerate(lines):
                    # If the count reaches the number of lines to iterate, exit the loop
                    if idx == max_lines_to_export:
                        break

                    prefix, json_string = line.split(': ', 1)
                    data = json.loads(json_string)
                    new_log_exporter_obj = LogExporter()
                    new_log_exporter_obj.log_message = str(data[1]['log'])
                    log_exporter_obj_list.append(new_log_exporter_obj)
                    log_exporter_send_list = True

                if log_exporter_send_list:
                    send_list_data(log_exporter_obj_list, LogExporter().__class__.__name__)

        except Exception as exc:
            self._logger.info(str(exc))

        return ServiceStatus.OK
