#
# Copyright (c) 2017-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import os

from netq_lib.orm.redisdb.models import Ntp
from netq_lib.common.utils import is_active_systemd_svc, netq_agent_process_status_write
from netq_lib.common.enums import NetqAgentStatus, NetqAgentState

is_msg_logged_once = False
def process_ntpq(output, logger, **kwargs):
    global is_msg_logged_once #pylint: disable=global-statement
    ''' Parse ntpq -p command output
         - Function extracts active server information
         - If activer server is found it returns ntp_sync = 'yes'
    ''' #pylint: disable=pointless-string-statement
    current_server = '-'
    stratum = 16
    ntp = Ntp()
    ntp_sync = 'no'
    is_ptp_active = False

    if output:
        for line in output.splitlines():
            columns = line.split()
            if len(columns) < 3:
                continue
            if '*' in columns[0]:
                current_server = columns[0][1:]
                if len(columns) >= 8:
                    stratum = columns[-8]
                    ntp_sync = 'yes'
                else:
                    stratum = 16

    if ntp_sync == 'no':
        if os.path.exists('/bin/systemctl'):
            is_ptp_active = is_active_systemd_svc('ptp4l.service')
            if is_ptp_active:
                if not is_msg_logged_once:
                    is_msg_logged_once = True
                    msg_string = "NTP not synced... PTP service is active"
                    logger.info(msg_string)
                    netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NTP_STATUS,
                                                    NetqAgentState.NETQ_AGENT_STATE_NA,
                                                    msg_string)
            else:
                '''
                PTP service is not active, handle transition of services
                say ptp is disabled and ntp was tried to enabled.
                and NTP sync is 'no', we should set the boolean to log the
                message to False again.
                ''' #pylint: disable=pointless-string-statement
                is_msg_logged_once = False

        if not is_msg_logged_once:
            logger.warn("NTP not synced")
            netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NTP_STATUS,
                                            NetqAgentState.NETQ_AGENT_STATE_WARNING,
                                            'NTP not synced, Agent may show up as rotten due to incorrect timestamps')

        if output:
            logger.debug("NTP server details are:")
            for line in output.splitlines():
                logger.debug("Server: {}".format(line))
    else:
        netq_agent_process_status_write(NetqAgentStatus.NETQ_AGENT_NTP_STATUS,
                                        NetqAgentState.NETQ_AGENT_STATE_SUCCESS,
                                        'NTP sync success')
    ntp.current_server = current_server
    ntp.stratum = stratum
    ntp.ntp_sync = ntp_sync
    ntp.ntp_app = 'ntpq'
    ntp.ptp_state = is_ptp_active
    ntp.save()

    return ntp_sync

def process_timedatectl(output, **kwargs):
    ''' Parse timedatectl status command output
         - checks for 'Network time on: yes' or 'NTP synchronized: yes'
         - checks for 'systemd-timesyncd.service active: yes' (ub1804)
         - If found, returns ntp_sync = 'yes'
         - verified on ub1604 w/ systemd v229
         - verified on ub1804 w/ systemd v237
    '''
    ntp = Ntp()
    ntp_sync = 'no'

    if output:
        for line in output.splitlines():
            columns = line.split(':')
            # ubuntu and debian datetimectl format differences
            if 'Network time on' in columns[0] or \
                    'NTP synchronized' in columns[0] or \
                    'systemd-timesyncd.service active' in columns[0]:
                ntp_sync = columns[1].lstrip()
                break

    # current server and stratum not known
    # when using timedatectl / systemd-timesyncd
    ntp.current_server = '-'
    ntp.stratum = 16
    ntp.ntp_sync = ntp_sync
    ntp.ntp_app = 'timedatectl'
    ntp.save()

    return ntp_sync

def process_chronyc(output, **kwargs):
    ''' Parse chronyc sources command output
         - Function extracts active server information
         - If active server is found it returns ntp_sync = 'yes'
         - Eg.
# chronyc  sources
210 Number of sources = 4
MS Name/IP address         Stratum Poll Reach LastRx Last sample
===============================================================================
^+ up2.com                       2  10   377   733   -815us[ -815us] +/-   95ms
^+ ntp1.csl.tjhsst.edu           3   8   377   228   +239us[ +239us] +/-   89ms
^? excalibur.prolixium.com       0  10     0     -     +0ns[   +0ns] +/-    0ns
^* resolver1.skyfiberintern>     2  10   377   939   +277us[ +404us] +/-   44ms

    '''
    current_server = '-'
    stratum = 16
    ntp = Ntp()
    ntp_sync = 'no'

    if output:
        for line in output.splitlines():
            columns = line.split()
            if len(columns) < 3:
                continue
            if '*' in columns[0]:
                current_server = columns[1]
                if len(columns) >= 2:
                    stratum = columns[2]
                    ntp_sync = 'yes'
                else:
                    stratum = 16

    ntp.current_server = current_server
    ntp.stratum = stratum
    ntp.ntp_sync = ntp_sync
    ntp.ntp_app = 'chronyc'
    ntp.save()

    return ntp_sync
