#
# Copyright (c) 2020-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import OspfNbr, OspfIf
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_lib.common import configuration
from netq_agent.parsers.common.render import BaseRender


class OspfIfRender(BaseRender):
    def __init__(self):
        super(OspfIfRender, self).__init__()
        self.model = OspfIf

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        retcode = ServiceStatus.OK
        fullstate = kwargs.get('fullstate')

        try:
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            # We treat invalid JSON as the infomation terminating
            json_out = []
            if fullstate is True:
                save_fullstate(json_out, OspfIf().__class__.__name__)
            retcode = ServiceStatus.FAIL

        interfaces = {}
        for vrf in json_out:
            try:
                # ospf json notorious for missing fields - just protect entire block
                if vrf != 'default':
                    # we only support default vrf
                    continue
                interfaces = json_out[vrf]['interfaces']
                break
            except Exception:
                pass

        # With OSPF, failed sessions just don't show up. So catch those
        m_ospfif = OspfIf.memmodel()
        json_out = interfaces
        for entry in list(m_ospfif.query.filter(hostname=HOSTNAME)):
            new_entry = OspfIf()
            key = entry.ifname
            if key not in json_out or json_out[key].get('area') != entry.area:
                # A key parameter changed or i/f is no longer enabled for ospf
                new_entry.copy(entry)
                new_entry.delete()

        # Handle new entries
        ospf_arr = []
        for key in json_out:
            try:
                # ospf json notorious for missing fields - just protect entire block
                entry = OspfIf()
                entry.ifname = key
                entry.router_id = json_out[key]['routerId']
                entry.area = json_out[key]['area']
                entry.if_up = json_out[key].get('ifUp', False)
                entry.is_unnumbered = json_out[key].get('ifUnnumbered', False)
                entry.network_type = json_out[key]['networkType']
                entry.nbr_count = json_out[key]['nbrCount']
                entry.nbr_adj_count = json_out[key]['nbrAdjacentCount']
                entry.cost = json_out[key]['cost']
                entry.is_passive = json_out[key].get('timerPassiveIface', False)
                entry.mtu = json_out[key]['mtuBytes']
                # Bug in code since time immemorial; the time supplied in JSON
                # isn't in msecs, but its 1000/real_value. So, 5s retransmit time
                # becomes 1000/5 = 200.
                # This is true in version 3.3. In future we'll need to check the
                # version of the OS this agent is running on
                # Fixed in CL 4
                if 'timerDeadMsecs' in json_out[key]:
                    entry.dead_time = int(1000 / json_out[key]['timerDeadMsecs'])
                elif 'timerDeadSecs' in json_out[key]:
                    entry.dead_time = int(json_out[key]['timerDeadSecs'])

                if 'timerRetransmitSecs' in json_out[key]:
                    entry.rexmit_time = int(json_out[key]['timerRetransmitSecs'])
                elif 'timerRetransmit' in json_out[key]:
                    entry.rexmit_time = int(1000 / json_out[key]['timerRetransmit'])

                if configuration.is_cl4():
                    entry.hello_time = int(json_out[key]['timerMsecs'] / 1000)
                else:
                    entry.hello_time = int(1000 / json_out[key]['timerMsecs'])
                # Can't really check status with this output, need to switch
                entry.save()
                if fullstate is True:
                    ospf_arr.append(entry)
            except Exception:
                pass

        if fullstate is True:
            save_fullstate(ospf_arr, OspfIf().__class__.__name__)

        return retcode


class OspfNbrRender(BaseRender):
    def __init__(self):
        super(OspfNbrRender, self).__init__()
        self.model = OspfNbr

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        retcode = ServiceStatus.OK
        fullstate = kwargs.get('fullstate')

        try:
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            # We treat invalid JSON as the infomation terminating
            json_out = []
            if fullstate is True:
                save_fullstate(json_out, OspfNbr().__class__.__name__)
            retcode = ServiceStatus.FAIL

        # With OSPF, failed sessions just don't show up. So catch those
        m_ospfnbr = OspfNbr.memmodel()
        m_ospfif = OspfIf.memmodel()
        neighbors = {}
        for vrf in json_out:
            try:
                # ospf json notorious for missing fields - just protect entire block
                if 'neighbors' in json_out[vrf]:
                    neighbors[vrf] = json_out[vrf]['neighbors']
                else:
                    neighbors[vrf] = json_out[vrf]
            except Exception:
                pass

        for entry in list(m_ospfnbr.query.filter(hostname=HOSTNAME)):
            new_entry = OspfNbr()
            nbr = entry.peer_id
            nbr_vrf = self.get_nbr_vrf(nbr, neighbors)
            if not nbr_vrf:
                if svc_failed:
                    new_entry.copy(entry)
                    new_entry.delete()
                else:
                    # check ospfif table for the same entry
                    # if its not there - then delete this entry
                    ospfif_present = False
                    ifentries = m_ospfif.query.filter(hostname=HOSTNAME,
                                                      ifname=entry.ifname)
                    for ifentry in ifentries:
                        ospfif_present = True
                        break
                    if not ospfif_present:
                        # delete
                        new_entry.copy(entry)
                        new_entry.delete()
                    elif entry.state != 'Fail':
                        new_entry.state = 'Fail'
                        new_entry.ifname = entry.ifname
                        new_entry.area = entry.area
                        new_entry.peer_id = entry.peer_id
                        new_entry.save()
            else:
                # one nbr can be learnt via multiple interfaces
                if type(neighbors[nbr_vrf][nbr]) == list: #pylint: disable=unidiomatic-typecheck
                    ifacelist = neighbors[nbr_vrf][nbr]
                else:
                    ifacelist = [neighbors[nbr_vrf][nbr]]

                for iface in ifacelist:
                    area_id = iface['areaId'].split(':')[0]
                    if entry.area != area_id:
                        new_entry.copy(entry)
                        new_entry.delete()

        # Handle new entries
        ospf_arr = []
        for vrf in neighbors:
            for nbr in neighbors[vrf]:
                try:
                    # ospf json notorious for missing fields - just protect entire block

                    # in old format - we need to ignore certain keys
                    if nbr in ('vrfName', 'vrfId'):
                        continue

                    # one nbr can be learnt via multiple interfaces
                    if type(neighbors[vrf][nbr]) == list: #pylint: disable=unidiomatic-typecheck
                        ifacelist = neighbors[vrf][nbr]
                    else:
                        ifacelist = [neighbors[vrf][nbr]]
                    for iface in ifacelist:
                        entry = OspfNbr()
                        entry.peer_id = nbr
                        entry.peer_addr = iface['ifaceAddress']
                        entry.state = iface['nbrState']
                        entry.ifname = iface['ifaceName'].split(':')[0]
                        entry.area = iface['areaId'].split(':')[0]
                        entry.vrf = vrf
                        # Can't really check status with this output, need to switch
                        entry.save()
                        if fullstate is True:
                            ospf_arr.append(entry)
                except Exception:
                    pass

        if fullstate is True:
            save_fullstate(ospf_arr, OspfNbr().__class__.__name__)

        return retcode

    def get_nbr_vrf(self, nbr, neighbors):
        for vrf in neighbors:
            if nbr in neighbors[vrf]:
                return vrf
        return None

