#
# Copyright (c) 2020-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#
import re
import subprocess
from os import path

import ujson as json
from netq_agent.parsers.common.render import BaseRender
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate, subprocess_wrapper
from netq_lib.orm.redisdb.models import CPU, Disk, Memory, Inventory
from netq_lib.common import configuration
from netq_lib.common.enums import ConfKw

OK = 0
MISSING = 1
OLD_SVC_STATUS = UNINITIALIZED = -1
LSBLK_REGEX = re.compile(
    r"NAME=\"(.*)\" SIZE=\"(.*)\" TYPE=\"(.*)\" VENDOR=\"(.*)\" "
    r"TRAN=\"(.*)\" REV=\"(.*)\" MODEL=\"(.*)\"\n?")

dummy_inventory = [
    {
        "platform_vendor": "A Systems",
        "platform_model": "A_VX",
        "asic_vendor": "AA Inc",
        "asic_model": "A-Z",
        "disk_total_size": "10 GB",
        "cpu_model": "A-SUBLEQ",
        "license_state": "ok",
    },
    {
        "platform_vendor": "B Systems",
        "platform_model": "B_VX",
        "asic_vendor": "BB Inc",
        "asic_model": "B-Z",
        "disk_total_size": "20 GB",
        "cpu_model": "B-SGI",
        "license_state": "bad",
    },
    {
        "platform_vendor": "C Systems",
        "platform_model": "C_VX",
        "asic_vendor": "CC Inc",
        "asic_model": "C-Z",
        "disk_total_size": "30 GB",
        "cpu_model": "C-CRAY",
        "license_state": "N/A",
    },
    {
        "platform_vendor": "D Systems",
        "platform_model": "D_VX",
        "asic_vendor": "DD Inc",
        "asic_model": "D-Z",
        "disk_total_size": "33 GB",
        "cpu_model": "D-CRAY",
        "license_state": "missing",
    },
    {
        "platform_vendor": "E Systems",
        "platform_model": "E_VX",
        "asic_vendor": "EE Inc",
        "asic_model": "E-Z",
        "disk_total_size": "39 GB",
        "cpu_model": "pentium",
        "license_state": "ok",
    },
    {
        "platform_vendor": "F Systems",
        "platform_model": "F_VX",
        "asic_vendor": "FF Inc",
        "asic_model": "F-Z",
        "disk_total_size": "10 GB",
        "cpu_model": "F-CRAY",
        "license_state": "bad",
    },
    {
        "platform_vendor": "G Systems",
        "platform_model": "G_VX",
        "asic_vendor": "GG Inc",
        "asic_model": "G-Z",
        "disk_total_size": "15 GB",
        "cpu_model": "G-CRAY",
        "license_state": "N/A",
    },
    {
        "platform_vendor": "H Systems",
        "platform_model": "H_VX",
        "asic_vendor": "HH Inc",
        "asic_model": "H-Z",
        "disk_total_size": "10 GB",
        "cpu_model": "H-CRAY",
        "license_state": "missing",
    }
]


class PlatformRender(BaseRender):
    def __init__(self): #pylint: disable=useless-super-delegation
        super(PlatformRender, self).__init__()

    def process_os_info(self, logger, os_release):
        raise NotImplementedError

    def process_asic_info(self, logger, board, platform_json, platform_detect):
        raise NotImplementedError

    def process_os_license_info(self, cmd_output, svc_status=OK, **kwargs):
        raise NotImplementedError

    def process_cpu_info(self, logger, lscpu):
        cpu = CPU()

        cpu.arch = 'N/A'
        cpu.nos = 'N/A'
        cpu.model = 'N/A'
        cpu.max_freq = 'N/A'
        cpu.mem_total = 'N/A'

        try:
            if lscpu is not None:
                for line in lscpu.splitlines():
                    if line.startswith('Architecture'):
                        cpu.arch = line.split(':')[1].strip()
                    if line.startswith('CPU(s)'):
                        cpu.nos = line.split(':')[1].strip()
                    if line.startswith('Model name'):
                        cpu_info = line.split(':')[1].strip()
                        model_freq = cpu_info.split('@')
                        if len(model_freq) == 1:
                            cpu.model = model_freq[0].replace('CPU', '').strip().replace('  ', ' ')
                        elif len(model_freq) == 2:
                            cpu.model = model_freq[0].replace('CPU', '').strip().replace('  ', ' ')
                            cpu.max_freq = model_freq[1].strip()
        except (IndexError, ValueError) as ex:
            logger.error("Failed to process lscpu output: {}".format(ex))
            raise ex

        return cpu

    def process_disk_info(self, logger, lsblk):
        """
        NAME="fd0" SIZE="4K" TYPE="disk" VENDOR="" TRAN="" REV="" MODEL=""
        NAME="sda" SIZE="28G" TYPE="disk" VENDOR="ATA     " TRAN="ata" REV="2.5+" MODEL="QEMU HARDDISK   "
        NAME="sr0" SIZE="952K" TYPE="rom" VENDOR="QEMU    " TRAN="ata" REV="2.5+" MODEL="QEMU DVD-ROM    "
        """
        all_disk = []
        subtotal_disk_size = 0.0

        try:
            for name, size, d_type, vendor, tran, rev, model in LSBLK_REGEX.findall(lsblk):
                disk = Disk()
                disk.name = name or 'N/A'
                disk.size = size or 'N/A'
                disk.d_type = d_type or 'N/A'
                disk.vendor = vendor or 'N/A'
                disk.transport = tran or 'N/A'
                disk.rev = rev or 'N/A'
                disk.model = model or 'N/A'
                try:
                    if disk.size.endswith('M'):
                        gb_size = float(disk.size.split('M')[0]) / 1024.0
                        disk.size = ('%0.2fG' % gb_size)
                        subtotal_disk_size += gb_size
                    elif disk.size.endswith('K'):
                        subtotal_disk_size += float(disk.size.split('K')[0]) / 1048576.0
                    elif disk.size.endswith('G'):
                        subtotal_disk_size += float(disk.size.split('G')[0])
                except ValueError:
                    pass

                disk.save()
                all_disk.append({'name': disk.name, 'size': disk.size, 'd_type': disk.d_type,
                                 'vendor': disk.vendor, 'transport': disk.transport, 'rev': disk.rev,
                                 'model': disk.model})
        except Exception as ex:
            logger.error("Failed to process lsblk output: {}".format(ex))
            raise ex

        return all_disk, subtotal_disk_size


    def process_mem_info(self, logger, meminfo, meminfo_dmi):
        mem_size = []
        mem_locator = []
        mem_b_locator = []
        mem_m_type = []
        mem_serial = []
        mem_mfg = []
        mem_speed = []
        try:
            if meminfo_dmi is not None:
                for line in meminfo_dmi.splitlines():
                    line = line.strip()
                    if line.startswith('Size:'):
                        _, val = line.split(': ')
                        mem_size.append(val)
                    elif line.startswith('Bank Locator:'):
                        _, val = line.split(': ')
                        mem_b_locator.append(val)
                    elif line.startswith('Locator:'):
                        _, val = line.split(': ')
                        mem_locator.append(val)
                    elif line.startswith('Speed:'):
                        _, val = line.split(': ')
                        mem_speed.append(val)
                    elif line.startswith('Type:'):
                        _, val = line.split(': ')
                        mem_m_type.append(val)
                    elif line.startswith('Serial Number:'):
                        try:
                            _, val = line.split(': ')
                        except Exception:
                            val = 'unknown'
                        mem_serial.append(val)
                    elif line.startswith('Manufacturer:'):
                        try:
                            _, val = line.split(': ')
                        except Exception:
                            val = 'unknown'
                        mem_mfg.append(val)
            if (len(mem_size) == 0) and (meminfo is not None): #pylint: disable=len-as-condition
                for line in meminfo.splitlines():
                    if line.startswith('MemTotal'):
                        mem_str = float(line.split(': ')[1].strip().split()[0])
                        mem_size.append('%.2f' % (mem_str / 1024) + 'MB')
                        mem_locator.append('N/A')
                        mem_b_locator.append('')
                        mem_m_type.append('N/A')
                        mem_mfg.append('N/A')
                        mem_serial.append('N/A')
                        mem_speed.append('N/A')
                        break

            all_mem = []
            subtotal_mem_size = 0.0
            for i in range(0, len(mem_size)): #pylint: disable=consider-using-enumerate
                if 'no' in mem_size[i].lower():
                    continue
                mem = Memory()
                mem.size = mem_size[i]
                if 'MB' in mem.size:
                    subtotal_mem_size += float(mem.size.strip(' MB'))
                elif 'GB' in mem.size:
                    subtotal_mem_size += float(mem.size.strip(' GB')) * 1024
                mem.name = mem_locator[i]
                if 'no' not in mem_b_locator[i].lower():
                    mem.name += mem_b_locator[i]
                mem.m_type = mem_m_type[i]
                if len(mem_mfg) > i:
                    mem.vendor = mem_mfg[i]
                    if 'bad' in mem.vendor.lower():
                        mem.vendor = 'N/A'
                if len(mem_serial) > i:
                    mem.serial_number = mem_serial[i]
                    if 'bad' in mem.serial_number.lower():
                        mem.serial_number = 'N/A'
                if len(mem_speed) > i:
                    mem.speed = mem_speed[i]
                mem.save()
                all_mem.append({'size': mem.size, 'name': mem.name, 'm_type': mem.m_type, 'vendor': mem.vendor,
                                'serial_number': mem.serial_number, 'speed': mem.speed})
        except (IndexError, ValueError) as ex:
            logger.error("Failed to process mem info: {}".format(ex))
            raise ex

        return all_mem, subtotal_mem_size

    def process_app_release_info(self, logger, app_release_info):
        appliance_version = 'N/A'
        manifest_hash = 'N/A'

        try:
            if app_release_info:
                for line in app_release_info.splitlines():
                    if line.startswith('APPLIANCE_VERSION'):
                        _, appliance_version = line.strip().split('=')
                    if line.startswith('APPLIANCE_MANIFEST_HASH'):
                        _, manifest_hash = line.strip().split('=')
        except ValueError as ex:
            logger.error("Failed to process app release info: {}".format(ex))
            raise ex

        return appliance_version, manifest_hash

    def _inventory_data_valid(self, os, board, asic):
        return True     # default True, in sonic override

    def process_platform_info(self, logger, random_platform, platform, os_release,
                              decode_syseeprom, lscpu, meminfo, platform_json, meminfo_dmi,
                              lsblk, agent_version, platform_detect, is_full_state,
                              app_release_info, service_nvue_enabled, service_nvue_active, is_nvue_config_saved,
                              service_nclu_enabled, opta_on_switch_enabled=None, lcm_executor_on_switch_enabled=None):

        inventory = Inventory()
        os = self.process_os_info(logger, os_release)

        all_cpu = []
        cpu = self.process_cpu_info(logger, lscpu)

        board = self.process_syseeprom_info(logger, decode_syseeprom)

        asic = self.process_asic_info(logger, board, platform_json, platform_detect)

        all_disk, subtotal_disk_size = self.process_disk_info(logger, lsblk)

        all_mem, subtotal_mem_size = self.process_mem_info(logger, meminfo, meminfo_dmi)

        appliance_version, manifest_hash = self.process_app_release_info(logger, app_release_info)

        license_obj = None
        license_dict = dict()
        global OLD_SVC_STATUS #pylint: disable=global-statement
        if OLD_SVC_STATUS == UNINITIALIZED:
            RC = OK
            try:
                subprocess_wrapper('/usr/sbin/switchd -lic'.split())
            except subprocess.CalledProcessError as ex:
                RC = ex.returncode
            except OSError as ex:
                RC = ex.errno
            finally:
                license_obj, license_dict = self.process_os_license_info(cmd_output=None, svc_failed=RC, platform=platform)
                OLD_SVC_STATUS = RC

        if not self._inventory_data_valid(os, board, asic):
            return ServiceStatus.FAIL

        if asic:
            inventory.asic_vendor = asic.vendor
            inventory.asic_model = asic.model
            inventory.asic_model_id = asic.model_id
            inventory.asic_core_bw = asic.core_bw
            inventory.asic_ports = asic.ports
            inventory.asic_data = json.dumps([]) #pylint: disable=no-member
        inventory.os_name = os.name
        inventory.os_version = os.version
        inventory.os_version_id = os.version_id
        inventory.cpu_arch = cpu.arch
        inventory.cpu_model = cpu.model
        inventory.cpu_nos = cpu.nos
        inventory.cpu_max_freq = cpu.max_freq
        inventory.cpu_data = json.dumps(all_cpu) #pylint: disable=no-member
        if board:
            inventory.platform_base_mac = board.base_mac
            inventory.platform_part_number = board.part_number
            inventory.platform_mfg_date = board.mfg_date
            inventory.platform_serial_number = board.serial_number
            inventory.platform_label_revision = board.label_revision
            inventory.platform_vendor = board.vendor
            inventory.platform_model = board.model

        inventory.agent_version = agent_version
        inventory.memory_data = json.dumps(all_mem) #pylint: disable=no-member
        inventory.memory_total_size = '{0:.2f} MB'.format(subtotal_mem_size)
        inventory.disk_data = json.dumps(all_disk) #pylint: disable=no-member
        inventory.disk_total_size = '{0:.2f} GB'.format(subtotal_disk_size)
        if license_obj and license_dict:
            inventory.license_state = license_obj.state
            inventory.license_data = json.dumps([license_dict]) #pylint: disable=no-member

        # do we need to randomize data (for QA)
        if random_platform:
            # NETQ-2251: random_platform is provided by test config.Currently, its range 1-8
            random_index = random_platform - 1
            dummy_data = dummy_inventory[random_index]
            inventory.platform_model = dummy_data['platform_model']
            inventory.platform_vendor = dummy_data['platform_vendor']
            inventory.asic_vendor = dummy_data['asic_vendor']
            inventory.asic_model = dummy_data['asic_model']
            inventory.disk_total_size = dummy_data['disk_total_size']
            inventory.cpu_model = dummy_data['cpu_model']
            inventory.license_state = dummy_data['license_state']

        #NVUE enabled
        if not service_nvue_enabled:
            service_nvue_enabled = "disabled"
        if not service_nvue_active:
            service_nvue_active = "inactive"
        inventory.nvue_enabled = service_nvue_enabled is not None and service_nvue_enabled.strip() == 'enabled'
        inventory.nvue_active = service_nvue_active is not None and service_nvue_active.strip() == 'active'
        inventory.is_nvue_config_saved = is_nvue_config_saved is not None and is_nvue_config_saved.strip() == 'true'
        if service_nclu_enabled is not None:
            inventory.nclu_enabled = "net could not connect to netd" not in service_nclu_enabled
        else:
            inventory.nclu_enabled = False

        inventory.appliance_version = appliance_version
        inventory.manifest_hash = manifest_hash

        #opta on switch flag
        if (opta_on_switch_enabled is not None and opta_on_switch_enabled.strip() == 'active' and lcm_executor_on_switch_enabled is not None and lcm_executor_on_switch_enabled.strip() == 'active'):
            inventory.opta_on_switch_enabled = "Both"
        elif (opta_on_switch_enabled is not None and opta_on_switch_enabled.strip() == 'active' and lcm_executor_on_switch_enabled is not None and lcm_executor_on_switch_enabled.strip() != 'active'):
            inventory.opta_on_switch_enabled = "Opta"
        else:
            inventory.opta_on_switch_enabled = "N/A"



        os.save()
        cpu.save()
        if asic:
            asic.save()
        if board:
            board.save()

        # In Opta, during init avoid sending inventory if manifest_hash is 'N/A'
        if path.isfile('/etc/app-release'):  # opta node
            if inventory.manifest_hash == 'N/A':
                logger.debug("skipping inventory data")
                return ServiceStatus.OK

        conf_dict = configuration.get_config('platform')
        inband_interface = conf_dict[ConfKw.NETQ_AGENT].get(ConfKw.INBAND_INTERFACE, ConfKw.DEFAULT_INBAND_INTERFACE)
        inventory.inband_interface=inband_interface
        if is_full_state:
            save_fullstate([inventory], inventory.__class__.__name__)
        else:
            inventory.save()

        return ServiceStatus.OK
