#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import subprocess
from netq_lib.orm.redisdb.models import Port
from netq_lib.common.utils import save_fullstate, subprocess_wrapper
from netq_lib.common.enums import Platform
from netq_agent.parsers.common.render import BaseRender


class PortsRender(BaseRender):
    def __init__(self):
        super(PortsRender, self).__init__()
        self.model = Port
        self.local_port_dict = dict()
        self._logger = None
        self._ethtool_bin_path = self._get_ethtool_path()

    @staticmethod
    def _get_ethtool_path():
        raise NotImplementedError

    def ethtool_fetch_port_info(self, port_name):
        ethtool_dict = dict()
        try:
            ethtool_dump = subprocess_wrapper([self._ethtool_bin_path, port_name])
            for line in ethtool_dump.splitlines():
                if ':' not in line:
                    continue
                key, value = line.split(':', 1)
                key = key.strip()
                value = value.strip()
                if key in list(ethtool_dict.keys()):
                    ethtool_dict[key] = ethtool_dict[key] + '| ' + value
                else:
                    ethtool_dict[key] = value
        except subprocess.CalledProcessError:
            pass

        if self.is_cl_platform:
            try:
                ethtool_dump = subprocess_wrapper([self._ethtool_bin_path, '--show-fec', port_name])

                for line in ethtool_dump.splitlines():
                    if ":" not in line:
                        continue
                    key, value = line.split(':', 1)
                    key = key.strip()
                    value = value.strip()
                    if key in ethtool_dict.keys():
                        ethtool_dict[key] = ethtool_dict[key] + '| ' + value
                    else:
                        ethtool_dict[key] = value
            except subprocess.CalledProcessError:
                pass

        return ethtool_dict

    def ethtool_fetch_test_serial_number(self, port_name):
        raise NotImplementedError

    def ethtool_fetch_eeprom_optical_info(self, port_name):
        ethtool_m_dict = dict()
        try:
            ethtool_m_dump = subprocess_wrapper(
                [self._ethtool_bin_path, '-m', port_name])
            for line in ethtool_m_dump.splitlines():
                if ':' not in line:
                    continue
                key, value = line.split(':', 1)
                key = key.strip()
                value = value.strip()
                if key.startswith('Length'):
                    key = 'length'
                value = value.strip()
                if key == 'length' and value == '0m' or value == '0km':
                    continue
                if key in list(ethtool_m_dict.keys()):
                    ethtool_m_dict[key] = ethtool_m_dict[key] + '| ' + value
                else:
                    ethtool_m_dict[key] = value
        except subprocess.CalledProcessError:
            pass

        return ethtool_m_dict

    def skip_port_platform_specific(self, port_name):
        return False

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        port_names = kwargs.get('ports', None)
        platform = kwargs.get('platform', '')
        optical_ports = kwargs.get('optical_ports', None)
        self._logger = kwargs.get('logger', None)
        last_ports = set(self.local_port_dict.keys())
        curr_ports = port_names
        del_ports = last_ports - curr_ports
        self.is_cl_platform = False
        if Platform.VX in platform or Platform.CUMULUS in platform:
            self.is_cl_platform = True
            # handle ports disappear, e.g., split ports
        if del_ports:
            for dport_name in del_ports:
                port = Port()
                port.ifname = dport_name
                port.delete()
                del self.local_port_dict[dport_name]

        if not port_names:
            return

        fullstate_arr = []
        for port_name in port_names:
            # On some platforms, due to slow I2C reads or related issues, we are
            # to read the port EEPROMs leading us to reach the false conclusion
            # that the port is empty. To avoid such issues, we save data only if
            # see the same data twice. "changes" keep track of that. Only if there
            # is no change do we try to "save" data.
            change = False
            # ignore PIM interfaces and 'swid0_eth' seen on MLX boxes.
            if (port_name.startswith('pimreg') or port_name == 'swid0_eth' or
                    port_name.startswith('usb')):
                continue

            if self.skip_port_platform_specific(port_name):
                continue

            if port_name not in list(self.local_port_dict.keys()):
                self.local_port_dict[port_name] = dict()
                self.local_port_dict[port_name]['identifier'] = 'n/a'
                self.local_port_dict[port_name]['test_serial_number'] = 'n/a'
                self.local_port_dict[port_name]['vendor_name'] = 'n/a'
                self.local_port_dict[port_name]['serial_number'] = 'n/a'
                self.local_port_dict[port_name]['part_number'] = 'n/a'
                self.local_port_dict[port_name]['transreceiver'] = 'n/a'
                self.local_port_dict[port_name]['connector'] = 'n/a'
                self.local_port_dict[port_name]['length'] = 'n/a'

            ethtool_dict = self.ethtool_fetch_port_info(port_name)

            port = Port()
            port.ifname = port_name
            eth_port = ethtool_dict.get('Port', 'n/a')
            eth_supp_port = ethtool_dict.get('Supported ports', 'n/a')
            identifier = 'empty'
            if eth_port.startswith('Twisted') and 'tp' in eth_supp_port.lower():
                identifier = 'RJ45'
            elif 'vx' in platform:
                identifier = 'RJ45'
            elif 'backplane' in eth_supp_port.lower():
                identifier = 'backplane'

            speed = ethtool_dict.get('Speed', 'n/a').strip('!')
            if 'Mb/s' in speed:
                try:
                    speed = '%dG' % (int(speed.split('M', 1)[0]) / 1000)
                except ValueError:
                    pass
            port.speed = speed

            port.autoneg = ethtool_dict.get('Auto-negotiation', 'n/a')
            port.state = ethtool_dict.get('Link detected', 'n/a')
            if port.state == 'yes':
                port.state = 'up'
            elif port.state == 'no':
                port.state = 'down'
            port.fec = ethtool_dict.get('FEC encodings', 'n/a')
            port.supported_fec = ethtool_dict.get('Supported FEC modes', 'n/a')
            port.advertised_fec = ethtool_dict.get('Advertised FEC modes', 'n/a')

            test_serial_number = self.ethtool_fetch_test_serial_number(port_name)
            if test_serial_number and test_serial_number != self.local_port_dict[port_name]['test_serial_number']:
                self.local_port_dict[port_name]['test_serial_number'] = test_serial_number
                change = True
                if not test_serial_number:
                    self.local_port_dict[port_name]['identifier'] = identifier
                    self.local_port_dict[port_name]['vendor_name'] = 'n/a'
                    self.local_port_dict[port_name]['serial_number'] = 'n/a'
                    self.local_port_dict[port_name]['part_number'] = 'n/a'
                    self.local_port_dict[port_name]['transreceiver'] = 'n/a'
                    self.local_port_dict[port_name]['connector'] = 'n/a'
                    self.local_port_dict[port_name]['length'] = 'n/a'
                else:
                    ethtool_m_dict = self.ethtool_fetch_eeprom_optical_info(port_name)
                    if self._is_optical_port(ethtool_m_dict):
                        optical_ports.add(port_name)

                    ident = ethtool_m_dict.get('Identifier', 'n/a').strip()
                    if ('no' in ident.lower() or
                            (identifier != 'RJ45' and ident == 'n/a')):
                        ident = 'empty'
                    if '(' in ident:
                        ident = ident.replace('(', '').replace(')', '')
                        p_id, p_type = ident.strip().split(' ', 1)
                        ident = p_type
                    self.local_port_dict[port_name]['identifier'] = ident
                    self.local_port_dict[port_name]['serial_number'] = ethtool_m_dict.get('Vendor SN', 'n/a')
                    self.local_port_dict[port_name]['vendor_name'] = ethtool_m_dict.get('Vendor name', 'n/a')
                    self.local_port_dict[port_name]['part_number'] = ethtool_m_dict.get('Vendor PN', 'n/a')
                    self.local_port_dict[port_name]['length'] = ethtool_m_dict.get('length', 'n/a').replace('|', ',')
                    trans = ethtool_m_dict.get('Transceiver type', 'n/a')
                    trans_type = set()
                    for x in trans.strip().split('|'):
                        if 'FC:' in x:
                            if 'MBytes/sec' not in x:
                                y = x.split(':')[1].strip()
                                trans_type.add(y.strip())
                        elif 'Ethernet:' in x:
                            y = x.split(':')[1].strip()
                            y = y.replace('BASE-', 'Base-')
                            trans_type.add(y.strip())
                        elif ('SONET' in x or 'Infiniband' in x or
                              'short reach' in x or 'M5' in x):
                            continue
                        else:
                            blacklist = ['SONET', 'Infiniband', 'short reach',
                                         'M5', 'OM3']
                            found = False
                            for b in blacklist:
                                if b in x:
                                    found = True
                            if not found:
                                trans_type.add(x.strip())
                    self.local_port_dict[port_name]['transreceiver'] = ','.join(
                        sorted(trans_type))

                    conn = ethtool_m_dict.get('Connector', 'n/a').strip()
                    if ' ' in conn:
                        conn = conn.split(' ', 1)[1].replace(
                            '(', '').replace(')', '')
                        if conn == 'No separable connector':
                            conn = 'n/a'
                        elif conn == 'MPO Parallel Optic':
                            conn = 'MPO Parallel'

                    self.local_port_dict[port_name]['connector'] = conn

            port.identifier = self.local_port_dict[port_name]['identifier']
            port.vendor_name = self.local_port_dict[port_name]['vendor_name']
            port.serial_number = self.local_port_dict[port_name]['serial_number']
            port.part_number = self.local_port_dict[port_name]['part_number']
            port.transreceiver = self.local_port_dict[port_name]['transreceiver']
            port.connector = self.local_port_dict[port_name]['connector']
            port.length = self.local_port_dict[port_name]['length']

            fullstate_arr.append(port)
            if not change:
                if not fullstate:
                    port.save()
        if fullstate:
            save_fullstate(fullstate_arr, Port().__class__.__name__,
                           should_send_gnmi=True)
