#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import subprocess
import time
import sys
import os
import yaml
import eventlet
import eventlet.pools
import greenlet

from netq_lib.orm.redisdb.models import ProcNetDevInterface, EthtoolInterface
from netq_lib.orm import HOSTNAME
from netq_agent.parsers.common.render import BaseRender, ReturnGroup


global_gthread_dict = {}
global_sub_process = {}
__CMD_TIMEOUT = 20
eth_dict = {}
ETHTOOL_TEST_DATA_PATH = '/etc/netq/.test_ethtool.yml'
if sys.version_info > (3,):
    long = int

class ProcRender(BaseRender):
    def __init__(self):
        super(ProcRender, self).__init__()
        self.test_data = self.read_test_data()
        self.interface_ethtool_op = {}
        self.procnetdev = {}
        self._logger = None
        # eth_dict stores a mapping of ethtool names to python-friendly variable names
        # {'HwIfInOctets': 'hw_if_in_octets', 'SoftInDrops': 'soft_in_drops', ...}
        # create_ethtool() is invoked only if the global eth_dict is empty
        global eth_dict #pylint: disable=global-statement
        if not eth_dict:
            eth_dict = self.create_ethtool_dict()

    def get_interfaces(self, **kwargs):
        raise NotImplementedError

    @staticmethod
    def read_test_data():
        test_data = {}
        if os.path.exists(ETHTOOL_TEST_DATA_PATH):
            try:
                with open(ETHTOOL_TEST_DATA_PATH, 'r') as f:
                    test_data = yaml.safe_load(f)
            except Exception as e:
                pass
        return test_data

    def stop_command_checker(self, green_thread, port_name, cmd):
        try:
            (out, err) = green_thread.wait()
            if cmd.returncode != 0:
                self.interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
            else:
                self.interface_ethtool_op[port_name] = ReturnGroup(success=True, op=out)
            return
        except greenlet.GreenletExit:
            self.interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
        except Exception:  # pylint: disable=broad-except
            self.interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
            eventlet.kill(None, *sys.exc_info())

    def run_command_safe(self, command, thread_pool, port_name):
        """Run time-limited command and update the proc dict self.interface_ethtool_op
        with the ethtool status and response
        """
        global global_gthread_dict #pylint: disable=global-statement
        global global_sub_process #pylint: disable=global-statement
        try:
            cmd = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        except (subprocess.CalledProcessError, OSError):
            self.interface_ethtool_op[port_name] = ReturnGroup(success=False, op="")
            return

        global_sub_process[port_name] = cmd
        green_thread = thread_pool.spawn(cmd.communicate)
        green_thread.link(self.stop_command_checker, port_name, cmd)
        global_gthread_dict[port_name] = green_thread

    def create_proc(self, ifname, rx_bytes=-1, rx_drop=-1, rx_errs=-1, rx_frame=-1, rx_multicast=-1, rx_packets=-1, tx_bytes=-1,
                    tx_carrier=-1, tx_colls=-1, tx_drop=-1, tx_errs=-1, tx_packets=-1, rx_broadcast=-1, tx_broadcast=-1, tx_multicast=-1):
        # fifo and compressed related counters are intentionally left out
        iface = ProcNetDevInterface()
        iface.ifname = ifname
        iface.rx_bytes = rx_bytes
        iface.rx_drop = rx_drop
        iface.rx_errs = rx_errs
        iface.rx_frame = rx_frame
        iface.rx_multicast = rx_multicast
        iface.rx_packets = rx_packets
        iface.tx_bytes = tx_bytes
        iface.tx_carrier = tx_carrier
        iface.tx_colls = tx_colls
        iface.tx_drop = tx_drop
        iface.tx_errs = tx_errs
        iface.tx_packets = tx_packets
        iface.rx_broadcast = rx_broadcast
        iface.tx_broadcast = tx_broadcast
        iface.tx_multicast = tx_multicast
        return iface

    def get_command(self, ifname):
        raise NotImplementedError

    def assign_procnetdev_value(self, ifname, procnetdev):
        # Update procnetdev dict with /proc/net/dev values only if it is empty
        if not procnetdev:
            with open("/proc/net/dev", "r") as f:
                txt = f.readlines()
            for t in txt:
                if ":" not in t:
                    continue
                k, v = t.split(":")
                try:
                    v = list(map(int, v.split()))
                except Exception:
                    pass
                procnetdev[k.strip()] = v
        proc_obj = None
        if ifname in procnetdev:
            # Assign procnetdev dictionary values for the interface to the ProcNetDevInterface object
            [rx_bytes, rx_packets, rx_errs, rx_drop, _, rx_frame, _, rx_multicast, tx_bytes, tx_packets, tx_errs, tx_drop, _,
             tx_colls, tx_carrier, _] = procnetdev[ifname]
            proc_obj = self.create_proc(ifname, rx_bytes=rx_bytes, rx_drop=rx_drop, rx_errs=rx_errs, rx_frame=rx_frame,
                                        rx_multicast=rx_multicast, rx_packets=rx_packets, tx_bytes=tx_bytes, tx_carrier=tx_carrier,
                                        tx_colls=tx_colls, tx_drop=tx_drop, tx_errs=tx_errs, tx_packets=tx_packets)
        else:
            self._logger.debug("Interface {} not found in procnetdev content".format(ifname))
        return proc_obj

    @staticmethod
    def create_ethtool_dict():
        # Returns a dictionary which has ethtool names mapped to python-friendly variable names
        # output = {'HwIfInOctets': 'hw_if_in_octets', 'HwIfInUcastPkts': 'hw_if_in_ucast_pkts', ...}
        import re
        k = ['HwIfInOctets', 'HwIfInUcastPkts', 'HwIfInBcastPkts', 'HwIfInMcastPkts', 'HwIfOutOctets', 'HwIfOutUcastPkts',
             'HwIfOutMcastPkts', 'HwIfOutBcastPkts', 'HwIfInDiscards', 'HwIfInL3Drops', 'HwIfInBufferDrops',
             'HwIfInAclDrops', 'HwIfInDot3LengthErrors', 'HwIfInErrors', 'SoftInErrors', 'SoftInDrops',
             'SoftInFrameErrors', 'HwIfOutDiscards', 'HwIfOutErrors', 'HwIfOutQDrops', 'HwIfOutNonQDrops',
             'SoftOutErrors', 'SoftOutDrops', 'SoftOutTxFifoFull', 'HwIfOutQLen', 'HwIfInDot3FrameErrors',
             'HwIfAlignmentErrors', 'HwIfSymbolErrors', 'HwIfUndersizeErrors', 'HwIfOversizeErrors', 'HwIfJabberErrors',
             'HwIfInPausePkt', 'HwIfOutPausePkt', 'HwIfInPfc0Pkt', 'HwIfOutPfc0Pkt', 'HwIfInPfc1Pkt', 'HwIfOutPfc1Pkt',
             'HwIfInPfc2Pkt', 'HwIfOutPfc2Pkt', 'HwIfInPfc3Pkt', 'HwIfOutPfc3Pkt', 'HwIfInPfc4Pkt', 'HwIfOutPfc4Pkt',
             'HwIfInPfc5Pkt', 'HwIfOutPfc5Pkt', 'HwIfInPfc6Pkt', 'HwIfOutPfc6Pkt', 'HwIfInPfc7Pkt', 'HwIfOutPfc7Pkt',
             'HwIfOutWredDrops', 'HwIfOutQ0WredDrops', 'HwIfOutQ1WredDrops', 'HwIfOutQ2WredDrops', 'HwIfOutQ3WredDrops',
             'HwIfOutQ4WredDrops', 'HwIfOutQ5WredDrops', 'HwIfOutQ6WredDrops', 'HwIfOutQ7WredDrops', 'HwIfOutQ8WredDrops',
             'HwIfOutQ9WredDrops']
        v = [re.sub('([A-Z]{1})', r'_\1', i).lower()[1:] for i in k]
        return dict(list(zip(k, v)))

    def create_eth(self, ifname, ethElems, test_data=None):
        global eth_dict #pylint: disable=global-statement

        eth_obj = EthtoolInterface()
        eth_obj.ifname = ifname
        eth_obj_var_list = list(eth_dict.values())

        for i in list(ethElems.keys()):
            if i.startswith("NIC"):
                continue
            if i in  eth_dict:
                eth_obj_var = eth_dict[i]
                count = ethElems[i] if test_data is None or self.test_data.get(i) is None else test_data.get(i)
                setattr(eth_obj, eth_obj_var, count)
                eth_obj_var_list.remove(eth_obj_var)

        for i in eth_obj_var_list:
            setattr(eth_obj, i, -1)

        return eth_obj

    def process_stats(self, ifname, retcode, cmd_out):
        if retcode == 0:
            ethElems = {}
            for ethtoolstat in cmd_out.op.splitlines():
                ethtoolstat = ethtoolstat.decode()
                if ":" not in ethtoolstat:
                    continue

                key, value = ethtoolstat.split(':', 1)
                key = key.strip()
                value = value.strip()
                ethElems[key] = value

            proc_list = ['HwIfInOctets', 'HwIfInUcastPkts', 'HwIfInBcastPkts', 'HwIfInMcastPkts', 'HwIfOutOctets',
                         'HwIfOutUcastPkts', 'HwIfOutMcastPkts', 'HwIfOutBcastPkts', 'HwIfInDiscards',
                         'HwIfInDot3LengthErrors', 'HwIfInErrors', 'HwIfOutDiscards', 'HwIfOutErrors']

            if all(i in ethElems for i in proc_list):
                eth_obj = self.create_eth(ifname, ethElems, self.test_data.get(ifname))
                eth_obj.save()
                proc_obj = self.create_proc(ifname,
                                            rx_bytes=ethElems['HwIfInOctets'],
                                            rx_drop=ethElems['HwIfInDiscards'],
                                            rx_errs=long(ethElems['HwIfInDot3LengthErrors']) + long(ethElems['HwIfInErrors']),
                                            rx_frame=long(ethElems['HwIfInUcastPkts']) + long(ethElems['HwIfInBcastPkts']) + long(
                                                ethElems['HwIfInMcastPkts']),
                                            rx_multicast=ethElems['HwIfInMcastPkts'],
                                            rx_packets=long(ethElems['HwIfInUcastPkts']) + long(ethElems['HwIfInBcastPkts']) + long(
                                                ethElems['HwIfInMcastPkts']),
                                            tx_bytes=ethElems['HwIfOutOctets'],
                                            tx_drop=ethElems['HwIfOutDiscards'],
                                            tx_errs=ethElems['HwIfOutErrors'],
                                            tx_packets=long(ethElems['HwIfOutUcastPkts']) + long(ethElems['HwIfOutMcastPkts']) + long(
                                                ethElems['HwIfOutBcastPkts']),
                                            rx_broadcast=ethElems['HwIfInBcastPkts'],
                                            tx_broadcast=ethElems['HwIfOutBcastPkts'],
                                            tx_multicast=ethElems['HwIfOutMcastPkts'])
            else:
                proc_obj = self.assign_procnetdev_value(ifname, self.procnetdev)
        else:
            proc_obj = self.assign_procnetdev_value(ifname, self.procnetdev)

        if proc_obj:
            proc_obj.save()

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        """
        Parse /proc/net/dev counters
        """
        global __CMD_TIMEOUT #pylint: disable=global-statement
        global global_gthread_dict #pylint: disable=global-statement
        global global_sub_process #pylint: disable=global-statement
        global eth_dict #pylint: disable=global-statement
        self._logger = kwargs.get('logger')
        self.interface_ethtool_op = {}
        __CMD_TIMEOUT = 20
        global_gthread_dict = {}
        global_sub_process = {}

        thread_pool = kwargs.get('pool', '*')
        if_names = self.get_interfaces(**kwargs)

        m_proc = ProcNetDevInterface.memmodel()
        old_proc_objs = []
        if m_proc:
            old_proc_objs = (
                list(m_proc.query.filter(hostname=HOSTNAME))
            )

        for old_proc_obj in old_proc_objs:
            if old_proc_obj.ifname not in if_names:
                old_proc_obj.delete()

        skipped_ifnames = []
        # Run ethtool command for each interface in a separate thread
        for ifname in if_names:
            command = self.get_command(ifname)
            if not command:
                skipped_ifnames.append(ifname)
                continue
            self.run_command_safe(command, thread_pool, ifname)

        for ifname in if_names:
            if ifname in skipped_ifnames:
                continue
            endtime = time.time() + int(__CMD_TIMEOUT)

            while True:
                cmd_out = ''
                if ifname in self.interface_ethtool_op:
                    cmd_out = self.interface_ethtool_op.pop(ifname)
                    retcode = 0 if cmd_out.success else -1
                    break

                if time.time() > endtime:
                    retcode = -1
                    try:
                        global_sub_process[ifname].terminate()
                        eventlet.sleep(0.3)
                        global_sub_process[ifname].wait()
                        eventlet.kill(global_gthread_dict[ifname], greenlet.GreenletExit)
                    except Exception as e:
                        break
                    break
                eventlet.sleep(0.3)

            self.process_stats(ifname, retcode, cmd_out)
