#
# Copyright (c) 2021-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#
import os
from subprocess import CalledProcessError
import yaml
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Roce, RoceConfig, NvueRoceConfig
from netq_lib.common import utils
from netq_lib.common.utils import save_fullstate, subprocess_wrapper, get_cl_version, \
    is_active_systemd_svc
from netq_lib.common.enums import Platform, ServiceStatus, DefaultPaths, roceConfigFiles
from netq_lib.common import configuration
from netq_lib.common.enums import ConfKw
from netq_agent.parsers.common.render import BaseRender
from netq_agent.handlers.port_handler import PortHandler

import ujson as json


class RoceRender(BaseRender):
    def __init__(self):
        super(RoceRender, self).__init__()
        self.port_id_name_mapping = dict()
        self.last_roce_config_modify_time = [None, None, None, None]
        self.last_nvue_roce_config_modify_time = [None, None, None, None]
        self.switchd_last_reset_time = [None, None]
        self.nvued_last_reset_time = [None, None]
        self.is_roce_new_ver = self.is_roce_new_version()
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)
        self.port_handler = PortHandler.get_instance(self.__logger, None)

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')
        if not configuration.is_cl4():
            return ServiceStatus.OK
        if Platform.MELLANOX not in platform:
            return ServiceStatus.OK

        if os.path.exists(DefaultPaths.ROCE_DUMMY_DATA_PATH):
            try:
                with open(DefaultPaths.ROCE_DUMMY_DATA_PATH, 'r') as f:
                    test_roce_data = yaml.safe_load(f)

                    roce_intf_counters = test_roce_data.get('roce_counters', [])
                    for intf_counter in roce_intf_counters:
                        roce_obj = Roce()
                        roce_obj_attr = dir(roce_obj)
                        for ifname, intf_counters in list(intf_counter.items()):
                            setattr(roce_obj, 'ifname', ifname)
                            cell_unit_size = intf_counters.get('cell_unit_size', 96)
                            for key, value in list(intf_counters.items()):
                                if key in roce_obj_attr:
                                    if 'usage_cells' in key:
                                        setattr(roce_obj, key, value * cell_unit_size)
                                    else:
                                        setattr(roce_obj, key, value)
                        roce_obj.save()

            except Exception as e:
                pass
            return ServiceStatus.OK

        if not self.is_roce_enabled():
            self.delete_all_roce_counters()
            return ServiceStatus.OK

        pool_size_mapping = self.roce_pool_size_json() if self.is_roce_new_ver else self.roce_pool_size()

        try:
            cmd_output = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json roce counters'.split())
            cmd_json_output = json.loads(cmd_output)  # pylint: disable=no-member
        except CalledProcessError as ex:
            self.__logger.debug("mlxcmd roce counters failed: {}".format(ex))
            self.delete_all_roce_counters()
            return ServiceStatus.FAIL
        except ValueError:
            self.delete_all_roce_counters()
            return ServiceStatus.FAIL

        fullstate_arr = []
        if 'exception' in cmd_json_output:
            return ServiceStatus.OK

        for per_intf_dict in list(cmd_json_output):
            roce_obj = Roce()
            roce_obj_attr = dir(roce_obj)
            cell_unit_size = per_intf_dict.get('cell_unit_size', 96)
            for key, value in list(per_intf_dict.items()):
                if value is None:
                    continue
                if key in roce_obj_attr:
                    if 'usage_cells' in key:
                        setattr(roce_obj, key, value * cell_unit_size)
                    else:
                        setattr(roce_obj, key, value)
                if key == 'port_id':
                    ifname = self._get_ifname_from_port_id(value)
                    setattr(roce_obj, 'ifname', ifname)

            for key, value in list(pool_size_mapping.items()):
                if key in roce_obj_attr:
                    setattr(roce_obj, key, value)

            fullstate_arr.append(roce_obj)
            if not fullstate:
                roce_obj.save()

        if fullstate:
            save_fullstate(fullstate_arr, Roce().__class__.__name__)

        return ServiceStatus.OK

    def is_roce_enabled(self):
        roce_enabled = False
        if os.path.exists(DefaultPaths.ROCE_STATUS_FILE_PATH):
            try:
                with open(DefaultPaths.ROCE_STATUS_FILE_PATH, 'r') as fd:
                    for line in fd.readlines():
                        if 'traffic.roce_mode' in line:
                            mode = int(line.strip().split('=')[1].strip())
                            if mode != 0:
                                roce_enabled = True
                            break
            except Exception as e:
                pass
        return roce_enabled

    def is_roce_new_version(self):
        try:
            version = get_cl_version()
            if version.startswith('4.4') or version.startswith('5'):
                return True
            else:
                return False
        except Exception:
            return False

    def is_switch_service_changed(self, is_nvue=False):
        service_changed = False
        if is_nvue:
            array_idx = 1
        else:
            array_idx = 0

        try:
            cmd_out = subprocess_wrapper('/bin/systemctl show -p StateChangeTimestamp switchd'.split())
            state_change_time = cmd_out.splitlines()[0].split('=')[1]
            if self.switchd_last_reset_time[array_idx] is None or state_change_time != self.switchd_last_reset_time[
                array_idx]:
                self.__logger.debug(
                    "switchd: is_nvue: {} old : {}, new {}".format(is_nvue, self.switchd_last_reset_time[array_idx],
                                                                   state_change_time))
                self.switchd_last_reset_time[array_idx] = state_change_time
                service_changed = True
        except ValueError:
            pass
        except CalledProcessError as ex:
            self.__logger.debug("systemctl show -p StateChangeTimestamp switchd failed {}".format(ex))
            pass


        try:
            cmd_out = subprocess_wrapper('/bin/systemctl show -p StateChangeTimestamp nvued'.split())
            state_change_time = cmd_out.splitlines()[0].split('=')[1]
            if self.nvued_last_reset_time[array_idx] is None or state_change_time != self.nvued_last_reset_time[
                array_idx]:
                self.__logger.debug(
                    "nvued: is_nvue: {} old : {}, new {}".format(is_nvue, self.nvued_last_reset_time[array_idx],
                                                                 state_change_time))
                self.nvued_last_reset_time[array_idx] = state_change_time
                service_changed = True
        except ValueError:
            pass
        except CalledProcessError as ex:
            self.__logger.debug("systemctl show -p StateChangeTimestamp nvued failed {}".format(ex))
            pass

        return service_changed

    def is_roce_config_changed(self, is_nvue=False):
        return self.is_config_changed(DefaultPaths.TRAFFIC_CONF_FILE_PATH, roceConfigFiles.TRAFFIC_CONF, is_nvue) or \
               self.is_config_changed(DefaultPaths.QOS_CONF_FILE_PATH, roceConfigFiles.QOS_CONF, is_nvue) or \
               self.is_config_changed(DefaultPaths.QOS_INFRA_FILE_PATH, roceConfigFiles.QOS_INFRA, is_nvue) or \
               self.is_config_changed(DefaultPaths.LLDP_AUTO_CONF_FILE, roceConfigFiles.LLDP_AUTO_CONF_FILE, is_nvue)

    def is_config_changed(self, roce_config_file_path, idx, is_nvue):
        roce_config_changed = False
        if os.path.exists(roce_config_file_path):
            try:
                current_modify_time = int(os.path.getmtime(roce_config_file_path))
                if is_nvue:
                    if self.last_nvue_roce_config_modify_time[idx] is None:
                        roce_config_changed = True
                    elif current_modify_time > self.last_nvue_roce_config_modify_time[idx]:
                        roce_config_changed = True

                    if roce_config_changed:
                        self.last_nvue_roce_config_modify_time[idx] = current_modify_time
                else:
                    if self.last_roce_config_modify_time[idx] is None:
                        roce_config_changed = True
                    elif current_modify_time > self.last_roce_config_modify_time[idx]:
                        roce_config_changed = True

                    if roce_config_changed:
                        self.last_roce_config_modify_time[idx] = current_modify_time
            except Exception as e:
                pass
        return roce_config_changed

    def parse_nvue_qos_roce(self, cmd_output, svc_failed=False, **kwargs):
        cfg_changed = False
        fullstate = kwargs.get("fullstate")

        if (not self.is_roce_enabled()) or (not configuration.is_cl5()) or (not configuration.is_nvue_active()):
            self.delete_all_nvue_roceconfig()
            return ServiceStatus.OK

        if self.is_roce_config_changed(is_nvue=True):
            cfg_changed = True

        '''
                service change | config change | full state | result
                      0                0            0          return
                      0                0            1          continue
                      0                1            0          continue
                      0                1            1          fullstate & continue
                      1                0            0          force a fullstate & continue
                      1                0            1          force a fullstate & continue
                      1                1            0          continue (update from cfg change)
                      1                1            1          fullstate & continue
        '''  # pylint: disable=pointless-string-statement

        switch_svc_changed = self.is_switch_service_changed(is_nvue=True)
        if switch_svc_changed or fullstate:
            self.__logger.debug("[NVUE] svcs_changed {} or fs {}".format(switch_svc_changed, fullstate))
            if not cfg_changed and not fullstate:
                self.__logger.debug("[NVUE] cfg changed False; force fullstate")
                fullstate = True
        elif not cfg_changed:
            return ServiceStatus.OK

        try:
            nvue_roce_status_output = subprocess_wrapper('nv show qos roce -o=json'.split())
            cmd_json_output = json.loads(nvue_roce_status_output)  # pylint: disable=no-member
        except ValueError:
            self.delete_all_nvue_roceconfig()
            return ServiceStatus.FAIL

        m_nvue_roceconfig = NvueRoceConfig.memmodel()
        old_config_objs = []
        new_config_objs = []
        if m_nvue_roceconfig:
            old_config_objs = (
                list(m_nvue_roceconfig.query.filter(hostname=HOSTNAME))
            )

        nvue_roce_config_obj = NvueRoceConfig()

        if 'congestion-control' in cmd_json_output:
            congestion_control = cmd_json_output.get('congestion-control')
            congestion_mode = congestion_control.get('congestion-mode')
            nvue_roce_config_obj.congestion_mode = congestion_mode

            min_threshold = congestion_control.get('min-threshold', 0)
            nvue_roce_config_obj.min_threshold = min_threshold

            max_threshold = congestion_control.get('max-threshold', 0)
            nvue_roce_config_obj.max_threshold = max_threshold

            enabled_tc = congestion_control.get('enabled-tc')
            nvue_roce_config_obj.enabled_tc = enabled_tc

            probability = congestion_control.get('probability')
            nvue_roce_config_obj.probability = probability

        if 'exception-list' in cmd_json_output:
            exception_list = cmd_json_output.get('exception-list', {})
            nvue_roce_config_obj.exception_list = exception_list

        if 'lldp-app-tlv' in cmd_json_output:
            lldp_app_tlv = cmd_json_output.get('lldp-app-tlv')
            lldp_priority = lldp_app_tlv.get('priority')
            nvue_roce_config_obj.lldp_priority = lldp_priority

            lldp_protocol_id = lldp_app_tlv.get('protocol-id')
            nvue_roce_config_obj.lldp_protocol_id = lldp_protocol_id

            lldp_selector = lldp_app_tlv.get('selector')
            nvue_roce_config_obj.lldp_selector = lldp_selector

        if 'mode' in cmd_json_output:
            nvue_roce_config_obj.roce_mode = cmd_json_output.get('mode')

        if 'pfc' in cmd_json_output:
            pfc = cmd_json_output.get('pfc')
            pfc_priority = pfc.get('pfc-priority')
            nvue_roce_config_obj.pfc_priority = pfc_priority

            rx_enabled = pfc.get('rx-enabled')
            nvue_roce_config_obj.rx_enabled = rx_enabled

            tx_enabled = pfc.get('tx-enabled')
            nvue_roce_config_obj.tx_enabled = tx_enabled

        if 'pool-map' in cmd_json_output:
            nvue_roce_config_obj.pool_map = cmd_json_output.get('pool-map', {})

        if 'prio-map' in cmd_json_output:
            nvue_roce_config_obj.priority_map = cmd_json_output.get('prio-map', {})

        if 'tc-map' in cmd_json_output:
            nvue_roce_config_obj.tc_map = cmd_json_output.get('tc-map', {})

        if 'trust' in cmd_json_output:
            trust = cmd_json_output.get('trust')
            nvue_roce_config_obj.trust_mode = trust.get('trust-mode')

        new_config_objs.append(nvue_roce_config_obj)

        # remove current elements from old list
        for new_ele in new_config_objs:
            for old_ele in old_config_objs:
                if old_ele.hostname == new_ele.hostname:
                    old_config_objs.remove(old_ele)

        # delete remaining stale entries
        for old_ele in old_config_objs:
            old_config_ent = NvueRoceConfig()
            old_config_ent.copy(old_ele)
            old_config_ent.delete()

        for nvue_roce_config_obj in new_config_objs:
            nvue_roce_config_obj.save()

        if fullstate:
            save_fullstate(new_config_objs, NvueRoceConfig().__class__.__name__)

        return ServiceStatus.OK

    def parse_roce_config(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')
        cmd_json_output = []
        cfg_changed = False

        if os.path.exists(DefaultPaths.ROCECONFIG_DUMMY_DATA_PATH):
            try:
                with open(DefaultPaths.ROCECONFIG_DUMMY_DATA_PATH, 'r') as f:
                    cmd_json_output = yaml.safe_load(f)
            except Exception as e:
                pass

        else:
            if (Platform.MELLANOX not in platform) or (not configuration.is_cl4()):
                return ServiceStatus.OK

            if not self.is_roce_enabled():
                self.delete_all_roceconfig(fullstate=True)
                return ServiceStatus.OK

            if self.is_roce_config_changed():
                cfg_changed = True

            '''
                service change | config change | full state | result
                      0                0            0          return
                      0                0            1          continue
                      0                1            0          continue
                      0                1            1          fullstate & continue
                      1                0            0          force a fullstate & continue
                      1                0            1          force a fullstate & continue
                      1                1            0          continue (update from cfg change)
                      1                1            1          fullstate & continue
            '''  # pylint: disable=pointless-string-statement

            switch_svc_changed = self.is_switch_service_changed()
            self.__logger.debug("cfg_changed {}".format(cfg_changed))
            if switch_svc_changed or fullstate:
                self.__logger.debug("svcs_changed {} or fs {}".format(switch_svc_changed, fullstate))
                if not cfg_changed and not fullstate:
                    self.__logger.debug("cfg changed False; force fullstate")
                    fullstate = True
            elif not cfg_changed:
                return ServiceStatus.OK

            if os.path.exists('/bin/systemctl'):
                is_active = is_active_systemd_svc('sx_sdk.service')
                if not is_active:
                    self.delete_all_roceconfig()
                    return ServiceStatus.FAIL

            try:
                roce_status_output = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json roce status'.split())
                cmd_json_output = json.loads(roce_status_output)  # pylint: disable=no-member
            except ValueError:
                self.delete_all_roceconfig()
                return ServiceStatus.FAIL

        congestion_control_from_qos = None
        ets_priority_mode = None

        if 'exception' in cmd_json_output:
            self.delete_all_roceconfig()
            return ServiceStatus.OK

        m_roceconfig = RoceConfig.memmodel()
        old_config_objs = []
        new_config_objs = []
        if m_roceconfig:
            old_config_objs = (
                list(m_roceconfig.query.filter(hostname=HOSTNAME))
            )

        for roce_status in list(cmd_json_output):
            roce_config_obj = RoceConfig()

            if 'congestion-control' in roce_status:
                congestion_control = roce_status.get('congestion-control')
                roce_config_obj.ecn_min_th = congestion_control.get('min_threshold_bytes', 0)
                roce_config_obj.ecn_max_th = congestion_control.get('max_threshold_bytes', 0)
                ecn_tcs = congestion_control.get('ecn_tcs')
                roce_config_obj.enabled_tc = ','.join(str(x) for x in ecn_tcs)
                ecn_enabled = congestion_control.get('ecn_enabled', 0) == 1
                red_enabled = congestion_control.get('red_enabled', 0) == 1

                mode = 'None'
                if ecn_enabled and not red_enabled:
                    mode = 'ECN'
                elif not ecn_enabled and red_enabled:
                    mode = 'RED'
                elif ecn_enabled and red_enabled:
                    mode = 'ECN+RED'
                roce_config_obj.mode = mode

            else:
                if congestion_control_from_qos is None:
                    congestion_control_from_qos = self.get_congestion_control()

                roce_config_obj.enabled_tc = congestion_control_from_qos.get('enabled_tc', 'None')
                roce_config_obj.mode = congestion_control_from_qos.get('mode', 'None')
                roce_config_obj.ecn_min_th = congestion_control_from_qos.get('min_th', 0)
                roce_config_obj.ecn_max_th = congestion_control_from_qos.get('max_th', 0)

            if 'roce_dscp' in roce_status:
                roce_config_obj.dscp_sp_map = roce_status.get('roce_dscp', {})
            else:
                roce_config_obj.dscp_sp_map = roce_status.get('dscp', {})

            if 'roce_pg_map' in roce_status:
                roce_config_obj.sp_pg_map = roce_status.get('roce_pg_map', {})
            else:
                roce_config_obj.sp_pg_map = roce_status.get('pg_map', {})

            if 'roce_tc_map' in roce_status:
                roce_config_obj.sp_tc_map = roce_status.get('roce_tc_map', {})
            else:
                roce_config_obj.sp_tc_map = roce_status.get('tc_map', {})

            if 'roce_pcp' in roce_status:
                roce_config_obj.pcp_sp_map = roce_status.get('roce_pcp', {})
            else:
                roce_config_obj.pcp_sp_map = roce_status.get('pcp', {})
            roce_config_obj.pcp_unmapped = 'roce_pcp' not in roce_status

            roce_config_obj.cnp_dscp = roce_status.get('cnp_dscp', {})
            roce_config_obj.cnp_pcp = roce_status.get('cnp_pcp', {})
            roce_config_obj.cnp_pg_map = roce_status.get('cnp_pg_map', {})
            roce_config_obj.cnp_tc_map = roce_status.get('cnp_tc_map', {})
            roce_config_obj.roce_mode = roce_status.get('roce_mode')

            if 'port' in roce_status:
                roce_config_obj.ifname = roce_status.get('port')
            else:
                port_id = roce_status.get('port_id')
                roce_config_obj.ifname = self._get_ifname_from_port_id(port_id)

            rx_enabled = roce_status.get('pfc_enabled', {}).get('rx_enabled', 'no') == 'yes'
            tx_enabled = roce_status.get('pfc_enabled', {}).get('tx_enabled', 'no') == 'yes'
            roce_config_obj.pfc_rx_enabled = rx_enabled
            roce_config_obj.pfc_tx_enabled = tx_enabled
            # PFC priority should be captured if traffic is present in any direction (not both)
            roce_config_obj.pfc_enabled_sp = roce_status.get('pfc_priority') if rx_enabled or tx_enabled else '-'

            if 'cnp_sched_map' in roce_status:
                roce_config_obj.cnp_ets_mode = roce_status.get('cnp_sched_map', {}).get('dwrr_mode')

            if 'roce_sched_map' in roce_status:
                roce_config_obj.ets_mode = roce_status.get('roce_sched_map', {}).get('dwrr_mode')
            else:
                if ets_priority_mode is None:
                    ets_priority_mode = self.get_ets_priority_mode()
                roce_config_obj.ets_mode = ets_priority_mode

            roce_config_obj.nvue_enabled = configuration.is_nvue_active()

            new_config_objs.append(roce_config_obj)

        # remove current elements from old list
        for new_ele in new_config_objs:
            for old_ele in old_config_objs:
                if old_ele.ifname == new_ele.ifname:
                    old_config_objs.remove(old_ele)

        # delete remaining stale entries
        for old_ele in old_config_objs:
            old_config_ent = RoceConfig()
            old_config_ent.copy(old_ele)
            old_config_ent.delete()

        for roce_config_obj in new_config_objs:
            roce_config_obj.save()

        if fullstate:
            save_fullstate(new_config_objs, RoceConfig().__class__.__name__)

        return ServiceStatus.OK

    def delete_all_roceconfig(self, fullstate=False):
        m_roceconfig = RoceConfig.memmodel()
        if m_roceconfig and not fullstate:
            for entry in m_roceconfig.query.filter(hostname=HOSTNAME):
                old_roce_config = RoceConfig()
                old_roce_config.copy(entry)
                old_roce_config.delete()
        else:
            save_fullstate([], RoceConfig().__class__.__name__)


    def delete_all_nvue_roceconfig(self):
        m_roceconfig = NvueRoceConfig.memmodel()
        if m_roceconfig:
            for entry in m_roceconfig.query.filter(hostname=HOSTNAME):
                old_roce_config = NvueRoceConfig()
                old_roce_config.copy(entry)
                old_roce_config.delete()

    def delete_all_roce_counters(self):
        m_roce = Roce.memmodel()
        if m_roce:
            for entry in m_roce.query.filter(hostname=HOSTNAME):
                old_roce = Roce()
                old_roce.copy(entry)
                old_roce.delete()

    def _get_ifname_from_port_id(self, port_id):
        if str(port_id) not in self.port_handler.roce_port_map:
            self.port_handler.refresh_port_map()
            self.port_handler._rebuild_port_maps()
            self.__logger.debug(
                "Refreshed Port Id to Interface name mapping : {}".format(self.port_handler.roce_port_map))
        return self.port_handler.roce_port_map.get(str(port_id))
