#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#
import re
from enum import IntEnum

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import SsdUtil
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.render import BaseRender


class SmartAttrs(IntEnum):
    AverageEraseCount = 167
    NandEndurance = 168
    RemainingLifeTimePerc = 169
    EraseCount = 173
    RemainingLifeLeft = 248


class SmartAttrsRegex:
    DEVICE_MODEL = re.compile(r'^(?:Device Model|Model Number):\s*(.+)')
    PERCENTAGE_USED = re.compile(r'^Percentage Used:\s*(.+)%?')


class SsdUtilRender(BaseRender):
    def __init__(self):
        super(SsdUtilRender, self).__init__()
        self.model = SsdUtil
        self._logger = None

    def build_ssd_util_objs_full_state(self, ssd_util_out, full_state_arr):
        full_state_arr.append(ssd_util_out)

    def build_ssd_util_objs(self, ssd_util_out, full_state_arr, fullstate):
        if fullstate:
            self.build_ssd_util_objs_full_state(ssd_util_out, full_state_arr)
        else:
            ssd_util_out.save()


    def get_default_ssd_util_obj(self):
        new_ssd_util_obj = SsdUtil()
        new_ssd_util_obj.health = -1.0
        new_ssd_util_obj.avg_pe_cycles = -1
        new_ssd_util_obj.device_name = 'NA'
        return new_ssd_util_obj

    def extract_build_info(self, cmd_output):

        '''
        The smartctl command has the following format
            root@mlx-2100-51:/# /usr/sbin/smartctl -a /dev/sda
            smartctl 6.4 2014-10-07 r4002 [x86_64-linux-4.1.0-cl-7-amd64] (local build)
            Copyright (C) 2002-14, Bruce Allen, Christian Franke, www.smartmontools.org

            === START OF INFORMATION SECTION ===
            Device Model:     M.2 (S42) 3ME3
            Serial Number:    20160526AA0918000001
            LU WWN Device Id: 5 24693e 201650001
            Firmware Version: S15A19
            User Capacity:    16,013,942,784 bytes [16.0 GB]
            Sector Size:      512 bytes logical/physical
            Rotation Rate:    Solid State Device
            Form Factor:      2.5 inches
            Device is:        Not in smartctl database [for details use: -P showall]
            ATA Version is:   ATA8-ACS (minor revision not indicated)
            SATA Version is:  SATA 3.0, 6.0 Gb/s (current: 6.0 Gb/s)
            Local Time is:    Mon Oct 14 07:25:27 2019 EDT
            SMART support is: Available - device has SMART capability.
            SMART support is: Enabled

            === START OF READ SMART DATA SECTION ===
            SMART overall-health self-assessment test result: PASSED

            General SMART Values:
            Offline data collection status:  (0x00)	Offline data collection activity
                                was never started.
                                Auto Offline Data Collection: Disabled.
            Total time to complete Offline
            data collection: 		(   32) seconds.
            Offline data collection
            capabilities: 			 (0x00) 	Offline data collection not supported.
            SMART capabilities:            (0x0003)	Saves SMART data before entering
                                power-saving mode.
                                Supports SMART auto save timer.
            Error logging capability:        (0x00)	Error logging NOT supported.
                                General Purpose Logging supported.
            SCT capabilities: 	       (0x0039)	SCT Status supported.
                                SCT Error Recovery Control supported.
                                SCT Feature Control supported.
                                SCT Data Table supported.

            SMART Attributes Data Structure revision number: 16
            Vendor Specific SMART Attributes with Thresholds:
            ID# ATTRIBUTE_NAME          FLAG     VALUE WORST THRESH TYPE      UPDATED  WHEN_FAILED RAW_VALUE
              1 Raw_Read_Error_Rate     0x0000   000   000   000    Old_age   Offline      -       0
              2 Throughput_Performance  0x0000   000   000   000    Old_age   Offline      -       0
              3 Spin_Up_Time            0x0000   000   000   000    Old_age   Offline      -       0
              5 Reallocated_Sector_Ct   0x0013   100   100   001    Pre-fail  Always       -       0
              7 Unknown_SSD_Attribute   0x0000   000   000   000    Old_age   Offline      -       0
              8 Unknown_SSD_Attribute   0x0000   000   000   000    Old_age   Offline      -       0
              9 Power_On_Hours          0x0002   255   000   000    Old_age   Always       -       255
             10 Unknown_SSD_Attribute   0x0000   000   000   000    Old_age   Offline      -       0
             12 Power_Cycle_Count       0x0002   232   000   000    Old_age   Always       -       232
            163 Unknown_Attribute       0x0000   000   000   000    Old_age   Offline      -       13
            168 Unknown_Attribute       0x0000   000   000   000    Old_age   Offline      -       0
            169 Unknown_Attribute       0x0000   083   000   000    Old_age   Offline      -       83
            175 Program_Fail_Count_Chip 0x0000   000   000   000    Old_age   Offline      -       0
            192 Power-Off_Retract_Count 0x0000   000   000   000    Old_age   Offline      -       0
            194 Temperature_Celsius     0x0000   030   100   000    Old_age   Offline      -       30 (2 100 0 0 0)
            197 Current_Pending_Sector  0x0012   000   100   000    Old_age   Always       -       0
            225 Unknown_SSD_Attribute   0x0000   000   251   000    Old_age   Offline      -       26656115
            240 Unknown_SSD_Attribute   0x0000   000   000   000    Old_age   Offline      -       0
            165 Unknown_Attribute       0x0002   020   001   000    Old_age   Always       -       532
            167 Unknown_Attribute       0x0002   240   001   000    Old_age   Always       -       496
            170 Unknown_Attribute       0x0003   100   001   000    Pre-fail  Always       -       65
            171 Unknown_Attribute       0x0002   000   001   000    Old_age   Always       -       0
            172 Unknown_Attribute       0x0002   000   001   000    Old_age   Always       -       0
            176 Erase_Fail_Count_Chip   0x0000   100   001   000    Old_age   Offline      -       0
            187 Reported_Uncorrect      0x0002   000   001   000    Old_age   Always       -       0
            229 Unknown_Attribute       0x0002   100   001   000    Old_age   Always       -       89569723997848
            232 Available_Reservd_Space 0x0003   100   001   000    Pre-fail  Always       -       0
            235 Unknown_Attribute       0x0002   000   000   000    Old_age   Always       -       0
            241 Total_LBAs_Written      0x0002   100   001   000    Old_age   Always       -       104125
            242 Total_LBAs_Read         0x0002   100   001   000    Old_age   Always       -       9488
        '''

        if " No such device" in str(cmd_output):
            self._logger.debug("SsdUtil skipping device: No such device")
            return None
        if "command not found" in str(cmd_output) or "Error response" in str(cmd_output):
            return ServiceStatus.FAIL

        skip_flag = True
        storfly_health = -1
        nand_endurance = None
        erase_count_avg = None
        new_ssd_util_obj = self.get_default_ssd_util_obj()
        if cmd_output is None:
            return new_ssd_util_obj
        for line in cmd_output.split("\n"):
            device_name = SmartAttrsRegex.DEVICE_MODEL.findall(line)
            if device_name:
                new_ssd_util_obj.device_name = device_name[0]

            percentage_used = SmartAttrsRegex.PERCENTAGE_USED.findall(line)
            if percentage_used:
                percentage_used = percentage_used[0].strip('%')
                if percentage_used.isdigit():
                    new_ssd_util_obj.health = 100 - float(percentage_used[0])
            if "ID# ATTRIBUTE_NAME          FLAG     VALUE WORST THRESH TYPE      UPDATED  WHEN_FAILED RAW_VALUE" in line:
                skip_flag = False
            if skip_flag:
                continue
            line_attributes = []
            idx = -1
            try:
                line_attributes = line.split()
                if len(line_attributes) > 0: #pylint: disable=len-as-condition
                    idx = int(line_attributes[0].strip())
            except ValueError:
                line_attributes.append("-1")
            if idx == SmartAttrs.EraseCount:  # For "InnoDisk Corp. - mSATA 3ME" devices
                if len(line_attributes) > 11:
                    raw_val = line_attributes[11].strip()
                    erase_count_avg = int(raw_val, 16) if raw_val.startswith('0x') else int(raw_val)
            if idx == SmartAttrs.NandEndurance:  # For StorFly devices
                raw_val = line_attributes[-1].strip()
                nand_endurance = float.fromhex(raw_val) if raw_val.startswith('0x') else float(raw_val)
            if idx == SmartAttrs.RemainingLifeTimePerc:
                raw_val = line_attributes[-1].strip()
                new_ssd_util_obj.health = float.fromhex(raw_val) if raw_val.startswith('0x') else float(raw_val)
            if idx == SmartAttrs.AverageEraseCount:
                raw_val = line_attributes[-1].strip()
                new_ssd_util_obj.avg_pe_cycles = int(raw_val, 16) if raw_val.startswith('0x') else int(raw_val)
            if idx == SmartAttrs.RemainingLifeLeft:
                raw_val = line_attributes[-1].strip()
                storfly_health = float.fromhex(raw_val) if raw_val.startswith('0x') else float(raw_val)

        if new_ssd_util_obj.avg_pe_cycles == -1 and erase_count_avg is not None:
            new_ssd_util_obj.avg_pe_cycles = erase_count_avg
        if ("StorFly" in new_ssd_util_obj.device_name and "-MLX" in new_ssd_util_obj.device_name):
            if nand_endurance is not None:
                new_ssd_util_obj.health = round(100 - (new_ssd_util_obj.avg_pe_cycles * 100 / nand_endurance), 2)
            else:
                new_ssd_util_obj.health = storfly_health

        return new_ssd_util_obj


    def parse(self, cmd_output, svc_failed=False, **kwargs):
        """Build ssd util info
        """
        fullstate = kwargs.get('fullstate')
        self._logger = kwargs.get('logger')

        try:
            ssd_util_out = self.extract_build_info(cmd_output)
            if not ssd_util_out:
                return ServiceStatus.OK
            if ssd_util_out == ServiceStatus.FAIL:
                self._logger.error("SsdUtil extract build info failed")
                return ServiceStatus.FAIL

        except ValueError:
            self._logger.error("SsdUtil parsing failed with ValueError")
            m_ssd_util = SsdUtil.memmodel()
            for entry in m_ssd_util.query.filter(hostname=HOSTNAME):
                old_ssd_util_ent = SsdUtil()
                old_ssd_util_ent.copy(entry)
                old_ssd_util_ent.delete()
            dummy = []
            if fullstate:
                save_fullstate(dummy, SsdUtil.__name__)
            return ServiceStatus.FAIL

        full_state_arr = []

        self.build_ssd_util_objs(ssd_util_out, full_state_arr, fullstate)

        if fullstate:
            save_fullstate(full_state_arr, SsdUtil.__name__)

        return ServiceStatus.OK
