#
# Copyright (c) 2019-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import BtrFsInfo
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.render import BaseRender


class CumulusBtrfsRender(BaseRender):
    def __init__(self):
        super(CumulusBtrfsRender, self).__init__()
        self.model = BtrFsInfo

    def build_btrfs_objs_full_state(self, btrfs_out, full_state_arr):
        full_state_arr.append(btrfs_out)

    def build_btrfs_objs(self, btrfs_out, full_state_arr, fullstate):
        if fullstate:
            self.build_btrfs_objs_full_state(btrfs_out, full_state_arr)
        else:
            btrfs_out.save()

    def get_value(self, string):
        return string.split(":")[1].strip()

    def get_default_btrfs(self):
        new_btrfs_obj = BtrFsInfo()
        new_btrfs_obj.device_size = -1
        new_btrfs_obj.device_allocated = -1
        new_btrfs_obj.device_unallocated = -1
        new_btrfs_obj.device_missing = -1
        new_btrfs_obj.used = -1
        new_btrfs_obj.free = -1
        new_btrfs_obj.data_ratio = -1.0
        new_btrfs_obj.metadata_ratio = -1.0
        new_btrfs_obj.global_reserve = -1
        new_btrfs_obj.device_single_size = -1
        new_btrfs_obj.device_single_used = -1
        new_btrfs_obj.device_metadata_size = -1
        new_btrfs_obj.device_metadata_used = -1
        new_btrfs_obj.device_system_size = -1
        new_btrfs_obj.device_system_used = -1
        return new_btrfs_obj

    def extract_build_info(self, cmd_output):

        '''
        The btrfs command has the following format
        cumulus@tor-1:~$ sudo btrfs fi usage -b /
        Overall:
            Device size:		        6170849280
            Device allocated:		        1733296128
            Device unallocated:		        4437553152
            Device missing:		                 0
            Used:			        1095913472
            Free (estimated):		        4784119808	(min: 2565343232)
            Data ratio:			              1.00
            Metadata ratio:		              2.00
            Global reserve:		          33554432	(used: 0)

        Data,single: Size:1293680640, Used:947113984
           /dev/vda4	1293680640

        Metadata,DUP: Size:186253312, Used:74383360
           /dev/vda4	 372506624

        System,DUP: Size:33554432, Used:16384
           /dev/vda4	  67108864

        Unallocated:
           /dev/vda4	4437553152
        '''



        new_btrfs_obj = self.get_default_btrfs()
        if cmd_output is None or "ERROR" in cmd_output or "command not found" in cmd_output or "No such file or directory" in cmd_output:
            return ServiceStatus.FAIL
        for line in cmd_output.split("\n"):
            if "Device size" in line:
                try:
                    new_btrfs_obj.device_size = int(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.device_size = -1
            if "Device allocated" in line:
                try:
                    new_btrfs_obj.device_allocated = int(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.device_allocated = -1

            if "Device unallocated" in line:
                try:
                    new_btrfs_obj.device_unallocated = int(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.device_unallocated = -1
            if "Device missing" in line:
                try:
                    new_btrfs_obj.device_missing = int(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.device_missing = -1
            if "Used" in line and line.count(":") == 1:
                try:
                    new_btrfs_obj.used = int(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.used = -1
            if "Free (estimated)" in line:
                try:
                    new_btrfs_obj.free = int(self.get_value(line).split()[0])
                except ValueError:
                    new_btrfs_obj.free = -1
            if "Data ratio" in line:
                try:
                    new_btrfs_obj.data_ratio = float(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.data_ratio = -1.0
            if "Metadata ratio" in line:
                try:
                    new_btrfs_obj.metadata_ratio = float(self.get_value(line))
                except ValueError:
                    new_btrfs_obj.metadata_ratio = -1.0
            if "Global reserve" in line:
                try:
                    new_btrfs_obj.global_reserve = int(self.get_value(line).split()[0])
                except ValueError:
                    new_btrfs_obj.global_reserve = -1

            if "Data,single" in line:
                try:
                    for single_data in line.split(":", 1)[1].split(","):
                        if "Size" in single_data:
                            new_btrfs_obj.device_single_size = int(self.get_value(single_data))
                        elif "Used" in single_data:
                            new_btrfs_obj.device_single_used = int(self.get_value(single_data))
                except ValueError:
                    new_btrfs_obj.device_single_size = -1
                    new_btrfs_obj.device_single_used = -1

            if "Metadata,DUP" in line:
                try:
                    for single_data in line.split(":", 1)[1].split(","):
                        if "Size" in single_data:
                            new_btrfs_obj.device_metadata_size = int(self.get_value(single_data))
                        elif "Used" in single_data:
                            new_btrfs_obj.device_metadata_used = int(self.get_value(single_data))
                except ValueError:
                    new_btrfs_obj.device_metadata_size = -1
                    new_btrfs_obj.device_metadata_used = -1

            if "System,DUP" in line:
                try:
                    for single_data in line.split(":", 1)[1].split(","):
                        if "Size" in single_data:
                            new_btrfs_obj.device_system_size = int(self.get_value(single_data))
                        elif "Used" in single_data:
                            new_btrfs_obj.device_system_used = int(self.get_value(single_data))
                except ValueError:
                    new_btrfs_obj.device_system_size = -1
                    new_btrfs_obj.device_system_used = -1

        return new_btrfs_obj

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get('fullstate')

        try:
            btrfs_out = self.extract_build_info(cmd_output)

        except ValueError:
            m_btrfs = BtrFsInfo.memmodel()
            for entry in m_btrfs.query.filter(hostname=HOSTNAME):
                oldbtrfs_ent = BtrFsInfo()
                oldbtrfs_ent.copy(entry)
                oldbtrfs_ent.delete()
            dummy = []
            if fullstate:
                save_fullstate(dummy, BtrFsInfo.__name__)
            return ServiceStatus.FAIL

        full_state_arr = []

        self.build_btrfs_objs(btrfs_out, full_state_arr, fullstate)

        if fullstate:
            save_fullstate(full_state_arr, BtrFsInfo.__name__)

        return ServiceStatus.OK


def process_btrfs(cmd_output, svc_failed=False, **kwargs):
    """Build btrfs info
    """
    render = CumulusBtrfsRender()
    return render.parse(cmd_output, svc_failed, **kwargs)
