#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import ClagSession
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.render import BaseRender


class CumulusClagRender(BaseRender):
    def __init__(self):
        super(CumulusClagRender, self).__init__()
        self.model = ClagSession

    def parse(self, cmd_output, is_clag_running=False, **kwargs):
        """Build clag info for node
        """
        retcode = ServiceStatus.OK
        fullstate = kwargs.get('fullstate')

        try:
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            json_out = []
            if fullstate is True:
                save_fullstate(json_out, ClagSession().__class__.__name__)
            retcode = ServiceStatus.FAIL

        if 'errorMsg' in json_out:
            if 'Unable to communicate with clagd' in json_out['errorMsg']:
                retcode = ServiceStatus.OK
                is_clag_running = False
            else:
                retcode = ServiceStatus.FAIL
                is_clag_running = True
            json_out = []

        m_clag_session = ClagSession.memmodel()
        for entry in m_clag_session.query.filter(hostname=HOSTNAME):
            old_clag_entry = ClagSession()
            if (not json_out or
                    json_out['status']['sysMac'] != entry.clag_sysmac):
                if not is_clag_running:
                    old_clag_entry.copy(entry)
                    old_clag_entry.delete()
                else:
                    old_clag_entry.peer_state = False
                    old_clag_entry.clag_sysmac = entry.clag_sysmac
                    old_clag_entry.conflicted_bonds = []
                    old_clag_entry.single_bonds = []
                    old_clag_entry.proto_down_bonds = []
                    old_clag_entry.dual_bonds = {}
                    old_clag_entry.save()

        if not json_out:
            if fullstate is True:
                save_fullstate(json_out, ClagSession().__class__.__name__)
            return retcode

        clag_conflicted_bonds = []
        clag_singly_attached_bonds = []
        clag_proto_down_bonds = []
        dual_attach_bonds = {}

        clag_session = ClagSession()

        clag_session.peer_state = json_out['status']['peerAlive']

        clag_session.peer_if = (
            json_out['status']['peerIf']
        )
        clag_session.peer_role = json_out['status']['peerRole']
        clag_session.role = json_out['status']['ourRole']
        clag_session.clag_sysmac = json_out['status']['sysMac']
        if json_out['status']['peerIp']:
            clag_session.peer_ip = (
                json_out['status']['peerIp']
            )
        else:
            clag_session.peer_ip = ''
        if json_out['status']['backupIp']:
            clag_session.backup_ip = (
                json_out['status']['backupIp']
            )
        else:
            clag_session.backup_ip = ''
        clag_session.backup_ip_active = json_out['status']['backupActive']
        clag_session.vxlan_anycast = (
            json_out['status'].get('vxlanAnycast', ''))

        for iface in json_out['clagIntfs']:
            iface = iface
            if json_out['clagIntfs'][iface]['status'] == 'dual':
                dual_attach_bonds[iface] = (
                    json_out['clagIntfs'][iface]['peerIf']
                )
            elif 'conflicts' in json_out['clagIntfs'][iface]:
                clag_conflicted_bonds.append(
                    [iface,
                     json_out['clagIntfs'][iface]['conflicts'][0]]

                )
                retcode = ServiceStatus.ERROR
            elif 'protoDown' in json_out['clagIntfs'][iface]:
                clag_proto_down_bonds.append([
                    iface,
                    json_out['clagIntfs'][iface]['protoDown'][0]])
                retcode = ServiceStatus.ERROR
            else:
                clag_singly_attached_bonds.append(iface)
                retcode = ServiceStatus.WARNING

        clag_session.conflicted_bonds = clag_conflicted_bonds
        clag_session.single_bonds = clag_singly_attached_bonds
        clag_session.proto_down_bonds = clag_proto_down_bonds
        clag_session.dual_bonds = dual_attach_bonds

        # If the CLAG sysmac changed, we need to delete the older entry.
        m_clag_session = ClagSession.memmodel()
        for entry in m_clag_session.query.filter(hostname=HOSTNAME):
            if entry.clag_sysmac != clag_session.clag_sysmac:
                old_clag_entry = ClagSession()
                old_clag_entry.copy(entry)
                old_clag_entry.delete()
        clag_session.save()

        clag_session_arr = []
        clag_session_arr.append(clag_session)
        if fullstate is True:
            save_fullstate(clag_session_arr, ClagSession().__class__.__name__)

        return retcode


def process_clagctl(cmd_output, is_clag_running=False, **kwargs):
    render = CumulusClagRender()
    render.parse(cmd_output, is_clag_running, **kwargs)
