#
# Copyright (c) 2020-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


from netq_lib.common.enums import Platform
from netq_lib.orm.redisdb.models import Dom
from netq_lib.common.enums import ServiceStatus
from netq_agent.parsers.common.dom import DomRender

class CumulusDomRender(DomRender):

    def _get_command(self):
        return "ethtool -m "

    def _process_command_output(self, cmd_out):
        eth_m_elems = dict()
        for eth_tool_dom in cmd_out.op.splitlines():
            eth_tool_dom = eth_tool_dom.decode()
            if ":" not in eth_tool_dom:
                continue

            key, value = eth_tool_dom.split(':', 1)
            key = key.strip()
            value = value.strip()
            eth_m_elems[key] = value

        if 'Optical diagnostics support' in eth_m_elems:
            if 'Yes' not in eth_m_elems['Optical diagnostics support']:
                return ServiceStatus.ERROR
        elif 'Transmitter technology' in eth_m_elems:
            if 'Copper' in eth_m_elems['Transmitter technology']:
                return ServiceStatus.ERROR

        return eth_m_elems, ServiceStatus.OK

    def _create_dom_object(self, ifname, eth_elems):
        dom_obj = Dom()
        dom_obj.ifname = ifname
        dom_obj.identifier = eth_elems.get('Identifier', 'n/a')

        dom_obj.laser_rx_power = {}
        dom_obj.laser_output_power = {}
        dom_obj.laser_bias_current = {}
        if 'Optical diagnostics support' in eth_elems:
            self._get_single_channel_values(dom_obj, eth_elems)
        elif 'Transmitter technology' in eth_elems:
            self._get_multi_channel_values(dom_obj, eth_elems)

        dom_obj.module_temp = eth_elems.get('Module temperature', 'n/a')
        dom_obj.module_voltage = eth_elems.get('Module voltage', 'n/a')

        dom_obj.laser_rx_power_high_alarm_th = eth_elems.get('Laser rx power high alarm threshold', 'n/a')
        dom_obj.laser_rx_power_low_alarm_th = eth_elems.get('Laser rx power low alarm threshold', 'n/a')
        dom_obj.laser_rx_power_high_warning_th = eth_elems.get('Laser rx power high warning threshold', 'n/a')
        dom_obj.laser_rx_power_low_warning_th = eth_elems.get('Laser rx power low warning threshold', 'n/a')

        dom_obj.laser_output_power_high_alarm_th = eth_elems.get('Laser output power high alarm threshold', 'n/a')
        dom_obj.laser_output_power_low_alarm_th = eth_elems.get('Laser output power low alarm threshold', 'n/a')
        dom_obj.laser_output_power_high_warning_th = eth_elems.get('Laser output power high warning threshold', 'n/a')
        dom_obj.laser_output_power_low_warning_th = eth_elems.get('Laser output power low warning threshold', 'n/a')

        dom_obj.laser_bias_current_high_alarm_th = eth_elems.get('Laser bias current high alarm threshold', 'n/a')
        dom_obj.laser_bias_current_low_alarm_th = eth_elems.get('Laser bias current low alarm threshold', 'n/a')
        dom_obj.laser_bias_current_high_warning_th = eth_elems.get('Laser bias current high warning threshold', 'n/a')
        dom_obj.laser_bias_current_low_warning_th = eth_elems.get('Laser bias current low warning threshold', 'n/a')

        dom_obj.module_temp_high_alarm_th = eth_elems.get('Module temperature high alarm threshold', 'n/a')
        dom_obj.module_temp_low_alarm_th = eth_elems.get('Module temperature low alarm threshold', 'n/a')
        dom_obj.module_temp_high_warning_th = eth_elems.get('Module temperature high warning threshold', 'n/a')
        dom_obj.module_temp_low_warning_th = eth_elems.get('Module temperature low warning threshold', 'n/a')

        dom_obj.module_voltage_high_alarm_th = eth_elems.get('Module voltage high alarm threshold', 'n/a')
        dom_obj.module_voltage_low_alarm_th = eth_elems.get('Module voltage low alarm threshold', 'n/a')
        dom_obj.module_voltage_high_warning_th = eth_elems.get('Module voltage high warning threshold', 'n/a')
        dom_obj.module_voltage_low_warning_th = eth_elems.get('Module voltage low warning threshold', 'n/a')

        return dom_obj

    @staticmethod
    def _get_single_channel_values(dom_obj, eth_elems):
        dom_obj.laser_rx_power['Channel 1'] = eth_elems.get('Receiver signal average optical power', 'n/a')
        dom_obj.laser_output_power['Channel 1'] = eth_elems.get('Laser output power', 'n/a')
        dom_obj.laser_bias_current['Channel 1'] = eth_elems.get('Laser bias current', 'n/a')

    def _get_multi_channel_values(self, dom_obj, eth_elems):
        for key, value in eth_elems.items():
            if 'Channel' in key:
                rx_power = self.LASER_RX_POWER_PATTERN.match(key)
                if rx_power:
                    dom_obj.laser_rx_power[rx_power.group(1)] = value
                else:
                    output_power = self.LASER_OUTPUT_POWER_PATTERN.match(key)
                    if output_power:
                        dom_obj.laser_output_power[output_power.group(1)] = value
                    else:
                        bias_current = self.LASER_BIAS_CURRENT_PATTERN.match(key)
                        if bias_current:
                            dom_obj.laser_bias_current[bias_current.group(1)] = value


def process_dom(output, svc_failed=False, **kwargs):
    platform = kwargs.get('platform', '')
    if not (Platform.VX in platform or Platform.CUMULUS in platform):
        return
    dom.process_dom_interfaces(kwargs)


# Creating Dom object, which holds the context for dom processing
dom = CumulusDomRender()
