#
# Copyright (c) 2017-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
import os
import ctypes
import subprocess
from subprocess import CalledProcessError
from distutils.util import strtobool
import ujson as json
from netq_lib.common.enums import Platform
from netq_lib.common.utils import is_cl3, is_cl4, is_active_systemd_svc, check_cl_version_onwards
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import EcmpHash, EcmpInfo, AREcmpInfo, ARConfigGlobalV2, ARConfigIntf, AsicMonitorHistogram
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.enums import AREcmpQhistP95State
from netq_lib.common import utils, configuration
from netq_lib.common.utils import save_fullstate, subprocess_wrapper
from netq_lib.common.utils import SWITCHD_SW_TABLE_PATH
from netq_lib.common.utils import get_cl_switchd_sw_tables_idx
from netq_lib.common.enums import ConfKw, DefaultPaths
from netq_agent.parsers.common.render import BaseRender
from netq_agent.handlers.sdk_utils import SdkUtils

class CumulusEcmpHashRender(BaseRender):

    def __init__(self):
        super(CumulusEcmpHashRender, self).__init__()
        self.model = EcmpHash
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)

    def search_field(self, field, text):
        return re.search(field + " = (.*)", text)

    def extract(self, field, text):
        commented_matched = self.search_field("\#.*?" + field, text) #pylint: disable=anomalous-backslash-in-string
        if commented_matched:
            return ''
        matched = self.search_field(field, text)
        if matched:
            return matched.groups()[0] if len(matched.groups()) > 0 else '' #pylint: disable=len-as-condition
        else:
            return ''

    def to_boolean(self, input_val, default=False):
        return strtobool(input_val) if input_val.strip() else default

    def to_int(self, input_val, default=-1):
        casted_unsigned_int = ctypes.c_uint32(int(input_val)).value if input_val.strip() else default
        return casted_unsigned_int

    def parse(self, cmd_output, svc_failed, **kwargs):
        # self.__logger.info("In CumulusEcmpHashRender svc_failed {} cmd_output {}".format(svc_failed, cmd_output))

        fullstate = kwargs.get('fullstate')
        platform = kwargs.get('platform', '')

        ecmp_hash_list = []

        if svc_failed:
            self.__logger.error("Ecmp hash lookup on traffic.conf failed")
            return ServiceStatus.FAIL
        try:
            hash_val = EcmpHash()
            hash_val.symmetric_hash_enable = self.to_boolean(self.extract("symmetric_hash_enable", cmd_output), True)
            hash_val.resilient_hash_enable = self.to_boolean(self.extract("resilient_hash_enable", cmd_output))
            hash_val.seed = self.to_int(self.extract("ecmp_hash_seed", cmd_output)) # Default is random in the config
            hash_val.max_paths = self.to_int(self.extract("ecmp_max_paths", cmd_output), 0) # Default is 0 in the config
            if hash_val.max_paths == 0:
                #SDK interprets Max path Config 0 as all available ports. report the max path from sx netdev instead of 0
                cmd = "ip -o link show type sx_netdev  | wc -l"
                hash_val.max_paths = self.to_int(subprocess.check_output(cmd, shell=True))

            hash_val.resilient_hash_entries = self.to_int(self.extract("resilient_hash_entries_ecmp", cmd_output), 64) # Default is 64 in the config
            hash_val.ip_prot = self.to_boolean(self.extract("hash_config.ip_prot", cmd_output))
            hash_val.sip = self.to_boolean(self.extract("hash_config.sip", cmd_output))
            hash_val.dip = self.to_boolean(self.extract("hash_config.dip", cmd_output))
            hash_val.sport = self.to_boolean(self.extract("hash_config.sport", cmd_output))
            hash_val.dport = self.to_boolean(self.extract("hash_config.dport", cmd_output))
            hash_val.ip6_label = self.to_boolean(self.extract("hash_config.ip6_label", cmd_output))
            hash_val.ing_intf = self.to_boolean(self.extract("hash_config.ing_intf", cmd_output))
            hash_val.inner_ip_prot = self.to_boolean(self.extract("hash_config.inner_ip_prot", cmd_output))
            hash_val.inner_sip = self.to_boolean(self.extract("hash_config.inner_sip", cmd_output))
            hash_val.inner_dip = self.to_boolean(self.extract("hash_config.inner_dip", cmd_output))
            hash_val.inner_sport = self.to_boolean(self.extract("hash_config.inner_sport", cmd_output))
            hash_val.inner_dport = self.to_boolean(self.extract("hash_config.inner_dport", cmd_output))
            hash_val.inner_ip6_label = self.to_boolean(self.extract("hash_config.inner_ip6_label", cmd_output))
            if fullstate is True:
                ecmp_hash_list.append(hash_val)
            else:
                hash_val.save()
        except Exception as e:
            self.__logger.error(e, exc_info=True)
            return ServiceStatus.FAIL

        if fullstate is True:
            save_fullstate(ecmp_hash_list, EcmpHash().__class__.__name__)

        return ServiceStatus.OK


def process_ecmp_hash(cmd_output, svc_failed=False, **kwargs):
    render = CumulusEcmpHashRender()
    render.parse(cmd_output, svc_failed, **kwargs)


class CumulusEcmpInfoRender(BaseRender):
    def __init__(self):
        super(CumulusEcmpInfoRender, self).__init__()
        self.model = EcmpInfo
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)
        self.intf_map_id = 0
        self.ecmp_sdk_id_table = 0
        self.sdk_utils = SdkUtils(self.__logger)

    def parse_ecmp_group_type(self, lines):
        result_dict = {}
        for line in lines:
            line = line.strip()
            if line.startswith("sdk-id"):
                group_id = int(line.split()[1])
            elif line.startswith("router-ecmp-type"):
                result_dict[group_id] = line.split()[2]
        return result_dict

    def get_ecmp_group_type_mapping(self):

        is_active = False
        result_dict = {}
        file_path = DefaultPaths.ECMP_GROUP_TYPE_DUMMY_DATA_PATH

        if os.path.exists(file_path):
            with open(file_path, 'r') as fd:
                lines = fd.readlines()

            result_dict = self.parse_ecmp_group_type(lines)
            return result_dict

        if self.ecmp_sdk_id_table == 0:
            self.ecmp_sdk_id_table = get_cl_switchd_sw_tables_idx('Backend ECMP SDK ID Table')

        if self.ecmp_sdk_id_table == 0:
            try:
                if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                    is_active = is_active_systemd_svc('switchd.service')
                    if not is_active:
                        self.__logger.debug('Cannot get ECMP SDK-ID table, switchd service not running?')
                        return None
            except Exception as e:
                self.__logger.debug('Check switchd status %s', str(e))
                return None

            '''
            When switchd is transition to active->inactive,
            there can be a scenario when the sw_tables are not found, but switchd is active.
            ''' #pylint: disable=pointless-string-statement
            self.__logger.info('Cannot get ECMP SDK-ID table {}'.format(is_active))
            return None

        sw_table_file = '{}/{}'.format(SWITCHD_SW_TABLE_PATH, self.ecmp_sdk_id_table)
        cmd = 'cat {} | grep "sdk-id\|router-ecmp-type"'.format(sw_table_file) #pylint: disable=anomalous-backslash-in-string

        if not os.path.exists(sw_table_file):
            return None

        ecmp_group_type_info = subprocess.check_output(cmd, shell=True)
        lines = ecmp_group_type_info.decode("utf-8").split("\n")

        result_dict = self.parse_ecmp_group_type(lines)

        return result_dict

    def get_rif_interface_mapping(self):

        is_active = False
        file_path = DefaultPaths.RIF_INTERFACE_DUMMY_DATA_PATH
        if os.path.exists(file_path):
            with open(file_path, 'r') as fd:
                matched_interfaces = fd.readlines()

            matched_interface_groups = re.findall("ifp entry\s+(.*?)\s+-.*?l3-interface-id\s+(\d+)", str(matched_interfaces)) #pylint: disable=anomalous-backslash-in-string
            self.__logger.info("rif mapping {}".format(matched_interface_groups))
            return {int(x[1]):x[0] for x in matched_interface_groups}

        if self.intf_map_id == 0:
            self.intf_map_id = get_cl_switchd_sw_tables_idx('Backend Interface Table')

        if self.intf_map_id == 0:
            try:
                if (is_cl3() or is_cl4()) and os.path.exists('/bin/systemctl'):
                    is_active = is_active_systemd_svc('switchd.service')
                    if not is_active:
                        self.__logger.debug('Cannot get intf map table, switchd service not running?')
                        return None
            except Exception as e:
                self.__logger.debug('Check switchd status %s', str(e))
                return None

            '''
            When switchd is transition to active->inactive,
            there can be a scenario when the sw_tables are not found, but switchd is active.
            ''' #pylint: disable=pointless-string-statement
            self.__logger.info('Cannot get intf map table {}'.format(is_active))
            return None

        sw_table_file = '{}/{}'.format(SWITCHD_SW_TABLE_PATH, self.intf_map_id)
        cmd = 'cat {} | grep "ifp entry\|l3-interface-id"'.format(sw_table_file) #pylint: disable=anomalous-backslash-in-string

        if not os.path.exists(sw_table_file):
            return None

        matched_interfaces = subprocess.check_output(cmd, shell=True)
        matched_interface_groups = re.findall("ifp entry\s+(.*?)\s+-.*?l3-interface-id\s+(\d+)", str(matched_interfaces)) #pylint: disable=anomalous-backslash-in-string
        return {int(x[1]):x[0] for x in matched_interface_groups}

    def calculate_p95(self, queue_hist_entry):
        ifname = queue_hist_entry.ifname
        bin_list = []
        bin_list.append(queue_hist_entry.bin0)
        bin_list.append(queue_hist_entry.bin1)
        bin_list.append(queue_hist_entry.bin2)
        bin_list.append(queue_hist_entry.bin3)
        bin_list.append(queue_hist_entry.bin4)
        bin_list.append(queue_hist_entry.bin5)
        bin_list.append(queue_hist_entry.bin6)
        bin_list.append(queue_hist_entry.bin7)
        bin_list.append(queue_hist_entry.bin8)
        bin_list.append(queue_hist_entry.bin9)

        bin_start = queue_hist_entry.minimum_queue_bytes
        bin_width = queue_hist_entry.queue_bytes_range

        total_frequency = sum(bin_entry for bin_entry in bin_list)

        p95_position = 0.95 * total_frequency

        cumulative_frequency = 0
        p95_value = None
        for index, bin_entry in enumerate(bin_list):
            cumulative_frequency += bin_entry
            if cumulative_frequency >= p95_position:
                bin_index_p95 = index
                p95_value = bin_start + bin_width * (bin_index_p95 + 1) - 1
                break

        return p95_value

    def get_ar_ecmp_associated_data(self):

        ar_capable = False
        ar_dict = {}
        qHist_dict = {}
        ar_global_enabled = False
        old_ar_ecmp_objs = []

        p95_dummy_dict = {}

        file_path = DefaultPaths.QHIST_P95_DUMMY_DATA_PATH

        if os.path.exists(file_path):
            with open(file_path, "r") as file:
                for line in file:
                    parts = line.strip().split(":")

                    if len(parts) == 2:
                        key = parts[0].strip()
                        value = int(parts[1].strip())
                        p95_dummy_dict[key] = value

        m_arconfigglobal_objs = ARConfigGlobalV2.memmodel()
        if m_arconfigglobal_objs:
            ar_capable = True
            arconfigglobal_objs = [old_ar_ecmp_obj for old_ar_ecmp_obj in m_arconfigglobal_objs.query.filter(hostname=HOSTNAME)]
            for entry in arconfigglobal_objs:
                ar_global_enabled = entry.is_enabled

        if ar_global_enabled:
            m_arconfigintf_objs = ARConfigIntf.memmodel()
            if m_arconfigintf_objs:
                arconfigintf_objs = [old_ar_ecmp_obj for old_ar_ecmp_obj in m_arconfigintf_objs.query.filter(hostname=HOSTNAME)]

                for entry in arconfigintf_objs:
                    ifname = entry.ifname
                    is_ar_enabled = entry.is_enabled
                    ar_dict[ifname] = is_ar_enabled

            m_ar_ecmp_info = AREcmpInfo.memmodel()
            if m_ar_ecmp_info:
                old_ar_ecmp_objs = [old_ar_ecmp_obj for old_ar_ecmp_obj in m_ar_ecmp_info.query.filter(hostname=HOSTNAME)]

            m_qHist_objs = AsicMonitorHistogram.memmodel()
            if m_qHist_objs:
                qHist_objs = [qHist_obj for qHist_obj in m_qHist_objs.query.filter(hostname=HOSTNAME)]

                for entry in qHist_objs:
                    ifname = entry.ifname
                    p95 = self.calculate_p95(entry)
                    qHist_dict[ifname] = p95

                    if ifname in p95_dummy_dict:
                        qHist_dict[ifname] = p95_dummy_dict[ifname]

        return ar_capable, ar_global_enabled, ar_dict, qHist_dict, old_ar_ecmp_objs

    def rif_interface_get_port(self, ifname):
        fields = ifname.split()
        if len(fields) == 1:
            return ifname
        if len(fields) >= 2 and fields[1].endswith(":"):
            return fields[1][:-1]

        return ifname

    def delete_ar_ecmp_info(self):
        m_ar_ecmp_info = AREcmpInfo.memmodel()
        if m_ar_ecmp_info:
            for entry in m_ar_ecmp_info.query.filter(hostname=HOSTNAME):
                old_ar_ecmp_info = AREcmpInfo()
                old_ar_ecmp_info.copy(entry)
                old_ar_ecmp_info.delete()

    def parse(self, cmd_output, svc_failed, **kwargs):

        fullstate = kwargs.get('fullstate')
        platform = kwargs.get('platform', '')

        file_path = DefaultPaths.ECMP_DUMMY_DATA_PATH
        if os.path.exists(file_path):
            with open(file_path, 'r') as json_file:
                json_out = json.load(json_file) #pylint: disable=no-member

            if not json_out:
                self.delete_ar_ecmp_info()
        else:
            try:
                ecmp_count = self.sdk_utils.get_ecmp_count()
                if ecmp_count and int(ecmp_count) == 0:
                    self.delete_ar_ecmp_info()
                    return ServiceStatus.OK
            except Exception as ex:
                self.__logger.debug("Got error while fetching ecmp_count: {}".format(ex))

            try:
                cmd_output = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json l3 ecmp_table'.split())
                json_out = json.loads(cmd_output)  # pylint: disable=no-member
            except ValueError:
                self.__logger.error("ECMP info lookup on mlxcmd failed")
                return ServiceStatus.ERROR
            except CalledProcessError as ex:
                self.__logger.info("ECMP info lookup on mlxcmd failed with {}".format(ex))
                return ServiceStatus.ERROR
            if len(json_out) == 0:
                return ServiceStatus.OK  # return if nothing to process

        try:
            rif_interface_mapping = self.get_rif_interface_mapping()
            if rif_interface_mapping is None:
                self.__logger.info("Ecmp info lookup on switchd software-tables l3-interface failed")
                return ServiceStatus.FAIL
        except Exception as e:
            self.__logger.info("Ecmp info lookup on switchd software-tables l3-interface failed")
            return ServiceStatus.FAIL

        # self.__logger.info("In CumulusEcmpInfoRender svc_failed {} cmd_output {} interfaces {}".format(svc_failed, cmd_output, rif_interface_mapping))

        ecmp_info_list = []
        new_ar_ecmp_info_objs = []
        fullstate_ar_ecmp_info_arr = []

        ar_dict = {}
        qHist_dict = {}
        ar_global_enabled = False

        ar_capable, ar_global_enabled, ar_dict, qHist_dict, old_ar_ecmp_objs = self.get_ar_ecmp_associated_data()
        if ar_capable:
            ecmp_group_type_mapping = {}
            try:
                ecmp_group_type_mapping = self.get_ecmp_group_type_mapping()
                if ecmp_group_type_mapping is None:
                    self.__logger.info("Ecmp info lookup on switchd software-tables ECMP SDK-ID failed")
                    return ServiceStatus.FAIL
            except Exception as e:
                self.__logger.info("Ecmp info lookup on switchd software-tables ECMP SDK-ID failed")
                return ServiceStatus.FAIL


        for group_id in json_out:
            mbr_list = []
            grp_ar_config = None
            mbr_qhist_p95 = None
            grp_ar_mbr_cfg_eq = True
            mbrs_qhist_p95_state = AREcmpQhistP95State.EQUAL.value
            for single_json_obj in json_out[group_id]['nh_list']:
                try:
                    ecmpObj = EcmpInfo()
                    ecmpObj.group_id = group_id
                    ecmpObj.rif = single_json_obj[0]
                    ecmpObj.weight = single_json_obj[1]
                    ecmpObj.bytes = single_json_obj[5]
                    ecmpObj.ifname = rif_interface_mapping.get(ecmpObj.rif, '')

                    if ar_capable:
                        ifname = self.rif_interface_get_port(ecmpObj.ifname)
                        if_ar_enabled = ar_dict.get(ifname, False)
                        p95 = qHist_dict.get(ifname, -1)

                        if not grp_ar_config:
                            grp_ar_config = if_ar_enabled
                            grp_ar_mbr_cfg_eq = True
                        else:
                            grp_ar_mbr_cfg_eq = grp_ar_mbr_cfg_eq and (grp_ar_config == if_ar_enabled)

                        if p95 == -1:
                            mbrs_qhist_p95_state = AREcmpQhistP95State.UNKNOWN.value

                        if mbrs_qhist_p95_state != AREcmpQhistP95State.UNKNOWN.value:
                            if not mbr_qhist_p95:
                                mbr_qhist_p95 = p95
                                mbrs_qhist_p95_state = AREcmpQhistP95State.EQUAL.value
                            else:
                                if mbr_qhist_p95 != p95:
                                    mbrs_qhist_p95_state = AREcmpQhistP95State.UNEQUAL.value

                        mbrObj = {
                            "ifname": ifname,
                            "ifArEnabled": if_ar_enabled,
                            "weight": ecmpObj.weight,
                            "p95": p95
                        }

                        mbr_list.append(mbrObj)

                    if fullstate is True:
                        ecmp_info_list.append(ecmpObj)
                    else:
                        ecmpObj.save()

                except Exception as e:
                    if(check_cl_version_onwards(major=5, minor=1) and Platform.MELLANOX in platform):
                        self.__logger.error(e, exc_info=True)
                    pass

            if ar_capable:
                ar_ecmp_obj = AREcmpInfo()
                ar_ecmp_obj.group_id = int(group_id)
                ar_ecmp_obj.ar_enabled = ar_global_enabled
                ar_ecmp_obj.mbr_list = mbr_list
                ar_ecmp_obj.mbrs_ar_config_eq = grp_ar_mbr_cfg_eq
                ar_ecmp_obj.mbrs_qhist_p95_state = mbrs_qhist_p95_state
                ar_ecmp_obj.ecmp_type = ecmp_group_type_mapping.get(ar_ecmp_obj.group_id, None)
                new_ar_ecmp_info_objs.append(ar_ecmp_obj)
                if not fullstate:
                    ar_ecmp_obj.save()
                else:
                    fullstate_ar_ecmp_info_arr.append(ar_ecmp_obj)

        if ar_capable:
            for old_ar_ecmp_obj in old_ar_ecmp_objs:
                if old_ar_ecmp_obj.group_id not in [new_obj.group_id for new_obj in new_ar_ecmp_info_objs]:
                    ar_ecmp_old_obj = AREcmpInfo()
                    ar_ecmp_old_obj.copy(old_ar_ecmp_obj)
                    ar_ecmp_old_obj.delete()
                    old_ar_ecmp_objs.remove(old_ar_ecmp_obj)

        if fullstate is True:
            save_fullstate(ecmp_info_list, EcmpInfo().__class__.__name__)

            if ar_capable:
                save_fullstate(fullstate_ar_ecmp_info_arr, AREcmpInfo().__class__.__name__)

        return ServiceStatus.OK

def process_ecmp_info(cmd_output, svc_failed=False, **kwargs):
    render = CumulusEcmpInfoRender()
    render.parse(cmd_output, svc_failed, **kwargs)
