#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import time
import ujson as json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import MstpInfo
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.mstp import MstpRender


class CumulusMstpRender(MstpRender):
    def __init__(self): #pylint: disable=useless-super-delegation
        super(CumulusMstpRender, self).__init__()

    def __is_bridge_vlan_filtering(self, name):
        """Determine if this bridge is a vlan filtering bridge or not
        """
        try:
            with open(
                    '/sys/class/net/%s/bridge/vlan_filtering' % name,
                    'r') as fname:
                return int(fname.readline().strip()) == 1
        except IOError:
            return False

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        retcode = ServiceStatus.OK
        fullstate = kwargs.get('fullstate')

        try:
            json_out = json.loads(cmd_output) #pylint: disable=no-member
        except ValueError:
            # We treat invalid JSON as the infomation terminating
            json_out = []
            if fullstate is True:
                save_fullstate(json_out, MstpInfo().__class__.__name__)
            retcode = ServiceStatus.FAIL

        now = time.time()
        m_mstp_session = MstpInfo.memmodel()
        for entry in m_mstp_session.query.filter(hostname=HOSTNAME):
            old_mstp_entry = MstpInfo()
            if not json_out:
                if not entry.state:
                    continue
                if not svc_failed:
                    old_mstp_entry.copy(entry)
                    old_mstp_entry.delete()
                else:
                    old_mstp_entry.state = False
                    old_mstp_entry.bridge_name = entry.bridge_name
                    old_mstp_entry.save()
            elif entry.bridge_name not in json_out:
                old_mstp_entry.copy(entry)
                old_mstp_entry.delete()

        mstp_session_arr = []
        for key in json_out:
            bridge = json_out[key][key]
            # UPDATE: We do handle vlan-unaware bridges now
            # We don't handle vlan-unaware bridges yet
            # if 'vlanFilter' not in bridge:
            #     continue
            mstp_session = MstpInfo()
            edge_ports = []
            network_ports = []
            disputed_ports = []
            bpduguard_ports = []
            bpduguard_err_ports = []
            bpdufilter_ports = []
            ba_inconsistent_ports = []
            ports = {}

            mstp_session.bridge_name = key
            if 'rootPortName' in bridge:
                mstp_session.root_port_name = (
                    bridge['rootPortName']
                )
            mstp_session.root_bridge = bridge['dsgnRoot'].lower()
            mstp_session.topo_chg_ports = [
                bridge['topoChngPort'],
                bridge['lastTopoChngPort']
            ]
            mstp_session.topo_chg_cntr = bridge['topoChngCounter']
            mstp_session.bridge_id = bridge['bridgeId'].lower()
            mstp_session.time_since_tcn = int(
                now - bridge['timeSinceLastTopoChng']
            )
            mstp_session.is_vlan_filtering = self.__is_bridge_vlan_filtering(key)

            for port in json_out[key]:
                if port in (key, 'bridgeIndex'):
                    continue

                # All the following sections assume that the key exists only if the
                # value is True. This is the current behavior of mstpctl.
                for portid in json_out[key][port]:
                    portinfo = json_out[key][port][portid]
                    if 'operEdgePort' in portinfo:
                        edge_ports.append(port)
                    if 'networkPort' in portinfo:
                        network_ports.append(port)
                    if 'baInconsistent' in portinfo:
                        ba_inconsistent_ports.append(port)
                        retcode = ServiceStatus.ERROR
                    if 'bpduGuardPort' in portinfo:
                        bpduguard_ports.append(port)
                    if 'bpduGuardError' in portinfo:
                        bpduguard_err_ports.append(port)
                        retcode = ServiceStatus.ERROR
                    if 'disputed' in portinfo:
                        disputed_ports.append(port)
                        retcode = ServiceStatus.ERROR
                    if 'bpduFilterPort' in portinfo:
                        bpdufilter_ports.append(port)

                    ports[port] = {}
                    ports[port]['role'] = portinfo['role']
                    ports[port]['state'] = portinfo['state']
                    ports[port]['clagDualConnMac'] = (
                        portinfo['clagDualConnMac'].lower()
                    )
                    ports[port]['sendRstp'] = portinfo['allInformation'].get('sendRstp', False)
                    ports[port]['clagIsl'] = 'clagIsl' in portinfo or False
                    ports[port]['clagRole'] = portinfo['clagRole']

            mstp_session.edge_ports = edge_ports
            mstp_session.network_ports = network_ports
            mstp_session.disputed_ports = disputed_ports
            mstp_session.bpduguard_ports = bpduguard_ports
            mstp_session.bpduguard_err_ports = bpduguard_err_ports
            mstp_session.bpdufilter_ports = bpdufilter_ports
            mstp_session.ba_inconsistent_ports = ba_inconsistent_ports
            mstp_session.ports = ports
            mstp_session.state = True

            if fullstate is True:
                mstp_session_arr.append(mstp_session)
            else:
                mstp_session.save()

        if fullstate is True:
            save_fullstate(mstp_session_arr, MstpInfo().__class__.__name__)

        return retcode


def process_mstpall(cmd_output, svc_failed=False, **kwargs):
    """Build mstp info for node
    """
    render = CumulusMstpRender()
    return render.parse(cmd_output, svc_failed, **kwargs)
