#
# Copyright (c) 2017-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import subprocess
import ujson as json
from netq_lib.orm.redisdb.models import OS, ASIC, License, Board
from netq_lib.common.enums import Platform
from netq_lib.common.utils import save_fullstate, subprocess_wrapper
from netq_agent.parsers.common.platform import PlatformRender, OK, MISSING
import netq_agent.parsers.common.platform as platform_module


class CumulusPlatformRender(PlatformRender):
    def __init__(self): #pylint: disable=useless-super-delegation
        super(CumulusPlatformRender, self).__init__()

    def process_os_info(self, logger, os_release):
        os = OS()

        os.name = 'N/A'
        os.version = 'N/A'
        os.version_id = 'N/A'

        try:
            if os_release is not None:
                for line in os_release.splitlines():
                    if line.startswith('NAME='):
                        os.name = line.split('=')[1].strip('\"')
                    if line.startswith('VERSION='):
                        os.version = line.split('=')[1].strip('\"')
                    if line.startswith('VERSION_ID='):
                        os.version_id = line.split('=')[1].strip('\"')
        except (IndexError, ValueError) as ex:
            logger.error("Failed to process os_release output: {}".format(ex))
            raise ex

        return os

    def process_asic_info(self, logger, board, platform_json, platform_detect):
        asic = ASIC()

        asic.vendor = 'N/A'
        asic.model = 'N/A'
        asic.model_id = 'N/A'
        asic.core_bw = 'N/A'
        asic.ports = 'N/A'

        try:
            if platform_json:
                asic.vendor = platform_json.get('soc', dict()).get('vendor', '')
                asic.model = platform_json.get('soc', dict()).get('model', '')
                asic.model_id = platform_json.get('soc', dict()).get('model_id', '')
                # CL 3.1.2 and before use /usr/share/cumulus/platform-info/<vendor.json>
                # that don't have the core_bw fields.
                if platform_json.get('soc', dict()).get('core_bw', None):
                    asic.core_bw = '%s%s' % (
                        platform_json.get('soc', dict()).get('core_bw', ''),
                        platform_json.get('soc', dict()).get('core_bw_unit', ''))
                asic.ports = platform_json.get('ports', dict()).get('layout', '')
                # This is a hack for NETQ-1759
                if platform_detect == 'mlnx,x86_MSN2700':
                    board.vendor = 'Mellanox'
                    board.model = 'MSN2700'
                elif platform_detect == 'mlnx,x86_MSN2100':
                    board.vendor = 'Mellanox'
                    board.model = 'MSN2100'
                elif "simx" in platform_detect:
                    board.vendor = 'Mellanox'
                    board.model = platform_detect.split(",", 1)[1].split("_", 1)[1].upper()
                else:
                    board.model = platform_json.get('model_name', '')
                    if board.vendor == 'N/A':
                        board.vendor = platform_json.get('vendor', '')
        except (ValueError, KeyError) as ex:
            logger.error("Failed to process platform information: {}".format(ex))
            raise ex

        return asic

    def process_os_license_info(self, cmd_output, svc_status=OK, **kwargs):
        fullstate = kwargs.get('fullstate')
        license_obj = License()
        license_obj.name = Platform.CL
        license_dict = dict()
        platform = kwargs.get('platform', '')
        # NETQ-1835: earlier we are doing inventory.save two places.
        if platform_module.OLD_SVC_STATUS != svc_status:
            # NETQ-2142 the inventory.license_state will be initialized at agent restart.
            license_obj.license = ''
            license_obj.state = 'N/A'
            license_dict = {'name': Platform.CL, 'license': ''}
            if 'cumulus' in platform:
                try:
                    license_obj.license = license_dict['license'] = subprocess_wrapper(
                        ['/usr/cumulus/bin/cl-license', ]
                    )
                except (subprocess.CalledProcessError, OSError):
                    pass
                if svc_status == OK:
                    if license_obj.license == "":
                        license_obj.state = 'missing'
                    else:
                        license_obj.state = 'ok'
                elif svc_status == MISSING:
                    license_obj.state = 'missing'
                else:
                    license_obj.state = 'invalid'
            platform_module.OLD_SVC_STATUS = svc_status
            license_obj.save()

        if fullstate:
            license_obj_arr = []
            license_obj_arr.append(license_obj)
            save_fullstate(license_obj_arr, License().__class__.__name__)
        return license_obj, license_dict

    def process_syseeprom_info(self, logger, decode_syseeprom):
        board = Board()
        board.vendor = 'N/A'
        board.model = 'N/A'
        board.base_mac = 'N/A'
        board.part_number = 'N/A'
        board.mfg_date = 'N/A'
        board.serial_number = 'N/A'
        board.label_revision = 'N/A'

        try:
            if decode_syseeprom is not None:
                json_op = json.loads(decode_syseeprom) #pylint: disable=no-member
                tlvs = json_op['tlv']
                if 'Base MAC Address' in tlvs:
                    board.base_mac = tlvs["Base MAC Address"]["value"].strip(chr(0))
                if 'Part Number' in tlvs:
                    board.part_number = tlvs["Part Number"]["value"].strip(chr(0))
                if 'Manufacture Date' in tlvs:
                    board.mfg_date = tlvs["Manufacture Date"]["value"][:10]. \
                        strip(chr(0))
                if 'Serial Number' in tlvs:
                    board.serial_number = tlvs["Serial Number"]["value"].strip(chr(0))
                if 'Label Revision' in tlvs:
                    board.label_revision = tlvs["Label Revision"]["value"].strip(chr(0))
                if 'Product Name' in tlvs:
                    board.model = tlvs["Product Name"]["value"].strip(chr(0))
                # vendor: recently observed from 4.3 CL, where Vendor Name is replaced with Manufacturer
                if 'Manufacturer' in tlvs:
                    board.vendor = tlvs["Manufacturer"]["value"].strip(chr(0))
                elif 'Vendor Name' in tlvs:
                    board.vendor = tlvs["Vendor Name"]["value"].strip(chr(0))

        except (KeyError, ValueError) as ex:
            logger.error("Failed to process syseeprom output: {}".format(ex))

        return board


def process_platform_info(
        logger,
        random_platform,
        platform,
        os_release,
        decode_syseeprom,
        lscpu,
        meminfo,
        platform_json,
        meminfo_dmi,
        lsblk,
        agent_version,
        platform_detect,
        is_full_state,
        app_release_info,
        service_nvue_enabled,
        service_nvue_active,
        is_nvue_config_saved,
        service_nclu_enabled,
        opta_on_switch_enabled,
        lcm_executor_on_switch_enabled
):
    render = CumulusPlatformRender()
    render.process_platform_info(logger, random_platform, platform,
                                 os_release, decode_syseeprom, lscpu,
                                 meminfo, platform_json, meminfo_dmi,
                                 lsblk, agent_version, platform_detect,
                                 is_full_state, app_release_info,
                                 service_nvue_enabled, service_nvue_active,
                                 is_nvue_config_saved, service_nclu_enabled,
                                 opta_on_switch_enabled, lcm_executor_on_switch_enabled)


def process_cl_license(cmd_output, svc_failed=OK, **kargs):
    render = CumulusPlatformRender()
    render.process_os_license_info(cmd_output, svc_failed, **kargs)
