#
# Copyright (c) 2022-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import ctypes
import re
import os
from datetime import datetime
from subprocess import CalledProcessError
import eventlet
from netq_lib.orm import HOSTNAME
from netq_lib.common.enums import Platform
from netq_lib.common.utils import subprocess_wrapper
from netq_lib.orm.redisdb.models import PtpGlobalConfig, PtpClockInfo, PtpCorrection, PtpPortStatus, PtpPortStatistics
from netq_lib.common.enums import ServiceStatus, DefaultPaths
from netq_lib.common import utils, configuration
from netq_lib.common.utils import save_fullstate, is_active_systemd_svc
from netq_lib.common.enums import ConfKw
from netq_agent.parsers.common.render import BaseRender


class CumulusPtpRender(BaseRender):
    __instance = None

    @staticmethod
    def get_instance():
        if CumulusPtpRender.__instance is None:
            CumulusPtpRender()
        return CumulusPtpRender.__instance

    def __init__(self):
        if CumulusPtpRender.__instance is None:
            CumulusPtpRender.__instance = self
        super(CumulusPtpRender, self).__init__()
        self.last_ptp_config_modify_time = None
        self.__logger = utils.get_logger(ConfKw.NETQ_AGENT, configuration)
        self.__platform = None


    def search_field(self, field, text): #pylint: disable=inconsistent-return-statements
        re_expression = field + "\t(.*)" + "|" + field + " (.*)"
        match = re.search(re_expression, text)
        if match:
            if match.groups()[0]:
                return match.groups()[0]
            if match.groups()[1]:
                return match.groups()[1]
        else:
            return None

    def extract(self, field, text):
        commented_matched = self.search_field("\#.*?" + field, text) #pylint: disable=anomalous-backslash-in-string
        if commented_matched:
            return ''
        matched = self.search_field(field, text)
        if matched:
            return matched if len(matched) > 0 else '' #pylint: disable=len-as-condition
        else:
            return ''

    def to_boolean(self, arg):
        return int(arg.strip()) == 1 if len(arg) else False #pylint: disable=len-as-condition

    def to_int(self, arg, default=-1):
        casted_int = ctypes.c_int32(int(arg, 0)).value if arg.strip() else default
        return casted_int

    def to_long(self, arg, default=-1):
        casted_long = ctypes.c_long(int(arg, 0)).value if arg.strip() else default
        return casted_long

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')

        if not configuration.is_cl4():
            return ServiceStatus.OK
        if Platform.MELLANOX not in platform:
            return ServiceStatus.OK

        if not self.is_ptp_enabled():
            self.delete_all_ptpconfig()
            return ServiceStatus.OK

        if not self.is_ptp_config_changed() and not fullstate:
            return ServiceStatus.OK

        ptp_config_list = []
        if svc_failed:
            self.__logger.warn("PTP config lookup on ptp4l.conf failed")
            return ServiceStatus.FAIL
        try:
            #self.__logger.info("ptp4l.conf file out: {}".format(cmd_output))
            ptp_global_config = PtpGlobalConfig()
            ptp_global_config.slave_only = self.to_boolean(self.extract("slaveOnly", cmd_output))
            ptp_global_config.priority1 = self.to_int(self.extract("priority1", cmd_output))
            ptp_global_config.priority2 = self.to_int(self.extract("priority2", cmd_output))
            ptp_global_config.domain_number = self.to_int(self.extract("domainNumber", cmd_output))
            ptp_global_config.two_step_flag = self.to_boolean(self.extract("twoStepFlag", cmd_output))
            ptp_global_config.ofm_min_threshold = self.to_int(
                self.extract("offset_from_master_min_threshold", cmd_output))
            ptp_global_config.ofm_max_threshold = self.to_int(
                self.extract("offset_from_master_max_threshold", cmd_output))
            ptp_global_config.mpd_threshold = self.to_int(self.extract("mean_path_delay_threshold", cmd_output))

            if fullstate is True:
                ptp_config_list.append(ptp_global_config)
            else:
                ptp_global_config.save()
        except Exception as e:
            self.__logger.error("ptp config parse caught exception: {}".format(e))
            return ServiceStatus.FAIL

        if fullstate:
            save_fullstate(ptp_config_list, PtpGlobalConfig().__class__.__name__)

        return ServiceStatus.OK

    def is_ptp_enabled(self):
        is_pt4l_active = False
        if os.path.exists('/bin/systemctl'):
            is_pt4l_active = is_active_systemd_svc('ptp4l.service')

        return is_pt4l_active

    def is_ptp_config_changed(self):
        return self.is_config_changed(DefaultPaths.PTP_CONF_FILE_PATH)

    def is_config_changed(self, ptp_config_file_path):
        ptp_config_changed = False
        if os.path.exists(ptp_config_file_path):
            try:
                current_modify_time = int(os.path.getmtime(ptp_config_file_path))
                if self.last_ptp_config_modify_time is None:
                    ptp_config_changed = True
                elif current_modify_time > self.last_ptp_config_modify_time:
                    ptp_config_changed = True

                if ptp_config_changed:
                    self.last_ptp_config_modify_time = current_modify_time
            except Exception:
                pass
        return ptp_config_changed

    def extract_ptp_counters(self, counter_obj, port, counters_list):
        # self.__logger.info("ptp counters: {}, len: {}".format(counters_list, len(counters_list)))
        setattr(counter_obj, 'ifname', port)
        if len(counters_list) < 9:
            return False
        for line in counters_list:
            items = [v.strip() for v in line.split()]
            if items[0] == "Announce":
                counter_obj.rx_announce, counter_obj.tx_announce = self.to_long(items[1]), self.to_long(items[2])
            if items[0] == "Sync":
                counter_obj.rx_sync, counter_obj.tx_sync = self.to_long(items[1]), self.to_long(items[2])
            if items[0] == "Follow-up":
                counter_obj.rx_follow_up, counter_obj.tx_follow_up = self.to_long(items[1]), self.to_long(items[2])
            if items[0] == "Delay" and  items[1] == "Request":
                counter_obj.rx_delay_req, counter_obj.tx_delay_req = self.to_long(items[2]), self.to_long(items[3])
            if items[0] == "Delay" and  items[1] == "Response":
                counter_obj.rx_delay_resp, counter_obj.tx_delay_resp = self.to_long(items[2]), self.to_long(items[3])
            if "Peer" in items and "Request" in items:
                counter_obj.rx_peer_delay_req, counter_obj.tx_peer_delay_req = self.to_long(items[3]), self.to_long(
                    items[4])
            if "Peer" in items and "Response" in items:
                counter_obj.rx_peer_delay_resp, counter_obj.tx_peer_delay_resp = self.to_long(items[3]), self.to_long(
                    items[4])
            if items[0] == "Management":
                counter_obj.rx_management, counter_obj.tx_management = self.to_long(items[1]), self.to_long(items[2])
            if items[0] == "Signaling":
                counter_obj.rx_signaling, counter_obj.tx_signaling = self.to_long(items[1]), self.to_long(items[2])

        return True

    def parse_ptp_counters(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')

        if not configuration.is_cl4():
            return ServiceStatus.OK

        if Platform.MELLANOX not in platform:
            return ServiceStatus.OK

        if not self.is_ptp_enabled():
            self.delete_all_ptp_counters()
            return ServiceStatus.OK

        try:
            '''Sample Output
            -------------------------------------------------------------------------------
            Packet Type         Received          Transmitted
            -------------------------------------------------------------------------------
            Port swp20
            Announce               29022                    0
            Sync                   58040                    0
            Follow-up              58040                    0
            Delay Request              0                57786
            Delay Response         57786                    0
            Peer Delay Request         0                    0
            Peer Delay Response        0                    0
            Management                 0                    0
            Signaling                  0                    0

            Port swp21
            Announce               29022                    0
            Sync                   58040                    0
            Follow-up              58040                    0
            Delay Request              0                    0
            Delay Response             0                    0
            Peer Delay Request         0                    0
            Peer Delay Response        0                    0
            Management                 0                    0
            Signaling                  0                    0
            ''' #pylint: disable=pointless-string-statement

            fullstate_arr = []
            counters_output = subprocess_wrapper('/usr/lib/cumulus/ptpdebug show counters'.split())
            # self.__logger.info("ptp counters out: {}".format(counters_output))
            # for each port there will be 9 counters associated
            # the code is based on above format assumption
            lines = counters_output.splitlines()[3:]
            while len(lines) > 0: #pylint: disable=len-as-condition
                if "Port" in lines[0]:
                    ptp_counter_obj = PtpPortStatistics()
                    port = lines[0].split()[1].strip()
                    counters = lines[1:10]
                    lines = lines[11:]
                    if not self.extract_ptp_counters(ptp_counter_obj, port, counters):
                        continue
                    fullstate_arr.append(ptp_counter_obj)
                    if not fullstate:
                        ptp_counter_obj.save()
                else:
                    lines = lines[1:]

        except Exception as e:
            self.__logger.error("ptp counter collection caught exception: {}".format(e))
            self.delete_all_ptp_counters()
            return ServiceStatus.FAIL

        if fullstate and fullstate_arr:
            save_fullstate(fullstate_arr, PtpPortStatistics().__class__.__name__)

        return ServiceStatus.OK

    def parse_ptp_clock(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')

        if not configuration.is_cl4():
            return ServiceStatus.OK
        if Platform.MELLANOX not in platform:
            return ServiceStatus.OK

        # cache the platform information
        if self.__platform is None:
            self.__platform = platform

        if not self.is_ptp_enabled():
            self.delete_all_ptp_clockinfo()
            return ServiceStatus.OK

        clock_info_list = []
        try:
            """ sample output:
            Clock Information

            Clock Identity       : 1c:34:da:ff:fe:2d:a1:48
            Clock Domain         : 0
            One Step/Two Step    : Two Step
            Number of PTP ports  : 1
            Priority1            : 255
            Priority2            : 255
            Clock Quality
                Class            : 248
                Accuracy         : 0xfe
            Offset (log variance): 0xffff
            Offset From Master   : 2.0 ns
            Mean Path Delay      : 19.0 ns
            Steps Removed        : 1
            """ #pylint: disable=pointless-string-statement
            ptp_clock_output = subprocess_wrapper('/usr/lib/cumulus/ptpdebug show clock'.split())

            # self.__logger.info("clock cmd out: {}".format(ptp_clock_output))

            """Sample output
            Parent Clock
            Parent Port identity             : 1c34da.fffe.2da048-1
            Parent Offset Scaled Log Variance: 0xffff
            Parent Clock Phase Change Rate   : 0x7fffffff
            Parent Stats                     : 0
            GM Identity                      : 1c:34:da:ff:fe:2d:a0:48
            GM Priority1                     : 255
            GM Priority2                     : 255
            GM Clock Quality
                Class                        : 248
                Accuracy                     : 254
            GM Offset(log variance)          : 0xff
            """ #pylint: disable=pointless-string-statement
            ptp_clock_parent = subprocess_wrapper('/usr/lib/cumulus/ptpdebug show clock parent'.split())
            # self.__logger.info("parent clock cmd out: {}".format(ptp_clock_parent))

            clock_info = PtpClockInfo()
            for line in ptp_clock_output.splitlines():
                if ":" not in line:
                    continue
                kv_list = [s.strip() for s in line.split(":", 1)]
                if kv_list[0] == "Clock Identity":
                    clock_info.clock_identity = kv_list[1]
                if kv_list[0] == "Clock Domain":
                    clock_info.clock_domain = self.to_int(kv_list[1])
                if kv_list[0] == "Number of PTP ports":
                    clock_info.num_ptp_ports = self.to_int(kv_list[1])
                if kv_list[0] == "Class":
                    clock_info.cq_quality = self.to_long(kv_list[1])
                if kv_list[0] == "Accuracy":
                    clock_info.cq_accuracy = self.to_long(kv_list[1])
                if kv_list[0] == "Steps Removed":
                    clock_info.steps_removed = self.to_int(kv_list[1])

            for line in ptp_clock_parent.splitlines():
                if ":" not in line:
                    continue
                kv_list = [s.strip() for s in line.split(":", 1)]
                if kv_list[0] == "Parent Port identity":
                    clock_info.parent_clock_identity = kv_list[1]
                if kv_list[0] == "GM Identity":
                    clock_info.gm_identity = kv_list[1]
                if kv_list[0] == "Class":
                    clock_info.gm_cq_class = kv_list[1]
                if kv_list[0] == "Accuracy":
                    clock_info.gm_cq_accuracy = kv_list[1]

            if fullstate is True:
                clock_info_list.append(clock_info)
            else:
                clock_info.save()
        except Exception as e:
            self.__logger.error("ptp show clock command parsing failed with error : {}".format(e))
            return ServiceStatus.FAIL

        if fullstate:
            save_fullstate(clock_info_list, PtpClockInfo().__class__.__name__)

        return ServiceStatus.OK


    def parse_ptp_status(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')

        if not configuration.is_cl4():
            return ServiceStatus.OK
        if Platform.MELLANOX not in platform:
            return ServiceStatus.OK

        if not self.is_ptp_enabled():
            self.delete_all_ptp_port_status()
            return ServiceStatus.OK

        m_ptp_status = PtpPortStatus.memmodel()
        old_ptp_status_list = []
        new_ptp_status_list = []
        if m_ptp_status:
            old_ptp_status_list = (
                list(m_ptp_status.query.filter(hostname=HOSTNAME))
            )

        try:
            '''Sample output
            ----------------------------------------------------
            Port State
            ----------------------------------------------------
            swp20 SLAVE
            swp21 PASSIVE
            swp22 PASSIVE
            ''' #pylint: disable=pointless-string-statement
            ptp_status_output = subprocess_wrapper('/usr/lib/cumulus/ptpdebug show status'.split())
            # self.__logger.info("ptp status out: {}".format(ptp_status_output))

            # ignore if the command failed
            if "failed" in ptp_status_output:
                return ServiceStatus.FAIL

            status_lines = ptp_status_output.splitlines()[3:]
            for line in status_lines:
                if "ptp_glbl_show:" in line:
                    return ServiceStatus.FAIL
                kv_list = [v.strip() for v in line.split()]
                ptp_status_obj = PtpPortStatus()
                ptp_status_obj.ifname = kv_list[0]
                ptp_status_obj.port_status = kv_list[1]
                new_ptp_status_list.append(ptp_status_obj)

        except (CalledProcessError, OSError):
            return ServiceStatus.FAIL
        except Exception as e:
            self.__logger.debug("ptp show status cmd parsing failed with error: {}".format(str(e)))
            return ServiceStatus.FAIL

        # remove current elements from old list
        for new_ele in new_ptp_status_list:
            for old_ele in old_ptp_status_list:
                if old_ele.ifname == new_ele.ifname:
                    old_ptp_status_list.remove(old_ele)

        # delete remaining stale entries
        for old_ele in old_ptp_status_list:
            old_config_ent = PtpPortStatus()
            old_config_ent.copy(old_ele)
            old_config_ent.delete()

        for ptp_status_obj in new_ptp_status_list:
            ptp_status_obj.save()

        if fullstate:
            save_fullstate(new_ptp_status_list, PtpPortStatus().__class__.__name__)

        return ServiceStatus.OK

    def parse_ptp_correction(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get("fullstate")
        platform = kwargs.get('platform', '')

        if not configuration.is_cl4():
            return ServiceStatus.OK
        if Platform.MELLANOX not in platform:
            return ServiceStatus.OK

        if not self.is_ptp_enabled():
            return ServiceStatus.OK

        try:
            ptp_correction_out = subprocess_wrapper('/usr/lib/cumulus/tsmon -o -j'.split())
            # self.__logger.info("ptp correction out: {}".format(ptp_correction_out))

            fullstate_arr = []
            for line in ptp_correction_out.splitlines():
                correction = PtpCorrection()
                seq, data = line.split("@", 1)
                for kv in data.split(','):
                    key, value = kv.split("=")
                    if key == "log-time":
                        correction.log_timestamp = datetime.strptime(value[:-1], "%Y-%m-%dT%H:%M:%S.%f").timestamp()
                    if key == "offset-from-master":
                        correction.offset_from_master = self.to_long(value)
                    if key == "mean-path-delay":
                        correction.mean_path_delay = self.to_long(value)

                fullstate_arr.append(correction)
                if not fullstate:
                    correction.save_grpc()

        except Exception as e:
            self.__logger.error("ptp correction command parsing failed with error: {}".format(e))
            return ServiceStatus.FAIL

        if fullstate and fullstate_arr:
            save_fullstate(fullstate_arr, PtpCorrection().__class__.__name__)

        return ServiceStatus.OK

    def poll_ptp_corrections(self):
        while True:
            try:
                if self.__platform:
                    self.parse_ptp_correction(cmd_output=None, svc_failed=False, fullstate=True, platform=self.__platform)
                eventlet.sleep(3)
            except Exception as e:
                self.__logger.warn("PTP Corrections polling failed: {}".format(e))
                break
        return

    def delete_all_ptpconfig(self):
        m_ptp_config = PtpGlobalConfig.memmodel()
        if m_ptp_config:
            for entry in m_ptp_config.query.filter(hostname=HOSTNAME):
                old_ptp_config = PtpGlobalConfig()
                old_ptp_config.copy(entry)
                old_ptp_config.delete()

    def delete_all_ptp_clockinfo(self):
        m_ptp_clock = PtpClockInfo.memmodel()
        if m_ptp_clock:
            for entry in m_ptp_clock.query.filter(hostname=HOSTNAME):
                old_ptp_clock = PtpClockInfo()
                old_ptp_clock.copy(entry)
                old_ptp_clock.delete()

    def delete_all_ptp_port_status(self):
        m_ptp_status = PtpPortStatus.memmodel()
        if m_ptp_status:
            for entry in m_ptp_status.query.filter(hostname=HOSTNAME):
                old_ptp_status = PtpPortStatus()
                old_ptp_status.copy(entry)
                old_ptp_status.delete()

    def delete_all_ptp_counters(self):
        m_ptp_counters = PtpPortStatistics.memmodel()
        if m_ptp_counters:
            for entry in m_ptp_counters.query.filter(hostname=HOSTNAME):
                old_ptp_counter = PtpPortStatistics()
                old_ptp_counter.copy(entry)
                old_ptp_counter.delete()


def process_ptp_config(cmd_output, svc_failed=False, **kwargs):
    ptp_render.parse(cmd_output, svc_failed, **kwargs)
    # ptp_render.parse_ptp_port_config(cmd_output, svc_failed, **kwargs)


def process_clock_info(cmd_output, svc_failed=False, **kwargs):
    ptp_render.parse_ptp_clock(cmd_output, svc_failed, **kwargs)


def process_ptp_counters(cmd_output, svc_failed=False, **kwargs):
    ptp_render.parse_ptp_counters(cmd_output, svc_failed, **kwargs)


def process_ptp_status(cmd_output, svc_failed=False, **kwargs):
    ptp_render.parse_ptp_status(cmd_output, svc_failed, **kwargs)


def process_ptp_correction(cmd_output, svc_failed=False, **kwargs):
    ptp_render.parse_ptp_correction(cmd_output, svc_failed, **kwargs)


ptp_render = CumulusPtpRender.get_instance()
