#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import os
import subprocess
import ujson as json
from netq_lib.common.utils import subprocess_wrapper, is_active_systemd_svc, get_logger, \
    get_first_up_port, get_first_existing_port
from netq_lib.common.enums import ServiceStatus
from netq_lib.common import configuration
from netq_lib.common.enums import ConfKw
from netq_agent.parsers.common.roce import RoceRender

class CumulusRoceRender(RoceRender):
    def __init__(self):
        super(CumulusRoceRender, self).__init__()
        self.__logger = get_logger(ConfKw.NETQ_AGENT, configuration)

    def roce_port_info(self):
        port_id_name_mapping = dict()

        if os.path.exists('/bin/systemctl'):
            is_active = is_active_systemd_svc('sx_sdk.service')
            if not is_active:
                return ServiceStatus.FAIL

        try:
            mlx_port_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json ports info'.split())
            try:
                mlx_port_info_json = json.loads(mlx_port_info) #pylint: disable=no-member
            except ValueError:
                return ServiceStatus.FAIL
            for port_id, port_status in list(mlx_port_info_json.items()):
                port_id_name_mapping[port_id] = port_status.get('ifname')
        except subprocess.CalledProcessError:
            pass
        return port_id_name_mapping


    def roce_pool_size(self):
        pool_size_mapping = dict()

        if os.path.exists('/bin/systemctl'):
            is_active = is_active_systemd_svc('sx_sdk.service')
            if not is_active:
                return None

        try:
            mlx_system_status = subprocess_wrapper('/usr/lib/cumulus/mlxcmd roce status --system'.split())
            for line in mlx_system_status.splitlines():
                split_line = line.split()
                if len(split_line) != 9:
                    continue
                pool_name, _, _, direction, _, pool_size, _, _, _ = line.split()
                pool_name_updated = pool_name.replace("-", "_") + '_size'
                pool_size_mapping[pool_name_updated] = int(pool_size) * 144
        except subprocess.CalledProcessError:
            pass
        return pool_size_mapping

    def roce_pool_size_for_interface(self, mlx_system_json):
        functional_port = get_first_up_port() if get_first_up_port() is not None else get_first_existing_port()
        pool_json = dict()
        try:
            pool_max_idx = len(mlx_system_json.get('pools'))
            found_key = False

            for idx in range(pool_max_idx):
                for key in mlx_system_json.get('pools')[idx].keys():
                    if key == functional_port:
                        found_key = True
                        break
                if found_key:
                    break

            if not found_key:
                idx = 0
                functional_port = list(mlx_system_json.get('pools')[idx].keys())[0]

            pool_json = mlx_system_json.get('pools')[idx].get(functional_port)
        except subprocess.CalledProcessError:
            pass
        return pool_json

    def roce_pool_size_json(self):
        pool_size_mapping = dict()

        if os.path.exists('/bin/systemctl'):
            is_active = is_active_systemd_svc('sx_sdk.service')
            if not is_active:
                return pool_size_mapping

        try:
            mlx_system_status = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json roce status --system'.split())
            try:
                mlx_system_json = json.loads(mlx_system_status) #pylint: disable=no-member
            except ValueError:
                return pool_size_mapping

            pool_json = mlx_system_json.get('pools')
            if type(pool_json) is list: #pylint: disable=unidiomatic-typecheck
                pool_json = self.roce_pool_size_for_interface(mlx_system_json)

            if pool_json:
                for pool_id, pool_status in list(pool_json.items()):
                    is_infinite = pool_status.get('infinite', 0) == 1
                    pool_size = -1 if is_infinite else pool_status.get('bytes', 0)
                    pool_name_updated = pool_status.get('name').replace("-", "_") + '_size'
                    pool_size_mapping[pool_name_updated] = pool_size
        except subprocess.CalledProcessError:
            pass
        return pool_size_mapping

    def get_congestion_control(self):
        congestion_control_map = dict()

        if os.path.exists('/bin/systemctl'):
            is_active = is_active_systemd_svc('sx_sdk.service')
            if not is_active:
                return congestion_control_map

        try:
            qos_redecn_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json qos redecn --port swp*'.split())
            try:
                qos_redecn_info_json_arr = json.loads(qos_redecn_info) #pylint: disable=no-member
            except ValueError:
                return congestion_control_map

            for qos_redecn_info_json in qos_redecn_info_json_arr:
                if 'config' in qos_redecn_info_json:
                    qos_configs = qos_redecn_info_json.get('config')
                    if qos_configs:
                        congestion_enabled_tc_lst = list(qos_configs.keys())
                        congestion_enabled_tc = ','.join(congestion_enabled_tc_lst)
                        qos_config = list(qos_configs.values())[0]
                        flow = list(qos_config.get('flow', {}).values())
                        if flow:
                            congestion_control_map['enabled_tc'] = congestion_enabled_tc
                            ecn_enabled = qos_config.get('ecn_enabled', 0) == 1
                            red_enabled = qos_config.get('red_enabled', 0) == 1

                            mode = 'None'
                            if ecn_enabled and not red_enabled:
                                mode = 'ECN'
                            elif not ecn_enabled and red_enabled:
                                mode = 'RED'
                            elif ecn_enabled and red_enabled:
                                mode = 'ECN+RED'

                            congestion_control_map['mode'] = mode
                            congestion_control_map['max_th'] = flow[0].get('max', 0)
                            congestion_control_map['min_th'] = flow[0].get('min', 0)
                            break

        except subprocess.CalledProcessError:
            pass

        return congestion_control_map

    def get_ets_priority_mode(self) -> str:
        dwrr_mode = ''

        if os.path.exists('/bin/systemctl'):
            is_active = is_active_systemd_svc('sx_sdk.service')
            if not is_active:
                return dwrr_mode

        try:
            roce_scheduler_info = subprocess_wrapper('/usr/lib/cumulus/mlxcmd --json scheduler show'.split())
            roce_scheduler_json = json.loads(roce_scheduler_info) #pylint: disable=no-member
        except ValueError:
            return dwrr_mode
        scheduler = roce_scheduler_json.get('list')

        if not scheduler:
            scheduler = roce_scheduler_json.get('scheduler-elements')

        if scheduler:
            roce_scheduler = next((item for item in list(scheduler) if (item.get('level') == 2 and item.get('index') == 3)), None)
            if roce_scheduler is not None:
                dwrr_mode = roce_scheduler.get('dwrr_mode')

        return dwrr_mode

def process_roce(cmd_output, svc_failed=False, **kwargs):
    roce_render.parse(cmd_output, svc_failed, **kwargs)

def process_roce_config(cmd_output, svc_failed=False, **kwargs):
    roce_render.parse_roce_config(cmd_output, svc_failed, **kwargs)

def process_nvue_roce_config(cmd_output, svc_failed=False, **kwargs):
    roce_render.parse_nvue_qos_roce(cmd_output, svc_failed, **kwargs)

roce_render = CumulusRoceRender()
