#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#



import ujson as json

from netq_lib.orm.redisdb.models import (
    Temp,
    Fan,
    PSU,
    Power)
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.sensors import SensorRender, SensorPowerRender, FAIL_TEMP, FAIL_FAN
# local_sensor_dict = dict()

dummy_sensor = [
    [
        {
            "s_power_in_input": 0.0,
            "s_voltage_in_max": 0.0,
            "s_voltage_out_input": 0.0,
            "s_power_out_max": 0.0,
            "s_power_in_max": 675.0,
            "s_voltage_in_input": 0.0,
            "s_voltage_out_max": 0.0,
            "s_adapter_name": "cumulus-vx-i2c-10-59",
            "s_voltage_in_min": 0.0,
            "s_voltage_out_min": 0.0,
            "s_power_out_input": 0.0
        },
        {
            "s_power_in_input": 72.375,
            "s_voltage_in_max": 0.0,
            "s_voltage_out_input": 11.957,
            "s_power_out_max": 0.0,
            "s_power_in_max": 675.0,
            "s_voltage_in_input": 214.0,
            "s_voltage_out_max": 0.0,
            "s_adapter_name": "cumulus-vx-i2c-10-58",
            "s_voltage_in_min": 0.0,
            "s_voltage_out_min": 0.0,
            "s_power_out_input": 57.5
        }
    ],
    [
        {
            "s_power_in_input": 3.0,
            "s_voltage_in_max": 0.0,
            "s_voltage_out_input": 230.0,
            "s_power_out_max": 0.0,
            "s_power_in_max": 175.0,
            "s_voltage_in_input": 200.0,
            "s_voltage_out_max": 0.0,
            "s_adapter_name": "cumulus40-vx-i2c-10-59",
            "s_voltage_in_min": 0.0,
            "s_voltage_out_min": 0.0,
            "s_power_out_input": 2.0
        },
        {
            "s_power_in_input": 72.375,
            "s_voltage_in_max": 0.0,
            "s_voltage_out_input": 11.957,
            "s_power_out_max": 0.0,
            "s_power_in_max": 375.0,
            "s_voltage_in_input": 218.0,
            "s_voltage_out_max": 0.0,
            "s_adapter_name": "cumulus40-vx-i2c-10-58",
            "s_voltage_in_min": 0.0,
            "s_voltage_out_min": 0.0,
            "s_power_out_input": 57.5
        }
    ]
]


class CumulusSensorRender(SensorRender):
    def __init__(self): #pylint: disable=useless-super-delegation
        super(CumulusSensorRender, self).__init__()

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get('fullstate')
        if svc_failed:
            return ServiceStatus.FAIL
        else:
            try:
                s_list = json.loads(cmd_output) #pylint: disable=no-member
            except ValueError:
                return ServiceStatus.FAIL

        temp_arr = []
        psu_arr = []
        fan_arr = []
        for s in s_list:
            sensor = None
            name = s['name']
            if not name:
                continue
            if s['type'] == 'temp':
                sensor = Temp()
                sensor.s_input = (float(s['input']) if s.get('input') is not None
                                  else float(FAIL_TEMP))
                sensor.s_crit = int(s['crit']) if s.get('crit') is not None else FAIL_TEMP
                sensor.s_max = int(s['max']) if s.get('max') is not None else FAIL_TEMP
                sensor.s_min = int(s['min']) if s.get('min') is not None else FAIL_TEMP
                sensor.s_lcrit = int(s['lcrit']) if s.get('lcrit') is not None else FAIL_TEMP
                sensor.s_desc = s['description'].lower()
                # Naga: Temporarily comment below lines to have stream all the time
                #if local_sensor_dict.get(name):
                #    prev_val, prev_st = local_sensor_dict[name]
                #    if ((cur_st == prev_st) and
                #            (abs(sensor.s_input - prev_val) < 2.5)):
                #        continue
                #local_sensor_dict[name] = (sensor.s_input, cur_st)
            elif s['type'] == 'fan':
                sensor = Fan()
                sensor.s_input = int(s['input']) if s.get('input') is not None else FAIL_FAN
                sensor.s_max = int(s['max']) if s.get('max') is not None else FAIL_FAN
                sensor.s_min = int(s['min']) if s.get('min') is not None else FAIL_FAN
                sensor.s_desc = s['description'].lower()
                # Naga: Temporarily comment below lines to have stream all the time
                #if local_sensor_dict.get(name):
                #    prev_val, prev_st = local_sensor_dict[name]
                #    if ((cur_st == prev_st) and
                #            abs(sensor.s_input - prev_val) < s_max_val):
                #        continue
                #local_sensor_dict[name] = (sensor.s_input, cur_st)
            elif s['type'] == 'power':
                sensor = PSU()

            if sensor is not None:
                sensor.s_name = s['name'].lower()
                if not s['msg']:
                    sensor.s_msg = ''
                else:
                    msg = s['msg']

                    msgs = msg.split('\n')
                    m = []
                    for msg in msgs:
                        if 'Unable to find driver path' in msg:
                            # ignore this msg
                            continue
                        if 'Unable to find pwm path' in msg:
                            # ignore this msg
                            continue
                        m.append(msg.split(':', 1)[1])
                    if m:
                        sensor.s_msg = ','.join(m)
                    else:
                        sensor.s_msg = ''
                sensor.s_state = s['state'].lower()
                sensor.s_prev_state = s['prev_state'].lower()

                if fullstate:
                    if sensor.__class__.__name__ == "Temp":
                        temp_arr.append(sensor)
                    elif sensor.__class__.__name__ == "Fan":
                        fan_arr.append(sensor)
                    else:
                        psu_arr.append(sensor)
                else:
                    sensor.save()
        if fullstate:
            save_fullstate(temp_arr, "Temp", False)
            save_fullstate(fan_arr, "Fan", False)
            save_fullstate(psu_arr, "PSU", False)

        return ServiceStatus.OK


def process_sensors(output, svc_failed=False, **kwargs):
    render = CumulusSensorRender()
    return render.parse(output, svc_failed=False, **kwargs)


class CumulusSensorPowerRender(SensorPowerRender):
    def __init__(self): #pylint: disable=useless-super-delegation
        super(CumulusSensorPowerRender, self).__init__()

    def handle_vx_platform(self, random_platform, power_arr, fullstate):
        if random_platform:
            random_index = int(random_platform) % 2
            random_power = dummy_sensor[random_index]
            for power_info in random_power:
                power_obj = Power()
                for k, v in list(power_info.items()):
                    setattr(power_obj, k, v)
                if not fullstate:
                    power_obj.save()
                power_arr.append(power_obj)

    def psu_found_check(self, psu_output_line):
        if 'vin' in psu_output_line or 'pin' in psu_output_line:
            return True
        return False

    def voltage_in_check(self, psu_output_line):
        if psu_output_line.startswith("vin"):
            return True
        return False

    def voltage_out_check(self, psu_output_line):
        if psu_output_line.startswith("vout"):
            return True
        return False

    def power_in_check(self, psu_output_line):
        if psu_output_line.startswith('pin'):
            return True
        return False

    def power_out_check(self, psu_output_line):
        if psu_output_line.startswith("pout"):
            return True
        return False


def process_power(cmd_output, svc_failed=False, **kwargs):
    render = CumulusSensorPowerRender()
    return render.parse(cmd_output, svc_failed, **kwargs)
