#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


import math
import re

from netq_agent.parsers.common.dom import DomRender
from netq_lib.common.enums import ServiceStatus
from netq_lib.orm.redisdb.models import Dom

class SonicDomRender(DomRender):

    def _get_command(self):
        return "show interfaces transceiver eeprom --dom "

    def _process_command_output(self, cmd_out):
        result = dict()
        prev_index = None
        prev_key = None
        for eth_tool_dom in cmd_out.op.splitlines():
            eth_tool_dom = eth_tool_dom.decode()
            if ":" not in eth_tool_dom:
                continue

            key, value = eth_tool_dom.split(':', 1)
            key = key.strip()
            value = value.strip()

            index = eth_tool_dom.find(key)
            if not prev_index:
                prev_index = index
                prev_key = key
            if index > prev_index:
                if result[prev_key]:
                    result[prev_key].update({key: value})
                else:
                    result[prev_key] = {key: value}
            else:
                prev_index = prev_index
                prev_key = key
                result[key] = value

        return result, ServiceStatus.OK

    @staticmethod
    def _get_laser_pwr_string(dBm_str):
        if dBm_str == 'n/a':
            return dBm_str
        dBm_val = float(dBm_str.strip('dBm'))
        mw_val = float("{:.4f}".format(1 * math.pow(10, dBm_val / 10)))
        return "{} mW / {} dBm".format(mw_val, dBm_val)

    @staticmethod
    def _get_temperature_string(c_temp_str):
        if c_temp_str == 'n/a':
            return c_temp_str
        c_temp_val = float(c_temp_str.strip('C'))
        f_temp_val = (c_temp_val * 9/5) + 32
        return "{} degrees C / {} degrees F".format(c_temp_val, f_temp_val)

    def _create_dom_object(self, ifname, eth_elems):
        dom_obj = Dom()
        dom_obj.ifname = ifname
        dom_obj.identifier = eth_elems.get('Identifier', 'n/a')

        dom_obj.laser_rx_power = {}
        dom_obj.laser_output_power = {}
        dom_obj.laser_bias_current = {}

        try:
            for ch_attr, attr_value in eth_elems.get('ChannelMonitorValues').items():
                if re.search('RX(.+?)Power', ch_attr):
                    ch_index = re.match('RX(.+?)Power', ch_attr).group(1)
                    dom_obj.laser_rx_power['Channel {}'.format(ch_index)] = self._get_laser_pwr_string(attr_value)
                elif re.search('TX(.+?)Power', ch_attr):
                    ch_index = re.match('TX(.+?)Power', ch_attr).group(1)
                    dom_obj.laser_output_power['Channel {}'.format(ch_index)] = self._get_laser_pwr_string(attr_value)
                elif re.search('TX(.+?)Bias', ch_attr):
                    ch_index = re.match('TX(.+?)Bias', ch_attr).group(1)
                    dom_obj.laser_bias_current['Channel {}'.format(ch_index)] = attr_value.strip('mA') + ' mA'
                else:
                    self._logger.debug("Skipped attribute {} under ChannelMonitorValues".format(ch_attr))

            dom_obj.module_temp = self._get_temperature_string(eth_elems.get('ModuleMonitorValues', {}).get('Temperature', 'n/a'))
            dom_obj.module_voltage = eth_elems.get('ModuleMonitorValues', {}).get('Vcc', 'n/a').strip('Volts') + ' V'

            dom_obj.laser_rx_power_high_alarm_th = self._get_laser_pwr_string(eth_elems.get('ChannelThresholdValues', {}).get('RxPowerHighAlarm', 'n/a'))
            dom_obj.laser_rx_power_low_alarm_th = self._get_laser_pwr_string(eth_elems.get('ChannelThresholdValues', {}).get('RxPowerLowAlarm', 'n/a'))
            dom_obj.laser_rx_power_high_warning_th = self._get_laser_pwr_string(eth_elems.get('ChannelThresholdValues', {}).get('RxPowerHighWarning', 'n/a'))
            dom_obj.laser_rx_power_low_warning_th = self._get_laser_pwr_string(eth_elems.get('ChannelThresholdValues', {}).get('RxPowerLowWarning', 'n/a'))

            dom_obj.laser_output_power_high_alarm_th = 'n/a'    # eth_elems.get('Laser output power high alarm threshold', 'n/a')
            dom_obj.laser_output_power_low_alarm_th = 'n/a'     # eth_elems.get('Laser output power low alarm threshold', 'n/a')
            dom_obj.laser_output_power_high_warning_th = 'n/a'  # eth_elems.get('Laser output power high warning threshold', 'n/a')
            dom_obj.laser_output_power_low_warning_th = 'n/a'   # eth_elems.get( 'Laser output power low warning threshold', 'n/a')

            dom_obj.laser_bias_current_high_alarm_th = eth_elems.get('ChannelThresholdValues', {}).get('TxBiasHighAlarm', 'n/a').strip('mA') + ' mA'
            dom_obj.laser_bias_current_low_alarm_th = eth_elems.get('ChannelThresholdValues', {}).get('TxBiasLowAlarm', 'n/a').strip('mA') + ' mA'
            dom_obj.laser_bias_current_high_warning_th = eth_elems.get('ChannelThresholdValues', {}).get('TxBiasHighWarning', 'n/a').strip('mA') + ' mA'
            dom_obj.laser_bias_current_low_warning_th = eth_elems.get('ChannelThresholdValues', {}).get('TxBiasLowWarning', 'n/a').strip('mA') + ' mA'

            dom_obj.module_temp_high_alarm_th = self._get_temperature_string(eth_elems.get('ModuleThresholdValues', {}).get('TempHighAlarm', 'n/a'))
            dom_obj.module_temp_low_alarm_th = self._get_temperature_string(eth_elems.get('ModuleThresholdValues', {}).get('TempLowAlarm', 'n/a'))
            dom_obj.module_temp_high_warning_th = self._get_temperature_string(eth_elems.get('ModuleThresholdValues', {}).get('TempHighWarning', 'n/a'))
            dom_obj.module_temp_low_warning_th = self._get_temperature_string(eth_elems.get('ModuleThresholdValues', {}).get('TempLowWarning', 'n/a'))

            dom_obj.module_voltage_high_alarm_th = eth_elems.get('ModuleThresholdValues', {}).get('VccHighAlarm', 'n/a').strip('Volts') + ' V'
            dom_obj.module_voltage_low_alarm_th = eth_elems.get('ModuleThresholdValues', {}).get('VccLowAlarm', 'n/a').strip('Volts') + ' V'
            dom_obj.module_voltage_high_warning_th = eth_elems.get('ModuleThresholdValues', {}).get('VccHighWarning', 'n/a').strip('Volts') + ' V'
            dom_obj.module_voltage_low_warning_th = eth_elems.get('ModuleThresholdValues', {}).get( 'VccLowWarning', 'n/a').strip('Volts') + ' V'

            return dom_obj
        except AttributeError:
            self._logger.debug("Failed to fetch dom information for {}".format(ifname))
            return None


def process_dom(output, svc_failed=False, **kwargs):
    dom.process_dom_interfaces(kwargs)


# Creating Dom object, which holds the context for dom processing
dom = SonicDomRender()
