#
# Copyright (c) 2017-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import OS, ASIC, Board, License
from netq_lib.common.enums import Platform
from netq_agent.parsers.common.platform import PlatformRender, OK


class SonicPlatformRender(PlatformRender):
    def __init__(self):
        super(SonicPlatformRender, self).__init__()

    def process_os_info(self, logger, os_release):
        os = OS()
        os.name = 'N/A'
        os.version = 'N/A'
        os.version_id = 'N/A'

        try:
            if os_release is not None:
                for line in os_release.splitlines():
                    if 'SONiC Software Version:' in line:
                        os.name = Platform.SONIC
                        os.version_id = line.split(':')[1].strip()
                        os.version = os.version_id.split('.')[1]
        except (IndexError, ValueError) as ex:
            logger.error("Failed to process os info: {}".format(ex))
            raise ex

        return os


    def process_asic_info(self, logger, board, show_version, platform_detect):
        asic = ASIC()

        asic.vendor = 'N/A'
        asic.model = 'N/A'
        asic.model_id = 'N/A'
        asic.core_bw = 'N/A'
        asic.ports = 'N/A'

        try:
            if show_version is not None:
                for line in show_version.splitlines():
                    if line.startswith('ASIC:'):
                        asic.vendor = line.split(':')[1].strip().capitalize()

        except (IndexError, ValueError) as ex:
            logger.error("Failed to process platform information: {}".format(ex))
            raise ex

        return asic

    def process_os_license_info(self, cmd_output, svc_status=OK, **kwargs):
        """
        fullstate = kwargs.get('fullstate')
        license_obj = License()
        license_obj.name = Platform.SONIC

        if platform_module.OLD_SVC_STATUS != svc_status or fullstate:
            license_obj.license = ''
            license_obj.state = 'N/A'
            license_dict = {'name': Platform.CL, 'license': ''}
            platform_module.OLD_SVC_STATUS = svc_status
            if not fullstate:
                license_obj.save()
            else:
                license_obj_arr = []
                license_obj_arr.append(license_obj)
                save_fullstate(license_obj_arr, License().__class__.__name__)
        return license_obj, license_dict
        """
        return None, None

    def process_syseeprom_info(self, logger, decode_syseeprom):
        board = Board()
        board.vendor = 'N/A'
        board.model = 'N/A'
        board.base_mac = 'N/A'
        board.part_number = 'N/A'
        board.mfg_date = 'N/A'
        board.serial_number = 'N/A'
        board.label_revision = 'N/A'

        if decode_syseeprom:
            """
            TLV Name             Code Len Value
            -------------------- ---- --- -----
            """
            header_str = "TLV Name             Code Len Value"
            header = header_str.split()
            idx = 0
            matches_positions = []
            table_margin = 0
            result = dict()
            required_fields = ['Product Name', 'Part Number', 'Serial Number', 'Base MAC Address', 'Manufacture Date',
                               'Device Version', 'Manufacturer']  # 'Vendor Name']
            for row in decode_syseeprom.split('\n'):
                if idx == 0:
                    if header_str not in row:
                        continue
                    if header != row.split():
                        logger.error("Invalid header 'show platform syseeprom'")
                        break
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif not len(row):
                    pass    # allow increment of idx
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].strip() for i in range(len(matches_positions)-1)]
                    if data[0] in required_fields:
                        try:
                            result[data[0]] = data[3]
                        except IndexError:
                            logger.error("Syseeprom info: IndexError Exception caught while retrieving inforation from {}".format(row))
                            pass    # allow increment of idx
                idx = idx + 1
            #board.vendor = result.get('Vendor Name', 'N/A')
            board.vendor = result.get('Manufacturer', 'N/A')
            board.model = result.get('Product Name', 'N/A')
            board.base_mac = result.get('Base MAC Address', 'N/A')
            board.part_number = result.get('Part Number', 'N/A')
            board.mfg_date = result.get('Manufacture Date', 'N/A')
            board.serial_number = result.get('Serial Number', 'N/A')
            board.label_revision = result.get('Device Version', 'N/A')

        return board

    def _inventory_data_valid(self, os, board, asic):
        m_os = OS.memmodel()
        old_os_objs = list(m_os.query.filter(hostname=HOSTNAME))
        if old_os_objs:
            old_os = old_os_objs[0]
            if old_os.name != 'N/A' and os.name == 'N/A':
                return False

        m_board = Board.memmodel()
        old_board_objs = list(m_board.query.filter(hostname=HOSTNAME))
        if old_board_objs:
            old_board = old_board_objs[0]
            if old_board.vendor != 'N/A' and board.vendor == 'N/A':
                return False

        m_asic = ASIC.memmodel()
        old_asic_objs = list(m_asic.query.filter(hostname=HOSTNAME))
        if old_asic_objs:
            old_asic = old_asic_objs[0]
            if old_asic.vendor != 'N/A' and asic.vendor == 'N/A':
                return False

        return True

def process_platform_info(
        logger,
        random_platform,
        platform,
        os_release,
        decode_syseeprom,
        lscpu,
        meminfo,
        show_version,
        meminfo_dmi,
        lsblk,
        agent_version,
        platform_detect,
        is_full_state,
        app_release_info,
        service_nvue_enabled,
        service_nvue_active,
        is_nvue_config_saved,
        service_nclu_enabled
):
    render = SonicPlatformRender()
    render.process_platform_info(logger, random_platform, platform,
                                 os_release, decode_syseeprom, lscpu,
                                 meminfo, show_version, meminfo_dmi,
                                 lsblk, agent_version, platform_detect,
                                 is_full_state, app_release_info,
                                 service_nvue_enabled, service_nvue_active,
                                 is_nvue_config_saved, service_nclu_enabled)
