#
# Copyright (c) 2020-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
import os
import subprocess
from netq_lib.common.utils import subprocess_wrapper
from netq_agent.parsers.common.ports import PortsRender


class SonicPortsRender(PortsRender):
    def __init__(self):
        super(SonicPortsRender, self).__init__()
        self._status_cmd_result = dict()
        self._sfp_port_map = dict()

    @staticmethod
    def _get_ethtool_path():
        return '/usr/bin/ethtool'

    def _build_sfp_port_map(self):
        # show version
        try:
            show_cmd_out = subprocess_wrapper('show platform summary'.split())
        except subprocess.CalledProcessError:
            self._logger.warn("Failed to execute command - show platform summary")
            return
        platform = None
        hwsku = None
        if show_cmd_out:
            try:
                for row in show_cmd_out.split('\n'):
                    if 'Platform' in row:
                        platform = row.split(':')[1].strip()
                    if 'HwSKU' in row:
                        hwsku = row.split(':')[1].strip()
            except IndexError:
                self._logger.warn("Failed to fetch Platform and HwSKU information from show platform summary output")
                return

        # Access the port ini file
        path = '/usr/share/sonic/device/{}/{}/port_config.ini'.format(platform, hwsku)
        spf_index = None
        lanes_index = None
        lanes_to_index_map = dict()
        idx = 0
        if os.path.exists(path):
            with open(path, 'r') as fd:
                for row in fd:
                    if idx == 0:
                        try:
                            if '#' in row:
                                header = row.strip('#').split()
                                spf_index = header.index('index')
                                lanes_index = header.index('lanes')

                        except ValueError:
                            self._logger.warn("Unable to process port_config.ini file: {}".format(path))
                            return
                    else:
                        row = row.split()
                        lanes_to_index_map[row[lanes_index]] = row[spf_index]
                    idx = idx + 1
        else:
            self._logger.warn("File not found {}".format(path))
            return

        for port_name, result in self._status_cmd_result.items():   # show interface status
            lanes = result['Lanes']
            for lanes_str, index in lanes_to_index_map.items():     # port_config.ini
                if lanes in lanes_str:
                    self._sfp_port_map[port_name] = 'sfp{}'.format(index)

        #self._logger.info("EthernetX to sfpX map : {}".format(self._sfp_port_map))


    def ethtool_fetch_port_info(self, port_name):
        if 'eth' in port_name:
            return super(SonicPortsRender, self).ethtool_fetch_port_info(port_name)
        else:
            if not self._status_cmd_result:
                try:
                    show_cmd_dump = subprocess_wrapper('show interface status'.split())
                except subprocess.CalledProcessError:
                    return dict()

                if show_cmd_dump:
                    #     Interface    Lanes    Speed    MTU    FEC       Alias           Vlan    Oper    Admin            Type    Asym PFC
                    # -------------  -------  -------  -----  -----  ----------  -------------  ------  -------  --------------  ----------
                    header = "    Interface    Lanes    Speed    MTU    FEC       Alias           Vlan    Oper    Admin            Type    Asym PFC".split()
                    idx = 0
                    matches_positions = []
                    table_margin = 0
                    for row in show_cmd_dump.split("\n"):
                        if idx == 0:
                            if header != row.split():
                                self._logger.error("Command 'show interface status' header validation failed")
                                break
                        elif idx == 1:
                            matches = re.finditer('  ', row)
                            matches_positions = [match.start() for match in matches]
                            matches_positions.append(None)
                            matches_positions.insert(0,0)
                            table_margin = len(row)
                        elif len(row) != table_margin:
                            self._logger.warn("Invalid row {}".format(row))
                        else:
                            result = dict()
                            data = [row[matches_positions[i]:matches_positions[i+1]].lstrip() for i in range(len(matches_positions)-1)]
                            result['Speed'] = data[2]
                            result['FEC encodings'] = data[4]

                            if data[7] == 'up':
                                result['Link detected'] = 'yes'
                            elif data[7] == 'down':
                                result['Link detected'] = 'no'
                            else:
                                result['Link detected'] = None

                            result['Lanes'] = data[1]

                            # result['Port'] = None
                            # result['Supported ports'] = None
                            # result['Auto-negotiation'] = None
                            # result['Supported FEC modes'] = None
                            # result['Advertised FEC modes'] = None
                            # result['Link detected'] = None
                            self._status_cmd_result[data[0]] = result
                        idx = idx + 1
                self._build_sfp_port_map()

            return self._status_cmd_result.get(port_name, dict())

    def ethtool_fetch_test_serial_number(self, port_name):
        test_serial_number = None
        if 'eth' in port_name:
            try:
                test_serial_number = subprocess_wrapper(
                    [self._ethtool_bin_path, '-m', port_name, 'offset', '168', 'length',
                     '16', 'raw', 'on'])
            except subprocess.CalledProcessError:
                test_serial_number = None
            except UnicodeDecodeError:
                test_serial_number = None
        else:
            test_serial_number = port_name  # TODO
        return test_serial_number

    def ethtool_fetch_eeprom_optical_info(self, port_name):
        if 'eth' in port_name:
            return super(SonicPortsRender, self).ethtool_fetch_eeprom_optical_info(port_name)

        result = dict()
        try:
            show_cmd_dump = subprocess_wrapper(
                ['show', 'interfaces', 'transceiver', 'eeprom', port_name])
        except subprocess.CalledProcessError:
            self._logger.info("Failed to execute show interfaces transceiver eeprom {}".format(port_name))
            return dict()

        prev_index = None
        prev_key = None
        app_adv_content = False
        for line in show_cmd_dump.splitlines():
            if not line:
                continue
            if ":" in line:
                key, value = line.split(':', 1)
            elif prev_key == "Application Advertisement":
                key = "Application Advertisement"
                app_adv_content = True
                value = line.strip()
            else:
                continue
            key = key.strip()
            value = value.strip()
            if value == 'SFP EEPROM Not detected':
                return dict()
            if value == 'SFP EEPROM detected':
                continue
            if key == 'Connector':
                app_adv_content = False
            index = line.find(key)

            if key == 'Vendor Name':
                key = 'Vendor name'
            elif key == 'Length Cable Assembly(m)':
                key = 'length'
                value = value + 'm'
            elif key == 'LengthOM3(UnitsOf10m)':
                key = 'length'
                value_in_mtrs = (float(value) if '.' in value else int(value))*10
                value = str(value_in_mtrs) + 'm'

            if not prev_index:
                prev_index = index
                prev_key = key

            if index > prev_index or app_adv_content:
                if prev_key in result:
                    result[prev_key] = result[prev_key] + ', ' + line.strip()
                else:
                    result[prev_key] = line.strip()
            else:
                prev_index = prev_index
                prev_key = key
                result[key] = value

        try:
            if 'Not supported for CMIS cables' not in result['Specification compliance']:
                for txt in result['Specification compliance'].split(','):
                    if txt.split(':', 1)[0].strip() in ['10/40G Ethernet Compliance Code',
                                                        'Extended Specification compliance',
                                                        'Gigabit Ethernet Compliant codes',
                                                        'EthernetComplianceCodes']:
                        if 'Transceiver type' in result:
                            result['Transceiver type'] = result['Transceiver type'] + ', ' + txt.split(':', 1)[1].strip().replace(' or', ',')
                        else:
                            result['Transceiver type'] = txt.split(':', 1)[1].strip().replace(' or', ',')
            elif result['Application Advertisement'] != 'N/A':
                result['Transceiver type'] = result['Application Advertisement']
        except (IndexError, KeyError):
            self._logger.debug("Failed to fetch Transceiver type for {}".format(port_name))

        return result

    def reset_status_cmd_result(self):
        self._status_cmd_result = dict()

    def skip_port_platform_specific(self, port_name):
        return port_name.startswith('sfp')

    @staticmethod
    def _is_optical_port(ethtool_m_dict):
        if not ethtool_m_dict:
            return False
        if 'Transceiver type' in ethtool_m_dict and \
                ('CR' in ethtool_m_dict['Transceiver type'] or '1000BASE-CX' in ethtool_m_dict['Transceiver type']):
            return False
        return True


def process_ports(cmd_output, svc_failed=False, **kwargs):
    ports_render.parse(cmd_output, svc_failed, **kwargs)
    ports_render.reset_status_cmd_result()


ports_render = SonicPortsRender()
