#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
import sys
import math

from netq_agent.parsers.common.proc import ProcRender
from netq_lib.orm.redisdb.models import EthtoolInterface


if sys.version_info > (3,):
    long = int


class SonicProcRender(ProcRender):
    def __init__(self):
        super(SonicProcRender, self).__init__()
        # self.err_data = self.read_interface_err_counters()
        self.first_run = True
        self.batch_size = 0
        self.total_buckets = 1
        self.cmd_periodicity = 0
        self.if_name_list = None
        self.PROC_PROCESS_DURATION = 30
        self.bucket_index = 0

    def get_interfaces(self, **kwargs):
        if self.bucket_index == 0:
            self.cmd_periodicity = kwargs.get('period', self.PROC_PROCESS_DURATION)
            if_names = kwargs.get('ports', None)
            self.if_name_list = [i for i in if_names if i.startswith('Ethernet') or i.startswith('eth')]
            if self.first_run:
                self.total_buckets = 1
                self.batch_size = len(self.if_name_list)
                self.first_run = False
            else:
                if self.PROC_PROCESS_DURATION < self.cmd_periodicity:
                    self.total_buckets = (self.cmd_periodicity // self.PROC_PROCESS_DURATION)
                self.batch_size = math.ceil(len(self.if_name_list) / self.total_buckets)

        start_idx = self.bucket_index * self.batch_size
        end_idx = start_idx + self.batch_size
        # Update bucket index
        self.bucket_index += 1
        if self.bucket_index == self.total_buckets:
            self.bucket_index = 0
        # Boundary check for indexes
        if start_idx >= len(self.if_name_list):
            process_if_list = list()
            return process_if_list
        elif end_idx > len(self.if_name_list):
            end_idx = len(self.if_name_list)
        process_if_list = self.if_name_list[start_idx:end_idx]
        # Process error data too.
        self.err_data = self.read_interface_err_counters()
        return process_if_list

    def get_command(self, ifname):
        if ifname.startswith('Ethernet'):
            return "show interfaces counters detailed {}".format(ifname)
        elif ifname.startswith('eth'):
            return "ethtool -S {}".format(ifname)
        else:
            self._logger.warn("Unknown interface {} to fetch interface counters".format(ifname))
            return None

    def read_interface_err_counters(self):
        err_data = {}
        cmd = 'show interfaces counters errors'
        result = self.run_command(cmd)
        if result.success and result.op:
            #      IFACE    STATE    RX_ERR    RX_DRP    RX_OVR    TX_ERR    TX_DRP    TX_OVR
            # ----------  -------  --------  --------  --------  --------  --------  --------
            header = "     IFACE    STATE    RX_ERR    RX_DRP    RX_OVR    TX_ERR    TX_DRP    TX_OVR".split()
            idx = 0
            matches_positions = []
            table_margin = 0
            for row in result.op.decode().split("\n"):
                if idx == 0:
                    if header != row.split():
                        self._logger.error("Command '{}' header validation failed".format(cmd))
                        return dict()
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif len(row) != table_margin:
                    #self._logger.warn("Invalid row - {}".format(row))
                    continue
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].strip() for i in range(len(matches_positions)-1)]
                    err_data[data[0]] = dict(
                        rx_drop=long(data[3].replace(',', '')),
                        rx_errs=long(data[2].replace(',', '')),
                        tx_drop=long(data[5].replace(',', '')),
                        tx_errs=long(data[6].replace(',', ''))
                    )
                idx = idx + 1
        return err_data

    def create_eth(self, ifname, hw_if_in_ucast_pkts=-1, hw_if_in_bcast_pkts=-1, hw_if_in_mcast_pkts=-1,
                   hw_if_in_discards=-1, hw_if_in_errors=-1, hw_if_out_ucast_pkts=-1, hw_if_out_mcast_pkts=-1,
                   hw_if_out_bcast_pkts=-1, hw_if_out_discards=-1, hw_if_out_errors=-1):
        eth_obj = EthtoolInterface()
        eth_obj.ifname = ifname
        eth_obj.hw_if_in_ucast_pkts = hw_if_in_ucast_pkts
        eth_obj.hw_if_in_bcast_pkts = hw_if_in_bcast_pkts
        eth_obj.hw_if_in_mcast_pkts = hw_if_in_mcast_pkts
        eth_obj.hw_if_in_discards = hw_if_in_discards
        eth_obj.hw_if_in_errors = hw_if_in_errors
        eth_obj.hw_if_out_ucast_pkts = hw_if_out_ucast_pkts
        eth_obj.hw_if_out_mcast_pkts = hw_if_out_mcast_pkts
        eth_obj.hw_if_out_bcast_pkts = hw_if_out_bcast_pkts
        eth_obj.hw_if_out_discards = hw_if_out_discards
        eth_obj.hw_if_out_errors = hw_if_out_errors
        return eth_obj

    def process_stats(self, ifname, retcode, cmd_out):
        if retcode == 0:
            if ifname.startswith('eth'):
                super(SonicProcRender, self).process_stats(ifname, retcode, cmd_out)
            if ifname.startswith('Ethernet'):
                rx_ucast = rx_mcast = rx_bcast = tx_ucast = tx_mcast = tx_bcast = -1

                for row in cmd_out.op.decode().split("\n"):
                    try:
                        if row.startswith('Unicast Packets Received'):
                            val = row.split('. ')[1]
                            rx_ucast = long(val.replace(',', ''))
                        elif row.startswith('Multicast Packets Received'):
                            val = row.split('. ')[1]
                            rx_mcast = long(val.replace(',', ''))
                        elif row.startswith('Broadcast Packets Received'):
                            val = row.split('. ')[1]
                            rx_bcast = long(val.replace(',', ''))
                        elif row.startswith('Unicast Packets Transmitted'):
                            val = row.split('. ')[1]
                            tx_ucast = long(val.replace(',', ''))
                        elif row.startswith('Multicast Packets Transmitted'):
                            val = row.split('. ')[1]
                            tx_mcast = long(val.replace(',', ''))
                        elif row.startswith('Broadcast Packets Transmitted'):
                            val = row.split('. ')[1]
                            tx_bcast = long(val.replace(',', ''))
                    except IndexError:
                        self._logger.debug("Error while processing counters detailed output row: {}".format(row))
                        continue

                eth_obj = self.create_eth(ifname,
                                          hw_if_in_ucast_pkts=rx_ucast, hw_if_in_bcast_pkts=rx_bcast, hw_if_in_mcast_pkts=rx_mcast,
                                          hw_if_in_discards=self.err_data.get(ifname, {}).get('rx_drop') if self.err_data.get(ifname, {}).get('rx_drop') is not None else -1,
                                          hw_if_in_errors=self.err_data.get(ifname, {}).get('rx_errs') if self.err_data.get(ifname, {}).get('rx_errs') is not None else -1,
                                          hw_if_out_ucast_pkts=tx_ucast, hw_if_out_mcast_pkts=tx_mcast, hw_if_out_bcast_pkts=tx_bcast,
                                          hw_if_out_discards=self.err_data.get(ifname, {}).get('tx_drop') if self.err_data.get(ifname, {}).get('tx_drop') is not None else -1,
                                          hw_if_out_errors=self.err_data.get(ifname, {}).get('tx_errs') if self.err_data.get(ifname, {}).get('tx_errs') is not None else -1)
                eth_obj.save()

        proc_obj = self.assign_procnetdev_value(ifname, self.procnetdev)
        if proc_obj:
            proc_obj.save()


def process_proc_net_dev(output, svc_failed=False, **kwargs):
    """
    Parse /proc/net/dev counters
    """
    proc.procnetdev = {}    # Reset counters read from /proc/net/dev before every run
    return proc.parse(output, svc_failed, **kwargs)


proc = SonicProcRender()
