#
# Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import greenlet

from swsssdk import ConfigDBConnector
from netq_lib.common.utils import is_sonic_ib_containerized


class SonicDB(object):
    APPL_DB = "APPL_DB"
    ASIC_DB = "ASIC_DB"
    COUNTERS_DB = "COUNTERS_DB"
    CONFIG_DB = "CONFIG_DB"
    STATE_DB = "STATE_DB"


class SonicTable(object):

    # APPL_DB Tables START
    ######################
    APPL_DB_ROUTE_TABLE = "ROUTE_TABLE"
    APPL_DB_INTF_TABLE = "INTF_TABLE"
    APPL_DB_IB_PORT_TABLE = "IB_PORT_TABLE"
    APPL_DB_ALIAS_PORT_MAP = "ALIAS_PORT_MAP"
    # APPL_DB Tables END
    ####################

    # STATE_DB Tables START
    #######################
    STATE_DB_INTERFACE_TABLE = "INTERFACE_TABLE"
    STATE_DB_TRANSCEIVER_DOM_SENSOR = "TRANSCEIVER_DOM_SENSOR"
    STATE_DB_IB_PORT_TABLE = "IB_PORT_TABLE"
    # STATE_DB Tables END
    #####################

    # CONFIG_DB Tables START
    ######################
    CONFIG_DB_IB_PORT = "IB_PORT"
    # CONFIG_DB Tables END
    ####################

    # COUNTERS_DB Tables START
    ##########################
    COUNTERS_DB_COUNTERS_PORT_NAME_MAP = "COUNTERS_PORT_NAME_MAP"
    COUNTERS_DB_COUNTERS = "COUNTERS"
    # COUNTERS_DB Tables END
    ########################

    # ASIC_DB Tables START
    ##########################
    ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_FDB_ENTRY = "ASIC_STATE:SAI_OBJECT_TYPE_FDB_ENTRY"
    ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_VLAN = "ASIC_STATE:SAI_OBJECT_TYPE_VLAN"
    ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_BRIDGE_PORT = "ASIC_STATE:SAI_OBJECT_TYPE_BRIDGE_PORT"
    # ASIC_DB Tables END
    ########################

def sonic_data_get(data, db_name, table_name):
    return data.get(db_name, {}).get(table_name, {})

class SonicDBConnector(ConfigDBConnector):
    REDIS_SCAN_BATCH_SIZE = 30

    def __init__(self, **kwargs):
        super(SonicDBConnector, self).__init__(use_unix_socket_path=True if is_sonic_ib_containerized() else False, **kwargs)

    def __get_config(self, client, pipe, table, data, cursor):
        """Read config data in batches of size REDIS_SCAN_BATCH_SIZE using Redis pipelines
        Args:
            client: Redis client
            pipe: Redis DB pipe
            data: config dictionary
            cursor: position to start scanning from
        Returns:
            cur: poition of next item to scan
        """
        pattern = '{}{}*'.format(table, self.TABLE_NAME_SEPARATOR)
        cur, keys = client.scan(cursor=cursor, match=pattern, count=self.REDIS_SCAN_BATCH_SIZE)
        keys = [key for key in keys if key != self.INIT_INDICATOR]
        for key in keys:
            pipe.hgetall(key)
        records = pipe.execute()

        for index, key in enumerate(keys):
            #(table_name, row) = key.split(self.TABLE_NAME_SEPARATOR, 1)
            row = key.split('{}{}'.format(table, self.TABLE_NAME_SEPARATOR))[-1]
            entry = self.raw_to_typed(records[index])
            if entry is not None:
                #data.setdefault(table_name, {})[self.deserialize_key(row)] = entry
                data.setdefault(table, {})[row] = entry
        return cur

    def get_table(self, table):
        """Read an entire table from config db.
        Args:
            table: Table name.
        Returns:
            Table data in a dictionary form of
            { 'row_key': {'column_key': value, ...}, ...}
            or { ('l1_key', 'l2_key', ...): {'column_key': value, ...}, ...} for a multi-key table.
            Empty dictionary if table does not exist.
        """
        client = self.get_redis_client(self.db_name)
        pipe = client.pipeline()
        data = {}
        cur = 0
        while True:
            cur = self.__get_config(client, pipe, table, data, cur)
            if not cur:
                break

        if not data and self.exists(self.db_name, table):
            data[table] = self.get_all(self.db_name, table)

        return data


class SonicDBRender(object):
    MODEL = None    # Define in derived class

    def __init__(self):
        self._logger = None

    def process(self, logger):
        self._logger = logger
        db_sources = self.db_sources()
        data = {}
        # TODO: Fetch using threads
        db = SonicDBConnector()
        for (db_name, table) in db_sources:
            db.db_connect(db_name)
            self._logger.debug("Fetch data from table {}:{}".format(db_name, table))
            table = db.get_table(table)
            if data.get(db_name):
                data[db_name].update(table)
            else:
                data.update({db_name: table})
        # process retrieved data
        self.process_fullstate(data)

    def handle_periodic_timeout(self):
        raise NotImplementedError

    def db_sources(self):
        raise NotImplementedError

    def process_fullstate(self):
        raise NotImplementedError


class SonicPerDBClient(object):
    def __init__(self, db_name, pool, logger):
        self._logger = logger
        self.db_name = db_name
        self.db = SonicDBConnector()
        self.db.db_connect(self.db_name)
        self.db_callbacks = dict()
        self._thread_pool = pool
        self._listener_thread = None
        self.table_with_separators = dict()

    def __stop_monitor_checker(self, green_thread):
        try:
            green_thread.wait()
        except greenlet.GreenletExit:
            pass
        except Exception as ex:
            self._logger.exception('Monitor thread terminated unexpectedly. Error: %s', ex)
        finally:
            if green_thread == self._listener_thread:
                self._logger.info('Listener thread ended for {}'.format(self.db_name))
                self._listener_thread = None

    def base_callback(self, table, key, value):
        self._logger.debug('Callback invoked for table {}:{}'.format(table, key))
        if table in self.table_with_separators:
            match = None
            for subkey in self.table_with_separators.get(table):
                if subkey in key:
                    if not match:
                        match = subkey
                    else:
                        if subkey.count(self.db.TABLE_NAME_SEPARATOR) > match.count(self.db.TABLE_NAME_SEPARATOR):
                            match = subkey
            if not match:
                self._logger.debug('table name derivation failed for {}:{}'.format(table, key))
                return
            table = '{}{}{}'.format(table, self.db.TABLE_NAME_SEPARATOR, match)
            key = key.split(match)[-1]
            key = key.lstrip(self.db.TABLE_NAME_SEPARATOR)
            #self._logger.info('Update table and key information {}, {}'.format(table, key))

        for handler in self.db_callbacks.get(table, []):
            # TODO: Scheduling on callback, to avoid overwrite
            green_thread = self._thread_pool.spawn(handler, table, key, value)
            green_thread.link(self.__stop_monitor_checker)

    def watch_table(self, table, handler):
        if table in self.db_callbacks:
            if handler not in self.db_callbacks[table]:
                self.db_callbacks[table] = self.db_callbacks[table].append(handler)
        else:
            self.db_callbacks[table] = [handler]
            if self.db.TABLE_NAME_SEPARATOR in table:
                key, subkey = table.split(self.db.TABLE_NAME_SEPARATOR, 1)
                self.db.subscribe(key, self.base_callback)
                self.table_with_separators.setdefault(key, []).append(subkey)

    def is_listener_running(self):
        return self._listener_thread is not None

    def _listen(self):
        self._logger.info('Starting listener for {}'.format(self.db_name))
        try:
            self.db.listen()
        except Exception as e:
            self._logger.error("failed to listen to client radis db table {}, please verify that syncd is running: {}".format(self.db_name, e))

    def start_listener(self):
        if not self._listener_thread:
            self._listener_thread = self._thread_pool.spawn(self._listen)
            self._listener_thread.link(self.__stop_monitor_checker)

    def stop_listener(self):
        self._logger.info("Stop listener thread {}".format(self.db_name))
        if self._listener_thread:
            self._listener_thread.kill()


class SonicDBSubscribeManager(object):
    def __init__(self, pool, logger):
        self.per_db_client = dict()
        self._logger = logger
        self._thread_pool = pool
        self._registered_models = list()

    def is_model_registered(self, model):
        return model in self._registered_models

    def register_callback(self, render):
        # Check if Model is already registered
        if self.is_model_registered(render.MODEL.__name__):
            return

        # Subscribe for DB sources
        db_sources = render.db_sources()
        for (db_name, table), handler in db_sources.items():
            if db_name not in self.per_db_client:
                self.per_db_client[db_name] = SonicPerDBClient(db_name, self._thread_pool, self._logger)
                self._logger.debug('PerDB Client for {} added'.format(db_name))
            self.per_db_client[db_name].watch_table(table, handler)
            if not self.per_db_client[db_name].is_listener_running():
                self._logger.debug('Starting Thread for {}'.format(db_name))
                self.per_db_client[db_name].start_listener()

        # Mark the model as registered
        self._registered_models.append(render.MODEL.__name__)
        self._logger.debug('Registered Models {}'.format(self._registered_models))

    def exit_gracefully(self):
        for _, db_client in self.per_db_client.items():
            db_client.stop_listener()
