#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
import math

from netq_agent.parsers.sonic.redis_db.client import SonicDBConnector, SonicDB, SonicTable, SonicDBRender, sonic_data_get
from netq_lib.common.utils import save_fullstate
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import Dom


class DomHandler(SonicDBRender):
    MODEL = Dom

    def __init__(self):
        super(DomHandler, self).__init__()
        self.first_run = True
        self._optical_ports = set()
        self._optical_ports_curr = set()

    def db_sources(self):
        return {
            (SonicDB.STATE_DB, SonicTable.STATE_DB_TRANSCEIVER_DOM_SENSOR): self.state_db_transceiver_dom_sensor
        }

    def state_db_transceiver_dom_sensor(self, table, key, value):
        pass

    @staticmethod
    def _get_temperature_string(c_temp_str):
        if c_temp_str == 'N/A':
            return c_temp_str
        c_temp_val = float(c_temp_str)
        f_temp_val = (c_temp_val * 9/5) + 32
        return "{} degrees C / {} degrees F".format(c_temp_val, f_temp_val)

    @staticmethod
    def _get_laser_pwr_string(dBm_str):
        if dBm_str == 'N/A':
            return dBm_str
        dBm_val = float(dBm_str)
        mw_val = float("{:.4f}".format(1 * math.pow(10, dBm_val / 10)))
        return "{} mW / {} dBm".format(mw_val, dBm_val)

    @staticmethod
    def _value_exists(val_str):
        if not val_str or val_str == 'N/A':
            return False
        return True

    def _create_dom_object(self, ifname, info):
        try:
            dom_obj = Dom()
            dom_obj.ifname = ifname
 
            dom_obj.laser_rx_power = {}
            dom_obj.laser_output_power = {}
            dom_obj.laser_bias_current = {}
 
            for attr, value in info.items():
                if re.search('^rx(.+?)power$', attr):
                    ch_index = re.match('rx(.+?)power', attr).group(1)
                    dom_obj.laser_rx_power['Channel {}'.format(ch_index)] = self._get_laser_pwr_string(value)
                elif re.search('^tx(.+?)power$', attr):
                    ch_index = re.match('tx(.+?)power', attr).group(1)
                    dom_obj.laser_output_power['Channel {}'.format(ch_index)] = self._get_laser_pwr_string(value)
                elif re.search('^tx(.+?)bias$', attr):
                    ch_index = re.match('tx(.+?)bias', attr).group(1)
                    dom_obj.laser_bias_current['Channel {}'.format(ch_index)] = 'N/A' if not self._value_exists(value) else value + ' mA'
 
            dom_obj.module_temp = self._get_temperature_string(info.get('temperature', 'N/A'))
            dom_obj.module_voltage = 'N/A' if not self._value_exists(info.get('voltage')) else info.get('voltage') + ' V'
 
            dom_obj.laser_rx_power_high_alarm_th = self._get_laser_pwr_string(info.get('rxpowerhighalarm', 'N/A'))
            dom_obj.laser_rx_power_low_alarm_th = self._get_laser_pwr_string(info.get('rxpowerlowalarm', 'N/A'))
            dom_obj.laser_rx_power_high_warning_th = self._get_laser_pwr_string(info.get('rxpowerhighwarning', 'N/A'))
            dom_obj.laser_rx_power_low_warning_th = self._get_laser_pwr_string(info.get('rxpowerlowwarning', 'N/A'))
 
            dom_obj.laser_output_power_high_alarm_th = self._get_laser_pwr_string(info.get('txpowerhighalarm', 'N/A'))
            dom_obj.laser_output_power_low_alarm_th = self._get_laser_pwr_string(info.get('txpowerlowalarm', 'N/A'))
            dom_obj.laser_output_power_high_warning_th = self._get_laser_pwr_string(info.get('txpowerhighwarning', 'N/A'))
            dom_obj.laser_output_power_low_warning_th = self._get_laser_pwr_string(info.get('txpowerlowwarning', 'N/A'))
 
            dom_obj.laser_bias_current_high_alarm_th = 'N/A' if not self._value_exists(info.get('txbiashighalarm')) else info.get('txbiashighalarm') + ' mA'
            dom_obj.laser_bias_current_low_alarm_th = 'N/A' if not self._value_exists(info.get('txbiaslowalarm')) else info.get('txbiaslowalarm') + ' mA'
            dom_obj.laser_bias_current_high_warning_th = 'N/A' if not self._value_exists(info.get('txbiashighwarning')) else info.get('txbiashighwarning') + ' mA'
            dom_obj.laser_bias_current_low_warning_th = 'N/A' if not self._value_exists(info.get('txbiaslowwarning')) else info.get('txbiaslowwarning') + ' mA'
 
            dom_obj.module_temp_high_alarm_th = self._get_temperature_string(info.get('temphighalarm', 'N/A'))
            dom_obj.module_temp_low_alarm_th = self._get_temperature_string(info.get('templowalarm', 'N/A'))
            dom_obj.module_temp_high_warning_th = self._get_temperature_string(info.get('temphighwarning', 'N/A'))
            dom_obj.module_temp_low_warning_th = self._get_temperature_string(info.get('templowwarning', 'N/A'))
 
            dom_obj.module_voltage_high_alarm_th = 'N/A' if not self._value_exists(info.get('vcchighalarm')) else info.get('vcchighalarm') + ' V'
            dom_obj.module_voltage_low_alarm_th = 'N/A' if not self._value_exists(info.get('vcclowalarm')) else info.get('vcclowalarm') + ' V'
            dom_obj.module_voltage_high_warning_th = 'N/A' if not self._value_exists(info.get('vcchighwarning')) else info.get('vcchighwarning') + ' V'
            dom_obj.module_voltage_low_warning_th = 'N/A' if not self._value_exists(info.get('vcclowwarning')) else info.get('vcclowwarning') + ' V'
            return dom_obj

        except KeyError as ex:
            self._logger.info("Failed to fetch dom information for {}, reason {}".format(ifname, ex))
            return None

    def update_optical_ports(self, optical_ports):
        self._optical_ports_curr = optical_ports

    def handle_outdated_objs(self):
        m_dom = Dom.memmodel()
        if self._optical_ports_curr != self._optical_ports:
            # if the current optical ports are different
            for entry in m_dom.query.filter(hostname=HOSTNAME):
                if entry.ifname not in self._optical_ports_curr:
                    old_dom_ent = Dom()
                    old_dom_ent.copy(entry)
                    old_dom_ent.delete()

        self._optical_ports = self._optical_ports_curr

    def process_fullstate(self, data):
        self.handle_outdated_objs()
        fullstate_arr = []
        for ifname, dom_info in sonic_data_get(data, SonicDB.STATE_DB, SonicTable.STATE_DB_TRANSCEIVER_DOM_SENSOR).items():
            if ifname in self._optical_ports:
                dom_obj = self._create_dom_object(ifname, dom_info)
                if dom_obj:
                    dom_obj.save()
                    if self.first_run:
                        fullstate_arr.append(dom_obj)
        if self.first_run:
            save_fullstate(fullstate_arr, Dom().__class__.__name__)
            self.first_run = False


def process_dom(*args, **kwargs):
    logger = kwargs.get('logger')
    optical_ports = kwargs.get('optical_ports', None)
    render.update_optical_ports(optical_ports)
    render.process(logger)


render = DomHandler()
