#
# Copyright (c) 2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import json
import eventlet
import greenlet
import inspect

from netq_lib.orm.objects import Field
from netq_agent.parsers.sonic.redis_db.client import SonicDBConnector, SonicDB, SonicTable, SonicDBRender, sonic_data_get
from netq_lib.orm.redisdb.models import MacFdb
from netq_lib.common.nlfullstate import NetlinkFullStateObject


class MacFdbHandler(SonicDBRender):
    MODEL = MacFdb

    def __init__(self):
        super(MacFdbHandler, self).__init__()
        # TODO: Decide if Lock required 
        self.vlan_table = {}
        self.port_map = {}
        self.bridge_port_table = {}
        self.periodic_thread = None
        self._redis_notification_keys = dict()
        self.json_fields = []
        for _, field in inspect.getmembers(MacFdb,
                                           lambda member: not inspect.isroutine(member)):
            if isinstance(field, Field):
                self.json_fields.append(field)

    def __stop_monitor_checker(self, green_thread):
        try:
            green_thread.wait()
        except greenlet.GreenletExit:
            pass
        except Exception as ex:
            self._logger.exception('Monitor thread terminated unexpectedly. Error: %s', ex)
        finally:
            if green_thread == self.periodic_thread:
                self._logger.debug('Periodic handler thread ended for MacFdb')
                self.periodic_thread = None

    def db_sources(self):
        return {
            (SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_FDB_ENTRY):
                self.asic_db_asic_state_sai_object_type_fdb_entry,
            (SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_VLAN):
                self.asic_db_asic_state_sai_object_type_vlan,
            (SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_BRIDGE_PORT):
                self.asic_db_asic_state_sai_object_type_bridge_port,
            (SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS_PORT_NAME_MAP):
                self.counters_db_counters_port_name_map,
        }

    def _fetch_table_data(self, db_name, table):
        db = SonicDBConnector()
        db.db_connect(db_name)
        data = db.get_table(table)
        return sonic_data_get(data, db_name, table)

    def _cache_redis_notification_keys(self, table, key):
        self._redis_notification_keys[(table, key)] = True

    def _get_redis_notification_keys(self):
        return self._redis_notification_keys

    def _reset_redis_notification_keys(self):
        self._redis_notification_keys.clear()

    def asic_db_asic_state_sai_object_type_fdb_entry(self, table, key, value):
        while self.periodic_thread:
            eventlet.sleep(0.01)
        self._cache_redis_notification_keys(table, key)

    def asic_db_asic_state_sai_object_type_vlan(self, table, key, value):
        pass

    def asic_db_asic_state_sai_object_type_bridge_port(self, table, key, value):
        pass

    def counters_db_counters_port_name_map(self, table, key, value):
        pass

    @staticmethod
    def _create_mac(key_json, info, vlan, nexthop):
        mac = MacFdb()
        mac.mac_address = key_json.get('mac')
        mac.vlan = vlan
        mac.nexthop = nexthop
        mac.is_static = info.get('SAI_FDB_ENTRY_ATTR_TYPE', 'SAI_FDB_ENTRY_TYPE_DYNAMIC') == 'SAI_FDB_ENTRY_TYPE_STATIC'
        mac.port = '0'
        return mac

    def _retrieve_vlan_details(self, key_json):
        vlan = None
        bvid = key_json['bvid']
        
        if bvid not in self.vlan_table:
            # Attempt to read DB
            # TODO: handle in callback instead
            self.vlan_table = self._fetch_table_data(SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_VLAN)

        try:
            vlan = int(self.vlan_table.get(bvid, {}).get('SAI_VLAN_ATTR_VLAN_ID'))
        except TypeError:
            return None

        return vlan

    def _retrieve_nexthop_details(self, fdb_info):
        nexthop = None

        if not fdb_info:
            return None

        bridge_port_id = fdb_info.get('SAI_FDB_ENTRY_ATTR_BRIDGE_PORT_ID')
        if bridge_port_id not in self.bridge_port_table:
            # Update from DB
            # TODO: handle in callback instead
            self.bridge_port_table = self._fetch_table_data(SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_BRIDGE_PORT)


        bridge_port = self.bridge_port_table.get(fdb_info.get('SAI_FDB_ENTRY_ATTR_BRIDGE_PORT_ID'))
        if not bridge_port:
            return None

        port_id = bridge_port.get('SAI_BRIDGE_PORT_ATTR_PORT_ID')
        if port_id not in self.port_map:
            # Update from DB
            # TODO: handle in callback instead
            self.port_map = self._fetch_table_data(SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS_PORT_NAME_MAP)
            self.port_map = {v: k for k, v in self.port_map.items()}

        nexthop = self.port_map.get(bridge_port.get('SAI_BRIDGE_PORT_ATTR_PORT_ID'), '')
        if not nexthop:
            return None

        return nexthop

    def process_fullstate(self, data):
        self.port_map = sonic_data_get(data, SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS_PORT_NAME_MAP)
        self.port_map = {v: k for k, v in self.port_map.items()}
        self.vlan_table = sonic_data_get(data, SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_VLAN)
        self.bridge_port_table = sonic_data_get(data, SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_BRIDGE_PORT)

        for fdb_key, info in sonic_data_get(data, SonicDB.ASIC_DB, SonicTable.ASIC_DB_ASIC_STATE_SAI_OBJECT_TYPE_FDB_ENTRY).items():
            key_json = json.loads(fdb_key)
            vlan = self._retrieve_vlan_details(key_json)
            if not vlan:
                continue
            nexthop = self._retrieve_nexthop_details(info)
            mac_obj = self._create_mac(key_json, info, vlan, nexthop)
            if mac_obj:
                mac_obj.save_memmodel(self.json_fields)
                NetlinkFullStateObject.push_message_queue(mac_obj, self.json_fields)

        # Send fullstate
        NetlinkFullStateObject.flush_fullstate('MacFdb')
        # reset any cached notifications
        self._reset_redis_notification_keys()

    def _periodic_handler(self):
        db = SonicDBConnector()
        db.db_connect(SonicDB.ASIC_DB)
        client = db.get_redis_client(SonicDB.ASIC_DB)
        self._logger.debug("Periodic handler processing redis keys count {}".format(len(self._get_redis_notification_keys())))
        for key in self._get_redis_notification_keys():
            # Fetch the information from redis DB.
            info = db.raw_to_typed(client.hgetall(':'.join(key)))
            key_json = json.loads(key[1])
            vlan = self._retrieve_vlan_details(key_json)
            if not vlan:
                continue
            nexthop = self._retrieve_nexthop_details(info)
            mac_obj = self._create_mac(key_json, info, vlan, nexthop)
            if mac_obj:
                if not info:
                    mac_obj.delete()
                else:
                    mac_obj.save()
        # reset any cached notifications
        self._reset_redis_notification_keys()

    def handle_periodic_timeout(self, pool):
        if self.periodic_thread:
            self._logger.debug('Periodic execution already in progress for MacFdb, skip periodic handling')
            return

        self.periodic_thread = pool.spawn(self._periodic_handler)
        self.periodic_thread.link(self.__stop_monitor_checker)


def process_macfdb(*args, **kwargs):
    logger = kwargs.get('logger')
    full_state = kwargs.get('fullstate')
    if full_state:
        logger.debug('fullstate process_macfdb invoked')
        render.process(logger)
        # Subscribe - Done only during Init
        subscribe_manager = kwargs.get('subscribe_manager')
        subscribe_manager.register_callback(render)
    else:
        logger.debug('periodic process_macfdb invoked')
        pool = kwargs.get('pool')
        # Read from MemModel and push over gRPC
        render.handle_periodic_timeout(pool)


render = MacFdbHandler()
