#
# Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

from netq_agent.parsers.sonic.redis_db.client import SonicDBConnector, SonicDB, SonicTable, SonicDBRender, sonic_data_get
from netq_lib.orm.redisdb.models import IbInterface, Link

class IbInterfaceHandler(SonicDBRender):
    MODEL = IbInterface

    def __init__(self):
        super(IbInterfaceHandler, self).__init__()

    def db_sources(self):
        return {
            (SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS_PORT_NAME_MAP): self.counters_db_counters_port_name_map,
            (SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS): self.counters_db_counters,
            (SonicDB.STATE_DB, SonicTable.STATE_DB_IB_PORT_TABLE): self.state_db_ib_port_table,
            (SonicDB.CONFIG_DB, SonicTable.CONFIG_DB_IB_PORT): self.config_db_ib_port_table,
            (SonicDB.APPL_DB, SonicTable.APPL_DB_IB_PORT_TABLE): self.appl_db_ib_port_table,
        }

    def counters_db_counters_port_name_map(self):
        pass

    def counters_db_counters(self):
        pass

    def state_db_ib_port_table(self):
        pass

    def config_db_ib_port_table(self):
        pass

    def appl_db_ib_port_table(self):
        pass

    def _attr_counter_db_key_mapping(self):
        return {
            'logical_port_state': 'SAI_PORT_STAT_INFINIBAND_LOGICAL_STATE',
            'physical_port_state': 'SAI_PORT_STAT_INFINIBAND_PHYSICAL_STATE',
            'ib_speed': 'SAI_PORT_STAT_INFINIBAND_SPEED_OPER',
            'speed': 'SAI_PORT_STAT_INFINIBAND_SPEED_OPER',
            'width': 'SAI_PORT_STAT_INFINIBAND_LANES_OPER',
            'mtu': 'SAI_PORT_STAT_INFINIBAND_MTU_OPER',
            'operational_vls': 'SAI_PORT_STAT_INFINIBAND_VL_OPER',
            'in_pkts': 'SAI_PORT_STAT_INFINIBAND_IF_IN_PKTS_EXT',
            'out_pkts': 'SAI_PORT_STAT_INFINIBAND_IF_OUT_PKTS_EXT',
            'in_errors': 'SAI_PORT_STAT_INFINIBAND_PC_ERR_RCV_F',
            'out_errors': 'SAI_PORT_STAT_INFINIBAND_ERR_XMTCONSTR_F',
            'in_octets': 'SAI_PORT_STAT_INFINIBAND_IF_IN_OCTETS_EXT',
            'out_octets': 'SAI_PORT_STAT_INFINIBAND_IF_OUT_OCTETS_EXT',
            'in_discards': 'SAI_PORT_STAT_INFINIBAND_PC_VL15_DROPPED_F',
            'out_discards': 'SAI_PORT_STAT_INFINIBAND_PC_XMT_DISCARDS_F',
            'port_symbol_errors': 'SAI_PORT_STAT_INFINIBAND_PC_ERR_SYM_F',
            'port_rcv_errors': 'SAI_PORT_STAT_INFINIBAND_PC_ERR_RCV_F',
            'port_vl15_dropped': 'SAI_PORT_STAT_INFINIBAND_PC_VL15_DROPPED_F',
            'port_xmit_wait': 'SAI_PORT_STAT_INFINIBAND_IF_OUT_WAIT',
        }

    def _attr_state_db_key_mapping(self):
        return {
            'supported_ib_speeds': 'speed_admin',
            'supported_widths': 'lanes_admin',
            'max_supported_MTUs': 'mtu_max',
            'ib_subnet': 'ib_subnet',
            'vl_capabilities': 'vl_admin',
            'oper_status': 'oper_status'
        }

    def _attr_config_db_key_mapping(self):
        return {
            'ifname': 'alias',
            'speed_negotiate': 'auto_neg',
            'admin_status': 'admin_status',
            'description': 'description',
            'ifindex': 'index',
        }

    def update_ib_interface_obj(self, ib_interface_obj, data, attr_map):
        if data:
            for attr_name, db_key in attr_map.items():
                try:
                    setattr(ib_interface_obj, attr_name, data.get(db_key, ""))
                except (ValueError, KeyError) as ex:
                    self._logger.debug(
                        'Failed to set attr {} for ib_interface_obj object'.format(attr_name))
            pass

    def _create_ib_interface_obj(self, counter_db_data, state_db_data, config_db_data, appl_db_data):
        ib_interface_obj = IbInterface()
        self.update_ib_interface_obj(ib_interface_obj, counter_db_data, self._attr_counter_db_key_mapping())
        self.update_ib_interface_obj(ib_interface_obj, state_db_data, self._attr_state_db_key_mapping())
        self.update_ib_interface_obj(ib_interface_obj, config_db_data, self._attr_config_db_key_mapping())
        return ib_interface_obj

    def process_fullstate(self, data):
        port_map = sonic_data_get(data, SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS_PORT_NAME_MAP)  # ifname -> oid
        if not port_map:
            self._logger.error('Failed to retrieve COUNTERS_PORT_NAME_MAP')
            return
        for ifname, oid in port_map.items():
            counter_db_record = sonic_data_get(data, SonicDB.COUNTERS_DB, SonicTable.COUNTERS_DB_COUNTERS).get(oid)
            state_db_record = sonic_data_get(data, SonicDB.STATE_DB, SonicTable.STATE_DB_IB_PORT_TABLE).get(ifname)
            config_db_record = sonic_data_get(data, SonicDB.CONFIG_DB, SonicTable.CONFIG_DB_IB_PORT).get(ifname)
            appl_db_record = sonic_data_get(data, SonicDB.APPL_DB, SonicTable.APPL_DB_IB_PORT_TABLE).get(ifname)

            ib_interface_record = self._create_ib_interface_obj(counter_db_record, state_db_record,
                                                                config_db_record, appl_db_record)
            if ib_interface_record:
                self._logger.debug(ib_interface_record)
                ib_interface_record.save()

def process_counters(*args, **kwargs):
    logger = kwargs.get('logger')
    render.process(logger)

render = IbInterfaceHandler()