#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import re
import json

from netq_lib.orm.redisdb.models import (
    Temp,
    Fan,
    PSU
)
from netq_lib.common.enums import ServiceStatus
from netq_lib.common.utils import save_fullstate
from netq_agent.parsers.common.sensors import SensorRender, SensorPowerRender, FAIL_TEMP, FAIL_FAN


class SonicSensorRender(SensorRender):
    def __init__(self):
        super(SonicSensorRender, self).__init__()

    def process_fan_sensors(self, fullstate, logger):
        fan_arr = []
        result = self.run_command("show platform fan")
        if result.success and result.op:
            # #         FAN     Speed    Direction     Presence    Status          Timestamp
            # # -----------  --------  -----------  -----------  --------  -----------------
            # header = "        FAN     Speed    Direction     Presence    Status          Timestamp".split()
            #  Drawer    LED          FAN    Speed    Direction     Presence    Status          Timestamp
            # --------  -----  -----------  -------  -----------  -----------  --------  -----------------
            header = "  Drawer    LED          FAN    Speed    Direction     Presence    Status          Timestamp".split()
            idx = 0
            matches_positions = []
            table_margin = 0
            for row in result.op.decode().split("\n"):
                if idx == 0:
                    if 'Fan Not detected' in row:
                        logger.debug('Fan sensors not detected')
                        return
                    if header != row.split():
                        logger.error("Command 'show platform fan' header validation failed. Details {}".format(row))
                        return
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif len(row) == 0:
                    idx += 1
                    continue
                elif len(row) != table_margin:
                    logger.warn("Invalid row {}".format(row))
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].lstrip() for i in range(len(matches_positions)-1)]
                    name = data[2]
                    status = data[6]
                    min_speed = FAIL_FAN
                    max_speed = FAIL_FAN
                    current_speed = FAIL_FAN
                    r_min_speed = self.run_command("cat /var/run/hw-management/thermal/{}_min".format(name))
                    if r_min_speed.success and r_min_speed.op:
                        min_speed = int(r_min_speed.op.decode().strip())
                    r_max_speed = self.run_command("cat /var/run/hw-management/thermal/{}_max".format(name))
                    if r_max_speed.success and r_max_speed.op:
                        max_speed = int(r_max_speed.op.decode().strip())
                    r_current_speed = self.run_command("cat /var/run/hw-management/thermal/{}_speed_get".format(name))
                    if r_current_speed.success and r_current_speed.op:
                        current_speed = int(r_current_speed.op.decode().strip())
                    sensor = Fan()
                    sensor.s_name = name
                    sensor.s_state = status.lower()
                    sensor.s_input = current_speed
                    sensor.s_max = max_speed
                    sensor.s_min = min_speed

                    sensor.save()
                    if fullstate:
                        fan_arr.append(sensor)

                idx += 1

            if fullstate:
                save_fullstate(fan_arr, Fan.__name__)


    def process_temp_sensors(self, fullstate, logger):
        temp_arr = []
        result = self.run_command("show platform temperature")
        if result.success and result.op:
            #                Sensor    Temperature    High TH    Low TH    Crit High TH    Crit Low TH    Warning          Timestamp
            # ---------------------  -------------  ---------  --------  --------------  -------------  ---------  -----------------
            header = "               Sensor    Temperature    High TH    Low TH    Crit High TH    Crit Low TH    Warning          Timestamp".split()
            idx = 0
            matches_positions = []
            table_margin = 0
            for row in result.op.decode().split("\n"):
                if idx == 0:
                    if 'Thermal Not detected' in row:
                        logger.debug('Temperature sensors not detected')
                        return
                    if header != row.split():
                        logger.error("Command 'show platform temperature' header validation failed. "
                                     "Details {}".format(row))
                        return
                elif idx == 1:
                    matches = re.finditer('  ', row)
                    matches_positions = [match.start() for match in matches]
                    matches_positions.append(None)
                    matches_positions.insert(0,0)
                    table_margin = len(row)
                elif len(row) == 0:
                    idx += 1
                    continue
                elif len(row) != table_margin:
                    logger.warn("Invalid row - {}".format(row))
                else:
                    data = [row[matches_positions[i]:matches_positions[i+1]].lstrip() for i in range(len(matches_positions)-1)]

                    if not (data[1] != 'N/A' or data[2] != 'N/A' or data[3] != 'N/A'
                            or data[4] != 'N/A' or data[5] != 'N/A'):
                        logger.debug('Skipped Temp Sensor {} as no data is available'.format(data[0]))
                        continue

                    sensor = Temp()
                    sensor.s_name = data[0]
                    sensor.s_input = float(data[1]) if data[1] != "N/A" else float(FAIL_TEMP)
                    sensor.s_max = int(float(data[2])) if data[2] != "N/A" else FAIL_TEMP
                    sensor.s_min = int(float(data[3])) if data[3] != "N/A" else FAIL_TEMP
                    sensor.s_crit = int(float(data[4])) if data[4] != "N/A" else FAIL_TEMP
                    sensor.s_lcrit = int(float(data[5])) if data[5] != "N/A" else FAIL_TEMP
                    if data[1] != 'N/A':
                        if data[2] != 'N/A' and sensor.s_input > sensor.s_max:
                            sensor.s_state = 'not ok'
                        elif data[3] != 'N/A' and sensor.s_input < sensor.s_min:
                            sensor.s_state = 'not ok'
                        elif data[6] != 'False':
                            sensor.s_state = 'not ok'
                        else:
                            sensor.s_state = 'ok'
                    else:
                        sensor.s_state = 'N/A'

                    sensor.save()

                    if fullstate:
                        temp_arr.append(sensor)
                idx += 1
            if fullstate:
                save_fullstate(temp_arr, Temp.__name__)


    def process_psu_sensors(self, fullstate, logger):
        psu_arr = []
        result = self.run_command("show platform psustatus --json")
        if result.success and result.op:
            try:
                output = json.loads(result.op.decode())
            except JSONDecodeError as e:
                logger.info('Failed to process psustatus information')
                return
            for item in output:
                sensor = PSU()
                sensor.s_name = item.get('name')
                sensor.s_state = item.get('status')
                sensor.save()
                if fullstate:
                    psu_arr.append(sensor)
            if fullstate:
                save_fullstate(psu_arr, PSU.__name__)

    def parse(self, cmd_output, svc_failed=False, **kwargs):
        fullstate = kwargs.get('fullstate')
        logger = kwargs.get('logger')
        if svc_failed:
            return ServiceStatus.FAIL

        self.process_fan_sensors(fullstate, logger)
        self.process_temp_sensors(fullstate, logger)
        self.process_psu_sensors(fullstate, logger)
        return ServiceStatus.OK


def process_sensors(cmd_output, svc_failed=False, **kwargs):
    render = SonicSensorRender()
    return render.parse(cmd_output, svc_failed, **kwargs)


class SonicSensorPowerRender(SensorPowerRender):
    def __init__(self):
        super(SonicSensorPowerRender, self).__init__()

    def psu_found_check(self, psu_output_line):
        if 'Rail (in)' in psu_output_line or 'Pwr (in)' in psu_output_line:
            return True
        return False

    def voltage_in_check(self, psu_output_line):
        if psu_output_line.endswith("Rail (in):"):
            return True
        return False

    def voltage_out_check(self, psu_output_line):
        if psu_output_line.endswith("Rail (out):"):
            return True
        return False

    def power_in_check(self, psu_output_line):
        if psu_output_line.endswith("Pwr (in):"):
            return True
        return False

    def power_out_check(self, psu_output_line):
        if psu_output_line.endswith("Pwr (out):"):
            return True
        return False


def process_power(cmd_output, svc_failed=False, **kwargs):
    render = SonicSensorPowerRender()
    return render.parse(cmd_output, svc_failed, **kwargs)

