#
# Copyright (c) 2018-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

from unittest import TestCase
import os
import time
import eventlet
import gevent
import traceback
from gevent import monkey
from gevent.queue import Full
monkey.patch_all

AF_BRIDGE = 6

from netq_lib.common import configuration
from netq_lib.common.enums import DefaultPaths, ConfKw, AGENTSTATUS
from netq_agent.cmd.agent import _ClDbPush
from netq_agent.cmd.netq_agent_discovery import NetQAgentDiscovery, ChannelState
from netq_agent.cmd.netq_agent_stream import NetQAgentStream, NetQAgentServicer
#from netq_agent.cmd.netq_pq import NetQPriorityQueue, QueueElem


from concurrent import futures
from threading import Thread

# GRPC Specfic Imports
# GRPC Client imports
import grpc
import netq_agent.interface.netq_pb2
from netq_agent.interface.netq_pb2_grpc import NetqAgentStreamServiceStub

# GRPC Server IMPORTS
from netq_agent.interface.netq_reg_pb2 import NetqAgentDiscoveryResp, ResponseStatus
from netq_agent.interface.netq_reg_pb2_grpc import add_NetqAgentDiscoveryServiceServicer_to_server, NetqAgentDiscoveryServiceServicer
from netq_agent.interface.netq_pb2 import NetqOPTARequest

# Logging
import logging
logger = logging.getLogger('netq-agent')
consoleHandler = logging.StreamHandler()
logFormatter = logging.Formatter(
    "%(asctime)s [%(filename)s:%(lineno)d] [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
consoleHandler.setFormatter(logFormatter)
logger.addHandler(consoleHandler)


# Test Utilities
from netq_agent.tests.netq_agent.lldp_msg import LLDP_STR

_PROC_NAME = 'netq_agent'


class NetQAgentTestOptaClient(object):

    def __init__(self, conf_dict, logger):
        self._conf = conf_dict
        self.agent_addr = self.get_agent_addr()
        self._logger = logger
        try:
            self._logger.info(
                "Sending OPTA Request to agent {}".format(self.agent_addr))
            self._channel = grpc.insecure_channel(self.agent_addr)
        except Exception as e:
            self._logger.error(
                "Unable to establish connection with OPTA {}".format(e))
            return AGENTSTATUS.FAILURE

    def get_agent_addr(self):
        serv_addr = "{}:{}".format(self._conf['netq-agent'].get(ConfKw.AGENT_URL, ConfKw.DEFAULT_AGENT_URL),
                                   self._conf['netq-agent'].get(ConfKw.AGENT_PORT, ConfKw.DEFAULT_AGENT_PORT))
        return serv_addr

    def send_start_stream(self):
        pass

    def send_stop_stream(self):
        pass

    def send_opta_req(self):
        req_msg = NetqOPTARequest()

        netq_agent_service_stub = NetqAgentStreamServiceStub(self._channel)

        try:
            rsp_message = netq_agent_service_stub.StreamMessage(req_msg)
            self._logger.info("OPTA Message Request sent")
        except grpc.RpcError as e:
            self._logger.error("GRPC Error Received {}".format(e))
            return AGENTSTATUS.FAILURE
        except Exception as e:
            self._logger.error("Unknown GRPC Exception {}".format(e))
            return AGENTSTATUS.FAILURE

        logger.info("Rsp {}".format(rsp_message))

        return rsp_message


class OptaServicer(NetqAgentDiscoveryServiceServicer):

    def __init__(self, service):
        self.service = service
        self.running_count = 0

    def build_discover_response(self, request, context):

        self.running_count = self.running_count + 1
        status = 'Success'
        rsp_message = NetqAgentDiscoveryResp(status=status,
                                             reconnect_opta_addr="127.0.0.1")

        logger.info("OPTA Response Dispatched")
        return rsp_message

    # This is a Inteface Message
    def DiscoverMessage(self, request, context):
        logger.info("Discover Message Received ")
        rsp_msg = self.build_discover_response(request, context)
        return rsp_msg


class OptaService(object):

    def __init__(self, conf_dict):
        self.url = conf_dict['netq-agent'][ConfKw.SERVER]
        self.port = conf_dict['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)
        self.wait_time = 60 * 60 * 24
        self.server = None

    def start_service(self):

        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        import pdb
        pdb.set_trace()

        add_NetqAgentDiscoveryServiceServicer_to_server(
            OptaServicer(self), self.server)
        server_addr = "{}:{}".format(self.url, self.port)
        logger.info("Start OPTA server {}".format(server_addr))
        self.server.add_insecure_port(server_addr)
        self.server.start()
        while True:
            time.sleep(self.wait_time)

    def stop_service(self):
        self.server.stop(0)


class NetQOPTA():

    def start_opta_server(self, conf_dict):
        self.opta_service = OptaService(conf_dict)
        self.opta_thread = Thread(target=self.opta_service.start_service)
        self.opta_thread.start()
        self.wait_time = 60 * 60 * 24
        while True:
            time.sleep(self.wait_time)

    def stop_opta_server(self):
        self.opta_service.stop_service()

    def get_common_config(self):
        conf_dict = configuration.get_config(_PROC_NAME)
        return conf_dict

    def process_opta_req(self):

        opta_client = NetQAgentTestOptaClient(conf_dict, logger)

        rsp = opta_client.send_opta_req()

        for res in rsp:
            logger.info("Response res {}".format(res))


def run():

    import pdb
    pdb.set_trace()
    opta_serv = NetQOPTA()
    conf_dict = opta_serv.get_common_config()
    opta_serv.start_opta_server(conf_dict)


if __name__ == '__main__':
    run()
