#
# Copyright (c) 2018-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

from unittest import TestCase
from mock import patch, Mock
import os
import time
import eventlet
import gevent
import traceback
from gevent import monkey
from gevent.queue import Full
monkey.patch_all

AF_BRIDGE = 6

from netq_lib.common import configuration
from netq_lib.common.enums import DefaultPaths, ConfKw, AGENTSTATUS
from netq_agent.cmd.agent import _ClDbPush
from netq_agent.cmd.netq_agent_discovery import NetQAgentDiscovery, ChannelState
from netq_agent.cmd.netq_agent_stream import NetQAgentStream, NetQAgentServicer
from netq_agent.cmd.netq_pq import NetQPriorityQueue, QueueElem


from concurrent import futures
from threading import Thread

# GRPC Specfic Imports
# GRPC Client imports
import grpc
import netq_agent.interface.netq_pb2
from netq_agent.interface.netq_pb2_grpc import NetqAgentStreamServiceStub

# GRPC Server IMPORTS
from netq_agent.interface.netq_reg_pb2 import NetqAgentDiscoveryResp, ResponseStatus
from netq_agent.interface.netq_reg_pb2_grpc import add_NetqAgentDiscoveryServiceServicer_to_server, NetqAgentDiscoveryServiceServicer
from netq_agent.interface.netq_pb2 import NetqOPTARequest


# Test Utilities
from tests.netq_agent.netq_test_thread_utils import TestThread
from tests.netq_agent.lldp_msg import LLDP_STR


# Logging
import logging
logger = logging.getLogger('netq-agent')
consoleHandler = logging.StreamHandler()
logFormatter = logging.Formatter(
    "%(asctime)s [%(filename)s:%(lineno)d] [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
consoleHandler.setFormatter(logFormatter)
logger.addHandler(consoleHandler)


_PROC_NAME = 'netq-agent'

DefaultPaths.DBCONFIG_RUNNING = './tests/netq_agent/config/netq_agent_running.json'


class TestDefaultPaths(object):

    TEST_PATH = './tests/netq_agent/config/'
    CONF_FILE = '{}netq.yml'.format(TEST_PATH)
    CONF_DIR = '{}config.d/'.format(TEST_PATH)
    DBCONFIG = '%snetq-agent-commands.yml' % CONF_DIR
    DBCONFIG_RUNNING = '/var/run/netq-agent-running.json'
    INTERFACE_FILE = '/etc/network/interfaces'
    UDS_FILE = '/var/run/netq-agent.sock'


def get_configuration_test_paths():
    conf_files = []
    if os.path.exists(TestDefaultPaths.CONF_FILE):
        conf_files.append(TestDefaultPaths.CONF_FILE)

    for root, dirs, files in os.walk(TestDefaultPaths.CONF_DIR):

        for filename in sorted(files):
            # only look for files with .yml or .yaml extensions
            if filename.endswith('.yml') or filename.endswith('.yaml'):
                conf_files.append(os.path.join(root, filename))
    return conf_files


class NetQAgentTestOptaClient(object):

    def __init__(self, conf_dict, logger):
        self._conf = conf_dict
        self.agent_addr = self.get_agent_addr()
        self._logger = logger
        try:
            self._logger.info(
                "Sending OPTA Request to agent {}".format(self.agent_addr))
            self._channel = grpc.insecure_channel(self.agent_addr)
        except Exception as e:
            self._logger.error(
                "Unable to establish connection with OPTA {}".format(e))
            return AGENTSTATUS.FAILURE

    def get_agent_addr(self):
        serv_addr = "{}:{}".format(self._conf['netq-agent'].get(ConfKw.AGENT_URL, ConfKw.DEFAULT_AGENT_URL),
                                   self._conf['netq-agent'].get(ConfKw.AGENT_PORT, ConfKw.DEFAULT_AGENT_PORT))
        return serv_addr

    def send_start_stream(self):
        pass

    def send_stop_stream(self):
        pass

    def send_opta_req(self):
        req_msg = NetqOPTARequest()

        netq_agent_service_stub = NetqAgentStreamServiceStub(self._channel)

        try:
            rsp_message = netq_agent_service_stub.StreamMessage(req_msg)
            self._logger.info("OPTA Message Request sent")
        except grpc.RpcError as e:
            self._logger.error("GRPC Error Received {}".format(e))
            return AGENTSTATUS.FAILURE
        except Exception as e:
            self._logger.error("Unknown GRPC Exception {}".format(e))
            return AGENTSTATUS.FAILURE

        logger.info("Rsp {}".format(rsp_message))

        return rsp_message


class OptaServicer(NetqAgentDiscoveryServiceServicer):

    def __init__(self, service, status):
        self.status = status
        self.service = service
        self.running_count = 0
        self.status = status

    def build_discover_response(self, request, context):

        if self.running_count > len(self.status)-1:
            logger.error("FAiled to build message")
            self.service.stop_service()
            return
        logger.info("Building Discovery response status {} running_count {}".
                    format(self.status, self.running_count))
        status = self.status[self.running_count]
        if status == "":
            self.service.stop_service()
            self.service.start_service()
            return
        self.running_count = self.running_count + 1
        rsp_message = NetqAgentDiscoveryResp(status=status,
                                             reconnect_opta_addr="127.0.0.1")

        logger.info("OPTA Response Dispatched")
        return rsp_message

    # This is a Inteface Message
    def DiscoverMessage(self, request, context):
        logger.info("Discover Message Received ")
        rsp_msg = self.build_discover_response(request, context)
        return rsp_msg


class OptaService(object):

    def __init__(self, conf_dict, status):
        self.url = conf_dict['netq-agent'][ConfKw.SERVER]
        self.port = conf_dict['netq-agent'].get(ConfKw.PORT, ConfKw.DEFAULT_OPTA_PORT)
        self.wait_time = 60 * 60 * 24
        self.server = None
        self.status = status

    def start_service(self):

        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

        add_NetqAgentDiscoveryServiceServicer_to_server(
            OptaServicer(self, self.status), self.server)
        server_addr = "{}:{}".format(self.url, self.port)
        logger.info("Start OPTA server {}".format(server_addr))
        self.server.add_insecure_port(server_addr)
        self.server.start()
        while True:
            time.sleep(self.wait_time)

    def stop_service(self):
        self.server.stop(0)


class NetQAgentTest(TestCase):

    def setup(self, conf_dict):
        self._pool = eventlet.GreenPool(200)
        opta_service = OptaService(conf_dict, _)
        from threading import Thread
        opta_thread = Thread(target=opta_service.start_service)
        opta_thread.start()
        #self.opta_thread = TestThread.spawn(int(time.time()), self._pool, opta_service.start_service)

    def start_opta_server(self, conf_dict, status):
        self.opta_service = OptaService(conf_dict, status)
        self.opta_thread = Thread(target=self.opta_service.start_service)
        self.opta_thread.start()

    def stop_opta_server(self):
        self.opta_service.stop_service()

    @patch('netq_lib.common.configuration.get_paths', Mock(return_value=get_configuration_test_paths()))
    def get_common_config(self):
        conf_dict = configuration.get_config(_PROC_NAME)
        return conf_dict

    @patch('netq_lib.common.configuration.get_paths', Mock(return_value=get_configuration_test_paths()))
    def test_configuration(self):
        conf_dict = configuration.get_config(_PROC_NAME)

        if conf_dict != None:
            self.setup(conf_dict)

        cldbpush_inst = _ClDbPush(conf_dict)
        return cldbpush_inst.run()

    def test_netq_agent_discovery_success(self):
        conf_dict = self.get_common_config()

        agent_discovery = NetQAgentDiscovery(conf_dict, logger=logger)

        self.start_opta_server(conf_dict, ['Success'])

        agent_discovery.start_discovery()

        self.stop_opta_server()

    def clean_up(self):
        time.sleep(2)
        pid = os.getpid()

        os.kill(pid, 9)

    def test_netq_agent_discovery_failure(self):
        conf_dict = self.get_common_config()

        agent_discovery = NetQAgentDiscovery(conf_dict, logger=logger)

        self.start_opta_server(conf_dict, ['Failure'])

        agent_discovery.start_discovery()

        self.stop_opta_server()

    def test_agent_without_opta(self):
        conf_dict = self.get_common_config()

        agent_discovery = NetQAgentDiscovery(conf_dict, logger=logger)

        agent_discovery.start_discovery()

    def test_agent_discovery_retry(self):
        conf_dict = self.get_common_config()

        agent_discovery = NetQAgentDiscovery(conf_dict, logger=logger)

        self.start_opta_server(conf_dict, ['Failure', '', 'Success'])

        agent_discovery.start_discovery()

        self.stop_opta_server()
        # self.clean_up()


class NetQAgentStreamTest(TestCase):

    def setup(self, conf_dict):
        self._pool = eventlet.GreenPool(200)
        opta_service = OptaService(conf_dict, _)
        from threading import Thread
        opta_thread = Thread(target=opta_service.start_service)
        opta_thread.start()
        #self.opta_thread = TestThread.spawn(int(time.time()), self._pool, opta_service.start_service)

    def start_opta_server(self, conf_dict, status):
        self.opta_service = OptaService(conf_dict, status)
        self.opta_thread = Thread(target=self.opta_service.start_service)
        self.opta_thread.start()

    def stop_opta_server(self):
        self.opta_service.stop_service()

    @patch('netq_lib.common.configuration.get_paths', Mock(return_value=get_configuration_test_paths()))
    def get_common_config(self):
        conf_dict = configuration.get_config(_PROC_NAME)
        return conf_dict

    @patch('netq_lib.common.configuration.get_paths', Mock(return_value=get_configuration_test_paths()))
    def test_configuration(self):
        conf_dict = configuration.get_config(_PROC_NAME)

        if conf_dict != None:
            self.setup(conf_dict)

        cldbpush_inst = _ClDbPush(conf_dict)
        return cldbpush_inst.run()

    @patch('netq_agent.cmd.netq_agent_stream.NetQAgentServicer.get_lldp_command_out', Mock(return_value=LLDP_STR))
    def test_agent_stream(self):
        try:
            conf_dict = self.get_common_config()

            logger.info("Starting OPTA Server")
            self.start_opta_server(conf_dict, ['Success'])

            agent_stream = NetQAgentStream(conf_dict, logger)

            # if agent_stream.get_discovery_status() != ChannelState.CONNECTED:
            #    logger.error("Agent Discovery Failed. Test Aborting...")
            #    return

            from threading import Thread
            stream_thread = Thread(target=agent_stream.stream_main)
            stream_thread.start()
            time.sleep(2)
            opta_client = NetQAgentTestOptaClient(conf_dict, logger)

            rsp = opta_client.send_opta_req()

            for res in rsp:
                logger.info("Response res {}".format(res))

            self.stop_opta_server()

        except Exception as e:
            logger.error("{}".format(e))
            traceback.print_exc()
        finally:
            self.clean_up()

    @patch('netq_agent.cmd.netq_agent_stream.NetQAgentServicer.get_lldp_command_out', Mock(return_value=LLDP_STR))
    def test_agent_stream_serv(self):
        try:
            conf_dict = self.get_common_config()

            agent_stream = NetQAgentStream(conf_dict, logger)

            # if agent_stream.get_discovery_status() != ChannelState.CONNECTED:
            #    logger.error("Agent Discovery Failed. Test Aborting...")
            #    return

            from threading import Thread
            stream_thread = Thread(target=agent_stream.stream_main)
            stream_thread.start()
            time.sleep(2)
            #opta_client = NetQAgentTestOptaClient(conf_dict, logger)

            #rsp = opta_client.send_opta_req()

            while True:
                time.sleep(100)
            self.stop_opta_server()

        except Exception as e:
            logger.error("{}".format(e))
            traceback.print_exc()
        finally:
            self.clean_up()

    @patch('netq_agent.cmd.netq_agent_stream.NetQAgentServicer.get_lldp_command_out', Mock(return_value=LLDP_STR))
    def test_agent_stream_comm(self):
        try:
            conf_dict = self.get_common_config()

            agent_stream = NetQAgentStream(conf_dict, logger)
            self.start_opta_server(conf_dict, ['Success'])

            # if agent_stream.get_discovery_status() != ChannelState.CONNECTED:
            #    logger.error("Agent Discovery Failed. Test Aborting...")
            #    return

            #from threading import Thread
            #stream_thread = Thread(target = agent_stream.stream_main)
            # stream_thread.start()
            # time.sleep(2)
            #opta_client = NetQAgentTestOptaClient(conf_dict, logger)
            agent_stream.stream_main()

            #rsp = opta_client.send_opta_req()

            while True:
                time.sleep(100)
            self.stop_opta_server()

        except Exception as e:
            logger.error("{}".format(e))
            traceback.print_exc()
        finally:
            self.clean_up()

    def clean_up(self):
        time.sleep(2)
        pid = os.getpid()

        os.kill(pid, 9)


class NetQAgentQueueTest(TestCase):

    def producer_thread(self, pq):

        while True:
            try:
                gevent.sleep(0.01)
                elems = [QueueElem(i, 4) for i in range(1, 2)]
                elems = elems + [QueueElem(i, 2) for i in range(2, 4)]
                elems = elems + [QueueElem(i, 3) for i in range(4, 5)]

                try:
                    #s = [pq.put((e.priority, e)) for e in elems]
                    s = [pq.putq(e) for e in elems]
                except Full:
                    logger.error("Queue Full")
                except Exception as e:
                    logger.error(e)
                    logger.info("Size of Queue {}".format(pq.len()))
                logger.info("Put complete successfully")
            except Exception as e:
                logger.error("Exception {}".format(e))

    def consumer_thread(self, pq):

        logger.info("Consumer thread started")
        while True:
            try:
                elem = pq.get_buffer()

                if elem is None:
                    continue

                for priority, e in elem:
                    logger.info("Elements fetched {} priority {}".format(
                        e.data, e.priority))
            except Exception as e:
                logger.error("Exception {}".format(e))

    def test_queue_getput(self):

        pq = NetQPriorityQueue(5, 5, logger)

        elems = [QueueElem(i, 1) for i in range(1, 5)]

        s = [pq.put(e) for e in elems]

        elem = pq.get_buffer()

        for e in elem:
            logger.info("Element {}".format(e))

    def test_priority(self):

        pq = NetQPriorityQueue(5, 5, logger)

        elems = [QueueElem(i, 4) for i in range(1, 2)]
        elems = elems + [QueueElem(i, 2) for i in range(2, 4)]
        elems = elems + [QueueElem(i, 3) for i in range(4, 5)]

        s = [pq.put((e.priority, e)) for e in elems]

        elem = pq.get_buffer()

        for priority, e in elem:
            logger.info("Elements fetched {} priority {}".format(
                e.data, e.priority))

    def test_queue_thread(self):
        pq = NetQPriorityQueue(5, 5, logger)

        from threading import Thread

        p_thread = Thread(target=self.producer_thread, args=(pq,))
        c_thread = Thread(target=self.consumer_thread, args=(pq,))
        c_thread.start()
        p_thread.start()

        p_thread.join()
        c_thread.join()

    def test_queue_multi(self):
        pq = NetQPriorityQueue(5, 5, logger)

        p1_thread = Thread(target=self.producer_thread, args=(pq,))
        p2_thread = Thread(target=self.producer_thread, args=(pq,))
        c_thread = Thread(target=self.consumer_thread, args=(pq,))
        c_thread.start()
        p1_thread.start()
        p2_thread.start()

        p1_thread.join()
        p2_thread.join()
        c_thread.join()
