#
# Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

import json
import os
import sys
import time
import random
import datetime
import argparse
import shutil

class QueueHistogramDataGenerator(object):
    def __init__(self, n_ports=4, interval=1, n_records=1000, start_time=None, sample_population_max=1000000, ifprefix="swp"):
        self.n_ports = random.randint(1, 128) if n_ports == 0 else n_ports
        self.interval = interval
        self.n_records = n_records
        self.start_time = self.get_timestamp(start_time) if start_time != None else start_time
        self.sample_population_max = sample_population_max
        self.ifprefix = ifprefix

    def get_timestamp(self, start_time=None):
        stime = 0
        if type(start_time) == str:
            time_format = '%Y-%m-%d %H:%M:%S.%f'
            stime = int(datetime.datetime.strptime(start_time.split('+')[0], time_format).timestamp())
        elif type(start_time) == int:
            stime = start_time
        else:
            stime = int(datetime.datetime.now().timestamp())

        print("Setting timestamp = {} - {}".format(stime, datetime.datetime.fromtimestamp(stime)))

        return stime

    def generate_bin_values(self, n_bins, sample_max=1000000, bin_values=[]):
        for bin in range(0, n_bins):
            bin_values.append(random.randint(1, sample_max - sum(bin_values) - n_bins + bin))
        return bin_values.copy()

    def generate_queue_histogram(self, time_step):
        q_hist_data = {}
        timestamp_str = datetime.datetime.fromtimestamp(time_step).strftime('%Y-%m-%d %H:%M:%S.%f')
        q_hist_data["timestamp_info"] = {"start_datetime": timestamp_str, "end_datetime": timestamp_str}
        q_hist_data["buffer_info"] = "null"
        q_hist_data["packet_info"] = "null"
        q_hist_data["histogram_info"] = {}
        for swp in range(1, self.n_ports + 1):
            swp_port = self.ifprefix + (str((swp - 1)) if self.ifprefix.endswith('s') else str(swp))
            q_histogram_range = ["0:863", "864:2303", "2304:3743", "3744:5183", "5184:6623", "6624:8063", "8064:9503", "9504:10943", "10944:12383", "12384:*"]
            q_histogram_data = {"0:863": 0, "864:2303": 0, "2304:3743": 0, "3744:5183": 0, "5184:6623": 0, "6624:8063": 0, "8064:9503": 0, "9504:10943": 0, "10944:12383": 0, "12384:*": 0}
            q_histogram_bins = random.randint(1, 10)
            q_histogram_bin_start_offset = random.randint(0, q_histogram_bins)
            #print("Offset: {}, Bins: {}".format(q_histogram_bin_start_offset, q_histogram_bins))
            if (q_histogram_bin_start_offset + q_histogram_bins) >= 10:
                q_histogram_bin_start_offset -= (q_histogram_bin_start_offset + q_histogram_bins) % 10

            q_histogram_bin_values = self.generate_bin_values(q_histogram_bins, self.sample_population_max, [])
            bin_idx = 0
            for off in range(q_histogram_bin_start_offset, q_histogram_bins):
                key = q_histogram_range[off]
                q_histogram_data[key] = q_histogram_bin_values[bin_idx]
                bin_idx += 1

            q_hist_data["histogram_info"][swp_port] = q_histogram_data

        return q_hist_data

    def generate_queue_histograms(self, file_count=200, file_path='/home/cumulus/histograms/'):
        count = 0
        fileidx = 0
        basefilename = file_path + "histogram_stats_"
        time_step = self.start_time if self.start_time != None else self.get_timestamp()
        while (count < self.n_records):
            json_data = json.dumps(self.generate_queue_histogram(time_step))
            filename = basefilename + str(fileidx)
            file = open(filename, "w")
            file.write(json_data)
            file.close()
            print("Timestep - {}".format(datetime.datetime.fromtimestamp(time_step).strftime('%Y-%m-%d %H:%M:%S.%f')))
            print("Wrote file {} @ {}".format(filename, datetime.datetime.now()))
            time.sleep(self.interval)
            count += 1
            fileidx = (fileidx + 1) % file_count
            if self.start_time != None:
                time_step += self.interval
            else:
                time_step = self.get_timestamp()

def do_setup_configuration(config_file):
    # Create histogram snapshot path
    snapshot_dir = "/home/cumulus/histograms"
    if not os.path.exists(snapshot_dir):
        os.mkdir(snapshot_dir)

    # Move the existing config file
    bck_config_file = config_file + ".bck"
    shutil.move(config_file, bck_config_file)
    histogram_config = """
    monitor.port_group_list = [histogram_pg]
    monitor.histogram_pg.port_set            = swp1-swp32
    monitor.histogram_pg.stat_type           = histogram
    monitor.histogram_pg.cos_list            = [0]
    monitor.histogram_pg.trigger_type        = timer
    monitor.histogram_pg.timer               = 1s
    monitor.histogram_pg.action_list         = [snapshot]
    monitor.histogram_pg.snapshot.file       = /home/cumulus/histograms/histogram_stats
    monitor.histogram_pg.snapshot.file_count = 200
    monitor.histogram_pg.histogram.minimum_bytes_boundary =   960
    monitor.histogram_pg.histogram.histogram_size_bytes   = 12288
    monitor.histogram_pg.histogram.sample_time_ns         =  1024
    """

    conf_file = open(config_file, "w")
    conf_file.write(histogram_config)
    conf_file.close()

def do_cleanup_configuration(config_file):
    snapshot_dir = "/home/cumulus/histograms"
    config_file = "/etc/cumulus/datapath/monitor.conf"
    bck_config_file = config_file + ".bck"
    shutil.move(bck_config_file, config_file)
    shutil.rmtree(snapshot_dir, ignore_errors=False)

def setup(args):
    if args.setupconf:
        do_setup_configuration("/etc/cumulus/datapath/monitor.conf")

def cleanup(args):
    if args.setupconf:
        do_cleanup_configuration("/etc/cumulus/datapath/monitor.conf")

if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--randseed", help="random seed")
    parser.add_argument("--nports", help="number of ports")
    parser.add_argument("--interval", help="data generation interval")
    parser.add_argument("--nrecords", help="total number of records to generate")
    parser.add_argument("--starttime", help="timestamp of the data")
    parser.add_argument("--filecount", help="file counts")
    parser.add_argument("--filepath", help="filepath where the data would be generated")
    parser.add_argument("--samplesize", help="Sample population size")
    parser.add_argument("--ifprefix", help="Interface prefix")
    parser.add_argument("--setupconf", help="Automatically setup configuration", action="store_true")

    args = parser.parse_args()

    rand_seed = int(datetime.datetime.now().timestamp()) if args.randseed == None else int(args.randseed)

    n_ports = 32 if args.nports == None else int(args.nports)
    interval = 1 if args.interval == None else int(args.interval)
    n_records = 2592000 if args.nrecords == None else int(args.nrecords) # 30 days default
    start_time = args.starttime if args.starttime != None else None
    sample_size = 1000000 if args.samplesize == None else int(args.samplesize)

    file_count = 200 if args.filecount == None else int(args.filecount)
    file_path = "/home/cumulus/histograms/" if args.filepath == None else args.filepath
    ifprefix = "swp" if args.ifprefix == None else args.ifprefix
    setupconf = True if args.setupconf else False

    setup(args)

    random.seed(rand_seed)
    q_hist_data_generator = QueueHistogramDataGenerator(n_ports, interval, n_records, start_time, sample_size, ifprefix)

    print("Configuration:")
    print(" Random seed         - {}".format(rand_seed))
    print(" Start time          - {}".format(start_time))
    print(" Number of ports     - {}".format(n_ports))
    print(" Interval            - {}".format(interval))
    print(" Number of records   - {}".format(n_records))
    print(" Sample size         - {}".format(sample_size))
    print(" File count          - {}".format(file_count))
    print(" File path           - {}".format(file_path))
    print(" Interface prefix    - {}".format(ifprefix))
    print(" Setup configuration - {}".format(setupconf))

    print(" Histogram generation started for every {} sec(s)...".format(interval))
    q_hist_data_generator.generate_queue_histograms(file_count=file_count, file_path=file_path)
    print("Histogram generation ended...")

    cleanup(args)