#
# Copyright (c) 2017-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


# CONFIGURATION SPECIFIC TO NETQ APPLICATION

from netq_lib.common import utils
from netq_lib.common.configuration import update_yaml_configuration, get_config
from netq_lib.common.enums import DefaultPaths, ConfKw

__REDIS_P = 6379
__SENTL_P = 26379


def get_path(path):
    return path if path else DefaultPaths.CONF_FILE


def update_config(conf_dict, path=None):
    path = path or DefaultPaths.CONF_FILE
    preamble = '## Netq configuration File.\n'
    path = get_path(path)
    rc = update_yaml_configuration(conf_dict, path, preamble)
    if rc:
        print((utils.RED_T('Failed to update backend server configuration %s' % path)))
    return rc

def get_config_section(section, path=None):
    conf_dict = get_config(paths=path)
    return conf_dict.get(section) if conf_dict else None

def _merge_config(conf_dict, section, keyname, value):
    if section not in conf_dict:
        conf_dict[section] = {}

    if keyname:
        conf_dict[section][keyname] = value
    else:
        conf_dict[section] = value

    return conf_dict

def update_config_section(section, value, path):
    # Don't merge in the whole config
    path = path or DefaultPaths.CONF_FILE
    return update_config(_merge_config(get_config(paths=path), section, None, value), path=path)

def conf_update_config(section, keyname, value):
    # We're writing the configs for the netq-agent here
    # TODO: update_config returns 1 which is coincidentally RC_FAIL on failure
    path = DefaultPaths.CONF_FILE
    return update_config(_merge_config(get_config(paths=path), section, keyname, value))


def conf_get_config(section, keyname):
    conf_dict = get_config()

    value = None
    if section in conf_dict:
        value = conf_dict[section].get(keyname, None)

    return value

##################
#
# REDIS BACKEND
#
#################

def add_backend_server(servers, server_addr, port=__REDIS_P, vrf=None,
                       role=None):
    servers = servers if servers is not None else []
    servers = delete_backend_server(servers, server_addr, role)
    servers.append({ConfKw.SERVER: server_addr,
                    ConfKw.PORT: port,
                    ConfKw.VRF: vrf,
                    ConfKw.ROLE: role})
    return servers


def delete_backend_server(servers, server_addr, role=None):
    servers = servers if servers is not None else []
    for i, server in enumerate(servers):
        if server_addr == server.get(ConfKw.SERVER):
            if role and role == server.get(ConfKw.ROLE):
                del servers[i]  # delete all if multiple matches in list
    return servers


def _load_backend_servers(path=None):
    cfg = get_config(paths=path)
    backend = cfg.get(ConfKw.BACKEND) if cfg else None
    if backend:
        if isinstance(backend, list):
            return backend

        if isinstance(backend, dict):
            return [backend]
    return []


def load_backend_servers(path=None):
    servers = _load_backend_servers(path)
    if servers:
        for i, server in enumerate(servers):
            if server.get(ConfKw.ROLE) is None:
                if len(servers) > 1:
                    servers[i][ConfKw.ROLE] = ConfKw.SENTINEL
                else:
                    servers[i][ConfKw.ROLE] = ConfKw.MASTER
    return servers


def save_backend_servers(servers, path=None):
    '''servers is list of dicts with host,port,vrf'''
    # TODO: Hack-attack
    if servers and len(servers) == 1:
        servers = servers[0]
    return update_config_section(ConfKw.BACKEND, servers, path=path)


def get_config_backend():
    '''gets backend config using default config paths'''
    conf = get_config()
    if ConfKw.BACKEND in conf:
        return conf[ConfKw.BACKEND]
    else:
        raise RuntimeError('Please specify backend server')


# Ensure that there is a default config
def ensure_default_backend():
    path = '/etc/netq/netq.yml'
    servers = load_backend_servers(path)
    if servers is None:
        save_backend_servers(add_backend_server([], '127.0.0.1',
                                                role=ConfKw.MASTER))
