#
# Copyright (c) 2018-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#


from multiprocessing import Queue
import time
from netq_lib.common.agent_stats_monitor import AgentStatsMon
import sys

# ###########Experimental Code ###################


global global_cl_inst


def setClAgentInst(clInst):
    global global_cl_inst
    global_cl_inst = clInst


def getClAgentInst():
    global global_cl_inst
    return global_cl_inst

###############################################


def create_message_queues(*args, **kwargs):
    logger = kwargs.get('logger')
    conf = kwargs.get('conf')
    pq = NetQPriorityQueue.create_priority_queue(conf, logger)
    bp_pq = NetQPriorityQueue.create_bp_queue(conf, logger)
    return pq, bp_pq


def create_gnmi_message_queues(*args, **kwargs):
    logger = kwargs.get('logger')
    conf = kwargs.get('conf')
    gnmi_pq = NetQPriorityQueue.create_gnmi_priority_queue(conf, logger)
    return gnmi_pq

def create_agent_info_queue(*args, **kwargs):

    logger = kwargs.get('logger')
    conf = kwargs.get('conf')
    pq = NetQPriorityQueue.create_agent_info_queue(conf, logger)
    return pq


class QueueElem(object):

    def __init__(self, priority, message_type, fullstate, data):
        self.priority = priority
        self.message_type = message_type
        self.fullstate = fullstate
        self.data = data


class NetQPriorityQueueWrapper(object):

    """NetQ Wrapper Priority Queue implementation"""

    def __init__(self, maxsize=1000, conf=None, max_num_queue=None, logger=None):
        self.maxsize = maxsize
        self._logger = logger
        self._conf = conf
        self.pq = []
        # Netlink Messages only
        self.message_types = ['Link', 'Address', 'Netconf', 'Route', 'Neighbor', 'MacFdb', 'VxlanRemoteDest', 'Nexthop']
        self._max_num_queue = max_num_queue if max_num_queue is not None else len(self.message_types)
        self.message_type_to_qid = {v: k + self._max_num_queue for k, v in enumerate(self.message_types)}
        self.default_pq_id = 1
        for priority in range(0, self._max_num_queue + len(self.message_type_to_qid)):
            priority_queue = Queue(maxsize)
            priority_queue.cancel_join_thread()
            self.pq.append(priority_queue)

    def get_pqid_by_message(self, message_type):
        if message_type == "Routes":
            message_type = "Route"
        try:
            pq_id = self.message_type_to_qid[message_type]
            return pq_id
        except Exception:
            self._logger.error("Unable to find the message type mapping {}".format(message_type))
            return self.default_pq_id

    @property
    def max_num_queue(self):
        return self._max_num_queue

    def is_full(self):
        for queue_id in range(0, self._max_num_queue):
            if self.pq[queue_id].qsize() >= self.maxsize:
                return True
        return False

    def is_empty(self):
        for queue_id in range(0, self._max_num_queue):
            if self.pq[queue_id].qsize() != 0:
                return False
        return True

    def get_reader_fds(self):
        fd_set = []
        for queue_id in range(0, self._max_num_queue):
            try:
                if self.pq[queue_id] is not None:
                    fd_set.append(self.pq[queue_id]._reader)
                else:
                    self._logger.debug("Queue Destroyed {}".format(queue_id))
            except Exception as e:
                self._logger.error("Exception Caught Idx {} {}".format(queue_id, e))
        return fd_set

    def len(self, priority_idx):
        return self.pq[priority_idx].qsize()

    def print_all(self):
        for priority_idx in range(0, self._max_num_queue + len(self.message_type_to_qid)):
            self._logger.info(" qid {} len {}".format(priority_idx, self.pq[priority_idx].qsize()))

    def len_all(self):
        sum = 0
        for priority_idx in range(0, self._max_num_queue + len(self.message_type_to_qid)):
            sum += self.pq[priority_idx].qsize()
        return sum

    def putq(self, elem):
        try:
            self.update_agent_stats(elem)
            self.pq[elem.priority].put((elem.priority, elem.message_type, elem.fullstate, elem))
        except Exception:
            agent_stats_object = AgentStatsMon.get_instance()
            agent_stats_object.pq_full_count += 1
            self._logger.debug("Queue full length:{}".format(self.len(elem.priority)))

    def get_buffer(self, priority_idx, counter=0):
        q_size = self.pq[priority_idx].qsize()
        retry = 3

        self._logger.debug("Counter mismtach message pqidx {} counter {} qsize {}".format(priority_idx, counter, q_size))
        while(retry):
            if (q_size != counter and counter != 0):
                self._logger.debug("Counter mismtach message pqidx {} counter {} qsize {}".format(priority_idx, counter, q_size))
                time.sleep(.1)
                retry = retry - 1
                q_size = self.pq[priority_idx].qsize()
            else:
                break

        num_elements = q_size if(q_size <= counter or counter == 0) else counter

        while (num_elements):
            try:
                msg = self.pq[priority_idx].get(True)
                yield msg
                num_elements = num_elements - 1
            except Exception as e:
                print (e)
                self._logger.debug("Queue is empty {}".format(e))
                time.sleep(0.1)

    def update_agent_stats(self, elem):

        agent_stats_object = AgentStatsMon.get_instance()

        msg_count_stats = agent_stats_object.message_count
        msg_count_stats['total_msg_count'] = msg_count_stats.get('total_msg_count', 0) + 1
        msg_count_stats['total_msg_bytes'] = msg_count_stats.get('total_msg_bytes', 0) + sys.getsizeof(elem.data)

        if elem.fullstate:
            agent_stats_object.msgtype_fullstate_count[elem.message_type] = agent_stats_object.msgtype_fullstate_count\
                                                                                .get(elem.message_type, 0) + 1
            msg_count_stats['full_state_count'] = msg_count_stats.get('full_state_count', 0) + 1
        else:
            msg_count_stats['partial_state_count'] = msg_count_stats.get('partial_state_count', 0) + 1

        if elem.message_type in self.message_types:
            msg_count_stats['netlink_msg_count'] = msg_count_stats.get('netlink_msg_count', 0) + 1
        else:
            msg_count_stats['non_netlink_msg_count'] = msg_count_stats.get('non_netlink_msg_count', 0) + 1

        agent_stats_object.message_count = msg_count_stats
        agent_stats_object.msg_type_count[elem.message_type] = \
            agent_stats_object.msg_type_count.get(elem.message_type, 0) + 1

    def total_pq_usage(self, pq_util_stats):

        current_queue_usage = 0
        for current_queue in self.pq:
            current_queue_usage += current_queue.qsize()

        if current_queue_usage >= pq_util_stats.get('max_pq_util', 0):
            pq_util_stats['max_pq_util'] = current_queue_usage

        if current_queue_usage <= pq_util_stats.get('min_pq_util', 0):
            pq_util_stats['min_pq_util'] = current_queue_usage

        pq_util_stats['count'] = pq_util_stats.get('count', 0) + 1
        pq_util_stats['sum_pq_util'] = pq_util_stats.get("sum_pq_util", 0) + current_queue_usage

        return pq_util_stats

    def empty_queues(self):
        for idx in range(0, self._max_num_queue):
            while not self.pq[idx].empty():
                self.pq[idx].get()

    def empty_pq(self, priority_idx):
        q = self.pq[priority_idx]
        while not q.empty():
            try:
                q.get(False)
            except Exception:
                continue


class NetQPriorityQueue(object):

    PQ = None
    GNMI_PQ = None
    # this is used to report any issues from grpc server to agent thread
    bp_PQ = None
    agent_PQ = None

    @classmethod
    def create_priority_queue(cls, conf, logger):
        try:
            cls.PQ = NetQPriorityQueueWrapper(maxsize=500000, conf=conf, logger=logger)
        except Exception as e:
            logger.error("Exception {}".format(e))
            cls.PQ = None
        return cls.PQ

    @classmethod
    def create_gnmi_priority_queue(cls, conf, logger):
        try:
            cls.GNMI_PQ = NetQPriorityQueueWrapper(maxsize=1000, conf=conf, logger=logger)
        except Exception as e:
            logger.error("Exception {}".format(e))
            cls.GNMI_PQ = None
        return cls.GNMI_PQ

    @classmethod
    def create_bp_queue(cls, conf, logger):
        try:
            cls.bp_PQ = NetQPriorityQueueWrapper(maxsize=10, conf=conf, logger=logger, max_num_queue=1)
        except Exception as e:
            logger.error("Exception {}".format(e))
            cls.bp_PQ = None
        return cls.bp_PQ

    @classmethod
    def create_agent_info_queue(cls, conf, logger):
        try:
            cls.agent_PQ = NetQPriorityQueueWrapper(maxsize=1024, conf=conf, logger=logger, max_num_queue=1)
        except Exception as e:
            logger.error("Exception {}".format(e))
            cls.agent_PQ = None
        return cls.agent_PQ
