#
# Copyright (c) 2016-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

"""ROM models
"""
import inspect
import time
from datetime import datetime

import hashlib
from netq_lib.common.query import GatewayQuery

from netq_lib.common.configuration import (
    get_config,
)
from netq_lib.common.enums import ConfKw, ServiceStatus
from netq_lib.common.netq_pq import NetQPriorityQueue, QueueElem
from netq_lib.common.nlfullstate import NetlinkFullStateObject
from netq_lib.orm import objects, HOSTNAME
from netq_lib.orm.memdb.models import MemModel
from netq_lib.common.utils import is_gnmi_enabled

# If you add any classes to this file that is not an object pushed to the
# backend DB, such as  _RedisModel and Connection, please add those classes
# to the ignored list in agent.py's reinitialize_db() function.

_KEY_TTL = 90  # TTL liveness is 90s


class _RedisModel(objects.NetQModel):
    __memmodel = None
    post_headers = dict()
    get_headers = dict()
    put_headers = dict()
    delete_headers = dict()
    agent_conf = get_config().get(ConfKw.NETQ_AGENT, dict())
    redis_enabled = agent_conf.get(ConfKw.IS_REDIS_ENABLED, ConfKw.DEFAULT_IS_REDIS_ENABLED)
    opta_enabled = agent_conf.get(ConfKw.IS_OPTA_ENABLED, ConfKw.DEFAULT_IS_OPTA_ENABLED)
    gnmi_enabled = is_gnmi_enabled(get_config())
    container_hostname = ''
    container_port = 8080
    query_cls = GatewayQuery

    _NETLINK_MSGS = ['Address', 'Route', 'Link', 'Netconf', 'Neighbor', 'MacFdb', 'VxlanRemoteDest']
    time_map = {'start': 'gt_timestamp', 'end': 'lt_timestamp', 'timestamp': 'eq_timestamp',
                'endtimestamp': 'lt_timestamp'}
    optional_args_map = dict()
    api_key_map = dict()
    api_filter_args = []
    app_name = ''
    delete_request_args = []
    post_request_args = []
    put_request_args = []
    base_url = ''

    active = objects.Boolean(json_key='active')
    hostname = objects.String(json_key='hostname')
    timestamp = objects.Float(json_key='timestamp')
    deleted = objects.Boolean(json_key='deleted')
    message_type = objects.String(json_key='messageType')

    def __init__(self, gateway=False):
        self.active = True
        self.deleted = False
        self.hostname = HOSTNAME
        self.timestamp = time.time()
        super(_RedisModel, self).__init__(gateway)
        # Fill the right message type after initialization
        self.message_type = self.__class__.__name__

    def __lt__(self, other):
        return self.hostname < other.hostname

    def __gt__(self, other):
        return self.hostname > other.hostname

    @classmethod
    def exclude_fmt(cls):
        """Return the list of fields to ignore when doing a comparison to
        find the differences between two instances of an object.
        See BgpSession for an example usage"""
        return cls.timestamp, cls.active

    @classmethod
    def memmodel(cls):
        if cls.__memmodel is None:
            cls.__memmodel = type(cls.__name__, (MemModel, cls), {})
        return cls.__memmodel

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @classmethod
    def url_fmt(cls, key=None):
        raise NotImplementedError

    @staticmethod
    def _get_queue_object():
        return NetQPriorityQueue.PQ

    @staticmethod
    def _get_gnmi_queue_object():
        return NetQPriorityQueue.GNMI_PQ

    def get_priority(self):
        return 1

    def copy(self, src):
        for _, field in inspect.getmembers(self.__class__):
            if (isinstance(field, objects.Field) and field.name != self.__class__.timestamp.name):
                src_attribute = getattr(src, field.name, None)
                if src_attribute is not None:
                    setattr(self, field.name, src_attribute)

    def delete(self, *args, **kwargs):
        # TODO: cleanup this function
        '''Mark the object as deleted OR purge the entry from the backend'''
        obj = self.__class__()
        obj.copy(self)
        obj.active = False
        obj.deleted = True
        obj.save()

    def get_objid(self):
        '''A unique for this object type/hostname ID'''
        obj = self.obj_to_key(self.__dict__).encode('utf8')
        return hashlib.md5(obj).hexdigest()

    def __is_netlink_full_state_ongoing(self):

        if self.message_type not in self._NETLINK_MSGS:
            return False

        if (self.message_type != NetlinkFullStateObject.msg_type):
            return False

        if (NetlinkFullStateObject.get_fullstate() is not True):
            return False
        else:
            return True

    def save(self, *args, **kwargs):
        # TODO: cleanup this function
        return

    def save_grpc(self):

        pq = self._get_queue_object()
        if pq is None:
            return False

        priority = self.get_priority()
        pq_data = self.to_object_json()
        if pq_data is None:
            return False

        if self.__is_netlink_full_state_ongoing():
            # NetlinkFullStateObject.increment_counter(self.message_type)
            # priority = pq.get_pqid_by_message(self.message_type)
            # Do not push the objects to the Queue
            # Wait till the Netlink initialization is complete
            return True

        try:
            q_elem = QueueElem(priority, self.message_type, False, pq_data)
            pq.putq(q_elem)
        except Exception:
            # Raise Exception
            return False
        return True

    def save_gnmi_grpc(self):
        pq = self._get_gnmi_queue_object()
        if pq is None:
            return False

        priority = self.get_priority()
        pq_data = self.to_object_json()
        if pq_data is None:
            return False

        try:
            q_elem = QueueElem(priority, self.message_type, False, pq_data)
            pq.putq(q_elem)
        except Exception:
            # Raise Exception
            return False
        return True

    def save_memmodel(self, json_fields=[]):

        cls = self.__class__
        mcls = cls.memmodel()
        key = cls.obj_to_key(self.__dict__)
        val = cls.obj_to_val(self.__dict__)
        score = cls.obj_to_score(self.__dict__)

        add = True
        delete = False
        old_obj = None

        if mcls is not None:
            old_objs = (
                list(mcls.query.range(start=0, end=10000, first=False,
                                      keys=[key]))
            )
            if old_objs:
                old_obj = old_objs[-1]

                if self.equals(old_obj, json_fields):
                    if old_obj.active == self.active:
                        add = False
                    elif (self.active and len(old_objs) >= 2) or self.deleted:
                        delete = True

                for obj in old_objs[:-2]:
                    obj.delete_object()

        if delete:
            del_score = cls.obj_to_score(old_obj.__dict__)
            mcls.delete(key, del_score)
        if add:
            mcls.save(key, score, val)

        return add

    def send(self):
        if self.opta_enabled:
            self.save_grpc()
        if self.should_save_gnmi():
            self.save_gnmi_grpc()

    def save_entry(self):
        add = self.save_memmodel()
        if add:
            self.send()

    def counter_fmt(self):
        return None

    def should_save_gnmi(self):
        return False

    def counter_retention_policy(self):
        return None

    def counter_continuous_query(self):
        return None

    def diff(self, other):
        '''
        Returns dict { k:(self,other) } differences between self and other
        instance same class. Only finds deltas between common keys
        '''

        if other is None:
            raise ValueError

        if not type(self) is type(other):
            raise NotImplementedError

        # WARNING: Using PREFIX from objects.
        exclude_fields = {'i_' + field.name for field in self.exclude_fmt()}
        return {k: (v, other.__dict__.get(k))
                for k, v in list(self.__dict__.items())
                if k not in exclude_fields and other.__dict__.get(k) != v}


class Netconf(_RedisModel):
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    forwarding = objects.Boolean(json_key='forwarding')
    app_name = 'netconf'
    container_hostname = 'netq-app-link'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_ipv6']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.is_ipv6)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.ifindex, cls.forwarding, cls.timestamp, cls.active, cls.deleted

    def save(self):
        super(Netconf, self).save_entry()

class Address(_RedisModel):
    ifname = objects.String(json_key='interface')
    prefix = objects.String(json_key='prefix')
    mask = objects.Integer(json_key='mask')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')
    app_name = 'address'
    api_key_map = {'prefix': 'prefix'}
    optional_args_map = {'diff_type': 'diff_type', 'return_only_diff': 'return_only_diff', 'group_by': 'group_by'}
    container_hostname = 'netq-app-address'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'diff':
                return (cls.optional_args_map.get(key, key), 'snapshot')
            elif val == 'history':
                return (cls.optional_args_map.get(key, key), 'timeline')
            else:
                return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_ipv6', 'diff_type', 'return_only_diff', 'group_by']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'prefix', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname, cls.prefix,
                cls.mask, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.message_type, cls.timestamp, cls.active, cls.deleted

    def save(self):
        super(Address, self).save_entry()


# TODO: class Inventory (notice the missing _) will replace all of these.
class _Inventory(_RedisModel):
    app_name = 'inventory'

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'type']


class ASIC(_Inventory):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    model_id = objects.String(json_key='modelId')
    core_bw = objects.String(json_key='coreBW')
    ports = objects.String(json_key='ports')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.model_id, cls.core_bw, cls.ports)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self):
        super(ASIC, self).save_entry()


class BgpSession(_RedisModel):
    """Store BGP Session specific info
    """
    peer_name = objects.String(json_key='peerName')
    state = objects.String(json_key='peerState')
    peer_router_id = objects.String(json_key='bgpRouterId')
    peer_asn = objects.Long(json_key='peerAsn')
    peer_hostname = objects.String(json_key='peerHostname')
    asn = objects.Long(json_key='bgpAsn')
    reason = objects.String(json_key='reason')
    ipv4_pfx_rcvd = objects.Integer(json_key='ipv4PrefixRx')
    ipv6_pfx_rcvd = objects.Integer(json_key='ipv6PrefixRx')
    evpn_pfx_rcvd = objects.Integer(json_key='evpnPrefixRx')
    ipv4_packet_queue_length = objects.Integer(json_key='ipv4PacketQueueLength')
    ipv6_packet_queue_length = objects.Integer(json_key='ipv6PacketQueueLength')
    evpn_packet_queue_length = objects.Integer(json_key='l2EvpnPacketQueueLength')
    last_reset_time = objects.Float(json_key='lastResetTime')
    up_time = objects.Float(json_key='upTime')
    conn_estd = objects.Integer(json_key='connEstd')
    conn_dropped = objects.Integer(json_key='connDropped')
    upd8_rx = objects.Integer(json_key='upd8Rx')
    upd8_tx = objects.Integer(json_key='upd8Tx')
    vrfid = objects.Integer(json_key='vrfId')
    vrf = objects.String(json_key='vrf')
    tx_families = objects.Json(json_key='txFamilies')
    rx_families = objects.Json(json_key='rxFamilies')
    objid = objects.String(json_key='objId')
    hold_time = objects.Integer(json_key='holdTime')
    keep_alive_interval = objects.Integer(json_key='keepAliveInterval')
    graceful_restart = objects.String(json_key='gracefulRestart')
    graceful_restart_remote_time = objects.Integer(json_key='gracefulRestartRemoteTime')
    address_families_by_peer = objects.String(json_key='addressFamiliesByPeer')
    local_gr_mode = objects.String(json_key="localGrMode")
    remote_gr_mode = objects.String(json_key='remoteGrMode')
    graceful_restart_rbit = objects.Boolean(json_key='gracefulRestartRBit')
    configured_restart_time = objects.Integer(json_key='configuredRestartTime')
    received_restart_time = objects.Integer(json_key='receivedRestartTime')
    ipv4_stale_path_time = objects.Integer(json_key='ipv4StalePathTime')
    ipv6_stale_path_time = objects.Integer(json_key='ipv6StalePathTime')
    api_key_map = {'peer_name': 'peername'}
    app_name = 'bgp'
    container_hostname = 'netq-app-bgp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'timers':
                return (cls.optional_args_map.get(key, key), 'timer')
            else:
                return (cls.optional_args_map.get(key, key), val)
        else:
            return ['view']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'peer_name', 'asn', 'vrf', 'check']

    @classmethod
    def exclude_fmt(cls):
        return (cls.timestamp, cls.active, cls.last_reset_time, cls.up_time,
                cls.upd8_tx, cls.upd8_rx)

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.peer_name, cls.asn, cls.vrf

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.state, cls.peer_router_id, cls.peer_asn, cls.peer_hostname,
                cls.reason, cls.ipv4_pfx_rcvd, cls.ipv6_pfx_rcvd,
                cls.evpn_pfx_rcvd, cls.timestamp, cls.last_reset_time,
                cls.conn_estd, cls.conn_dropped, cls.upd8_rx, cls.vrfid,
                cls.upd8_tx, cls.up_time, cls.tx_families, cls.objid,
                cls.rx_families, cls.active, cls.deleted, cls.ipv4_packet_queue_length, cls.ipv6_packet_queue_length,
                cls.evpn_packet_queue_length, cls.hold_time, cls.keep_alive_interval, cls.graceful_restart,
                cls.graceful_restart_remote_time, cls.address_families_by_peer, cls.local_gr_mode, cls.remote_gr_mode,
                cls.graceful_restart_rbit, cls.configured_restart_time, cls.received_restart_time, cls.ipv4_stale_path_time,
                cls.ipv6_stale_path_time)

    def save(self):
        super(BgpSession, self).save_entry()

class StreamingCheck(_RedisModel):
    app_name = 'streamcheck'
    container_hostname = 'netq-app-check'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['checktype', 'tests']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['check_filter_id', 'hostnames']

class StreamingCheckMlag(StreamingCheck):
    app_name = 'streamcheckmlag'

class StreamingCheckBgp(StreamingCheck):
    app_name = 'streamcheckbgp'

class StreamingCheckInterface(StreamingCheck):
    app_name = 'streamcheckinterfaces'

class StreamingCheckMtu(StreamingCheck):
    app_name = 'streamcheckmtu'

class StreamingCheckVlan(StreamingCheck):
    app_name = 'streamcheckvlan'

class StreamingCheckAgent(StreamingCheck):
    app_name = 'streamcheckagents'
    container_hostname = 'netq-app-node'

class StreamingCheckNtp(StreamingCheck):
    app_name = 'streamcheckntp'
    container_hostname = 'netq-app-ntp'

class StreamingCheckSensors(StreamingCheck):
    app_name = 'streamchecksensors'
    container_hostname = 'netq-app-sensors'

class StreamingCheckEvpn(StreamingCheck):
    app_name = 'streamcheckevpn'
    container_hostname = 'netq-app-check-nvo'

class StreamingCheckVxlan(StreamingCheck):
    app_name = 'streamcheckvxlan'
    container_hostname = 'netq-app-check-nvo'

class StreamingCheckRoCE(StreamingCheck):
    app_name = 'streamcheckroce'
    container_hostname = 'netq-app-procdevstats'

class StreamingCheckArConfig(StreamingCheck):
    app_name = 'streamcheckarconfig'
    container_hostname = 'netq-app-procdevstats'

class Board(_Inventory):
    vendor = objects.String(json_key='vendor')
    model = objects.String(json_key='model')
    base_mac = objects.String(json_key='baseMac')
    part_number = objects.String(json_key='partNo')
    mfg_date = objects.String(json_key='mfgDate')
    serial_number = objects.String(json_key='serialNo')
    label_revision = objects.String(json_key='rev')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vendor, cls.model,
                cls.base_mac, cls.part_number, cls.mfg_date, cls.serial_number,
                cls.label_revision)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp,)

    def save(self):
        super(Board, self).save_entry()


class ClagSession(_RedisModel):
    role = objects.String(json_key='role')
    peer_role = objects.String(json_key='peerRole')
    peer_state = objects.Boolean(json_key='peerState')
    peer_if = objects.String(json_key='clagPeerIf')
    peer_ip = objects.String(json_key='peerIp')
    clag_sysmac = objects.String(json_key='sysMac')
    backup_ip_active = objects.Boolean(json_key='backupIpActive')
    backup_ip = objects.String(json_key='backupIp')
    single_bonds = objects.Json(json_key='singleBonds')
    dual_bonds = objects.Json(json_key='dualAttachedBonds')
    conflicted_bonds = objects.Json(json_key='conflictedBonds')
    proto_down_bonds = objects.Json(json_key='protoDownBonds')
    vxlan_anycast = objects.String(json_key='vxlanAnycast')
    api_key_map = {'clag_sysmac': 'sysmac'}
    app_name = 'clag'
    container_hostname = 'netq-app-clagsession'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'clag_sysmac']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.clag_sysmac

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.peer_role, cls.role, cls.peer_state, cls.peer_if,
                cls.backup_ip, cls.backup_ip_active, cls.single_bonds,
                cls.dual_bonds, cls.timestamp, cls.conflicted_bonds,
                cls.vxlan_anycast, cls.proto_down_bonds, cls.active,
                cls.deleted, cls.peer_ip)

    def save(self):
        super(ClagSession, self).save_entry()


class Command(_RedisModel):
    key = objects.String(json_key='key')
    output = objects.String(json_key='output')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.key

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.output,

    def save(self):
        self.save_grpc()


class CPU(_Inventory):
    arch = objects.String(json_key='arch')
    nos = objects.String(json_key='cores')
    model = objects.String(json_key='model')
    max_freq = objects.String(json_key='freq')
    mem_total = objects.String(json_key='memory')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.arch, cls.nos, cls.model,
                cls.max_freq, cls.mem_total)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self):
        super(CPU, self).save_entry()


class Description(_RedisModel):
    objtype = objects.String(json_key='objType')
    descrid = objects.String(json_key='descrId')
    description = objects.String(json_key='description')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.objtype, cls.descrid

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.description, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(Description, self).save_entry()


class Disk(_Inventory):
    name = objects.String(json_key='name')
    size = objects.String(json_key='size')
    d_type = objects.String(json_key='type')
    vendor = objects.String(json_key='vendor')
    transport = objects.String(json_key='transport')
    rev = objects.String(json_key='rev')
    model = objects.String(json_key='model')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.d_type,
                cls.vendor, cls.transport, cls.rev, cls.model)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self):
        super(Disk, self).save_entry()


class DockerContainer(_RedisModel):
    container_id = objects.String(json_key='containerId')
    gw = objects.String(json_key='gateway')
    image = objects.String(json_key='image')
    ip = objects.String(json_key='containerIp')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='containerName')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    prefix_len = objects.Integer(json_key='prefixLen')
    port_list = objects.String(json_key='portList')
    service_id = objects.String(json_key='serviceId')
    service_name = objects.String(json_key='serviceName')
    status = objects.String(json_key='status')
    start_time = objects.String(json_key='startTime')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.image,
                cls.network_name, cls.ip, cls.mac,
                cls.service_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.status,
                cls.network_id, cls.gw, cls.prefix_len,
                cls.port_list, cls.service_id,
                cls.start_time, cls.active, cls.deleted)

    def save(self):
        super(DockerContainer, self).save_entry()


class DockerPortMap(_RedisModel):
    container_port = objects.String(json_key='containerPort')
    container_id = objects.String(json_key='containerId')
    container_ip = objects.String(json_key='containerIp')
    host_port = objects.String(json_key='hostPort')
    host_ip = objects.String(json_key='hostIp')
    image = objects.String(json_key='image')
    mac = objects.String(json_key='mac')
    name = objects.String(json_key='containerName')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    node_id = objects.String(json_key='nodeId')
    proto = objects.String(json_key='proto')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.name, cls.container_ip, cls.proto, cls.container_port,
                cls.host_ip, cls.host_port, cls.network_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.container_id, cls.network_id,
                cls.image, cls.mac, cls.node_id, cls.active, cls.deleted)

    def save(self):
        super(DockerPortMap, self).save_entry()


class DockerSwarmCluster(_RedisModel):
    cluster_version = objects.String(json_key='clusterVersion')
    cluster_name = objects.String(json_key='clusterName')
    cluster_id = objects.String(json_key='clusterId')
    docker_version = objects.String(json_key='version')
    nodes = objects.String(json_key='workerNodes')
    num_nodes = objects.Integer(json_key='numNodes')
    num_managers = objects.Integer(json_key='numManagers')
    managers = objects.String(json_key='managerNodes')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.cluster_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.docker_version, cls.cluster_version, cls.cluster_id,
                cls.num_nodes, cls.num_managers, cls.managers,
                cls.timestamp, cls.nodes, cls.active, cls.deleted)

    def save(self):
        super(DockerSwarmCluster, self).save_entry()


class DockerSwarmNode(_RedisModel):
    cluster_id = objects.String(json_key='clusterId')
    cluster_name = objects.String(json_key='clusterName')
    docker_version = objects.String(json_key='dockerVersion')
    node_id = objects.String(json_key='nodeId')
    node_name = objects.String(json_key='nodeName')
    node_state = objects.String(json_key='nodeState')
    plugins = objects.String(json_key='plugins')
    role = objects.String(json_key='role')
    availability = objects.String(json_key='availability')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.cluster_name, cls.node_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.docker_version, cls.cluster_id, cls.node_id,
                cls.node_state, cls.role, cls.plugins,
                cls.availability, cls.active, cls.deleted)

    def save(self):
        super(DockerSwarmNode, self).save_entry()


class DockerNetwork(_RedisModel):
    container_list = objects.String(json_key='containerList')
    driver = objects.String(json_key='driver')
    host_binding = objects.String(json_key='hostBinding')
    ipv6_enabled = objects.Boolean(json_key='ipv6Enabled')
    ip_masquerade = objects.Boolean(json_key='ipMasq')
    ipam_driver = objects.String(json_key='ipamDriver')
    encrypted = objects.Boolean(json_key='encrypted')
    default_bridge = objects.Boolean(json_key='defaultBridge')
    network_name = objects.String(json_key='networkName')
    network_id = objects.String(json_key='networkId')
    subnet = objects.String(json_key='subnet')
    gateway = objects.String(json_key='gateway')
    parent_interface = objects.String(json_key='parentInterface')
    vxlan_id = objects.Integer(json_key='vxlanId')
    mtu = objects.Integer(json_key='mtu')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.network_name, cls.driver)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.gateway, cls.parent_interface, cls.vxlan_id, cls.network_id,
                cls.mtu, cls.host_binding, cls.ipv6_enabled, cls.ip_masquerade,
                cls.encrypted, cls.default_bridge, cls.ipam_driver, cls.subnet,
                cls.container_list, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(DockerNetwork, self).save_entry()


class DockerService(_RedisModel):
    image = objects.String(json_key='image')
    mode = objects.String(json_key='mode')
    parallelism = objects.Integer(json_key='parallelism')
    port_list = objects.String(json_key='portMapping')
    network_list = objects.String(json_key='networkList')
    replicas = objects.Integer(json_key='replicas')
    service_name = objects.String(json_key='serviceName')
    service_id = objects.String(json_key='serviceId')
    version = objects.Integer(json_key='version')
    vip = objects.String(json_key='virtualIp')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.service_name, cls.mode)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.image, cls.replicas, cls.parallelism, cls.service_id,
                cls.port_list, cls.network_list, cls.vip, cls.version,
                cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(DockerService, self).save_entry()


class DockerHost(_RedisModel):
    containers = objects.Integer(json_key='installed')
    containers_running = objects.Integer(json_key='running')
    docker_version = objects.String(json_key='version')
    images = objects.Integer(json_key='images')
    ip_forwarding = objects.Boolean(json_key='ipForwarding')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.docker_version)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.images, cls.containers, cls.containers_running,
                cls.ip_forwarding, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(DockerHost, self).save_entry()


class EsxHost(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    pnic = objects.String(json_key='pnic')
    dvs = objects.String(json_key='dvs')
    pg = objects.String(json_key='pg')
    encaptype = objects.Integer(json_key='encaptype')
    encap = objects.Integer(json_key='encap')
    vmname = objects.String(json_key='vmname')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    mtu = objects.Integer(json_key='mtu')
    pnic_admin_state = objects.String(json_key='pnic_admin_state')
    pnic_oper_state = objects.String(json_key='pnic_oper_state')
    peer_hostname = objects.String(json_key='peer_hostname')
    peer_ifname = objects.String(json_key='peer_ifname')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname,
                cls.vcenter, cls.datacenter, cls.cluster, cls.esxhost,
                cls.pnic, cls.dvs, cls.pg, cls.encaptype, cls.encap)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.ip, cls.mac, cls.mtu, cls.pnic_admin_state,
                cls.pnic_oper_state, cls.peer_hostname, cls.peer_ifname,
                cls.active, cls.deleted)

    def save(self):
        super(EsxHost, self).save_entry()


class EsxVmAttr(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    vnic = objects.String(json_key='vnic')
    pg = objects.String(json_key='portgroup')
    ip = objects.String(json_key='ip')
    mac = objects.String(json_key='mac')
    esxhost = objects.String(json_key='esxhost')
    mem = objects.String(json_key='memory')
    disk = objects.String(json_key='disk')
    cpus = objects.String(json_key='numcpus')
    vmpath = objects.String(json_key='vmpath')
    state = objects.String(json_key='powerstate')
    annotation = objects.String(json_key='annotation')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.vmname, cls.ostype, cls.vnic, cls.pg, cls.ip, cls.mac)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.esxhost, cls.mem, cls.disk, cls.cpus, cls.vmpath,
                cls.state, cls.annotation, cls.active, cls.deleted)

    def save(self):
        super(EsxVmAttr, self).save_entry()


class EsxVmBind(_RedisModel):
    vcenter = objects.String(json_key='vcenter')
    datacenter = objects.String(json_key='datacenter')
    cluster = objects.String(json_key='cluster')
    esxhost = objects.String(json_key='esxhost')
    vmname = objects.String(json_key='vmname')
    ostype = objects.String(json_key='ostype')
    state = objects.String(json_key='state')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.vcenter, cls.datacenter,
                cls.cluster, cls.esxhost, cls.vmname, cls.ostype, cls.state)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.active, cls.deleted)

    def save(self):
        super(EsxVmBind, self).save_entry()


class EvpnVniInfo(_RedisModel):
    '''Contains VNI info obtained from bgp evpn'''
    vni = objects.Integer(json_key='vni')
    origin_ip = objects.String(json_key='originIp')
    rd = objects.String(json_key='rd')
    export_rt = objects.Json(json_key='exportRT')
    import_rt = objects.Json(json_key='importRT')
    in_kernel = objects.Boolean(json_key='inKernel')
    adv_all_vni = objects.Boolean(json_key='advAllVni')
    adv_gw_ip = objects.String(json_key='advGatewayMacIp')
    adv_pip = objects.String(json_key='advertisePip')
    adv_svi_macip = objects.String(json_key='advertiseSviMacIp')
    sys_ip = objects.String(json_key='sysIP')
    sys_mac = objects.String(json_key='sysMAC')
    rmac = objects.String(json_key='rmac')
    is_l3 = objects.Boolean(json_key='isL3')
    mapping = objects.String(json_key='mapping')
    flooding = objects.Boolean(json_key='flooding')
    app_name = 'evpn'
    container_hostname = 'netq-app-evpn'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin_ip, cls.rd, cls.export_rt, cls.import_rt,
                cls.in_kernel, cls.timestamp, cls.adv_all_vni, cls.adv_gw_ip,
                cls.active, cls.deleted, cls.is_l3, cls.adv_pip, cls.adv_svi_macip,
                cls.sys_ip, cls.sys_mac, cls.rmac, cls.mapping, cls.flooding)

    def save(self):

        super(EvpnVniInfo, self).save_entry()

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'vni']


class _Sensor(_RedisModel):
    app_name = 'sensor'
    api_key_map = {'s_name': 'name'}
    container_hostname = 'netq-app-sensors'

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'type', 's_name']


class Fan(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='description')
    s_input = objects.Integer(json_key='speed')
    s_max = objects.Integer(json_key='max')
    s_min = objects.Integer(json_key='min')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.s_state, cls.s_prev_state, cls.s_input,
                cls.s_msg, cls.s_max, cls.s_min, cls.active, cls.deleted)

    def save(self):
        super(Fan, self).save_entry()
        # pq = self._get_queue_object()
        # priority = self.get_priority()
        # pq_data = self.to_json()
        # q_elem = QueueElem(priority, pq_data)
        # pq.putq(q_elem)


class Inventory(_RedisModel):
    agent_version = objects.String(json_key='agent_version')
    asic_vendor = objects.String(json_key='asic_vendor')
    asic_model = objects.String(json_key='asic_model')
    asic_model_id = objects.String(json_key='asic_modelId')
    asic_core_bw = objects.String(json_key='asic_coreBW')
    asic_ports = objects.String(json_key='asic_ports')
    asic_data = objects.String(json_key='asic_data')
    platform_vendor = objects.String(json_key='platform_vendor')
    platform_model = objects.String(json_key='platform_model')
    platform_base_mac = objects.String(json_key='platform_baseMac')
    platform_part_number = objects.String(json_key='platform_partNo')
    platform_mfg_date = objects.String(json_key='platform_mfgDate')
    platform_serial_number = objects.String(json_key='platform_serialNo')
    platform_label_revision = objects.String(json_key='platform_rev')
    cpu_arch = objects.String(json_key='cpu_arch')
    cpu_nos = objects.String(json_key='cpu_cores')
    cpu_model = objects.String(json_key='cpu_model')
    cpu_max_freq = objects.String(json_key='cpu_freq')
    cpu_data = objects.String(json_key='cpu_data')
    disk_total_size = objects.String(json_key='disk_size')
    disk_data = objects.String(json_key='disk_data')
    license_state = objects.String(json_key='license_state')
    license_data = objects.String(json_key='license_data')
    memory_total_size = objects.String(json_key='memory_size')
    memory_data = objects.String(json_key='memory_data')
    os_name = objects.String(json_key='os_name')
    nvue_enabled = objects.Boolean(json_key='nvue_enabled')
    nvue_active = objects.Boolean(json_key='nvue_active')
    is_nvue_config_saved = objects.Boolean(json_key='is_nvue_config_saved')
    nclu_enabled = objects.Boolean(json_key='nclu_enabled')
    appliance_version = objects.String(json_key='applianceVersion')
    manifest_hash = objects.String(json_key='manifestHash')
    os_version = objects.String(json_key='os_fullVersion')
    os_version_id = objects.String(json_key='os_version')
    opta_on_switch_enabled =objects.String(json_key='opta_on_switch_enabled')
    inband_interface = objects.String(json_key='inband_interface')

    app_name = 'inventory'
    api_filter_args = ['hostname', 'asic_vendor', 'asic_model', 'asic_model_id', 'platform_vendor', 'platform_model',
                       'cpu_arch', 'disk_name', 'disk_transport', 'disk_vendor', 'license_name', 'license_state',
                       'memory_vendor', 'memory_m_type', 'os_name', 'os_version']
    container_hostname = 'netq-app-inventory'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def val_fmt(cls, key_type=None):
        return (cls.asic_vendor, cls.asic_model, cls.asic_model_id, cls.asic_core_bw, cls.asic_ports, cls.asic_data,
                cls.platform_vendor, cls.platform_model, cls.platform_base_mac, cls.platform_part_number,
                cls.platform_mfg_date, cls.platform_serial_number, cls.platform_label_revision,
                cls.cpu_arch, cls.cpu_nos, cls.cpu_model, cls.cpu_max_freq, cls.cpu_data,
                cls.disk_total_size, cls.disk_data, cls.appliance_version,
                cls.license_state, cls.license_data, cls.manifest_hash,
                cls.memory_total_size, cls.memory_data, cls.nvue_enabled, cls.nvue_active, cls.is_nvue_config_saved, cls.nclu_enabled,
                cls.os_name, cls.os_version, cls.os_version_id, cls.agent_version,
                cls.message_type, cls.timestamp, cls.opta_on_switch_enabled, cls.inband_interface)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    def save(self):

        super(Inventory, self).save_entry()

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'type']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['filter_opta']


class _Kube(_RedisModel):
    @classmethod
    def url_fmt(cls, key=None):
        raise NotImplementedError

    @classmethod
    def key_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def score_fmt(cls, key_type=None):
        raise NotImplementedError

    @classmethod
    def val_fmt(cls):
        raise NotImplementedError

    app_name = 'kubernetes'
    container_hostname = 'netq-app-kube'


class KubeCluster(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    nodes = objects.String(json_key='nodes')
    controller_manager = objects.String(json_key='controller_manager')
    scheduler = objects.String(json_key='scheduler')
    etcd = objects.String(json_key='etcd')
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'api_server', 'cluster_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.api_server, cls.cluster_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.nodes, cls.timestamp,
                cls.controller_manager,
                cls.scheduler, cls.etcd,
                cls.active, cls.deleted)

    def save(self):

        super(KubeCluster, self).save_entry()


class KubeDeployment(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    replicas = objects.Integer(json_key='replicas')
    ready_replicas = objects.Integer(json_key='ready_replicas')
    start_time = objects.String(json_key='start_time')
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.replicas, cls.ready_replicas, cls.start_time,
                cls.active, cls.deleted)

    def save(self):

        super(KubeDeployment, self).save_entry()


class KubeDaemonSet(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    desired_count = objects.Integer(json_key='desired_count')
    ready_count = objects.Integer(json_key='ready_count')
    start_time = objects.String(json_key='start_time')
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.desired_count, cls.ready_count, cls.start_time,
                cls.active, cls.deleted)

    def save(self):

        super(KubeDaemonSet, self).save_entry()


class KubeNode(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    node_name = objects.String(json_key='node_name')
    labels = objects.String(json_key='labels')
    role = objects.String(json_key='role')
    pod_cidr = objects.String(json_key='pod_cidr')
    kube_proxy_version = objects.String(json_key='kube_proxy_version')
    kubelet_version = objects.String(json_key='kubelet_version')
    status = objects.String(json_key='status')
    machine_id = objects.String(json_key='machine_id')
    container_runtime = objects.String(json_key='container_runtime')
    start_time = objects.String(json_key='start_time')
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.node_name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.role, cls.labels, cls.pod_cidr,
                cls.kube_proxy_version, cls.kubelet_version,
                cls.machine_id, cls.status, cls.container_runtime,
                cls.timestamp, cls.start_time,
                cls.active, cls.deleted)

    def save(self):

        super(KubeNode, self).save_entry()


class KubePod(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    template_hash = objects.String(json_key='template_hash')
    node_name = objects.String(json_key='node_name')
    containers = objects.String(json_key='containers')
    ip_address = objects.String(json_key='ip_address')
    owner_kind = objects.String(json_key='owner_kind')
    owner_name = objects.String(json_key='owner_name')
    start_time = objects.String(json_key='start_time')
    status = objects.String(json_key='status')
    api_key_map = {'cluster_name': 'clustername'}
    optional_args_map = {'ip_address': 'podip', 'node_name': 'nodename'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label', 'ip_address', 'node_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name,
                cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels, cls.template_hash,
                cls.node_name, cls.containers, cls.ip_address,
                cls.owner_kind, cls.owner_name, cls.start_time,
                cls.status, cls.active, cls.deleted)

    def save(self):

        super(KubePod, self).save_entry()


class KubeReplicationController(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    replicas = objects.Integer(json_key='replicas')
    ready_replicas = objects.Integer(json_key='ready_replicas')
    start_time = objects.String(json_key='start_time')
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.replicas, cls.ready_replicas, cls.start_time,
                cls.active, cls.deleted)

    def save(self):

        super(KubeReplicationController, self).save_entry()


class KubeReplicaSet(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    pod_template_hash = objects.String(json_key='pod_template_hash')
    replicas = objects.Integer(json_key='replicas')
    ready_replicas = objects.Integer(json_key='ready_replicas')
    start_time = objects.String(json_key='start_time')
    api_key_map = {'cluster_name': 'clustername'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.replicas, cls.ready_replicas, cls.pod_template_hash,
                cls.start_time, cls.active, cls.deleted)

    def save(self):

        super(KubeReplicaSet, self).save_entry()


class KubeService(_Kube):
    api_server = objects.String(json_key='api_server')
    cluster_name = objects.String(json_key='cluster_name')
    namespace = objects.String(json_key='namespace')
    name = objects.String(json_key='name')
    labels = objects.String(json_key='labels')
    service_type = objects.String(json_key='service_type')
    cluster_ip = objects.String(json_key='cluster_ip')
    external_ip = objects.String(json_key='external_ip')
    ports = objects.String(json_key='port')
    status = objects.String(json_key='status')
    start_time = objects.String(json_key='start_time')
    api_key_map = {'cluster_name': 'clustername'}
    optional_args_map = {'cluster_ip': 'clusterip', 'external_ip': 'externalip'}

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname', 'cluster_name', 'namespace', 'name']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['label', 'cluster_ip', 'external_ip']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.api_server,
                cls.cluster_name, cls.namespace, cls.name)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.labels,
                cls.service_type, cls.cluster_ip,
                cls.external_ip, cls.ports, cls.status,
                cls.start_time, cls.active, cls.deleted)

    def save(self):

        super(KubeService, self).save_entry()


class License(_Inventory):
    name = objects.String(json_key='name')
    state = objects.String(json_key='state')
    license = objects.String(json_key='license')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.state, cls.license, cls.timestamp

    def save(self):
        super(License, self).save_entry()


class Link(_RedisModel):
    message_type = objects.String(json_key='messageType')
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    mac_address = objects.MacAddress(json_key='macAddress')
    admin_state = objects.String(json_key='adminState')
    oper_state = objects.String(json_key='operState')
    managed = objects.Boolean(json_key='managed')
    kind = objects.String(json_key='type')
    mtu = objects.Integer(json_key='mtu')
    master = objects.String(json_key='master')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    vlans = objects.String(json_key='vlans')
    access_vlan = objects.Integer(json_key='pvid')
    dstport = objects.Integer(json_key='dstPort')
    localip = objects.String(json_key='localIp')
    vni = objects.Integer(json_key='vni')
    vrf = objects.String(json_key='vrf')
    rt_table_id = objects.Integer(json_key='routeTableId')
    parent_if = objects.String(json_key='parentIf')
    stp_state = objects.Integer(json_key='stpState')
    down_reason = objects.String(json_key='downReason')
    learning_en = objects.Boolean(json_key='learningEnabled')
    arp_suppress_en = objects.Boolean(json_key='arpSuppressionEnabled')
    objid = objects.String(json_key='objId')
    is_svd = objects.Boolean(json_key='isSvd')
    pvid_vni_map = objects.String(json_key='pvidVniMap')
    ipv4_forwarding = objects.Boolean(json_key='ipv4Forwarding')
    ipv6_forwarding = objects.Boolean(json_key='ipv6Forwarding')
    app_name = 'link'
    container_hostname = 'netq-app-link'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'kind', 'master', 'vni', 'state']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        """We need to query sometimes by VNI and so we stick VNI in the key.
        Its 0 for anything but Vxlan.
        """
        return (cls.__name__, cls.hostname, cls.ifname, cls.kind, cls.vni,
                cls.master)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.admin_state, cls.oper_state, cls.managed,
                cls.mtu, cls.ifindex, cls.is_vlan_filtering, cls.timestamp,
                cls.vlans, cls.access_vlan, cls.localip, cls.down_reason,
                cls.vrf, cls.rt_table_id, cls.parent_if, cls.stp_state,
                cls.mac_address, cls.dstport, cls.learning_en, cls.objid,
                cls.is_svd, cls.pvid_vni_map, cls.arp_suppress_en, cls.active,
                cls.deleted, cls.ipv4_forwarding, cls.ipv6_forwarding)

    def save(self):

        super(Link, self).save_entry()


class Liveness(_RedisModel):

    def __str__(self):
        return self.hostname

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.hostname,

    def save(self):
        self.save_grpc()


class ConfigMon(_RedisModel):
    file_name = objects.String(json_key='fileName')
    file_content = objects.String(json_key='fileContent')
    last_modify_time = objects.Float(json_key='lastModifyTime')
    app_name = 'configmon'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'file_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.file_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.file_name, cls.file_content, cls.timestamp, cls.last_modify_time,
                cls.active, cls.deleted)

    def save(self):

        super(ConfigMon, self).save_entry()

class NvueMon(_RedisModel):
    file_name = objects.String(json_key='fileName')
    file_content = objects.String(json_key='fileContent')
    last_modify_time = objects.String(json_key='lastRevModifyTime')
    rev_id = objects.String(json_key='rev_id')
    app_name = 'nvuemon'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'file_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.file_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.file_name, cls.file_content, cls.timestamp, cls.last_modify_time,
                cls.active, cls.deleted, cls.rev_id)

    def save(self):

        super(NvueMon, self).save_entry()


class RunningConfigMon(_RedisModel):
    commandname = objects.String(json_key='commandname')
    result = objects.String(json_key='result')
    last_modify_time = objects.Float(json_key='lastModifyTime')
    app_name = 'runningconfigmon'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'commandname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.commandname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.commandname, cls.result, cls.timestamp, cls.last_modify_time,
                cls.active, cls.deleted)

    def save(self):

        super(RunningConfigMon, self).save_entry()


class ClSupport(_RedisModel):
    file_abs_name = objects.String(json_key='fileAbsName')
    create_time = objects.Float(json_key='createTime')
    app_name = 'clsupport'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'file_abs_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.file_abs_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return cls.message_type, cls.file_abs_name, cls.create_time, cls.timestamp, cls.active, cls.deleted

    def save(self):

        super(ClSupport, self).save_entry()


class AgentStats(_RedisModel):
    agent_util = objects.Json(json_key='agentUtil')
    grpc_util = objects.Json(json_key='grpcUtil')
    message_count = objects.Json(json_key='messageCount')
    msg_type_count = objects.Json(json_key='msgTypeCount')
    msgtype_fullstate_count = objects.Json(json_key='msgtypeFullstateCcount')
    pq_full_count = objects.Integer(json_key='pqFullCount')
    pq_util = objects.Json(json_key='pqUtil')
    agent_pause_count = objects.Integer(json_key='agentPauseCount')
    chunk_count = objects.Json(json_key='chunkCount')
    opta_grps_reset_count = objects.Integer(json_key='optaGrpsResetCount')
    app_name = 'agentstats'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.message_count, cls.msg_type_count, cls.msgtype_fullstate_count, cls.grpc_util,
                cls.agent_util, cls.pq_full_count, cls.pq_util, cls.agent_pause_count, cls.chunk_count,
                cls.opta_grps_reset_count, cls.timestamp, cls.active, cls.deleted)

    def save(self):

        super(AgentStats, self).save_entry()


class ResourceUtil(_RedisModel):
    cpu_utilization = objects.Float(json_key='cpuUtilization')
    cpu_user = objects.Float(json_key='cpuUser')
    cpu_nice = objects.Float(json_key='cpuNice')
    cpu_idle = objects.Float(json_key='cpuIdle')
    cpu_system = objects.Float(json_key='cpuSystem')
    cpu_iowait = objects.Float(json_key='cpuIOWait')
    cpu_irq = objects.Float(json_key='cpuIrq')
    cpu_softirq = objects.Float(json_key='cpuSoftIrq')
    cpu_steal = objects.Float(json_key='cpuSteal')
    cpu_guest = objects.Float(json_key='cpuGuest')
    cpu_guest_nice = objects.Float(json_key='cpuGuestNice')
    interval = objects.Integer(json_key='interval')
    mem_utilization = objects.Float(json_key='memUtilization')
    mem_total = objects.Long(json_key='memTotal')
    mem_used = objects.Long(json_key='memUsed')
    disk_utilization = objects.Json(json_key='diskUtilization')
    root_disk = objects.String(json_key='rootDisk')
    is_disk_read_only = objects.Boolean(json_key='isDiskReadOnly')
    app_name = 'resource'
    container_hostname = 'netq-app-resource-util'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.cpu_utilization, cls.cpu_user,
                cls.cpu_nice, cls.cpu_idle, cls.cpu_system, cls.cpu_iowait,
                cls.cpu_irq, cls.cpu_softirq, cls.cpu_steal, cls.cpu_guest,
                cls.cpu_guest_nice, cls.interval, cls.mem_utilization,
                cls.mem_total, cls.mem_used, cls.disk_utilization,
                cls.root_disk, cls.is_disk_read_only, cls.timestamp,
                cls.active, cls.deleted)

    def save(self):

        super(ResourceUtil, self).save_entry()

    def should_save_gnmi(self):
        return self.gnmi_enabled


class ForwardingResource(_RedisModel):
    host_0_entry = objects.Json(json_key='host0Entry')
    host_1_entry = objects.Json(json_key='host1Entry')
    route_0_entry = objects.Json(json_key='route0Entry')
    route_1_entry = objects.Json(json_key='route1Entry')
    route_total_entry = objects.Json(json_key='routeTotalEntry')
    ecmp_nh_entry = objects.Json(json_key='ecmpNhEntry')
    mac_entry = objects.Json(json_key='macEntry')
    mroute_total_entry = objects.Json(json_key='mrouteTotalEntry')
    app_name = 'cl-resource-forwarding'
    container_hostname = 'netq-app-resource-util'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.host_0_entry, cls.host_1_entry, cls.route_0_entry,
                cls.route_1_entry, cls.route_total_entry, cls.ecmp_nh_entry, cls.mac_entry,
                cls.mroute_total_entry, cls.active, cls.timestamp, cls.deleted)

    def save(self):

        super(ForwardingResource, self).save_entry()


class ACLResource(_RedisModel):
    in_acl_entry = objects.Json(json_key='inAclEntries')
    eg_acl_entry = objects.Json(json_key='outAclEntries')
    in_acl_counter = objects.Json(json_key='inAclCounters')
    eg_acl_counter = objects.Json(json_key='outAclCounters')
    in_acl_meter = objects.Json(json_key='inAclMeters')
    eg_acl_meter = objects.Json(json_key='outAclMeters')
    in_acl_slice = objects.Json(json_key='inAclSlices')
    eg_acl_slice = objects.Json(json_key='outAclSlices')
    in_acl_v4mac_filter = objects.Json(json_key='inAclIpv4Filter')
    in_acl_v4mac_mangle = objects.Json(json_key='inAclIpv4Mangle')
    in_acl_v6_filter = objects.Json(json_key='inAclIpv6Filter')
    in_acl_v6_mangle = objects.Json(json_key='inAclIpv6Mangle')
    eg_acl_v4mac_filter = objects.Json(json_key='outAclIpv4Filter')
    eg_acl_v4mac_mangle = objects.Json(json_key='outAclIpv4Mangle')
    eg_acl_v6_filter = objects.Json(json_key='outAclIpv6Filter')
    eg_acl_v6_mangle = objects.Json(json_key='outAclIpv6Mangle')
    in_acl_mirror_filter = objects.Json(json_key='inAclMirror')
    in_acl_8021x_filter = objects.Json(json_key='inAcl8021xFilter')
    in_pbr_v4mac_filter = objects.Json(json_key='inPbrIpv4Filter')
    in_pbr_v6_filter = objects.Json(json_key='inPbrIpv6Filter')
    acl_l4_port_range_checkers = objects.Json(json_key='aclL4PortRangeCheckers')
    acl_mlx_regions = objects.Json(json_key='aclRegions')
    acl_mlx_18b_rules_key = objects.Json(json_key='acl18BRulesKey')
    acl_mlx_32b_rules_key = objects.Json(json_key='acl32BRulesKey')
    acl_mlx_54b_rules_key = objects.Json(json_key='acl54BRulesKey')
    app_name = 'cl-resource-acl'
    container_hostname = 'netq-app-resource-util'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.active, cls.in_acl_entry, cls.eg_acl_entry,
                cls.in_acl_counter, cls.eg_acl_counter, cls.in_acl_meter,
                cls.eg_acl_meter, cls.in_acl_slice, cls.eg_acl_slice, cls.in_acl_v4mac_filter,
                cls.in_acl_v4mac_mangle, cls.in_acl_v6_filter, cls.in_acl_v6_mangle,
                cls.eg_acl_v4mac_filter, cls.eg_acl_v4mac_mangle, cls.eg_acl_v6_filter,
                cls.eg_acl_v6_mangle, cls.in_acl_mirror_filter, cls.in_acl_8021x_filter,
                cls.in_pbr_v4mac_filter, cls.in_pbr_v6_filter, cls.acl_l4_port_range_checkers,
                cls.acl_mlx_regions, cls.acl_mlx_18b_rules_key, cls.acl_mlx_32b_rules_key,
                cls.acl_mlx_54b_rules_key, cls.timestamp, cls.deleted)

    def save(self):

        super(ACLResource, self).save_entry()


class Lldp(_RedisModel):
    ifname = objects.String(json_key='interface')
    peer_hostname = objects.String(json_key='peerHostname')
    peer_ifname = objects.String(json_key='peerInterface')
    lldp_peer_bridge = objects.Boolean(json_key='lldpPeerBridge')
    lldp_peer_router = objects.Boolean(json_key='lldpPeerRouter')
    lldp_peer_station = objects.Boolean(json_key='lldpPeerStation')
    lldp_peer_os = objects.String(json_key='lldpPeerOS')
    lldp_peer_osv = objects.String(json_key='lldpPeerOsVersion')
    api_key_map = {'peer_hostname': 'peerhostname'}
    app_name = 'lldp'
    container_hostname = 'netq-app-lldp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['filter_deleted']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'peer_hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.peer_ifname, cls.lldp_peer_bridge, cls.lldp_peer_router,
                cls.lldp_peer_station, cls.lldp_peer_os, cls.lldp_peer_osv,
                cls.timestamp, cls.active, cls.deleted)

    def save(self):

        super(Lldp, self).save_entry()


class Ptm(_RedisModel):
    cbl_status = objects.String(json_key='cblStatus')
    port = objects.String(json_key='port')
    exp_nbr = objects.String(json_key='expNbr')
    act_nbr = objects.String(json_key='actNbr')

    app_name = 'ptm'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'port']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.port

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.cbl_status, cls.port, cls.exp_nbr, cls.act_nbr,
                cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(Ptm, self).save_entry()


class MacFdb(_RedisModel):
    mac_address = objects.MacAddress(json_key='macAddress')
    origin = objects.Boolean(json_key='origin')
    nexthop = objects.String(json_key='egressPort')
    vlan = objects.Integer(json_key='vlan')
    dst = objects.String(json_key='dst')
    port = objects.String(json_key='port')
    is_remote = objects.Boolean(json_key='isRemote')
    is_static = objects.Boolean(json_key='isStatic')
    api_key_map = {'mac_address': 'macaddress'}
    app_name = 'macs'
    optional_args_map = {'diff_type': 'diff_type', 'return_only_diff': 'return_only_diff', 'group_by': 'group_by'}
    container_hostname = 'netq-app-macs'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'diff':
                return (cls.optional_args_map.get(key, key), 'snapshot')
            elif val == 'history':
                return (cls.optional_args_map.get(key, key), 'timeline')
            else:
                return (cls.optional_args_map.get(key, key), val)
        else:
            return ['origin', 'diff_type', 'return_only_diff', 'group_by']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'mac_address', 'vlan', 'commentary']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.mac_address, cls.vlan

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.origin, cls.nexthop, cls.dst, cls.port, cls.timestamp,
                cls.is_remote, cls.is_static, cls.active, cls.deleted)

    def save(self):
        super(MacFdb, self).save_entry()


class Memory(_Inventory):
    vendor = objects.String(json_key='vendor')
    size = objects.String(json_key='size')
    speed = objects.String(json_key='speed')
    name = objects.String(json_key='name')
    m_type = objects.String(json_key='type')
    serial_number = objects.String(json_key='serialNo')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.size, cls.speed,
                cls.m_type, cls.vendor, cls.serial_number)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp,)

    def save(self):
        super(Memory, self).save_entry()


class MstpInfo(_RedisModel):
    """Store MSTP non-port-specific info for troubleshooting
    """
    bridge_name = objects.String(json_key='bridgeName')
    state = objects.Boolean(json_key='state')
    root_port_name = objects.String(json_key='rootPortName')
    root_bridge = objects.String(json_key='rootBridge')
    topo_chg_ports = objects.Json(json_key='topoChangePort')
    time_since_tcn = objects.Long(json_key='timeSinceLastTcn')
    topo_chg_cntr = objects.Long(json_key='toptChangeCntr')
    bridge_id = objects.String(json_key='bridgeId')
    edge_ports = objects.Json(json_key='edgePorts')
    network_ports = objects.Json(json_key='networkPorts')
    disputed_ports = objects.Json(json_key='disputedPorts')
    bpduguard_ports = objects.Json(json_key='bpduguardPorts')
    bpduguard_err_ports = objects.Json(json_key='bpduguardErrPorts')
    ba_inconsistent_ports = objects.Json(json_key='baInconsistentPorts')
    bpdufilter_ports = objects.Json(json_key='bpdufilterPorts')
    is_vlan_filtering = objects.Boolean(json_key='isVlanFiltering')
    ports = objects.Json(json_key='ports')
    app_name = 'mstp'
    api_key_map = {'bridge_name': 'bridgename'}
    container_hostname = 'netq-app-mstpinfo'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'bridge_name']

    @classmethod
    def exclude_fmt(cls):
        return cls.timestamp, cls.active, cls.time_since_tcn

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.bridge_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.root_port_name, cls.topo_chg_ports, cls.time_since_tcn,
                cls.topo_chg_cntr, cls.ports, cls.edge_ports, cls.state,
                cls.network_ports, cls.disputed_ports, cls.bpduguard_ports,
                cls.root_bridge, cls.bridge_id, cls.bpduguard_err_ports,
                cls.ba_inconsistent_ports, cls.bpdufilter_ports,
                cls.is_vlan_filtering, cls.active, cls.deleted,
                cls.timestamp)

    def save(self):

        super(MstpInfo, self).save_entry()


class OS(_Inventory):
    name = objects.String(json_key='name')
    version = objects.String(json_key='fullVersion')
    version_id = objects.String(json_key='version')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.timestamp, cls.name, cls.version,
                cls.version_id)

    def save(self):
        super(OS, self).save_entry()


class OspfNbr(_RedisModel):
    peer_id = objects.String(json_key='peerId')
    peer_addr = objects.String(json_key='peerAddress')
    state = objects.String(json_key='state')
    ifname = objects.String(json_key='interface')
    is_ipv6 = objects.String(json_key='isIPv6')
    area = objects.String(json_key='area')
    vrf = objects.String(json_key='vrf')
    app_name = 'ospfnbr'
    container_hostname = 'netq-app-ospf'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.peer_id, cls.area, cls.vrf

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'peer_id', 'area', 'vrf']

    @classmethod
    def val_fmt(cls):
        return (cls.state, cls.timestamp, cls.ifname, cls.is_ipv6,
                cls.peer_addr, cls.area, cls.vrf, cls.active, cls.deleted)

    def save(self):

        super(OspfNbr, self).save_entry()


class OspfIf(_RedisModel):
    ifname = objects.String(json_key='interface')
    is_unnumbered = objects.Boolean(json_key='isUnnumbered')
    if_up = objects.Boolean(json_key='ifUp')
    area = objects.String(json_key='area')
    network_type = objects.String(json_key='networkType')
    nbr_count = objects.Integer(json_key='nbrCount')
    nbr_adj_count = objects.Integer(json_key='nbrAdjCount')
    cost = objects.Integer(json_key='cost')
    is_passive = objects.Boolean(json_key='isPassive')
    mtu = objects.Integer(json_key='mtu')
    dead_time = objects.Integer(json_key='deadTime')
    rexmit_time = objects.Integer(json_key='rexmitTime')
    hello_time = objects.Integer(json_key='helloTime')
    router_id = objects.String(json_key='routerId')
    app_name = 'ospfif'
    container_hostname = 'netq-app-ospf'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname, cls.area

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'area']

    @classmethod
    def val_fmt(cls):
        return (cls.network_type, cls.timestamp, cls.nbr_count, cls.if_up,
                cls.nbr_adj_count, cls.is_unnumbered, cls.is_passive, cls.cost,
                cls.mtu, cls.dead_time, cls.rexmit_time, cls.hello_time,
                cls.router_id, cls.area, cls.active, cls.deleted)

    def save(self):

        super(OspfIf, self).save_entry()


class Neighbor(_RedisModel):
    message_type = objects.String(json_key='messageType')
    ifname = objects.String(json_key='interface')
    ifindex = objects.Integer(json_key='ifIndex')
    ip_address = objects.String(json_key='ipAddress')
    mac_address = objects.MacAddress(json_key='macAddress')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    vrf = objects.String(json_key='vrf')
    is_remote = objects.Boolean(json_key='isRemote')
    api_filter_args = ['ifname', 'mac_address']
    api_key_map = {'ip_address': 'ipaddress', 'mac_address': 'mac'}
    optional_args_map = {'diff_type': 'diff_type', 'return_only_diff': 'return_only_diff',
                         'group_by': 'group_by', 'ip_address': 'ipaddress', 'mac_address': 'mac'}
    app_name = 'neighbor'
    container_hostname = 'netq-app-neighbor'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'diff':
                return (cls.optional_args_map.get(key, key), 'snapshot')
            elif val == 'history':
                return (cls.optional_args_map.get(key, key), 'timeline')
            else:
                return (cls.optional_args_map.get(key, key), val)
        else:
            return ['ifname', 'ip_address', 'mac_address', 'vrf', 'is_ipv6', 'diff_type', 'return_only_diff', 'group_by']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ip_address', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ifname,
                cls.ip_address, cls.is_ipv6, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ifindex, cls.timestamp, cls.is_remote,
                cls.active, cls.mac_address, cls.deleted)

    def save(self):

        super(Neighbor, self).save_entry()

class LogExporter(_RedisModel):
    log_message = objects.String(json_key='log_message')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'log_message']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.log_message

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.log_message, cls.timestamp)

    def save(self):
        super(LogExporter, self).save_grpc()

class ARConfigGlobalV2(_RedisModel):
    app_name = 'arconfigglobalv2'
    container_hostname = 'netq-app-procdevstats'
    is_enabled = objects.Boolean(json_key='is_enabled')
    is_roce_enabled = objects.Boolean(json_key='is_roce_enabled')
    roce_mode = objects.String(json_key='roce_mode')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.is_enabled, cls.is_roce_enabled, cls.roce_mode,
                cls.active, cls.deleted, cls.timestamp)

    def save(self):
        super(ARConfigGlobalV2, self).save_entry()

class ARConfigIntf(_RedisModel):
    app_name = 'arconfigintf'
    container_hostname = 'netq-app-procdevstats'
    ifname = objects.String(json_key='ifname')
    is_enabled = objects.Boolean(json_key='is_enabled')
    link_util_thresh_disabled = objects.Boolean(json_key='link_util_thresh_disabled')
    link_util_thresh = objects.Integer(json_key='link_util_thresh')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.is_enabled, cls.link_util_thresh_disabled, cls.link_util_thresh,
                cls.active, cls.deleted, cls.timestamp)

    def save(self):
        super(ARConfigIntf, self).save_entry()

class AREcmpInfo(_RedisModel):
    app_name = 'arecmpinfo'
    container_hostname = 'netq-app-procdevstats'
    group_id = objects.Integer(json_key='group_id')
    ar_enabled = objects.Boolean(json_key='ar_enabled')
    mbr_list = objects.Json(json_key='mbr_list')
    ecmp_type = objects.String(json_key='ecmp_type')
    mbrs_ar_config_eq = objects.Boolean(json_key='mbrs_ar_config_eq')
    mbrs_qhist_p95_state = objects.Integer(json_key='mbrs_qhist_p95_state')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'group_id']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.group_id

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ar_enabled, cls.mbr_list, cls.ecmp_type,
                cls.mbrs_ar_config_eq, cls.mbrs_qhist_p95_state, cls.active, cls.deleted, cls.timestamp)

    def save(self):
        super(AREcmpInfo, self).save_entry()

class Node(_RedisModel):
    lastboot = objects.Float(json_key='lastBoot')
    sys_uptime = objects.Float(json_key='sysUptime')
    last_reinit = objects.Float(json_key='lastReinit')
    ntp_state = objects.String(json_key='ntpState')
    ptp_state = objects.Boolean(json_key ='ptpState')
    version = objects.String(json_key='version')
    appliance_version = objects.String(json_key='applianceVersion')
    manifest_hash = objects.String(json_key='manifestHash')
    app_name = 'node'
    post_request_args = ['node_names']
    container_hostname = 'netq-app-node'

    def __str__(self):
        return self.hostname

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['filter_rotten', 'filter_dead', 'filter_opta']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    def get_priority(self):
        return 1

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.lastboot, cls.sys_uptime, cls.last_reinit,
                cls.appliance_version, cls.manifest_hash,
                cls.ntp_state, cls.ptp_state, cls.cls.version, cls.timestamp)

    def save(self):
        self.save_grpc()


class Ntp(_RedisModel):
    message_type = objects.String(json_key='messageType')
    current_server = objects.String(json_key='currentServer')
    stratum = objects.Integer(json_key='stratum')
    ntp_sync = objects.String(json_key='ntpSync')
    ntp_app = objects.String(json_key='ntpApp')
    ptp_state = objects.Boolean(json_key='ptpState')
    optional_args_map = {'sync_state': 'in_sync'}
    app_name = 'ntp'
    container_hostname = 'netq-app-ntp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'in-sync':
                return (cls.optional_args_map.get(key, key), 'yes')
            elif val == 'out-of-sync':
                return (cls.optional_args_map.get(key, key), 'no')
        else:
            return ['sync_state']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.current_server, cls.stratum,
                cls.ntp_sync, cls.timestamp, cls.ntp_app, cls.ptp_state,
                cls.active, cls.deleted)

    def save(self):

        super(Ntp, self).save_entry()


class Port(_RedisModel):
    ifname = objects.String(json_key='interface')
    identifier = objects.String(json_key='module')
    speed = objects.String(json_key='speed')
    autoneg = objects.String(json_key='autoneg')
    state = objects.String(json_key='state')
    transreceiver = objects.String(json_key='transceiver')
    connector = objects.String(json_key='connector')
    vendor_name = objects.String(json_key='vendor')
    part_number = objects.String(json_key='partNo')
    serial_number = objects.String(json_key='serialNo')
    length = objects.String(json_key='length')
    supported_fec = objects.String(json_key='supportedFEC')
    advertised_fec = objects.String(json_key='advertisedFEC')
    fec = objects.String(json_key='FEC')
    app_name = 'port'
    container_hostname = 'netq-app-port'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.identifier, cls.speed, cls.autoneg, cls.state,
                cls.transreceiver, cls.connector, cls.length, cls.vendor_name,
                cls.part_number, cls.serial_number, cls.fec, cls.supported_fec, cls.advertised_fec,
                cls.active, cls.deleted, cls.timestamp)

    def save(self):

        super(Port, self).save_entry()

    def should_save_gnmi(self):
        return self.gnmi_enabled

class Dom(_RedisModel):
    ifname = objects.String(json_key='interface')
    identifier = objects.String(json_key='module')

    laser_rx_power = objects.Json(json_key='laserRxPower')
    laser_output_power = objects.Json(json_key='laserOpPower')
    laser_bias_current = objects.Json(json_key='laserBiasCurrent')
    module_temp = objects.String(json_key='modTemp')
    module_voltage = objects.String(json_key='modVoltage')

    laser_rx_power_high_alarm_th = objects.String(json_key='laserRxPowerHighAlarmTh')
    laser_rx_power_low_alarm_th = objects.String(json_key='laserRxPowerLowAlarmTh')
    laser_rx_power_high_warning_th = objects.String(json_key='laserRxPowerHighWarningTh')
    laser_rx_power_low_warning_th = objects.String(json_key='laserRxPowerLowWarningTh')

    laser_output_power_high_alarm_th = objects.String(json_key='laserOpPowerHighAlarmTh')
    laser_output_power_low_alarm_th = objects.String(json_key='laserOpPowerLowAlarmTh')
    laser_output_power_high_warning_th = objects.String(json_key='laserOpPowerHighWarningTh')
    laser_output_power_low_warning_th = objects.String(json_key='laserOpPowerLowWarningTh')

    module_temp_high_alarm_th = objects.String(json_key='modTempHighAlarmTh')
    module_temp_low_alarm_th = objects.String(json_key='modTempLowAlarmTh')
    module_temp_high_warning_th = objects.String(json_key='modTempHighWarningTh')
    module_temp_low_warning_th = objects.String(json_key='modTempLowWarningTh')

    laser_bias_current_high_alarm_th = objects.String(json_key='laserBiasCurrentHighAlarmTh')
    laser_bias_current_low_alarm_th = objects.String(json_key='laserBiasCurrentLowAlarmTh')
    laser_bias_current_high_warning_th = objects.String(json_key='laserBiasCurrentHighWarningTh')
    laser_bias_current_low_warning_th = objects.String(json_key='laserBiasCurrentLowWarningTh')

    module_voltage_high_alarm_th = objects.String(json_key='modVoltageHighAlarmTh')
    module_voltage_low_alarm_th =  objects.String(json_key='modVoltageLowAlarmTh')
    module_voltage_high_warning_th = objects.String(json_key='modVoltageHighWarningTh')
    module_voltage_low_warning_th = objects.String(json_key='modVoltageLowWarningTh')

    app_name = 'dom'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'path']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['ifname', 'type', 'channel_id']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.identifier, cls.laser_rx_power, cls.laser_output_power, cls.module_temp,
                cls.laser_bias_current, cls.module_voltage,
                cls.laser_rx_power_high_alarm_th, cls.laser_rx_power_low_alarm_th,
                cls.laser_rx_power_high_warning_th, cls.laser_rx_power_low_warning_th,
                cls.laser_output_power_high_alarm_th, cls.laser_output_power_low_alarm_th,
                cls.laser_output_power_high_warning_th, cls.laser_output_power_low_warning_th,
                cls.module_temp_high_alarm_th, cls.module_temp_low_alarm_th,
                cls.module_temp_high_warning_th, cls.module_temp_low_warning_th,
                cls.laser_bias_current_high_alarm_th, cls.laser_bias_current_low_alarm_th,
                cls.laser_bias_current_high_warning_th, cls.laser_bias_current_low_warning_th,
                cls.module_voltage_high_alarm_th, cls.module_voltage_low_alarm_th,
                cls.module_voltage_high_warning_th, cls.module_voltage_low_warning_th,
                cls.active, cls.deleted, cls.timestamp)

    def save(self):
        super(Dom, self).save_entry()


class BtrFsInfo(_RedisModel):
    device_size = objects.Long(json_key='device_size')
    device_allocated = objects.Long(json_key='device_allocated')
    device_unallocated = objects.Long(json_key='device_unallocated')
    device_missing = objects.Long(json_key='device_missing')
    used = objects.Long(json_key='used')
    free = objects.Long(json_key='free')
    data_ratio = objects.Float(json_key='data_ratio')
    metadata_ratio = objects.Float(json_key='metadata_ratio')
    global_reserve = objects.Long(json_key='global_reserve')
    device_single_size = objects.Long(json_key='device_single_size')
    device_single_used = objects.Long(json_key='device_single_used')
    device_metadata_size = objects.Long(json_key='device_metadata_size')
    device_metadata_used = objects.Long(json_key='device_metadata_used')
    device_system_size = objects.Long(json_key='device_system_size')
    device_system_used = objects.Long(json_key='device_system_used')
    app_name = 'btrfs'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.device_size, cls.device_allocated, cls.device_unallocated,
                cls.device_missing, cls.used, cls.free,
                cls.data_ratio, cls.metadata_ratio, cls.global_reserve,
                cls.device_single_size, cls.device_single_used,
                cls.device_metadata_size, cls.device_metadata_used,
                cls.device_system_size, cls.device_system_used,
                cls.active, cls.deleted,
                cls.timestamp)

    def save(self):

        super(BtrFsInfo, self).save_entry()


class SsdUtil(_RedisModel):
    health = objects.Float(json_key='health')
    avg_pe_cycles = objects.Integer(json_key='avg_pe_cycles')
    device_name = objects.String(json_key='device_name')
    app_name = 'ssdutil'
    container_hostname = 'netq-app-resource-util'
    optional_args_map = {'details': 'details'}

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['details']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.health, cls.avg_pe_cycles, cls.active,
                cls.device_name, cls.deleted, cls.timestamp)

    def save(self):

        super(SsdUtil, self).save_entry()


class PackageInfo(_RedisModel):
    package_name = objects.String(json_key='package_name')
    version = objects.String(json_key='version')
    status = objects.String(json_key='status')
    cl_version = objects.String(json_key='cl_version')
    app_name = 'packageinfo'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'package_name']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.package_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.package_name, cls.version, cls.status,
                cls.cl_version, cls.active, cls.deleted,
                cls.timestamp)

    def save(self):

        super(PackageInfo, self).save_entry()


class Manifest(_RedisModel):
    app_name = 'manifest'
    container_hostname = 'netq-app-inventory'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['release_id', 'cpu_arch', 'asic_vendor', 'package_name']


class PortStatsInterface(_RedisModel):
    ifname = objects.String(json_key='interface')
    rx_bytes = objects.Long(json_key='rxBytes')
    rx_drop = objects.Long(json_key='rxDrop')
    rx_errs = objects.Long(json_key='rxErrs')
    rx_frame = objects.Long(json_key='rxFrame')
    rx_multicast = objects.Long(json_key='rxMulticast')
    rx_packets = objects.Long(json_key='rxPackets')
    rx_util = objects.Float(json_key='rxUtil')
    tx_bytes = objects.Long(json_key='txBytes')
    tx_carrier = objects.Long(json_key='txCarrier')
    tx_colls = objects.Long(json_key='txColls')
    tx_drop = objects.Long(json_key='txDrop')
    tx_errs = objects.Long(json_key='txErrs')
    tx_packets = objects.Long(json_key='txPackets')
    tx_util = objects.Float(json_key='txUtil')
    port_speed = objects.String(json_key='portSpeed')
    rx_broadcast = objects.Long(json_key='rx_broadcast')
    tx_broadcast = objects.Long(json_key='tx_broadcast')
    tx_multicast = objects.Long(json_key='tx_multicast')
    app_name = 'procdevstats'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.rx_bytes, cls.rx_drop, cls.rx_errs,
                cls.rx_frame, cls.rx_multicast, cls.rx_packets,
                cls.tx_bytes, cls.tx_carrier, cls.tx_colls,
                cls.tx_drop, cls.tx_errs, cls.tx_packets,
                cls.active, cls.deleted, cls.timestamp)


class ProcNetDevInterface(_RedisModel):
    ifname = objects.String(json_key='interface')
    rx_bytes = objects.Long(json_key='rxBytes')
    rx_drop = objects.Long(json_key='rxDrop')
    rx_errs = objects.Long(json_key='rxErrs')
    rx_frame = objects.Long(json_key='rxFrame')
    rx_multicast = objects.Long(json_key='rxMulticast')
    rx_packets = objects.Long(json_key='rxPackets')
    tx_bytes = objects.Long(json_key='txBytes')
    tx_carrier = objects.Long(json_key='txCarrier')
    tx_colls = objects.Long(json_key='txColls')
    tx_drop = objects.Long(json_key='txDrop')
    tx_errs = objects.Long(json_key='txErrs')
    tx_packets = objects.Long(json_key='txPackets')
    rx_broadcast = objects.Long(json_key='rx_broadcast')
    tx_broadcast = objects.Long(json_key='tx_broadcast')
    tx_multicast = objects.Long(json_key='tx_multicast')
    app_name = 'procdevstatsraw'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.rx_bytes, cls.rx_drop, cls.rx_errs,
                cls.rx_frame, cls.rx_multicast, cls.rx_packets,
                cls.tx_bytes, cls.tx_carrier, cls.tx_colls,
                cls.tx_drop, cls.tx_errs, cls.tx_packets,
                cls.rx_broadcast, cls.tx_broadcast, cls.tx_multicast,
                cls.active, cls.deleted, cls.timestamp)

    def counter_fmt(self):
        return {
            'measurement': 'procnetdev',
            'time': datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S.%f'),
            'tags': {
                'hostname': self.hostname,
                'interface': self.ifname,
            },
            'fields': {
                'rxBytes': self.rx_bytes,
                'rxDrop': self.rx_drop,
                'rxErrs': self.rx_errs,
                'rxFrame': self.rx_frame,
                'rxMulticast': self.rx_multicast,
                'rxPackets': self.rx_packets,
                'txBytes': self.tx_bytes,
                'txCarrier': self.tx_carrier,
                'txColls': self.tx_colls,
                'txDrop': self.tx_drop,
                'txErrs': self.tx_errs,
                'txPackets': self.tx_packets,
            },
        }

    def counter_retention_policy(self):
        # Delete the downsampled 5m data after one year
        return 'CREATE RETENTION POLICY "one_year" ON "netq" DURATION 52w REPLICATION 1'

    def counter_continuous_query(self):

        # Aggregate the raw 15s data from the previous 5m, this runs ever 5m
        return """
CREATE CONTINUOUS QUERY "cq_5m" ON "netq" BEGIN
  SELECT
    MEAN("rxBytes") AS "mean_rxBytes",
    MEAN("rxDrop") AS "mean_rxDrop",
    MEAN("rxErrs") AS "mean_rxErrs",
    MEAN("rxFrame") AS "mean_rxFrame",
    MEAN("rxMulticast") AS "mean_rxMulticast",
    MEAN("rxPackets") AS "mean_rxPackets",
    MEAN("txBytes") AS "mean_txBytes",
    MEAN("txCarrier") AS "mean_txCarrier",
    MEAN("txColls") AS "mean_txColls",
    MEAN("txDrop") AS "mean_txDrop",
    MEAN("txErrs") AS "mean_txErrs",
    MEAN("txPackets") AS "mean_txPackets"
  INTO "one_year"."downsampled_stats"
  FROM "procnetdev"
  GROUP BY time(5m)
END
"""

    def should_save_gnmi(self):
        return self.gnmi_enabled

    def save(self):
        if self.opta_enabled:
            super(ProcNetDevInterface, self).save_grpc()
        if self.should_save_gnmi():
            super(ProcNetDevInterface, self).save_gnmi_grpc()


class IbInterface(_RedisModel):
    ifname = objects.String(json_key='interface')
    logical_port_state = objects.String(json_key='logicalPortState')
    physical_port_state = objects.String(json_key='physicalPortState')
    supported_ib_speeds = objects.String(json_key='supportedIbSpeeds')
    ib_speed = objects.String(json_key='ibSpeed')
    speed = objects.String(json_key='speed')
    speed_negotiate = objects.String(json_key='speedNegotiate')
    supported_widths = objects.String(json_key='supportedWidths')
    width = objects.String(json_key='width')
    max_supported_MTUs = objects.String(json_key='maxSupportedMTUs')
    mtu = objects.String(json_key='mtu')
    ib_subnet = objects.String(json_key='ibSubnet')
    vl_capabilities = objects.String(json_key='vlCapabilities')
    operational_vls = objects.String(json_key='operationalVls')

    #General interface
    ifindex = objects.String(json_key='ifIndex')
    oper_status = objects.String(json_key='operStatus')
    admin_status = objects.String(json_key='adminStatus')
    description = objects.String(json_key='description')
    in_pkts = objects.String(json_key='inPkts')
    out_pkts = objects.String(json_key='outPkts')
    in_errors = objects.String(json_key='inErrors')
    out_errors = objects.String(json_key='outErrors')
    in_octets = objects.String(json_key='inOctets')
    out_octets = objects.String(json_key='outOctets')
    in_discards = objects.String(json_key='inDiscards')
    out_discards = objects.String(json_key='outDiscards')

    #IB errors
    port_symbol_errors = objects.String(json_key='portSymbolErrors')
    port_rcv_errors = objects.String(json_key='PortRcvErrors')
    port_vl15_dropped = objects.String(json_key='PortVL15Dropped')
    port_xmit_wait = objects.String(json_key='PortXmitWait')

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.logical_port_state, cls.physical_port_state, cls.supported_ib_speeds,
                cls.ib_speed, cls.speed, cls.speed_negotiate,
                cls.supported_widths, cls.width, cls.max_supported_MTUs,
                cls.mtu, cls.ib_subnet, cls.vl_capabilities, cls.operational_vls)

    def counter_fmt(self):
        return {
            'measurement': 'ib_interface',
            'time': datetime.fromtimestamp(self.timestamp).strftime('%Y-%m-%d %H:%M:%S.%f'),
            'tags': {
                'hostname': self.hostname,
                'interface': self.ifname,
            },
            'fields': {
                'logicalPortState': self.logical_port_state,
                'physicalPortState': self.physical_port_state,
                'supportedIbSpeeds': self.supported_ib_speeds,
                'ibSpeed': self.ib_speed,
                'speed': self.speed,
                'speedNegotiate': self.speed_negotiate,
                'supportedWidths': self.supportedWidths,
                'width': self.width,
                'maxSupportedMTUs': self.max_supported_MTUs,
                'mtu': self.mtu,
                'ibSubnet': self.ib_subnet,
                'vlCapabilities': self.vl_capabilities,
                'operationalVls': self.operational_vls
            },
        }

    def should_save_gnmi(self):
        return self.gnmi_enabled

    def save(self):
        if self.should_save_gnmi():
            super(IbInterface, self).save_gnmi_grpc()

class PSU(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_name = objects.String(json_key='name')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')
    power_voltage_attributes = {}

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.s_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp, cls.s_state, cls.s_prev_state,
                cls.s_msg, cls.active, cls.deleted)

    def set_extra_attributes(self, key, value):
        self.power_voltage_attributes[key] = value

    def get_extra_attributes(self):
        return self.power_voltage_attributes

    def save(self):
        super(PSU, self).save_entry()


"""
The Power model is an extended version of the PSU we retrive it from the sensors -u command
"""


class Power(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_adapter_name = objects.String(json_key='adapter')
    s_voltage_in_input = objects.Float(json_key='vin')
    s_voltage_in_min = objects.Float(json_key='vinMin')
    s_voltage_in_max = objects.Float(json_key='vinMax')
    s_voltage_out_input = objects.Float(json_key='vout')
    s_voltage_out_min = objects.Float(json_key='voutMin')
    s_voltage_out_max = objects.Float(json_key='voutMax')
    s_power_in_input = objects.Float(json_key='pin')
    s_power_in_max = objects.Float(json_key='pinMax')
    s_power_out_input = objects.Float(json_key='pout')
    s_power_out_max = objects.Float(json_key='poutMax')

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.s_adapter_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp, cls.s_adapter_name, cls.s_voltage_in_input,
                cls.s_voltage_in_min, cls.s_voltage_in_max,
                cls.s_voltage_out_input, cls.s_voltage_out_min, cls.s_voltage_out_max,
                cls.s_power_in_input, cls.s_power_in_max,
                cls.s_power_out_input, cls.s_power_out_max,
                cls.active, cls.deleted)

    def save(self):
        super(Power, self).save_entry()


class Route(_RedisModel):
    message_type = objects.String(json_key='messageType')
    prefix = objects.String(json_key='prefix')
    protocol = objects.String(json_key='protocol')
    route_type = objects.Integer(json_key='routeType')
    origin = objects.Boolean(json_key='origin')
    nexthops = objects.Json(json_key='nextHops')
    src = objects.String(json_key='src')
    is_ipv6 = objects.Boolean(json_key='isIPv6')
    rt_table_id = objects.Integer(json_key='routeTableId')
    vrf = objects.String(json_key='vrf')
    priority = objects.Integer(json_key='priority')
    nh_id = objects.Integer(json_key='nhId')
    api_key_map = {'prefix': 'ipaddress'}
    optional_args_map = {'origin': 'is_origin'}
    app_name = 'route'
    container_hostname = 'netq-app-route'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['origin', 'is_ipv6', 'prefix_len']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'lpm', 'prefix', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.prefix, cls.is_ipv6, cls.origin, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.nexthops, cls.src, cls.timestamp, cls.active,
                cls.deleted, cls.priority, cls.route_type, cls.rt_table_id,
                cls.protocol, cls.nh_id)

    def save(self):

        super(Route, self).save_entry()


class Nexthop(_RedisModel):
    message_type = objects.String(json_key='messageType')
    nh_id = objects.Integer(json_key='nhId')
    is_group = objects.Boolean(json_key='isGroup')
    is_blackhole = objects.Boolean(json_key='isBlackhole')
    nexthops = objects.Json(json_key='nextHops')
    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.nh_id, cls.is_group, cls.is_blackhole,
                cls.nexthops, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(Nexthop, self).save_entry()

class Services(_RedisModel):
    name = objects.String(json_key='service')
    is_enabled = objects.Boolean(json_key='isEnabled')
    is_active = objects.Boolean(json_key='isActive')
    is_monitored = objects.Boolean(json_key='isMonitored')
    status = objects.Integer(json_key='status')
    vrf = objects.String(json_key='vrf')
    pid = objects.String(json_key='pid')
    start_time = objects.Float(json_key='startTime')
    optional_args_map = {'is_active': 'active', 'is_monitored': 'monitored', 'stat_string': 'status'}
    api_key_map = {'name': 'type'}
    app_name = 'services'
    container_hostname = 'netq-app-services'
    os_name = objects.String(json_key='os_name')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            if val == 'active' or val == 'monitored':
                return (cls.optional_args_map.get(key, key), 'yes')
            elif val == 'ok':
                return (cls.optional_args_map.get(key, key), ServiceStatus.OK)
            elif val == 'warning':
                return (cls.optional_args_map.get(key, key), ServiceStatus.WARNING)
            elif val == 'error':
                return (cls.optional_args_map.get(key, key), ServiceStatus.ERROR)
            elif val == 'fail':
                return (cls.optional_args_map.get(key, key), ServiceStatus.FAIL)
            else:
                return (cls.optional_args_map.get(key, key), val)
        else:
            return ['is_active', 'is_monitored', 'stat_string']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'name', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.is_enabled, cls.is_active, cls.status, cls.is_monitored,
                cls.start_time, cls.pid,
                cls.timestamp, cls.active, cls.deleted, cls.os_name)

    def save(self):

        super(Services, self).save_entry()


class Temp(_Sensor):
    message_type = objects.String(json_key='messageType')
    s_name = objects.String(json_key='name')
    s_desc = objects.String(json_key='description')
    s_input = objects.Float(json_key='temp')
    s_crit = objects.Integer(json_key='critical')
    s_lcrit = objects.Integer(json_key='lcritical')
    s_max = objects.Integer(json_key='max')
    s_min = objects.Integer(json_key='min')
    s_state = objects.String(json_key='state')
    s_prev_state = objects.String(json_key='prevState')
    s_msg = objects.String(json_key='message')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.s_name,
                cls.s_desc)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp, cls.s_state, cls.s_prev_state,
                cls.s_input, cls.active, cls.deleted,
                cls.s_msg, cls.s_crit, cls.s_max, cls.s_min, cls.s_lcrit)

    def save(self):
        super(Temp, self).save_entry()


class VxlanRemoteDest(_RedisModel):
    vni = objects.Integer(json_key='vni')
    rdst = objects.String(json_key='remoteDest')
    app_name = 'vxlan'
    container_hostname = 'netq-app-vxlan'

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.vni, cls.rdst

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['streaming']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'vni', 'rdst']

    @classmethod
    def val_fmt(cls):
        return (cls.vni, cls.rdst, cls.active, cls.deleted, cls.timestamp)

    def save(self):

        super(VxlanRemoteDest, self).save_entry()


# DATA FROM APPS ONLY #


class Vlan(_RedisModel):
    vlan = objects.String(json_key='vlan')
    ports = objects.String(json_key='ports')
    svi = objects.String(json_key='svi')
    app_name = 'vlan'
    container_hostname = 'netq-app-vlan'

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'vlan']


class Interface(_RedisModel):
    details = objects.String(json_key='details')
    ifname = objects.String(json_key='interface')
    state = objects.String(json_key='state')
    type = objects.String(json_key='vlan')
    vrf = objects.String(json_key='vrf')
    is_svd = objects.Boolean(json_key='isSvd')
    app_name = 'interface'
    api_filter_args = ['ifname', 'kind', 'state']
    container_hostname = 'netq-app-interface'

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'kind', 'state']


class Notification(_RedisModel):
    app_name = 'notifier'
    container_hostname = 'netq-app-notifier'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['filter', 'enable_slack', 'enable_pagerduty', 'proxy_url']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']



class EventsConfig(_RedisModel):
    app_name = 'eventsconfig'
    container_hostname = 'netq-app-notifier'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']



class CheckFilter(_RedisModel):
    app_name = 'checkfilter'
    container_hostname = 'netq-app-notifier'

    check_config_id = objects.String(json_key='checkConfigId')
    check_name = objects.String(json_key='checkName')
    function = objects.String(json_key='function')
    is_enabled = objects.String(json_key='isEnabled')
    is_event_suppressed = objects.String(json_key='isEventSuppressed')
    test_name = objects.String(json_key='testName')
    scope = objects.String(json_key='scope')
    suppress_until = objects.String(json_key='suppressUntil')
    is_active = objects.String(json_key='isActive')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, '')
        else:
            return ['path']

class CheckFilterObject(CheckFilter):
    check_filter_id = objects.String(json_key='checkFilterId')
    check_name = objects.String(json_key='checkName')
    function = objects.String(json_key='function')
    is_enabled = objects.String(json_key='isEnabled')
    is_event_suppressed = objects.String(json_key='isEventSuppressed')
    test_name = objects.String(json_key='testName')
    scope = objects.String(json_key='scope')
    suppress_until = objects.String(json_key='suppressUntil')
    is_active = objects.String(json_key='isActive')

    delete_request_args = ['check_filter_id']
    post_request_args = ['is_enabled', 'is_event_suppressed', 'function',
                         'is_active', 'suppress_until', 'scope']
    put_request_args = ['check_filter_id', 'is_enabled',
                        'is_event_suppressed', 'function',
                         'is_active', 'suppress_until', 'scope']

    @classmethod
    def val_fmt(cls):
        return (cls.check_filter_id,  cls.check_name, cls.scope,
                cls.suppress_until, cls.is_active)

class CheckCatalog(CheckFilter):
    check_name = objects.String(json_key='checkName')
    test_name = objects.String(json_key='testName')
    filter_attributes = objects.String(json_key='filterAttributes')
    description = objects.String(json_key='description')

    @classmethod
    def val_fmt(cls):
        return (cls.check_name, cls.test_name, cls.filter_attributes,
                cls.description)

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['check_type']


class Notifier(Notification):
    format = objects.String(json_key='format')
    message_type = objects.String(json_key='messageType')
    severity = objects.String(json_key='severity')
    message_id = objects.Integer(json_key='messageId')
    message_tag = objects.String(json_key='messageTag')

    @classmethod
    def val_fmt(cls):
        return (cls.format, cls.message_type, cls.severity)

class EventsHierarchy(EventsConfig):
    message_type = objects.String(json_key='messageType')
    filter_condition = objects.String(json_key='filterCondition')
    filter_condition_hierarchy = objects.String(json_key='filterConditionHierarchy')
    filter_condition_desc = objects.String(json_key='filterConditionDesc')

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.filter_condition, cls.filter_condition_hierarchy,
                cls.filter_condition_desc)

class EventsConfigObject(EventsConfig):

    eventsconfig_id = objects.String(json_key='eventsconfigId')
    eventsconfig_name = objects.String(json_key='eventsconfigName')
    events_id = objects.String(json_key='eventsId')
    scope = objects.String(json_key='scope')
    suppress_until = objects.String(json_key='suppressUntil')
    is_active = objects.String(json_key='isActive')

    delete_request_args = ['eventsconfig_id']
    post_request_args = ['eventsconfig_name', 'message_type', 'events_id',
                         'is_active', 'suppress_until', 'scope']
    put_request_args = ['eventsconfig_id', 'eventsconfig_name', 'message_type', 'events_id',
                         'is_active', 'suppress_until', 'scope']

    @classmethod
    def val_fmt(cls):
        return (cls.eventsconfig_id, cls.eventsconfig_name, cls.events_id, cls.scope,
                cls.suppress_until, cls.is_active)


class EventsConfigScope(EventsConfig):
    delete_request_args = ['id']
    post_request_args = ['scope_name', 'scope_help']



class Tca(_RedisModel):
    app_name = 'tcaconfig'
    container_hostname = 'netq-app-notifier'

    # delete_request_args = ['tca']
    # post_request_args = ['event_id', 'scope', 'severity', 'is_active', 'suppress_until', 'channels', 'threshold']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Event(_RedisModel):
    severity = objects.String(json_key='severity')
    message_id = objects.Integer(json_key='messageId')
    filter_message = objects.String(json_key='filterMessage')
    message = objects.String(json_key='message')
    app_name = 'events'
    container_hostname = 'netq-app-events'

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return []

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['hostnames', 'severity', 'message_types']


class EventIds(Tca):
    delete_request_args = ['id']
    post_request_args = ['id', 'help']


class TcaConfig(Tca):
    delete_request_args = ['tca_id']
    post_request_args = ['tca_id', 'event_id', 'severity', 'scope', 'is_active', 'suppress_until',
                         'channels', 'conditions', 'terminate_on_match', 'tca_name', 'is_auto_tca']


class TcaScope(Tca):
    delete_request_args = ['id']
    post_request_args = ['scope_name', 'scope_help']


class Filter(Notification):
    delete_request_args = ['filter']
    post_request_args = ['filter_name', 'channel_names', 'rule_names', 'terminate_on_match', 'severity', 'order_type',
                         'order_filter_name']


class Rule(Notification):
    delete_request_args = ['rule']
    post_request_args = ['rule_name', 'rule_json']


class Channel(Notification):
    delete_request_args = ['channel']
    post_request_args = ['channel_type', 'channel_name', 'severity', 'channel_info_json']


class Search(_RedisModel):
    hostname = objects.String(json_key='hostname')
    item_type = objects.String(json_key='itemType')
    ip_address = objects.String(json_key='ipAddress')
    app_name = 'search'
    container_hostname = 'netq-app-search'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['prefix']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Validation(_RedisModel):
    app_name = 'check'
    container_hostname = 'netq-app-tracecheck-scheduler'
    validation_type = objects.Json(json_key='validationType')
    post_request_args = ['validation_type', 'cadence_min', 'alert_on_failure',
                         'start_time', 'validation_name', 'action']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['type', 'by', 'proto', 'time', 'duration']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Trace(_RedisModel):
    app_name = 'trace'
    container_hostname = 'netq-app-tracecheck-scheduler'
    post_request_args = ['data']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['type', 'time', 'duration', 'bucket', 'by']

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path']


class Label(_RedisModel):
    app_name = 'label'
    container_hostname = 'netq-app-node'
    put_request_args = ['labelname', 'labeltype', 'labelmap']

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        return []

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['labelname']


class WjhL1DropAggregate(_RedisModel):
    app_name = 'wjh'
    container_hostname = 'netq-app-wjh'

    ingress_port = objects.String(json_key='ingressPort')
    is_port_up = objects.Boolean(json_key='isPortUp')
    port_down_reason = objects.String(json_key='portDownReason')
    description = objects.String(json_key='description')
    state_change_count = objects.Long(json_key='stateChangeCount')
    symbol_error_count = objects.Long(json_key='symbolErroriCount')
    crc_error_count = objects.Long(json_key='crcErrorCount')
    first_timestamp = objects.Float(json_key='firstTimeStamp')
    last_timestamp = objects.Float(json_key='lastTimeStamp')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['ingress_port', 'reason', 'port_aggregate']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'l1-drops']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ingress_port)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.is_port_up, cls.port_down_reason, cls.description,
                cls.state_change_count, cls.symbol_error_count,
                cls.crc_error_count, cls.first_timestamp,
                cls.last_timestamp, cls.active, cls.deleted, cls.timestamp)

    def save(self):

        super(WjhL1DropAggregate, self).save_entry()

    def should_save_gnmi(self):
        return self.gnmi_enabled

class WjhDropAggregate(_RedisModel):
    app_name = 'wjh'
    container_hostname = 'netq-app-wjh'

    reason_id = objects.Integer(json_key='reason_id')
    reason = objects.String(json_key='reason')
    drop_type = objects.String(json_key='dropType')
    ingress_port = objects.String(json_key='ingressPort')
    ingress_lag = objects.String(json_key='ingressLag')
    egress_port = objects.String(json_key='egressPort')
    traffic_class = objects.Integer(json_key='trafficClass')
    acl_rule_id = objects.Long(json_key='ruleId')
    acl_bind_point = objects.Integer(json_key='bindingPoint')
    acl_name = objects.String(json_key='aclName')
    acl_rule = objects.String(json_key='rule')
    smac = objects.String(json_key='srcMac')
    dmac = objects.String(json_key='dstMac')
    sip = objects.String(json_key='sourceIp')
    dip = objects.String(json_key='destIp')
    proto = objects.Integer(json_key='proto')
    sport = objects.Integer(json_key='srcPort')
    dport = objects.Integer(json_key='dstPort')
    agg_count = objects.Long(json_key='aggregateCount')
    original_occupancy = objects.Integer(json_key='originalOccupancy')
    original_latency = objects.Integer(json_key='originalLatency')
    first_timestamp = objects.Float(json_key='firstTimeStamp')
    last_timestamp = objects.Float(json_key='lastTimeStamp')
    severity = objects.String(json_key='severity')
    vlan = objects.Integer(json_key="vlan")


    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return ['ingress_port', 'reason_id', 'reason', 'src_ip', 'dst_ip', 'proto', 'src_port', 'dst_port', 'src_mac', 'dst_mac',
                    'egress_port', 'traffic_class', 'acl_rule_id', 'summary', 'severity', "vlan"]

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', "l2-drops", "router-drops", "tunnel-drops", "buffer-drops", "acl-drops"]

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ingress_port, cls.reason,
                cls.drop_type, cls.smac, cls.dmac, cls.sip, cls.dip, cls.proto,
                cls.sport, cls.dport, cls.traffic_class, cls.egress_port,
                cls.acl_rule_id)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.first_timestamp, cls.last_timestamp, cls.agg_count,
                cls.acl_bind_point, cls.acl_name, cls.acl_rule,
                cls.active, cls.deleted, cls.timestamp)

    def save(self):
        if self.opta_enabled:
            super(WjhDropAggregate, self).save_grpc()
        if self.should_save_gnmi():
            super(WjhDropAggregate, self).save_gnmi_grpc()

    def should_save_gnmi(self):
        return self.gnmi_enabled

class WjhDropRawInfo(_RedisModel):
    app_name = 'wjh'
    container_hostname = 'netq-app-wjh'

    reason_id = objects.Long(json_key='reasonId')
    reason = objects.String(json_key='reason')
    severity = objects.String(json_key='severity')
    description = objects.String(json_key='description')
    ingress_port = objects.String(json_key='ingressPort')
    packet = objects.String(json_key='packet')
    packet_size = objects.Long(json_key='packetSize')
    drop_type = objects.String(json_key='dropType')

    acl_bind_point = objects.Integer(json_key='bindingPoint')
    acl_rule_id = objects.Long(json_key='ruleId')
    acl_name = objects.String(json_key='aclName')
    acl_rule = objects.String(json_key='rule')

    egress_port = objects.String(json_key='egressPort')
    traffic_class = objects.String(json_key='trafficClass')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return (cls.optional_args_map.get(key, key), val)
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'reason_id', 'ingress_port', 'drop_type']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.ingress_port, cls.reason_id, cls.drop_type)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.reason, cls.severity, cls.description, cls.acl_bind_point,
                cls.acl_name, cls.acl_rule, cls.acl_rule_id, cls.egress_port,
                cls.traffic_class, cls.packet, cls.packet_size, cls.active,
                cls.deleted, cls.timestamp)

    def save(self):

        super(WjhDropRawInfo, self).save_entry()


class EthtoolInterface(_RedisModel):
    ifname = objects.String(json_key='interface')
    hw_if_in_octets = objects.Long(json_key='HwIfInOctets')
    hw_if_in_ucast_pkts = objects.Long(json_key='HwIfInUcastPkts')
    hw_if_in_bcast_pkts = objects.Long(json_key='HwIfInBcastPkts')
    hw_if_in_mcast_pkts = objects.Long(json_key='HwIfInMcastPkts')
    hw_if_out_octets = objects.Long(json_key='HwIfOutOctets')
    hw_if_out_ucast_pkts = objects.Long(json_key='HwIfOutUcastPkts')
    hw_if_out_mcast_pkts = objects.Long(json_key='HwIfOutMcastPkts')
    hw_if_out_bcast_pkts = objects.Long(json_key='HwIfOutBcastPkts')
    hw_if_in_discards = objects.Long(json_key='HwIfInDiscards')
    hw_if_in_l3_drops = objects.Long(json_key='HwIfInL3Drops')
    hw_if_in_buffer_drops = objects.Long(json_key='HwIfInBufferDrops')
    hw_if_in_acl_drops = objects.Long(json_key='HwIfInAclDrops')
    hw_if_in_dot3_length_errors = objects.Long(json_key='HwIfInDot3LengthErrors')
    hw_if_in_errors = objects.Long(json_key='HwIfInErrors')
    soft_in_errors = objects.Long(json_key='SoftInErrors')
    soft_in_drops = objects.Long(json_key='SoftInDrops')
    soft_in_frame_errors = objects.Long(json_key='SoftInFrameErrors')
    hw_if_out_discards = objects.Long(json_key='HwIfOutDiscards')
    hw_if_out_errors = objects.Long(json_key='HwIfOutErrors')
    hw_if_out_q_drops = objects.Long(json_key='HwIfOutQDrops')
    hw_if_out_non_q_drops = objects.Long(json_key='HwIfOutNonQDrops')
    soft_out_errors = objects.Long(json_key='SoftOutErrors')
    soft_out_drops = objects.Long(json_key='SoftOutDrops')
    soft_out_tx_fifo_full = objects.Long(json_key='SoftOutTxFifoFull')
    hw_if_out_q_len = objects.Long(json_key='HwIfOutQLen')
    hw_if_in_dot3_frame_errors = objects.Long(json_key='HwIfInDot3FrameErrors')
    hw_if_symbol_errors = objects.Long(json_key='HwIfSymbolErrors')
    hw_if_alignment_errors = objects.Long(json_key='HwIfAlignmentErrors')
    hw_if_undersize_errors = objects.Long(json_key='HwIfUndersizeErrors')
    hw_if_oversize_errors = objects.Long(json_key='HwIfOversizeErrors')
    hw_if_jabber_errors = objects.Long(json_key='HwIfJabberErrors')
    hw_if_in_pause_pkt = objects.Long(json_key='HwIfInPausePkt')
    hw_if_out_pause_pkt = objects.Long(json_key='HwIfOutPausePkt')
    hw_if_in_pfc0_pkt = objects.Long(json_key='HwIfInPfc0Pkt')
    hw_if_out_pfc0_pkt = objects.Long(json_key='HwIfOutPfc0Pkt')
    hw_if_in_pfc1_pkt = objects.Long(json_key='HwIfInPfc1Pkt')
    hw_if_out_pfc1_pkt = objects.Long(json_key='HwIfOutPfc1Pkt')
    hw_if_in_pfc2_pkt = objects.Long(json_key='HwIfInPfc2Pkt')
    hw_if_out_pfc2_pkt = objects.Long(json_key='HwIfOutPfc2Pkt')
    hw_if_in_pfc3_pkt = objects.Long(json_key='HwIfInPfc3Pkt')
    hw_if_out_pfc3_pkt = objects.Long(json_key='HwIfOutPfc3Pkt')
    hw_if_in_pfc4_pkt = objects.Long(json_key='HwIfInPfc4Pkt')
    hw_if_out_pfc4_pkt = objects.Long(json_key='HwIfOutPfc4Pkt')
    hw_if_in_pfc5_pkt = objects.Long(json_key='HwIfInPfc5Pkt')
    hw_if_out_pfc5_pkt = objects.Long(json_key='HwIfOutPfc5Pkt')
    hw_if_in_pfc6_pkt = objects.Long(json_key='HwIfInPfc6Pkt')
    hw_if_out_pfc6_pkt = objects.Long(json_key='HwIfOutPfc6Pkt')
    hw_if_in_pfc7_pkt = objects.Long(json_key='HwIfInPfc7Pkt')
    hw_if_out_pfc7_pkt = objects.Long(json_key='HwIfOutPfc7Pkt')
    hw_if_out_wred_drops = objects.Long(json_key='HwIfOutWredDrops')
    hw_if_out_q0_wred_drops = objects.Long(json_key='HwIfOutQ0WredDrops')
    hw_if_out_q1_wred_drops = objects.Long(json_key='HwIfOutQ1WredDrops')
    hw_if_out_q2_wred_drops = objects.Long(json_key='HwIfOutQ2WredDrops')
    hw_if_out_q3_wred_drops = objects.Long(json_key='HwIfOutQ3WredDrops')
    hw_if_out_q4_wred_drops = objects.Long(json_key='HwIfOutQ4WredDrops')
    hw_if_out_q5_wred_drops = objects.Long(json_key='HwIfOutQ5WredDrops')
    hw_if_out_q6_wred_drops = objects.Long(json_key='HwIfOutQ6WredDrops')
    hw_if_out_q7_wred_drops = objects.Long(json_key='HwIfOutQ7WredDrops')
    hw_if_out_q8_wred_drops = objects.Long(json_key='HwIfOutQ8WredDrops')
    hw_if_out_q9_wred_drops = objects.Long(json_key='HwIfOutQ9WredDrops')
    hw_if_in_64_octets_pkt = objects.Long(json_key='HwIfIn64OctetsPkts')
    hw_if_in_65_to_127_octets_pkt = objects.Long(json_key='HwIfIn65To127OctetsPkts')
    hw_if_in_128_to_255_octets_pkt = objects.Long(json_key='HwIfIn128To255OctetsPkts')
    hw_if_in_256_to_511_octets_pkt = objects.Long(json_key='HwIfIn256To511OctetsPkts')
    hw_if_in_512_to_1023_octets_pkt = objects.Long(json_key='HwIfIn512To1023OctetsPkts')
    hw_if_in_512_to_1518_octets_pkt = objects.Long(json_key='HwIfIn1024To1518OctetsPkts')

    app_name = 'ethtoolraw'

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.hw_if_in_octets,
                cls.hw_if_in_ucast_pkts,
                cls.hw_if_in_bcast_pkts,
                cls.hw_if_in_mcast_pkts,
                cls.hw_if_out_octets,
                cls.hw_if_out_ucast_pkts,
                cls.hw_if_out_mcast_pkts,
                cls.hw_if_out_bcast_pkts,
                cls.hw_if_in_discards,
                cls.hw_if_in_l3_drops,
                cls.hw_if_in_buffer_drops,
                cls.hw_if_in_acl_drops,
                cls.hw_if_in_dot3_length_errors,
                cls.hw_if_in_errors,
                cls.soft_in_errors,
                cls.soft_in_drops,
                cls.soft_in_frame_errors,
                cls.hw_if_out_discards,
                cls.hw_if_out_errors,
                cls.hw_if_out_q_drops,
                cls.hw_if_out_non_q_drops,
                cls.soft_out_errors,
                cls.soft_out_drops,
                cls.soft_out_tx_fifo_full,
                cls.hw_if_out_q_len,
                cls.hw_if_in_dot3_frame_errors,
                cls.hw_if_in_pause_pkt,
                cls.hw_if_out_pause_pkt,
                cls.hw_if_in_pfc0_pkt,
                cls.hw_if_out_pfc0_pkt,
                cls.hw_if_in_pfc1_pkt,
                cls.hw_if_out_pfc1_pkt,
                cls.hw_if_in_pfc2_pkt,
                cls.hw_if_out_pfc2_pkt,
                cls.hw_if_in_pfc3_pkt,
                cls.hw_if_out_pfc3_pkt,
                cls.hw_if_in_pfc4_pkt,
                cls.hw_if_out_pfc4_pkt,
                cls.hw_if_in_pfc5_pkt,
                cls.hw_if_out_pfc5_pkt,
                cls.hw_if_in_pfc6_pkt,
                cls.hw_if_out_pfc6_pkt,
                cls.hw_if_in_pfc7_pkt,
                cls.hw_if_out_pfc7_pkt,
                cls.hw_if_out_wred_drops,
                cls.hw_if_out_q0_wred_drops,
                cls.hw_if_out_q1_wred_drops,
                cls.hw_if_out_q2_wred_drops,
                cls.hw_if_out_q3_wred_drops,
                cls.hw_if_out_q4_wred_drops,
                cls.hw_if_out_q5_wred_drops,
                cls.hw_if_out_q6_wred_drops,
                cls.hw_if_out_q7_wred_drops,
                cls.hw_if_out_q8_wred_drops,
                cls.hw_if_out_q9_wred_drops,
                cls.active, cls.deleted, cls.timestamp)

    def save(self):
        if self.opta_enabled:
            super(EthtoolInterface, self).save_grpc()
        if self.should_save_gnmi():
            super(EthtoolInterface, self).save_gnmi_grpc()

    def should_save_gnmi(self):
        return self.gnmi_enabled


class Roce(_RedisModel):
    ifname = objects.String(json_key='interface')
    port_id = objects.Long(json_key='portId')
    tx_cnp_tc_packets = objects.Long(json_key='txCnpTcPackets')
    tx_cnp_buffer_max_usage_cells = objects.Long(json_key='txCnpBufferMaxUsageCells')
    rx_roce_pg_max_usage_cells = objects.Long(json_key='rxRocePgMaxUsageCells')
    rx_roce_pg_packets = objects.Long(json_key='rxRocePgPackets')
    rx_roce_pfc_pause_packets = objects.Long(json_key='rxRocePfcPausePackets')
    tx_roce_tc_usage_cells = objects.Long(json_key='txRoceTcUsageCells')
    tx_cnp_buffer_usage_cells = objects.Long(json_key='txCnpBufferUsageCells')
    rx_cnp_pg_packets = objects.Long(json_key='rxCnpPgPackets')
    rx_roce_pfc_pause_duration = objects.Long(json_key='rxRocePfcPauseDuration')
    tx_roce_tc_bytes = objects.Long(json_key='txRoceTcBytes')
    rx_cnp_pg_bytes = objects.Long(json_key='rxCnpPgBytes')
    tx_roce_buffer_max_usage_cells = objects.Long(json_key='txRoceBufferMaxUsageCells')
    rx_cnp_buffer_max_usage_cells = objects.Long(json_key='rxCnpBufferMaxUsageCells')
    rx_cnp_no_buffer_discard = objects.Long(json_key='rxCnpNoBufferDiscard')
    tx_roce_pfc_pause_duration = objects.Long(json_key='txRocePfcPauseDuration')
    tx_ecn_marked_packets = objects.Long(json_key='txEcnMarkedPackets')
    rx_roce_no_buffer_discard = objects.Long(json_key='rxRoceNoBufferDiscard')
    cell_unit_size = objects.Long(json_key='cellUnitSize')
    rx_roce_pg_usage_cells = objects.Long(json_key='rxRocePgUsageCells')
    rx_roce_buffer_usage_cells = objects.Long(json_key='rxRoceBufferUsageCells')
    tx_cnp_tc_bytes = objects.Long(json_key='txCnpTcBytes')
    tx_cnp_tc_max_usage_cells = objects.Long(json_key='txCnpTcMaxUsageCells')
    tx_roce_unicast_no_buffer_discard = objects.Long(json_key='txRroceUnicastNoBufferDiscard')
    tx_roce_tc_max_usage_cells = objects.Long(json_key='txRoceTcMaxUsageCells')
    rx_cnp_pg_usage_cells = objects.Long(json_key='rxCnpPgUsageCells')
    tx_cnp_unicast_no_buffer_discard = objects.Long(json_key='txCnpUnicastNoBufferDiscard')
    rx_roce_buffer_max_usage_cells = objects.Long(json_key='rxRoceBufferMaxUsageCells')
    rx_cnp_pg_max_usage_cells = objects.Long(json_key='rxCnpPgMaxUsageCells')
    tx_roce_pfc_pause_packets = objects.Long(json_key='txRocePfcPausePackets')
    rx_roce_pg_bytes = objects.Long(json_key='rxRocePgBytes')
    tx_roce_buffer_usage_cells = objects.Long(json_key='txRoceBufferUsageCells')
    rx_cnp_buffer_usage_cells = objects.Long(json_key='rxCnpBufferUsageCells')
    tx_cnp_tc_usage_cells = objects.Long(json_key='txCnpTcUsageCells')
    tx_roce_tc_packets = objects.Long(json_key='txRoceTcPackets')
    lossy_default_ingress_size = objects.Long(json_key='lossyDefaultIngressSize')
    roce_reserved_ingress_size = objects.Long(json_key='roceReservedIngressSize')
    lossy_default_egress_size = objects.Long(json_key='lossyDefaultEgressSize')
    roce_reserved_egress_size = objects.Long(json_key='roceReservedEgressSize')
    app_name = 'roce'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def exclude_fmt(cls):
        return []

    @classmethod
    def val_fmt(cls):
        return (cls.message_type,
                cls.ifname,
                cls.port_id,
                cls.tx_cnp_tc_packets,
                cls.tx_cnp_buffer_max_usage_cells,
                cls.rx_roce_pg_max_usage_cells,
                cls.rx_roce_pg_packets,
                cls.rx_roce_pfc_pause_packets,
                cls.tx_roce_tc_usage_cells,
                cls.tx_cnp_buffer_usage_cells,
                cls.rx_cnp_pg_packets,
                cls.rx_roce_pfc_pause_duration,
                cls.tx_roce_tc_bytes,
                cls.rx_cnp_pg_bytes,
                cls.tx_roce_buffer_max_usage_cells,
                cls.rx_cnp_buffer_max_usage_cells,
                cls.rx_cnp_no_buffer_discard,
                cls.tx_roce_pfc_pause_duration,
                cls.tx_ecn_marked_packets,
                cls.rx_roce_no_buffer_discard,
                cls.cell_unit_size,
                cls.rx_roce_pg_usage_cells,
                cls.rx_roce_buffer_usage_cells,
                cls.tx_cnp_tc_bytes,
                cls.tx_cnp_tc_max_usage_cells,
                cls.tx_roce_unicast_no_buffer_discard,
                cls.tx_roce_tc_max_usage_cells,
                cls.rx_cnp_pg_usage_cells,
                cls.tx_cnp_unicast_no_buffer_discard,
                cls.rx_roce_buffer_max_usage_cells,
                cls.rx_cnp_pg_max_usage_cells,
                cls.tx_roce_pfc_pause_packets,
                cls.rx_roce_pg_bytes,
                cls.tx_roce_buffer_usage_cells,
                cls.rx_cnp_buffer_usage_cells,
                cls.tx_cnp_tc_usage_cells,
                cls.tx_roce_tc_packets,
                cls.lossy_default_ingress_size,
                cls.roce_reserved_ingress_size,
                cls.lossy_default_egress_size,
                cls.roce_reserved_egress_size,
                cls.timestamp,
                cls.active,
                cls.deleted)

    def save(self):
        super(Roce, self).save_entry()


class HostDevicesRoce(_RedisModel):
    device_name = objects.String(json_key='DeviceName')
    port_name = objects.String(json_key='PortName')
    rx_prio3_bytes = objects.Long(json_key='RxPrio3Bytes')
    rx_prio3_packets = objects.Long(json_key='RxPrio3Packets')
    rx_prio3_discards = objects.Long(json_key='RxPrio3Discards')
    tx_prio3_bytes = objects.Long(json_key='TxPrio3Bytes')
    tx_prio3_packets = objects.Long(json_key='TxPrio3Packets')
    rx_prio3_buf_discard = objects.Long(json_key='RxPrio3BufDiscard')
    rx_prio3_cong_discard = objects.Long(json_key='RxPrio3CongDiscard')
    np_cnp_sent = objects.Long(json_key='NpCnpSent')
    np_ecn_marked_roce_packets = objects.Long(json_key='NpEcnMarkedRocePackets')
    rp_cnp_handled = objects.Long(json_key='RpCnpHandled')
    rp_cnp_ignored = objects.Long(json_key='RpCnpIgnored')

    app_name = 'hostdroce'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ['hostname', 'device_name', 'device_type']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return []

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.device_name

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def exclude_fmt(cls):
        return []

    @classmethod
    def val_fmt(cls):
        return (cls.message_type,
                cls.device_name,
                cls.port_name,
                cls.rx_prio3_bytes,
                cls.rx_prio3_packets,
                cls.rx_prio3_discards,
                cls.tx_prio3_bytes,
                cls.tx_prio3_packets,
                cls.rx_prio3_buf_discard,
                cls.rx_prio3_cong_discard,
                cls.np_cnp_sent,
                cls.np_ecn_marked_roce_packets,
                cls.rp_cnp_handled,
                cls.rp_cnp_ignored,
                cls.timestamp,
                cls.active,
                cls.deleted)

    def save(self):
        super(HostDevicesRoce, self).save_entry()


class NvueRoceConfig(_RedisModel):
    ifname = objects.String(json_key='interface')
    congestion_mode = objects.String(json_key='congestionMode')
    min_threshold = objects.Long(json_key='minThreshold')
    max_threshold = objects.Long(json_key='maxThreshold')
    enabled_tc = objects.String(json_key='enabledTc')
    probability = objects.String(json_key='probability')
    exception_list = objects.Json(json_key='exceptionList')
    lldp_priority = objects.Long(json_key='lldpPriority')
    lldp_protocol_id = objects.Long(json_key='lldpPriority')
    lldp_selector = objects.String(json_key='lldpSelector')
    roce_mode = objects.String(json_key='roceMode')
    pfc_priority = objects.String(json_key='pfcPriority')
    rx_enabled = objects.String(json_key='rxEnabled')
    tx_enabled = objects.String(json_key='txEnabled')
    pool_map = objects.Json(json_key='poolMap')
    priority_map = objects.Json(json_key='priorityMap')
    tc_map = objects.Json(json_key='tcMap')
    trust_mode = objects.String(json_key='trustMode')

    app_name = 'nvueroceconfig'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ifname, cls.congestion_mode, cls.min_threshold, cls.max_threshold, cls.enabled_tc,
                cls.probability, cls.exception_list, cls.lldp_priority, cls.lldp_protocol_id, cls.lldp_selector,
                cls.roce_mode, cls.pfc_priority, cls.rx_enabled, cls.tx_enabled, cls.pool_map, cls.priority_map,
                cls.tc_map, cls.trust_mode, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(NvueRoceConfig, self).save_entry()


class RoceConfig(_RedisModel):
    ifname = objects.String(json_key='interface')
    enabled_tc = objects.String(json_key='enabledTc')
    roce_mode = objects.String(json_key='roceMode')
    mode = objects.String(json_key='mode')
    ecn_min_th = objects.Long(json_key='ecnMinTh')
    ecn_max_th = objects.Long(json_key='ecnMaxTh')
    pfc_tx_enabled = objects.Boolean(json_key='pfcTxEnabled')
    pfc_rx_enabled = objects.Boolean(json_key='pfcRxEnabled')
    pfc_enabled_sp = objects.String(json_key='pfcEnabledCp')
    dscp_sp_map = objects.Json(json_key='dscpSpMap')
    pcp_sp_map = objects.Json(json_key='pcpSpMap')
    pcp_unmapped = objects.Boolean(json_key='pcpUnmapped')
    cnp_dscp = objects.Json(json_key='cnpDscp')
    cnp_pcp = objects.Json(json_key='cnpPcp')
    cnp_pg_map = objects.Json(json_key='cnpPgMap')
    cnp_tc_map = objects.Json(json_key='cnpTcMap')
    sp_tc_map = objects.Json(json_key='spTcMap')
    sp_pg_map = objects.Json(json_key='spPgMap')
    ets_mode = objects.String(json_key='etsMode')
    cnp_ets_mode = objects.String(json_key='cnpEtsMode')
    nvue_enabled = objects.Boolean(json_key='nvueEnabled')
    app_name = 'roceconfig'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ifname, cls.roce_mode, cls.enabled_tc, cls.mode, cls.ecn_min_th,
                cls.cnp_dscp, cls.cnp_pcp, cls.pcp_unmapped, cls.cnp_pg_map, cls.cnp_tc_map, cls.cnp_ets_mode,
                cls.ecn_max_th, cls.pfc_tx_enabled, cls.pfc_rx_enabled, cls.pfc_enabled_sp, cls.dscp_sp_map,
                cls.pcp_sp_map, cls.sp_tc_map, cls.sp_pg_map, cls.ets_mode, cls.nvue_enabled, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(RoceConfig, self).save_entry()


class EcmpHash(_RedisModel):
    symmetric_hash_enable = objects.Boolean(json_key='symmetricHashEnable')
    resilient_hash_enable = objects.Boolean(json_key='resilientHashEnable')
    seed = objects.Long(json_key='seed')
    max_paths = objects.Long(json_key='maxPaths')
    resilient_hash_entries = objects.Long(json_key='resilientHashEntries')
    ip_prot = objects.Boolean(json_key='ipProt')
    sip = objects.Boolean(json_key='sip')
    dip = objects.Boolean(json_key='dip')
    sport = objects.Boolean(json_key='sport')
    dport = objects.Boolean(json_key='dport')
    ip6_label = objects.Boolean(json_key='ip6Label')
    ing_intf = objects.Boolean(json_key='ingIntf')
    inner_ip_prot = objects.Boolean(json_key='innerIpProt')
    inner_sip = objects.Boolean(json_key='innerSip')
    inner_dip = objects.Boolean(json_key='innerDip')
    inner_sport = objects.Boolean(json_key='innerSport')
    inner_dport = objects.Boolean(json_key='innerDport')
    inner_ip6_label = objects.Boolean(json_key='innerIp6Label')
    app_name = 'ecmp'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'hash']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.symmetric_hash_enable, cls.resilient_hash_enable, cls.seed,
                cls.max_paths, cls.resilient_hash_entries, cls.ip_prot, cls.sip, cls.dip, cls.sport, cls.dport,
                cls.ip6_label, cls.ing_intf, cls.inner_ip_prot, cls.inner_sip, cls.inner_dip, cls.inner_sport,
                cls.inner_dport, cls.inner_ip6_label, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(EcmpHash, self).save_entry()


class EcmpInfo(_RedisModel):
    group_id = objects.Integer(json_key='groupId')
    rif = objects.Integer(json_key='rif')
    bytes = objects.Long(json_key='bytes')
    weight = objects.Integer(json_key='weight')
    ifname = objects.String(json_key='ifname')
    app_name = 'ecmp'
    container_hostname = 'netq-app-procdevstats'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ["group_id", "ifname"]

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'info-group-aggregate', 'info-path-aggregate']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.group_id, cls.rif, cls.bytes, cls.weight, cls.ifname, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(EcmpInfo, self).save_grpc()

class SpiceAmberGlobal(_RedisModel):
    spice_path = objects.String(json_key='spice_path')
    spice_value = objects.Json(json_key='spice_value')
    spice_ts = objects.Float(json_key='spice_ts')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'spice_path']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.spice_path

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.spice_value, cls.spice_ts, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        if self.opta_enabled:
            super(SpiceAmberGlobal, self).save_grpc()


class SpiceAmberPortLevel(_RedisModel):
    port_num = objects.Integer(json_key='port_num')
    spice_path = objects.String(json_key='spice_path')
    spice_value = objects.Json(json_key='spice_value')
    spice_ts = objects.Float(json_key='spice_ts')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'port_num', 'spice_path']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.port_num, cls.spice_path

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.spice_value, cls.spice_ts, cls.timestamp, cls.active, cls.deleted)

    def should_save_gnmi(self):
        return False

    def save(self):
        if self.opta_enabled:
            super(SpiceAmberPortLevel, self).save_grpc()

class PtpGlobalConfig(_RedisModel):
    slave_only = objects.Boolean(json_key='slaveOnly')
    priority1 = objects.Integer(json_key='priority1')
    priority2 = objects.Integer(json_key='priority2')
    domain_number = objects.Integer(json_key='domainNumber')
    two_step_flag = objects.Boolean(json_key='twoStepFlag')
    ofm_min_threshold = objects.Integer(json_key='ofmMinThreshold')
    ofm_max_threshold = objects.Integer(json_key='ofmMaxThreshold')
    mpd_threshold = objects.Integer(json_key='mpdThreshold')
    app_name = 'ptp'
    container_hostname = 'netq-app-ptp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'path', 'domain_number']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.slave_only, cls.priority1, cls.priority2, cls.domain_number,
                cls.two_step_flag, cls.ofm_min_threshold, cls.ofm_max_threshold, cls.mpd_threshold,
                cls.timestamp, cls.active, cls.deleted)

    def save(self):
        if self.opta_enabled:
            super(PtpGlobalConfig, self).save_entry()

class PtpClockInfo(_RedisModel):
    clock_identity = objects.String(json_key='clockIdentity')
    parent_clock_identity = objects.String(json_key='parentClockIdentity')
    gm_identity = objects.String(json_key='gmIdentity')
    clock_domain = objects.Integer(json_key='clockDomain')
    num_ptp_ports = objects.Integer(json_key='numPtpPorts')
    cq_quality = objects.Long(json_key='cqQuality')
    cq_accuracy = objects.Long(json_key='cqAccuracy')
    steps_removed = objects.Integer(json_key='stepsRemoved')
    gm_cq_class = objects.Long(json_key='gmCqClass')
    gm_cq_accuracy = objects.Long(json_key='gmCqAccuracy')
    app_name = 'ptp'
    container_hostname = 'netq-app-ptp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ['hostname']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'path']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.clock_identity, cls.parent_clock_identity, cls.gm_identity,
                cls.clock_domain, cls.num_ptp_ports,cls.cq_quality, cls.cq_accuracy,
                cls.steps_removed, cls.gm_cq_class, cls.gm_cq_accuracy, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(PtpClockInfo, self).save_entry()

class PtpCorrection(_RedisModel):
    log_timestamp = objects.Float(json_key='logTimestamp')
    offset_from_master = objects.Long(json_key='offset_from_master')
    mean_path_delay = objects.Long(json_key='mean_path_delay')
    app_name = 'ptp'
    container_hostname = 'netq-app-ptp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ['hostname']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'path']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.log_timestamp, cls.offset_from_master, cls.mean_path_delay,
                cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(PtpCorrection, self).save_grpc()


class PtpPortStatus(_RedisModel):
    ifname = objects.String(json_key='interface')
    port_status = objects.String(json_key='portStatus')
    app_name = 'ptp'
    container_hostname = 'netq-app-ptp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ['hostname', 'ifname']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'path']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ifname, cls.port_status, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(PtpPortStatus, self).save_entry()

class PtpPortStatistics(_RedisModel):
    ifname = objects.String(json_key='interface')
    rx_announce = objects.Long(json_key='rx_announce')
    rx_delay_req = objects.Long(json_key='rx_delay_req')
    rx_delay_resp = objects.Long(json_key='rx_delay_resp')
    rx_follow_up = objects.Long(json_key='rx_follow_up')
    rx_management = objects.Long(json_key='rx_management')
    rx_peer_delay_req = objects.Long(json_key='rx_peer_delay_req')
    rx_peer_delay_resp = objects.Long(json_key='rx_peer_delay_resp')
    rx_signaling = objects.Long(json_key='rx_signaling')
    rx_sync = objects.Long(json_key='rx_sync')
    tx_announce = objects.Long(json_key='tx_announce')
    tx_delay_req = objects.Long(json_key='tx_delay_req')
    tx_delay_resp = objects.Long(json_key='tx_delay_resp')
    tx_follow_up = objects.Long(json_key='tx_follow_up')
    tx_management = objects.Long(json_key='tx_management')
    tx_peer_delay_req = objects.Long(json_key='tx_peer_delay_req')
    tx_peer_delay_resp = objects.Long(json_key='tx_peer_delay_resp')
    tx_signaling = objects.Long(json_key='tx_signaling')
    tx_sync = objects.Long(json_key='tx_sync')
    app_name = 'ptp'
    container_hostname = 'netq-app-ptp'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ['hostname', 'ifname']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'ifname', 'path']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def exclude_fmt(cls):
        return []

    @classmethod
    def val_fmt(cls):
        return (cls.message_type,
                cls.ifname,
                cls.rx_announce, cls.rx_delay_req, cls.rx_delay_resp, cls.rx_follow_up,
                cls.rx_management, cls.rx_peer_delay_req, cls.rx_peer_delay_resp, cls.rx_signaling, cls.rx_sync,
                cls.tx_announce, cls.tx_delay_req, cls.tx_delay_resp, cls.tx_follow_up,
                cls.tx_management, cls.tx_peer_delay_req, cls.tx_peer_delay_resp, cls.tx_signaling, cls.tx_sync,
                cls.timestamp,
                cls.active,
                cls.deleted)

    def save(self):
        super(PtpPortStatistics, self).save_entry()

class AsicMonitorHistogram(_RedisModel):
    ifname = objects.String(json_key='ifname')
    bin0 = objects.Long(json_key='bin0')
    bin1 = objects.Long(json_key='bin1')
    bin2 = objects.Long(json_key='bin2')
    bin3 = objects.Long(json_key='bin3')
    bin4 = objects.Long(json_key='bin4')
    bin5 = objects.Long(json_key='bin5')
    bin6 = objects.Long(json_key='bin6')
    bin7 = objects.Long(json_key='bin7')
    bin8 = objects.Long(json_key='bin8')
    bin9 = objects.Long(json_key='bin9')
    maximum_queue_bytes = objects.Long(json_key='maximum_queue_bytes')
    minimum_queue_bytes = objects.Long(json_key='minimum_queue_bytes')
    peer_hostname = objects.String(json_key='peer_hostname')
    queue_bytes_range = objects.Long(json_key='queue_bytes_range')
    sampling_time = objects.Long(json_key='sampling_time')
    start_timestamp = objects.Long(json_key='timestamp')
    end_timestamp = objects.Long(json_key='timestamp')

    app_name = 'queuehistogram'
    container_hostname = 'netq-app-asicmonitoring'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'interface']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.ifname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.ifname, cls.timestamp, cls.start_timestamp, cls.end_timestamp,
                cls.maximum_queue_bytes, cls.minimum_queue_bytes, cls.peer_hostname, cls.queue_bytes_range,
                cls.sampling_time, cls.bin0, cls.bin1, cls.bin2, cls.bin3, cls.bin4, cls.bin5, cls.bin6, cls.bin7,
                cls.bin8, cls.bin9)

    def save(self):
        super(AsicMonitorHistogram, self).save_memmodel()

class AsicMonitorBufferStats(_RedisModel):
    buffer_stats = objects.Json(json_key='buffer_stats')

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'buffer_stats']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname, cls.buffer_stats

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.buffer_stats, cls.timestamp)

    def save(self):
        super(AsicMonitorBufferStats, self).save_grpc()

class CustomerPreference(_RedisModel):
    app_name = 'customer'
    container_hostname = 'netq-app-customermgmt'
    put_request_args = ['id', 'is_sso_enabled', 'preferences']
    base_url = 'https://{}:{}/netq/customer/v1/preferences'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return ['hostname']

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['path', 'hostname']

    @classmethod
    def key_fmt(cls, key_type=None):
        return cls.__name__, cls.hostname

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.message_type, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(CustomerPreference, self).save_entry()

class ServiceStatistics(_RedisModel):
    name = objects.String(json_key='service')
    vrf = objects.String(json_key='vrf')
    cpu_percent = objects.Float(json_key='cpu_percent')
    memory_percent = objects.Float(json_key='memory_percent')
    app_name = 'process'
    container_hostname = 'netq-app-services'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'name', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return (cls.cpu_percent, cls.memory_percent, cls.timestamp, cls.active, cls.deleted)

    def save(self):
        super(ServiceStatistics, self).save_grpc()

class ServiceStatisticsAggrgate(_RedisModel):
    name = objects.String(json_key='process_name')
    is_enabled = objects.Boolean(json_key='enabled')
    cpu_five_min = objects.String(json_key='cpu_five_min')
    cpu_one_min = objects.String(json_key='cpu_one_min')
    memory_one_min = objects.String(json_key='memory_one_min')
    memory_five_min = objects.String(json_key='memory_five_min')
    vrf = objects.String(json_key='vrf')
    up_time = objects.Float(json_key='up_time')
    pid = objects.String(json_key='pid')
    last_changed = objects.Float(json_key='last_changed')
    app_name = 'process'
    container_hostname = 'netq-app-services'

    @classmethod
    def optional_args(cls, key=None, val=None):
        if key:
            return cls.optional_args_map.get(key, key), val
        else:
            return []

    @classmethod
    def url_fmt(cls, key=None):
        if key:
            return cls.api_key_map.get(key, key)
        else:
            return ['hostname', 'name', 'vrf']

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.name, cls.vrf)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.last_changed

    @classmethod
    def val_fmt(cls):
        return (cls.cpu_one_min, cls.cpu_five_min, cls.memory_one_min, cls.memory_five_min)


