#!/usr/bin/env python
#
# Copyright (c) 2020-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# LicenseRef-NvidiaProprietary
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited
#

"""
Copyright (C) 2016, 2017, 2018 Cumulus Networks, Inc. All rights reserved.

This software is subject to the Cumulus Networks End User License Agreement available
at the following locations:

Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
Cumulus Linux systems: /usr/share/cumulus/EULA.txt
"""

import logging
import os
import re
import sys
import subprocess
import unittest
from network_docopt.match import (
    custom_matchers,
    dscp_mapping,
    is_loose_matching_variable,
    re_hex_range_variable,
    re_number_range_variable,
    tcp_flags,
    tcp_port_mapping,
    udp_port_mapping,
)
from network_docopt.node import Node, tree_nodes
from pprint import pformat

log = logging.getLogger(__name__)


re_help_key_value = re.compile(r"""^(.*?):(.*)$""")


def sort_for_humans(list_foo):
    """
    Sort list_foo in the way that humans expect.
    http://nedbatchelder.com/blog/200712/human_sorting.html

    >>> sort_for_humans(['swp10', 'swp2', 'swp1', 'swp20', 'swp0', 'swp11'])
    ['swp0', 'swp1', 'swp2', 'swp10', 'swp11', 'swp20']
    """

    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)]
    list_foo.sort(key=alphanum_key)
    return list_foo


def sort_ifaces(ifnames):
    """
    Given a list of interface names, return a list of those names sorted
    in the following order:
        lo -> eth -> swp -> bonds, bridges, etc

    >>> sort_ifaces(['bonds', 'bridges', 'swp10', 'eth0', 'swp1', 'swp20', 'swp2', 'eth1', 'lo'])
    ['lo', 'eth0', 'eth1', 'swp1', 'swp2', 'swp10', 'swp20', 'bonds', 'bridges']
    """
    sorted_variables = []
    sorted_interfaces_lo = []
    sorted_interfaces_eth = []
    sorted_interfaces_swp = []
    sorted_interfaces_nonswp = []

    for x in sort_for_humans(ifnames):
        if x.startswith('<'):
            sorted_variables.append(x)

        elif x.startswith('lo'):
            sorted_interfaces_lo.append(x)

        elif x.startswith('eth'):
            sorted_interfaces_eth.append(x)

        elif re.match(r'^sw(\d)*p', x):
            sorted_interfaces_swp.append(x)

        else:
            sorted_interfaces_nonswp.append(x)

    return (sorted_variables +
            sorted_interfaces_lo +
            sorted_interfaces_eth +
            sorted_interfaces_swp +
            sorted_interfaces_nonswp)


def get_bash_prompt():
    """
    This takes 40-50ms, we have not found a faster way to get the user's prompt
    """
    try:
        subprocess.check_output("/bin/echo /bin/false | /bin/bash -i",
                                stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        for line in e.output.splitlines():
            line = line.decode(encoding='UTF-8').strip()
            line = line.replace('/bin/false', '')

            if not line:
                continue

            if 'stty: standard input: Inappropriate ioctl for device' in line:
                continue

            return line
    return None


def merge_dict(x, y):
    """
    Merge dictionary y into dictionary x.  If the same key is in both the
    value from y is used.

    >>> merge_dict({'mario' : 'plumber', 'luigi' : 'marios brother'}, {'zelda' : 'princess', 'mario' : 'luigis brother'})
    {'mario': 'luigis brother', 'zelda': 'princess', 'luigi': 'marios brother'}

    >>> merge_dict({'mario' : 'plumber', 'luigi' : 'marios brother'}, {})
    {'mario': 'plumber', 'luigi': 'marios brother'}

    >>> merge_dict({}, {'zelda' : 'princess', 'mario' : 'luigis brother'})
    {'mario': 'luigis brother', 'zelda': 'princess'}
    """
    assert isinstance(x, dict), "x '%s' is %s, it must be a dict" % (pformat(x), type(x))
    assert isinstance(y, dict), "y '%s' is %s, it must be a dict" % (pformat(y), type(y))

    if not x:
        return y

    if not y:
        return x

    for (key, value) in list(y.items()):
        x[key] = value

    return x


def tree_tab_complete_leaf_nodes(node, argv, argv_index):
    """
    Return a list of the Node objects at the end of each branch that is
    either fully or partially matched by argv. This is used for TAB completion.
    """
    leaf_nodes = []

    # 'node' matches the word at argv[argv_index]
    if node.match(argv[argv_index], use_fuzzy=False, keywords_only=False, for_tab_complete=True):

        # Is this the last word in argv?
        if argv_index == len(argv) - 1:
            leaf_nodes.append(node)
        else:
            for child_node in list(node.children.values()):
                leaf_nodes.extend(tree_tab_complete_leaf_nodes(child_node, argv, argv_index + 1))

    # 'node' does not match the word at argv[argv_index]. If the Node is not
    # required then we keep exploring. If the Node is required then we stop
    # exploring and return an empty leaf_nodes list.
    elif not node.required:

        for child_node in list(node.children.values()):
            leaf_nodes.extend(tree_tab_complete_leaf_nodes(child_node, argv, argv_index))

    return leaf_nodes


def tree_tab_words(root_node, argv, args, ended_with_space):
    """
    Return a list of words for TAB completion
    """
    ENTER_ok = False

    # Find all of the branches that are either a full or partial match for argv
    tab_complete_leaf_nodes = tree_tab_complete_leaf_nodes(root_node, argv, 0)

    # If there were no matching branches go ahead and return
    if not tab_complete_leaf_nodes:
        return ((), args, ENTER_ok, [])

    if ended_with_space:
        '''
        There are times when argv can match multiple branches in the tree. For example
        say you have a program named foo with the following __doc__:

        Usage:
            foo show ip neighbor bart
            foo show ip neighbor homer
            foo show ipv6 neighbor marge
            foo show ipv6 neighbor lisa

        Which translates to the following parse tree

        foo
         +-- show
             +-- ip
             |   +-- neighbor
             |       +-- bart
             |       +-- homer
             +-- ipv6
                 +-- neighbor
                     +-- lisa
                     +-- marge

        If the user enters "foo show ip neighbor <TAB>" that will partial match
        the following branches

        foo
         +-- show
             +-- ip
             |    +-- neighbor
             +-- ipv6
                  +-- neighbor

        We would return bart, homer, lisa and marge as the TAB complete options.
        That would be confusing for the user because they expect to only see bart
        and homer since they specified "ip".  The problem is that "ip" is a subset
        of "ipv6" so the "show ipv6 neighbor" branches of the tree are also matches.

        The fix is to only return the TAB options for the least ambiguous branches.
        In our simpsons scenario "foo show ip neighbor" will have an ambiguity score
        of 2 ('show' vs 'sh' has a score of 2 and that is the only ambiguous node in
        the branch) where "foo show ipv6 neighbor" will have an ambiguity score of 4
        (2 for 'show' vs 'sh' and another 2 for 'ipv6' vs 'ip').

        Re-define 'tab_complete_leaf_nodes' to be the least ambiguous branches. Note
        that we only do this if ended_with_space is True. If we did this all the time
        then when the user did "foo show ip<TAB>" (no space after the "ip") we would
        only return "ip" instead of "ip" and "ipv6" which would be incorrect.
        '''
        min_ambiguity_score = None
        min_ambiguity_score_leaf_nodes = []

        for leaf_node in tab_complete_leaf_nodes:
            ambiguity_score = leaf_node.command_ambiguity_score()

            if min_ambiguity_score is None or ambiguity_score < min_ambiguity_score:
                min_ambiguity_score = ambiguity_score
                min_ambiguity_score_leaf_nodes = [leaf_node]

            elif ambiguity_score == min_ambiguity_score:
                min_ambiguity_score_leaf_nodes.append(leaf_node)

        tab_complete_leaf_nodes = min_ambiguity_score_leaf_nodes

    tab_words = []
    args = {}
    last_argv_word = argv[-1]

    for leaf_node in tab_complete_leaf_nodes:

        if ended_with_space:

            # Say leaf_node has words (mario|martian|maverick)
            #
            # - If argv ends with "m <TAB>" we must return () as the match is ambiguous
            # - If argv ended with "mar <TAB>" we must return () as the match is ambiguous
            # - If argv ended with "mav <TAB>" we must return leaf_node.tab_words
            if leaf_node.ambiguous_match(last_argv_word):
                return ((), args, ENTER_ok, [])

            else:
                tab_words.extend(leaf_node.tab_words)

                # If the leaf node is a <wildcard> we want to append <wildcard> as an
                # option so that the custom tab_completer for <wildcard> fires.
                #for word in leaf_node.words:
                #    if word.startswith('<wildcard'):
                #        tab_words.append(word)

                if leaf_node.ENTER_ok:
                    ENTER_ok = True

        else:

            for word in leaf_node.words:
                if not word.startswith('<') and word.startswith(last_argv_word):
                    tab_words.append(word)

                elif word in custom_matchers:
                    tab_words.append(word)

            # If the command is
            #
            #   nes show megaman <ipv4/prefixlen>
            #
            # and the user does
            #
            #   nes show megaman 1.1.1.1/32<TAB>
            #
            # We need to return '1.1.1.1/32` as a TAB choice in order for the
            # cursor to move over.  The next time the user hits TAB they will
            # see <ENTER> as an option.
            #
            # Am commenting this out for now, it causes issues in other scenarios.
            # If you have:
            #
            #    net (add|del) vrf <interface>
            #    net (add|del) vrf mgmt
            #
            # When the user does "net add vrf mg<TAB>" the if check below causes us
            # to return "mg" and "mgmt" as tab_words.  This breaks TAB complete, it
            # displays both of those options with their help text.

            #if not tab_words:
            #    tab_words.append(last_argv_word)

        # Even though we are only looking for tab complete options we still
        # populate args based on the options tree...we do this so any
        # tab_complete_XXX() hooks can be given args.  The hooks may need
        # to know what args matched in order to determine what to TAB complete.
        leaf_node.command_populate_args(args)

    # Remove duplicates
    tab_words = tuple(sorted(list(set(tab_words))))

    return (tab_words, args, ENTER_ok, tab_complete_leaf_nodes)


def format_helptext(text, width):
    """
    Take a really long string of help text and break it down into multiple
    lines all less than or equal to 'width'
    """
    lines = []
    line = []
    line_length = 0

    if not text:
        text = 'help text would go here'
        #return lines

    for word in text.split():

        # Too long
        if (line_length + len(word)) > width:
            lines.append(' '.join(line))
            line = []
            line_length = 0

        line.append(word)
        line_length += len(word) + 1  # +1 for the whitespace after the word

    if line:
        lines.append(' '.join(line))

    return lines


def tab_words_expand(tab_words, ended_with_space, last_argv, tab_complete_hook={}, args={}):
    """
    At this point we have all of the words to print as options. There
    are some of these we want to expand on though. <interfaces> for
    instance, we want users to be able to TAB complete the interface names.

    >>> tab_words_expand(('mario', 'zelda'), True, 'zel')
    (['mario', 'zelda'], {})

    >>> tab_words_expand(('mario', 'zelda'), False, 'zel')
    (['mario', 'zelda'], {})

    >>> tab_words_expand(('ip', 'ipv6'), False, 'ip')
    (['ip', 'ipv6'], {})

    # ended_with_space is True so expand all options
    >>> tab_words_expand(('<tcp-flags>',), True, 'zel')
    (['<text>', 'ACK', 'ALL', 'FIN', 'NONE', 'PSH', 'RST', 'SYN', 'URG'], {'ALL': 'All TCP flags', 'URG': 'Urgent pointer field is significant', 'ACK': 'Acknowledgment bit', 'PSH': 'Push function. Asks to push the buffered data to the receiving application', '<text>': 'Comma separated list of TCP flags', 'NONE': 'No TCP flags', 'SYN': 'Synchronize sequence numbers', 'RST': 'Reset the connection', 'FIN': 'No more data from sender'})

    # ended_with_space is False, only expand the options that start with last_argv ('zel')
    >>> tab_words_expand(('<tcp-flags>',), False, 'zel')
    ([], {})

    # ended_with_space is False, only expand the options that start with last_argv ('A')
    >>> tab_words_expand(('<tcp-flags>',), False, 'A')
    (['ACK', 'ALL'], {'ACK': 'Acknowledgment bit', 'ALL': 'All TCP flags'})

    # ended_with_space is False, only expand the options that start with last_argv ('S')
    >>> tab_words_expand(('<tcp-flags>',), False, 'S')
    (['SYN'], {'SYN': 'Synchronize sequence numbers'})

    # This is case sensitive
    >>> tab_words_expand(('<tcp-flags>',), False, 's')
    ([], {})

    >>> tab_words_expand(('',), False, 'zel')
    ([''], {})

    >>> tab_words_expand((), False, 'zel')
    ([], {})
    """
    assert isinstance(tab_words, tuple), "type is %s" % type(tab_words)
    assert isinstance(ended_with_space, bool), "type is %s" % type(ended_with_space)
    assert isinstance(last_argv, str), "type is %s" % type(last_argv)
    assert isinstance(tab_complete_hook, dict), "type is %s" % type(tab_complete_hook)

    final_words = []
    helptext = {}

    for word_text in tab_words:

        # Only variables can be expanded so if this is a keyword go ahead
        # and continue to the next word
        if not word_text.startswith('<'):
            final_words.append(word_text)
            continue

        if word_text in tab_complete_hook:
            completer = tab_complete_hook[word_text]

            for (keyword, keyword_help) in completer(args, ended_with_space, last_argv):
                final_words.append(keyword)

                if keyword_help:
                    helptext[keyword] = keyword_help

        elif word_text.startswith('<interface'):
            interfaces = sort_for_humans([x for x in os.listdir('/sys/class/net/') if x != 'bonding_masters'])

            if ended_with_space:
                final_words.extend(interfaces)
            else:
                for ifname in interfaces:
                    if ifname.startswith(last_argv):
                        final_words.append(ifname)

            final_words.append(word_text)

        elif word_text.startswith('<dscp'):

            if ended_with_space:
                for (dscp_option, dscp_helptext) in list(dscp_mapping.items()):
                    final_words.append(dscp_option)
                    helptext[dscp_option] = dscp_helptext
            else:
                for (dscp_option, dscp_helptext) in list(dscp_mapping.items()):
                    if dscp_option.startswith(last_argv):
                        final_words.append(dscp_option)
                        helptext[dscp_option] = dscp_helptext

        elif word_text.startswith('<tcp-flags'):

            if ended_with_space:
                helptext['<text>'] = 'Comma separated list of TCP flags'
                final_words.append('<text>')

                for (tcp_flag_option, tcp_flag_helptext) in list(tcp_flags.items()):
                    final_words.append(tcp_flag_option)
                    helptext[tcp_flag_option] = tcp_flag_helptext
            else:
                for (tcp_flag_option, tcp_flag_helptext) in list(tcp_flags.items()):
                    if tcp_flag_option.startswith(last_argv):
                        final_words.append(tcp_flag_option)
                        helptext[tcp_flag_option] = tcp_flag_helptext

        elif word_text.startswith('<tcp-port'):

            if ended_with_space:
                for (tcp_port_option, tcp_port_helptext) in list(tcp_port_mapping.items()):
                    final_words.append(tcp_port_option)
                    helptext[tcp_port_option] = tcp_port_helptext
            else:
                for (tcp_port_option, tcp_port_helptext) in list(tcp_port_mapping.items()):
                    if tcp_port_option.startswith(last_argv):
                        final_words.append(tcp_port_option)
                        helptext[tcp_port_option] = tcp_port_helptext

        elif word_text.startswith('<udp-port'):

            if ended_with_space:
                for (udp_port_option, udp_port_helptext) in list(udp_port_mapping.items()):
                    final_words.append(udp_port_option)
                    helptext[udp_port_option] = udp_port_helptext
            else:
                for (udp_port_option, udp_port_helptext) in list(udp_port_mapping.items()):
                    if udp_port_option.startswith(last_argv):
                        final_words.append(udp_port_option)
                        helptext[udp_port_option] = udp_port_helptext
        else:
            final_words.append(word_text)

    if '<interface>' in final_words:
        final_words = sort_ifaces(list(set(final_words)))
    else:
        final_words = sort_for_humans(list(set(final_words)))

    return (final_words, helptext)


def tab_words_without_helptext(tab_words, helptext):
    """
    >>> tab_words_without_helptext(['bowser', 'mario', 'luigi', 'zelda'], {'mario' : 'plumber', 'luigi' : 'brother of mario'})
    ['bowser', 'zelda']

    >>> tab_words_without_helptext(['<ipv4>', '<1-10>', '<0x0000-0xFFFF>', '<ipv6>'], {'<ipv4>' : 'IPv4 Address', '<ipv6>' : 'IPv6 Address'})
    []

    >>> tab_words_without_helptext(['bowser', 'mario', 'luigi', 'zelda'], {})
    ['bowser', 'mario', 'luigi', 'zelda']

    >>> tab_words_without_helptext([], {'mario' : 'plumber', 'luigi' : 'brother of mario'})
    []
    """
    result = []

    for word in tab_words:
        if word not in helptext:

            # 37, <1-10>, <0x0000-0xFFFF>, etc help text will be generated on the fly so we can ignore these
            if not (word.isdigit() or re_number_range_variable.match(word) or re_hex_range_variable.match(word)):
                result.append(word)

    return result


def tab_words_scrub_loose_options(tab_words, ended_with_space, last_argv):
    """
    # This scenario
    prompt# net clear ip bgp pref --completions 4
        <interface>  :  An interface name
        prefix       :  IP prefix
        <ENTER>

    # needs to return just 'prefix' so that it can tab complete
    prompt# net clear ip bgp pref --completions 4
        prefix       :  IP prefix

    >>> tab_words_scrub_loose_options(['<interface>', 'prefix', '<ENTER>'], True, 'pref')
    ['<interface>', 'prefix', '<ENTER>']

    >>> tab_words_scrub_loose_options(['<interface>', 'prefix', '<ENTER>'], False, 'pref')
    ['prefix']
    """

    if ended_with_space:
        return tab_words

    all_vars_are_loose = True
    all_keywords_start_with_last_argv = True
    start_with_last_argv = []

    for word in tab_words:
        if word.startswith('<'):
            if word == '<ENTER>' or is_loose_matching_variable(word):
                continue
            else:
                all_vars_are_loose = False
                break
        else:
            if word.startswith(last_argv):
                start_with_last_argv.append(word)
            else:
                all_keywords_start_with_last_argv = False
                break

    if all_vars_are_loose and all_keywords_start_with_last_argv:
        return start_with_last_argv
    else:
        return tab_words


def tab_words_print_fish(tab_words, helptext, ENTER_ok):
    """
    For fish we always print options (with help text) to STDOUT
    """

    for word in tab_words:
        if word in helptext['global']:
            print(("%s\t%s" % (word, helptext[word])))
        else:
            print(word)

    if ENTER_ok:
        print('<ENTER>')


def tab_words_print_bash(argv, last_argv, final_words, ended_with_space,
                         helptext, daemon, ENTER_ok,
                         leaf_node_commands):
    """
    For bash we do one of two things:
    - print options without help text to STDOUT
    - print options with help text to STDOUT but flag with USE_STDERR
    """
    use_stdout = False

    if last_argv:
        all_start_with_last_argv = True

        for word in final_words:
            if not word.startswith(last_argv):
                all_start_with_last_argv = False
                break

        if all_start_with_last_argv:
            '''
            Now check to see if this would actually make us tab complete more
            If the user does "mst<TAB>" when our options are 'mstpctl-bpduguard'
            and 'mstpctl-portautoedge' then we want to print these to stdout.
            But if the user does "mstpctl-<TAB>" we do not want to print these because
            we have already TABed out as far as we can go so it would be better to
            print the help text.
            '''
            first_characters = []
            for word in final_words:
                tmp_word = word.replace(last_argv, '')
                if tmp_word:
                    first_characters.append(tmp_word[0])
                else:
                    first_characters.append(None)

            first_characters = list(set(first_characters))

            if len(first_characters) == 1:
                use_stdout = True

    if use_stdout:
        for word in final_words:
            print(word)

    else:
        output = []

        # If we are running as a daemon the daemon should catch everything we print
        # so write to stdout but use the USE_STDERR output to signal to the daemon
        # that it should print this to stderr.
        if daemon:
            output.append("USE_STDERR")
        else:
            output.append('')

        max_word_len = 0
        for word in final_words:
            if len(word) > max_word_len:
                max_word_len = len(word)

        # 4 for the leading whitespace and 5 for the '  :  '
        width = 4 + max_word_len + 5
        max_width = 120

        command_specific_helptext = {}

        if 'per-command' in helptext:
            for command in leaf_node_commands:
                command_specific_helptext = merge_dict(command_specific_helptext, helptext['per-command'].get(command, {}))

        for word in final_words:

            # See if there is command specific helptext for this word
            word_helptext = command_specific_helptext.get(word)

            if word_helptext is None:
                word_helptext = helptext['global'].get(word)

            if word_helptext is not None:
                for (index, line) in enumerate(format_helptext(word_helptext, max_width - width)):
                    if index == 0:
                        output.append("    %s  :  %s" % (word.ljust(max_word_len), line))
                    else:
                        output.append("    %s     %s" % (''.ljust(max_word_len), line))
            else:
                re_number_range = re.match('<(\d+)-(\d+)>', word)

                if re_number_range:
                    output.append("    %s  :  An integer from %s to %s" % (word.ljust(max_word_len), re_number_range.group(1), re_number_range.group(2)))

                elif word.isdigit():
                    output.append("    %s  :  The number %s" % (word.ljust(max_word_len), word))
                else:
                    output.append("    %s  :  add help text" % word.ljust(max_word_len))

        if ENTER_ok:
            output.append('    <ENTER>')

        if daemon:
            # If we are running as a daemon it will be up to the client to run
            # get_bash_prompt() so that we get the client's bash prompt instead
            # of the deamon's bash prompt.
            print(('\n'.join(output)))

        else:
            # If we are here then we are printing TAB complete options,
            # with help text, to STDERR
            output.append("\033[93mWARN: NetQ Daemon is not running\033[0m\n\n"+
                          "\033[90mPlease configure netq cli:\n"
                          "sudo netq config add cli (config arguments)\n\n"
                          "and, start netqd:\n"
                          "sudo netq config restart cli\033[0m")
            cmd = argv[0]
            argv = argv[1:]

            if argv:
                output.append('%s%s %s%s' % (get_bash_prompt(), cmd, ' '.join(argv), ' ' if ended_with_space else ''))
            else:
                output.append("%s%s " % (get_bash_prompt(), cmd))

            sys.stderr.write('\n'.join(output))


class TestTabComplete(unittest.TestCase):

    # ./foobar show bgp 1.1.1.1/32<TAB>

    def setUp(self):
        self.root = Node(0, None, "nes")

        commands = """
            nes show [dodongo|zelda] [json]
            nes show (mario|luigi) json
            nes add (mario|martian|maverick) mustache
            nes add zelda sword
            nes clear character (mario|luigi)
            nes clearing character (zelda|link)
            nes del (mario hat|mario mustache|zelda sword) json
            nes load mario [mustache trimmer]
            nes show megaman <ipv4/prefixlen>
            nes show samus [detail] [json]
        """

        for command in commands.splitlines():
            self.root.add_branch(command)

        # Build a tuple of nodes and leaf nodes
        self.nodes = tree_nodes(self.root)
        self.leaf_nodes = tuple([node for node in self.nodes if node.leaf_node()])

        for node in self.nodes:
            node.populate_tab_words()

        for leaf_node in self.leaf_nodes:
            leaf_node.populate_node_ENTER_ok()

    def test_root_options(self):

        # nes<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes'], {}, False)
        self.assertEqual(tab_words, ('nes',))

        # nes <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes'], {}, True)
        self.assertEqual(tab_words, ('add', 'clear', 'clearing', 'del', 'load', 'show'))

    def test_no_match(self):
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'foo', 'bar'], {}, True)
        self.assertEqual(tab_words, ())

        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'foo', 'bar'], {}, False)
        self.assertEqual(tab_words, ())

    def test_keyword_completion(self):

        # nes ad<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'ad'], {}, False)
        self.assertEqual(tab_words, ('add',))

        # nes add<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add'], {}, False)
        self.assertEqual(tab_words, ('add',))

        # nes ad <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'ad'], {}, True)
        self.assertEqual(tab_words, ('mario', 'martian', 'maverick', 'zelda'))

        # nes add <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add'], {}, True)
        self.assertEqual(tab_words, ('mario', 'martian', 'maverick', 'zelda'))

        # nes add m<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add', 'm'], {}, False)
        self.assertEqual(tab_words, ('mario', 'martian', 'maverick'))

        # nes add mar<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add', 'mar'], {}, False)
        self.assertEqual(tab_words, ('mario', 'martian'))

        # nes add mav<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add', 'mav'], {}, False)
        self.assertEqual(tab_words, ('maverick', ))

        # nes add m <TAB>
        # NO MATCH
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add', 'm'], {}, True)
        self.assertEqual(tab_words, ())

        # nes add mar <TAB>
        # NO MATCH
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add', 'mar'], {}, True)
        self.assertEqual(tab_words, ())

        # nes add mav <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'add', 'mav'], {}, True)
        self.assertEqual(tab_words, ('mustache', ))

    def test_ENTER_option(self):
        """
        ENTER_ok must be True when the user can press <ENTER>

        nes show [dodongo|zelda] [json]
        nes show (mario|luigi) json
        nes show megaman <ipv4/prefixlen>
        nes show samus [detail] [json]
        """
        show = self.root.find_node('show')
        self.assertTrue(show.ENTER_ok)

        # nes sh<TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'sh'], {}, False)
        self.assertEqual(tab_words, ('show',))
        self.assertFalse(ENTER_ok)

        # nes show<TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show'], {}, False)
        self.assertEqual(tab_words, ('show',))
        self.assertFalse(ENTER_ok)

        # nes sh <TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'sh'], {}, True)
        self.assertEqual(tab_words, ('dodongo', 'json', 'luigi', 'mario', 'megaman', 'samus', 'zelda'))
        self.assertTrue(ENTER_ok)

        # nes show <TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show'], {}, True)
        self.assertEqual(tab_words, ('dodongo', 'json', 'luigi', 'mario', 'megaman', 'samus', 'zelda'))
        self.assertTrue(ENTER_ok)

        # nes show mario json<TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'mario', 'json'], {}, False)
        self.assertEqual(tab_words, ('json',))
        self.assertFalse(ENTER_ok)

        # nes show mario json <TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'mario', 'json'], {}, True)
        self.assertEqual(tab_words, ())
        self.assertTrue(ENTER_ok)

        # nes show megaman 1.1.1.1/32<TAB>
        #(tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'megaman', '1.1.1.1/32'], {}, False)
        #self.assertEquals(tab_words, ('1.1.1.1/32',))
        #self.assertFalse(ENTER_ok)

        # nes show megaman 1.1.1.1/32 <TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'megaman', '1.1.1.1/32'], {}, True)
        self.assertEqual(tab_words, ())
        self.assertTrue(ENTER_ok)

        # nes show samus
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'samus'], {}, True)
        self.assertEqual(tab_words, ('detail', 'json'))
        self.assertTrue(ENTER_ok)

        # nes show samus detail
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'samus', 'detail'], {}, True)
        self.assertEqual(tab_words, ('json',))
        self.assertTrue(ENTER_ok)

        # nes show samus detail json
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'samus', 'detail', 'json'], {}, True)
        self.assertEqual(tab_words, ())
        self.assertTrue(ENTER_ok)

        # nes show samus json
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'show', 'samus', 'json'], {}, True)
        self.assertEqual(tab_words, ())
        self.assertTrue(ENTER_ok)

    def test_least_ambiguous_branch(self):
        """
        nes clear character (mario|luigi)
        nes clearing character (zelda|link)
        """

        # nes cle character <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'cle', 'character'], {}, True)
        self.assertEqual(tab_words, ('luigi', 'mario'))

        # nes clear character <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'clear', 'character'], {}, True)
        self.assertEqual(tab_words, ('luigi', 'mario'))

        # nes clear character<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'clear', 'character'], {}, False)
        self.assertEqual(tab_words, ('character',))

        # nes cleari character <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'cleari', 'character'], {}, True)
        self.assertEqual(tab_words, ('link', 'zelda'))

        # nes clearing character <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'clearing', 'character'], {}, True)
        self.assertEqual(tab_words, ('link', 'zelda'))

    def test_multiword_node(self):
        """
        nes del (mario hat|mario mustache|zelda sword) json
        """

        # nes del<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del'], {}, False)
        self.assertEqual(tab_words, ('del',))

        # nes del <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del'], {}, True)
        self.assertEqual(tab_words, ('mario', 'zelda'))

        # nes del m<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'm'], {}, False)
        self.assertEqual(tab_words, ('mario',))

        # nes del m <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'm'], {}, True)
        self.assertEqual(tab_words, ('hat', 'mustache'))

        # nes del mario<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'mario'], {}, False)
        self.assertEqual(tab_words, ('mario',))

        # nes del mario <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'mario'], {}, True)
        self.assertEqual(tab_words, ('hat', 'mustache'))

        # nes del z<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'z'], {}, False)
        self.assertEqual(tab_words, ('zelda',))

        # nes del zelda<TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'zelda'], {}, False)
        self.assertEqual(tab_words, ('zelda',))

        # nes del zelda <TAB>
        (tab_words, _, _, _) = tree_tab_words(self.root, ['nes', 'del', 'zelda'], {}, True)
        self.assertEqual(tab_words, ('sword',))

    def test_optional_multiword_node(self):
        """
        nes load mario [mustache trimmer]

        Nodes that contain a phrase with a space are a beast to TAB complete.
        To make this a little easier we expand those commands out into multiple
        commands.  Example:
        - nes load mario
        - nes load mario mustache trimmer
        """
        # nes load mario<TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'load', 'mario'], {}, False)
        self.assertEqual(tab_words, ('mario',))
        self.assertFalse(ENTER_ok)

        # nes load mario <TAB>
        (tab_words, _, ENTER_ok, _) = tree_tab_words(self.root, ['nes', 'load', 'mario'], {}, True)
        self.assertEqual(tab_words, ('mustache',))
        self.assertTrue(ENTER_ok)


if __name__ == '__main__':
    import doctest

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(filename)16s %(levelname)8s: %(message)s')
    log = logging.getLogger(__name__)

    # Color the errors and warnings in red
    logging.addLevelName(logging.ERROR, "\033[91m   %s\033[0m" % logging.getLevelName(logging.ERROR))
    logging.addLevelName(logging.WARNING, "\033[91m %s\033[0m" % logging.getLevelName(logging.WARNING))

    doctest.testmod()

    test_suites_to_run = []
    test_suites_to_run.append(TestTabComplete)

    for test_suite in test_suites_to_run:
        suite = unittest.TestLoader().loadTestsFromTestCase(test_suite)
        unittest.TextTestRunner(verbosity=2).run(suite)
